/*
    DataSource.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824

    -------------------- Edit History (as of 11/22/04) ----------------------
    01/31/12 rdb: Added support for creating a SparseDataSource
    01/07/09 pjr: Added getStatistics() method (taken from DynamicChunking).
    08/25/08 kjr: Added support getting metadata through the FDLDescriptor.
                  Added a method to get the DataSource's FDLDescriptor.
    01/31/06 dlipsa: Added support for creating chunked data sources through
             an xfdl file
    11/28/05 rdb: create( String ): Major weakness: how can we tell whether file name
                    is an xfdl file or ddl file? (ddl was not previously supported
                    in this method.) For now just use extension convention: if it ends
                    in "ddl", invoke XDDLReader, else invoke XFDLReader.
                  deleted checkDataFileName;  now in FDLDescriptor.getDataFileURL
                  deleted a bunch of deprecated methods that are no longer used.
                  revised create( MRDataSource ) filename handling to remove '@'.
    10/27/05 rdb: replaced DataSource.createFDLDescriptor with FDLDescriptor.create
    03/29/05 rdb: Added support for MRDataSource in create methods
    03/14/05 rdb: Moved createFDLDescriptor code to FDLDescriptor.create
    02/16/05 rdb: Renamed createRemoteDataSources to createRemote with extra URL arg
                      and moved DSName to first argument
    02/11/05 rdb: Renamed createPhysicalDataSource to create.
                  Modified createFDLDescriptor, to check for existence of file
                     and if fails, look for arg.xfdl and arg.fdl and arg.xml
                  Removed FDLReader access; allowed ".fdl" to specify XFDL file
    02/07/05 rdb: Added constructor with name argument only, deleted default ctor.
    01/25/05 rdb: Removed abstract redefinitions of datum/subblock methods that
                    are already defined in DataCollection
                    as well as implementations that are now done in DataCollection,
                    such as DataBlock return version of subblock and Datum return
                    versions of datum.
                  Rewrote all the constructors to use super() and this() correctly.
    01/24/05 rdb: Added getValues()
                  Makes for consistency with DataCollection -- but it is very
                     expensive for large files.
    12/11/04 rdb: Renamed the "create" methods to "createBlock". These seem to be
                  package methods, so just changed them everywhere.
    12/05/04 rdb: Moved dimensionality, recordDesc, bounds, numAttributes
                  up to DataCollection. Also associated access methods.
    11/24/04 rdb:
            Added array versions of the get<type> methods.
    11/22/04 rdb:
            Added definitions and default implementations of
               getDouble, getFloat, getInt, getShort, getByte
*/

package edu.unh.sdb.datasource;

//import java.io.*;

import edu.unh.sdb.common.*;

import java.net.*;

/**
 * The DataSource class provides a provides a view of the data in which
 * each datum has a unique location within an N-dimensional index space.
 * Many types of DataSource can be connected to one or more component
 * DataSources to form a tree. The Root DataSource will communicate
 * directly with the Lattice, while the various leaf DataSources are
 * responsible for reading data from a file or network socket. The
 * various internal DataSources perform tasks like taking a subset,
 * Attribute Join or Spatial Join.
 */

 abstract public class DataSource extends DataCollection
 {
    /** The name of this DataSource object. */
    protected String name;

    protected String wsName;

    protected boolean isActivated=false;

    protected FDLDescriptor fdlDescriptor = null; // if DS created with FDL
                               // Not clear this should be here
                               // Probably should have a DSDescriptor that is parent
                               // to FDLDescriptor and that's what should be here.
                               // for now, though, we'll leave it this way.

    //----------------------- Constructors ------------------------------
    /** Default constructor.
     */
    protected DataSource( String dsname )
    {
       this.name = dsname;
    }

    /** Constructor sets the bounds and recordDesc members to
     *   the provided values.
     */
    protected DataSource( String aname, RecordDescriptor r, ISBounds db )
    {
       super( db, r );
       this.name = aname;
    }

    /** Constructor sets the bounds member to
     *   the provided value and the recordDesc member to
     *   a single float attribute.
     */
    protected DataSource( String aname, ISBounds db )
    {
       this( aname, new RecordDescriptor( "f" ), db );
    }

    /** Constructor sets all data members to copies of the
     *   corresponding members from the argument DataSource.
     */
    protected DataSource ( DataSource ds )
    {
       this( ds.name, ds.recordDesc, ds.bounds );
    }

    /** The DataSource will perform any initialization required
     *   in order to begin reading/processing data. It is
     *   important that all child classes call super.activate()
     */
    public  void activate()
    {
        this.isActivated = true;
    }


    /** The DataSource will release any resources used during the
	 *	readling or processing of data. It is important
     *  that all child classes call super.deactivate()
     */
    public void deactivate()
    {
        this.isActivated = false;
    }

    /** Returns true if the datasource is activated, false otherwise.
     */
	public final boolean isActivated(){

        return this.isActivated;
    }

    /** Returns the name of this DataSource object. */
    public String getName() {

         return this.name;
    }

    /** Return a Values array of all attributes in the entire DataSource in whatever
     *  type and order they are stored internally (either attribute order or point order
     *  for Granite-defined objects). This default implementation is really bad for
     *  large datasources -- it just creates a single DataBlock and does a
     *  getValues on that.
     */
    public Values getValues()
    {
       return this.subblock( this.bounds ).getValues();
    }

    /** 
     * Returns the MetadataCollection associated with this DataSource's FDLDescriptor.
	 * 
	 * @return returns the MetadataCollection associated with this DataSource's
	 *         FDLDescriptor.
	 */
	public MetadataCollection getMetadataCollection() {
		return fdlDescriptor.getMetadata();
	}

	/** 
     * Sets the MetadataCollection associated with this DataSource's FDLDescriptor.
	 */
	public void setMetadataCollection(MetadataCollection mc){
		fdlDescriptor.setMetadata(mc);
	}

	/**
	 * returns the FDLDescriptor associated with this DataSource.
	 * 
	 * @return the FDLDescriptor associated with this DataSource.
	 */
	public FDLDescriptor getFDLDescriptor() {
		return fdlDescriptor;
	}

	/**
	
    
    /** Make a copy of the argument using a chosen subclass
        of IndexSpaceID. The subclass is chosen according to
        a combination of time and space considerations. The
        chosen subclass will always be able to represent
        this DataSource's Index Space or Base Resolution.
    */
    public IndexSpaceID copyBestID(IndexSpaceID id){

        if(bounds==null) {
            // We have no idea how big the Index Space for
            // this DataSource will be, so we must use
            // the ID with largest possible Index Space.
            return new IntegerIndexSpaceID(id);

        } else {
            //This is where we decide what the best kind of
            // ID is. We might decide based on some mix of
            //    space and time considerations.
            //
            // Let's punt for now, so Feng can get on with
            // his code.

            return new IntegerIndexSpaceID(id);
        }
    }

    /** Write the values of the entire DataSource index space to a string in
        a nice format. Uses the DataSource axisOrdering and bounds.*/
    public String dumpData(){

        int i,d=this.dim();
        ISIterator iter=new ISIterator(this.bounds);
        String retVal=new String();


        for(iter.init();iter.valid();iter.next()){

            retVal+=""+this.datum(iter)+"\t";

            for(i=d-1;i>0;i--){

                if(iter.getCoord(i)==this.bounds.copyUpper().getCoord(i)){
                    retVal+="\n";
                } else {
                    break;
                }
            }
        }
        return retVal;
    }

    /** Return a String representing the object. */
    @Override
    public String toString(){

        return this.toString("");
    }

    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){

        String indent2=indent+Globals.toStringIndent;
        StringBuffer result=new StringBuffer();

        result.append(indent+"DataSource:\n");
        result.append(indent+"name="+name);
        result.append("\n");

        result.append(indent+"dimensionality=");
        result.append(dimensionality);
        result.append("\n");

        result.append(indent+"bounds=");
        result.append(bounds);
        result.append("\n");

        result.append(indent+"number of attributes=");
        result.append(numAttributes);
        result.append("\n");

        result.append(indent+"recordDescriptor=");
        result.append(recordDesc.toString(indent2));
        result.append("\n");

        return new String(result);
    }

    /** Returns true if this DataSource has the same bounds and
        RecordDescriptor as the argument DataSource.
        Currently, the DataSource name is not considered in the
        comparison.*/
    @Override
    public boolean equals(Object o)
    {
        DataSource d=(DataSource) o;

        return 	this.bounds.equals(d.bounds) &&
            this.recordDesc.equals(d.recordDesc);
    }

    /** Creates and returns a DataSource. The first argument is an arbitrary name
     *  for the datasource. The second argment should be the name of
	 *	an xfdl or fdl file. The child class of DataSource is determined by the
	 *  file description file.
	 *
	 * If the passed in file name does not exist, and that name does not end
	 * in one of the typical extensions (.xfdl, .fdl, .xml), this method will
	 * append those extensions in that order and check for the existence of
	 * the file.
	 *
	 * If the filename includes directory specifications, the root of the path
	 * will be saved. If the file name  specified in this FDL descriptor file,
	 * cannot be found, we will append the fdl file path root to the data file
	 * name and try to open that.
    */
    public static final DataSource create( String dsName, String xmlFile )
    {

       
       if ( xmlFile.endsWith( "ddl" ) )
       {
          XDDLReader ddlReader = new XDDLReader( xmlFile );
          DataSource[] allDS = ddlReader.getAllPublicDS();
          if ( allDS == null )
          {
             new SDBError( SDBError.Fatal, "DDL file has no specified 'root'; " +
             " cannot be used to create a DataSource.");
          }
          else if ( allDS.length == 0 )
          {
             new SDBError( SDBError.Fatal, "DDL file has no specified 'root'; " +
             " cannot be used to create a DataSource.");
          }
          else if ( allDS.length > 1 )
          {
             new SDBError( SDBError.Warning, "DDL file has " + allDS.length + " 'roots'; "
                    + " first one chosen; may not be correct.");
          }
          return allDS[ 0 ];
       }
       else
       {
          FDLDescriptor fdld = FDLDescriptor.create( xmlFile );
          return create( dsName, fdld);
       }
    }

    /** Creates and returns a RemoteBFDS where the fileServerURL argument identifies
     *  the location of the DataSource
     */
    public static final DataSource create( String dsName,
                                           String fdlfile,
                                           String fileServerURL )
    {
        FDLDescriptor fdld = FDLDescriptor.create( fdlfile );
        return createRemote( dsName, fdld, fileServerURL );
    }

    /** Creates and returns a RemoteBFDS where the fileServerURL argument identifies
     *  the location of the DataSource
     */
    public  static final DataSource createRemote( String dsName,
                                                  FDLDescriptor fdld,
                                                  String fileServerURL )
    {
        ISBounds dsBounds=fdld.getBounds();
        RecordDescriptor dsRecordDescriptor = fdld.getRecordDescriptor();
        DataSource ds=null;

        AxisOrdering dsOrdering=fdld.getAxisOrdering();

        String dsFilename=fdld.getDataFileURL();
       // int numAttributes=dsRecordDescriptor.size();

		ds = new RemoteBFDS( dsName,   dsRecordDescriptor, dsBounds,
		                     dsOrdering, dsFilename, fileServerURL);
		ds.fdlDescriptor = fdld;
       return ds;
    }
    /** Creates and returns a RemoteBFDS where the fileServerURL argument identifies
     *  the location of the DataSource
     */
    public static final DataSource createRemote( String dsName,
                                                 String fdlfile,
                                                 String fileServerURL )
    {
        FDLDescriptor fdld = FDLDescriptor.create( fdlfile );
        return createRemote( dsName, fdld, fileServerURL );
    }

    public  static final DataSource createRemoteUDP( String dsName,
                                                  FDLDescriptor fdld,
                                                  String fileServerURL )
    {
        ISBounds dsBounds=fdld.getBounds();
        RecordDescriptor dsRecordDescriptor = fdld.getRecordDescriptor();
        DataSource ds=null;

        AxisOrdering dsOrdering=fdld.getAxisOrdering();

        String dsFilename=fdld.getDataFileURL();

        //added by sridhar
        //System.out.println("***dsFilename*** " + dsFilename);

       // int numAttributes=dsRecordDescriptor.size();

		ds = new RemoteUDPClient( dsName,   dsRecordDescriptor, dsBounds,
		                     dsOrdering, dsFilename, fileServerURL);
		ds.fdlDescriptor = fdld;
       return ds;
    }


     public static final DataSource createRemoteUDP( String dsName,
                                                  String fdlfile,
                                                  String fileServerURL )
    {
        //System.out.println("Hello Sridhar I am here in createRemoteUDP.");
        FDLDescriptor fdld=FDLDescriptor.create( fdlfile );
        return createRemoteDataSourceUDP( fdld, dsName, fileServerURL );
    }


    public  static final DataSource createRemoteDataSourceUDP( FDLDescriptor fdld,
                                                            String dsName,
                                                            String fileServerURL )
    {
       return createRemoteUDP( dsName, fdld, fileServerURL );
    }


 //////////////////////////////////////////////////////////////////////
//    public  static final DataSource createRemoteUDT( String dsName,
//                                                  FDLDescriptor fdld,
//                                                  String fileServerURL )
//    {
//        ISBounds dsBounds=fdld.getBounds();
//        RecordDescriptor dsRecordDescriptor = fdld.getRecordDescriptor();
//        DataSource ds=null;
//
//        AxisOrdering dsOrdering=fdld.getAxisOrdering();
//
//        String dsFilename=fdld.getDataFileURL();
//
//        //added by sridhar
//        //System.out.println("***dsFilename*** " + dsFilename);
//
//       // int numAttributes=dsRecordDescriptor.size();
//
//		ds = new RemoteUDTClient( dsName,   dsRecordDescriptor, dsBounds,
//		                     dsOrdering, dsFilename, fileServerURL);
//		ds.fdlDescriptor = fdld;
//       return ds;
//    }
//
//
//     public static final DataSource createRemoteUDT( String dsName,
//                                                  String fdlfile,
//                                                  String fileServerURL )
//    {
//        //System.out.println("Hello Sridhar I am here in createRemoteUDP.");
//        FDLDescriptor fdld=FDLDescriptor.create( fdlfile );
//        return createRemoteDataSourceUDT( fdld, dsName, fileServerURL );
//    }
//
//
//    public  static final DataSource createRemoteDataSourceUDT( FDLDescriptor fdld,
//                                                            String dsName,
//                                                            String fileServerURL )
//    {
//       return createRemoteUDT( dsName, fdld, fileServerURL );
//    }

 //////////////////////////////////////////////////////////////////////
    public  static final DataSource createRemoteClient( String dsName,
                                                  FDLDescriptor fdld,
                                                  String fileServerURL )
    {
       return createRemoteDataSource(dsName, fdld, fileServerURL);
    }


     public static final DataSource createRemoteClient( String dsName,
                                                  String fdlfile,
                                                  String fileServerURL )
    {
        //System.out.println("Hello Sridhar I am here in createRemoteUDP.");
        FDLDescriptor fdld=FDLDescriptor.create( fdlfile );
        return createRemoteClient( dsName, fdld, fileServerURL );
    }

		/*###########				Interface that user should call to create remote datasource	if not called from create()				############*/
    public  static final DataSource createRemoteDataSource( String dsName,
                                                  String fdlfile,
                                                  String fileServerURL )
    {
       FDLDescriptor fdld=FDLDescriptor.create( fdlfile );
	     return createRemoteDataSource(dsName, fdld, fileServerURL);
    }

    public  static final DataSource createRemoteDataSource( String dsName,
                                                  FDLDescriptor fdld,
                                                  String fileServerURL )
    {
       ISBounds dsBounds=fdld.getBounds();
       RecordDescriptor dsRecordDescriptor = fdld.getRecordDescriptor();
       AxisOrdering dsOrdering=fdld.getAxisOrdering();
			 String dsURL = fdld.getDataFileURL();

       URI remoteSourceURI=null;
       try{
	       remoteSourceURI = new URI(fdld.getDataFileURL());
	     }
	     catch(NullPointerException npe){npe.printStackTrace();}
	     catch(URISyntaxException userror){userror.printStackTrace();}

       String scheme = remoteSourceURI.getScheme();
       DataSource ds = null;
       //using the old version of UDP client by Sridhar
       if(scheme.equalsIgnoreCase("sudp")){
	       	ds = createRemoteUDP( dsName, fdld, fileServerURL );
	     }
	     //rearchitectured new version of remote datasource with both UDP and UDT
       //by baoqiang
	     else if(scheme.equalsIgnoreCase("udt") ||
                     scheme.equalsIgnoreCase("tcp") ||
	             scheme.equalsIgnoreCase("udp"))
             {
                 //System.out.println("I am here in the createRemoteDataSource method of datasource.");
                 if(dsURL.toLowerCase().endsWith("_hdf5.xfdl"))
                 {
                       ds = new ChomboRemoteClient(dsName, dsRecordDescriptor, dsBounds,
                                                   dsOrdering, dsURL, "ChomboFoobar");
                 }
                 else
                 {
                       ds = new GraniteRemoteClient(dsName, dsRecordDescriptor, dsBounds,
                                                    dsOrdering, dsURL, "foobar");
                 }
	     }
       
       ds.fdlDescriptor = fdld;

       return ds;
    }
 //////////////////////////////////////////////////////////////////////

    /** Creates and returns a PhysicalDataSource, given an FDLDescriptor that has
	 *  already been initialized from an FDL file and an arbitrary name for
	 *  the datasource. The type of DataSource returned depends on the file
	 *  descriptor.
	 */
    public static final DataSource create( String dsName, FDLDescriptor fdld )
   {
      ISBounds dsBounds = fdld.getBounds();
      RecordDescriptor dsRecordDescriptor = fdld.getRecordDescriptor();
      DataSource ds = null;

      AxisOrdering dsOrdering = fdld.getAxisOrdering();

      String dsFilename = fdld.getDataFileString();

      if ( dsFilename.startsWith( "@") )
      {
         fdld.setDataFileURL( dsFilename.substring( 1 )); // delete '@'
         ds = new MRDataSource( dsName, fdld );
         return ds;
      }
      else if ( dsFilename.endsWith(".hdf5"))
      {
          ds = new ChomboDataSource(dsName, fdld);
          return ds;
      }


      // Handling remote datasources needs to be implemented
      String dsURL = fdld.getDataFileURL();



      /*################
      	Remote Cases
      ###############*/
      if(    dsURL.startsWith( "sudp") 
          || dsURL.startsWith( "udp://") 
          || dsURL.startsWith( "udt://") 
          || dsURL.startsWith( "tcp://"))
      {
			return createRemoteDataSource(dsName, fdld, dsURL);
	   }
/*		//details of creating remote data source are moved to createRemoteDataSource(dsName, fdld, dsURL)
      if( dsURL.startsWith( "sudp")){
      	 //using the old version of UDP client by Sridhar
         return createRemoteUDP(dsName, fdld, dsURL);
      }

			//remote client case
      if ( dsURL.startsWith( "udp://") || dsURL.startsWith( "udt://"))
      {
      		//rearchitectured new version of remote datasource with both UDP and UDT
      		//by baoqiang
		      return new GraniteRemoteClient(dsName, dsRecordDescriptor, dsBounds,
                                    dsOrdering, dsURL, "foobar") ;
      }
*/
      /*################
      	Local Cases
      ###############*/

      if ( fdld.getDataFileType().equals( "ascii" ) )
      {
         // We don't have an XTextFileDataSource yet
         ds = new TextFileDataSource( dsName, dsRecordDescriptor, dsBounds,
               dsOrdering, dsFilename );
      }
      else if ( fdld.getDataFileType().equals( "binary" ) )
      {
         int fieldLayout = fdld.getFieldLayout();
//         if ( fdld.isSparse() )
//         {
//        	 // abi - 04.26.2014
//        	 ds = new SparseDataSourceChunkExtraction ( dsName, fdld );
//          
//         }
         else if ( fdld.isChunkedFileFormat() )
         {
            // rdb 03/05/12: this could be replaced now with invocation of the
            //               FDLDescriptor constructor for ChunkedDataSource
            IndexSpaceID chunkDimensions = fdld.getChunkDimensions ();
            ISBounds partitioningBounds = new ISBounds(dsBounds.getDimensions ().divide (chunkDimensions));
            ds = new ChunkedDataSource( dsName,
                                        partitioningBounds,
                                        dsBounds,
                                        dsFilename,
                                        dsOrdering,
                                        dsRecordDescriptor );
         }
         else if ( fdld.isSparse() )
         {
            // For now, sparse data sources can only be in pointOrder
            // As with ChunkedDataSource, SparseDataSource probably should become
            //    a wrapper than go around any other datasource
           
        	 // abi - 04.26.2014
        	 ds = new SparseDataSourceChunkExtraction( dsName, fdld );
         }
         else
         {
             switch ( fieldLayout )
             {
                case DataCollection.pointOrder:
                   ds = new PointOrderBFDS( dsName,
                                            dsRecordDescriptor,
                                            dsBounds,
                                            dsOrdering,
                                            dsFilename,
                                            fdld.getByteOrder() );
                   break;
                case DataCollection.attributeOrder:
                   ds = new AttributeOrderBFDS( dsName,
                                                dsRecordDescriptor,
                                                dsBounds,
                                                dsOrdering,
                                                dsFilename,
                                                fdld.getByteOrder() );
                   break;
                default:
                   new SDBError( SDBError.Fatal, "Unknown file format" );
             }
             ((BinaryFileDataSource) ds).setFileOffset( fdld.getFileOffset() );
         }
      }
      else
         new SDBError( SDBError.Warning, "Unknown file type." );
      if ( ds != null )
         ds.fdlDescriptor = fdld;
      return ds;
    }
    /*************** Deprecated: use FDLDescriptor.create( fdlfile ) *******
    /**
     * create an FDLDescriptor object from a file with the name specified by the
     * argument.
     *
	 * If the passed in file name does not exist, and that name does not end
	 * in one of the typical extensions (.xfdl, .fdl, .xml), this method will
	 * append those extensions in that order and check for the existence of
	 * the file.
	 *
	 * If the filename includes directory specifications, the root of the path
	 * will be saved. If the file name  specified in this FDL descriptor file,
	 * cannot be found, we will append the fdl file path root to the data file
	 * name and try to open that.
     *
     * @param fdlfile -
     *           name of an fdl file from which to extract a DataSource
     *           description
     * @return The DataSource description as an FDLDescriptor object.
     *
     * deprecated Use FDLDescriptor.create( fdlfile )
     *
	public  static FDLDescriptor createFDLDescriptor( String fdlfile )
	{
	   FDLDescriptor fdld = null;
	   fdld = FDLDescriptor.create( fdlfile );

	   if ( fdld == null )
          new IllegalArgumentError( SDBError.Recoverable,
                "couldn't determine file type from extension. filename=" + fdlfile);
       return fdld;
	}
    /**************************************/
    /**************************************/

    /** Creates and returns a PhysicalDataSource. The first argument is an arbitrary
     *  name for the datasource. The second argment should be the name of
     *  an xfdl or fdl file. The child class of DataSource is determined by the
     *  file description file.
     * deprecated
     * Same as create( dsName, fdlfile ), i.e., parameters are reversed
    *
    public static final DataSource createPhysicalDataSource( String fdlfile,
                                                             String dsName)
    {
        return create( dsName, fdlfile );
    }
    /**************************************/

   /** Creates and returns a RemoteBFDS where the fileServerURL argument identifies
     *  the location of the DataSource
     * deprecated Use createRemote( dsName, fdlfile, fileServerURL )
     *
    public static final DataSource createRemoteDataSource( String fdlfile,
                                                           String dsName,
                                                           String fileServerURL )
    {
       return createRemote( dsName, fdlfile, fileServerURL );
    }
    /**************************************/

    /** Creates and returns a RemoteBFDS where the fileServerURL argument identifies
     *  the location of the DataSource
     * deprecated Use create( dsName, fdld, fileServerURL )
     *
    public  static final DataSource createRemoteDataSource( FDLDescriptor fdld,
                                                            String dsName,
                                                            String fileServerURL )
    {
       return createRemote( dsName, fdld, fileServerURL );
    }
    /**************************************/



   /** Generate a string indicating statistics for performance evaluation.
    */
   public String getStatistics ()
   {
      return "";
   }


    /** Attempts to clear the filesystem cache executing a shell script at $HOME/bin/trc.
     */
    public void clearCache()
    {
        Runtime r=Runtime.getRuntime();
        try
        {
           String trcpath = System.getProperty("user.home")+"/bin/trc";
           
           //System.out.println("Executing "+trcpath);
           
           (r.exec(trcpath)).waitFor();
        }
        catch(Exception e)
        {
            System.out.println(e);
        }
    }


	/**---------------------- main ----------------------------------
	// just a little code to test the code that tries to be smarter about
	// finding files.
     *
	 */

	public static void main( String[] args )
	{

	}


   /**
    * Check if the DataSource has been activated before trying to access data.
    * If it has not been activated, activate it.*
    */
   protected final void checkActivated ()
   {
      if (!isActivated())
      {
         new SDBError (SDBError.Warning, "Implicit activate() of " + this.name);
         activate();
      }
   }
}

