package ar;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.vecmath.Point3f;

import com.application.particles.flow.Streamlet;

import edu.unh.sdb.datasource.AxisOrdering;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.ISBoundsIterator;
import edu.unh.sdb.datasource.IntegerIndexSpaceID;
import edu.unh.sdb.datasource.RegularISPartitioning;
import edu.unh.sdb.datasource.SparseBlockInfo;
import edu.unh.sdb.datasource.SparseStorage;

public class ARSparseStorage { // ARSparseStorage

	private ArrayList<ARNode> blockList;
	private ARDataSource arDataSource;
	private DataSource[] dsArr; // instead have array of SparseStorage
	Map<String,Integer>[] regionMapper  = new HashMap[8] ; 

	private boolean tableCreated;

	/*
	 * We are defining this field with the sparse file xfdl. For each resolution
	 * file we need to create a sparseStorage class based on the respective
	 * resolution sparse xfdl file.
	 * 
	 * SPARSESTORAGE ( Important fields for getting sparse data ) sparseTable <
	 * Integer , SparseBlockInfo > : Hash table < sparseIndex, sparse block info
	 * > SPARSEBLOCKINFO -> sparse Index, start position of the block, seek
	 * address in sparse file, block size, original block index
	 */
	private SparseStorage[] sparseStorageArr;
	private int noOfRes;
	ARNode treeRoot;

	private Hashtable<Integer, ARStorageBlock> arStorageBlockTable = new Hashtable<Integer, ARStorageBlock>();
	private AxisOrdering axisOrdering;
	private RegularISPartitioning[] blockPartioning;

	public static ARStorage[] arStorage = new ARStorage[3];

	public ARSparseStorage(String[] arFilePrefixSuffixArr,
			String sparseFilePrefix) {

		// for building AR tree
		ARBuilder arBuilder = new ARBuilder();
		arBuilder.processData(arBuilder.generateMRHierarchy(arFilePrefixSuffixArr), 0.1f);
		arDataSource = arBuilder.getDataSource() ; 
		blockList = new ArrayList<ARNode>();
		axisOrdering = arBuilder.getAxisOrdering();
		noOfRes = arDataSource.getResolutions();
		blockPartioning = new RegularISPartitioning[noOfRes];
		treeRoot = arDataSource.getRoot() ;
		regionMapper = new RegionMapper().getRegionMapper() ; 

		createDataSourceArray();
		createPartioningArray(new IntegerIndexSpaceID(3, 3, 3));
		// commenting this for now, just to test the PArticles flow for particle current leaf update.
		createSparseStorageArray(sparseFilePrefix);

	}
	
	private void createDataSourceArray() {
		int nResolutions = arDataSource.getResolutions();
		String[] dsNames = arDataSource.getDataFDLNames();
		dsArr = new DataSource[nResolutions];
		// Note : Need to Replace with SparseStorage class .. ? 
		for (int m = 0; m < nResolutions; m++) {
			dsArr[m] = DataSource.create("dsResolutions_" + m, dsNames[m]);
			dsArr[m].activate();
		}

	}

	private void createPartioningArray(IntegerIndexSpaceID spaceId) {
		// each res DB will be the space bounds, not sure about the partition
		// space in each res.
		// Need to handle it later. For now setting only blockPartioning[of res
		// 1]
		blockPartioning[0] = new RegularISPartitioning(spaceId,
				dsArr[0].getBounds(), dsArr[0].getFDLDescriptor()
						.getAxisOrdering());

	}

	/**
	 * Creating the array of Sparse Storage for each resolution. For now we are
	 * creating only at res 0.
	 */
	private void createSparseStorageArray(String sparseFilePrefix) {
		sparseStorageArr = new SparseStorage[noOfRes];
	//	sparseStorageArr[0] = new SparseStorage(sparseFilePrefix, "SparseDS");

	}


	public AxisOrdering getAxisOrdering() {
		return this.axisOrdering;
	}

	public RegularISPartitioning getBlockPartioning(int axis) {
		return blockPartioning[axis];
	}

	public DataSource getDsArr(int res) {
		return dsArr[res];
	}

	public boolean isTableCreated() {
		return tableCreated;
	}

	public void setTableCreated(boolean tableCreated) {
		this.tableCreated = tableCreated;
	}

	public SparseStorage[] getSparseArr() {
		return sparseStorageArr;
	}

	public void setSparseArr(SparseStorage[] sparseArr) {
		this.sparseStorageArr = sparseArr;
	}

	public int getNoOfRes() {
		return noOfRes;
	}

	public void setNoOfRes(int noOfRes) {
		this.noOfRes = noOfRes;
	}

	public ARNode getTreeRoot() {
		return treeRoot;
	}

	public void setTreeRoot(ARNode treeRoot) {
		this.treeRoot = treeRoot;
	}


	public ArrayList<ARNode> getBlockList() {
		return blockList;
	}

	public ARDataSource getArDataSource() {
		return arDataSource;
	}

	public Hashtable<Integer, ARStorageBlock> getArStorageBlockTable() {
		return arStorageBlockTable;
	}

	/**
	 * This method traverse through each children of the node passing and adds
	 * the leaf Node's ARNodeBlock to the BlockList
	 * 
	 * @param n
	 *            ARNode variable
	 * @param list
	 *            An ArrayList of ARNodeBlock where the leaf ARNodeBlocks are
	 *            added
	 */
	public void createValidARNodeBlockList(ARNode n, ArrayList<ARNodeBlock> list) {

		if (n == null) {
			return;
		}

		for (int i = 0; i < 7; i++) {
			ARNode child = (ARNode) n.children[i];
			if (child != null && child.isLeaf()) {
				list.add(child.getBlock());

			}
			createValidARNodeBlockList((ARNode) n.children[i], list);
		}
	}

	
	/*
	 *  Method that returns a map of sparse Data Blocks with sparse Bounds of each Block as key. 
	 *  Returned map contains only the parts that was found in the sparse data.
	 *  If given bounds are not present in sparse then empty map is returned.
	 *  
	 */
	public Map<ISBounds, DataBlock> subblock(ISBounds bounds, int resolution){
		// first we need to check whether the given bounds are valid and find the 
		// valid and invalid sparse block parts for it.
		
		// based on the AR node resolution  we will be getting the valid parts from 
		//the respective resolution sparse file.
		Map<Integer, ISBounds> partsMap = sparseStorageArr[resolution].getParts(bounds) ; 
		Map<ISBounds, DataBlock> blocksMap  = new HashMap<ISBounds, DataBlock>() ; 
		
		if(!partsMap.isEmpty()){ 
			
			
			for(Map.Entry<Integer, ISBounds> mapEntry : partsMap.entrySet()){
				DataBlock dBlock = sparseStorageArr[resolution].readPartition(mapEntry.getKey()) ; 
				SparseBlockInfo sparseBlock = sparseStorageArr[resolution].getSparseDataTable().get(mapEntry.getKey()); 
				Integer[] sparseLower =  sparseBlock.getStartPosition() ; 
				ISBounds entryBounds = mapEntry.getValue() ; 
				int[] entryBoundsLower = entryBounds.getLowerArray(); 
				//int[] dbLower = dBlock.getLowerBoundsArray() ; 
				int[] lower = { entryBoundsLower[0] - sparseLower[0], 
								entryBoundsLower[1] - sparseLower[1], 
								entryBoundsLower[2] - sparseLower[2]} ; 
				
				int[] entryBoundsSize = entryBounds.getDimensionsArray(); 
				int[] upper = { (lower[0] + entryBoundsSize[0]) - 1 , 
								(lower[1] + entryBoundsSize[1]) - 1 ,
								(lower[2] + entryBoundsSize[2]) - 1  } ;
						
						    
			
				ISBounds currPartBounds = new ISBounds(lower, upper) ; 
				dBlock = dBlock.subblock(currPartBounds) ; 
				
				blocksMap.put(entryBounds, dBlock) ; 
			}
		}
		return blocksMap ; 
	}

	
	
	//---------------------------------------------------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------------------------------------------------
	/**
	 * This method creates a Hash table ARStorageBlock class. Extracts block
	 * data with respective block bounds and resolution Datasource adds in the
	 * datablock variable of ARStorageBlock
	 * 
	 * @param list
	 *            ArrayList of ARNodeBlock that contains the ARNode data
	 */
	public void createARDataBlocksAndAddToARTable() {

		for (int i = 0; i < blockList.size(); i++) {
			ARNode blk = blockList.get(i);
			ARStorageBlock arBlock = new ARStorageBlock();
			int res = blk.getBlock().getResolution();
			int[] dimArr = blk.getBounds().getDimensionsArray();

			// Treating this ARStorage as cache block so not adding now.
			DataBlock dataBlock = dsArr[res].subblock(new ISBounds(dimArr));
			// System.out.println("count : " +
			// dataBlock.getValues().getFloats().length );
			dsArr[res].subblock(dataBlock, blk.getBounds());

			arBlock.setArDataBlock(dataBlock);
			arBlock.setArBlockBounds(blk.getBounds());

			// arStorageBlockTable.put(ConversionUtilities.arrayToString(arBlock.getArBlockBounds().getLowerArray()),
			// arBlock);
			// System.out.println(" Hash bounds " + blk.getBounds());
			arStorageBlockTable.put(i, arBlock);
		}
		this.setTableCreated(true);
	}

	public static void main(String args[]) throws IOException {
		/**
		 * For pruning according to given error tolerance
		 */
		// Float error = 50.0f ;
		// ar_root.prune( new ARFloatError( error ));
		// ar_root.condense();

		// String[] testArr = null ;
		// // need to assign the test Arr values - just added now to avoid
		// compilation errors
		// ARStorage arStorage = new ARStorage(testArr) ;
		// arStorage.createValidARNodeBlockList(arStorage.getArDataSource().getRoot(),
		// arStorage.getBlockList());
		// System.out.println("List size " + arStorage.getBlockList().size() ) ;
		//
		//
		// arStorage.createARDataBlocksAndAddToARTable(arStorage.getBlockList());

		/**
		 * Lines of code written for easy understanding of code
		 */

		// file.createNewFile();
		// writer = new FileWriter(file);
		//
		// writeNode(arDataSource_x.getRoot());
		//
		// writer.flush();
		// writer.close();
		//

		String[] file_x = new String[2];
		file_x[0] = "ARData/Solar_vx/vx-64x64x64-bin.";
		file_x[1] = "xfdl";
		arStorage[0] = new ARStorage(file_x);

		String[] file_y = new String[2];
		file_y[0] = "ARData/Solar_vy/vy-64x64x64-bin.";
		file_y[1] = "xfdl";
		arStorage[1] = new ARStorage(file_y);

		String[] file_z = new String[2];
		file_z[0] = "ARData/Solar_vz/vz-64x64x64-bin.";
		file_z[1] = "xfdl";
		arStorage[2] = new ARStorage(file_z);

		for (int y = 0; y < 64; y++) {
			for (int x = 0; x < 64; x++) {

				System.out.println("x, y : " + x + ", " + y);
				IntegerIndexSpaceID id = new IntegerIndexSpaceID(x, y, 0);
				float dx = getPosValueInARDatasource(id, 0);
				float dy = getPosValueInARDatasource(id, 1);

				System.out.println("dx : " + dx);
				System.out.println("dy : " + dy);

			}
		}

	}

	private static float getPosValueInARDatasource(IntegerIndexSpaceID id,
			int index) {

		// Code used for Regular Partitioning - not used now since the AR block
		// partitions were not equal.
		// RegularPartitionISID rpISID ;
		// RegularISPartitioning partitioning =
		// arStorage[index].getBlockPartioning(0) ;
		//
		// ISBounds containingBounds =
		// partitioning.getContainingPartitionBounds(id) ;
		// System.out.println("containingBounds " + containingBounds ) ;
		// ARStorageBlock arblock =
		// arStorage[index].getArStorageBlockTable().get(index);
		// DataBlock dBlock = arblock.getArDataBlock() ;
		// Datum dtm = arStorage[index].dsArr[index].createDatum();
		// rpISID = new
		// RegularPartitionISID(arStorage[index].getBlockPartioning(0), id );
		// dBlock.datum( dtm, rpISID.getOffsetISID() );
		// return dtm.getFloat(0) ; // one only attribute available in current
		// test data

		// Data accessing based on the position

		ArrayList<ARNodeBlock> list = arStorage[index].getBlockList();
		System.out.println("list.size " + list.size());
		int h = 0;
		for (h = 0; h < list.size(); h++) {
			boolean withinBlock = false;
			ISBounds blkbounds = list.get(h).getBounds();
			// index space dimensions for loop
			for (int d = 0; d < id.dim(); d++) {
				if (id.getCoord(d) >= blkbounds.getLower(d)
						&& id.getCoord(d) <= blkbounds.getUpper(d)) {
					withinBlock = true;
				} else {
					withinBlock = false;
					break;
				}
			}

			if (withinBlock == true) {
				break;
			}

		}

		ARStorageBlock arBlock = arStorage[index].getArStorageBlockTable().get(
				h);
		ISBounds arBounds = arBlock.getArBlockBounds();
		DataBlock dBlock = arBlock.getArDataBlock();

		int[] exacLoc = new int[3];
		exacLoc[0] = id.getCoord(0) - arBounds.getLower(0);
		exacLoc[1] = id.getCoord(1) - arBounds.getLower(1);
		exacLoc[2] = id.getCoord(2) - arBounds.getLower(2);

		IntegerIndexSpaceID offsetCalcuID = new IntegerIndexSpaceID(exacLoc[0],
				exacLoc[1], exacLoc[2]);

		return dBlock.getFloat(offsetCalcuID, 0);

	}

	/**
	  * 
	  */
	public ISBounds traverseAndGetBlockBounds(ARNode node,
			IntegerIndexSpaceID id) {

		for (int i = 0; i < 8; i++) {
			ARNode child = (ARNode) node.children[i];
			if (child != null) {
				if (child.getNumChildren() > 0) {
					// this is the node we need
					boolean withinBlock = false;
					ISBounds blkbounds = child.getBounds();
					for (int d = 0; d < id.dim(); d++) {
						if (id.getCoord(d) >= blkbounds.getLower(d)
								&& id.getCoord(d) <= blkbounds.getUpper(d)) {
							withinBlock = true;
						} else {
							withinBlock = false;
							break;
						}
					}
					if (withinBlock == true) {
						traverseAndGetBlockBounds(child, id);
					}
				} else {
					// Found the exact block ;
					return child.getBounds();
				}
			}
		}
		return node.getBounds();
	}

	
	public Integer updateSibling(ISBounds currRegionBounds, Integer currRegionNumber, Point3f point3f) {

		String newRegion = "";
		boolean change = false;
			if (point3f.x < currRegionBounds.getLower(0)) {
				newRegion =  "L";
				change = true;
			} else if (point3f.x >= currRegionBounds.getUpper(0)) {
				newRegion =  "R";
				change = true;
			}
			if (point3f.y < currRegionBounds.getLower(1)) {
				newRegion = newRegion + "B";
				change = true;
			} else if (point3f.y >= currRegionBounds.getUpper(1)) {
				newRegion = newRegion + "T";
				change = true;
			}
			if (point3f.z < currRegionBounds.getLower(2)) {
				newRegion = newRegion + "F";
				change = true;
			} else if (point3f.z >= currRegionBounds.getUpper(2)) {
				newRegion = newRegion + "E";
				change = true;
			}
			if (change == true) {
				Map<String, Integer> map = regionMapper[currRegionNumber] ;
				return map.get(newRegion) ; 
				
			} // still in same region 
				
			return currRegionNumber;
			
	}
	
	
	public ARNode updateLeaf(ARNode residingNode, Point3f point3f){
		
		// ** same node **
		// if the position is inside the residing node then returning the same node. 
		if(withinBlock(residingNode.getBlock().getBounds(), point3f) == true){
			return residingNode ; 
		}
		
		// ** new node ==> sibling node **
		// if not inside the same residingNode then checking with the parent node bounds to see 
		// whether the new position is in any of the sibling Node. 
		ARNode newLeafNode ;
		while(true){
		ARNode parent = (ARNode) residingNode.getParent() ; 
		
		
		if(withinBlock(parent.getBlock().getBounds(), point3f) == true ){
		   newLeafNode = findChildNode(parent, residingNode.getBlock().getBounds(), point3f) ; 
			if(newLeafNode != null){
				break ;  
			}
		}
		
		residingNode = (ARNode) parent.getParent() ; 
		}
		
		
		if(newLeafNode.isLeaf()){
			return newLeafNode ;
		}else{ 
			// traverse through the node and find the leaf that is matching. 
			return traverseAndFindLeaf(newLeafNode, point3f) ; 
			
		}
		
		
	}
	
	private ARNode findChildNode(ARNode parent, ISBounds residingBounds,  Point3f point3f){
		
		ARNode child = null ;
		boolean found = false; 
		for(int i = 0 ; i <= 7 ;i++ ){
			child = (ARNode) parent.children[i] ; 
			if(child != null && child.getBlock().getBounds() != residingBounds){
				if(withinBlock(child.getBlock().getBounds(), point3f) == true){
					found = true ; 
					break ; 
				}
			}	
		}
		
		if(found == true){
			return child ;
		}
		return null ;
	}
	
	private boolean withinBlock(ISBounds nbnds, Point3f point3f) {
		boolean withinBlock_x = false;
		boolean withinBlock_y = false;
		boolean withinBlock_z = false;

		
		if (point3f.x >= nbnds.getLower(0) && point3f.x < nbnds.getUpper(0)) {
			withinBlock_x = true;
		}
		if (point3f.x >= nbnds.getLower(1) && point3f.x < nbnds.getUpper(1)) {
			withinBlock_y = true;
		}
		if (point3f.z >= nbnds.getLower(2) && point3f.z < nbnds.getUpper(2)) {
			withinBlock_z = true;
		}

		return (withinBlock_x && withinBlock_y && withinBlock_z);
	}
	
	public ARNode traverseAndFindLeaf(ARNode currNode,  Point3f point3f){
	
		for (int i = 0; i <= 7; i++) {
			ARNode child = (ARNode) currNode.children[i];
			if (child != null) {
				if (withinBlock(child.getBlock().getBounds(), point3f) == true) {
					if (child.isLeaf()) {
						return child;
					}
				}
				traverseAndFindLeaf(child, point3f);
			}
		}

		return null;
	}
}
