/** ByteValues. java
 * 
 * -----------------------------------------------------------
 * Edit history (started 11/27/04)
 * -------------------------------
 * 10/21/05 rdb: Removed unused variables
 * 02/07/05 rdb: Added getDataReference() to replace byteValues()
 * 02/03/05 rdb: Added get<type>s( <type> [] )
 * 01/26/05 rdb: Added setValues( int start, Values )
 * 01/25/05 rdb: Changed parms for spawn( int, int, int ) to (start,stride,count)
 *                   so they match order of get<type>s( start, stride, count )
 * 01/24/05 rdb: Added reOrder( int )
 * 01/21/05 rdb: Added getBytes( start, stride, count ) 
 *                 and getBytes( start, stride )        methods.
 * 01/19/05 rdb: Added getBytes(), left byteValues() which returns reference.
 * 01/13/05 rdb: Added: set<type> methods and especially setValues
 * 11/30/04 rdb: Removed "mixed" values functionality. This is now
 *               a class that supports byte variables.
 * 11/27/04 rdb: Added getByte method, implemented MixedValues
 *               for replacing old ByteValues functionality. 
 *               ByteValues should support single byte variables
 *               to match ShortValues, IntValues, etc. Did not 
 *               clean up this class yet.
 * -----------------------------------------------------------
 */
package edu.unh.sdb.common;

/** ByteValues is Values that store bytes in a 1-D byte array */
public class ByteValues extends Values
{
    private byte[] data;

    /** construct a ByteValues with no data */
    public ByteValues()
    {
       super();
       data = null;
    }

    /** construct a ByteValues for a given byte array */
    public ByteValues( byte[] byteArray )
    {
         super();
         this.data = byteArray;
    }

    /** construct a ByteValues, given the size of the byteArray*/
    public ByteValues( int size)
    {
         super();
         this.data = new byte[size];
    }

    /** create a copy of this object  */
    public Values spawn()
    {
       return new ByteValues( data.clone() );
    }
        
    /** Create a new instance of ByteValues with a subsest
     *  of the data in this object. Subsets can be contiguous or
     *  can have a stride as is done in point order storage.
     *  If there aren't "count" values to copy, the returned 
     *  object will only be as large as the available values.
     * @param start - first value to copy
     * @param stride - steps between values
     * @param count - number of values to copy
     */
    public Values spawn( int start, int stride, int count )
    {
       return new ByteValues( this.getBytes( start, stride, count ) );
    }

    /** set the data at given index  */
    final public void setData(int index, byte byteData)
     {
           data[index] = byteData;
     }

    /** get StorageType */
    public int getStorageType()
    {
         return StorageType.Byte;
     }   

    /** return the size of the 1 D array, in which the data is stored ( number of bytes) */
    public int size()
    {
         return data.length;
    }

    /** Return a new Values object with the values shuffled.
     *  This method starts at positions 0 through step-1, copying every 
     *  "step" values to the new array. This has the effect of converting 
     *  an attribute ordered array to a point ordered array or vice versa.
     *  The Values object does not actually need to know the axisOrdering,
     *  the conversion is the same. 
     */
    public Values  reorder( int step )
    {
       ByteValues ret = new ByteValues( this.size() );
       int  newIndex = 0;
       
       for ( int i = 0; i < step; i++ )
          for ( int j = i; j < this.size(); j += step )
             ret.data[ newIndex++ ] = this.data[ j ];
       return ret;
    }

    //+++++++++++++++ set<type> methods ++++++++++++++++++++++++++++++++++++++++
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setDouble( int index, double val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setFloat( int index, float val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setInt( int index, int val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setShort( int index, short val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setByte( int index, byte val )
    {
       data[ index ] = val;
    }
    
    /** Assign the data from the argument Values object to this object.
     *  This object will now have a new data size that is the same as the arguments.  
     *  (This version clones the data; do we want to just reference it?)  
     */
    public void setValues( Values newVals )
    {
       // I don't know if it really is faster to clone than copy, 
       // but this avoids the extra copy that would otherwise happen.
       if ( newVals instanceof ByteValues )
          this.data = ((ByteValues) newVals ).data.clone();  
       else
       {
	       byte newData[] = newVals.getBytes();
	       this.data = new byte[ newVals.size() ];
	       for ( int i = 0; i< newVals.size(); i++ )
	          data[ i ] = newData[ i ];
	   }
    }
    /** Assign elements of the passed in Values object to this one beginning
     *  at the start index . Do conversions as needed.
     *  If the argument size is less than is available beginning at the start 
     *  position, only the remaining values will be changed.
     */
    public void setValues( int start, Values newVals )
    {
       int  copySize = newVals.size();
       if ( copySize > this.size() - start )
          copySize = this.size() - start;
       byte newData[] = newVals.getBytes();
       
       for ( int i = 0; i < copySize; i++ )
          this.data[ i + start ] = newData[ i ];
    }
    //--------------------- get<type>s() -------------------------------
    /** Return double array representing all elements in the Values object */
    public  double[]   getDoubles()
    {
       double ret[] = new double[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** Return float array representing all elements in the Values object */
    public  float[]   getFloats()
    {
       float ret[] = new float[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** Return int array representing all elements in the Values object */
    public  int[]   getInts()
    {
       int ret[] = new int[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** Return a short array representing all elements in the Values object */
    public  short[]   getShorts()
    {
       short ret[] = new short[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** return a reference to the byte array this object is representing*/
    public byte[] getBytes()
    {
        return  data;
    }
       
    //------------- get <type>s( <type>[] ) -----------------------------
    /** fill the passed array with this data */
    @Override
    public double[] getDoubles( double[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getDoubles( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    @Override
    public float[] getFloats( float[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getFloats( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    @Override
    public int[] getInts( int[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getInts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    @Override
    public short[] getShorts( short[] filler )
    {
       if ( data.length != filler.length )
          return super.getShorts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    @Override
    public byte[] getBytes( byte[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getBytes( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    //+++++++++++++++ get<type>( int )  methods ++++++++++++++++++++++++++++++++++++
    /** return the float value representing the byte value
     *  of the given index
     */
    public float getFloat( int index ) 
    {
       if (index < 0)
          return 0;
       return data[ index ];
    }
    /** return the double value represent the byte value of the given index*/
    public double getDouble( int index) 
    {
       return data[ index ];
    }
    
    /** return the int value represent the byte value of the given index*/
    public int getInt( int index) 
    {
       return data[ index ];
    }
    
    /** return the short value represent the byte value of the given index*/
    public short getShort( int index) 
    {
       return data[ index ];
    }

    /** return the byte value at of the item at the given index of the byte array */
    public byte getByte(int index)
    {
        return data[index];
    }

     /** Return a reference to the byte data array this object contains.  
        * @deprecated 
        *  Use getDataReference()
      */
      public byte[] byteValues()
      {
          return data;
      }

      /** Return a reference to the data array this object contains.  
        */
       public byte[] getDataReference()
       {
           return data;
       }
        
      /** Return byte array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      @Override
      public byte[] getBytes( int start, int stride, int num )
      {
         int lastRequestedIndex = start + (num-1)*stride;
        
         if ( start >= this.size() )
            num = 0;
         else if ( lastRequestedIndex >= this.size() ) // not enough entries
            num = 1 + ( this.size() - start - 1 ) / stride; // find all that exist.
         byte ret[] = new byte[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            ret[ i ] = data[ offset ];
         return ret;
      }
}
