package edu.unh.sdb.common;

import edu.unh.sdb.datasource.SliceUtil;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

/**
 * Class that stores orientation using a set of Euler angles (heading,pitch,bank) 
 * and offers conversion routines to other ways of specifying orientation.
 * This is described in 3D Math Primer for Graphics and Game Development, chapter 10.
 * @author dlipsa
 */
public abstract class EulerAngles
{
   
   /** Creates a new instance of EulerAngles */
   public EulerAngles ()
   {
      heading = pitch = bank = 0;
   }
   
   public EulerAngles (double heading, double pitch, double bank)
   {
      this.heading = heading;
      this.pitch = pitch;
      this.bank = bank;
   }
   
   public EulerAngles (Vector3d v)
   {
      Matrix4d m = new Matrix4d ();
      SliceUtil.calculateRotation (getInitialOrientation (), v, m);
      setFromMatrix (m);
      bank = 0;
   }
   
   public EulerAngles create (Vector3d v)
   {
      if (this instanceof EulerAnglesYXZ)
         return new EulerAnglesYXZ (v);
      else if (this instanceof EulerAnglesZYX)
         return new EulerAnglesZYX (v);
      else
         return null;
   }
   
   public EulerAngles create (double heading, double pitch, double bank)
   {
      if (this instanceof EulerAnglesYXZ)
         return new EulerAnglesYXZ (heading, pitch, bank);
      else if (this instanceof EulerAnglesZYX)
         return new EulerAnglesZYX (heading, pitch, bank);
      else 
         return null;
   }
   
   @Override
   public String toString ()
   {
      return "EulerAngles(heading=" + heading + ", pitch=" +
         pitch + ", bank=" + bank + ")";
   }

   public abstract void rotateHeading (Matrix4d m);
   public abstract void rotatePitch (Matrix4d m);
   public abstract void rotateBank (Matrix4d m);
   public abstract Vector3d getInitialOrientation ();
   public abstract void setFromMatrix (Matrix4d m);
   
   public Vector3d vectorFromHeadingPitch ()
   {
      Matrix4d h = new Matrix4d ();
      Matrix4d p = new Matrix4d ();
      Vector3d orientation = getInitialOrientation ();
    
      rotateHeading (h);
      rotatePitch (p);
      p.transform (orientation);      
      h.transform (orientation);
      return orientation;
   }
         
   public Matrix4d calculateRotationHeadingPitchBank ()
   {
      Matrix4d r = new Matrix4d ();
      Matrix4d t = new Matrix4d ();
      rotateHeading (r);
      rotatePitch (t);
      r.mul (t);
      rotateBank (t);
      r.mul (t);
      return r;      
   }   
   
   /**
    * Calculate the normal on the rotation circle determined when 
    * heading is constant and pitch varies.
    */
   public Vector3d normalOnPitchRotation ()
   {      
      Vector3d first = vectorFromHeadingPitch ();
      pitch += 10;
      Vector3d second = vectorFromHeadingPitch ();
      pitch -= 10;
      Vector3d v = new Vector3d ();
      v.cross (first, second);
      return v;
   }
   
   public void setHeading (double heading)
   {
      this.heading = heading;
   }

   public void setBank (double bank)
   {
      this.bank = bank;
   }

   public void setPitch (double pitch)
   {
      this.pitch = pitch;
   }

   public double getHeading ()
   {
      return heading;
   }

   public double getPitch ()
   {
      return pitch;
   }

   public double getBank ()
   {
      return bank;
   }
   
   public void set (Angle angle, double value)
   {
      switch (angle)
      {
         case HEADING:
            setHeading (value);
            break;
         case PITCH:
            setPitch (value);
            break;
         case BANK:
            setBank (value);
            break;
         default:
            assert false;
      }
   }
   
   public void add (Angle angle, double value)
   {
      value += get (angle);
      set (angle, value);
   }

   public double get (Angle angle)
   {
      switch (angle)
      {
         case HEADING:
            return getHeading ();
         case PITCH:
            return getPitch ();
         case BANK:
            return getBank ();
         default:
            assert false;
            return 0;
      }
   }
            
   public enum Angle
   {
      HEADING,
      PITCH,
      BANK
   };
   
   private double heading;
   private double pitch;
   private double bank;
}
