/*    

    FieldIDMapper.java
    last modified 12/1/2001
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    ----------------- Edit history since 01/18/05 -----------------------------
    10/21/05 rdb: Remved unused variables and unused deprecated constructor
    02/20/05 rdb: Added FieldIDMapper( int[] )
    02/14/05 rdb: Fixed equals to accept null and return true if this is identity
    02/12/05 rdb: Added accessor methods based on src/dst 
                      getSources == getComponents
                      getDestinations == getComposites
                      toSource( int ) == toComponent( int )
                      toDestination( int ) == toComposite( int )
                  Got rid of 2 constructors replaced with static methods
                     createCrossMap
                     createComponentMap
    01/18/05 rdb: Added private checkIDArrays so checkComponentFields 
                      and checkCompositeFields would share the code.
                  Added a test to the sort method to see if ComponentFields already
                      sorted. We expect sorted array and our sort is costly so it 
                      might pay to make the test first.
*/

package edu.unh.sdb.common;

//import java.io.*;
import java.util.*;

/**
 *  This class records an array of field numbers from a component DataSource
 *  of a composite DataSource and a corresponding array of field numbers 
 *  from that composite DataSource. It is meant to be used in the methods
 *  datum() and subblock() to facilitate the datum retrieval.
 */

public class FieldIDMapper 
{
    /** The field numbers from a component DataSource. The numbers
     *  are sorted in ascending order. The numbers should be distinct.
     * Also can be thought of as the "source" of the mapping.
     */
    protected int [] componentFields;

    /** The one-to-one corresponding field numbers from the composite 
     *  DataSource. The numbers should be distinct.
     * Also can be thought of as the "destination" of the mapping. 
     */
    protected int [] compositeFields;

    /** Constructs a FieldIDMapper with two parallel arrays. 
     * component -> composite
     * source    -> destination 
     */
    public FieldIDMapper( int componentFIDs[], int compositeFIDs[] ) 
    {
        int size = componentFIDs.length;

        if ( size != compositeFIDs.length ) 
        {
            new IllegalArgumentError(SDBError.Recoverable,
                              "The sizes of the arrays don't match.");
        }
        else 
        {
            this.componentFields = componentFIDs.clone();
            this.compositeFields = compositeFIDs.clone();
            sort(); 
        }
    }

    /** Constructs an identity FieldIDMapper of specified size. */
    public FieldIDMapper( int numFields ) 
    {
        this.componentFields = identityArray( numFields );
        this.compositeFields = identityArray( numFields );
    }

    /** Constructs a FieldIDMapper by composing two related FieldIDMappers.
     * 
     *  The mapping is between fm1.compositeFields and fm2.componentFields.
     *  Useful for AttributeJoinDataSource.
     * 
     */
    public static FieldIDMapper createCrossMap( FieldIDMapper src, 
                                                  FieldIDMapper dst ) 
    {
       int [] cpnFids1 = src.getComponentFields();
       int [] cpsFids1 = src.getCompositeFields();
       int size1 = cpsFids1.length;
       int [] cpnFids2 = dst.getComponentFields();
       int [] cpsFids2 = dst.getCompositeFields();
       int size2 = cpsFids2.length;
       int [] cpsFids3 = new int[size2];
       int [] cpnFids3 = new int[size2];
       int size3 = 0;

       /** Find out mapping fieldID pairs where the mapping is
        *  from src.compositeFields[] to dst.componentFields*/
       for ( int i=0; i<size1; i++) {
           for ( int j=0; j<size2; j++) {
               if ( cpnFids1[i] == cpsFids2[j]) {
                   cpsFids3[size3] = cpsFids1[i];
                   cpnFids3[size3] = cpnFids2[j];
                   size3++;
               }
           }
       }
       FieldIDMapper cross = new FieldIDMapper( size3 );
       if ( size3 == 0) {
           cross.componentFields = null;
           cross.compositeFields = null;
       }
       else {
           cross.componentFields = new int[size3];
           cross.compositeFields = new int[size3];
           for ( int j=0; j<size3; j++) {
               cross.componentFields[j] = cpnFids3[j];
               cross.compositeFields[j] = cpsFids3[j];
           }
           cross.sort();
       }
       return cross;
    }


    /** Constructs a FieldIDMapper by composing two related FieldIDMappers.
     *  The mapping is between src.componentFields and dst.componentFields.
     *  Useful for BlockedDataSource.
     */
    public static FieldIDMapper createComponentMap( FieldIDMapper src, FieldIDMapper dst ) 
    {
        int [] cpnFids1 = src.getComponentFields();
        int [] cpsFids1 = src.getCompositeFields();
        int size1 = cpsFids1.length;
        int [] cpnFids2 = dst.getComponentFields();
        int [] cpsFids2 = dst.getCompositeFields();
        int size2 = cpsFids2.length;
        int [] cpsFids3 = new int[size2];
        int [] cpnFids3 = new int[size2];
        int size3 = 0;

        // Find out mapping fieldID pairs where the mapping is
        //  from src.compositeFields[] to dst.componentFields
        for ( int i=0; i<size1; i++) {
            for ( int j=0; j<size2; j++) {
                if ( cpsFids1[i] == cpsFids2[j]) {
                    cpsFids3[size3] = cpnFids2[j];
                    cpnFids3[size3] = cpnFids1[i];
                    size3++;
                }
            }
        }
        
        FieldIDMapper newMap = new FieldIDMapper( size3 );
        
        if ( size3 == 0 ) {
            newMap.componentFields = null;
            newMap.compositeFields = null;
        }
        else {
            newMap.componentFields = new int[size3];
            newMap.compositeFields = new int[size3];
            for ( int j=0; j<size3; j++) {
                newMap.componentFields[j] = cpnFids3[j];
                newMap.compositeFields[j] = cpsFids3[j];
            }
            newMap.sort();
        }
        return newMap;
    }

    /** Constructs a FieldIDMapper from a RecordSpec which specifies the 
     *  attributes from a DataSource. The RecordSpec entries become the
     * "source" or component fields; the destination (composite fields)
     * is the "identity" mapping: 0, 1, 2, ...
     * 
     * This.compositeFields records 
     *  the index positions of the field numbers in the RecordSpec argument.
     */
    public FieldIDMapper( RecordSpec recSpec ) 
    {
        this.componentFields = recSpec.getFieldNumbers();
        this.compositeFields = identityArray( recSpec.size() );
        sort();
    }

    /** Constructs a FieldIDMapper from a simple array of ints.
     *  the int array is the source (components). 
     *  the destination (composite fields) is 
     *  the "identity" mapping: 0, 1, 2, ...
     * 
     * This.compositeFields records 
     *  the index positions of the field numbers in the RecordSpec argument.
     */
    public FieldIDMapper( int[] source ) 
    {
        this.componentFields = source.clone();
        this.compositeFields = identityArray( source.length );
        sort();
    }

    /** Constructs a FieldIDMapper from a subset of given FieldIDMapper.
     *  The given RecordSpec specifies a subset of compositeFields (source) of
     *  the given FieldIDMapper. 
     */
    public FieldIDMapper(FieldIDMapper fm, RecordSpec cpsSpec) 
    {
        int [] fmcpsFIDs = fm.getCompositeFields();
        int [] fmcpnFIDs = fm.getComponentFields();
        int size = fmcpsFIDs.length;
        int size1 = cpsSpec.size();

        this.componentFields = new int[size1];
        this.compositeFields = cpsSpec.getFieldNumbers();
        /* Find out field numbers mapped to those in cpsSpec. */
        for ( int i=0; i<size1; i++) {
            for ( int j=0; j<size; j++) {
                if ( this.compositeFields[i] == fmcpsFIDs[j])
                    this.componentFields[i] = fmcpnFIDs[j];
            }
        }
        sort();
    }
    
    /** Create a integer array of specified length, initialized
     *  to the "identity" mapping: 0,1, ...
     */
    final protected static int[] identityArray( int len )
    {
       int[] temp = new int[ len ];
       identity( temp );
       return temp;
    }

    /** Assign identity mapping to the argument array
     *  "identity" mapping: 0,1, ...
     */
    final protected static void identity( int[] array )
    {
       for ( int i = 0; i < array.length; i++ )
          array[ i ] = i;
    }

    /** Return a copy of this FieldIDMapper object
    */
    public Object clone()
    {
        return new FieldIDMapper( this.componentFields, this.compositeFields );
    }

    /** 
     * sort the arrays according to field numbers in componentFields.
     *  Used by constructors.
     */
    private void sort() 
    {   
       //return;
       /*********/ 
        int size = this.componentFields.length;
        
        //rdb: This does a lot of work when we ask user to send in a sorted
        //     array. It's probably worthwhile to check if already sorted first.
        boolean sorted = true;
        for ( int i = 1; i < size && sorted; i++ )
           if ( this.componentFields[ i ] < this.componentFields[ i-1 ] )
              sorted = false;
           
        if ( sorted )   // don't bother to sort again
           return; 

        int  mult = 10000;
        
        // rdb: Highly unlikely that we'll ever get more than 10000 fields,
        //      but if we did, this will be an impossible bug to find, so
        //      I put in a test for exceeding the size
        //
        if ( size > mult )
           mult = size * 100; // size should be all we need, but we'll be safe

        // Convert this two arrays into one array.
        for ( int i=0; i<size; i++) {
            this.componentFields[i] = this.componentFields[i]*mult +
                                      this.compositeFields[i];
        }
        Arrays.sort( this.componentFields );
        // Recover the two arrays.
        for ( int i=0; i<size; i++) {
            this.compositeFields[i] = this.componentFields[i] % mult;
            this.componentFields[i] = this.componentFields[i] / mult;
        }
        /****/
    }
    
        /** Put two FieldIDMappers together. 
         * 
         */
    public FieldIDMapper combine(FieldIDMapper fmap1, FieldIDMapper fmap2) 
    {
        int[] cpsID1 = fmap1.getCompositeFields();
        int[] cpnID1 = fmap1.getComponentFields();
        int[] cpsID2 = fmap2.getCompositeFields();
        int[] cpnID2 = fmap2.getComponentFields();
        int size1 = cpsID1.length;
        int size2 = cpsID2.length;
        int size = size1 + size2;
        int[] allcpsID = new int[size];
        int[] allcpnID = new int[size];

        for ( int i=0; i<size1; i++) {
            allcpsID[i] = cpsID1[i];
            allcpnID[i] = cpnID1[i];
        }
        for ( int i=size1; i<size; i++) {
            allcpsID[i] = cpsID2[i-size1];
            allcpnID[i] = cpnID2[i-size1];
        }
        return new FieldIDMapper(allcpnID, allcpsID);
    }

    /** Returns true if it is an identity FieldIDMapper. */
    public boolean isIdentityMapper() 
    {
        for ( int i=0; i<this.componentFields.length; i++) 
        {
            if ( this.compositeFields[i] != this.componentFields[i])
                return false;
        }
        return true;
    }

    //-----------------Accessors---------------------------------

    /** Returns the field numbers of the component DataSource. */ 
    public int [] getComponentFields() 
    {
        if ( this.componentFields == null) 
            return null;
        return this.componentFields.clone();
    }

    /** Returns the field numbers of the composite DataSource. */
    public int [] getCompositeFields() 
    {
        if ( this.compositeFields == null)
            return null;
        return this.compositeFields.clone();
    }

    /** Returns the source identifier array
     *  This is the same as getComponentFields(), but FieldIDMapper
     *  need not be restricted to Component/Composite mapping */ 
    public int [] getSources() 
    {
        if ( this.componentFields == null) 
            return null;
        return this.componentFields.clone();
    }

    /** Returns the field numbers of the composite DataSource.
     *  This is the same as getCompositeFields(), but FieldIDMapper
     *  need not be restricted to Component/Composite mapping */ 
    public int [] getDestinations() 
    {
        if ( this.compositeFields == null)
            return null;
        return this.compositeFields.clone();
    }

    /** Returns the maximum field number of this.componentFields. */
    public int getMaxCpnID() 
    {
       int max=0;
       for ( int i=0; i < this.componentFields.length; i++ ) 
       {
           max = Math.max(max, this.componentFields[i]);
       }
       return max;
       
        //return this.componentFields[this.componentFields.length-1];
    }

    /** Returns the maximum field number of this.compositeFields. */
    public int getMaxCpsID() 
    {
        int max=0;
        for ( int i=0; i < this.compositeFields.length; i++ ) 
        {
            max = Math.max(max, this.compositeFields[i]);
        }
        return max;
    }

    /** Returns the number of fields. */
    public int size() 
    {
        if ( this.compositeFields == null) 
            return 0;
        return this.compositeFields.length;
    }
    
    /** return the component index given a composite index. */
    public int toComponent( int compositeIndex )
    {
        for ( int i=0; i < this.compositeFields.length; i++ )
        {
            if ( this.compositeFields[i] == compositeIndex )
                return this.componentFields[i];
        }
        new IllegalArgumentError( SDBError.Recoverable, 
                                   "Composite Index not found in FieldIDMapper");
        return -1;
    }
 
    /** Returns the composite index given the component index. */
    public int toComposite( int componentIndex )
    {
        for ( int i=0; i < this.componentFields.length; i++)
        {
            if ( this.componentFields[i] == componentIndex )
                return this.compositeFields[i];
        }
        new IllegalArgumentError(SDBError.Recoverable, 
                                "Component Index not found in FieldIDMapper");
        return -1;
    }

    /** return the source index given a destination index. 
     * Identical to toComponent( int ).
     * */
    public final int toSource( int destinationIndex )
    {
       return toComponent( destinationIndex );
    }
 
    /** Returns the destination index given the source index. 
     * Identical to toComposite( int )
     * */
    public final int toDestination( int sourceIndex )
    {
       return toComposite( sourceIndex );
    }

    /** Returns a RecordSpec mapped to the given RecordSpec whose
     *  field numbers could be found in this.compositeFields. It is
     *  useful for BlockedDataSource since the components and the
     *  composite DataSource have the same attributes.
     */
    public RecordSpec toComponent(RecordSpec spec1) {

        int size = this.compositeFields.length;
        int size1 = spec1.size();
        int [] fnum = spec1.getFieldNumbers();
        int [] fields = new int[size1];

        for ( int i=0; i<size1; i++) {
            for ( int j=0; j<size; j++) {
                if ( fnum[i] == this.compositeFields[j])
                    fields[i] = this.componentFields[j];
            }
        }
        return new RecordSpec(fields);
    }

    //-------------------equals()-----------------------
    /** Return true if parameter equals this object, false otherwise. */
    public boolean equals( Object obj ) 
    {
        FieldIDMapper fidm = (FieldIDMapper) obj;
        if ( fidm == null )  // they are equal if this is the identity mapping
           return isIdentityMapper();
        else
           return (this.componentFields).equals(fidm.getComponentFields()) 
                 &&
                  (this.compositeFields).equals(fidm.getCompositeFields());
    }

    /**
     * Check whether  all entries in the array are unique.
     * 
     * @param idArray
     * @return true if all elements in array are unique, else return false.
     */
    private boolean checkIDArray( int [] idArray ) 
    {
       int id;
       int len = idArray.length - 1;

       for ( int i = 0; i < len; i++ ) 
       {
           id = idArray[i];
           for ( int j = i + 1; j <= len; j++ ) 
           {
               if ( id == idArray[j] )
                   return false;
           }
       }
       return true;
   }

    /** Check whether the numbers in this.componentFields are distinct.*/
    public void checkComponentID() 
    {
       if ( checkIDArray( this.componentFields ))
          new IllegalArgumentError( SDBError.Recoverable, 
                                    "ComponentField IDs are not distinct." );
       
       /*** code revised 01/18/05 rdb
        int id;
        int len = this.componentFields.length - 1;

        for ( int i=0; i<len; i++) {
            id = this.componentFields[i];
            for ( int j=i+1; j<=len; j++) {
                if ( id == this.componentFields[j])
                    new IllegalArgumentError(SDBError.Recoverable,
                                             "Field IDs must be distict");
            }
        }
        /************************/
    }

    /** Check whether the numbers in this.compositeFields are distinct.*/
    public void checkCompositeID()     
    {
       if ( checkIDArray( this.compositeFields ))
          new IllegalArgumentError( SDBError.Recoverable,
                                    "CompositeField IDs are not distinct." );
       
       /*** code revised 01/18/05 rdb
        int id;
        int len = this.compositeFields.length - 1;

        for ( int i=0; i<len; i++) {
            id = this.compositeFields[i];
            for ( int j=i+1; j<=len; j++) {
                if ( id == this.compositeFields[j])
                    new IllegalArgumentError(SDBError.Recoverable,
                                             "Field IDs must be distict");
            }
        /************************/
    }

    /** to String */
    public String toString()
    {
        String cpnNumber = new String();
        String cpsNumber = new String();
        for ( int i = 0; i < componentFields.length; i++)
         {
            cpnNumber = cpnNumber +" " + componentFields[i];
            cpsNumber = cpsNumber +" " + compositeFields[i];
         }
         return "\n Component :[" + cpnNumber + "]\n"+" Composite :[" +cpsNumber +"]";
    }  
    
    //---------------------- main -----------------------------
    public static void main( String args[] )
    {
       //rdb: Just some dumb tests to make sure my "trivial" edits actually work
       
       // set up some int arrays to use later
       int arr1len6[] = new int[] { 0, 1, 2, 3, 4, 5 };
       int arr2len6[] = new int[] { 1, 2, 3, 4, 5, 0 };
       //rdb int arr1len5[] = new int[] { 1, 2, 3, 4, 5 };
       //rdb int arr2len5[] = new int[] { 2, 4, 6, 10, 8 };
       int arrDupl [] = new int[] { 0, 1, 2, 3, 3 };
              
       // test checkIDArray method
       FieldIDMapper goodFID = new FieldIDMapper( arr1len6, arr2len6 );
       goodFID.checkCompositeID();
       goodFID.checkComponentID();
       
       System.out.println( "check good composite/component arrays passed test. ");
       FieldIDMapper badFID = new FieldIDMapper( arrDupl, arrDupl );
       try {
          badFID.checkCompositeID();
       } catch (Exception ex)
       {
          System.out.println( ex.getMessage());
       }
       try {
           badFID.checkComponentID();
       } catch (Exception ex)
       {
          System.out.println( ex.getMessage());
       }
             
       // test use of pre-test in sort() method
       FieldIDMapper unsorted = new FieldIDMapper( arr2len6, arr1len6 );
       System.out.println( "Component array was unsorted. It should be sorted now.");
       System.out.println( unsorted );
    }
}













