
/************************************************************ 
   FloatValues.java
   Copyright 2000 University of New Hampshire
   Durham, NH 03824

   Xuan Tang xtang@cs.unh.edu
   last update: 4/17/03
************************************************************/
/** FloatValues.java
 * 
 * -----------------------------------------------------------
 * Edit history (started 11/27/04)
 * -------------------------------
 * 02/07/05 rdb: Added getDataReference() to replace floatValues()
 * 01/31/05 rdb: Added get<type>s( type[] )
 * 01/26/05 rdb: Added setValues( int start, Values )
 * 01/25/05 rdb: Changed parms for spawn( int, int, int ) to (start,stride,count)
 *                   so they match order of get<type>s( start, stride, count )
 * 01/24/05 rdb: Added reOrder( int )
 * 01/21/05 rdb: Added getFloats( start, stride, count ) 
 *                 and getFloats( start, stride )      methods.
 * 01/19/05 rdb: Added getFloats; left floatValues() which returns a reference.
 * 01/13/05 rdb: Added: set<type> methods and especially setValues
 * 11/27/04 rdb: Added getByte method
 * -----------------------------------------------------------
 */

package edu.unh.sdb.common;
import java.io.*;

/** FloatValues store floats in a 1 D float array */
public class FloatValues extends Values
{
    private float[] data;

    /** construct a FloatValues with no data */
    public FloatValues()
    {
       super();
       data = null;
    }

    /** construct a FloatValues for a given float array */
    public FloatValues( float[] floatArray )
    {
         super();
         this.data = floatArray;
    }

    /** construct a FloatValues for a given float array with the given number of floats.  */
    public FloatValues( int size )
    {
        super();
        this.data = new float[size];
         
        for( int i=0; i<size; i++)
        {
            this.data[i]=Float.NaN;
        }        
    }

    /** set the data at given index  NEW NEW NEW */
    final public void setData(int index, float floatData)
     {
           data[index] = floatData;
     }
 
    /** return the storageType of this FloatValues, The data type of the array used to store  
        the actual data 
     */
    public int getStorageType()
    {
        return StorageType.Float;
    } 

    /** retrun the size of the 1 D array, in which the data is stored */
    public int size()
    {
         return data.length;
    }

    /** Return a new Values object with the values shuffled.
     *  This method starts at positions 0 through step-1, copying every 
     *  "step" values to the new array. This has the effect of converting 
     *  an attribute ordered array to a point ordered array or vice versa.
     *  The Values object does not actually need to know the axisOrdering,
     *  the conversion is the same. 
     */
    public Values  reorder( int step )
    {
       FloatValues ret = new FloatValues( this.size() );
       int  newIndex = 0;
       
       for ( int i = 0; i < step; i++ )
          for ( int j = i; j < this.size(); j += step )
             ret.data[ newIndex++ ] = this.data[ j ];
       return ret;
    }

    //--------------------- get<type>s() -------------------------------
    /** Return double array representing all elements in the Values object */
    public  double[]   getDoubles()
    {
       double ret[] = new double[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** return a reference to the float array this object is representing*/
    public float[] getFloats()
    {
        return  data;
    }
       
    /** Return int array representing all elements in the Values object */
    public  int[]   getInts()
    {
       int ret[] = new int[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = (int) this.data[i];
       return ret;
    }
    
    /** Return a short array representing all elements in the Values object */
    public  short[]   getShorts()
    {
       short ret[] = new short[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = (short) this.data[i];
       return ret;
    }
    
    /** Return a byte array representing all elements in the Values object */
    public  byte[]   getBytes()
    {
       byte ret[] = new byte[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = (byte) this.data[i];
       return ret;
    }

    //------------- get <type>s( <type>[] ) -----------------------------
    /** fill the passed array with this data */
    public double[] getDoubles( double[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getDoubles( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public float[] getFloats( float[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getFloats( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public int[] getInts( int[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getInts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = (int) data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public short[] getShorts( short[] filler )
    {
       if ( data.length != filler.length )
          return super.getShorts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = (short) data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public byte[] getBytes( byte[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getBytes( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = ( byte ) data[ i ];
       return filler;
    }
       
    //--------------- get<type>( int ) -----------------------------------
    /** return the float value at the given index of the array */
    public float getFloat(int index)
    {
        return data[index];
    }
    
    /** return the int value of the float at the given index of the float array */
    public int getInt(int index)
    {
        return (int) data[index];
    }

    /** return the short value of the float at the given index of the float array */
    public short getShort(int index)
    {
        return (short) data[index];
    }

    /** return the byte value of the float at the given index of the float array */
    public byte getByte( int index )
    {
        return (byte) data[index];
    }

    /** return the double value of the float at the given index of the float array */
    public double getDouble( int index)
    {
        return data[index];
    }

    /** Return a reference to the float data array this object contains.  
     * @deprecated 
     * Use getDataReference()
     */
    public float[] floatValues()
    {
        return data;
    }
    /** Return a reference to the data array this object contains.  
      */
     public float[] getDataReference()
     {
         return data;
     }
      
    /** Return float array representing a subset of elements in the Values object
     *  Copying starts at the index specified by the start argument, 
     *  the 2nd item copied will be at index start + stride where stride is the 
     *  second argument; the 3rd is at start + 2*stride and so forth until 
     *  the specified number of values have been copied (3rd argument). 
     *  If there are not enough elements in the Values object to satisfy the 
     *  request, the returned array will have only as many values as could be
     *  found given the start and stride arguments. 
     */
    public float[] getFloats( int start, int stride, int num )
    {
       int lastRequestedIndex = start + (num-1)*stride;
      
       if ( start >= this.size() )
          num = 0;
       else if ( lastRequestedIndex >= this.size() ) // not enough entries
          num = 1 + ( this.size() - start - 1 ) / stride; // find all that exist.
       float ret[] = new float[ num ];
       int offset = start;
       for ( int i = 0; i < num; i++, offset += stride )
          ret[ i ] = data[ offset ];
       return ret;
    }
    
    /** create a copy of this object  */
    public Values spawn()
    {
       return new FloatValues( data.clone() );
    }
    
    /** Create a new instance of FloatValues with a subsest
     *  of the data in this object. Subsets can be contiguous or
     *  can have a stride as is done in point order storage.
     *  If there aren't "count" values to copy, the returned 
     *  object will only be as large as the available values.
     * @param start - first value to copy
     * @param stride - steps between values
     * @param count - number of values to copy
     */
    public Values spawn( int start, int stride, int count )
    {
       return new FloatValues( this.getFloats( start, stride, count ) );
    }

    //+++++++++++++++ set<type> methods ++++++++++++++++++++++++++++++++++++++++
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setDouble( int index, double val )
    {
       data[ index ] = (float) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setFloat( int index, float val )
    {
       data[ index ] = val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setInt( int index, int val )
    {
       data[ index ] = val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setShort( int index, short val )
    {
       data[ index ] = val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setByte( int index, byte val )
    {
       data[ index ] = val;
    }
    
    /** Assign the data from the argument Values object to this object.
     *  This object will now have a new data size that is the same as the arguments.  
     *  (This version clones the data; do we want to just reference it?)  
     */
    public void setValues( Values newVals )
    {
       // I don't know if it really is faster to clone than copy, 
       // but this avoids the extra copy that would otherwise happen.
       if ( newVals instanceof FloatValues )
          this.data = ((FloatValues) newVals ).data.clone();  
       else
       {
	       float newData[] = newVals.getFloats();
	       this.data = new float[ newVals.size() ];
	       for ( int i = 0; i< newVals.size(); i++ )
	          data[ i ] = newData[ i ];
	   }
    }
    /** Assign elements of the passed in Values object to this one beginning
     *  at the start index . Do conversions as needed.
     *  If the argument size is less than is available beginning at the start 
     *  position, only the remaining values will be changed.
     */
    public void setValues( int start, Values newVals )
    {
       int  copySize = newVals.size();
       if ( copySize > this.size() - start )
          copySize = this.size() - start;
       float newData[] = newVals.getFloats();
       
       for ( int i = 0; i < copySize; i++ )
          this.data[ i + start ] = newData[ i ];
    }
    //----------------------- main -------------------------------
    
    static public void  main( String args[] )
    {
       FloatValues fv = new FloatValues( new float[] 
                                {  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
					             10,11,12,13,14,15,16,27,18,19,
					             20,21,22,23,24,25,26,27,28,29,
					             30,31,32,33,34,35,36,37,38,39,
					             40,41,42,43,44,45,46,47,48,49,
					             50,51,52,53,54,55,56,57,58,59,
					             60,61,62,63,64,65,66,67,68,69,
					             70,71,72,73,74,75,76,77,78,79,
					             80,81,82,83,84,85,86,87,88,89,
					             90,91,92,93,94,95,96,97,98,99 } );
       float ar[];
       
       System.out.println( " Start    Len      Stride   ResultLength" );
       
       for ( int stride = 1; stride < 11; stride += 3 )
       {
          for ( int start = 0; start < fv.size(); start += 6 )
          {
             ar = fv.getFloats( start, stride, fv.size()/stride );
             System.out.println( "   " + start + 
                                "         " + fv.size() / stride +
                                "         " + stride + 
                                "         " + ar.length );
          }
       }
       
       ar = fv.getFloats( 9, 10, 10 );
       printArray( System.out, ar );

       ar = fv.getFloats( 10, 10, 10 );
       printArray( System.out, ar );

       ar = fv.getFloats( 6, 4, 25 );
       printArray( System.out, ar );
}
    private static void printArray( PrintStream out, float ar[] )
    {
       out.println( "-----------------");
       out.print( "[ " );
       if ( ar.length != 0 )
          out.print( ar[0] );
       for ( int i=1; i < ar.length; i++ )
          out.print( ", " + ar[i] );
       out.println( "]" );
    }
}
