/** IntValues.java
 * 
 * -----------------------------------------------------------
 * Edit history (started 11/27/04)
 * -------------------------------
 * 02/07/05 rdb: Added getDataReference() to replace intValues()
 * 02/03/05 rdb: Added get<type>s( <type>[] )
 * 01/26/05 rdb: Added setValues( int start, Values )
 * 01/25/05 rdb: Changed parms for spawn( int, int, int ) to (start,stride,count)
 *                   so they match order of get<type>s( start, stride, count )
 *               Added spawn( start, stride )
 * 01/24/05 rdb: Added reOrder( int )
 * 01/21/05 rdb: Added getInts( start, stride, count )
 *                 and getInts( start, stride )
 * 01/19/05 rdb: Added getInts; left intValues() which returns a reference
 * 01/14/05 rdb: Added getStorageType(), byteSize()
 * 01/13/05 rdb: Added: set<type> methods and especially setValues
 * 11/27/04 rdb: Added getByte method
 * 
 * -----------------------------------------------------------
 */
package edu.unh.sdb.common;

/** IntValues is Values that store ints in a 1 D float array */
public class IntValues extends Values
{
    private int[] data;

    /** construct a IntValues with no data */
    public IntValues()
    {
       super();
       data = null;
    }

    /** construct a IntValues for a given int array */
    public IntValues( int[] intArray )
    {
         super();
         this.data = intArray;
    }

    /** construct a IntValues for a given int array with the given number of ints.
        set all the values to 0
     */
    public IntValues( int size )
    {
        super();
        this.data = new int[size];
        for( int i=0; i<size; i++ ){
            this.data[i] = 0;
        }        
    }

    /** set the data at given index  */
    final public void setData(int index, int intData)
     {
           data[index] = intData;
     }

    /** get storage Type */
    public int getStorageType()
    {
       return StorageType.Int;
    }

    /** retrun the size of the 1 D array, in which the data is stored */
    public int size()
    {
         return data.length;
    }

    /** retrun the size in bytes of all the attributes in this datum. */
    public int byteSize()
    {
         return data.length*4;
    }

    /** Return a new Values object with the values shuffled.
     *  This method starts at positions 0 through step-1, copying every 
     *  "step" values to the new array. This has the effect of converting 
     *  an attribute ordered array to a point ordered array or vice versa.
     *  The Values object does not actually need to know the axisOrdering,
     *  the conversion is the same. 
     */
    public Values  reorder( int step )
    {
       IntValues ret = new IntValues( this.data.length );
       int  newIndex = 0;
       
       for ( int i = 0; i < step; i++ )
          for ( int j = i; j < this.size(); j += step )
             ret.data[ newIndex++ ] = this.data[ j ];
       return ret;
    }

    //--------------------- get<type>s() -------------------------------
    /** Return double array representing all elements in the Values object */
    public  double[]   getDoubles()
    {
       double ret[] = new double[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** Return float array representing all elements in the Values object */
    public  float[]   getFloats()
    {
       float ret[] = new float[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = this.data[i];
       return ret;
    }
    
    /** return a reference to the int array this IntValues object is representing*/
    public int[] getInts()
    {
        return  data;
    }
       
    /** Return a short array representing all elements in the Values object */
    public  short[]   getShorts()
    {
       short ret[] = new short[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = (short) this.data[i];
       return ret;
    }
    
    /** Return a byte array representing all elements in the Values object */
    public  byte[]   getBytes()
    {
       byte ret[] = new byte[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = (byte) this.data[i];
       return ret;
    }

    //------------- get <type>s( <type>[] ) -----------------------------
    /** fill the passed array with this data */
    public double[] getDoubles( double[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getDoubles( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public float[] getFloats( float[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getFloats( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public int[] getInts( int[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getInts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public short[] getShorts( short[] filler )
    {
       if ( data.length != filler.length )
          return super.getShorts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = (short) data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public byte[] getBytes( byte[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getBytes( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = ( byte ) data[ i ];
       return filler;
    }
       
    //+++++++++++++++ get<type>( int ) methods ++++++++++++++++++++++++++++++++++++++++

    /** return the float value of the int at the given index of the array */
    public float getFloat(int index)
    {
        return  data[index];
    }
    
    /** return the int value of the int at the given index of the int array */
    public int getInt(int index)
    {
        return  data[index];
    }

    /** return the short value of the int at the given index of the int array */
    public short getShort(int index)
    {
        return (short) data[index];
    }

    /** return the short value of the int at the given index of the float array */
    public byte getByte( int index )
    {
        return (byte) data[index];
    }

    /** return the double value of the int at the given index of the int array */
    public double getDouble(int index)
    {
        return data[index];
     }

    /** Return a reference to the int data array this object contains.  
     * @deprecated 
     * Use getDataReference()
     */
   public int[] intValues()
    {
        return data;
    }

   /** Return a reference to the data array this object contains.  
     */
    public int[] getDataReference()
    {
        return data;
    }
     

   /** Return int array representing a subset of elements in the Values object
    *  Copying starts at the index specified by the start argument, 
    *  the 2nd item copied will be at index start + stride where stride is the 
    *  second argument; the 3rd is at start + 2*stride and so forth until 
    *  the specified number of values have been copied (3rd argument). 
    *  If there are not enough elements in the Values object to satisfy the 
    *  request, the returned array will have only as many values as could be
    *  found given the start and stride arguments. 
    */
   public int[] getInts( int start, int stride, int num )
   {
      int lastRequestedIndex = start + (num-1)*stride;
     
      if ( start >= this.size() )
         num = 0;
      else if ( lastRequestedIndex >= this.size() ) // not enough entries
         num = 1 + ( this.size() - start - 1 ) / stride; // find all that exist.
      int ret[] = new int[ num ];
      int offset = start;
      for ( int i = 0; i < num; i++, offset += stride )
         ret[ i ] = data[ offset ];
      return ret;
   }
   
    /** create a copy of this object  */
    public Values spawn()
    {
       return new IntValues( data.clone() );
    }
    
    /** Create a new instance of IntValues with a subsest
     *  of the data in this object. Subsets can be contiguous or
     *  can have a stride as is done in point order storage.
     *  If there aren't "count" values to copy, the returned 
     *  object will only be as large as the available values.
     * @param start - first value to copy
     * @param stride - steps between values
     * @param count - number of values to copy
     */
    public Values spawn( int start, int stride, int count )
    {
       return new IntValues( this.getInts( start, stride, count ));
    }

    //+++++++++++++++ set<type> methods ++++++++++++++++++++++++++++++++++++++++
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setDouble( int index, double val )
    {
       data[ index ] = (int) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setFloat( int index, float val )
    {
       data[ index ] = (int) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setInt( int index, int val )
    {
       data[ index ] = val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setShort( int index, short val )
    {
       data[ index ] = val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed. */
    public void setByte( int index, byte val )
    {
       data[ index ] = val;
    }


    /** Assign the data from the argument Values object to this object.
     *  This object will now have a new data size that is the same as the arguments.
     *  (This version clones the data; do we want to just reference it?)  
     */
    public void setValues( Values newVals )
    {
       // I don't know if it really is faster to clone than copy, 
       // but this avoids the extra copy that would otherwise happen.
       if ( newVals instanceof IntValues )
          this.data = ((IntValues) newVals ).data.clone();  
       else
       {
	       int newData[] = newVals.getInts();
	       this.data = new int[ newVals.size() ];
	       for ( int i = 0; i< newVals.size(); i++ )
	          data[ i ] = newData[ i ];
	   }
    }
    /** Assign elements of the passed in Values object to this one beginning
     *  at the start index . Do conversions as needed.
     *  If the argument size is less than is available beginning at the start 
     *  position, only the remaining values will be changed.
     */
    public void setValues( int start, Values newVals )
    {
       int  copySize = newVals.size();
       if ( copySize > this.size() - start )
          copySize = this.size() - start;
       int newData[] = newVals.getInts();
       
       for ( int i = 0; i < copySize; i++ )
          this.data[ i + start ] = newData[ i ];
    }   
}
