/** MixedValues. java
 * 
 * @TODO correctly implement all the get/set methods so that they use the 
 *       RecordDescriptor to find appropriate fields in the set of bytes.
 *       Also to make it a true array of Records. 
 * -----------------------------------------------------------
 * Edit history (started 11/27/04)
 * -------------------------------
 * 02/07/05 rdb: Added getDataReference() to replace byteValues()
 * 02/03/05 rdb: Added get<type>s( <type>[] )
 * 01/26/05 rdb: Added setValues( int start, Values )
 * 01/25/05 rdb: Changed parms for spawn( int, int, int ) to (start,stride,count)
 *                   so they match order of get<type>s( start, stride, count )
 * 01/24/05 rdb: Added dummy reOrder( int )
 * 01/15/05 rdb: Added RecordDescriptor. This allows the class to extract 
 *               fields in the same way that the other Values classes can.
 * 01/13/05 rdb: Added: set<type> methods and especially setValues
 *               Realized that this won't work without a RecordDescriptor.
 *                 Do we want to do that??
 * 11/28/04 rdb: Created from ByteValues.java
 *               ByteValues becomes a uniform array of bytes
 * 11/27/04 rdb: Added getByte method
 * 
 * -----------------------------------------------------------
 */
package edu.unh.sdb.common;

/** MixedValues stores attributes as byte sequences in a 1 D byte array */
public class MixedValues extends Values
{
    private byte[] data;
    private RecordDescriptor recordDesc = null;

    /** construct a MixedValues with no data */
    protected MixedValues()
    {
       super();
       this.data = null;
    }

    /** construct a MixedValues with no data */
    public MixedValues( RecordDescriptor rd )
    {
       this( rd, null );
    }

    /** construct a MixedValues for a given byte array */
    public MixedValues( RecordDescriptor rd, byte[] byteArray )
    {
         super();
         this.data = byteArray;
    }

    /** construct a MixedValues, given the size of the byteArray*/
    public MixedValues( RecordDescriptor rd, int size )
    {
         this( rd, new byte[ size ] );
    }

    /** create a copy of this object  */
    public Values spawn()
    {
       return new MixedValues( this.recordDesc, data.clone() );
    }
        
    /** Create a new instance of MixedValues with a subsest
     *  of the data in this object. Subsets can be contiguous or
     *  can have a stride as is done in point order storage.
     *  If there aren't "count" values to copy, the returned 
     *  object will only be as large as the available values.
     * @param start - first value to copy
     * @param stride - steps between values
     * @param count - number of values to copy
     */
    public Values spawn( int start, int stride, int count )
    {
       byte newData[] = new byte[ count ];
       int src = start;
       int dst = 0;
       try {
          for ( dst = 0; dst < count; dst++, src += stride )
             newData[ dst ] = data[ src ];
       }
       catch ( ArrayIndexOutOfBoundsException ex )
       {
          new SDBError(SDBError.Recoverable, "Values.spawn: array bounds error."  );
          for ( ; dst < count; dst++ )
             newData[ dst ] = 0;
       }
       return new MixedValues( this.recordDesc, newData );
    }


    /** get StorageType */
    public int getStorageType()
    {
         return StorageType.Record;
     }   

    /** return the size of the 1 D array, in which the data is stored ( number of bytes) */
    public int size()
    {
         return data.length;
    }

    /** Return a new Values object with the values shuffled.
     *  This method starts at positions 0 through step-1, copying every 
     *  "step" values to the new array. This has the effect of converting 
     *  an attribute ordered array to a point ordered array or vice versa.
     *  The Values object does not actually need to know the axisOrdering,
     *  the conversion is the same.  
     */
    public Values  reorder( int step )
    {
       return this;
    }

    //--------------------- get<type>s() -------------------------------
    // rdb: these aren't really working right since the get<type>( int ) aren't
    /** Return double array representing all elements in the Values object */
    public  double[]   getDoubles()
    {
       double ret[] = new double[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = getDouble( i );
       return ret;
    }
    
    /** Return float array representing all elements in the Values object */
    public  float[]   getFloats()
    {
       float ret[] = new float[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = getFloat( i );
       return ret;
    }
    
    /** Return int array representing all elements in the Values object */
    public  int[]   getInts()
    {
       int ret[] = new int[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = getInt( i );
       return ret;
    }
    
    /** Return a short array representing all elements in the Values object */
    public  short[]   getShorts()
    {
       short ret[] = new short[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = getShort( i );
       return ret;
    }
    
    /** Return a byte array representing all elements in the Values object */
    public  byte[]   getBytes()
    {
       byte ret[] = new byte[ size() ];
       for ( int i = 0; i < size(); i++ )
          ret[ i ] = getByte( i );
       return ret;
    }

    //------------- get <type>s( <type>[] ) -----------------------------
    /** fill the passed array with this data */
    public double[] getDoubles( double[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getDoubles( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public float[] getFloats( float[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getFloats( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] =  data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public int[] getInts( int[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getInts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public short[] getShorts( short[] filler )
    {
       if ( data.length != filler.length )
          return super.getShorts( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] =  data[ i ];
       return filler;
    }
       
    /** fill the passed array with this data */
    public byte[] getBytes( byte[] filler )
    {
       if ( filler == null || data.length != filler.length )
          return super.getBytes( filler );    // let parent handle
       for ( int i = 0; i < filler.length; i++ )
          filler[ i ] = data[ i ];
       return filler;
    }
       
    //---------------- get<type>( int ) -----------------------
    /** return the float value represented by 4 adjacent bytes, starting 
     * with the byte of the given index
     */
    public float getFloat( int index ) throws NumberFormatException 
    {
       int bits;
      //System.out.println("Starting MixedValues.getFloat()");
       if( (index +3) <= this.size()-1 )
       {
           bits = (  (data[index]   & 0xFF) << 24 ) |
                  (  (data[index+1] & 0xFF) << 16 ) |
                  (  (data[index+2] & 0xFF) << 8  ) |
                  (  (data[index+3] & 0xFF));

         }
        else
            throw new NumberFormatException("Not enough bytes to convert a float ");

        return Float.intBitsToFloat(bits);
    }
    /** return the double value represented by 8 adjacent bytes, starting with the byte of the given index*/
    public double getDouble(int index) throws NumberFormatException
    {
        long lbits;
        if( (index +7) <= this.size()-1) 
         {
                lbits=  ((data[index]   & 0xFF)        << 56 ) |
                        ((data[index+1] & 0xFF)        << 48 ) |
                        ((data[index+2] & 0xFF)        << 40 ) |
                        ((data[index+3] & 0xFF)        << 32 ) |
                        ((data[index+4] & 0xFF)        << 24 ) |
                        ((data[index+5] & 0xFF)        << 16 ) |
                        ((data[index+6] & 0xFF)        << 8  ) |
                        ((data[index+7] & 0xFF));
         }
        else throw new NumberFormatException("Not enough bytes to convert a double");
        return Double.longBitsToDouble(lbits);
    }
    

    /** return the int value represented by 4 adjacent bytes, starting with the byte of 
        the given index.
        If there are not enought bytes to convert to a int, NumberFormatException is thrown 
     */
    public int getInt(int index) throws NumberFormatException
    {
       int bits;
       if( (index +3) <= this.size()-1 ){
		 bits=	(  (data[index]   & 0xFF) 	<< 24 ) |
                        (  (data[index+1] & 0xFF) 	<< 16 ) |
                        (  (data[index+2] & 0xFF) 	<< 8  ) |
                        (  (data[index+3] & 0xFF));

         }
        else
            throw new NumberFormatException("Not enough bytes to convert an int ");

        return bits;
    }

   /** return the short value represented by 2 adjacent bytes, starting with the byte of 
        the given index.
        If there are not enought bytes to convert to a short, NumberFormatException is thrown 
     */
    public short getShort( int index )
    {
       int bits;
       if( (index +1) <= this.size()-1 ){
		 bits=	(  (data[index]   & 0xFF) 	<< 8 ) |
                        (  (data[index+1] & 0xFF));

         }
        else
            throw new NumberFormatException("Not enough bytes to convert a short ");

        return (short) bits;
    }

    /** return the byte value of the item at the given index of the byte array */
    public byte getByte(int index)
    {
        return data[index];
    }

    /** Return a reference to the byte data array this object contains. 
     *  Note that each attribute value in the object uses a different number
     *  of bytes. 
     * @deprecated
     *  Use getDataReference()
     */
    public byte[] byteValues()
    {
        return data;
    }
    /** Return a reference to the data array this object contains.  
      */
     public byte[] getDataReference()
     {
         return data;
     }
         
    /** return the float array representation of the byte array.
     *  @deprecated
     *  Use getFloats();
     */
    public float[] floatValues()
    {

      //new InternalFaultError(SDBError.Recoverable,"Unimplemented Method");
      float fdata[] = new float[ size() / 4 ];
      for ( int i = 0; i < fdata.length; i++ )
         fdata[ i ] = getFloat( i*4 );
      return fdata;
    }   
    //+++++++++++++++ set<type> methods ++++++++++++++++++++++++++++++++++++++++
    /** Set the specified entry to the specified value after conversion, if needed. 
     * Questionable implementation 
     */
    
    public void setDouble( int index, double val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed.  
     * Questionable implementation 
     */
    public void setFloat( int index, float val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed.  
     * Questionable implementation 
     */
    public void setInt( int index, int val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed.  
     * Questionable implementation 
     */
    public void setShort( int index, short val )
    {
       data[ index ] = (byte) val;
    }
    
    /** Set the specified entry to the specified value after conversion, if needed.  
     * Questionable implementation 
     */
    public void setByte( int index, byte val )
    {
       data[ index ] = val;
    }

    /** Assign the data from the argument Values object to this object.
     *  This object will now have a new data size that is the same as the arguments.  
     *  (This version clones the data; do we want to just reference it?)  
     */
    public void setValues( Values newVals )
    {
       // I don't know if it really is faster to clone than copy, 
       // but this avoids the extra copy that would otherwise happen.
       if ( newVals instanceof MixedValues )
          this.data = ((MixedValues) newVals ).data.clone();  
       else
       {
	       byte newData[] = newVals.getBytes();
	       this.data = new byte[ newVals.size() ];
	       for ( int i = 0; i< newVals.size(); i++ )
	          data[ i ] = newData[ i ];
	   }
    }
    /** Assign elements of the passed in Values object to this one beginning
     *  at the start index . Do conversions as needed.
     *  If the argument size is less than is available beginning at the start 
     *  position, only the remaining values will be changed.
     */
    public void setValues( int start, Values newVals )
    {
       int  copySize = newVals.size();
       if ( copySize > this.size() - start )
          copySize = this.size() - start;
       byte newData[] = newVals.getBytes();
       
       for ( int i = 0; i < copySize; i++ )
          this.data[ i + start ] = newData[ i ];
    }
}
