//
//  StorageType.java
//  DS2Test
//
//  Created by Philip Rhodes on Mon Jan 21 2002.
//  Copyright (c) 2001 Univerity of New Hampshire. All rights reserved.
//
//------------------ Edit history (as of 10/21/05) ------------------------------
// 10/21/05 rdb: added getTypeCode method (to remove warning -- nobody uses it yet)
//-------------------------------------------------------------------------------

package edu.unh.sdb.common;

/** The StorageType class represents the various basic types that are native
 *	to Java.  User defined types are ultimately defined in terms of storage
 *  types, since these are the units that are written and read from disk or
 *  other sources of raw data. 
 */
public class StorageType extends Type
{
    public static final int Undefined=0x00;
    public static final int Boolean=0x01;
    public static final int Byte=0x02;
    public static final int Unsigned_Byte=0x03;
    public static final int UnsignedByte =0x03; 
    public static final int Char=0x04;
    public static final int Short=0x05;
    public static final int Unsigned_Short=0x06;
    public static final int UnsignedShort=0x06;
    public static final int Int=0x07;
    public static final int Float=0x08;
    public static final int Double=0x09;
    public static final int Long=0x0A;
    public static final int Record=0x0B;

    // Note that any valid StorageType must be included in the typeNames
    // array since it is used below to determine counts etc.
    private static String [] typeNames= {"Undefined",     // 0
                                         "boolean",       // 1
                                         "byte",          // 2
                                         "unsigned byte", // 3
                                         "char",          // 4
                                         "short",         // 5
                                         "unsigned short",// 6
                                         "int",           // 7
                                         "float",         // 8
                                         "double",        // 9
                                         "long",          // 10
                                         "record"         // 11
                                         };

    private int typeCode;

    // Package access for now.
    StorageType( String name )
    {
       super( name );
       this.typeCode = typeStringtoInt(name);
    }
    
    // Package access for now.
    StorageType( int typeCode )
    {
        super( intToTypeString(typeCode) );
        this.typeCode = typeCode;
    }
    
    int getTypeCode()
    {
       return this.typeCode;
    }
    
    /** Returns a valid type constant if the String argument is a valid type string.
        Otherwise, returns 0.
        */
    public static int typeStringtoInt( String type )
    {
        for ( int i=1; i<StorageType.typeNames.length; i++)  // 
        {
            if (StorageType.typeNames[i].equals(type))
                return i;
        }
        return 0;
    }

    /** returns a String of the name of the type indicated by the argument. */
    public static final String intToTypeString( int type )
    {
       return name( type );
    }
    
    /** returns a String of the name of the type indicated by the argument. */
    public static String name( int type )
    {
       try {
          return StorageType.typeNames[ type ];
         } catch ( Exception e )
           {
              return "Undefined Type" ;
           }
    }
    
    /** Returns true if the argument is a valid StorageType value, false otherwise.    */
    public static final boolean isStorageType( int type )
    {
        return (type >= 1) && (type <= typeNames.length );
    }

     /** Returns true if the argument is a valid StorageType value, except
         for StorageType.Record. For this value and any non-StorageType value
         this method returns false.
    */
    public static final boolean isSimpleStorageType(int type)
    {
        return ( type >= StorageType.Boolean ) && (type <= StorageType.Long );
    }
    
    /** Return this StorageType represented as a nicely formatted String. */
    public String toString()
    {
        StringBuffer result=new StringBuffer("<< ");
        result.append( this.name );
        result.append(" >>");
        return new String( result );
    }
    
    /** Return the number of distinct StorageTypes. */
    public static final int getNumStorageTypes()
    {
        return StorageType.Record;
    }
    
    /** Return the length in bytes of the "basic" StorageTypes.
     *  The length of StorageType.Record and all other types is returned as 0.
     */
    public static int byteSize( int type )
    {
        switch( type )
        {
           case StorageType.Boolean:
           case StorageType.Byte:
           case StorageType.Char:
           case StorageType.UnsignedByte:
              return 1; 
           case StorageType.Short:
           case StorageType.UnsignedShort:
              return 2; 
           case StorageType.Int:
           case StorageType.Float:
              return 4; 
           case StorageType.Long:
           case StorageType.Double:
              return 8; 
           default:
              return 0;
        }
    }
    
    public static void main(String [] args)
    {
        System.out.println("**********StorageType Tests*********");
        System.out.println();
        System.out.println("The type code for boolean is "+StorageType.typeStringtoInt("boolean"));
        System.out.println("The type code for byte is "+StorageType.typeStringtoInt("byte"));
        System.out.println("The type code for unsigned byte is "+StorageType.typeStringtoInt("unsigned byte"));
        System.out.println("The type code for char is "+StorageType.typeStringtoInt("char"));
        System.out.println("The type code for short is "+StorageType.typeStringtoInt("short"));
        System.out.println("The type code for unsigned short is "+StorageType.typeStringtoInt("unsigned short"));
        System.out.println("The type code for int is "+StorageType.typeStringtoInt("int"));
        System.out.println("The type code for float is "+StorageType.typeStringtoInt("float"));
        System.out.println("The type code for double is "+StorageType.typeStringtoInt("double"));
        System.out.println("The type code for long is "+StorageType.typeStringtoInt("long"));
    }
}
