//
//  TypeTable.java
//  DS2Test
//
//  Created by Philip Rhodes on Sat Jan 19 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
// ---------------------------------------------------
// Edit history (since 06/05/08)
// -----------------------------
// 06/05/08 rdb: makde ArrayList and HashMapgeneric


package edu.unh.sdb.common;

import java.util.*;


/** The TypeTable class keeps track of user defined types in an internal data structure. Types may
 *   be retrieved by name or by typecode, an int greater than zero. Typecodes are assigned
 *   when a new user defined type is inserted into the typetable. 
 */

public class TypeTable {

    /** Returned by insert when an error occurs. */
    public static final int invalidTypeCode=-1;
    
    protected HashMap<String,Integer> nameHash;
    protected ArrayList<Type> numberList;

    protected int nextTypeNumber=1;



    /** Default constructor.*/
    public TypeTable(){

        this.nameHash=new HashMap<String,Integer>(15);
        this.numberList=new ArrayList<Type>(15);
        
        for(int i=StorageType.Boolean;i<=StorageType.Record;i++){
            this.insert_type(new StorageType(i));
        }
    }

    /** This constructor allows an initial capacity to be specified. It should be
        more efficient than the default.
    */
    public TypeTable(int capacity){

        this.nameHash=new HashMap<String,Integer>(capacity+15);
        this.numberList=new ArrayList<Type>(capacity+15);
        
        for(int i=StorageType.Boolean;i<=StorageType.Record;i++){
            this.insert_type(new StorageType(i));
        }

        
        
    }

    /** Adds a user defined type to the typetable. If the insertion is successful,
        this method will return the typeCode assigned to the new type. Otherwise,
        an SDBError is generated, and invalidTypeCode is returned. Failure is most
        likely caused by attempting to insert a type with name that is already being
        used, and that is not structurally equivalent to the type being inserted.
    */
    public int insert(UserType d){
    
        return this.insert_type(d);
    }
    

    /** Adds a Type to the typetable. If the insertion is successful,
        this method will return the typeCode assigned to the new type. Otherwise,
        an SDBError is generated, and invalidTypeCode is returned. Failure is most
        likely caused by attempting to insert a type with name that is already being
        used, and that is not structurally equivalent to the type being inserted.
    */
    private int insert_type(Type d){
        
        String name=d.getName();
            
        if(this.nameHash.containsKey(name)){
            Type existingType=this.get(name);

            if(!existingType.equals(d)) {
                new IllegalArgumentError(SDBError.Recoverable,"There is already a type named \""+
                                          name+"\" in the type table that is not structurally equivalent");
                return TypeTable.invalidTypeCode;
            } else {
                return this.getTypeCode(name);
            }
            
        } else {
            this.numberList.add(d);
            this.nameHash.put(name, new Integer(this.nextTypeNumber));
            
            return this.nextTypeNumber++;
        }
    }

    /** Return the Type corresponding to the name.*/
    public final Type get(String name){

        //added by Gang Lu to fix the case where the defined type is not 
        // in the list
        
        if(this.nameHash.get(name) == null)
          return null;
          
        return this.numberList.get( this.nameHash.get(name).intValue()-1);
    }

    /** Return the Type corresponding to the typeCode.*/
    public final Type get(int typeCode){
    
    // FIXME: check for out of bounds
       // Type returnVal=(Type) this.numberList.get(typeCode-1);

        return this.numberList.get(typeCode-1);
    }

    /** Return the typecode corresponding to the type name.*/
    public final int getTypeCode(String name){

        Integer returnVal=this.nameHash.get(name);
        
        if (returnVal==null){
            return invalidTypeCode;
        } else {
            return returnVal.intValue();
        }
        
    }

    /** Return the type name corresponding to the typecode.*/
    public final String getTypeName(int typeCode){

        return this.numberList.get(typeCode-1).getName();
    }

    /** Return the number of types stored in the TypeTable. */
    public final int size(){

        return this.numberList.size();
    }


    /** Return a representation of this object as a nicely formatted String. This method is meant to be called
        by the toString method of child classes only.*/
    public String toString(){
        
        StringBuffer result=new StringBuffer("TypeTable:\n\ttypecode\ttypename\n");
        int i;
        int size=this.size();
        
        for(i=1;i<=size;i++){
            result.append("\t");
            result.append(i);
            result.append("\t\t\t");
            result.append(this.get(i));
            result.append("\n");
        }
        
        return new String(result);
        
    }
    

    /** For testing and debugging.*/
    public static void main(String args[]){

        TypeTable ttable=new TypeTable(5);
        int pNum,aNum;

        RangedType probability=new RangedType("Probability",StorageType.Float,0,1);
        RangedType luminosity=new RangedType("Luminosity",StorageType.Int,1,10);
        

        pNum=ttable.insert(probability);
        aNum=ttable.insert(luminosity);


        System.out.println("typeCode "+pNum+" corresponds to:"+ttable.get(pNum));
        System.out.println("typeCode "+aNum+" corresponds to:"+ttable.get(aNum));
        System.out.println();
        System.out.println("The name \"Probability\" corresponds to:"+ttable.get("Probability"));
        System.out.println("The name \"Luminosity\" corresponds to:"+ttable.get("Luminosity"));


        RangedType luminosity2=new RangedType("Luminosity",StorageType.Int,1,10);

        int aNum2=ttable.insert(luminosity2);
        System.out.println("After inserting a duplicate type, typeCode "+aNum2+" corresponds to:"+ttable.get(aNum2));

        
        RangedType luminosity3=new RangedType("Luminosity",StorageType.Unsigned_Byte,1,10);
        
        try{
            ttable.insert(luminosity3);
        }catch(SDBException s){
            System.out.println("SDB Exception Properly thrown:\n"+s);
        }
        
        
        
        System.out.println("typeCode "+2+" corresponds to:"+ttable.get(2));
        
        System.out.println();
        
        // Gang Lu's Problem Case
        System.out.println("Gang Lu's Problem Case:");
        
        AttributeDescriptor[] attrs6 = new AttributeDescriptor[2];
        attrs6[0] = new AttributeDescriptor(ttable, "int", "a");
        attrs6[1] = new AttributeDescriptor(ttable, "float", "b");
        RecordDescriptor r6 = new RecordDescriptor("definedRecord", ttable, attrs6);
        ttable.insert(r6);  
        
              
        System.out.println(ttable);                      
        
        AttributeDescriptor[] attrs4 = new AttributeDescriptor[3];
        attrs4[0] = new AttributeDescriptor(ttable, "int", "x");
        attrs4[1] = new AttributeDescriptor(ttable, "float", "y");
        attrs4[2] =  new AttributeDescriptor(ttable, "definedRecord", "z"); 
    
        RecordDescriptor r4 = new RecordDescriptor("foo", ttable, attrs4);
        ttable.insert(r4);

        
        
        
        
        
        
        System.out.println(ttable);
    }
    
}
