/**     Values.java
 * 
 * Abstract class to serve as template for storing an array of values
 * of different types. The elements of the arrays can be retrieved singly
 * as any type -- children must implement the conversions.
 * 
 * -----------------------------------------------------------
 *  Edit history (started 11/27/04)
 * --------------------------------
 *  01/26/05 rdb: Added setValues( int start, Values )
 *  01/24/05 rdb: Changed arg order of spawn( int, int, int ) to match order
 *                  used in get<type>s to be (start, stride, count) instead of
 *                  (start, count, stride).
 *                Also create spawn(start, stride ) that spawns until it gets to end.
 *                   The implementation here works for children.
 *  01/24/05 rdb; Added reOrder( int step )
 *  01/21/05 rdb: Added get<type>s( start, stride, count ) 
 *                  and get<type>s( start, stride ).
 *  01/13/05 rdb: Added set<type> methods and setValues.
 *                Implement get<type> and get<type>s() methods as default 
 *                    implementations that convert. Children will override
 *                    if types match.
 *                Changed size to return int. There isn't enough memory to store
 *                    an float array that is longer than an int!
 *  11/27/04 rdb: Added getByte method.
 * 
 * -----------------------------------------------------------
 */

package edu.unh.sdb.common;

abstract public class Values
{
      public Values(){}
      
      public abstract double     getDouble( int index );
      public abstract float      getFloat(  int index );
      public abstract int        getInt  (  int index );
      public abstract short      getShort(  int index );
      public abstract byte       getByte (  int index );

      public abstract void       setDouble( int index, double val );
      public abstract void       setFloat(  int index, float val );
      public abstract void       setInt  (  int index, int val );
      public abstract void       setShort(  int index, short val );
      public abstract void       setByte (  int index, byte val );
 
      public abstract int        size();
      public abstract int        getStorageType();
      
      /** create a new copy of this Values class, but NO data is copied */
      public abstract Values     spawn();  // new instance of same type and data
      
      /** create a new instance of this Values class; copy a subset of the data to new. */
      public abstract Values     spawn( int start, int stride, int count ); // new instance
                                    // with subset of data
      
      /** Create a new instance of ShortValues with a subsest
       *  of the data in this object. Subsets can be contiguous or
       *  can have a stride as is done in point order storage.
       *  The new instance will copy all data from the existing object
       *  as defined by the stride.
       * @param start - first value to copy
       * @param stride - steps between values
       */
      public Values spawn( int start, int stride )
      {
         int count = 1 + ( this.size() - start - 1 ) / stride;
         return spawn( start, stride, count );
      }

      /** Return a new Values object with the values shuffled.
       *  This method starts at positions 0 through step-1, copying every 
       *  "step" values to the new array. This has the effect of converting 
       *  an attribute ordered array to a point ordered array or vice versa.
       *  The Values object does not actually need to know the axisOrdering,
       *  the conversion is the same. 
       */
      public abstract Values     reorder( int step );
      
      /** Assign the data from the argument Values object to this object.
       *  This object will now have a new data size that is the same as the arguments.  
       *  (This version clones the data; do we want to just reference it?)  
       */
      public abstract void setValues( Values newVal );
      /***************************************************************
      {
         int  sType    = this.getStorageType();
         int  copySize = this.size();
         if ( copySize > newVal.size() )
            copySize = newVal.size();
         for ( int i = 0; i < copySize; i++ )
         {
            double temp = newVal.getDouble( i ); // always retrieve as a double
            switch ( sType )
            {
              case StorageType.Double: setDouble( i, temp ); break;
              case StorageType.Float:  setFloat(  i, (float)temp ); break;
              case StorageType.Int:    setInt  (  i, (int)temp ); break;
              case StorageType.Short:  setShort(  i, (short)temp ); break;
              case StorageType.Byte:   setByte (  i, (byte)temp ); break;
            }
         }
      }
      /***************************************************************/
         
      /** Assign elements of the passed in Values object to this one beginning
       *  at the start index . Do conversions as needed.
       *  If the argument size is less than is available beginning at the start 
       *  position, only the remaining values will be changed.
       */
      abstract public void setValues( int start, Values newVal );
      /****
      {
         int  sType    = this.getStorageType();
         int  copySize = newVal.size();
         if ( copySize > this.size() - start )
            copySize = this.size() - start;
         for ( int i = start; i < start + copySize; i++ )
         {
            double temp = newVal.getDouble( i ); // always retrieve as a double
            switch ( sType )
            {
               case StorageType.Double: setDouble( i, temp ); break;
               case StorageType.Float:  setFloat(  i, (float)temp ); break;
               case StorageType.Int:    setInt  (  i, (int)temp ); break;
               case StorageType.Short:  setShort(  i, (short)temp ); break;
               case StorageType.Byte:   setByte (  i, (byte)temp ); break;
            }
         }
      }
      /**********************************************************/
      //+++++++++++++++++++++++++++ template methods ++++++++++++++++++++++++++++++
      // The following are all template methods that serve as default implementations
      // for these methods. The expectation is that some of these will be overridden
      // by each child.
      
      //--------------------- get<type>s() -------------------------------
      /** Return double array representing all elements in the Values object */
      abstract public  double[]   getDoubles();
      
      /** Return float array representing all elements in the Values object */
      abstract public  float[]   getFloats();
      
      /** Return int array representing all elements in the Values object */
      abstract public  int[]   getInts();
      
      /** Return a short array representing all elements in the Values object */
      abstract public  short[]   getShorts();
      
      /** Return a byte array representing all elements in the Values object */
      abstract public  byte[]   getBytes();

      /** Fill double array representing all elements in the Values object 
       *  If the passed array is not the right size, a 
       *  a correct size array is allocated and returned. 
       *  This is should be overridden by children -- it allocates and copies.
       */
      public  double[]   getDoubles( double[] filler )
      {
         double[] temp = getDoubles();
         if ( filler == null || filler.length != temp.length )
            return temp;
         else
            for ( int i = 0; i < temp.length; i++ )
               filler[ i ] = temp[ i ];
         return filler;
      }
      
      /** Fill float array representing all elements in the Values object 
       *  If the passed array is not the right size, a 
       *  a correct size array is allocated and returned. 
       *  This is should be overridden by children -- it allocates and copies.
       */
      public  float[]   getFloats( float[] filler )
      {
         float[] temp = getFloats();
         if ( filler == null || filler.length != temp.length )
            return temp;
         else
            for ( int i = 0; i < temp.length; i++ )
               filler[ i ] = temp[ i ];
         return filler;
      }
      
      /** Fill int array representing all elements in the Values object 
       *  If the passed array is not the right size, a 
       *  a correct size array is allocated and returned. 
       *  This is should be overridden by children -- it allocates and copies.
       */
      public  int[]   getInts( int[] filler )
      {
         int[] temp = getInts();
         if ( filler == null || filler.length != temp.length )
            return temp;
         else
            for ( int i = 0; i < temp.length; i++ )
               filler[ i ] = temp[ i ];
         return filler;
      }
      
      /** Fill short array representing all elements in the Values object 
       *  If the passed array is not the right size, an warning is issued and 
       *  a correct size array is allocated and returned. 
       *  This is should be overridden by children -- it allocates and copies.
       */
      public  short[]   getShorts( short[] filler )
      {
         short[] temp = getShorts();
         if ( filler == null || filler.length != temp.length )
            return temp;
         else
            for ( int i = 0; i < temp.length; i++ )
               filler[ i ] = temp[ i ];
         return filler;
      }
      
      /** Fill byte array representing all elements in the Values object 
       *  If the passed array is not the right size, a 
       *  a correct size array is allocated and returned. 
       *  This is should be overridden by children -- it allocates and copies.
       */
      public  byte[]   getBytes( byte[] filler )
      {
         byte[] temp = getBytes();
         if ( filler == null || filler.length != temp.length )
            return temp;
         else
            for ( int i = 0; i < temp.length; i++ )
               filler[ i ] = temp[ i ];
         return filler;
      }
      
      //------------------get<type>s( start, stride )-----------------------
      
      /** Return double array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until we
       *  get to the end of the array.
       */
      public double[] getDoubles( int start, int stride )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getDoubles( start, stride, num );
      }    

      /**
       * Reference version of getDoubles( start, stride )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  double[]   getDoubles( int start, int stride, double[] fill )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getDoubles( start, stride, num, fill );
      }

      /** Return float array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until we
       *  get to the end of the array.
       */
      public float[] getFloats( int start, int stride )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getFloats( start, stride, num );
      }
      
      /**
       * Reference version of getFloats( start, stride )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  float[]   getFloats( int start, int stride, float[] fill )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getFloats( start, stride, num, fill );
      }

      /** Return int array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until we
       *  get to the end of the array.
       */
      public int[] getInts( int start, int stride )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getInts( start, stride, num );
      }    

      /**
       * Reference version of getInts( start, stride )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  int[]   getInts( int start, int stride, int[] fill )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getInts( start, stride, num, fill );
      }

      /** Return short array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until we
       *  get to the end of the array.
       */
      public short[] getShorts( int start, int stride )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getShorts( start, stride, num );
      }    
      
      /**
       * Reference version of getShorts( start, stride )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  short[]   getShorts( int start, int stride, short[] fill )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getShorts( start, stride, num, fill );
      }

      /** Return byte array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until we
       *  get to the end of the array.
       */
      public byte[] getBytes( int start, int stride )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getBytes( start, stride, num );
      }
      
      /**
       * Reference version of getBytes( start, stride )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  byte[]   getBytes( int start, int stride, byte[] fill )
      {
         int num = 1 + ( this.size() - start - 1 ) / stride;
         return getBytes( start, stride, num, fill );
      }

      //------------------get<type>s( start, stride, len )-----------------------
      // extract a slice of the Values -- works for either point-order or attribute order. 
      // These are default implementations. Children should especially override 
      // the one that matches their type.
      //
      /** Return double array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      // Coding explanation for all these methods:
      //   In order to find the largest number of entries to return, given
      //   the array size and the start and stride values, here is what the
      //   code is based on:
      //     As long as the start is not past the end of the array, we
      //     certainly get 1 value to return. After the start position, there
      //     are still (size - (start+1)) entries left in the array. If we
      //     divide this number by the stride, we know how many additional
      //     steps we can take in the array. So the total number of entries
      //     we can copy is 
      //               1 + ( size - start - 1 ) / stride
      //
      public  double[]   getDoubles( int start, int stride, int num )
      {
         return getDoubles( start, stride, num, null );
      }
      
      /** Return floats array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      public float[] getFloats( int start, int stride, int num )
      {
         return getFloats( start, stride, num, null );
      }
      
      /** Return int array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      public int[] getInts( int start, int stride, int num )
      {
         return getInts( start, stride, num, null );
      }
      
      /** Return short array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      public short[] getShorts( int start, int num, int stride )
      {
         return getShorts( start, stride, num, null );
      }
      
      /** Return byte array representing a subset of elements in the Values object
       *  Copying starts at the index specified by the start argument, 
       *  the 2nd item copied will be at index start + stride where stride is the 
       *  second argument; the 3rd is at start + 2*stride and so forth until 
       *  the specified number of values have been copied (3rd argument). 
       *  If there are not enough elements in the Values object to satisfy the 
       *  request, the returned array will have only as many values as could be
       *  found given the start and stride arguments. 
       */
      public byte[] getBytes( int start, int stride, int num )
      {
         return getBytes( start, stride, num, null );
      }      

      /**
       * Reference version of getDoubles( start, stride, num )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  double[]   getDoubles( int start, int stride, int num, double[] fill )
      {
         if ( start >= this.size() )
            num = 0;
         else if ( start + ( num-1 )*stride >= this.size() )
            num = 1 + ( this.size() - start -1 ) / stride; // find all that exist.
         if  ( fill == null || fill.length != num )
            fill = new double[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            fill[ i ] = getDouble( offset );
         return fill;
      }

      /**
       * Reference version of getFloats( start, stride, num )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  float[]   getFloats( int start, int stride, int num, float[] fill )
      {
         if ( start >= this.size() )
            num = 0;
         else if ( start + ( num-1 )*stride >= this.size() )
            num = 1 + ( this.size() - start -1 ) / stride; // find all that exist.
         if  ( fill == null || fill.length != num )
            fill = new float[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            fill[ i ] = getFloat( offset );
         return fill;
      }

      /**
       * Reference version of getInts( start, stride, num )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  int[]   getInts( int start, int stride, int num, int[] fill )
      {
         if ( start >= this.size() )
            num = 0;
         else if ( start + ( num-1 )*stride >= this.size() )
            num = 1 + ( this.size() - start -1 ) / stride; // find all that exist.
         if  ( fill ==null || fill.length != num )
            fill = new int[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            fill[ i ] = getInt( offset );
         return fill;
      }

      /**
       * Reference version of getShorts( start, stride, num )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  short[]   getShorts( int start, int stride, int num, short[] fill )
      {
         if ( start >= this.size() )
            num = 0;
         else if ( start + ( num-1 )*stride >= this.size() )
            num = 1 + ( this.size() - start -1 ) / stride; // find all that exist.
         if  ( fill == null || fill.length != num )
            fill = new short[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            fill[ i ] = getShort( offset );
         return fill;
      }

      /**
       * Reference version of getBytes( start, stride, num )
       * The passed array is returned for convenience, but if it is not the right
       * size, a new array is allocated and returned. 
       */
      public  byte[]   getBytes( int start, int stride, int num, byte[] fill )
      {
         if ( start >= this.size() )
            num = 0;
         else if ( start + ( num-1 )*stride >= this.size() )
            num = 1 + ( this.size() - start -1 ) / stride; // find all that exist.
         if  ( fill == null || fill.length != num )
            fill = new byte[ num ];
         int offset = start;
         for ( int i = 0; i < num; i++, offset += stride )
            fill[ i ] = getByte( offset );
         return fill;
      }
      
      //-------------- main -------------------------------------
      public static void main( String[] args )
      {
         // really need to have some tests here
      }
}