/*
    ARRSInternalNode.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/
 
/* ---------------------- Edit history as of 12/10/05 -----------------------
 * 06/05/08 rdb: Made ArrayLists generic
 * 12/12/05 rdb: Added getMaxChildren abstract method.
 *               Major revision to signature and code of addNode and renamed to
 *                  addChild
 *               Changed getEldest and setEldest to getFirstChildID and setFirstChildID
 *                  and stored as an short instead of int
 *               Added lastChildID and getLastChildID, setLastChildID methods
 *               Added numChildren field and getNumChildren()
 *              
 * 12/10/05 rdb: Added copy constructor
 */
/**
 * The ARRSInternalNode class develop an ARRS internal node structure 
 * based on the concepts of the ARRS node in a ARRS tree. It is extended 
 * from the ARRSNode.
 */

package edu.unh.sdb.datasource;

import java.util.*;
//import java.io.*;

import edu.unh.sdb.common.*;

abstract public class ARRSInternalNode extends ARRSNode 
{
   protected int firstChildId;
   protected int lastChildId;
   protected short numChildren;

   /** Constructor without argument */
   public ARRSInternalNode()
   {
      this( (ISBounds)null, (CellIndexable) null );
   }
  
   /** Constructor takes a single ISBounds */
   public ARRSInternalNode( ISBounds bounds )
   {
      this( bounds, (CellIndexable) null );
   }
 
   /** Constructor takes a ISBounds and a CellIndexable */
   public ARRSInternalNode( ISBounds bounds, CellIndexable cellSource )
   {
      super( bounds, cellSource );
      this.firstChildId = -1;
      this.lastChildId = -1;
      this.numChildren = 0;
   }
 
   /** Copy constructor  */
   public ARRSInternalNode( ARRSInternalNode src )
   {
      super( src );
      this.firstChildId = -1;
   }
   
   /** return maximum allowed number of children, determined by child class */
   abstract public int maxChildren();
       
   /** Set first child's id */
   public void setFirstChildID( int eldest )
   {
      this.firstChildId = eldest;
   }

   /** Get the first child's id */
   public int getFirstChildID()
   {
      return this.firstChildId;
   }
   
   /** Set last child's id */
   public void setLastChildID( int youngest )
   {
      this.firstChildId = youngest;
   }

   /** get number of children */
   public int getNumChildren()
   {
      return this.numChildren;
   }

   /** Get the last child's id */
   public int getLastChildID()
   {
      return this.firstChildId;
   }
   
   /** Add a node as a child of this node. */
   public boolean addChild( ArrayList treeNodes, ARRSNode node )
   {
      int         numKids = this.getNumChildren();         // # siblings
      int         newId = node.getID();
      boolean     success = true;
      
      node.setSibling( this.nodeId ); // The last child points back to parent
      
      //No child case
      if ( numKids >= this.maxChildren() ) 
      {
         new SDBError( SDBError.Warning, "Tried to add child to full parent node." );
         success = false;
      }
      else if ( numKids == 0 )
         setFirstChildID( newId );
      else
      {
         ARRSNode olderSibling = (ARRSNode)( treeNodes.get( this.lastChildId )); 
         olderSibling.setSibling( newId);
      }
      if ( success )
      {
         this.lastChildId = newId;;
         this.numChildren++;
      }
      return success;
   }


   /** Add a node as a child of this node. */
   public boolean addNode( ArrayList<ARRSNode> treeNodes, ARRSNode node, 
                           int parentId, int id, int max_child_num ) 
   {
      //rdb: Major recoding. Old code passed in parentId, and looked up parent node
      //     of "node", but "this" is the parent and it knows its own id. 
      //     So, parentId is a spurious argument and there is no need to search for
      //     the parent node.
      //     max_child_num is also spurious. It is inherent in the type of "this"
      //        node (and the dimensionality); I added this information to the 
      //        ARRSInternalNode class.
      int index;
      ARRSNode[]       siblings;        // existing children of parent
      ARRSNode         olderSibling;    // last added sibling
      int              numSibs;         // # siblings

      
      //No child case
      if ( getFirstChildID() == -1 ) 
      {
         setFirstChildID( id );
         node.setID( id );
         treeNodes.set( parentId, this );
         treeNodes.add( node );
         return true;
      }

      //at least one child case. get all the children of the parent
      siblings = this.getChildren( treeNodes );
      numSibs  = siblings.length;

      //chilren positions are full and fail to add the new node into the tree
      if ( numSibs >= max_child_num )
         return false;

      //set some values for the new node
      node.setID( id );
      if ( numSibs ==  max_child_num - 1 ) 
         node.setSibling( parentId );

      //set the sibling value for the last child
      index = siblings[ numSibs - 1 ].getID();
      olderSibling = treeNodes.get( index );
      olderSibling.setSibling( id );
      treeNodes.set( index, olderSibling );

      //add the new node into the tree
      treeNodes.add( node );

      return true;
   }


   /** Get the specified child */
   public ARRSNode getChild( ArrayList tree, int i )
   {
      int num = 0, id;
      ARRSNode temp;

      id = this.firstChildId;

      //no children
      if ( id == -1 )
         return null;

	  if ( id >= tree.size() )
		 return null;
      //first child
      temp = (ARRSNode)tree.get( id );
      while ( true ) 
      {
         if ( num == i )
            return temp;
         //other children
         else 
         {
            num++;
            //sibling points to parent or the specified child does not exit
            if ( temp.getSiblingID() < id )
               return null;
            id = temp.getSiblingID();
            temp = (ARRSNode)tree.get( id );
         }
      }
   }


   /** Get all the children */
   public ARRSNode[] getChildren( ArrayList<ARRSNode> tree )
   {
      int num, id;
      ARRSNode temp;
      ArrayList<ARRSNode> children = new ArrayList<ARRSNode>();

      id = this.firstChildId;
      //no children
      if ( id == -1 )
         return null;

      //first child
      temp = tree.get( id );
      while ( true ) 
      {
         children.add( temp );
         //sibling points to parent or the specified child does not exit
         if ( temp.getSiblingID() < id )
            break;
         id = temp.getSiblingID();
         temp = tree.get( id );
      }

      if ( children.isEmpty() ) 
      {
         System.out.println( "no children" );  
         return null;
      }

      num = children.size();
      ARRSNode[] child = new ARRSNode[ num ];
      for ( int i = 0; i < num; i++) 
      {
         child[ i ] = children.get( i );
      }
      return child;
   }
  
    /** Get the id of the specified child */
   public int getChildID( ArrayList tree, int i )
   {
      ARRSNode node;
      node = this.getChild( tree, i );

      //node does not exit
      if ( node == null )
         return -1;
 
      return ( this.getChild( tree, i) ).getID();
   }
}
