/************************************************************ 
 AttributeOrderBFDS.java
 Copyright 2000 University of New Hampshire
 Durham, NH 03824

 ------------- Edit history (started 11/28/04 ------------------
 12/06/05 rdb: Fixed bugs that have been around since the major refactoring or earlier.
 12/04/05 rdb: Refactor to move public subblock methods to parent and implement
                  at this level only subBasicBlock and subCompositeBlock
               Added checkActivated calls.
 10/21/05 rdb: Deleted unused method, setByteArray and toBinaryString, if needed 
                  should use parent version
               Deleted lots of code previously commented out.
 06/01/05 rdb: Replaced direct file reads for 'datum' methods with read into
                 ByteBuffer that is pre-allocated to read 1 value at a time.
                 Very inefficient, but was worse before and we didn't handle
                 endian conversion correctly.
 04/23/05 rdb: check return code from allocateBuffer, Handle multiple
               reads per rod.
 01/30/05 rdb: Deleted subblock( ISBounds ) - parent version does it
 01/25/05 rdb: null allowed for RecordSpec and FieldIDMapper args to subblock
 01/06/05 rdb: create Datum objects from RecordDescriptor
 11/28/04 rdb: Replaced references to ByteValues with MixedValues
 
 ---------------------------------------------------------------
 Xuan Tang xtang@cs.unh.edu
 last update: 10/30/03

 ByteOrder endian

 updata comments: attribute axisOrdering implementation
 
 1. require a new method from RecordDescriptor attributeByteSize(int attributeID)
 ************************************************************/

/*------------------- Edit history as of 02/06/05 --------------------------
 * 02/06/05 rdb: Changed ByteBuffer allocation to use new parent allocateBuffer methods.
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.io.*;
import java.nio.*;
//import java.nio.channels.*;

/**
 * This class will read from a file of binary data, providing an n-dimensional
 * datasource view of that file. this class represents a BinaryFileDataSource
 * that is in attribute order
 */

public class AttributeOrderBFDS extends BinaryFileDataSource
{
   private ByteBuffer datumBuffer = null; // buffer for reading a datum. It's not
                                          // the size of a Datum, it's the size of
                                          // a double. largest field that can read.
   private int[] fullBufferSize;    // portion of buffer usable for normal size reads
                                // this could be less than the buffer capacity
                                // since we want the buffer to hold even multiples of
                                // data -- also we may be re-using a buffer that is
                                // bigger than we want for this case.
   private int[] lastBufferSize; // If a rod does not fit into the
                             // I/O buffer, it will take multiple reads to get the
                             // rod data. The last read need not be a full buffer 
                             // Need to know this limit for the last buffer read.
   private int[]        readsPerRod = null;  // # IO buffer reads to fill a rod
   
   /**
    * create a AttributeOrderBFDS , set the initial file offset at 0
    */
   public AttributeOrderBFDS( String name, RecordDescriptor recordDesc,
         ISBounds theBounds, AxisOrdering ordering, String URL, ByteOrder endian )
   {
      this( name, recordDesc, theBounds, ordering, URL, endian, 0 ); 
   }

   /**
    * create a AttributeOrderBFDS , set the initi file offset at the given value
    */
   public AttributeOrderBFDS( String name, RecordDescriptor recordDesc,
         ISBounds theBounds, AxisOrdering ordering, String URL,
         ByteOrder endian, int offset )
   {
      super( name, recordDesc, theBounds, ordering, URL, endian, offset ); // call
                                                                           // inherited
                                                                           // constructor
      this.dataOrder = DataCollection.attributeOrder;
      
      // we're going to have 1 buffer for each attribute; we'll try to read
      // the attributes as if they are in 3 separate files.
      int numAttr = this.recordDesc.getNumAttributes();
      initializeBuffers( numAttr );

      this.lastBufferSize = new int[ numAttr ];
      this.fullBufferSize = new int[ numAttr ];
      this.readsPerRod    = new int[ numAttr ];
      
      for ( int i = 0; i < lastBufferSize.length; i++ )
      {
         this.lastBufferSize[ i ] = 0;
         this.fullBufferSize[ i ] = 0;
         this.readsPerRod[ i ] = 1;
      }
   }

   /** Return the Datum corresponding to the given IndexSpaceID. */
   public Datum datum( IndexSpaceID did )
   {
      Datum returnVal = this.recordDesc.createDatum();

      this.datum( returnVal, did );
      return returnVal;
   }
   
   /** 
    * Read bytes for a value from the file into the ByteBuffer
    */
   private void readBytes( IndexSpaceID isid, int nBytes, int offset )
   {
      long fileOffset;
      if ( datumBuffer == null )
      {
         this.datumBuffer = ByteBuffer.allocateDirect( 
                                       this.recordDesc.getMaxAttributeByteSize() );
         this.datumBuffer.order( this.endian );       
      }
  	 try
	 {
	    //Position file stream
        fileOffset = AxisOrdering.computeOffset( isid, this.axisWeights )
                                   * nBytes + ( offset * this.volume() )
                                              + this.iniOffset;

        if ( fileOffset != ( (int) raFile.getFilePointer() ) )
           raFile.seek( fileOffset );
      
	    this.datumBuffer.clear();
	    fChannel.read( this.datumBuffer );
	   
	    this.datumBuffer.position( 0 ); // PJR
	 } 
	 catch ( IOException e )
	 {
	    new FileError( SDBError.Recoverable,
	         "BinaryFileDataSource.datum(Datum, IndexSpaceID):: IOException: " + e );
	 }      
   }

   /** Load the data at the given index into the provided datum. */
   public void datum( Datum datum, IndexSpaceID did )
   {
      checkActivated(); 
      int[] storageType = recordDesc.getStorageTypes();
      int[] byteSize = recordDesc.attributeByteSizes();
      int[] attributeOffset = recordDesc.byteOffsets();

      if ( datum.size() != this.recordNumFields )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Datum is not the right size for this datasource." );
         return;
      }
      for ( int i = 0; i < this.recordNumFields; i++ )
      {
         readBytes( did, byteSize[i], attributeOffset[i] );
         setDatum( i, storageType[i], datum, this.datumBuffer );
      }
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding RecordSpec.
    */
   public void datum( Datum datum, IndexSpaceID did, RecordSpec recSpec )
   {
      checkActivated(); 
      int[] storageType = recordDesc.getStorageTypes();
      int[] byteSize = recordDesc.attributeByteSizes();
      int[] attributeOffset = recordDesc.byteOffsets();
      int[] fieldNumbers = recSpec.getFieldNumbers();
      
      if ( datum.size() != recSpec.size() )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Datum is not the right size for this request." );
         return;
      }

      for ( int i = 0; i < recSpec.size(); i++ )
      {
         readBytes( did, byteSize[ fieldNumbers[i] ], 
                         attributeOffset[ fieldNumbers[i] ] );
         setDatum( fieldNumbers[i], storageType[ fieldNumbers[i] ], 
                        datum, this.datumBuffer );
      }
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding FieldIDMapper.
    */
   public void datum( Datum datum, IndexSpaceID did, FieldIDMapper mp )
   {
      checkActivated(); 
      int[] storageType = recordDesc.getStorageTypes();
      int[] byteSize = recordDesc.attributeByteSizes();
      int[] attributeOffset = recordDesc.byteOffsets();
      int[] componentFields = mp.getComponentFields();
      int[] compositeFields = mp.getCompositeFields();
      
      if ( datum.size() != mp.size() )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Datum is not the right size for this request." );
         return;
      }

      for ( int i = 0; i < mp.size(); i++ )
      {
         readBytes( did, byteSize[ componentFields[i] ], 
                         attributeOffset[ componentFields[i] ] );
         setDatum( compositeFields[i], storageType[ componentFields[i] ], 
                        datum, this.datumBuffer );
      }
   }

   /**
    * Return the Datum corresponding to the given IndexSpaceID. The Datum will
    * only contain attributes specified by the RecordSpec argument.
    */
   public Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      int specsize = spec.size();
      Datum returnVal = this.recordDesc.createDatum( spec );

      this.datum( tempDatum, did );

      for ( int i = 0; i < specsize; i++ )
      {
         returnVal.setFloat( i, tempDatum.getFloat( spec.getFieldNumber( i ) ) );
      }
      return returnVal;
   }

   /**
    * allocate an I/O buffer for i-th attribute. If the system cannot allocate a
    * buffer large enough for a rod, we will need multiple reads. The last
    * read need not be a full buffer. If this happens we allocate an extra
    * buffer for that last read.
    */
   private void allocateIOBuffer( int i, int unitByteSize, int rodSize )
   {
      long neededSize = unitByteSize * rodSize;
      
      if ( neededSize > Integer.MAX_VALUE )
         allocateBuffer( i, Integer.MAX_VALUE );
      else
         allocateBuffer( i, (int) neededSize );
      
      int  gotSize = getBufferSize( i );
      if ( gotSize == neededSize )
      {
         this.lastBufferSize[ i ] = gotSize;
         this.readsPerRod[ i ] = 1;
      }
      else
      {
         if ( gotSize % unitByteSize != 0 ) // be sure size is mult of unit size
            this.fullBufferSize[ i ] = gotSize / unitByteSize * unitByteSize;
         gotSize = this.fullBufferSize[ i ];
         
         this.readsPerRod[ i ] = (int) ( neededSize / gotSize );
         long lastSize = neededSize - gotSize;
         if ( lastSize > 0 )
         {
            this.readsPerRod[ i ]++;
            this.lastBufferSize[ i ] = (int) lastSize;
         }
      }     
   }
   /**
    * subBasicBlock
    * 
    * @param dstBlock
    * @param dstBounds
    * @param srcBounds
    * @param mp
    */
   // basicBlock implementation java 1.4
   // dstBlock is considered as composite, the calling binaryFileDataSource is
   // considered as component
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                               ISBounds srcBounds, FieldIDMapper mp )
   {
      int[] intData = null;
      float[] floatData = null;
      double[] doubleData = null;
      short[] shortData = null;
      byte[] byteData = null;

      int dataArrayOffset;
      long fileOffset;
      int dstIndex;
      int compositeNumFields = dstBlock.getRecordDescriptor().getNumAttributes();
      int numFields = mp.size();

      // a rod is a sequence of contiguous file data
      // rodSize: # data units in a rod
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis );
      
      ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
      int[] componentFields = mp.getComponentFields();
      int[] compositeFields = mp.getCompositeFields();

      int[] attributeByteSizes = this.recordDesc.attributeByteSizes();
      int[] attributeOffsets = this.recordDesc.byteOffsets();

      int dataType = dstBlock.getStorageType();

      // get the data array of dstblock
      if ( dataType == StorageType.Int )
         intData = ( (IntValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Float )
         floatData = ( (FloatValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Short
            || dataType == StorageType.Unsigned_Short )
         shortData = ( (ShortValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Double )
         doubleData = ( (DoubleValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Record )
         byteData = ( (MixedValues) dstBlock.getValues() ).getDataReference();
      else // the data is boolean, char, byte or unsigned byte
         byteData = ( (ByteValues) dstBlock.getValues() ).getDataReference();

      // allocate all buffers
      for ( int i = 0; i < numFields; i++ )
         allocateIOBuffer( i, attributeByteSizes[componentFields[ i ]], rodSize );
      
      try
      {
         componentFields = mp.getComponentFields();
         compositeFields = mp.getCompositeFields();

         for ( iter.init(); iter.valid(); iter.next() )
         {
            for ( int attr = 0; attr < numFields; attr++ )
            {
               fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                     * attributeByteSizes[componentFields[ attr ]]
                     + ( attributeOffsets[componentFields[ attr ]] * this.volume() )
                     + iniOffset;
               iterCopy.assign( iter );
               dstIndex = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy ) );
               dataArrayOffset = compositeNumFields * dstIndex + compositeFields[ attr ];
 
               if ( fileOffset != ( (int) raFile.getFilePointer() ) )
                  raFile.seek( fileOffset );
               
               //rdb for ( int b = 0; b < this.readsPerRod[ b ]; b++ )
               for ( int b = 0; b < this.readsPerRod[ attr ]; b++ )
               {    
                  ByteBuffer iobuf = getBuffer( attr );
                  iobuf.clear();
                  if ( b < this.readsPerRod[ attr ] - 1 )
                     iobuf.limit( this.fullBufferSize[ attr ]);
                  else 
                     iobuf.limit( lastBufferSize[ attr ] );
                  //fChannel.read(byteBuffer, 0, rodSize*
                  // attributeByteSizes[componentFields[i]] );
                  this.fChannel.read( iobuf );
                  
                  iobuf.position( 0 );
                  
                  while( iobuf.position() < iobuf.limit() )
                  {
                     switch ( dataType )
                     {
                        case StorageType.Char:
                        case StorageType.Unsigned_Byte:
                        case StorageType.Byte:
                           byteData[dataArrayOffset] = iobuf.get();
                           break;
                        case StorageType.Int:
                           intData[dataArrayOffset] = iobuf.getInt();
                           break;
                        case StorageType.Short:
                           shortData[dataArrayOffset] = iobuf.getShort();
                           break;
                        case StorageType.Double:
                           doubleData[dataArrayOffset] = iobuf.getDouble();
                           break;
                        case StorageType.Float:
                           floatData[dataArrayOffset] = iobuf.getFloat();
                           break;
                        case StorageType.Record:
                           new InternalFaultError( SDBError.Recoverable,
                                           "Records not implemented." );
                           break;
                        default:
                           new InternalFaultError(SDBError.Recoverable,
                                         "Unsupported StorageType: " + dataType );
                           break;
                     } // end of switch
                     dataArrayOffset += compositeNumFields;
                 }                  
               }
            }
         }
      } 
      catch ( IOException e )
      {
         new FileError( SDBError.Recoverable, "IO Exception in " +
         		"BinaryFileDataSource.subblock(DB, ISB, ISB,FieldIDMapper)::" + e );
      }
   }

   // compositeBlock implementation
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
                                    ISBounds inputSrcBounds, FieldIDMapper mp )
   {
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      
      int dstIndex;
      long fileOffset;
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 

      //System.out.println("BinaryFileDatasource rode size :" + rodSize );
      int datumArrayStep = (int) dstBlock.getAxisWeight( this.leastSignificantAxis );
      
      if ( mp == null )
         mp = new FieldIDMapper( dstBlock.getNumAttributes());

      int numFields = mp.size();

      int[] componentFields = mp.getComponentFields();

      int attributeByteSizes[] = this.recordDesc.attributeByteSizes();
      int attributeOffsets[] = this.recordDesc.byteOffsets();
      int storageTypes[] = this.recordDesc.getStorageTypes();

      ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );

      for ( int i = 0; i < numFields; i++ )
      {
         allocateIOBuffer( i, attributeByteSizes[ componentFields[i]], rodSize );
      }

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {
            for ( int attr = 0; attr < numFields; attr++ )
            {
               fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                     * attributeByteSizes[componentFields[ attr ]]
                     + ( attributeOffsets[componentFields[ attr ]] * this.volume() )
                     + iniOffset;

               iterCopy.assign( iter );
               dstIndex = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy ));
               if ( fileOffset != ( (int) raFile.getFilePointer() ) )
                  raFile.seek( fileOffset );
              
               int thisDstAttr = componentFields[ attr ];
               int thisByteSize = attributeByteSizes[ thisDstAttr ];
               int thisType = storageTypes[ thisDstAttr ];
               
               int bufUnitSize = getBuffer( 0 ).limit() / thisByteSize;
               for ( int b = 0; b < this.readsPerRod[ b ]; b++ )
               {    
                  ByteBuffer iobuf = getBuffer( attr );
                  iobuf.clear();

                  if ( b < this.readsPerRod[ b ] - 1  )
                     iobuf.limit( fullBufferSize[ attr ] );
                  {
                     iobuf.limit( lastBufferSize[ attr ] );
                     bufUnitSize = lastBufferSize[ attr ] / thisByteSize;
                  }
                  fChannel.read( iobuf );
                  iobuf.position( 0 );
                  
                  dstBlock.setAttributeOrderRod( iobuf, bufUnitSize, dstIndex, 
                                                datumArrayStep, thisDstAttr,                                         
                                                thisByteSize, thisType );
                  dstIndex += datumArrayStep * bufUnitSize; // update dstIndex by # read                
               }
               /**********************************************************/
            } 
         }
      } catch ( IOException e )
      {
         new FileError(
               SDBError.Recoverable,
               "IO Exception in BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }

   }

   // this is based on a new java 1.4
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                               ISBounds srcBounds, RecordSpec recSpec )
   {
      subBasicBlock( dstBlock, dstBounds, srcBounds, new FieldIDMapper( recSpec ));
      
   }

   // composite implementation , assume dstBlock is point order
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
         ISBounds inputSrcBounds, RecordSpec recSpec )
   {
      subCompositeBlock( dstBlock, dstBounds, inputSrcBounds, 
                                                     new FieldIDMapper( recSpec ));
      
   }

   /**
    * this is an implementation for BasicBlock based on java 1.4 Note: The
    * BasicBlock is assumed to be created and a valid data array is already
    * allocated. attribute Order, basicBlock with byte data array for mixed
    * type is not supported at this moment. will work on it
    */
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
      subBasicBlock( dstBlock, dstBounds, srcBounds, 
                                    new RecordSpec( dstBlock.getNumAttributes() ));
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. Note that the DataBlock is assumed to
    * already have a valid data array. NOTE: DataBlock must be a CompositeBlock,
    * and assume dstBlock is point order
    */
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
         ISBounds inputSrcBounds )
   {
      subCompositeBlock( dstBlock, dstBounds, inputSrcBounds, 
                                    new FieldIDMapper( dstBlock.getNumAttributes()));
      
   }


   /** Print the data members of the DataSource in an easy to read fashion. */
   public String toString()
   {

      StringBuffer result = new StringBuffer(
            "Attribute Order BinaryFileDataSource: raFile=" );

      result.append( raFile );
      result.append( "\n" );

      if ( ( this.endian ).equals( ByteOrder.BIG_ENDIAN ) )
      {
         result.append( "            format=big-endian\n" );
      }
      else
      {
         result.append( "            format=little-endian\n" );
      }

      result.append( super.toString() );

      return new String( result );
   }

   /**
    * Returns true if this BinaryFileDataSource has the same pointOrder as the
    * argument BinaryFileDataSource, as well as meeting inherited criteria for
    * equality.
    */
   public boolean equals( Object o )
   {
      if ( o instanceof AttributeOrderBFDS )
      {
         AttributeOrderBFDS d = (AttributeOrderBFDS) o;
         return ( this.endian ).equals( d.endian ) && super.equals( d );
      }
      else
         return false;
   }
}