/************************************************************ 
   BasicBlock.java
   Copyright 2000 University of New Hampshire
   Durham, NH 03824

   Xuan Tang xtang@cs.unh.edu
   
   ----------------- Edit History (as of 11/27/04 ----------------------------
   12/07/05 rdb: Refactored all the getIndexed<type>Attribute methods to be package
                   access and to remove the "Attribute" suffix.
   11/13/05 rdb: -Found a bug in subblock: it didn't work if destination block
                  does not have its lower left corner at origin. Now fixed.
                 -Consolidated the RecordSpec and FieldIDMapper versions of subblock
                  into a single method -- reduces replicated code. Not sure that
                  the code works "correctly" in all cases, but the old code didn't
                  either!
                 -Deleted lots of deprecated code that is no longer used anywhere.
                 -made copyArray package access and extended it to work for all types
                 
      ---------->  However, I couldn't figure out how to force it to get executed?
                   It's only called by BlockListCache.lookup, probably for multi-block
                   caches. It does not get called by slicerJOGL and the test code
                   in BlockListCache blows up earlier on even very simple data??
                  
   10/25/05 dlipsa: used recordDesc instead of getRecordDescription() (which 
                    does a clone) in datum()
   10/21/05 rdb: removed unused variables
                 commented out unused private method (attributeMatch)
   04/04/05 pjr: deleted setDatum( ISIterator iter, Datum datum, IntegerIndexSpaceID scratchID )
                  in favor of changes in ISIterator that allow the same functionality.
   04/03/05 pjr: added setDatum( ISIterator iter, Datum datum, IntegerIndexSpaceID scratchID )
   01/31/05 rdb: deleted get<type>s methods that are done in parents.
   01/27/05 rdb: changed subblock methods with RecordSpec or FieldIDMapper arguments
                    so that null is valid.
   01/25/05 rdb: changed Values.spawn( int, int, int ) to reflect new param order.
                   (actually replaced the call to the new spawn( start, stride ) version.
                 Moved DatBlock subblock( ISBounds ) implementation to DataCollection.
   01/19/05 rdb; deprecated get<type>Attributes( float[], recordSpec, pointOrder )
                    replaced with get<type>s( ... )
                    for Double, Float, Int, Short, Byte
                 deprecated <type>Values() methods
                    replaced with get<type>s(). intValues() -> getInts() etc.
                    This included a semantics change. Previously null was returned
                    if the type was not a "proper" match. Now, a conversion is done,
                    via the newly changed Values class hierarchy.
                 
   01/13/05 rdb:
                - revised handling of StorageType; using new RecordDescriptor interface
                  for that.
                - revised order of parameters of constructors to follow rule
                  that all required arguments have same position and order in all
                  versions. This was not true of the RecordDescriptor argument.
                  Deprecated versions that did not comply.
                - Made constructors that have the Values argument package visible --
                  I don't think they should be public -- I think we should avoid
                  having users be aware of the Values class. If we want that kind
                  of visibility, we should use the Java DataBuffer class.
                - revised constructor code so that code is shared whenever possible.
                - made subBasicBlock package access -- not sure why it's even there.
                  so put the implementation in subblock.
   12/14/04 rdb: changed size() calls to volume() calls.
   12/05/04 rdb: move size method to DataCollection
   11/27/04 rdb: Added getFieldPos private methods to localize the code that finds 
                    data positions in the 1D arrays. This code was replicated in 
                    lots of places. It is defined as "final", so hopefully inline 
                    code will be generated.
                 Added getDoubles( attr ), getFloats( attr ) etc.
                 Added getDoubles(), getFloats()
                 Added getValues( attr )
   
   last update before history started: 8/20/03
   
************************************************************/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.nio.*;

/* BasicBlock stores data in one dimensional array. if this datablock has multiple 
 * attributes with different data types, then the data array is a byte array. 
 * If the attributes are the same type, a data array is create according 
 * to this data type
*/

public class BasicBlock extends DataBlock
{
   private class Field 
   {

       public int typeCode;
       public short stride;
       public short offset;
       public Field(int code, short s, short o )
         {
              typeCode = code;
              stride   = s;
              offset   = o;
         }
   }

  private Values    data;
  private Field[]  fields;
  private int storageType;


  //used for fast datum access
  Datum nullDatum;
   
  // variables used for better performace
  private float[] floatArray;

  protected int leastSignificantAxis; // most rapidly changing axis in datablock.
  protected int [] steps;             //Used for iterating through a subblock.
  
  IntegerIndexSpaceID iterCopy;
  
  // ----------- Following variables are used for copying subblock data ----------
  private int   [] dstintData = null;
  private int   [] srcintData = null;
  
  private float [] dstfloatData = null;
  private float [] srcfloatData = null;
  
  private double[] dstdoubleData = null;
  private double[] srcdoubleData = null;
  
  private short [] dstshortData  = null;
  private short [] srcshortData  = null;
  
  private byte  [] dstbyteData   = null;  
  private byte  [] srcbyteData   = null;
  
  //++++++++++++++++++++ Constructors +++++++++++++++++++++++++++++++++++++++
  //
  // ----------------- BasicBlock( isb, recordDesc, axisOrdering, data ) --------------
  // most general constructor. All others call this to do most of the work.
  // Don't want to replicate code.
  
    /**
     * Construct a BasicBlock with the given data, ISBounds, RecordDescriptor and 
     * axis axisOrdering. The bounds will be moved to the origin.
     */
    BasicBlock( ISBounds isb, RecordDescriptor recordDesc, 
                       AxisOrdering ordering, Values data )
    {
        super( isb, recordDesc, ordering );
        
        this.storageType = recordDesc.getStorageType();
       
        //if ( bounds.volume() != data.length )
        //    new IllegalArgumentError( SDBError.Fatal,
        //                      "Volume of ISBounds does not match array length."); 

        if ( data == null )
           initializeValue( this.storageType );
        else 
           this.data = data;
        nullDatum = this.createDatum();
        initializeFields();
        this.leastSignificantAxis = this.axisOrdering.getLeastSignificantAxis();
        this.iterCopy = new IntegerIndexSpaceID( new int[ this.dimensionality ] );
        iniStep( isb );
    }

    /** Construct a BasicBlock with the given data and ISBounds. 
     *  The default axis axisOrdering is used, and the bounds will be moved to the origin.
     */    
    public BasicBlock( ISBounds isb, RecordDescriptor recordDesc, Values data )
    { 
        this( isb, recordDesc, null, data );
    }

  /** Construct a datablock with the given bounds and number of attributes. The data
   *  array for the block will be created based on the data type in recordDesc
   */

    public BasicBlock( ISBounds isb, RecordDescriptor recordDesc )
	{
	   this( isb, recordDesc, null, null );  // invoke general constructor
    }
	
    /** Construct a datablock with the given bounds and axisordering. The data
        array for the block will be created based on the data type in recordDesc
     */
    public BasicBlock( ISBounds isb, RecordDescriptor recordDesc, AxisOrdering ordering )
    {
       this( isb, recordDesc, ordering, null );
    }
    // ---------------------------------------------------------------------------

    private void iniStep( ISBounds isb )
    {
        // System.out.println("******* iniStep is called ************");
        // We have to set the steps array so that it iterates 1 step at a time
        // through most dimensions, except for the one that changes most rapidly.
        // For this special dimension, we'll step by the size of the dimension. 
        // This means we will only "visit" the beginning of a row, column, etc. 
        // This visit is all we need to read an entire line of contiguous data
        // from the data array. 
        // This is usfull for subblock functions 

        int d = this.dim();
        IndexSpaceID dimensions = isb.getDimensions();
        
        this.steps = new int[d];
        this.leastSignificantAxis = this.axisOrdering.getLeastSignificantAxis();

        this.steps[ this.axisOrdering.getAxis( d-1 )] 
                            = dimensions.getCoord( this.axisOrdering.getAxis( d - 1 ));
        for ( int i=0; i<d-1 ;i++ )
            this.steps[ this.axisOrdering.getAxis(i) ] = 1;            
     }


   /** set the proper value (FloatValues, IntValues...) according to the storageTypes 
    *  of all the fields. 
    */
    private void initializeValue( int storageType )
    {
        int size = this.bounds.volume();
        int fieldsPerRecord = this.recordDesc.size();
        int bytesPerRecord = this.recordDesc.byteSize();
        
        switch ( storageType )
        {
           case StorageType.Double: 
              this.data = new DoubleValues( size * fieldsPerRecord );
              break;
           case StorageType.Float: 
              this.data = new FloatValues( size * fieldsPerRecord );
              break;
           case StorageType.Int: 
              this.data = new IntValues( size * fieldsPerRecord );
              break;
           case StorageType.Short: 
           case StorageType.UnsignedShort: 
              this.data = new ShortValues( size * fieldsPerRecord );
              break;
           case StorageType.Byte: 
           case StorageType.UnsignedByte: 
           case StorageType.Char: 
              this.data = new ByteValues( size * fieldsPerRecord );
              break;
           case StorageType.Record: 
              this.data = new ByteValues( size * bytesPerRecord );
              break;
           default: 
              new SDBError( SDBError.Warning, "Unsupported Storage Type: " 
                    + storageType );
              this.data = new ByteValues( size * bytesPerRecord );
              break;
        }
    }

   
    /**
     * Construct a DatumBlock from the given DataBlock. Bounds, Ordering, and
     * RecordDescriptor will be copied directly from the DataBlock argument.
     * Also, the resulting DatumBlock will REFER to the data contained in the
     * DataBlock argument. This is faster than using one of the constructors
     * above.
     */
/*    DatumBlock(DataBlock b){
    
        super(b);
        
        this.data=b.data();
        nullDatum=this.getRecordDescriptor().createDatum();
    }
*/


    private void initializeFields()
    {   
           // initiate Fields , ByteValues is treated differentely from others
        int type;               // used to store the storageType
        short offset = 0;         // offset of each fields in the byte array
        short stride;             // the stride , infact it's the byte size of all the attributes of a datum
        this.fields = new Field[numAttributes]; 
        
        if (this.data==null || this.recordDesc==null){
            new InternalFaultError(SDBError.Fatal,"BasicBlock had null data or RecordDescriptor during initializeFields.");
        }
        
        
      if ( this.data instanceof MixedValues )
      {
         stride = (short) this.recordDesc.byteSize();
         for ( int i = 0; i < numAttributes; i++ )
         {
            type = this.recordDesc.getStorageType( i );
            offset = (short) this.recordDesc.byteOffset( i );
            fields[i] = new Field( type, stride, offset ); 
         }
         //System.out.println("MixedValues stored ");
      }
      else
      // for IntValues, FloatValues, ShortValues
      {
         stride = (short) numAttributes;
         for ( int i = 0; i < numAttributes; i++ )
         {
            type = recordDesc.getStorageType( i );
            fields[i] = new Field( type, stride, offset++ );
         }
      }
    }

   // ---------------------- local utility methods
   // -------------------------------
    final protected int getFieldPos( int oneDIndex, int fieldNumber )
    {
       return oneDIndex * ( this.fields[fieldNumber].stride ) + this.fields[fieldNumber].offset;
    }
   // ------------------------------------------------------------------------------


    /** copy the data array of the argument block into the 
        the data array of this basicBlock. recordDesc is not copied.
        This is intended for use by the caching and so this block should
        have the same types as the argument block.
        NOTE: So far the float version is implemented. 
     */ 
    void copyArray( BasicBlock block )
    {
       try {
       switch ( this.storageType )
       {
          case StorageType.Double:
             double[] dSrc = ((DoubleValues) block.data).getDataReference();
             double[] dDest = ((DoubleValues) this.data).getDataReference();
             System.arraycopy( dSrc, 0, dDest, 0, dSrc.length );
             break;
          case StorageType.Float:
             System.err.println( "BB.copyArray" );
             float[] fSrc = ((FloatValues) block.data).getDataReference();
             float[] fDest = ((FloatValues) this.data).getDataReference();
             System.arraycopy( fSrc, 0, fDest, 0, fSrc.length );
             float[] array =  block.getFloats(); 
             //rdb System.arraycopy( array, 0, this.getFloatAttributes(), 0, array.length );
             if ( array != fSrc || this.getFloats() != fDest )
                System.err.println( "new copyArray ERROR" );
             break;
          case StorageType.Int:
             int[] iSrc = ((IntValues) block.data).getDataReference();
             int[] iDest = ((IntValues) this.data).getDataReference();
             System.arraycopy( iSrc, 0, iDest, 0, iSrc.length );
             break;
          case StorageType.Short:
             short[] sSrc = ((ShortValues) block.data).getDataReference();
             short[] sDest = ((ShortValues) this.data).getDataReference();
             System.arraycopy( sSrc, 0, sDest, 0, sSrc.length );
             break;
          case StorageType.Byte:
             byte[] bSrc = ((ByteValues) block.data).getDataReference();
             byte[] bDest = ((ByteValues) this.data).getDataReference();
             System.arraycopy( bSrc, 0, bDest, 0, bSrc.length );
             break;
       }
       } catch ( ClassCastException cce )
       {
          new SDBError( SDBError.Warning, 
              "BasicBlock.copyArray: block data type mismatch. " + cce.getMessage() +
              "\n ---------> Trying to copy data using float; if it works, it's slow." );
          this.setValues( block.getValues() );
       }
    }

    /** return the Values (data) store in this basicBlock 
     */
   @Override
    public  Values getValues()
     {
         return data;
     }

    /** return a Values array consisting of the data for the specified
     *  attribute. 
     */
   @Override
    public  Values getValues( int attr )
     {
        if ( attr < 0 || attr > getNumAttributes( )) 
        {
           new SDBError(SDBError.Recoverable, "getValues: invalid attribute index: " + attr  );
           return null;
        }
        else if ( getNumAttributes() == 1 && attr == 0 )
           return data.spawn();
        else // need to copy the values for this attribute
           return data.spawn( this.fields[ attr ].offset, 
                              this.fields[ attr ].stride );
     }


     /** set a int field value at certain index( index of conceptual datum for an one D array )*/
     public final void setField(int index, int intData, int fieldID)
      {
            if ( this.getStorageType() != StorageType.Int )
               new InternalFaultError( SDBError.Recoverable,"unmatched storageType." );
            else
               ( (IntValues) data ).setData(( index * numAttributes +fieldID) , intData );
      }

     /** set a double field value at certain index( index of conceptual datum for an one D array )*/
     public final void setField( int index, double doubleData, int fieldID )
      {
            if ( this.getStorageType() != StorageType.Double )
               new InternalFaultError(SDBError.Recoverable,"unmatched storageType.");
            else
              {
                //System.out.println( "index :" + index + "  numAttributes :" + numAttributes +"  fieldID :" +fieldID +"    doublevalue :" + doubleData );
               ( (DoubleValues) data ).setData( (index * numAttributes +fieldID) , doubleData);
         
              }
      }

    /** set a short field value at certain index( index of conceptual datum for an one D array )*/
     public final void setField( int index, short shortData, int fieldID)
      {
            if ( this.getStorageType() != StorageType.Short)
               new InternalFaultError(SDBError.Recoverable,"unmatched storageType.");
            else
               ( (ShortValues) data ).setData((index * numAttributes +fieldID ) , shortData );
      }

     /** set a byte field value at certain index( index of conceptual datum for an one D array )*/
     public final void setField( int index, byte byteData, int fieldID)
      {
            if ( this.getStorageType() != StorageType.Byte )
               new InternalFaultError(SDBError.Recoverable,"unmatched storageType.");
            else
               ( (ByteValues) data ).setData((index * numAttributes + fieldID), byteData );
      }

     /** set a float field value at certain index( index of conceptual datum for an one D array )*/
     public final void setField(int index, float floatData, int fieldID)
      {
            if ( this.getStorageType() != StorageType.Float)
               new InternalFaultError(SDBError.Recoverable,"unmatched storageType.");
            else
               ( (FloatValues) data ).setData( (index * numAttributes +fieldID), floatData );
           // System.out.println( "index:" +index + "value" +floatData);
      }


     /** set a field value, the index is the index of the field in this  1D array */
     public  void setField(int fieldIndex, float floatData)
      {
        
        if ( this.getStorageType() != StorageType.Float)
               new InternalFaultError(SDBError.Recoverable,"unmatched storageType.");
            else
               //( (FloatValues) data ).setData( fieldIndex , floatData);
                {  
                    floatArray = (( FloatValues)data).getDataReference();                         
                    floatArray[fieldIndex] = floatData;
                }
    }
      

     /** get indexed float attribute , OneDIndex is the conceptual index of the datum in the 
         one D array . If the data type of specified attributes is not float, it will be cast to float
      */
     float getIndexedFloat( int oneDIndex, int fieldNumber )
     {  
       //return data.getFloat( OneDIndex * ( this.fields[fieldNumber].stride ) + this.fields[fieldNumber].offset );
       return data.getFloat( getFieldPos( oneDIndex, fieldNumber ) );
     }

   /** get indexed Short attribute, OneDIndex is the conceptual index of the datum in the one D array
       If the data type of specified attributes is not short, it will be cast to short
     */
    short getIndexedShort(int oneDIndex, int fieldNumber)
     {  
        // return data.getShort( OneDIndex * ( this.fields[fieldNumber].stride ) + this.fields[fieldNumber].offset ); 
        return data.getShort( getFieldPos( oneDIndex, fieldNumber ) ); 
     }


   /** get indexed int attribue, OneDIndex is the conceptual index of the datum in the one D array
       If the data type of specified attributes is not int, it will be cast to int
   */
    int getIndexedInt(int oneDIndex, int fieldNumber)
     {  
         //return data.getInt( OneDIndex * ( this.fields[fieldNumber].stride ) + this.fields[fieldNumber].offset ); 
         return data.getInt( getFieldPos( oneDIndex, fieldNumber ) ); 
     }

   /** get indexedFieldShort , OneDIndex is the conceptual index of the datum in the one D array
       If the data type of specified attributes is not double, it will be cast to double
    */
    double getIndexedDouble(int oneDIndex, int fieldNumber)
     {  
         //return data.getDouble( OneDIndex * ( this.fields[fieldNumber].stride ) + this.fields[fieldNumber].offset ); 
         return data.getDouble( getFieldPos( oneDIndex, fieldNumber ) ); 
     }



    /** get the float attribute for a given IndexSpaceID and attributeID, convert if needed */
   @Override
    public float getFloat( IndexSpaceID did, int fieldNumber )
    {
	   return this.data.getFloat( getFieldPos( this.computeOffset(did), fieldNumber ) );
    }
    
    //-------------------- get<type>( isid, attr ) --------------------------------
    // Base versions of these are defined in DataCollection, but these are more
    // efficient implementations.
    
    /** get the field value at the ISID position as an int. Convert type if necessary. */
   @Override
    public double getDouble( IndexSpaceID did, int fieldNumber )
    {
	   return this.data.getDouble( getFieldPos( this.computeOffset(did), fieldNumber ) );
    }

    /** get the field value at the ISID position as an int. Convert type if necessary. */
   @Override
    public int getInt( IndexSpaceID did, int fieldNumber )
    {
       return this.data.getShort( getFieldPos( this.computeOffset(did), fieldNumber ) );
    }

    /** get the field value at the ISID position as a short. Convert type if necessary. */
   @Override
    public short getShort( IndexSpaceID did, int fieldNumber )
    {
        return this.data.getShort( getFieldPos( this.computeOffset(did), fieldNumber ) );
    }

    /** get the field value at the ISID position as a short. Convert type if necessary. */
   @Override
    public byte getByte( IndexSpaceID did, int fieldNumber )
    {
       return this.data.getByte( getFieldPos( this.computeOffset(did), fieldNumber ) );
    }

    /** User passes in a pre-allocated float array, attributes specified by recordSpec are copied 
        from the BasicBlock to the dstArray. attribute IDs and their storage order are specified in RecordSpec.
        the boolean parameter indicates if the attribute in the dstArray in stored in a Point Order or AttributeOrder 
        Attributes are changed into floats if they are not stored as float in this BasicBlock. 
    */
    public void getFloats(float[] dstArray, RecordSpec recordSpec, boolean pointOrder)
    {
         //if ( this.storageType != StorageType.Float)
         //  new InternalFaultError(SDBError.Recoverable,"unmatched storageType, Requested Type: Float , BasicBlock storageType:" + this.storageType ); 
         //int length = this.volume() * recordSpec.size();
         int numAttri = recordSpec.size() ;
         //int blockNumAttributes = this.numAttributes;
         
         //float[] floatArray = new float[ this.volume() * numAttri ];
         int numberOfDatums = this.volume();
         int index =0;
         if ( pointOrder)
          {
             for ( int i =0; i<numberOfDatums; i++)
             {
                 for ( int j =0 ;j<numAttri;j++)
                   {
                      dstArray[index ++]= 
                         this.data.getFloat(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         else // store the attribtes in attributeOrder
          {
             for( int j =0 ;j<numAttri;j++)
             {
                for( int i =0; i<numberOfDatums; i++)
                   {
                      dstArray[index ++]= 
                         this.data.getFloat(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         //return dstArray;
    }


    /** User passes in a pre-allocated int array, attributes specified by recordSpec 
     * are copied from the BasicBlock to the dstArray. attribute IDs and their storage 
     * order are specified in RecordSpec. the boolean paramerter indicates if the 
     * attribute in the dstArray in stored in a Point Order or AttributeOrder 
     * Attributes are changed into ints if they are not stored as ints in this BasicBlock. 
     */
    public void getInts( int[] dstArray, RecordSpec recordSpec, boolean pointOrder )
    {
         //if ( this.storageType != StorageType.Int)
         //  new InternalFaultError(SDBError.Recoverable,"unmatched storageType, Requested Type: int , BasicBlock storageType:" + this.storageType ); 
         //int length = this.volume() * recordSpec.size();
         int i,j;
         int numAttri = recordSpec.size() ;
         //int blockNumAttributes = this.numAttributes;
         

         //int[] intArray = new int[ this.volume() * numAttri ];
         int numberOfDatums = this.volume();
         int index =0;
         if ( pointOrder)
          {
             //if ( attributeMatch(recordSpec) )
              //  return ((IntValues)this.data).intValues();

             for( i =0; i<numberOfDatums; i++)
             {
                 for( j =0 ;j<numAttri;j++)
                   {
                      dstArray[index ++]= 
                         this.data.getInt(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         else // store the attribtes in attributeOrder
          {
             for( j =0 ;j<numAttri;j++)
             {
                for( i =0; i<numberOfDatums; i++)
                   {
                      dstArray[index ++]= 
                         this.data.getInt(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         //return intArray;
    }


    /** User passes in a pre-allocated short array, attributes specified by recordSpec are copied 
	     from the BasicBlock to the dstArray. attribute IDs and their storage order are specified in RecordSpec.
	     the boolean paramerter indicates if the attribute in the dstArray in stored in a Point Order or AttributeOrder 
	     Attributes are changed into shorts if they are not stored as shorts in this BasicBlock. 
     */
    public void getShorts( short[] dstArray, RecordSpec recordSpec, boolean pointOrder )
    {
         int i,j;
         int numAttri = recordSpec.size() ;

         int numberOfDatums = this.volume();
         int index =0;
         if ( pointOrder)
          {
             for( i =0; i<numberOfDatums; i++)
             {
                 for( j =0 ;j<numAttri;j++)
                   {
                      dstArray[index ++]= 
                         this.data.getShort(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         else // store the attribtes in attributeOrder
          {
             for( j =0 ;j<numAttri;j++)
             {
                for( i =0; i<numberOfDatums; i++)
                   {
                      dstArray[index ++]= 
                         this.data.getShort(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
    }

   /** User passes in a pre-allocated double array, attributes specified by recordSpec are copied 
        from the BasicBlock to the dstArray. attribute IDs and their storage order are specified in RecordSpec.
        the boolean paramerter indicates if the attribute in the dstArray in stored in a Point Order or AttributeOrder 
        Attributes are changed into doubles if they are not stored as doulbes in this BasicBlock. 
    */
    public void getDoubles( double[] dstArray, RecordSpec recordSpec, boolean pointOrder)
    {
         int i,j;
         int numAttri = recordSpec.size() ;
         int numberOfDatums = this.volume();
         int index =0;
         if ( pointOrder)
          {
             for( i =0; i<numberOfDatums; i++)
             {
                 for( j =0 ;j<numAttri;j++)
                   {
                      dstArray[index ++]= 
                         this.data.getDouble(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
         else // store the attribtes in attributeOrder
          {
             for( j =0 ;j<numAttri;j++)
             {
                for( i =0; i<numberOfDatums; i++)
                   {
                      dstArray[index ++]= 
                         this.data.getDouble(i*this.numAttributes + recordSpec.getFieldNumber(j) );
                   }    
             }
          }
    }


    /******** rdb: unused method 10/21/05
    // test if the requested recordSpec maches the attributes of this BasicBlock   
    private boolean attributeMatch( RecordSpec recordSpec)
    {
        int recNumAttri = recordSpec.size();
        if ( recNumAttri != this.numAttributes)
           return false;
        for( int i=0; i< recNumAttri; i++)
          {
              if (recordSpec.getFieldNumber(i) != i )
                return false;
          }
        return true;
    }
    /****************************/
    
    /** Set all values at once. The input Values must be of the correct
     * length (which is tested) and are assumed to have the correct data order 
     * and axis order. Return of true indicates that implementation occurred.
     */
   @Override
    public boolean setValues( Values vals )
    {
       if ( vals.size() != this.volume() * this.numAttributes )
          return false;
       if ( this.storageType == vals.getStorageType() ) 
          this.data = vals;
       else
          this.data.setValues( vals );
       return true;
    }

    //---------------- set<type>s( <type>[] )---------------------
    /** set the entire attribute array block to the passed argument.
     *  If the array size is not correct, should generate an error.
     *  If storage type is not float, do nothing (should convert).
     */
    public void setFloats( float newVals[] )
    {
      if ( this.getStorageType() == StorageType.Float
            && newVals.length == this.data.size() )
         this.data = new FloatValues( newVals );
      else 
         new SDBError(SDBError.Recoverable,
                "BasicBlock.setFloats: array size mismatch." );
    }
    

     //------------------ set<type>s( int attr, <type>[] ) --------------------------
     // override parent for efficiency
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setDoubles( int attr, double[] newValues )
     {
        if ( this.getNumAttributes() == 1 
              && this.data instanceof DoubleValues 
              && newValues.length == this.data.size() )
           this.data = new DoubleValues( newValues );
        else
           super.setDoubles( attr, newValues );
     }
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setFloats( int attr, float[] newValues )
     {
         if ( this.getNumAttributes() == 1 
               && this.data instanceof FloatValues 
               && newValues.length == this.data.size() )
            this.data = new FloatValues( newValues );
         else 
            super.setFloats( attr, newValues );
     }
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setInts( int attr, int[] newValues )
     {
         if ( this.getNumAttributes() == 1 
               && this.data instanceof IntValues 
               && newValues.length == this.data.size() )
            this.data = new IntValues( newValues );
         else 
            super.setInts( attr, newValues );
     }
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setShorts( int attr, short[] newValues )
     {
         if ( this.getNumAttributes() == 1 
               && this.data instanceof ShortValues 
               && newValues.length == this.data.size() )
            this.data = new ShortValues( newValues );
         else 
            super.setShorts( attr, newValues );
     }
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setBytes( int attr, byte[] newValues )
     {
         if ( this.getNumAttributes() == 1 && this.data instanceof ByteValues 
             && newValues.length == this.data.size() )
            this.data = new ByteValues( newValues );
         else 
            super.setBytes( attr, newValues );
     }
     
     /** Assign attribute values for all points in the block */
   @Override
     public void setValues( int attr, Values newValues )
     {
         if ( ( this.getNumAttributes() == 1 )
               &&  ( this.data.getStorageType() == newValues.getStorageType() )
               && ( newValues.size() == this.data.size() ) )
            this.data = newValues;
         else 
            super.setValues( attr, newValues );
     }
     
     //------------------ datum methods ----------------------------
    /** Given an IndexSpaceID, return the corresponding datum.( this only works for float implementation )*/
    //  Do We still need it for BasicBlock
   @Override
    public  Datum datum(IndexSpaceID did){
    
        int NtoOneOffset=this.computeOffset(did);
        Datum returnVal=recordDesc.createDatum();
        
        for(int i=0; i < numAttributes ; i++)
           returnVal.setFloat( i, data.getFloat( getFieldPos( NtoOneOffset, i )));
           // returnVal.setFloat( i, data.getFloat( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
        return returnVal;
    }


    /** Given IndexSpaceID, load the corresponding datum into the datum argument.
     *  (this only works for float implementation )   
     */
   @Override
    public  void datum(Datum datum, IndexSpaceID did){

        int NtoOneOffset=this.computeOffset(did);

        for(int i=0; i < numAttributes ; i++ )
           datum.setFloat( i, data.getFloat( getFieldPos( NtoOneOffset, i )));    
           //datum.setFloat( i, data.getFloat( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
    }


    /** Given an array of coordinates, load the corresponding datum into the datum 
        argument.*/
   @Override
    public  void datum(Datum datum, int [] coords){
        int NtoOneOffset=this.computeOffset(coords);

        for(int i=0; i < numAttributes ; i++)
           datum.setFloat( i, data.getFloat( getFieldPos( NtoOneOffset, i )));    
           //datum.setFloat( i, data.getFloat( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
    }

    /** Given an array of coordinates, load the corresponding datum into the datum 
        argument.*/
    public  void datum(ShortDatum datum, int [] coords){
        int NtoOneOffset=this.computeOffset(coords);

        for(int i=0; i < numAttributes ; i++)
           datum.setShort( i, data.getShort( getFieldPos( NtoOneOffset, i )) );    
            //datum.setShort( i, data.getShort( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
    }

    /** Given an array of coordinates, load the corresponding datum into the datum 
        argument.*/
    public  void datum(ByteDatum datum, int [] coords){
        int NtoOneOffset=this.computeOffset(coords);

        for(int i=0; i < numAttributes ; i++)
           datum.setByte( i, data.getByte( getFieldPos( NtoOneOffset, i )) );    
            //datum.setShort( i, data.getShort( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
    }
    
    
    /** Given an IndexSpaceID, return the corresponding datum with attributes
        specified in the RecordSpec. */    
   @Override
    public  Datum datum(IndexSpaceID did, RecordSpec spec){
    
        int NtoOneOffset=this.computeOffset(did);
        for(int i=0; i < numAttributes ; i++)
           nullDatum.setFloat( i, data.getFloat( getFieldPos( NtoOneOffset, i )) );    
            //nullDatum.setFloat( i, data.getFloat( NtoOneOffset * this.fields[i].stride + this.fields[i].offset) );    
        return nullDatum.subset( spec );
    }
    
        
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
    {
      int NtoOneOffset = this.computeOffset(did);
      int[] cpnID = mp.getComponentFields();
      int[] cpsID = mp.getCompositeFields();
      for ( int i=0; i<cpnID.length; i++ )
	     dtm.setFloat( cpsID[i], this.getIndexedFloat( NtoOneOffset, cpnID[i]));
    }


   /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
   @Override
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec){
      int NtoOneOffset = this.computeOffset(did);
      int[] attributeID = recSpec.getFieldNumbers();
      int len = recSpec.size();
      for(int i=0; i<len; i++)
	dtm.setFloat(i, this.getIndexedFloat(NtoOneOffset, attributeID[i]) );
    }

/** this  is the old interface, the query bounds is check to ensure correctness
*/ 
 public  void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
    {
         //if ( this.storageType != dstBlock.getStorageType() )
         //          new InternalFaultError(SDBError.Recoverable,"Data Storage type mismatch ");
         if ( dstBounds.volume() != srcBounds.volume())
            {
              new IllegalArgumentError(SDBError.Fatal, "**** srcBounds does not match the dstBounds\n" +
                                        "srcBounds: " + srcBounds  +"\n"+
                                        "dstBounds: " + dstBounds );
             }
         if (  ! (dstBlock.getBounds()).contains(dstBounds) )
            {
               new IllegalArgumentError(SDBError.Fatal, "**** dstBounds is out of the bounds of dstBlock\n" +
                                        "dstBlock Bounds: " + dstBlock.getBounds()  +"\n"+
                                        "datBounds:" + dstBounds );
            } 
         if ( ! (this.getBounds()).contains(srcBounds) )
           {
               new IllegalArgumentError(SDBError.Fatal, "**** srcBounds is out of the bounds of calling basicBlock\n" +
                                        "srcBounds: " +srcBounds +"\n"+
                                        "Bounds of this BasicBlock:" + this.getBounds() );
           }

       if ( dstBlock instanceof BasicBlock){
           subblock( (BasicBlock)dstBlock, dstBounds, srcBounds );
       }
       else 
       {
           new InternalFaultError(SDBError.Recoverable,"Subblock from BasicBlock to CompositeBlock is not implemented .");
       }// end of else
    }

  /** this subblock does not check the bounds
   */
   public void psubblock( DataBlock dstBlock, ISBounds dstBounds, 
                          ISBounds srcBounds)
   {
         this.subblock( (BasicBlock)dstBlock,dstBounds,srcBounds );
   }

   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
         defined by dstBounds.  Note: currently we assume the datablock is BasicBlock, 
         and the space of this basicblock has been allocated in advance 
         AND is correctly sized.
      */
//     protected void subblock( BasicBlock dstBlock,  ISBounds dstBounds, 
//                                   ISBounds srcBounds )
//     {
//        subblock( dstBlock, dstBounds, srcBounds, (RecordSpec)null, (FieldIDMapper)null);
//     }
     
    protected void subblock( BasicBlock dstBlock,  ISBounds inDestBounds, 
                                  ISBounds srcBounds)
    {
       // 1. Deal with odd destination block bounds
       //
       // Need to deal with the destination as if its minimum location is at origin
       // We'll do that by saving the lower array, moving the block bounds to the origin
       //   and adjusting the destination bounds to make it relative to the origin.
       // Then just before returning, we'll restore the original min position.
       // Save the destinations min location:
       int[] srcLower = dstBlock.getBounds().getLowerArray();
       int[] dstLower = new int[srcLower.length];
       System.arraycopy (srcLower, 0, dstLower, 0, srcLower.length);
       
       int    numDestFields = dstBlock.getNumAttributes();
       
       // move destination bounds to be relative to the lower corner of block
       ISBounds dstBounds = inDestBounds.copy();
       dstBounds.makeRelativeTo( dstLower );
       
       // move the block bounds to the origin
       dstBlock.getBounds().moveToOrigin();       
       
       
       // We can iterate over rods; could save lots of copying time
       ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );        
       int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis );
       int rodDataSize = rodSize * numAttributes;     
       
      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
          
       // 3. Do the copy
        switch ( this.storageType )
       {
          case StorageType.Double:
             copyDoubles(dstBlock, iter, boundsMapper, rodDataSize, numDestFields);
             break;
          case StorageType.Float:
             copyFloats(dstBlock, iter, boundsMapper, rodDataSize, numDestFields);
             break;
          case StorageType.Int:
             copyInts(dstBlock, iter, boundsMapper, rodDataSize, numDestFields);
             break;
          case StorageType.Short:
             copyShorts(dstBlock, iter, boundsMapper, rodDataSize, numDestFields);
             break;
          case StorageType.Byte:
             copyBytes(dstBlock, iter, boundsMapper, rodDataSize, numDestFields);
             break;
          default:
             new IllegalArgumentError(SDBError.Fatal,
                   "Unsupported storage type code in BasicBlock :"+ this.storageType );
       }

       // 5. Now we need to restore the destination data block bounds to original
       dstBlock.getBounds().moveTo( dstLower );
    }

   private final void copyDoubles(DataBlock dstBlock, ISIterator iter, ISIDMapper boundsMapper, int rodDataSize, int numDestFields){

      double [] dstData = ((DoubleValues)  dstBlock.getValues() ).getDataReference();
      double [] srcData = ((DoubleValues) this.data).getDataReference();

      int srcOffset;           
      int dstOffset;        

      if(rodDataSize == 1){

         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            
            dstData[dstOffset] = srcData[srcOffset];
         }  
      } else {
      
         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            System.arraycopy( srcData, srcOffset, dstData, dstOffset, rodDataSize );
         }  
      } 
   }


   private final void copyFloats(DataBlock dstBlock, ISIterator iter, ISIDMapper boundsMapper, int rodDataSize, int numDestFields){

      float [] dstData = ((FloatValues)  dstBlock.getValues() ).getDataReference();
      float [] srcData = ((FloatValues) this.data).getDataReference();

      int srcOffset;           
      int dstOffset;        

      if(rodDataSize == 1){

         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            
            dstData[dstOffset] = srcData[srcOffset];
         }  
      } else {
      
         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            System.arraycopy( srcData, srcOffset, dstData, dstOffset, rodDataSize );
         }  
      } 
   }


   private final void copyInts(DataBlock dstBlock, ISIterator iter, ISIDMapper boundsMapper, int rodDataSize, int numDestFields){

      int [] dstData = ((IntValues)  dstBlock.getValues() ).getDataReference();
      int [] srcData = ((IntValues) this.data).getDataReference();

      int srcOffset;           
      int dstOffset;        

      if(rodDataSize == 1){

         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            
            dstData[dstOffset] = srcData[srcOffset];
         }  
      } else {
      
         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            System.arraycopy( srcData, srcOffset, dstData, dstOffset, rodDataSize );
         }  
      } 
   }


   private final void copyShorts(DataBlock dstBlock, ISIterator iter, ISIDMapper boundsMapper, int rodDataSize, int numDestFields){

      short [] dstData = ((ShortValues)  dstBlock.getValues() ).getDataReference();
      short [] srcData = ((ShortValues) this.data).getDataReference();

      int srcOffset;           
      int dstOffset;        

      if(rodDataSize == 1){

         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            
            dstData[dstOffset] = srcData[srcOffset];
         }  
      } else {
      
         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            System.arraycopy( srcData, srcOffset, dstData, dstOffset, rodDataSize );
         }  
      } 
   }


   private final void copyBytes(DataBlock dstBlock, ISIterator iter, ISIDMapper boundsMapper, int rodDataSize, int numDestFields){

      byte [] dstData = ((ByteValues)  dstBlock.getValues() ).getDataReference();
      byte [] srcData = ((ByteValues) this.data).getDataReference();

      int srcOffset;           
      int dstOffset;        

      if(rodDataSize == 1){

         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            
            dstData[dstOffset] = srcData[srcOffset];
         }  
      } else {
      
         for ( iter.init(); iter.valid(); iter.next() ){
         
            srcOffset = this.computeOffset( iter ) * this.numAttributes;
            this.iterCopy.assign( iter );
            dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                                  * numDestFields;
            System.arraycopy( srcData, srcOffset, dstData, dstOffset, rodDataSize );
         }  
      } 
   }


  /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds.  Note: currently we assume the datablock is BasicBlock, 
        and the space of this basicblock has been allocated in advance 
        AND is correctly sized.
     */
    protected void subblock( BasicBlock dstBlock,  ISBounds inDestBounds, 
                                  ISBounds srcBounds,
                                  RecordSpec rs, FieldIDMapper fim )
    {
       // 1. Deal with odd destination block bounds
       //
       // Need to deal with the destination as if its minimum location is at origin
       // We'll do that by saving the lower array, moving the block bounds to the origin
       //   and adjusting the destination bounds to make it relative to the origin.
       // Then just before returning, we'll restore the original min position.
       // Save the destinations min location:
       int[] srcLower = dstBlock.getBounds().getLowerArray();
       int[] dstLower = new int[srcLower.length];
       System.arraycopy (srcLower, 0, dstLower, 0, srcLower.length);
       
       // move destination bounds to be relative to the lower corner of block
       ISBounds dstBounds = inDestBounds.copy();
       dstBounds.makeRelativeTo( dstLower );
       
       // move the block bounds to the origin
       dstBlock.getBounds().moveToOrigin();       
       
       // 2. Set up fields for alternative cases: RecordSpec, FieldIDMapper and normal
       int rodSize = 0;
       int rodDataSize = 0;
       ISIterator iter = null;        
       int    numDestFields = dstBlock.getNumAttributes();
       int[]  rsFieldMapping = null;              // RecordSpec mapper
       int[]  fimComponentFields = null; // FIM arrays
       int[]  fimCompositeFields = null;
       
       if ( rs != null && fim != null )
          new SDBError( SDBError.Warning, "Internal error: BasicBlock.subblock " 
                              + " got RecordSpec AND FieldIDMapper, using RS." );
       if ( rs != null )  // RecordSpec case
       {
          // need to iterate over each location
          iter = new ISIterator( srcBounds, this.axisOrdering ); 
          rsFieldMapping = rs.getFieldNumbers();
       }
       else if ( fim != null )
       {
          // need to iterate over each location
          iter = new ISIterator( srcBounds, this.axisOrdering ); 
          fimComponentFields = fim.getComponentFields();
          fimCompositeFields = fim.getCompositeFields();
       }
       else  // default case
       {
          // We can iterate over rods; could save lots of copying time
          iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );        
          rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis );
          rodDataSize = rodSize * numAttributes;         
       }
          
       // 3. Set up copy pointers
       //
        switch ( this.storageType )
       {
          case StorageType.Double:
             dstdoubleData = ((DoubleValues)  dstBlock.getValues() ).getDataReference();
             srcdoubleData = ((DoubleValues) this.data).getDataReference();
             break;
          case StorageType.Float:
             dstfloatData = ((FloatValues) dstBlock.getValues() ).getDataReference();
             srcfloatData = ((FloatValues) this.data).getDataReference();
             break;
          case StorageType.Int:
             dstintData =   ((IntValues) dstBlock.getValues() ).getDataReference();
             srcintData =   ((IntValues) this.data).getDataReference();
             break;
          case StorageType.Short:
             dstshortData  = ((ShortValues)  dstBlock.getValues() ).getDataReference();
             srcshortData  = ((ShortValues) this.data).getDataReference();
             break;
          case StorageType.Byte:
             dstbyteData = ((ByteValues)  dstBlock.getValues() ).getDataReference();
             srcbyteData = ((ByteValues) this.data).getDataReference();
             break;
          default:
             new IllegalArgumentError(SDBError.Fatal,
                   "Unsupported storage type code in BasicBlock :"+ this.storageType );
       }

       //4. Set up mapping and do the copying
       ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
       
       int srcOffset;           
       int dstOffset;        
       for ( iter.init(); iter.valid(); iter.next() )
       {
          srcOffset = this.computeOffset( iter ) * this.numAttributes;
          this.iterCopy.assign( iter );
          dstOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy )) 
                                                               * numDestFields;
          if ( rs != null )  // We have a record spec, only copy selected attrs
             copyData( srcOffset, dstOffset, rsFieldMapping );
          else if (fim != null )
             copyData( srcOffset, dstOffset,  fimComponentFields, fimCompositeFields );
          else
             copyData( srcOffset, dstOffset, rodDataSize );
       }
       
       // 5. Now we need to restore the destination data block bounds to original
       dstBlock.getBounds().moveTo( dstLower );
    }
    
    protected void copyData( int srcOffset, int dstOffset,
                                  int[] componentFields, int[] compositeFields )
    {
       for ( int i = 0; i < componentFields.length ; i++ )
       {
          int doff = dstOffset + compositeFields[i];
          int soff = srcOffset + componentFields[i];
          switch ( this.storageType )
          {
             case StorageType.Float:
                dstfloatData[ doff ] = srcfloatData[ soff ];
                break;
             case StorageType.Int:
                dstintData[ doff ] = srcintData[ soff ];
                break;
             case StorageType.Short:
                dstshortData[ doff ] = srcshortData[ soff ];
                break;
             case StorageType.Byte:
                dstbyteData[ doff ] = srcbyteData[ soff ];
                break;
             case StorageType.Double:
                dstdoubleData[ doff ] = srcdoubleData[ soff ];
             break;
                default:
                new IllegalArgumentError(SDBError.Fatal,
                      "Unsupported storage type code in BasicBlock :"+ this.storageType );
          } 
       }
    }
                                  
    protected void copyData( int srcOffset, int dstOffset, int[] rsMapping )
    {   
       int doff = dstOffset;
       for ( int i = 0; i < rsMapping.length ; i++, doff++ )
       {
          switch ( this.storageType )
          {
             case StorageType.Float:
                dstfloatData[ doff ] = srcfloatData[ srcOffset + rsMapping[i] ];
                break;
             case StorageType.Int:
                dstintData[ doff ] = srcintData[ srcOffset + rsMapping[i] ];
                break;
             case StorageType.Short:
                dstshortData[ doff ] = srcshortData[ srcOffset + rsMapping[i] ];
                break;
             case StorageType.Byte:
                dstbyteData[ doff ] = srcbyteData[ srcOffset + rsMapping[i] ];
                break;
             case StorageType.Double:
                dstdoubleData[ doff ] = srcdoubleData[ srcOffset + rsMapping[i] ];
                break;
             default:
                new IllegalArgumentError(SDBError.Fatal,
                 "Unsupported storage type code in BasicBlock :"+ this.storageType );
          } 
       } 
    }
    
    protected void copyData( int srcOffset, int dstOffset, int rodDataSize )
    {     
       switch ( this.storageType )
       {
          case StorageType.Float:
             System.arraycopy( srcfloatData, srcOffset, dstfloatData, dstOffset, rodDataSize );
             break;
          case StorageType.Int:
             System.arraycopy( srcintData, srcOffset, dstintData, dstOffset, rodDataSize );
             break;
          case StorageType.Short:
             System.arraycopy( srcshortData, srcOffset, dstshortData, dstOffset, rodDataSize );
             break;
          case StorageType.Double:
             System.arraycopy( srcdoubleData, srcOffset, dstdoubleData, dstOffset, rodDataSize );
             break;
          case StorageType.Byte:
             System.arraycopy( srcbyteData, srcOffset, dstbyteData, dstOffset, rodDataSize );
             break;                   
          default:
             new IllegalArgumentError( SDBError.Fatal,
                   "Unsupported storage type code in BasicBlock :" + this.storageType );
       } 
    }

   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between 
        datums in this DataBlock and datums in the argument DataBlock.
        It is valid to pass a null argument for FieldIDMapper; this means use all 
        attributes. 
        ( Non-BasicBlock arguments are not implemented.) 
     */    
    public void subblock( DataBlock    dstBlock, 
                         ISBounds      dstBounds, 
                         ISBounds      srcBounds, 
                         FieldIDMapper mp)
    {
       if ( dstBlock instanceof BasicBlock)
           subblock( (BasicBlock)dstBlock,dstBounds,srcBounds, mp);
       else 
           new InternalFaultError(SDBError.Recoverable,
                 "Subblock from BasicBlock to CompositeBlock is not implemented .");
    }

    private void subblock( BasicBlock    dstBlock, 
                           ISBounds      dstBounds, 
                           ISBounds      srcBounds,
                           FieldIDMapper mp )
    {
       subblock( dstBlock, dstBounds, srcBounds, null, mp );
       /************** 11/13/05 deleted code below; merged functionality into
        * a common method -- more code shared.
        //int dataType = this.storageType;
        int   [] dstintData = null;
        int   [] srcintData = null;

        float [] dstfloatData = null;
        float [] srcfloatData = null;
  
        double[] dstdoubleData = null;
        double[] srcdoubleData = null;
 
        short [] dstshortData  = null;
        short [] srcshortData  = null;

        byte  [] dstbyteData   = null;  
        byte  [] srcbyteData   = null;
 
       switch ( this.getStorageType() ){
          case StorageType.Double:
             dstdoubleData = ((DoubleValues)  dstBlock.getValues() ).getDataReference();
             srcdoubleData = ((DoubleValues) this.data).getDataReference();
             break;
          case StorageType.Float:
             dstfloatData = ((FloatValues) dstBlock.getValues() ).getDataReference();
             srcfloatData = ((FloatValues) this.data).getDataReference();
             break;
          case StorageType.Int:
             dstintData =   ((IntValues) dstBlock.getValues() ).getDataReference();
             srcintData =   ((IntValues) this.data).getDataReference();
             break;
          case StorageType.Short:
             dstshortData  = ((ShortValues)  dstBlock.getValues() ).getDataReference();
             srcshortData  = ((ShortValues) this.data).getDataReference();
             break;
          case StorageType.Byte:
             dstbyteData  = ((ByteValues)  dstBlock.getValues() ).getDataReference();
             srcbyteData  = ((ByteValues) this.data).getDataReference();
             break;
          default:
              new IllegalArgumentError(SDBError.Fatal,
                    "Unsupported storage type code in BasicBlock :"+ this.getStorageType() );
         }

        int srcOffset, dstIndex, i;    
        ISIterator iter=new ISIterator(srcBounds,this.ordering);        

        ISIDMapper boundsMapper = new ISIDMapper(srcBounds,dstBounds);
        
        if ( mp == null )
           mp = new FieldIDMapper( this.getNumAttributes() ); // allocated identity mapper

        int    numFields = mp.size();
        int[]  componentFields = mp.getComponentFields();
        int[]  compositeFields = mp.getCompositeFields();
        int    dstNumFields = dstBlock.getNumAttributes();

        for(iter.init(); iter.valid(); iter.next())
        {
           srcOffset = this.computeOffset( iter ) * this.numAttributes;
           this.iterCopy.assign( iter );
           dstIndex = dstBlock.computeOffset(boundsMapper.toComposite(iterCopy))* dstNumFields; 
           
           for ( i =0; i< numFields ; i++)
           {
              switch (this.getStorageType())
              {
                 case StorageType.Float:
                    dstfloatData[dstIndex + compositeFields[i] ] = srcfloatData[srcOffset + componentFields[i] ];
                    break;
                 case StorageType.Int:
                    dstintData[dstIndex + compositeFields[i] ] = srcintData[srcOffset + componentFields[i] ];
                    break;
                 case StorageType.Short:
                    dstshortData[dstIndex + compositeFields[i] ] = srcshortData[srcOffset + componentFields[i] ];
                    break;
                 case StorageType.Byte:
                    dstbyteData[dstIndex + compositeFields[i] ] = srcbyteData[srcOffset + componentFields[i] ];
                    break;
                 case StorageType.Double:
                    dstdoubleData[dstIndex + compositeFields[i] ] = srcdoubleData[srcOffset + componentFields[i] ];
                 break;
                    default:
                    new IllegalArgumentError(SDBError.Fatal,
                          "Unsupported storage type code in BasicBlock :"+ this.getStorageType() );
              } 
           } 
        } 
        /******************* end delete **********************************/             
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the attributes
        of this DataSource to be written into the DataBlock datums.( not implemented) */
   @Override
    public void subblock( DataBlock dstBlock, 
                          ISBounds dstBounds, ISBounds srcBounds, 
                          RecordSpec spec)
    {
       if ( dstBlock instanceof BasicBlock)
          subblock( (BasicBlock)dstBlock,dstBounds,srcBounds, spec);
          //rdb subblock( (BasicBlock)dstBlock,dstBounds,srcBounds, new FieldIDMapper( spec );
       else 
       {
           new InternalFaultError(SDBError.Recoverable,
                 "Subblock from BasicBlock to CompositeBlock is not implemented .");
       }// end of else
    }

   private void subblock( BasicBlock dstBlock, 
                          ISBounds dstBounds, ISBounds srcBounds, 
                          RecordSpec spec)
   {
      subblock( dstBlock, dstBounds, srcBounds, spec, null );
      /************ deleted 11/13/05
        //int dataType = this.storageType;
        int   [] dstintData = null;
        int   [] srcintData = null;

        float [] dstfloatData = null;
        float [] srcfloatData = null;
  
        double[] dstdoubleData = null;
        double[] srcdoubleData = null;
 
        short [] dstshortData  = null;
        short [] srcshortData  = null;

        byte  [] dstbyteData   = null;  
        byte  [] srcbyteData   = null;
        
        if ( spec == null )
           spec = new RecordSpec( this.getNumAttributes() ); // allocated identity map
 
       switch ( this.getStorageType() ){
          case StorageType.Double:
             dstdoubleData = ((DoubleValues)  dstBlock.getValues() ).getDataReference();
             srcdoubleData = ((DoubleValues) this.data).getDataReference();
             break;
          case StorageType.Float:
             dstfloatData = ((FloatValues) dstBlock.getValues() ).getDataReference();
             srcfloatData = ((FloatValues) this.data).getDataReference();
             break;
          case StorageType.Int:
             dstintData = ((IntValues) dstBlock.getValues() ).getDataReference();
             srcintData = ((IntValues) this.data).getDataReference();
             break;
          case StorageType.Short:
             dstshortData  = ((ShortValues) dstBlock.getValues() ).getDataReference();
             srcshortData  = ((ShortValues) this.data).getDataReference();
             break;
          case StorageType.Byte:
             dstbyteData  = ((ByteValues)  dstBlock.getValues() ).getDataReference();
             srcbyteData  = ((ByteValues) this.data).getDataReference();
             break;
          default:
             new IllegalArgumentError(SDBError.Fatal,
                   "Unsupported storage type code in BasicBlock :"+ this.getStorageType() );
         }

        int srcOffset, dstIndex;    
        ISIterator iter=new ISIterator(srcBounds,this.ordering);        

        ISIDMapper boundsMapper=new ISIDMapper(srcBounds,dstBounds);

        int   numFields = spec.size();
        int   dstNumFields = dstBlock.getNumAttributes();

        int[] attributeID = spec.getFieldNumbers();

	for ( iter.init(); iter.valid(); iter.next() )
	{
	   srcOffset=this.computeOffset(iter) * this.numAttributes;
	   this.iterCopy.assign(iter);
	   dstIndex = dstBlock.computeOffset(boundsMapper.toComposite(iterCopy))* dstNumFields; 
	   
	   for ( int i = 0; i < numFields ; i++)
	   {
	      switch (this.getStorageType())
	      {
	         case StorageType.Float:
	            dstfloatData[dstIndex +i] = srcfloatData[srcOffset + attributeID[i] ];
	            break;
	         case StorageType.Int:
	            dstintData[dstIndex +i] = srcintData[srcOffset + attributeID[i] ];
	            break;
	         case StorageType.Short:
	            dstshortData[dstIndex +i] = srcshortData[srcOffset + attributeID[i] ];
	            break;
	         case StorageType.Byte:
	            dstbyteData[dstIndex +i] = srcbyteData[srcOffset + attributeID[i] ];
	            break;
	         case StorageType.Double:
	            dstdoubleData[dstIndex +i] = srcdoubleData[srcOffset + attributeID[i] ];
	            break;
	         default:
	            new IllegalArgumentError(SDBError.Fatal,
	             "Unsupported storage type code in BasicBlock :"+ this.getStorageType() );
	      } 
	   } 
	} 
    /**************** end delete ***********************************/             
   }
   
   //-------------------- set Attribute methods ---------------------------

   

 /** Set the location specified by the IndexSpaceID to the given datum. */
   public void setDatum( IndexSpaceID isid, Datum d )
   {
      int dataType = this.getStorageType();
      int offset = this.computeOffset( isid );
      //System.out.println("offset;" + offset);

      for ( int i = 0; i < this.numAttributes; i++ )
      {
         switch ( dataType )
         {
            case StorageType.Float:
               this.setField( offset, d.getFloat( i ), i );
               // System.out.println("visited here");
               break;
            case StorageType.Int:
               //this.setField( offset, (int) d.getFloat( i ), i );
               this.setField( offset, d.getInt( i ), i );
               break;
            case StorageType.Short:
               //this.setField( offset, (short) d.getFloat( i ), i );
               this.setField( offset, d.getShort( i ), i );
               break;
            case StorageType.Byte:
               //this.setField( offset, (byte) d.getFloat( i ), i );
               this.setField( offset, d.getByte( i ), i );
               break;
            case StorageType.Double:
               //this.setField( offset, (double) d.getFloat( i ), i );
               this.setField( offset, d.getDouble( i ), i );
               break;
            default:
               new IllegalArgumentError( SDBError.Fatal,
                     "Unsupported storage type code in BasicBlock :" + dataType );
         }
      }
   }



  /** set field */
  private void setField( int sourcetype, int datumArrayOffset, 
                        ByteBuffer byteBuffer, int bufferOffset, int attriID )
  {
      if ( sourcetype == this.getStorageType() )
      {
          switch(sourcetype)
          {
            case StorageType.Int:
                 this.setField( datumArrayOffset, byteBuffer.getInt(bufferOffset), attriID  );
                 break;
            case StorageType.Short:
                 this.setField( datumArrayOffset, byteBuffer.getShort(bufferOffset), attriID  );
                 break;
            case StorageType.Unsigned_Byte:
            case StorageType.Char:
            case StorageType.Byte:
                 this.setField( datumArrayOffset, byteBuffer.get(bufferOffset), attriID  );
                 break;
            case StorageType.Double:
                 this.setField( datumArrayOffset, byteBuffer.getDouble(bufferOffset), attriID  );
                 break;
            case StorageType.Float:
                 this.setField( datumArrayOffset, byteBuffer.getFloat(bufferOffset), attriID  );
                 break;
            case StorageType.Record:
                 new InternalFaultError( SDBError.Recoverable,
                                               "Records are not yet implemented.");
                 break;
            default:
                 new InternalFaultError( SDBError.Recoverable,"Unknown StorageType");
                 break;
           }
       } 
       else
       {
            new InternalFaultError(SDBError.Recoverable,"unmatched data storage type.");
       }
  }

 /** given a rod of data ( byte buffer), copy the specified field into the 
  * corresponding location in the BasicBlock, dstBlock is considered as 
  * compositeBlock, the srcBlock is the componentBlock
  */
    protected void setRod( ByteBuffer       byteBuffer, 
                           int              rodSize,
                           int              startDatumArrayOffset, 
                           int              datumArrayStep, 
                           FieldIDMapper    mp, 
                           RecordDescriptor sourceRecordDesc )
    {
        int sourcetype;
        int fieldPos=0;
        int datumPos=0;
        int datumArrayOffset = startDatumArrayOffset;
        int numFields = mp.size();
        int [] srcFields = mp.getComponentFields();
        int [] dstFields = mp.getCompositeFields();
        int datumByteSize= sourceRecordDesc.byteSize();

        for(int i=0;i<rodSize;i++,datumArrayOffset+=datumArrayStep)
        {
           for(int j=0;j<numFields;j++)
           {
              fieldPos =   sourceRecordDesc.byteOffset(srcFields[j]);
              sourcetype = sourceRecordDesc.getStorageType(srcFields[j]);
              setField( sourcetype, datumArrayOffset, byteBuffer, 
                                            (datumPos + fieldPos), dstFields[j] );
           }
           datumPos+=datumByteSize;
        }
    }
    
  /** given a rod of data of one attribute, copy the spedified attribute into the corresponding location in the BasicBlock , used for attribute order binaryFiledataSoure
   */ 
   protected void setAttributeOrderRod( ByteBuffer byteBuffer, 
                                 int rodSize, 
                                 int startDatumArrayOffset, 
                                 int datumArrayStep,                                                               
                                 int dstAttributeID, 
                                 int attributeByteSize, 
                                 int sourcetype )
    {
        int bufferAttributePos=0;
        int datumArrayOffset = startDatumArrayOffset;

        for( int i=0; i<rodSize; i++, datumArrayOffset += datumArrayStep )
        {
            setField( sourcetype, datumArrayOffset, 
                                  byteBuffer, bufferAttributePos, dstAttributeID );
            bufferAttributePos += attributeByteSize;
        }
    }


 /** given a rod of data ( byte buffer), copy the specified field in corresponding location in the 
      BasicBlock, dstBlock is considered as composite , the srcBlock is the component 
   */
    protected void setRod( ByteBuffer byteBuffer, int rodSize, int startDatumArrayOffset, int datumArrayStep,  RecordSpec recSpec, RecordDescriptor sourceRecordDesc )
    {
        int sourcetype;
        int fieldPos = 0;
        int datumPos = 0;
        int datumArrayOffset = startDatumArrayOffset;

        int numFields = recSpec.size();
        int [] fieldNumbers = recSpec.getFieldNumbers();
        int datumByteSize = sourceRecordDesc.byteSize();

        for( int i=0; i<rodSize; i++, datumArrayOffset += datumArrayStep )
        {
          for( int j=0; j<numFields; j++ )
             {
                 fieldPos   = sourceRecordDesc.byteOffset(fieldNumbers[j]);
                 sourcetype = sourceRecordDesc.getStorageType(fieldNumbers[j]);
                 setField( sourcetype, datumArrayOffset, byteBuffer,  
                                               (datumPos + fieldPos), j );
             }
          datumPos+=datumByteSize;
        }
   }
}
