/************************************************************ 
   BinaryFileDataSource.java
   Copyright 2000 University of New Hampshire
   Durham, NH 03824

   Xuan Tang xtang@cs.unh.edu
   
   ----------------- Edit history as of 12/14/04 -------------------
   03/04/12 rdb: Added constructor that uses FDLDescriptor argument.
   03/26/09 jpf: Overrode setAxisOrdering method to accomodate the user changing the
                 ordering on the fly.  This is required in order for the 'steps' array
                 to be updated.
                 Added helper method setSteps to update the steps
   11/29/05 rdb: Force URL access to go through getURL method, so can find files better.
   10/21/05 rdb: Added dummy main to reference some unused methods so no warnings:
                     setByteArray and some debugging print methods
   06/01/05 rdb: Removed TypeConverter object and the setDatum method that used it.
                   functionality now handled by ByteBuffers.
   02/22/05 rdb: Consolidated the two constructors
   02/07/05 rdb: renamed setOrder( DataBuffer )  to setBufferByteOrder 
                 moved endian variable and setBufferByteOrder and getByteOrder to 
                     PhysicalDataSource   
   01/06/05 rdb: create Datum objects from RecordDescriptor
   12/14/04 rdb -Added support for Byte storage type
                -for setDatum methods, do not force to float now, but let Datum
                   object do conversion if needed.
   -----------------------------------------------------------------
************************************************************/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/** This class will read from a file of binary data, providing an n-dimensional
    datasource view of that file. For the moment, we'll assume that all data is
    in big-endian Java Native format. 
*/

public abstract class BinaryFileDataSource extends PhysicalDataSource
{    
    protected RandomAccessFile raFile;
    protected FileChannel fChannel;
    protected int []      steps; //Used for iterating through a subblock.
    protected int         leastSignificantAxis; // Dimension that changes most rapidly in file.
    protected int         recordByteSize; // size of file records in bytes.
    protected int         recordNumFields; // number of fields in a complete record.
    protected Datum       tempDatum; // Used in Datum(IndexSpaceID,RecordSpec) to conserve memory.

    protected IntegerIndexSpaceID iterCopy;
    protected IntegerIndexSpaceID bufferIterCopy ;
    protected IndexSpaceID        dimensions;
    protected IndexSpaceID        bufferUpper;

    protected long   iniOffset;   // starting point of the file, by default the value is 0;
    
    //protected TypeConverter typeConverter; 

    protected int           srcRodSize;
    protected int           srcRodByteSize;

  /**
    * Create a BinaryFileDataSource from an fdl
    * 
    * @param name
    * @param fdl 
    */
   public BinaryFileDataSource( String name, FDLDescriptor fdl )
   {
      this( name, 
            fdl.getRecordDescriptor(), 
            fdl.getBounds(), 
            fdl.getAxisOrdering(), 
            null,   // String URL 
            fdl.endianness
           );
   }
   /** 
     * create a BinaryFileDataSource, set the initial file offset at 0 
     */
    public BinaryFileDataSource(String name,
                                RecordDescriptor recordDesc, 
                                ISBounds theBounds,
                                AxisOrdering ordering, 
                                String URL, 
                                ByteOrder endian ) 
    {
        this( name, recordDesc, theBounds, ordering, URL, endian, 0 );
    }

    /** crate a BinaryFileDataSource, set the initial file offset at the given value 
     */
    public BinaryFileDataSource( String name,
                                 RecordDescriptor recordDesc, 
                                 ISBounds theBounds,
                                 AxisOrdering ordering, 
                                 String URL, 
                                 ByteOrder endian, 
                                 int offset )
   {
      super( name, recordDesc, theBounds, ordering, URL ); // call inherited constructor

      this.recordByteSize = recordDesc.byteSize();
      this.recordNumFields = recordDesc.size();
      this.tempDatum = this.recordDesc.createDatum();
      this.endian = endian;
      //this.typeConverter = new TypeConverter();

      int d = this.dim();
      dimensions = theBounds.getDimensions();
      this.leastSignificantAxis = this.axisOrdering.getLeastSignificantAxis();


      this.setSteps(); // jpf - added helper method to set steps

      this.iterCopy = new IntegerIndexSpaceID( new int[ d ] );
      this.bufferIterCopy = new IntegerIndexSpaceID( new int[ d ] );
      this.bufferUpper = new IntegerIndexSpaceID( new int[ d ] );
      this.iniOffset = offset;

      this.srcRodSize = theBounds.getDimensions().getCoord( 
                          this.leastSignificantAxis );
      this.srcRodByteSize = srcRodSize * recordDesc.byteSize();
   }

    /** 
     * Open the binary data file for this datasource. 
     */
   @Override
    public void activate()
    {      
        super.activate();
        
        try {
           //rdb this.raFile = new RandomAccessFile( this.url,"r" ); 
           this.raFile = new RandomAccessFile( this.getURL(), "r" ); 
           this.fChannel = raFile.getChannel();

        } catch (IOException e){
            new FileError(SDBError.Recoverable,
              "Error opening RandomAccessFile: "+e);
        }
    }	
	
   /** 
    * Close the binary data file for this datasource, and the file can not 
    * be reopened 
    */
   @Override
   public void deactivate()
    {
        super.deactivate();
        
        try {
            this.fChannel.close(); 
            this.raFile.close();
            this.raFile = null;
            
         } catch (Exception e){
         
         // PJR: I won't bother with an error here, since the most likely cause
         // of trouble is deactivating twice. We probably don't care.
         
//            new SDBError(SDBError.Recoverable,
//              "Error closing RandomAccessFile in BinaryFileDataSource.BinaryFileDataSource():"+e);
         }                                      
    }

   /** set the file initial offset 
    */
   public void setFileOffset( long offset)
    {
         this.iniOffset = offset;
    }
   
   /** get the file initial offset 
    */
   public long getFileOffset( )
    {
         return this.iniOffset;
    }


   /** Set a new the AxisOrdering for this datasource.
    *  Note that this operation will have no affect on previously constructed
    *  Iterators or DataBlocks constructed for this DataSource.
    */
   public void setAxisOrdering( AxisOrdering newOrder )
   {
      super.setAxisOrdering(newOrder);
      this.setSteps();
   }



//   /** This method is used by both AttributeOrderBFDS and PointOrderBFDS, Datum methods
//    *  this method is used together with typeConverter to convert data type. will be deprecated
//    * @deprecated
//    */
//   protected void setDatum( int i, int storageType, Datum datum )
//    {
//      switch ( storageType )
//      {
//         case StorageType.Float:
//            datum.setFloat( i, this.typeConverter.floatResult() );
//            break;
//         case StorageType.Double:
//            datum.setDouble( i, this.typeConverter.doubleResult() );
//            break;
//         case StorageType.Int:
//            datum.setInt( i, this.typeConverter.intResult() );
//            break;
//         case StorageType.Short:
//            datum.setShort( i, this.typeConverter.shortResult() );
//            break;
//         case StorageType.Byte:
//            datum.setByte( i, this.typeConverter.byteResult() );
//            break;
//         default:
//            new InternalFaultError( SDBError.Fatal,
//                  "Unsupported storage type code in BinaryFileDataSource.readValue:"
//                        + storageType );
//            break;
//      }
//    }
    /****************/
   
   /** this method is used by both AttributeOrderBFDS and PointOrderBFDS in Datum methods 
    *  This method will replace setDatum(int i, int storageType, Datum datum ), and used 
    *  java.14 byteBuffer
    */
   protected void setDatum( int i, int storageType, Datum datum, ByteBuffer buffer )
    {
          switch(storageType)
          {
	         case StorageType.Float:
	 	        datum.setFloat( i, buffer.getFloat() );
	            break;
	         case StorageType.Double:
	      	    datum.setDouble( i, buffer.getDouble() );
	            break;
	         case StorageType.Int:
	   	        datum.setInt( i, buffer.getInt() );
	            break;
	         case StorageType.Short:
	    	         datum.setShort( i, buffer.getShort() );
	             break;
	         case StorageType.Byte:
   	             datum.setByte( i, buffer.get() );
                 break;
	         default:
	 	        new InternalFaultError(SDBError.Fatal,
                        "Unsupported storage type code in BinaryFileDataSource.readValue:"+storageType);
	         break;
	      }
    }

    // copy bytes from a byte array to specified section of a byte array
    // used to copy byte if the datablock stores data in byte array 
    private void setByteArray ( byte [] bytes, int index, int storageType, 
                                byte [] destArray, int destIndex )
    {
        int i;
        
        switch (storageType){
            case StorageType.Float:
                for ( i =0; i< 4; i++ )
                   destArray[destIndex++] = bytes[index++];
                break;
                
            case StorageType.Double:
                for ( i =0; i< 8; i++ )
                   destArray[destIndex++] = bytes[index++];
                break;
                
            case StorageType.Int:
                for ( i =0; i< 4; i++ )
                   destArray[destIndex++] = bytes[index++];
                break;

            case StorageType.Short:
                 for ( i =0; i< 2; i++ )
                   destArray[destIndex++] = bytes[index++];
                 break;

            case StorageType.Byte:
                 destArray[destIndex++] = bytes[index++];
                 break;
                
            default:
                new IllegalArgumentError(SDBError.Fatal,
                    "Unsupported storage type code in BinaryFileDataSource.readValue:"+storageType);
        }
    }


    /** Print the data members of the DataSource in an easy to read fashion.*/
   @Override
    public String toString(){

        StringBuffer result=new StringBuffer("BinaryFileDataSource: raFile=");

        result.append(raFile);
        result.append("\n");

        if( (this.endian).equals(ByteOrder.BIG_ENDIAN) ){
            result.append("            format=big-endian\n");
        } else {
            result.append("            format=little-endian\n");
        }
        
        result.append(super.toString());

        return new String(result);
    }

    /** Returns true if this BinaryFileDataSource has the same
        pointOrder as the argument BinaryFileDataSource, as
        well as meeting inherited criteria for equality. */
   @Override
    public boolean equals(Object o){

        if(o instanceof BinaryFileDataSource){
            
            BinaryFileDataSource d=(BinaryFileDataSource) o;
    
            return	(this.endian).equals(d.endian) &&
                    super.equals(d);

        } else {
            return false;
        }
            
    }

   @Override
   public int hashCode ()
   {
      throw new UnsupportedOperationException ("Not yet implemented");      
   }
    

    /** For debugging only. */
    private static String toBinaryString(int value, int fieldWidth)
    {
        int mask = (int) ( Math.pow( 2, fieldWidth) - 1 );
        value = value & mask;
        
        String binaryString = Integer.toBinaryString( value );
        int numPadChars = fieldWidth-binaryString.length();
        
        for ( int i=0; i<numPadChars; i++ )
        {
            binaryString="0"+binaryString;
        }
        
       return binaryString;
    }
 

   /** create a BinaryFileDataSource, set the initial file offset at 0
       this is only used for remove compiling errors.
       @deprecated use DataSource.create()
     */
    static BinaryFileDataSource createBinaryFileDataSource(
                                             String name,
                                             RecordDescriptor recordDesc, 
                                             ISBounds theBounds,
                                             AxisOrdering ordering, 
                                             String URL,
                                             ByteOrder    endian,
					                              int          dataOrder ) 
	{
		if( dataOrder == DataCollection.pointOrder )
			return new PointOrderBFDS( name, recordDesc, theBounds, ordering, URL, endian);
		else 
			return new AttributeOrderBFDS(name, recordDesc, theBounds, ordering, URL, endian);
   }

    //debugging tool
    private static void printArray( int[][] a )
      {
        for( int i=0; i < a.length; i++)
          {
             System.out.print("[");
             for( int j =0; j < a[0].length; j++){
               System.out.print(a[i][j] +" ");
              }
             System.out.println("]");
          }
      }
    private static void printArray_oneD( int[] a)
      {
       System.out.print("[");
       for( int i=0; i<a.length; i++)
          {
             System.out.print( a[i] +" " );
          }
       System.out.println( "]" );

      }


    private void setSteps( ) {
        int d = this.dim();
        this.dimensions = this.bounds.getDimensions(); 
        // We have to set the steps array so that it iterates 1 step at a time
        // through most dimensions, except for the one that changes most rapidly.
        // For this special dimension, we'll step by the size of the dimension. 
        // This means we will only "visit" the beginning of a row, column, etc. 
        // This visit is all we need to read an entire line of contiguous data
        // from the file. 
        this.steps = new int[d];
        for ( int i=0; i < d-1; i++ )
           this.steps[ this.axisOrdering.getAxis(i) ] = 1; 
        
        this.steps[ this.axisOrdering.getAxis(d-1) ] 
                    = dimensions.getCoord( this.axisOrdering.getAxis( d-1 ) );
    }




    //--------------- main NOT USABLE -----------------------------------
    // The only purpose of the main is to reference methods that are not
    // currently used, but we might want to have, including debug methods
    // and the setByteArray methods.
    public static void main( String[] args )
    {
       int a[] =  {0,1};
       int a2[][] = {{0,1},{2,3}};
       printArray_oneD( a );
       printArray( a2 );
       
       toBinaryString( 4, 4 );
       
       BinaryFileDataSource bfds = new AttributeOrderBFDS(
                               null, null,null, null, null, ByteOrder.BIG_ENDIAN );
       bfds.setByteArray( (byte[]) null, 0, 0, (byte[]) null, 0 );
    }
 
}
