//
//  BlockCache.java
//  CacheTest
//
//  Created by Philip Rhodes on Tue Apr 15 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//

package edu.unh.sdb.datasource;

/** The BlockCache class is the parent class for all caches that store blocks
  * of data in memory. Typically, these blocks are read from disk in order to
  * achieve a speedup by reducing the total number of reads made to disk. 
  */
public abstract class BlockCache extends Cache {
    
    // The bounds of the CacheBlock
    protected ISBounds              cacheBlockBounds;
    protected int[]                 cacheBlockDimensions;
    protected RegularISPartitioning spacePartitioning; 
    protected long []               tagAxisWeights;
    protected AxisOrdering          spacePartitioningOrdering;
    protected int []                partitioningDimensions;
    protected int                   numCacheBlocks;
    


    BlockCache( Indexable ds, ISBounds iterationSpace, IndexSpaceID blockDimensions, IndexSpaceID cacheArrayDimensions ){
    
        super( ds, iterationSpace );
        
        int numBlocks = 1;
        int d = this.dimensionality;
        
        for ( int i = 0; i < d; i++ ){
        
            numBlocks *= cacheArrayDimensions.getCoord(i);
        }

        
        this.init(blockDimensions, numBlocks, new AxisOrdering(d));
        
     }

    BlockCache( Indexable ds, ISBounds iterationSpace, IndexSpaceID blockDimensions, IndexSpaceID cacheArrayDimensions, AxisOrdering ordering){
       
       super( ds, iterationSpace );
       
       int numBlocks = 1;
       int d = this.dimensionality;
       
       for ( int i = 0; i < d; i++ ){
          
          numBlocks *= cacheArrayDimensions.getCoord(i);
       }
       
       
       this.init(blockDimensions, numBlocks, ordering);
       
    }
    

    BlockCache( Indexable ds, ISBounds iterationSpace, IndexSpaceID blockDimensions, int numBlocks ){
    
        super( ds, iterationSpace );
        
        this.init(blockDimensions, numBlocks, new AxisOrdering(this.dimensionality));
        
    }

    BlockCache( Indexable ds, ISBounds iterationSpace, IndexSpaceID blockDimensions, int numBlocks, AxisOrdering ordering ){
       
       super( ds, iterationSpace );
       
       this.init(blockDimensions, numBlocks, ordering);
       
    }
    

    private void init(IndexSpaceID blockDimensions, int numBlocks, AxisOrdering ordering) {

        int d=this.dimensionality;

        this.cacheBlockBounds     = new ISBounds( blockDimensions );
        this.cacheBlockDimensions = blockDimensions.getCoords( new int[d]);
        this.numCacheBlocks       = numBlocks;
        
//        this.partitioningDimensions = new int[d];
        
//        for ( int i=0; i<d; i++){
//        
//            this.partitioningDimensions[i] = (int) Math.ceil((float)this.iterationSpaceDimensions[i] / this.cacheBlockDimensions[i]);
//            
//            // PJR: We may need this to handle the case where the partition is larger than half the iteration space
//            //this.partitioningDimensions[i] = (this.partitioningDimensions[i] == 0) ? 1 : this.partitioningDimensions[i];
//        }
//
//        this.cachePartitioning = new RegularISPartitioning(new ISBounds(this.partitioningDimensions), iterationSpaceBounds);

         this.spacePartitioning = new RegularISPartitioning(blockDimensions, this.iterationSpaceBounds, ordering);
         
         this.tagAxisWeights = this.spacePartitioning.getAxisWeights();
         
         this.spacePartitioningOrdering = this.spacePartitioning.getOrdering();
         
         this.partitioningDimensions = this.spacePartitioning.getPartitioningDimensions(new int[d]);
         
    }
    
    
    /** Return an ISBounds describing the dimensions of the cache blocks. */
    ISBounds getCacheBlockBounds(){
    
        return (ISBounds) this.cacheBlockBounds.clone();
    }
    
    /** Return an IndexSpaceID describing the dimensions of the cache blocks. */
    IndexSpaceID getCacheBlockDimensions(){
    
        return this.cacheBlockBounds.getDimensions();
    }

    /** Return a reference to the cache block dimensions array. */
    int [] getCacheBlockDimensions(int [] dimensions){
    
        return this.cacheBlockDimensions;
    }

   /** Return the bounds of the region in the index space corresponding to the
     * given tag.
     */
    public final ISBounds tagToBounds(long tag, ISBounds bounds){
    
      int [] coords = new int[3];
      int d = this.dimensionality;
      int [] lowerCoords=bounds.getLowerArray();
      int [] upperCoords=bounds.getUpperArray();
    
      spacePartitioningOrdering.computeCoords(tag, coords, this.tagAxisWeights);
      
      for(int i=0; i<d; i++){
      
         lowerCoords[i] = lowerIterationSpaceCoords[i] + coords[i] * this.cacheBlockDimensions[i];
         upperCoords[i] = lowerCoords[i] + this.cacheBlockDimensions[i] - 1;
      }
      
      return bounds;
    }


   public final int [] tagToPartitioningCoords(long tag, int [] pcoords){
   
      spacePartitioningOrdering.computeCoords(tag, pcoords, this.tagAxisWeights);
      
      return pcoords;
   }

    /** Returns a new ISBounds representing the bounds of the cacheBlock that contains the given id. */
    ISBounds getCacheBlockBounds(IndexSpaceID id){
    
        return this.spacePartitioning.getContainingPartitionBounds(id);
    }

    
    
    /** Sets the ISBounds argument to the bounds of the cacheBlock that contains the given id. */
    void getCacheBlockBounds(IndexSpaceID id, ISBounds pBounds){
    
        this.spacePartitioning.getContainingPartitionBounds(id, pBounds);
    }
        
    /** Returns the ISPartitioning for this cache.*/
    ISPartitioning getCachePartitioning(){
        
        return this.spacePartitioning;
    }
    
    /** Return the number of bytes consumed by the cache. */
    long getCacheMemoryUsage(){
    
      return this.numCacheBlocks * this.getCacheBlockBounds().volume()
             * this.theSource.getRecordDescriptor().byteSize();
      
    }

    /** Returns an ISBoundsIterator that produces bounds corresponding to the cache blocks.*/
    ISBoundsIterator getCacheBlockBoundsIterator(){
        
        return this.spacePartitioning.getPartitionIterator(this.iterationSpaceBounds);
    }


    /** Sets the ISBoundsIterator argument so it produces bounds corresponding to the cache blocks.*/
    void getCacheBlockBoundsIterator(ISBoundsIterator iter){
        
        this.spacePartitioning.getPartitionIterator(this.iterationSpaceBounds, iter);
    }

}
