package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import ncsa.hdf.hdf5lib.*;
import ncsa.hdf.hdf5lib.exceptions.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

public class ChomboDataSource extends DataSource
{
    public final int defaultComponent = 0;
    
    private H5 h5;
    private HDF5Constants h5c;
    
    private int startLv, maxLv, component;
    private boolean server_init;
    private byte[] dataBuffer;
    
    // HDF5 management variables
    private int          fid;
    private String[]     components;
    private String       filename;
    private int          iteration, numLevels;
    private int[]        numBoxes, refRatio;
    private double       time;
    private ISBounds[]   probDomain;
    private ISBounds[][] boxBounds;
    private double[]     tempBuff;
    private int          tempBuffMax;
    private int          bufferSize;

    // Currently, we support reading only a single attribute. This data
    // member specifies which one of the components in the chombo file
    // will be read.
    private int desiredAttribute=0;
    
    public ChomboDataSource(String name)
    {
        this(name, new String());
    }
    
    public ChomboDataSource(String name, String fname)
    {
        super(name);
        filename = fname;
        h5 = new H5();
        h5c = new HDF5Constants();
        startLv = maxLv = -1;
        server_init = false;
        component   = defaultComponent;
        bufferSize  = 0;
    }
    
    public ChomboDataSource(String name, FDLDescriptor fdld)
    {
        this(name, fdld.getDataFileURL());
        this.setRecordDescriptor(fdld.getRecordDescriptor());
        this.desiredAttribute = fdld.getChomboDesiredAttribute();
    }


   /** When this method is called, this subblock methods of this collection will 
    *  read data as bytes, without performing any type conversion. This is meant 
    *  to be used by servers, since they send raw bytes over the network anyway,
    *  and any conversion is wasted effort.
    *
    *  Since Chombo files are already read as bytes, we override this method, and
    *  do nothing.
    */
   public void useBytesOnly(){
   
   }


    
    /**
     * Activates the data source by opening the currently addressed file.
     */
    public void activate()
    {
        if(filename != null)
        {
            this.openHDF5File();
            
            // Set startLv to 0, maxLv to numLevels.
            // This is temporary until we can account for specific level range selection.
            this.startLv = 0;
            this.maxLv   = this.numLevels - 1;
            
            // We need to store the highest resolution problem domain
            this.bounds = this.getProbDomain(this.maxLv);
            
            // Activate only if the file was valid.
            if(this.fid >= 0)
                super.activate();
        }
        else
            System.err.println("No file name is specified.  Cannot activate this data source.");
    }
        
    /**
     * Deactivates the data source by closing the file and clearing all variables.
     */
    public void deactivate()
    {
        if(this.isActivated)
        {
            this.closeHDF5File();
            this.boxBounds      = null;
            this.components     = null;
            this.numBoxes       = null;
            this.probDomain     = null;
            this.refRatio       = null;
            this.dimensionality = 0;
            this.iteration      = 0;
            this.numLevels      = 0;
            this.time           = 0;
            this.axisOrdering   = null;
            super.deactivate();
        }
        else
            System.err.println("This data source has not been activated yet.");
    }
     
    private void closeHDF5File()
    {
        try
        {
            if(this.isActivated)
            {
                h5.H5Fclose(fid);
                fid = 0;
                
                // do a deactivate here
            }
            else
                System.err.println("Chombo data source not activated.  No file is opened.");
        }
        catch(HDF5LibraryException h5le)
        {
            System.out.println("\nError:  Could not close file.  Program must exit.");
            System.exit(ChomboErrors.L_FILE_CLOSE_ERROR);
        }
    }
    
    public void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp)
    {}
    
    public ISBounds getBoxBounds(int lv, int boxNum)
    {
        return this.boxBounds[lv][boxNum];
    }
    
    public String[] getDataComponents()
    {
        return this.components;
    }
    
    public int getDimensionality()
    {
        return this.dimensionality;
    }
    
    public String getFilename()
    {
        return this.filename;
    }
    
    public int getIteration()
    {
        return this.iteration;
    }
    
    public int getNumBoxes(int lv)
    {
        return this.numBoxes[lv];
    }
    
    public int getNumLevels()
    {
        return this.numLevels;
    }
    
    public ISBounds getProbDomain(int lv)
    {
        return this.probDomain[lv];
    }
    
    public int getRefRatio(int lv)
    {
        return this.refRatio[lv];
    }
    
    public double getTime()
    {
        return this.time;
    }
    
    private ISBounds[] loadBoxDims(int lv)
    {
        long[] numBoxes    = new long[1];
        long[] maxNumBoxes = new long[1];
        int[]  allBoxCoords;
        int[] thisCoords = new int[this.dimensionality * 2];
        ISBounds[] allBoxDims = null;
        
        try
        {
            int lvlGrpID      = h5.H5Gopen(fid, ("level_" + lv));
            int boxesDsetID   = h5.H5Dopen(lvlGrpID, "boxes");
            int boxesDspaceID = h5.H5Dget_space(boxesDsetID);
            h5.H5Sget_simple_extent_dims(boxesDspaceID, numBoxes, maxNumBoxes);
            int memSpaceID    = h5.H5Screate_simple(1, numBoxes, maxNumBoxes);
            allBoxCoords = new int[(int)numBoxes[0] * (this.dimensionality * 2)];
            
            // Create a compound data type.
            // 2D boxes have 4 ints; 3D boxes have 6 ints
            int boxesDtypeID = h5.H5Tcreate(h5c.H5T_COMPOUND, ((this.dimensionality * 2) * 4));
            int offset = 0;
            if(this.dimensionality == 2)
            {
                h5.H5Tinsert(boxesDtypeID, "lo_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "lo_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "hi_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "hi_j", (4 * (offset)),   h5c.H5T_NATIVE_INT);
            }
            else
            {
                h5.H5Tinsert(boxesDtypeID, "lo_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "lo_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "lo_k", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "hi_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "hi_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(boxesDtypeID, "hi_k", (4 * (offset)),   h5c.H5T_NATIVE_INT);
            }
            
            // Read all box coordinates from the file
            h5.H5Dread(boxesDsetID, boxesDtypeID, memSpaceID, boxesDspaceID,
                       h5c.H5P_DEFAULT, allBoxCoords);
            
            allBoxDims = new ISBounds[(int)numBoxes[0]];
            for(int b = 0; b < (int)numBoxes[0]; b++)
            {
                for(int i = 0; i < this.dimensionality * 2; i++)
                    thisCoords[i] = allBoxCoords[b * this.dimensionality * 2 + i];
                
                allBoxDims[b] = new ISBounds(this.dimensionality, thisCoords);
            }
            
            h5.H5Sclose(memSpaceID);
            h5.H5Sclose(boxesDspaceID);
            h5.H5Tclose(boxesDtypeID);
            h5.H5Dclose(boxesDsetID);
            h5.H5Gclose(lvlGrpID);
            
            return allBoxDims;
        }
        catch(HDF5LibraryException h5le)
        {
            new SDBError(SDBError.Fatal, "Error:  This file does not contain the appropriate data box.\n" +
                               "        The program must exit.");
            //System.exit(ChomboErrors.L_BOX_NOT_FOUND);
        }
        catch(HDF5Exception h5e)
        {
            new SDBError(SDBError.Fatal,"Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            //System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return null;
    }
    
    private String[] loadDataComponents()
    {
        try
        {
            String[] components = null;
            byte[] b_charBuffer;
            int thisCompAttrID, numChars, typeID;
            int[] numComponents = new int[1];

            // First, get the number of components
            int numCompAttrID = h5.H5Aopen_name(fid, "num_components");
            h5.H5Aread(numCompAttrID, h5c.H5T_NATIVE_INT, numComponents);
            components = new String[numComponents[0]];
            h5.H5Aclose(numCompAttrID);

            // Next, we get the actual names of each data component
            for(int i = 0; i < numComponents[0]; i++)
            {
                thisCompAttrID = h5.H5Aopen_name(fid, ("component_" + i));
                typeID   = h5.H5Aget_type(thisCompAttrID);
                numChars = h5.H5Tget_size(typeID);
                b_charBuffer = new byte[numChars];
                h5.H5Aread(thisCompAttrID, typeID, b_charBuffer);
                components[i] = new String(b_charBuffer);
                h5.H5Aclose(thisCompAttrID);
            }
            return components;
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate field\n" +
                               "        for specifying the number of data components and/or the component\n" +
                               "        fields.  The program must exit.");
            System.exit(ChomboErrors.L_INVALID_COMPONENTS);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return null;
    }
    
    private int loadDimensionality()
    {   
        try
        {
            int[] dimBuffer = new int[1];
            int globalGrpID = h5.H5Gopen(fid, "/Chombo_global");
            int dimsAttrID = h5.H5Aopen_name(globalGrpID, "SpaceDim");
            h5.H5Aread(dimsAttrID, h5c.H5T_NATIVE_INT, dimBuffer);
            if(dimBuffer[0] < 2 || dimBuffer[0] > 3)
            {
                System.err.println("Error:  This file does not contain 2D or 3D data, and is\n" +
                                   "        not a valid Chombo file.  Please choose another file.\n" +
                                   "        The program must exit.");
                System.exit(ChomboErrors.L_INVALID_DIMENSIONS);
            }
            h5.H5Aclose(dimsAttrID);
            h5.H5Gclose(globalGrpID);
            return dimBuffer[0];
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate field\n" +
                               "        for specifying the number of dimensions.  The program must\n" +
                               "        exit.");
            System.exit(ChomboErrors.L_BAD_FILE_FORMAT);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return 0;
    }
    
    private int loadIteration()
    {        
        try
        {
            int[] iterationBuffer = new int[1];
            int rootGrpID = h5.H5Gopen(fid, "/");
            int iterAttrID = h5.H5Aopen_name(fid, "iteration");
            h5.H5Aread(iterAttrID, h5c.H5T_NATIVE_INT, iterationBuffer);
            h5.H5Aclose(iterAttrID);
            h5.H5Gclose(rootGrpID);
            return iterationBuffer[0];
        }
        catch(HDF5LibraryException h5le)
        {
            // It is not an error if "iteration" doesn't exist.  It just means
            // there is no specified cycle.  -1 represents this value for us.
            return -1;
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return 0;
    }
    
    private int loadNumBoxes(int lv)
    {       
        try
        {
            int lvGrpID = h5.H5Gopen(fid, ("level_" + lv));
            int boxesDsID = h5.H5Dopen(lvGrpID, "boxes");
            int boxesSpaceID = h5.H5Dget_space(boxesDsID);
            long[] numBoxesBuffer = new long[1];
            h5.H5Sget_simple_extent_dims(boxesSpaceID, numBoxesBuffer, null);
            h5.H5Dclose(boxesDsID);
            h5.H5Gclose(lvGrpID);
            return (int)numBoxesBuffer[0];
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate level group\n" +
                               "        for data boxes.  The program must exit.");
            System.exit(ChomboErrors.L_LEVEL_NOT_FOUND);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return -1;
    }
    
    private int loadNumLevels()
    {        
        try
        {
            int[] numLvlBuffer = new int[1];
            int numLvlAttrID = h5.H5Aopen_name(fid, "num_levels");
            h5.H5Aread(numLvlAttrID, h5c.H5T_NATIVE_INT, numLvlBuffer);
            h5.H5Aclose(numLvlAttrID);
            return numLvlBuffer[0];
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate field\n" +
                               "        for specifying the number of refinement levels.  The program must\n" +
                               "        exit.");
            System.exit(ChomboErrors.L_NUM_LEVELS_NOT_FOUND);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return -1;
    }
    
    private ISBounds loadProbDomain(int lv, int rank)
    { 
        try
        {
            int lvlGrpID = h5.H5Gopen(fid, ("level_" + lv));
            int probDomainAttrID = h5.H5Aopen_name(lvlGrpID, "prob_domain");
            int probDomainTypeID = h5.H5Tcreate(h5c.H5T_COMPOUND, ((rank * 2) * 4));
            int offset = 0;
            if(rank == 2)
            {
                h5.H5Tinsert(probDomainTypeID, "lo_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "lo_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "hi_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "hi_j", (4 * (offset)),   h5c.H5T_NATIVE_INT);
            }
            else
            {
                h5.H5Tinsert(probDomainTypeID, "lo_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "lo_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "lo_k", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "hi_i", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "hi_j", (4 * (offset++)), h5c.H5T_NATIVE_INT);
                h5.H5Tinsert(probDomainTypeID, "hi_k", (4 * (offset)),   h5c.H5T_NATIVE_INT);
            }            
            
            int[] probDomain = new int[rank * 2];
            h5.H5Aread(probDomainAttrID, probDomainTypeID, probDomain);
            
            ISBounds thisISBounds = new ISBounds(rank, probDomain);
            
            h5.H5Tclose(probDomainTypeID);
            h5.H5Aclose(probDomainAttrID);
            h5.H5Gclose(lvlGrpID);
            
            return thisISBounds;
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate data for\n" +
                               "        problem domain.  The program must exit.");
            System.exit(ChomboErrors.L_PROB_DOMAIN_NOT_FOUND);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return null;
    }
    
    private int loadRefRatio(int lv)
    {
        try
        {
            int[] refRatioBuffer = new int[1];
            int lvlGrpID = h5.H5Gopen(fid, ("level_" + lv));
            int refRatioAttrID = h5.H5Aopen_name(lvlGrpID, "ref_ratio");
            h5.H5Aread(refRatioAttrID, h5c.H5T_NATIVE_INT, refRatioBuffer);
            h5.H5Aclose(refRatioAttrID);
            h5.H5Gclose(lvlGrpID);
            return refRatioBuffer[0];
        }
        catch(HDF5LibraryException h5le)
        {
            System.err.println("Error:  This file does not contain the appropriate data for\n" +
                               "        refinement ratio.  The program must exit.");
            System.exit(ChomboErrors.L_REF_RATIO_NOT_FOUND);
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return 0;
    }
    
    private double loadTime()
    {
        int rootGrpID = 0, level0GrpID, timeAttrID;
        double[] timeBuffer = new double[1];
        
        try
        {
            // First, try the root group.
            rootGrpID = h5.H5Gopen(fid, "/");
            timeAttrID = h5.H5Aopen_name(rootGrpID, "time");
            h5.H5Aread(timeAttrID, h5c.H5T_NATIVE_DOUBLE, timeBuffer);
            h5.H5Aclose(timeAttrID);
            h5.H5Gclose(rootGrpID);
            return timeBuffer[0];
        }
        catch (HDF5LibraryException h5le)
        {
            try
            {
                // Time wasn't found in root group; Close root group and try "level_0".
                h5.H5Gclose(rootGrpID);
                level0GrpID = h5.H5Gopen(fid, "/level_0");
                timeAttrID = h5.H5Aopen_name(level0GrpID, "time");
                h5.H5Aread(timeAttrID, h5c.H5T_NATIVE_DOUBLE, timeBuffer);
                h5.H5Aclose(timeAttrID);
                h5.H5Gclose(level0GrpID);
                return timeBuffer[0];
            }
            catch(HDF5LibraryException h5le_2)
            {
                System.err.println("Error:  This file does not contain the appropriate field\n" +
                                   "        for specifying timestep.  The program must exit.");
                System.exit(ChomboErrors.L_TIMESTEP_NOT_FOUND);
            }
            catch(HDF5Exception h5e_2)
            {
                System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                                   "        associated with an attribute in the file.  Check your code.\n\n" +
                                   "NOTICE:  This error should NOT appear to application users.");
                System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
            }
        }
        catch(HDF5Exception h5e)
        {
            System.err.println("Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        return 0.0;
    }
    
    // Rename this to "activate"
    private void openHDF5File()
    {
        try
        {
            if(!this.isActivated)
            {
                // Obtain the fid for future access by the user (if necessary)
                // Put this in the constructor
                this.fid = h5.H5Fopen(filename, h5c.H5F_ACC_RDONLY, h5c.H5P_DEFAULT);
                
                // Load all the HDF5 file information
                this.numLevels      = this.loadNumLevels();
                this.dimensionality = this.loadDimensionality();
                this.components     = this.loadDataComponents();
                this.iteration      = this.loadIteration();
                this.time           = this.loadTime();
                int[] axes = new int[this.dimensionality];
                for(int d = 0; d < this.dimensionality; d++)
                    axes[d] = (this.dimensionality - 1) - d;
                
                this.axisOrdering   = new AxisOrdering(axes);
                this.numBoxes       = new int[numLevels];
                this.refRatio       = new int[numLevels];
                this.probDomain     = new ISBounds[numLevels];
                this.boxBounds      = new ISBounds[numLevels][];
                for(int lv = 0; lv < numLevels; lv++)
                {
                    this.numBoxes[lv] = this.loadNumBoxes(lv);
                    this.refRatio[lv] = this.loadRefRatio(lv);
                    this.probDomain[lv] = this.loadProbDomain(lv, this.dimensionality);
                    this.boxBounds[lv]  = this.loadBoxDims(lv);
                }
            }
            else
                System.err.println("\nChombo data source already activated.  Please close\n" +
                                   "the currently opened file first before opening a new one.");
        }
        catch(HDF5LibraryException h5le)
        {
            System.out.printf("\nError (%2d.%2d):  Could not open \"" + filename + "\".\n" +
                              "                    File may not be a valid HDF5 file or may not exist.  Try another file.\n",
                              h5le.getMajorErrorNumber(), h5le.getMinorErrorNumber());
            this.fid = -1;
        }
        catch(HDF5Exception h5e)
        {
            System.out.printf("\nGeneral HDF5 Exception:\n" +
                              "%s\n\n",h5e.getMessage());
            this.fid = -1;
        }
    }
    
    public void setStartLv(int sLv)
    {
        startLv = sLv;
    }
    
    public void setMaxLv(int mLv)
    {
        maxLv = mLv;
    }
    
    public void setComponent(int cmp)
    {
        component = cmp;
    }
    
    /**
     * Creates a new byte buffer and returns it, based on the start and maximum
     * levels specified, and the number of query boxes requested.
     * 
     * @param startLv Starting refinement level chosen
     * @param maxLv Ending refinement level chosen
     * @param queries Query box bounds
     * @return Bytes array large enough to hold the requested data
     */
    public byte[] makeByteBuffer(byte[] bytes, DSPartitionedStencil stencil)
    {
        int numBytes = 0;
        
        // First, calculate the number of bytes needed for the header.
        numBytes += 8;                                      // numDims, numLevels
        numBytes += (this.maxLv - this.startLv + 1) * 4;    // refRatios for requested levels
        numBytes += 4;                                      // number of query boxes
        
        // Next, calculate the number of bytes needed for each query box
        for(int q = 0; q < stencil.size(); q++)
        {
            // No longer necessary; see server_subblock (line 706) for details.
            //numBytes += (this.dimensionality * 2 * 4);          // Query box bounds
            
            // Calculate the size of the per-level data needed
            for(int lv = startLv; lv <= maxLv; lv++)
            {
                ISBounds thisBox = stencil.getIndexedBounds(q).copy();
                
                // Calculate total refinement ratio
                for(int r = lv; r < maxLv; r++)
                    thisBox.shrinkByFactor(this.refRatio[r]);
                
                numBytes += 8;                                  // numBoxes, numDataElems
                numBytes += (this.numBoxes[lv] *                // Bounds for each box in this level
                             this.dimensionality * 2 * 4);
                
                // Calculate the number of data elements for this level.
                int numElems = calculateNumElems(lv, thisBox);
                
                numBytes += numElems * this.recordDesc.size();
            }
        }
        
        //numBytes += 1000000 ; // PJR Fudgefactor 
        
        // If the bytes array is uninitialized or too small, make a new one
        if(bytes == null || bytes.length < numBytes)
            bytes = new byte[numBytes];
        
        // Store the current buffer size
        bufferSize = numBytes;
        
        // Return the bytes array
        return bytes;
    }
    
    public int writeHeader(byte[] bytes, DSPartitionedStencil stencil)
    {
        ByteBuffer buffer = ByteBuffer.wrap(bytes);
        
        // Write number of dimensions and levels
        buffer.putInt(this.dimensionality);
        buffer.putInt(this.maxLv - this.startLv + 1);
        
        // Write the refinement ratio values for each chosen level
        for(int lv = startLv; lv <= maxLv; lv++)
        {
            buffer.putInt(this.refRatio[lv]);
        }
        
        // Write the number of query boxes
        buffer.putInt(stencil.size());
        
        // Store the buffer into this datasource
        this.dataBuffer = buffer.array();
        
        // Return the current offset in the buffer.
        return buffer.position();
    }
    
    public int calculateNumElems(int lv, ISBounds bounds)
    {
        int numElems = 0;
        
        for(int b = 0; b < this.numBoxes[lv]; b++)
        {
            // Get the intersection of the current box and the bounds
            ISBounds thisBox = bounds.intersection(this.boxBounds[lv][b]);
            
            // If there is a valid intersection, add its volume to the number of elements.
            if(thisBox != null)
                numElems += thisBox.volume();
        }
        
        return numElems;
        //return numElems * this.recordDesc.size(); //PJR
    }




    
    public int server_subblock(byte[] bytes, ISBounds inputSrcBounds, int byteOffset)
    {
        ByteBuffer buffer = ByteBuffer.wrap(bytes);
        buffer.position(byteOffset);

        
        // This is no longer necessary, since we're using the bounds given by a
        // stencil anyway, which will be reused on the client.
//        // Write the bounds for this query box (lower first, then upper)
//        buffer.asIntBuffer().put(inputSrcBounds.getLowerArray());
//        buffer.position(buffer.position() + (inputSrcBounds.dim() * 4));
//        buffer.asIntBuffer().put(inputSrcBounds.getUpperArray());
//        buffer.position(buffer.position() + (inputSrcBounds.dim() * 4));
        
        for(int lv = startLv; lv <= maxLv; lv++)
        {            
            // Scale down the inputSrcBounds for this level
            ISBounds inputScaled = inputSrcBounds.copy();
            for(int r = lv; r < maxLv; r++)
                inputScaled.shrinkByFactor(this.refRatio[r]);
            
//            System.out.printf("\nOriginal query:     %s\nScaled to level %d: %s\n",
//                              inputSrcBounds, lv, inputScaled);
            
            
            
            // Write the number of valid boxes for this level
            buffer.putInt(getNumValidBoxes(lv, inputScaled));
            
            // Write the number of valid data elements for this level
            buffer.putInt(calculateNumElems(lv, inputScaled));
            
            for(int b = 0; b < this.numBoxes[lv]; b++)
            {
                // Write the intersections of each box and the inputSrcBounds
                // if the intersection is valid
                ISBounds thisBox = inputScaled.intersection(this.boxBounds[lv][b]);
                if(thisBox != null)
                {
                    buffer.asIntBuffer().put(thisBox.getLowerArray());
                    buffer.position(buffer.position() + (thisBox.dim() * 4));
                    buffer.asIntBuffer().put(thisBox.getUpperArray());
                    buffer.position(buffer.position() + (thisBox.dim() * 4));
                    
                    
                }
                
                
            }
            
            
            
            // Get the data values for this level and write them into the buffer
            tempBuff = getData(lv, inputScaled);
            
            // Switch byte order if necessary
            buffer.order(ByteOrder.nativeOrder());
            
            buffer.asDoubleBuffer().put(tempBuff, 0, tempBuffMax);
            // Restore buffer ordering before continuiung.
            buffer.order(ByteOrder.BIG_ENDIAN);
            
            
            buffer.position(buffer.position() + tempBuffMax * 8); 
        }
        
        // Store the buffer into this datasource
        this.dataBuffer = buffer.array();
        
        
        return buffer.position();
    }
    
    /**
     * @return The data buffer currently stored in the data source.  This is
     *         typically the most recently used buffer, whether a header was
     *         written or server_subblock was called to write data.
     */
    public byte[] getBuffer()
    {
        return this.dataBuffer;
    }
    
    /**
     * @return The size limit of the previously created data buffer
     */
    public int getBufferSize()
    {
        return this.bufferSize;
    }
    
    public double[] getData(int lv, ISBounds bound)
    {
        long[] start  = new long[1];
        long[] stride = new long[1];
        long[] count  = new long[1];
        long[] block  = new long[1];
        
        long[] startData = {0};
        long[] countData = new long[1];
        
        
        this.axisOrdering= new AxisOrdering(new int[] {2,1,0}); //PJR: Fortran ORDER!!
        
        int attrOffset; //PJR offset of desired attribute from beginning of box data
        int boxOffset=0; //PJR offset of current box data
        
        tempBuffMax = calculateNumElems(lv, bound);
        
        if(tempBuff == null || tempBuff.length < tempBuffMax)
            tempBuff = new double[tempBuffMax];
        
        for(int b = 0; b < this.numBoxes[lv]; b++)
        {


        
            // Create a new reference to a dataspace since we are creating a new
            // hyperslab for a new box.
            int dataspace = -1;
            
            // Get the intersection of the bound and current box.
            ISBounds thisBox = bound.intersection(this.boxBounds[lv][b]);
            
            if(thisBox != null)
            {

               // This is the offset for the desired attribute from the
               // beginning of the Chombo box data. We're just skipping over the
               // preceeding attributes.
               attrOffset = this.boxBounds[lv][b].volume() * this.desiredAttribute;
            
                // If the intersection exists, move the intersecting box to
                // correspond to the original box sitting at the origin.
                int[] origOffset = this.boxBounds[lv][b].copyLowerArray();
                for(int i = 0; i < origOffset.length; i++)
                    origOffset[i] = -origOffset[i];
                thisBox.move(origOffset);
                 
            
                // Set the axisWeights for the original box
                axisWeights = axisOrdering.makeAxisWeights(this.boxBounds[lv][b]);
            
                // Set the countData value to the volume of thisBox
                countData[0] = thisBox.volume();
                
            
                // Set stride to the original box's rod length.
                // Set count  to the number of rods for slice of thisBox.
                // Set block  to thisBox's rod length
                // NB: Chombo uses Fortran order {2,1,0}

                stride[0] = this.boxBounds[lv][b].getDimension(0);
                count[0]  = thisBox.getDimension(1);
                block[0]  = thisBox.getDimension(0);

            
                if(this.dimensionality == 3)
                {
                  int sliceVolume = this.boxBounds[lv][b].getDimension(0) * this.boxBounds[lv][b].getDimension(1);
                  
                  //offset of lower corner of intersection box in the current box dataspace
                  int thisBoxOffset = computeOffset(thisBox.getLowerArray());
                  
                  for(int slice = 0; slice < thisBox.getDimension(2); slice++)
                    {
                        // In 3D, the starting offset is affected by the size of
                        // all previous boxes, the starting location of the current
                        // box, and the current plane slice to be selected.

                        start[0] = boxOffset + attrOffset + thisBoxOffset +
                                   (slice * sliceVolume) ;
                                   
                    
                        // Select the current hyperslab.
                        dataspace = selectHyperslab(lv, dataspace, start, stride, count, block);
                    }
                }
                else
                {
                    // In 2D, the starting offset is affected only by the size of
                    // previous boxes and the starting location of the current box,
                    // since 2D exists on a single plane.
                  start[0] = boxOffset + attrOffset  + computeOffset(thisBox.getLowerArray());
                
                    // Select the hyperslab
                    dataspace = selectHyperslab(lv, dataspace, start, stride, count, block);
                }
            
                // Write data into the raw data array
                tempBuff = writeHyperslab(tempBuff, lv, startData, countData, dataspace);

            
                // When finished with this box, increment the startData by countData.
                startData[0] += countData[0];
            }
            else
            {
//                System.out.println("This is an empty box; no data to read for it.");
            }
            
            
            //PJR  Each box consists of nAttr attributes, so in order to advance
            // boxOffset to the beginning of the next box's data, we add
            // this.numAttributes*currentBoxVolume 
            boxOffset += this.numAttributes * this.boxBounds[lv][b].volume();

            
        }
        
        // Return the raw data array.
        return tempBuff;
    }
    
    public double[] writeHyperslab(double[] data, int lv, long[] startData, long[] countData, int dataspace)
    {
        try
        {
            int levelGrp  = h5.H5Gopen(fid, ("level_" + lv));
            int dataset   = h5.H5Dopen(levelGrp, "data:datatype=0");
            int datatype  = h5.H5Dget_type(dataset);
            long[] memspaceSize = {data.length};
            

            
            // Create memory space so HDF5 can write to the buffer
            int memspace = h5.H5Screate_simple(1, memspaceSize, null);
            
            // Select the entire temporary buffer memory space
            h5.H5Sselect_hyperslab(memspace, h5c.H5S_SELECT_SET,
                                   startData, null, countData, null); 
                                               
                                                                                                                   
            // Read from file to the temporary buffer
            h5.H5Dread(dataset, datatype, memspace, dataspace,
                          h5c.H5P_DEFAULT, data);
                        
            h5.H5Sclose(dataspace);
            h5.H5Tclose(datatype);
            h5.H5Dclose(dataset);
            h5.H5Gclose(levelGrp);
            return data;
        }
        catch(HDF5LibraryException h5le)
        {
            //new SDBError(SDBError.Fatal,"Error:  This file does not contain the appropriate data box.\n" +
            //                   "        The program must exit.");
            new SDBError(SDBError.Fatal,"HDF5 Error:  "+h5le);
                   
                               
            //System.exit(ChomboErrors.L_INVALID_DATA);
        }
        catch(HDF5Exception h5e)
        {
            new SDBError(SDBError.Fatal,"Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            //System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        
        return data;
    }
    
    public int selectHyperslab(int lv, int dataspace, long[] start, long[] stride, long[] count, long[] block)
    {
        try
        {
            int levelGrp  = h5.H5Gopen(fid, ("level_" + lv));
            int dataset   = h5.H5Dopen(levelGrp, "data:datatype=0");
            int datatype  = h5.H5Dget_type(dataset);
            
            if(dataspace < 0)
            {
                dataspace = h5.H5Dget_space(dataset);
            
                // Make a new hyperslab selection
                h5.H5Sselect_hyperslab(dataspace, h5c.H5S_SELECT_SET,
                                       start, stride, count, block);
            }
            else
            {
                // Add to an old hyperslab selection
                h5.H5Sselect_hyperslab(dataspace, h5c.H5S_SELECT_OR,
                                       start, stride, count, block);
            }
            
            h5.H5Tclose(datatype);
            h5.H5Dclose(dataset);
            h5.H5Gclose(levelGrp);
            
            return dataspace;
        }
        catch(HDF5LibraryException h5le)
        {
            new SDBError(SDBError.Fatal,"Error:  This file does not contain the appropriate data box.\n" +
                               "        The program must exit.");
            //System.exit(ChomboErrors.L_INVALID_DATA);
        }
        catch(HDF5Exception h5e)
        {
            new SDBError(SDBError.Fatal,"Error:  Programming bug.  The specified attribute ID is not\n" +
                               "        associated with an attribute in the file.  Check your code.\n\n" +
                               "NOTICE:  This error should NOT appear to application users.");
            //System.exit(ChomboErrors.G_INCORRECT_ID_TYPE);
        }
        
        return dataspace;
    }
    
    public int getNumValidBoxes(int lv, ISBounds inputSrcBounds)
    {
        int numBoxes = 0;
        for(int b = 0; b < this.numBoxes[lv]; b++)
        {
            if(inputSrcBounds.intersection(this.boxBounds[lv][b]) != null)
                numBoxes++;
        }
        return numBoxes;
    }
    
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds)
    {}

    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, FieldIDMapper mp)
    {}
}