/*******************************************************************************
 * CompositeBlock.java Copyright 2000 University of New Hampshire Durham, NH
 * 03824
 * 
 * -------------- Edit history (as of 11/28/04) ---------------
 * 07/13/05 rdb: Trivial correction of major error in getValues(int)
 * 01/27/05 rdb: changed subblock methods with RecordSpec or FieldIDMapper arguments
 *                  so that null is valid.
 * 01/21/05 rdb: Added constructor specifying attribute order, rather than point order.
 * 01/13/05 rdb: changed to reflect new RecordDescriptor StorageType interface.
 *               Add convenience constructors with RecordSpec and FieldIDMapper
 *                   to match those of BasicBlock.
 * 01/06/05 rdb: create Datum objects from RecordDescriptor
 * 
 * 11/28/04 rdb: -getValues( int attr ) get <type>( isid, attr ) methods 
 *               -replaced 2-element field array with private FieldInfo class 
 *               -changed setDatum method to support byte 
 *                and to use Datum.get <type>code instead of assuming FloatDatum
 *                and doing cast here
 * ------------------------------------------------------------
 * 
 * Xuan Tang xtang@cs.unh.edu update: 8/20/03 update datum, subblock functions,
 * test setDatum
 * 
 * last update 12/12/03 add public DataBlock subblock(RecordSpec rs, ISBounds
 * srcBounds)
 *  
 ******************************************************************************/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;
import java.nio.*;

/*
 * Composite Block contains a list a BasicBlcoks. Attributes with the same
 * datatype( int, float.. ) are stored in the same BasicBlock. Therefore,
 * CompositeBlock can used to store attributes with mixed datatypes
 */
public class CompositeBlock extends DataBlock
{

   private BasicBlock[]    basicBlocks;
   private FieldIDMapper[] idMapper;
   private int             numOfAttributes;
   private Datum           nullDatum;       //used for fast datum access
   private int             numBasicBlocks;

   private class FieldInfo
   {
      public int bb   = 0; // which basic block is the attribute in
      public int attr = 0; // what is this attributes index in its basic block
   }

   /**
    * Construct a datablock with the given bounds and RecordDescriptor.
    * Internal data arrays will be allocated according to the bounds and recordDesc.
    * The attributes will be stored in attribute order, which means that each 
    * attribute will be in its own BasicBlock. This will be more efficient if the
    * applications want to extract all the values for a particular attribute.
    */
   public CompositeBlock( ISBounds isb, RecordDescriptor recordDesc )
   {
      this( isb, recordDesc, true );
   }
   
   /**
    * Construct a datablock with the given bounds and RecordDescriptor.
    * Internal data arrays will be allocated according to the bounds and recordDesc.
    * The attributes will be stored by type. If the third argument is true, they will
    * be stored entirely by attribute, which means that each attribute will be in its
    * own BasicBlock.  This will be more efficient if the
    * applications want to extract all the values for a particular attribute.
    * If the third attribute is false, the attributes of the same storage type will
    * be stored together in a single array in point order. That is, the attributes of
    * the same type for a given spatial position will be adjacent in the array.
    */
   public CompositeBlock( ISBounds isb, RecordDescriptor recordDesc, boolean attrOrder )
   {
      super( isb, recordDesc );
      numOfAttributes = this.getNumAttributes();
      if ( attrOrder )
      {
         this.setDataOrder( DataCollection.attributeOrder );
         attrOrder( isb );
      }
      else
      {
        this.setDataOrder( DataCollection.pointOrder );
        pointOrder( isb );
      }
      nullDatum = this.createDatum();      
   }

   /** Initialize based on attribute Order ***/
   private void attrOrder( ISBounds isb )
   {
      numBasicBlocks = this.getNumAttributes();
      basicBlocks = new BasicBlock[ numBasicBlocks ];
      idMapper = new FieldIDMapper[ numBasicBlocks ];

      int[] componentFIDs = new int[] {0};
      int[] compositeFIDs = new int[1];

      AttributeDescriptor bbAttrDesc[] = new AttributeDescriptor[1];

      for ( int i = 0; i < numBasicBlocks; i++ )
      {
            compositeFIDs[ 0 ] = i;
            idMapper[i] = new FieldIDMapper( componentFIDs, compositeFIDs );
            bbAttrDesc[ 0 ] = recordDesc.getAttribute( i );
            this.basicBlocks[i] = new BasicBlock( isb, new RecordDescriptor( bbAttrDesc ));
      }
   }
   
   /** Initialize based on point order by storage type ***/
   private void pointOrder( ISBounds isb )
   {
      int size = this.bounds.volume();
      Values valueOfCompact;

      int numOfStorageType = recordDesc.getNumUsedTypes();
      basicBlocks = new BasicBlock[numOfStorageType];
      idMapper = new FieldIDMapper[numOfStorageType];
      numBasicBlocks = numOfStorageType;

      int storageType, numFieldsOfType;
      int[] componentFIDs, compositeFIDs;
      int componentIndex;  // use as the index for componet and composite FieldIDMapper
      AttributeDescriptor[] attriDesc;

      for ( int i = 0; i < numOfStorageType; i++ )
      {
         storageType = recordDesc.getUsedType( i );
         numFieldsOfType = recordDesc.getUsedTypeCount( i );
         compositeFIDs = new int[numFieldsOfType];
         componentFIDs = new int[numFieldsOfType];
         attriDesc = new AttributeDescriptor[numFieldsOfType];

         if ( storageType == StorageType.Double )
            valueOfCompact = new DoubleValues( size * numFieldsOfType );
         else if ( storageType == StorageType.Float )
            valueOfCompact = new FloatValues( size * numFieldsOfType );
         else if ( storageType == StorageType.Int )
            valueOfCompact = new IntValues( size * numFieldsOfType );
         else if ( storageType == StorageType.Short )
            valueOfCompact = new ShortValues( size * numFieldsOfType );
         else if ( storageType == StorageType.Byte )
            valueOfCompact = new ByteValues( size * numFieldsOfType );
         else
         {
            valueOfCompact = new FloatValues(); // just for an initialization
            new InternalFaultError( SDBError.Recoverable,
                  "Unknown storageType." );
         }
         componentIndex = 0;
         for ( int j = 0; j < numOfAttributes; j++ )
         {
            if ( storageType == recordDesc.getStorageType( j ) )
            {
               compositeFIDs[componentIndex] = j;
               componentFIDs[componentIndex] = componentIndex;
               attriDesc[componentIndex] = recordDesc.getAttribute( j );
               componentIndex++ ;
            }
         }
         idMapper[i] = new FieldIDMapper( componentFIDs, compositeFIDs );

         //System.out.println("*** fieldIDMapper is :" +
         // idMapper[i].toString() );
         //System.out.println(" *** basicblock dataType is : " +
         // storageType );
         this.basicBlocks[i] = new BasicBlock( isb,
               new RecordDescriptor( attriDesc ), valueOfCompact );
      }
      nullDatum = this.createDatum();
   }

   /**
    * Construct a datablock with the given bounds and RecordDescriptor. Select a 
    * subset of attributes based on the RecordSpec.
    * Appropriate internal data arrays will be allocated.
    */
   public CompositeBlock( ISBounds isb, RecordDescriptor recordDesc, RecordSpec rs )
   {
      this( isb, recordDesc.project( rs ));
   }

   /**
    * Construct a datablock with the given bounds and RecordDescriptor. Select a 
    * subset of attributes based on the RecordSpec.
    * Appropriate internal data arrays will be allocated.
    */
   public CompositeBlock( ISBounds isb, RecordDescriptor recordDesc, FieldIDMapper fm )
   {
      this( isb, recordDesc.project( fm ));
   }

   /**
    * given a fieldID (fieldID of this compositeBlock, return a array with
    * length of two, first element of the array represents the basicBlock
    * number, and second element represents the corresponding fieldID in the
    * basicBlock.
    */
   public FieldInfo locateAttribute( int fieldID )
   {
      FieldInfo fieldLocation = new FieldInfo();

      //int fieldLocation = new int[2];

      int numOfMapper = idMapper.length;
      int[] compositeFid;
      int componentFidNum;
      for ( int i = 0; i < numOfMapper; i++ )
      {
         compositeFid = idMapper[i].getCompositeFields();

         componentFidNum = compositeFid.length;
         for ( int j = 0; j < componentFidNum; j++ )
            if ( compositeFid[j] == fieldID )
            {
               fieldLocation.bb = i;
               fieldLocation.attr = j;
               //fieldLocation[0] = i;
               //fieldLocation[1] = j;
               //System.out.println("*** FieldLocation "+ i + " "+ j);
               return fieldLocation;
            }
      }
      new InternalFaultError( SDBError.Recoverable, "undefined FieldID." );
      return fieldLocation;
   }

   /**
    * given a BasicBlock ID, return a reference to the corresponding basicBlock
    */
   public BasicBlock getBasicBlock( int bbID )
   {
      try
      {
         return basicBlocks[ bbID ];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable,
               "getValues: invalid attribute index: " + bbID );
         if ( getNumBasicBlock() == 0 )
            return null;
         else return basicBlocks[0];
      }
   }

   /**
    * return number of basicBlocks contained in this Compositeblock
    */
   public int getNumBasicBlock()
   {
      return basicBlocks.length;
   }

   /** set a rod of data in the corresponding location of this datablock */
   protected void setrod( ByteBuffer rod, int rodSize, int iniDataArrayOffset,
         int dataArrayStep, FieldIDMapper mp, RecordDescriptor sourceRecordDesc )
   {
      //System.out.println("Composite Block setRod FieldIDMapper is :"+ mp);
      int numBasicB = basicBlocks.length;
      FieldIDMapper[] mp1 = new FieldIDMapper[numBasicB];
      for ( int i = 0; i < numBasicB; i++ )
      {
         //mp1[i] = new FieldIDMapper( this.idMapper[i], mp, 1);
         mp1[i] = FieldIDMapper.createComponentMap( mp, this.idMapper[i] );

         //System.out.println("BasicBlock setrod FieldIDMapper for block
         // number" + i+ " is:"+ mp1[i]);
         basicBlocks[i].setRod( rod, rodSize, iniDataArrayOffset,
               dataArrayStep, mp1[i], sourceRecordDesc );
      }
   }

   /** set a rod of data in the corresponding location of this datablock */
   protected void setrod( ByteBuffer rod, int rodSize, int iniDataArrayOffset,
         int dataArrayStep, RecordSpec recSpec,
         RecordDescriptor sourceRecordDesc )
   {
      //System.out.println("Composite Block setRod FieldIDMapper is :"+ mp);
      int numBasicB = basicBlocks.length;
      FieldIDMapper[] mp1 = new FieldIDMapper[numBasicB];
      int srcFields[] = recSpec.getFieldNumbers();
      int dstFields[] = new int[srcFields.length];
      for ( int j = 0; j < srcFields.length; j++ )
      {
         dstFields[j] = j;
      }
      FieldIDMapper mp = new FieldIDMapper( srcFields, dstFields );

      for ( int i = 0; i < numBasicB; i++ )
      {
         mp1[i] =  FieldIDMapper.createComponentMap( mp, this.idMapper[i] );
         //System.out.println("BasicBlock setrod FieldIDMapper for block
         // number" + i+ " is:"+ mp1[i]);
         basicBlocks[i].setRod( rod, rodSize, iniDataArrayOffset,
               dataArrayStep, mp1[i], sourceRecordDesc );
      }
   }

   /** set a rod of data in the corresponding location of this datablock */
   protected void setrod( ByteBuffer rod, int rodSize, int iniDataArrayOffset,
         int dataArrayStep, RecordDescriptor sourceRecordDesc )
   {
      int numBasicB = basicBlocks.length;
      FieldIDMapper[] mp1 = new FieldIDMapper[numBasicB];
      int numOfFields = sourceRecordDesc.size();
      int Fields[] = new int[numOfFields];

      for ( int j = 0; j < numOfFields; j++ )
      {
         Fields[j] = j;
      }
      FieldIDMapper mp = new FieldIDMapper( Fields, Fields );

      for ( int i = 0; i < numBasicB; i++ )
      {
         mp1[i] = FieldIDMapper.createComponentMap( mp, this.idMapper[i] );
         basicBlocks[i].setRod( rod, rodSize, iniDataArrayOffset,
               dataArrayStep, mp1[i], sourceRecordDesc );
      }
   }

   /**
    * set a rod of data in the corresponding location of this datablock, assume
    * the rod contains only one attribute, works for attributeOrder
    */
   protected void setAttributeOrderRod( ByteBuffer rod, 
                                 int rodSize,
                                 int iniDataArrayOffset, 
                                 int dataArrayStep, 
                                 int attributeID,
                                 int attributeByteSize, 
                                 int storageType )
   {
      FieldInfo attributeLocation = locateAttribute( attributeID );
      //System.out.println( "in CompositeBlock , rod Size :" + rodSize + "
      // dataum offset : " + iniDataArrayOffset );
      basicBlocks[attributeLocation.bb].setAttributeOrderRod( 
                                    rod, 
                                    rodSize,
                                    iniDataArrayOffset, 
                                    dataArrayStep, 
                                    attributeLocation.attr,
                                    attributeByteSize, 
                                    storageType );
   }

   /**
    * set float data of a given attribute at a given index ( index of the datum
    * in the one D array )
    */
   public final void setField( int index, float floatData, int fieldID )
   {
      FieldInfo attributeLocation = locateAttribute( fieldID );
      this.basicBlocks[attributeLocation.bb].setField( index, floatData,
            attributeLocation.attr );
   }

   /**
    * set int data of a given attribute at a given index ( index of the datum in
    * the one D array )
    */
   public final void setField( int index, int intData, int fieldID )
   {
      FieldInfo attributeLocation = locateAttribute( fieldID );
      this.basicBlocks[attributeLocation.bb].setField( index, intData,
            attributeLocation.attr );
   }

   /**
    * set double data of a given attribute at a given index ( index of the datum
    * in the one D array )
    */
   public final void setField( int index, double doubleData, int fieldID )
   {
      FieldInfo attributeLocation = locateAttribute( fieldID );
      this.basicBlocks[attributeLocation.bb].setField( index, doubleData,
            attributeLocation.attr );
   }

   /**
    * set short data of a given attribute at a given index ( index of the datum
    * in the one D array )
    */
   public final void setField( int index, short shortData, int fieldID )
   {
      FieldInfo attributeLocation = locateAttribute( fieldID );
      this.basicBlocks[attributeLocation.bb].setField( index, shortData,
            attributeLocation.attr );
   }

   /**
    * return a Values array consisting of the data for the specified attribute.
    */
   public Values getValues( int attr )
   {
      FieldInfo fieldLoc = locateAttribute( attr );
      BasicBlock block   = basicBlocks[ fieldLoc.bb ];
      return  block.getValues( fieldLoc.attr );
      
      /***** replaced by rdb
      if ( attr < 0 || attr > getNumAttributes() )
      {
         new SDBError( SDBError.Recoverable,
               "getValues: invalid attribute index: " + attr );
         return null;
      }
      else 
         return ( getBasicBlock( attr ) ).getValues( attr );
      /***************************/
   }

   //-------------------- get<type>( isid, attr )
   // --------------------------------
   // Base versions of these are defined in DataCollection, but these are more
   // efficient implementations.

   /**
    * get the field value at the ISID position as a double. Convert type if
    * necessary.
    */
   public double getDouble( IndexSpaceID did, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getDouble( did, fieldLoc.attr );
   }

   /**
    * get the field value at the ISID position as a float. Convert type if
    * necessary.
    */
   public float getFloat( IndexSpaceID did, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getFloat( did, fieldLoc.attr );
   }

   /**
    * get the field value at the ISID position as an int. Convert type if
    * necessary.
    */
   public int getInt( IndexSpaceID did, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getInt( did, fieldLoc.attr );
   }

   /**
    * get the field value at the ISID position as a short. Convert type if
    * necessary.
    */
   public short getShort( IndexSpaceID did, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getShort( did, fieldLoc.attr );
   }

   /**
    * get the field value at the ISID position as a byte. Convert type if
    * necessary.
    */
   public byte getByte( IndexSpaceID did, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getByte( did, fieldLoc.attr );
   }

   // ************ get<type>Attribute methods for attributes access
   // ***************
   // rdb: to be deprecated

   /** get the double attribute for a given IndexSpaceID and attributeID */
   public final double getDoubleAttribute( IndexSpaceID did, int fieldNumber )
   {
      return getDouble( did, fieldNumber );
   }

   /** get the float attribute for a given IndexSpaceID and attributeID */
   public final float getFloatAttribute( IndexSpaceID did, int fieldNumber )
   {
      return getFloat( did, fieldNumber );
   }

   /** get the int attribute for a given IndexSpaceID and attributeID */
   public final int getIntAttribute( IndexSpaceID did, int fieldNumber )
   {
      return getInt( did, fieldNumber );
   }

   /** get the short attribute for a given IndexSpaceID and attributeID */
   public final short getShortAttribute( IndexSpaceID did, int fieldNumber )
   {
      return getShort( did, fieldNumber );
   }

   /** get the byte attribute for a given IndexSpaceID and attributeID */
   public final byte getByteAttribute( IndexSpaceID did, int fieldNumber )
   {
      return getByte( did, fieldNumber );
   }

   /**
    * Given an indexSpaceID and the location of the attribute within the
    * compositeBlock.the location is represented by an array with length of 2,
    * the first element of the array represents the id of the basicBlock, in
    * which the attribute is stored. The second element of the array represents
    * the component fieldID, which matches the given compositefieldID This
    * method can be used when a certain attribute is accessed extensively. In
    * this case, there is not need to compute the location of the attribute
    * every time the attribute is accessed.
    * 
    * rdb note: the savings entailed in providing this additional "efficiency"
    * is miniscule -- one index into a single array. In exchange we are exposing
    * our internal implementation to the user which prevents us from changing
    * that implementation in the future.
    * 
    * This should be deprecated or made "package".
    */
   float getFloatAttribute( IndexSpaceID did, int[] componentFieldLoc )
   {
      return basicBlocks[componentFieldLoc[0]].getFloat( did,
            componentFieldLoc[1] );
   }

   //-------------------- getIndexed<type>Attribute methods
   // -----------------------
   // rdb: Should these be package?

   /**
    * get indexed float attribute , OneDIndex is the conceptual index of the
    * datum in the one D array
    */
   public float getIndexedFloatAttribute( int OneDIndex, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getIndexedFloat( OneDIndex, fieldLoc.attr );
   }

   /**
    * get indexed Short attribute, OneDIndex is the conceptual index of the
    * datum in the one D array
    */
   public short getIndexedShortAttribute( int OneDIndex, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getIndexedShort( OneDIndex, fieldLoc.attr );
   }

   /**
    * get indexed int attribue, OneDIndex is the conceptual index of the datum
    * in the one D array
    */
   public int getIndexedIntAttribute( int OneDIndex, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getIndexedInt( OneDIndex, fieldLoc.attr );
   }

   /**
    * get indexedFieldShort , OneDIndex is the conceptual index of the datum in
    * the one D array
    */
   public double getIndexedDoubleAttribute( int OneDIndex, int fieldNumber )
   {
      FieldInfo fieldLoc = locateAttribute( fieldNumber );
      BasicBlock block = basicBlocks[fieldLoc.bb];
      return block.getIndexedDouble( OneDIndex, fieldLoc.attr );
   }

   // some private method for fast attribute access
   private float getIndexedFloatAttribute( int OneDIndex, FieldInfo fieldLoc )
   {
      return basicBlocks[fieldLoc.bb].getIndexedFloat( OneDIndex,
            fieldLoc.attr );
   }

   private short getIndexedShortAttribute( int OneDIndex, FieldInfo fieldLoc )
   {
      return basicBlocks[fieldLoc.bb].getIndexedShort( OneDIndex,
            fieldLoc.attr );
   }

   private int getIndexedIntAttribute( int OneDIndex, FieldInfo fieldLoc )
   {
      return basicBlocks[fieldLoc.bb].getIndexedInt( OneDIndex,
            fieldLoc.attr );
   }

   private double getIndexedDoubleAttribute( int OneDIndex, FieldInfo fieldLoc )
   {
      return basicBlocks[fieldLoc.bb].getIndexedDouble( OneDIndex,
            fieldLoc.attr );
   }

   /**
    * User passes in a pre-allocated short array, attributes specified by
    * recordSpec are copied from the BasicBlock to the dstArray. attribute IDs
    * and their storage order are specified in RecordSpec. the boolean
    * paramerter indicates if the attribute in the dstArray in stored in a Point
    * Order or AttributeOrder Attributes are changed into shorts if they are not
    * stored as shorts in this BasicBlock.
    */
   public void getShortAttributes( short[] dstArray, RecordSpec recordSpec,
         boolean pointOrder )
   {
      int i, j;
      int numAttri = recordSpec.size();
      int numberOfDatums = this.volume();
      int index = 0;
      FieldInfo fieldLoc[] = this.fieldLocate( recordSpec );

      //fieldLoc[i] = locateAttribute( recordSpec.getIndexedFieldNumber(i) );

      if ( pointOrder )
      {
         for ( i = 0; i < numberOfDatums; i++ )
         {
            for ( j = 0; j < numAttri; j++ )
            {
               dstArray[index++ ] = this.getIndexedShortAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
      else
      // store the attribtes in attributeOrder
      {
         for ( j = 0; j < numAttri; j++ )
         {
            for ( i = 0; i < numberOfDatums; i++ )
            {
               dstArray[index++ ] = this.getIndexedShortAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
   }

   // locate the attribuate storage location, return a 2d array, with each row
   // representing a attribute, first column is the block number
   //   of the attribute, the second column is the id of the attribute within the
   // component block .
   private FieldInfo[] fieldLocate( RecordSpec recordSpec )
   {
      int numAttri = recordSpec.size();
      FieldInfo fieldLoc[] = new FieldInfo[numAttri];
      for ( int i = 0; i < numAttri; i++ )
      {
         fieldLoc[i] = locateAttribute( recordSpec.getFieldNumber( i ) );
      }
      return fieldLoc;
   }

   /**
    * User passes in a pre-allocated double array, attributes specified by
    * recordSpec are copied from this compositeBlock to the dstArray. attribute
    * IDs and their storage order are specified in RecordSpec. the boolean
    * paramerter indicates if the attribute in the dstArray in stored in a Point
    * Order or AttributeOrder Attributes are changed into doubles if they are
    * not stored as doulbes in this CompositeBlock.
    */
   public void getDoubleAttributes( double[] dstArray, RecordSpec recordSpec,
         boolean pointOrder )
   {
      int i, j;
      int numAttri = recordSpec.size();
      int numberOfDatums = this.volume();
      int index = 0;
      FieldInfo fieldLoc[] = this.fieldLocate( recordSpec );
      if ( pointOrder )
      {
         for ( i = 0; i < numberOfDatums; i++ )
         {
            for ( j = 0; j < numAttri; j++ )
            {
               dstArray[index++ ] = this.getIndexedDoubleAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
      else
      // store the attribtes in attributeOrder
      {
         for ( j = 0; j < numAttri; j++ )
         {
            for ( i = 0; i < numberOfDatums; i++ )
            {
               dstArray[index++ ] = this.getIndexedDoubleAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
   }

   /**
    * User passes in a pre-allocated float array, attributes specified by
    * recordSpec are copied from this CompositeBlock to the dstArray. attribute
    * IDs and their storage order are specified in RecordSpec. the boolean
    * paramerter indicates if the attribute in the dstArray in stored in a Point
    * Order or AttributeOrder Attributes are changed into floats if they are not
    * stored as float in this CompositeBlock.
    */
   public void getFloatAttributes( float[] dstArray, RecordSpec recordSpec,
         boolean pointOrder )
   {
      int i, j;
      int numAttri = recordSpec.size();
      int numberOfDatums = this.volume();
      int index = 0;
      FieldInfo fieldLoc[] = this.fieldLocate( recordSpec );

      if ( pointOrder )
      {
         for ( i = 0; i < numberOfDatums; i++ )
         {
            for ( j = 0; j < numAttri; j++ )
            {
               dstArray[index++ ] = this.getIndexedFloatAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
      else
      // store the attribtes in attributeOrder
      {
         for ( j = 0; j < numAttri; j++ )
         {
            for ( i = 0; i < numberOfDatums; i++ )
            {
               dstArray[index++ ] = this.getIndexedFloatAttribute( i,
                     fieldLoc[j] );
            }
         }
      }
   }

   /**
    * User passes in a pre-allocated int array, attributes specified by
    * recordSpec are copied from this Compositeblock to the dstArray. attribute
    * IDs and their storage order are specified in RecordSpec. the boolean
    * paramerter indicates if the attribute in the dstArray in stored in a Point
    * Order or AttributeOrder Attributes are changed into ints if they are not
    * stored as ints in this CompositeBlock.
    */
   public void getIntAttributes( int[] dstArray, RecordSpec recordSpec,
         boolean pointOrder )
   {
      int i, j;
      int numAttri = recordSpec.size();
      int numberOfDatums = this.volume();
      int index = 0;
      FieldInfo fieldLoc[] = this.fieldLocate( recordSpec );
      if ( pointOrder )
      {
         for ( i = 0; i < numberOfDatums; i++ )
         {
            for ( j = 0; j < numAttri; j++ )
            {
               dstArray[index++ ] = this
                     .getIndexedIntAttribute( i, fieldLoc[j] );
            }
         }
      }
      else
      // store the attribtes in attributeOrder
      {
         for ( j = 0; j < numAttri; j++ )
         {
            for ( i = 0; i < numberOfDatums; i++ )
            {
               dstArray[index++ ] = this
                     .getIndexedIntAttribute( i, fieldLoc[j] );
            }
         }
      }
   }

   /** Given an IndexSpaceID, return the corresponding datum. */
   public Datum datum( IndexSpaceID did )
   {
      this.datum( nullDatum, did );
      return nullDatum;
   }

   /**
    * Given IndexSpaceID, load the corresponding datum into the datum argument.
    * data will be converted into float if it is not stored as float in this
    * compositeBlock
    */
   public void datum( Datum datum, IndexSpaceID did )
   {
      int NtoOneOffset = this.computeOffset( did );
      FieldInfo fieldLoc;
      BasicBlock block;

      for ( int i = 0; i < this.numOfAttributes; i++ )
      {
         fieldLoc = locateAttribute( i );
         block = basicBlocks[fieldLoc.bb];
         datum.setFloat( i, block.getIndexedFloat( NtoOneOffset,
               fieldLoc.attr ) );
      }
   }

   /**
    * Given an array of coordinates, load the corresponding datum into the datum
    * argument. Note: only works for floats
    */
   public void datum( Datum datum, int[] coords )
   {
      int NtoOneOffset = this.computeOffset( coords );
      FieldInfo fieldLoc;
      BasicBlock block;

      for ( int i = 0; i < this.numOfAttributes; i++ )
      {
         fieldLoc = locateAttribute( i );
         block = basicBlocks[fieldLoc.bb];
         datum.setFloat( i, block.getIndexedFloat( NtoOneOffset,
               fieldLoc.attr ) );
      }
   }

   /**
    * Given an IndexSpaceID, return the corresponding datum with attributes
    * specified in the RecordSpec.
    */
   public Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      Datum returnDatum = this.getRecordDescriptor().createDatum();
      datum( returnDatum, did, spec );
      return returnDatum;
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding FieldIDMapper.
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      int oneDIndex = this.computeOffset( did );
      int[] cpnID = mp.getComponentFields();
      int[] cpsID = mp.getCompositeFields();
      int len = cpnID.length;
      FieldInfo fieldLoc;
      BasicBlock block;

      for ( int i = 0; i < len; i++ )
      {
         fieldLoc = locateAttribute( cpnID[i] );
         block = basicBlocks[fieldLoc.bb];
         dtm.setFloat( cpsID[i], block.getIndexedFloat( oneDIndex,
               fieldLoc.attr ) );
      }
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding RecordSpec.
    */
   public void datum( Datum dtm, IndexSpaceID did, RecordSpec recSpec )
   {
      int oneDIndex = this.computeOffset( did );
      int[] attributeIDs = recSpec.getFieldNumbers();
      int len = recSpec.size();
      FieldInfo fieldLoc;
      BasicBlock block;

      for ( int i = 0; i < len; i++ )
      {
         fieldLoc = locateAttribute( attributeIDs[i] );
         block = basicBlocks[fieldLoc.bb];
         dtm.setFloat( i, block.getIndexedFloat( oneDIndex,
               fieldLoc.attr ) );
      }
   }

   /** Set the location specified by the IndexSpaceID to the given datum. */
   public void setDatum( IndexSpaceID isid, Datum d )
   {
      BasicBlock blk = null;
      FieldInfo fieldLoc = null;
      int NtoOneOffset = this.computeOffset( isid );
      int storageType;

      for ( int i = 0; i < this.numOfAttributes; i++ )
      {
         fieldLoc = locateAttribute( i );
         blk = this.basicBlocks[fieldLoc.bb];
         storageType = this.recordDesc.getStorageType( i );
         switch ( storageType )
         {
            case StorageType.Float:
               blk.setField( NtoOneOffset, d.getFloat( i ), fieldLoc.attr );
               break;
            case StorageType.Double:
               blk.setField( NtoOneOffset, d.getDouble( i ), fieldLoc.attr );
               break;
            case StorageType.Int:
               blk.setField( NtoOneOffset, d.getInt( i ), fieldLoc.attr );
               break;
            case StorageType.Short:
               blk.setField( NtoOneOffset, d.getShort( i ), fieldLoc.attr );
               break;
            case StorageType.Byte:
               blk.setField( NtoOneOffset, d.getByte( i ), fieldLoc.attr );
               break;
            default:
               new InternalFaultError( SDBError.Fatal,
                     "Unsupported storage type code in BinaryFileDataSource.readValue:"
                           + storageType );
               break;
         }
      }
   }

   /**
    * Return a new DataBlock( currently it returns a CompositeBlock) that is a
    * subset of this datablock. Data is copied from this DataBlock to the return
    * DataBlock
    */
   public DataBlock subblock( ISBounds db )
   {
      CompositeBlock returnBlock = new CompositeBlock( db, this.recordDesc );
      ISBounds dstBounds = returnBlock.getBounds();
      subblock( returnBlock, dstBounds, db );
      return returnBlock;
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The dstBlock has to be a compositeBlock.
    * Currently we do not support subblocking from CompositeBlock to BasicBlock
    */
   public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
      if ( dstBlock instanceof CompositeBlock )
      {
         subblock( (CompositeBlock) dstBlock, dstBounds, srcBounds );
      }
      else
      {
         new InternalFaultError( SDBError.Recoverable,
               "Subblock from CompositeBlock to BasicBlock is not implemented ." );
      }// end of else

   }

   /**
    * assume the sub-compositeBlock has the same data format as this
    * compositeBlock
    */
   private void subblock( CompositeBlock dstBlock, 
                          ISBounds dstBounds, ISBounds srcBounds )
   {
      int numBlocks = basicBlocks.length;
      for ( int i = 0; i < numBlocks; i++ )
      {
         this.basicBlocks[i].subblock( dstBlock.basicBlocks[i], dstBounds,
               srcBounds );
      }
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The FieldIDMapper argument determines the
    * mapping between datums in this DataBlock and datums in the argument
    * DataBlock.
    */
   public void subblock( DataBlock dstBlock, ISBounds dstBounds,
         ISBounds srcBounds, FieldIDMapper mp )
   {
      if ( dstBlock instanceof CompositeBlock )
      {
         subblock( (CompositeBlock) dstBlock, dstBounds, srcBounds, mp );
      }
      else
      {
         new InternalFaultError( SDBError.Recoverable,
               "Subblock from CompositeBlock to BasicBlock is not implemented ." );
      }// end of else
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The FieldIDMapper argument determines the
    * mapping between datums in this DataBlock and datums in the argument
    * DataBlock. It is valid for FieldIDMapper argument to be null, which 
    * means that all attributes are selected.
    */
   private void subblock( CompositeBlock dstBlock, 
                          ISBounds dstBounds, ISBounds srcBounds, 
                          FieldIDMapper mp )
   {
      FieldInfo srcLoc;
      FieldInfo dstLoc;
      BasicBlock srcBlk;
      BasicBlock dstBlk;
      
      if ( mp == null )
         mp = new FieldIDMapper( this.getNumAttributes() ); // create identity mapping

      int[] componentFields = mp.getComponentFields();
      int[] compositeFields = mp.getCompositeFields();
      int size = mp.size();
      int[] cpn = new int[1];
      int[] cps = new int[1];
      
      FieldIDMapper idmapper;

      for ( int i = 0; i < size; i++ )
      {
         srcLoc = this.locateAttribute( componentFields[i] );
         dstLoc = dstBlock.locateAttribute( compositeFields[i] );
         cpn[0] = srcLoc.attr;
         cps[0] = dstLoc.attr;
         idmapper = new FieldIDMapper( cpn, cps );
         srcBlk = this.basicBlocks[srcLoc.bb];
         dstBlk = dstBlock.basicBlocks[dstLoc.bb];
         srcBlk.subblock( dstBlk, dstBounds, srcBounds, idmapper );
      }
   }

   /** the returning DataBlock is a CompositeBlock */
   public DataBlock subblock( ISBounds srcBounds, RecordSpec rs )
   {
      if ( rs == null )
         rs = new RecordSpec( this.getNumAttributes() );
      int[] componentFields = rs.getFieldNumbers();
      RecordDescriptor recdescriptor = this.getRecordDescriptor();
      AttributeDescriptor[] attriDesc = new AttributeDescriptor[numOfAttributes];
      for ( int i = 0; i < numOfAttributes; i++ )
      {
         attriDesc[i] = recdescriptor.getAttribute( componentFields[i] );
      }
      RecordDescriptor rdescriptor = new RecordDescriptor( attriDesc );
      CompositeBlock cBlock = new CompositeBlock( srcBounds, rdescriptor );
      this.subblock( cBlock, cBlock.getBounds(), srcBounds, rs );
      return cBlock;
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The RecordSpec argument specifies a subset
    * of the attributes of this DataSource to be written into the DataBlock
    * datums. currently the dstBlock has to be a compositeBlock
    */
   public void subblock( DataBlock dstBlock, ISBounds dstBounds,
         ISBounds srcBounds, RecordSpec spec )
   {
      if ( dstBlock instanceof CompositeBlock )
      {
         subblock( (CompositeBlock) dstBlock, dstBounds, srcBounds, spec );
      }
      else
      {
         new InternalFaultError( SDBError.Recoverable,
               "Subblock from CompositeBlock to BasicBlock is not implemented ." );
      }// end of else
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The RecordSpec argument specifies a subset
    * of the attributes of this DataSource to be written into the DataBlock
    * datums.
    */
   private void subblock( CompositeBlock dstBlock, 
                          ISBounds dstBounds, ISBounds srcBounds, 
                          RecordSpec spec )
   {
      FieldInfo srcLoc;
      FieldInfo dstLoc;
      BasicBlock srcBlk;
      BasicBlock dstBlk;

      if ( spec == null )
         spec = new RecordSpec( getNumAttributes() );
      int[] componentFields = spec.getFieldNumbers();
      int size = spec.size();
      int[] cpn = new int[1];
      int[] cps = new int[1];
      FieldIDMapper idmapper;
      for ( int i = 0; i < size; i++ )
      {
         srcLoc = this.locateAttribute( componentFields[i] );
         dstLoc = dstBlock.locateAttribute( i );
         cpn[0] = srcLoc.attr;
         cps[0] = dstLoc.attr;
         idmapper = new FieldIDMapper( cpn, cps );
         srcBlk = this.basicBlocks[srcLoc.bb];
         dstBlk = dstBlock.basicBlocks[dstLoc.bb];
         srcBlk.subblock( dstBlk, dstBounds, srcBounds, idmapper );
      }
   }
   
   //------------------ set<type>s( int attr, <type>[] ) --------------------------
   // override parent for efficiency
   
   /** Assign attribute values for all points in the block */
   public void setDoubles( int attr, double[] newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setDoubles( where.attr, newValues );
   }
   
   /** Assign attribute values for all points in the block */
   public void setFloats( int attr, float[] newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setFloats( where.attr, newValues );
   }
   
   /** Assign attribute values for all points in the block */
   public void setInts( int attr, int[] newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setInts( where.attr, newValues );
   }
   
   /** Assign attribute values for all points in the block */
   public void setShorts( int attr, short[] newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setShorts( where.attr, newValues );
   }
   
   /** Assign attribute values for all points in the block */
   public void setBytes( int attr, byte[] newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setBytes( where.attr, newValues );
   }
   
   /** Assign attribute values for all points in the block */
   public void setValues( int attr, Values newValues )
   {
      FieldInfo where = locateAttribute( attr );
      basicBlocks[ where.bb ].setValues( where.attr, newValues );
   }
}