//
//  DSBlockStencil.java
//  ObliqueCacheTest
//
//  Created by Philip Rhodes on 6/3/06.
//  Copyright 2006 Philip J Rhodes. All rights reserved.
//

package edu.unh.sdb.datasource;

import java.util.*;
import java.nio.*;
import edu.unh.sdb.common.*;



/** This is an implementation of a partitioned stencil that associates a
 *  Datablock with active elements. It uses an array to
 *  maintain the record of active elements. As a result, the set and get
 *  methods are extremely fast, but iteration might be a little less efficient
 *  than the list implementation. However, it does not tax the garbage collector,
 *  so it may be more appropriate for repeated use.
 */
public class DSBlockStencil extends DSPartitionedStencil{
   
   protected DataBlock [] inStencil;
   protected RecordDescriptor recordDescriptor;
   
   protected int size = 0;
   
   protected int  lastIndexRequested=-1;  // last index given to a getIndexed()
   protected int lastActiveOffset=-1; // last offset of an active element
                                         // found by getIndexed()
                                         
   protected int dataBytesUsed =0;  // Data bytes used by last call to getData()
   
   protected Datum tempDatum; // Spare datum used in AMR code
   

   //protected int offset = 4;        // length of the header when writing to a byte array.


    // Old
   /** Construct a stencil that identifies areas of interest using a partitioning.
    *  The first argument is the dimensions of the partitioning space, meaning
    *  that the index space bounded by the second argument will be divided up into
    *  a matching number of partitions in each dimension. For example, if we build
    *  an 8x8 partitioning of a 128x128 index space, each partition will have
    *  dimensions 16x16.  In this situation, the first argument for this constructor
    *  should be [8,8] and the second argument should be {<0,0>, <128,128>}.
    */
    
   /** Construct a stencil that identifies areas of interest using a partitioning.
    *  The first argument is the dimensions of an element of the partitioning, 
    *  meaning that the index space bounded by the second argument will be divided
    *  up into pieces of the shape described by the first argument. For example, 
    *  if we build an 8x8 partitioning of a 128x128 index space, each element
    *  of the partitioning will have dimensions 16x16. In this situation, 
    *  the first argument for this constructor should be [16,16] and the second
    *  argument should be {<0,0>, <127,127>}.
    */   
//   public DSBlockStencil( int [] elementDimensions, ISBounds space){
//
//      super(new RegularISPartitioning(elementDimensions, space));
//      this.inStencil = new boolean[(int) this.partitioning.volume()];
//   }


   public DSBlockStencil( RegularISPartitioning partitioning, RecordDescriptor recordDescriptor){

      super(partitioning);
      
      this.recordDescriptor = recordDescriptor;
      
      this.inStencil = new DataBlock[(int) this.partitioning.volume()];
      
      this.tempDatum = this.recordDescriptor.createDatum();
   }

   /** Fill the given byte array with the data contained in the active elements
    *  of this stencil. If the given array is not long enough to contain the stencil
    *  data, a new array is allocated and returned. The number of bytes actually used
    *  in the array can be retrieved by calling getDataBytesUsed().
    */
   public byte [] getData(byte [] data){
   
      int requiredCapacity = this.size * (int) this.partitionVolume();
      int dataOffset = 0;
      int volume;
      DataBlock block;
      
      if(data.length < requiredCapacity) {
      
         data = new byte[requiredCapacity];
      }
      
      for(int i = 0; i< this.size; i++){
      
         block = this.getIndexedBlock(i);
         
         volume = block.volume();
      
         System.arraycopy(block.getValues().getBytes(), 0, data, dataOffset, volume);
         
         dataOffset += volume;
         
      }
   
      this.dataBytesUsed=dataOffset;
      
      return data;
   }

   /** Return the number of bytes actually used in the last call to getData().
    */
   public int getDataBytesUsed(){
   
      return this.dataBytesUsed;
   }


   /** Populate the active elements (Datablocks) of this stencil with the 
    *  data found in the given array. This method assumes that the array 
    *  data was generated by the getData() method of a stencil exactly equivalent
    *  to this one. In particular, the stencil dimensions must all be identical,
    *  and the same elements must already be active.
    */
   public void setData(byte [] data){

      int dataOffset = 0;
      int volume;
      DataBlock block;
            
      for(int i = 0; i< this.size; i++){
      
         block = this.getIndexedBlock(i);
         
         volume = block.volume();
      
         System.arraycopy(data, dataOffset, block.getValues().getBytes(), 0, volume);
         
         dataOffset += volume;
         
      }
   }
   
   /** Populate the active elements (Datablocks) of this stencil with the
    *  data found in the given array.  This method assumes that the array
    *  data was generated by the getData() method of a stencil exactly equivalent
    *  to this one.  In particular, the stencil dimensions must all be identical,
    *  and the same elements must already be active.
    *
    *  This AMR code is meant to be used by GraniteRemoteClient to handle Chombo
    *  data sent over the wire by a GraniteServer employing a ChomboDataSource
    *  to read chombo files via hdf5. The main counterpart to this code can be found
    *  in ChomboDataSource.getData()
    *
    * TODO: this code needs to move somewhere else. Probably, we need a new class
    * that interprets byte arrays, encapsulating things like AMR formats more
    * cleanly.
    */
   public void setDoubleAMRData(byte[] data)
   {
       ByteBuffer buffer = ByteBuffer.wrap(data);
       int dims, numLevels, numQBoxes;
       int[] refRatios;
       int[] numBoxes, numDataElems;
       int totalRefRatio;
       
       double currentValue;
       IndexSpaceID loc;
       int[] blockShape = new int[this.dimensionality];
       int[] dataShape  = new int[this.dimensionality];
       Arrays.fill(blockShape, 1);
       Arrays.fill(dataShape, 1);
       
       ISBoundsIterator boxIter;
       ISIterator dataIter;

//       System.out.printf("Reading header information from data... ");
       dims      = buffer.getInt();
       numLevels = buffer.getInt();
       refRatios = new int[numLevels];
       for(int i = 0; i < numLevels; i++)
           refRatios[i] = buffer.getInt();
       numQBoxes = buffer.getInt();
//       System.out.println("Done.");
//       System.out.println("Header information:\n");
//       
//       System.out.printf("Number of dimensions:  %d\n", dims);
//       System.out.printf("Number of levels:      %d\n", numLevels);
//       System.out.println("Refinement ratios per level:");
//       for(int i = 0; i < refRatios.length; i++)
//           System.out.printf("  Level %d: %dx\n", i, refRatios[i]);
//       
//       System.out.printf("Number of query boxes: %d\n", numQBoxes);
//       System.out.println("-------------------------------------------\n");
       
//       System.out.println("Populating data blocks from files...");
       
       int volume;
       int[] upper = new int[this.dimensionality];
       int[] lower = new int[this.dimensionality];
       DataBlock qBlock       = null;
       ISBounds qBounds       = null;
       ISBounds[][] boxBounds = null;
       numBoxes     = new int[numLevels];
       numDataElems = new int[numLevels];
       boxBounds    = new ISBounds[numLevels][];
       
       // Create the box and data iterators with dummy values; they will be changed later.
       // Chombo uses fortran ordering, so we'll set the axis ordering appropriately
       boxIter  = new ISBoundsIterator(this.getSpaceBounds(), blockShape, new AxisOrdering(new int[]{2, 1, 0})); // PJR
       dataIter = new ISIterator(this.getSpaceBounds()); 
       
       // For each block in the stencil
       for(int q = 0; q < this.size; q++)
       {
           qBlock  = this.getIndexedBlock(q);
           qBounds = this.getIndexedBounds(q);
           
           volume = qBlock.volume();
           IndexSpaceID qBoundsLower = qBounds.getLower();


           
           //For each resolution level
           for(int lv = 0; lv < numLevels; lv++){


               // Get metadata from the buffer.
               numBoxes[lv]     = buffer.getInt();
               numDataElems[lv] = buffer.getInt();
        
               boxBounds[lv]    = new ISBounds[numBoxes[lv]];
               
               // Box sizes
               for(int b = 0; b < numBoxes[lv]; b++)
               {
                   buffer.asIntBuffer().get(lower, 0, this.dimensionality);
                   buffer.position(buffer.position() + (this.dimensionality * 4));
                   buffer.asIntBuffer().get(upper, 0, this.dimensionality);
                   buffer.position(buffer.position() + (this.dimensionality * 4));
                   boxBounds[lv][b] = new ISBounds(lower, upper);
               }
               
               // For each chombo box
               for(int b = 0; b < numBoxes[lv]; b++)
               {
                   // Calculate the total refinement ratio from this level to max level
                   // Omit the final level, since we cannot refine higher than that.
                   totalRefRatio = 1;
                   for(int thisRr = lv; thisRr < (numLevels - 1); thisRr++){
                       totalRefRatio *= refRatios[thisRr];
                   }
                   
                   // Block shape specifies the shape of a sub-partition in n-dimensions.
                   Arrays.fill(blockShape, totalRefRatio);
                   ISBounds scaledBox = boxBounds[lv][b].copyExpandByFactor(totalRefRatio);

                   // The box iterator will iterate over the intersection of the
                   // chombo box and the query bounds. At each step of the iteration
                   // we will read a data value, and use a dataIter to write this
                   // value at multiple locations in the datablock.
                   
                   // Box iteration area is over the intersection of the query block, anchored to
                   // the box from the buffer scaled to query box resolution.  Block shape is
                   // pre-calculated based on refinement ratio.
                   
                   ISBounds boxIterBounds = scaledBox.intersection(qBounds).makeRelativeTo(qBoundsLower);
                   IndexSpaceID anchor = scaledBox.getLower().subtract(qBoundsLower);
                   
                   boxIter.setAll(boxIterBounds, anchor, blockShape);
 
                   for(boxIter.init(); boxIter.valid(); boxIter.next()) 
                   {
                       // read a value from the buffer. This value will be duplicated
                       // by the dataIter loop to form the high resolution data in 
                       // the current datablock
                       currentValue = buffer.getDouble();
                       
                       // write the value to a spare datum, so we can use 
                       // the setDatum method of basicblock
                       this.tempDatum.setDouble(0, currentValue);
                       
                       // Now iterate over the interior of the box iterator's current
                       // shape, writing the value in each position.
                       dataIter.setBounds(boxIter);
                       for(dataIter.init(); dataIter.valid(); dataIter.next())
                       {
                           qBlock.setDatum(dataIter, this.tempDatum);
                       }
                   }
               }
           }
           
//           // Print per-level metadata.
//           System.out.printf("This block's bounds:          %s\n", qBounds);
//           //System.out.printf("Corresponding bounds in data: %s\n", dataBounds);
//           System.out.printf("This block's volume:          %d elements\n", volume);
//           System.out.println("For each level:");
//           for(int lv = 0; lv < numLevels; lv++)
//           {
//               totalRefRatio = 1;
//               for(int thisRr = lv; thisRr < numLevels; thisRr++)
//                   totalRefRatio *= refRatios[thisRr];
//               
//               System.out.printf("  Level %d:\n", lv);
//               System.out.printf("    Number of data elements:       %d\n", numDataElems[lv]);
//               System.out.printf("    Number of boxes:               %d\n", numBoxes[lv]);
//               System.out.printf("    Refinement ratio to max level: %d\n", totalRefRatio);
//               for(int b = 0; b < numBoxes[lv]; b++)
//               {
//                   System.out.printf("\n        Box %d: %s\n", b, boxBounds[lv][b]);
//                   if(lv != numLevels - 1)
//                       System.out.printf("           Scales to: %s as max level\n",
//                                         boxBounds[lv][b].copyExpandByFactor(totalRefRatio));
//               }
//           }
//           System.out.println();
       }
       

   }
   
   /** Populate the active elements (Datablocks) of this stencil with the 
    *  data found in the given array. This method assumes that the array 
    *  data was generated by the getData() method of a stencil exactly equivalent
    *  to this one. In particular, the stencil dimensions must all be identical,
    *  and the same elements must already be active.
    */
   public void setShortData(byte [] data){

      //int dataOffset = 0;
      int volume;
      DataBlock block;
      ShortBuffer buffer = ByteBuffer.wrap(data).asShortBuffer();
      
      int datalength = data.length;
            
      for(int i = 0; i< this.size; i++){
      
         block = this.getIndexedBlock(i);
         
         volume = block.volume();
      
         //System.arraycopy(data, dataOffset, block.getValues().getBytes(), 0, volume);
         buffer.get(block.getValues().getShorts(), 0, volume);
         
         //dataOffset += volume;
         
      }
   }
   

//   public IndexSpaceID getPartitionDimensions(){
//
//     return new IntegerIndexSpaceID(this.partitioning.getPartitionDimensionsArray());
//   }
//
//   /** Returns the dimensions of a partition as an int array.
//    */
//   public int [] getPartitionDimensionsArray(){
//   
//      return this.partitioning.getPartitionDimensionsArray();
//   }

   public int getStorageType(){
   
      return this.recordDescriptor.getStorageType();
   }



   /** Set the stencil element containing the id to "active". If it is already
     * active, then the method returns false. Returns true otherwise.
     * Not implemented.
     */
   public boolean set(IndexSpaceID id){
   

//      new SDBError(SDBError.Fatal, "Not Implemented");
//   
//      int offset = (int) this.partitioning.containingOffset(id);
//      
//      boolean wasntSet = this.inStencil[offset] == null;
//      
//      if(wasntSet) {
//         this.partitioning.getOffsetBounds(new ISBounds(this.dimensionality, offset));
//         this.inStencil[offset] = new BasicBlock(stencilBounds, this.recordDescriptor);         
//         this.size++;
//      }
//   
//      return wasntSet;
      
         return false;
   }


   /** Set the stencil element with index corresponding to id to "active.  id 
     * should already have been mapped to the partitioning space. If it is already
     * active, then the method returns false. Returns true otherwise.
     * Not implemented.
     */
   public  boolean setElement(IndexSpaceID pid){
   
      new SDBError(SDBError.Fatal, "Not Implemented");
      
      
//      int offset = (int) this.partitioning.elementOffset(pid);
//      
//      boolean wasntSet = this.inStencil[offset] == null;
//      
//      if(wasntSet) {
//      
//         this.inStencil[offset] = true;         
//         this.size++;
//      }
//   
//      return wasntSet;

         return false;
   }


   
   /** Set the stencil element containing the id to "active", associating the
    *  given DataBlock with that element. If it is already
     * active, then the method does not associate the new DataBlock with the 
     * element, and returns false. Returns true otherwise.
     */
   public boolean set(IndexSpaceID id, DataBlock block){
   
   
      int offset = (int) this.partitioning.containingOffset(id);
      
      boolean wasntSet = this.inStencil[offset] == null;
      
      if(wasntSet) {
      
         this.inStencil[offset] = block;         
         this.size++;
      }
   
      return wasntSet;
   }


   /** Set the stencil element with index corresponding to id to "active", 
     * associating the given DataBlock with that element.  id 
     * should already have been mapped to the partitioning space. If the
     * element is already active, then the new DataBlock is not associated,
     * and the method returns false.
     *  Returns true otherwise.
     */
   public  boolean setElement(IndexSpaceID pid, DataBlock block){
   
      try {
         int offset = (int) this.partitioning.elementOffset(pid);
         
         boolean empty = this.inStencil[offset] == null;
         
         if(empty) {
         
            this.inStencil[offset] = block;         
            this.size++;
         } else {
         
            System.out.println("    ***** DSBlockStencil.setElement(): element already set.");
         }
      
         return empty;
         
      } catch (ArrayIndexOutOfBoundsException e) {
      
         
         e.printStackTrace();
         
         new SDBError(SDBError.Warning, "DSBlockStencil.setElement(): Tried to "
         +"set element with id="+pid+", which caused an array out of bounds exception "
         +"because the stencil partitioning="+this.partitioning);
      
         return false;
      }
   }




   /** Return true if the given id maps to an active stencil element, false otherwise.*/
   public boolean get(IndexSpaceID id){

      int offset = (int) this.partitioning.containingOffset(id);
      
      return this.inStencil[offset] != null;
   }

   /** Return true if the given id maps to an active stencil element, false otherwise.*/
   public DataBlock getDataBlock(IndexSpaceID id){

      int offset = (int) this.partitioning.containingOffset(id);
      
      return this.inStencil[offset];
   }




   /** Sets the argument ISBounds to the bounds of the specified stencil element.
   *   The ordering of the stencil elements is arbitrary. If the specified index
   *   is not less than the number of elements, then the bounds argument is
   *   not changed, and the method returns null. Otherwise, the bounds are 
   *   modified and a reference to the argument bounds is returned.
   */
   public ISBounds getIndexedBounds( ISBounds stencilBounds, int index)
   {
      long offset;
      
      if( index < this.size){
      
         offset = this.getIndexedOffset(index);
         this.partitioning.getOffsetBounds(stencilBounds, offset);
         return stencilBounds;
      } else {
      
         return null;
      }
   }

   /** Return the DataBlock corresponding to the specified stencil element.
   *   The ordering of the stencil elements is arbitrary. If the specified index
   *   is not less than the number of elements, then the method returns null.
   */
   public DataBlock getIndexedBlock(int index)
   {
      long offset;
      
      if( index < this.size){
      
         offset = this.getIndexedOffset(index);
         return inStencil[(int) offset];
      } else {
      
         return null;
      }
   }

   /** Return the DataBlock corresponding to the specified stencil element.
   *   The ordering of the stencil elements is arbitrary. If the specified index
   *   is not less than the number of elements, then the method returns null.
   *   The ISBounds argument will be set to the bounds in the stencil space
   *   corresponding to this DataBlock.
   */
   public DataBlock getIndexedBlockAndBounds(ISBounds stencilBounds, int index)
   {
      long offset;
      
      if( index < this.size){
      
         offset = this.getIndexedOffset(index);
         this.partitioning.getOffsetBounds(stencilBounds, offset);
         return inStencil[(int) offset];
      } else {
      
         return null;
      }
   }

   /** Return the partition offset corresponding to the given index.*/
   protected  final long getIndexedOffset(int index){

      int offset;
      int startIndex;
      
      if(index > this.lastIndexRequested) {
      
         offset = this.lastActiveOffset + 1;
         startIndex = this.lastIndexRequested + 1; // for loop runs (index - lir) times
      } else if(index == this.lastIndexRequested){
      
         offset = this.lastActiveOffset;
         startIndex = index + 1; // for loop won't run
      } else {
      
         offset = 0;
         startIndex = 0; // for loop runs (index + 1) times
      }
             
      for(int i=startIndex; i<=index; i++){
      
         while(inStencil[offset] == null)
         {
            offset++;
         }
      }
      
      this.lastIndexRequested = index;
      this.lastActiveOffset = offset;
       
//       // EXPERIMENTAL REPLACEMENT CODE:
//       int offset = 0;
//       
//       for(int i = 0; i < index; i++)
//       {
//           offset++;
//           while(inStencil[offset] == null && offset < inStencil.length)
//               offset++;
//           
//           if(offset >= inStencil.length)
//               offset = 0;
//       }
       
       return offset;   
   }



   /** Set the element corresponding to the 1D offset to be active.
     * Not implemented.
     */
   protected void setElementFromOffset(int offset){
   
      new SDBError(SDBError.Fatal, "Not Implemented");
   
//      this.inStencil[offset] = true;
//      this.size++;
   }



   /** return number of active elements contained by this stencil
   */
   public int size()
   {
     return this.size;
   }

   /** Clear this stencil of all elements.
   */
   public void clear(){

      Arrays.fill(this.inStencil, null);
      this.size = 0;
      
      this.lastIndexRequested = -1;
      this.lastActiveOffset = -1;
   }

   /** Return a String representing the object. */
   public String toString ()
   {
      
      return this.toString ("");
   }

   public String toString(String indent){
   
      String indent2=indent+Globals.toStringIndent;
      int length = this.inStencil.length;
      
      StringBuffer result = new StringBuffer(indent + "{ DSBlockStencil:\n");
      result.append(super.toString(indent2)+"\n");
      
      result.append(indent2 + "Stencil size =" + size +"\n");
      
      result.append(indent2 + "array= ");
      
      for (int i=0; i< length; i++){
      
         if(this.inStencil[i] != null){
            result.append("1 ");
         } else {
            result.append("0 ");
         }
      
      }
      
      result.append("\n");
      
      result.append(indent + "}");
      
      return new String(result);
   
   }


   //_____________________________________________________________



    public static void  main(String [] argv){

//        ISBounds spaceBounds=new ISBounds(new int [] {10,10});
//
//        //DSBlockStencil stencil = new DSBlockStencil(new int[] {5,5}, spaceBounds);
//        DSBlockStencil stencil = new DSBlockStencil(new RegularISPartitioning(new IntegerIndexSpaceID(new int[] {5,5}), spaceBounds));
//
//
//        stencil.set(new IntegerIndexSpaceID(2,2));
//
//        stencil.set(new IntegerIndexSpaceID(8,8));
//
//
//        System.out.println("2,2 yields "+stencil.get(new IntegerIndexSpaceID(2,2)));
//        System.out.println("2,8 yields "+stencil.get(new IntegerIndexSpaceID(2,8)));
//        System.out.println("8,2 yields "+stencil.get(new IntegerIndexSpaceID(8,2)));
//        System.out.println("8,8 yields "+stencil.get(new IntegerIndexSpaceID(8,8)));

    }


}

