//
//  DSListStencil.java
//  StencilTest
//
//  Created by Philip Rhodes on Sat Jan 17 2004.
//  Copyright (c) 2004 University of New Hampshire. All rights reserved.
//
//--------------------- Edit history ----------------------------
// 06/05/08 rdb: made ArrayList generic


package edu.unh.sdb.datasource;

import java.util.*;
//import java.io.*;
//import edu.unh.sdb.common.*;

/** This is an implementation of a partitioned stencil that uses a list of
 *  ISIDs to maintain the record of active elements. As a result, the set and get
 *  methods are O(n), but iteration is efficient, and less memory is used 
 *  than with the array implementation. However, it taxes the garbage collector
 *  by creating and destroying lots of IndexSpaceIDs, so it may not be so
 *  suitable for repeated use.
 */
public class DSListStencil extends DSPartitionedStencil{

   protected ArrayList<IndexSpaceID>  partitionList;


   /** Construct a stencil that identifies areas of interest using a partitioning.
    *  The first argument is the dimensions of an element of the partitioning, 
    *  meaning that the index space bounded by the second argument will be divided
    *  up into pieces of the shape described by the first argument. For example, 
    *  if we build an 8x8 partitioning of a 128x128 index space, each element
    *  of the partitioning will have dimensions 16x16. In this situation, 
    *  the first argument for this constructor should be [16,16] and the second
    *  argument should be {<0,0>, <127,127>}.
    */   
//   public DSListStencil( int [] elementDimensions, ISBounds space){
//
//     super(new RegularISPartitioning(elementDimensions, space));
//     this.partitionList = new ArrayList();
//   }

   /** Construct a stencil that identifies areas of interest using a partitioning. */
   public DSListStencil( RegularISPartitioning partitioning){
   
   
      super(partitioning);
      this.partitionList = new ArrayList<IndexSpaceID>();
   }




   
   /** Set the stencil element containing the id to "active". If it is already
     * active, then the method returns false. Returns true otherwise.
     */
   public boolean set(IndexSpaceID id){

      IndexSpaceID pid = this.partitioning.map(id);

      if( ! this.partitionList.contains(pid) ) {

         this.partitionList.add(pid);
         //System.out.println("Added "+pid+" to stencil.");
         
         return true;
      } else {

         return false;
      }
   }


   /** Set the stencil element with index corresponding to id to "active.  id 
     * should already have been mapped to the partitioning space. If it is already
     * active, then the method returns false. Returns true otherwise.
     */
   public  boolean setElement(IndexSpaceID pid){

      if( ! this.partitionList.contains(pid) ) {

         this.partitionList.add(pid);
         //System.out.println("Added "+pid+" to stencil.");
         
         return true;
      } else {

         return false;
      }
   }




   /** Return true if the given id maps to an active stencil element, false otherwise.*/
   public boolean get(IndexSpaceID id){

     IndexSpaceID pid = this.partitioning.map(id);

     return this.partitionList.contains(pid);
   }

   Object [] getPartitionListArray(){

     return this.partitionList.toArray();
   }



//   /** Fill the given array with a representation of the partition list.
//    *  This representation is meant to be parsed by setPartitionListFromBytes().
//    *  If the array argument is null or refers to an array of insufficient
//    *  length, a new byte array will be allocated and a reference returned.
//    *  After this method returns, the number of bytes consumed by the representation
//    *  can be retrieved by a call to getPartitionListBytesLength().
//    */
//   public byte [] getPartitionListBytes(byte [] bytes){
//
//
//      int len = this.partitionList.size();
//      int d = this.partitioning.dim();
//      int [] coords;
//
//      // Set the length data member so it can be retrieved using
//      // getPartitionListBytesLength().
//
//      this.partitionListBytesLength = len + this.offset;
//
//      // If the argument was either null, or points to an array
//      // that is too short, we will allocate a new array.
//      if(bytes == null || bytes.length < (len * d * this.bytesPerCoord + this.offset)) {
//
//         bytes = new byte [len * d * this.bytesPerCoord + this.offset];
//      }
//
//      // Write the header.
//      // First, we write the number of active elements to the beginning
//      // of the array as a short.
//      bytes[0] = (byte) ((len & 0x0000FF00) >>> 8);
//      bytes[1] = (byte) ((len & 0x000000FF));
//
//      // Next, we write the number of bytes per coordinate as a short.
//      // This is too much space for a small value, but we can preserve
//      // fullword alignment if we use two bytes.
//
//      bytes[2] = (byte) (0);
//      bytes[3] = (byte) (this.bytesPerCoord);
//
//
//      // Now write the partition coordinates to the array.
//
//      int coord;
//
//      for(int i=0; i<len; i++){
//
//         coords = ((IntegerIndexSpaceID)this.partitionList.get(i)).refCoords();
//
//         for(int j=0; j<d; j++){
//
//            coord = coords[j];
//
//            for(int k = this.bytesPerCoord - 1; k >= 0; k--){
//
//
//               bytes[this.bytesPerCoord * (i * d + j) + k + this.offset] =
//                     (byte) (coord & 0x000000FF);
//
//               coord >>>= 8;
//            }
//         }
//      }
//
//      return bytes;
//   }
//
//
//
//   /** Set the partition list according to the data found in the
//    *  given array of bytes.
//    */
//   public byte [] setPartitionListFromBytes(byte [] bytes){
//
//      int d = this.partitioning.dim();
//      int [] coords = new int[d];
//
//
//      // First, read the number of active partitions from the array.
//      int len = ( (bytes[0] & 0xFF) << 8 | (bytes[1] & 0xFF));
//
//      // Now read the number of bytes per coordinate
//      this.bytesPerCoord = ( (bytes[2] & 0xFF) << 8 | (bytes[3] & 0xFF));  // should be 4 for now.
//
//      this.clear();
//
//      for(int i=0; i<len; i++){
//         for(int j=0; j<d; j++){
//
//            coords[j] = bytes[this.offset + this.bytesPerCoord * (i * d + j) + 0];
//
//            for(int k=1; k< this.bytesPerCoord; k++){
//
//               coords[j] <<= 8;
//
//               coords[j] |= bytes[this.offset + this.bytesPerCoord * (i * d + j) + k] & 0xFF;
//
//            }
//         }
//
//         this.partitionList.add(new IntegerIndexSpaceID(coords));
//      }
//
//      return bytes;
//   }

   /** Set the element corresponding to the 1D offset to be active.*/
   protected void setElementFromOffset(int offset){
   
      this.partitionList.add(this.partitioning.getOffsetID(offset));
   }


   /** 
    * Return the partition offset corresponding to the given index.
    * The index identifies the desired partition. The returned value is
    * the 1D index of the start of the partition in 
    */
   protected  long getIndexedOffset(int index){

      return this.partitioning.partitionOffset(this.partitionList.get(index));
   }


   /** Set the given ISBounds to a shape equal to the indexed stencil element. */
   public ISBounds getIndexedBounds( ISBounds stencilBounds, int index)
   {
     IndexSpaceID partitionIndexID = this.partitionList.get(index);
     this.partitioning.getIndexedBounds( stencilBounds, partitionIndexID);
     
     return stencilBounds;
   }

   /** return number of element contained by this stencil
   */
   public int size()
   {
     return this.partitionList.size();
   }

   /** Clear this stencil of all elements.
   */
   public void clear(){

      this.partitionList.clear();
   }


   //_____________________________________________________________



    public static void  main(String [] argv){

        ISBounds spaceBounds=new ISBounds(new int [] {10,10});

        DSListStencil stencil = new DSListStencil(new RegularISPartitioning(new IntegerIndexSpaceID(new int[] {5,5}), spaceBounds));


        stencil.set(new IntegerIndexSpaceID(2,2));

        stencil.set(new IntegerIndexSpaceID(8,8));


        System.out.println("2,2 yields "+stencil.get(new IntegerIndexSpaceID(2,2)));
        System.out.println("2,8 yields "+stencil.get(new IntegerIndexSpaceID(2,8)));
        System.out.println("8,2 yields "+stencil.get(new IntegerIndexSpaceID(8,2)));
        System.out.println("8,8 yields "+stencil.get(new IntegerIndexSpaceID(8,8)));

    }


}
