/* 
    DataBlock.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824

 * ------------------------- Edit history as of 12/5/04 ---------------------
 * 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
 * 01/31/05 rdb: Added set<type>s( <type>[] )
 * 01/25/05 rdb: Moved abstract datum method (and comments to DataCollection)
 * 01/21/05 rdb: Added set<type>s( int attr, <type>[] )
 * 01/13/05 rdb: Revised constructor args. Bounds and RecordDescriptor are always required, 
 *                  so they should always be the first two. 
 *                -Added:       DataBlock( ISB, RecordDescriptor, AxisOrdering )
 *                -Deprecated : DataBlock( ISB, AxisOrdering, RecordDescriptor )
 *                -Also revised DataBlock( ISB, RD ) and DataBlock( DataBlock ),
 *                 so both call the principal constructor and share code, rather than
 *                 replicate code.
 *                  
 * 12/05/04 rdb: Deleted (or moved) dimensionality, recordDesc, etc. and associated 
 *               access methods since these are now inherited from DataCollection   
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

/** An abstract parent class for datablocks--used to transport blocks of data around the system */

public abstract class DataBlock extends DataCollection
{    
    /** Empty Constructor  */
    protected DataBlock() {}

    /** Construct a datablock with the given bounds, axis axisOrdering and RecordDescriptor. 
     * The bounds will be moved to the origin.
     * @deprecated  Use DataBlock( ISBounds, RecordDescriptor, AxisOrdering )
     */
    public DataBlock(ISBounds isb, AxisOrdering ordering, RecordDescriptor recordDesc)
    {
       this( isb, recordDesc, ordering );
    }
    /** Construct a datablock with the given bounds, axis axisOrdering and RecordDescriptor. 
     * The bounds will be moved to the origin.
     */
    public DataBlock( ISBounds isb, RecordDescriptor recordDesc, AxisOrdering ordering )
    {
       super( isb, recordDesc, ordering );
        
//       if ( ordering == null )
//          this.axisOrdering=new AxisOrdering(isb.dim());
//       else 
//       {
//          if (ordering.dim() == isb.dim())
//             this.axisOrdering=ordering.copy();
//          else 
//          {
//             new DimensionalityError(SDBError.Recoverable,
//                 "Dimensionalites of ISBounds and Axisordering arguments do not match.");
//             this.axisOrdering=new AxisOrdering(isb.dim());
//          }
//       }
    }

    /** Construct a datablock with the given bounds and number of attributes. 
        The default axis axisOrdering is used, and 
        the bounds will be moved to the origin.
        */
    public DataBlock(ISBounds isb, RecordDescriptor recordDesc)
    {
       this( isb, recordDesc, null );
    }

    /** Construct a DataBlock from the given DataBlock. Bounds, Ordering, and
        RecordDescriptor will be copied directly from the DataBlock argument.
        This is faster than using one of the constructors above.
    */
    public DataBlock( DataBlock b )
    {
       this( b.bounds, b.recordDesc, b.axisOrdering );
       //this.setValues( b.getValues() );
    }

    //-------------------- get methods -------------------------------------
    /** Returns a Values array which encapsulates a 1D array of the data
     *  values in the block with one entry for every point in the block.
     *  The type of the Values object is determined by the underlying storage
     *  type(s) needed. This default implementation returns the Values in the 
     *  order and type defined by the block. This implementation is expected  
     *  to be overridden by the child classes for efficiency.
     */
    public Values getValues()
    {
       Values ret = this.recordDesc.createValues( this.volume() );
       int    iVal = 0; // index into Values data
       int    nAttr = this.getNumAttributes();
       
       if ( this.getDataOrder() == DataCollection.attributeOrder )
       {
          for ( int attr = 0; attr < nAttr; attr++ )
          {
             Values aVals = this.getValues( attr );
             ret.setValues( iVal, aVals );
             iVal += this.volume();
          }
       }
       else 
       {
          Datum dat = this.createDatum();
          ISIterator iter = new ISIterator( this.bounds );
          for ( ; iter.valid(); iter.next() )
          {
             datum( dat, iter );
             Values datVals = dat.getValues();
             ret.setValues( iVal, datVals );
             iVal += nAttr;
          }
       }
       return ret;
    }
    //-------------------- set Attribute methods ---------------------------
    /** Set the location specified by the IndexSpaceID to the given datum.*/
    abstract public void setDatum( IndexSpaceID isid, Datum d );   

    /** For the position of the 1st argument, set the attribute specified by the 
     *  2nd argument to the value of the 3rd argument.  
     */    
    public void setDouble( IndexSpaceID isid, int attr, double v )
    {
       Datum   d = this.datum( isid );
       d.setDouble( attr, v );
       setDatum( isid, d );
    }

    /** For the position of the 1st argument, set the attribute specified by the 
     *  2nd argument to the value of the 3rd argument.  
     */    
    public void setFloat( IndexSpaceID isid, int attr, float v )
    {
       Datum   d = this.datum( isid );
       d.setFloat( attr, v );
       setDatum( isid, d );
    }

    /** For the position of the 1st argument, set the attribute specified by the 
     *  2nd argument to the value of the 3rd argument.  
     */    
    public void setInt( IndexSpaceID isid, int attr, int v )
    {
       Datum   d = this.datum( isid );
       d.setInt( attr, v );
       setDatum( isid, d );
    }

    /** For the position of the 1st argument, set the attribute specified by the 
     *  2nd argument to the value of the 3rd argument.  
     */    
    public void setShort( IndexSpaceID isid, int attr, short v )
    {
       Datum   d = this.datum( isid );
       d.setShort( attr, v );
       setDatum( isid, d );
    }
 	
    /** For the position of the 1st argument, set the attribute specified by the 
     *  2nd argument to the value of the 3rd argument.  
     */    
    public void setByte( IndexSpaceID isid, int attr, byte v )
    {
       Datum   d = this.datum( isid );
       d.setByte( attr, v );
       setDatum( isid, d );
    }

    //---------------- set<type> array methods ------------------
    // These are "template methods". They utilize child-specific methods
    // but they work for children -- however, for performance reasons the
    // child classes might want to override.
    
    /** Set all attributes of the Datum at the specified isid to the values 
     *  in passed array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setDoubles( IndexSpaceID isid, double[] newVals )
    {
       Datum d = this.datum( isid );
       d.setDoubles( newVals );
       this.setDatum( isid, d );
    }
       
    /** Set all attributes of the Datum at the specified isid to the values 
     *  in passed array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setFloats( IndexSpaceID isid, float[] newVals )
    {
       Datum d = this.datum( isid );
       d.setFloats( newVals );
       this.setDatum( isid, d );
    }
    
    /** Set all attributes of the Datum at the specified isid to the values 
     *  in passed array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setInts( IndexSpaceID isid, int[] newVals )
    {
       Datum d = this.datum( isid );
       d.setInts( newVals );
       this.setDatum( isid, d );
    }
    
    /** Set all attributes of the Datum at the specified isid to the values 
     *  in passed array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setShorts( IndexSpaceID isid, short[] newVals )
    {
       Datum d = this.datum( isid );
       d.setShorts( newVals );
       this.setDatum( isid, d );
    }
    
    /** Set all attributes of the Datum at the specified isid to the values 
     *  in passed array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setBytes( IndexSpaceID isid, byte[] newVals )
    {
       Datum d = this.datum( isid );
       d.setBytes( newVals );
       this.setDatum( isid, d );
    }

   /** Set the specified attribute of the Datum at all points to the values 
     *  in passed attribute array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setDoubles( int attr, double[] newVals )
    {
       ISIterator iter = new ISIterator( this.bounds );
       for ( int i = 0; i < newVals.length && iter.valid(); iter.next() )
          this.setDouble( iter, attr, newVals[ i++ ] );
    }

    /** Set the specified attribute of the Datum at all points to the values 
     *  in passed attribute array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setFloats( int attr, float[] newVals )
    {
       ISIterator iter = new ISIterator( this.bounds );
       for ( int i = 0; i < newVals.length && iter.valid(); iter.next() )
          this.setFloat( iter, attr, newVals[ i++ ] );
    }
       
    /** Set the specified attribute of the Datum at all points to the values 
     *  in passed attribute array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setInts( int attr, int[] newVals )
    {
       ISIterator iter = new ISIterator( this.bounds );
       for ( int i = 0; i < newVals.length && iter.valid(); iter.next() )
          this.setInt( iter, attr, newVals[ i++ ] );
    }
       
    /** Set the specified attribute of the Datum at all points to the values 
     *  in passed attribute array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setShorts( int attr, short[] newVals )
    {
       ISIterator iter = new ISIterator( this.bounds );
       for ( int i = 0; i < newVals.length && iter.valid(); iter.next() )
          this.setShort( iter, attr, newVals[ i++ ] );
    }
       
    /** Set the specified attribute of the Datum at all points to the values 
     *  in passed attribute array. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setBytes( int attr, byte[] newVals )
    {
       ISIterator iter = new ISIterator( this.bounds );
       for ( int i = 0; i < newVals.length && iter.valid(); iter.next() )
          this.setByte( iter, attr, newVals[ i++ ] );
    }
    
    /** Set the specified attribute of the Datum at all points to the values 
     *  in passed Values object. Convert as needed to internal type.
     *  If the passed array is too short, remaining attributes are left
     *  unchanged. If the passed array is too long, the values are updated, 
     *  and a recoverable error is generated.
     */    
    public void setValues( int attr, Values newVals )
    {
       switch ( this.getStorageType() )
       {
          case StorageType.Double: 
             this.setDoubles( attr, newVals.getDoubles() );
             break;
          case StorageType.Float: 
             this.setFloats( attr, newVals.getFloats() );
             break;
          case StorageType.Int: 
             this.setInts( attr, newVals.getInts() );
             break;
          case StorageType.Short: 
             this.setShorts( attr, newVals.getShorts() );
             break;
          case StorageType.Byte: 
             this.setBytes( attr, newVals.getBytes() );
             break;
       }
    }
    
    //----------------- set<type>sByPoint/Attribute( <type>[] ) --------------------   
	/**
	 * Set all the attributes of all points in the block. The input Values object is
	 * organized by point. 
	 * Convert as needed to internal storage order and internal type. 
	 * If the input array is not the right size an exception is thrown.
	 */ 
    public void setValuesByPoint( Values vals )
	{
       Values temp = vals;
	   if ( vals.size() != this.getValues().size() )
	   {
	      new SDBError( SDBError.Recoverable, "Incorrect array length. Set aborted.");
	      return;
	   }
	   
	   if ( this.dataOrder == DataCollection.attributeOrder )
	      temp = vals.reorder( this.numAttributes );  // reorder to attribute order
	   if ( !this.setValues( temp ) )   // false means child can't do it
	   {   
	      if ( temp == vals )  // if we didn't reorder to attribute order
	         temp = vals.reorder( this.numAttributes );  // do it now
	      setByAttribute( vals );    // reorder probably faster than doing each point
       }
	}

	/**
	 * Set all the attributes of all points in the block. The input Values object is
	 * organized by attribute. 
	 * Convert as needed to internal storage order and internal type. 
	 * If the input array is not the right size an exception is thrown.
	 */ 
    public void setValuesByAttribute( Values vals )
	{
       Values temp = vals;
	   if ( vals.size() != this.getValues().size() )
	   {
	      new SDBError( SDBError.Recoverable, "Incorrect array length. Set aborted.");
	      return;
	   }
	   if ( this.dataOrder == DataCollection.pointOrder )
	      temp = vals.reorder( this.volume() );  // reorder to point order
	   if ( !this.setValues( temp ) )   // false means child can't do it
	      setByAttribute( vals );     // so assign by attribute
    }
    
    /*
     * Assign a Values object in Attribute order by Attribute
     */
    protected void setByAttribute( Values vals )
    {
       for ( int start = 0; start < vals.size(); start += this.volume() )
          for ( int i = 0; i < this.numAttributes; i++ )
          {
             Values attrVals = vals.spawn( start, 1, this.volume() );
             this.setValues( i, attrVals );
          }         
    }

    /** Set all the attributes of all points in the block. The input array
     *  is organized by point. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setDoublesByPoint( double[] newVals )
    {
       setValuesByPoint( new DoubleValues( newVals ) );
    }
    
    /** Set all the attributes of all points in the block. The input array
     *  is organized by attribute. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setDoublesByAttribute( double[] newVals )
    {
       setValuesByAttribute( new DoubleValues( newVals ) );
    }

    /** Set all the attributes of all points in the block. The input array
     *  is organized by point. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setFloatsByPoint( float[] newVals )
    {
       setValuesByPoint( new FloatValues( newVals ) );
    }
    
    /** Set all the attributes of all points in the block. The input array
     *  is organized by attribute. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setFloatsByAttribute( float[] newVals )
    {
       setValuesByAttribute( new FloatValues( newVals ) );
    }

    /** Set all the attributes of all points in the block. The input array
     *  is organized by point. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setIntsByPoint( int[] newVals )
    {
       setValuesByPoint( new IntValues( newVals ) );
    }
    
    /** Set all the attributes of all points in the block. The input array
     *  is organized by attribute. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setIntsByAttribute( int[] newVals )
    {
       setValuesByAttribute( new IntValues( newVals ) );
    }

    /** Set all the attributes of all points in the block. The input array
     *  is organized by point. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setShortsByPoint( short[] newVals )
    {
       setValuesByPoint( new ShortValues( newVals ) );
    }
    
    /** Set all the attributes of all points in the block. The input array
     *  is organized by attribute. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setShortsByAttribute( short[] newVals )
    {
       setValuesByAttribute( new ShortValues( newVals ) );
    }

    /** Set all the attributes of all points in the block. The input array
     *  is organized by point. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setBytesByPoint( byte[] newVals )
    {
       setValuesByPoint( new ByteValues( newVals ) );
    }
    
    /** Set all the attributes of all points in the block. The input array
     *  is organized by attribute. Convert as needed to internal storage order and
     *  internal type. If the array is not the right size, an exception is thrown. 
     */ 
    public void setBytesByAttribute( byte[] newVals )
    {
       setValuesByAttribute( new ByteValues( newVals ) );
    }

    /** Set all the attributes of all points in the block. Convert as needed to 
     *  internal type. If the class does not support this mechanism, the
     *  return value is false, which is the default implementation.
     */ 
    protected boolean setValues( Values vals )
	{
	   return false;
	}

    //--------------------- subblock methods ----------------------------

    /** Write the values of the entire DataBlock index space to a string in
        a nice format. Uses the default axisOrdering. */
    public String toString(){
    
        int i,d=this.dim();
        ISIterator iter=new ISIterator(this.bounds);
        int [] upperBlockBounds=this.getUpperBoundsArray();
        StringBuffer retVal=new StringBuffer();
        
        
        for(iter.init();iter.valid();iter.next()){
        
            retVal.append(this.datum(iter));
            retVal.append("\t");
            
            for(i=d-1;i>0;i--){
               
                if(iter.getCoord(i)==upperBlockBounds[i]){
                    retVal.append("\n");
                } else {
                    break;
                }
            }
        }
        
        return new String(retVal);
    }
}
