//
//  DataSink.java
//  DataSinkTest
//
//  Created by Philip Rhodes on Mon May 05 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//

package edu.unh.sdb.datasource;

import java.io.*;

import edu.unh.sdb.common.*;

public class DataSink 
{
	protected RandomAccessFile raFile;
	protected String filename;
	protected long [] axisWeights;
	protected RecordDescriptor recordDescriptor;
	protected AxisOrdering ordering;
	protected ISBounds bounds;
	protected int dimensionality;
	protected int numFields;
	protected int datumSize;
	
	public DataSink(String filename, RecordDescriptor rd, ISBounds bounds)
   {	
		this.filename=new String(filename);	
		this.recordDescriptor=rd;
		this.numFields=rd.size();
		this.bounds=bounds;
		this.dimensionality=bounds.dim();
		
		this.ordering=new AxisOrdering(this.dimensionality);
		this.axisWeights=this.ordering.makeAxisWeights(this.bounds);
		this.datumSize=this.recordDescriptor.byteSize();
	}

	protected DataSink(String filename, RecordDescriptor rd)
   {
      this.filename=new String(filename);	
		this.recordDescriptor=rd;
		this.numFields=rd.size();
		this.bounds=new ISBounds(new IntegerIndexSpaceID(1)); 
		this.dimensionality=1;
		
		this.ordering=new AxisOrdering(this.dimensionality);
		this.axisWeights=this.ordering.makeAxisWeights(this.bounds);
		this.datumSize=this.recordDescriptor.byteSize();		
	}

    
	
	
	public ISBounds getBounds()
   {
		return (ISBounds) this.bounds.clone();
	}
	
	
	public void activate()
   {
		this.open();
	}

	public void deactivate()
   {	
		this.close();
	}

	public void write(Datum d, IndexSpaceID id)
   {	
		int offset=this.computeOffset(id)*this.datumSize;
		
		float value;
		int storageType;
		int i;
		
		//this.currentOffset=offset;
		
		try {
		
			this.raFile.seek(offset);
			
			for(i=0;i<this.numFields;i++){
			
				storageType=recordDescriptor.getAttribute(i).getStorageType();
				
				value=d.getFloat(i);
				
				switch(storageType){
				
					case StorageType.Float:
						this.raFile.writeFloat(value);
						break;
						
					case StorageType.Double:
						this.raFile.writeDouble(value);
						break;
						
					case StorageType.Int:
						this.raFile.writeInt((int)value);
						break;
						
					case StorageType.Short:
						this.raFile.writeShort((short)value);
						break;
						
					default:
						new InternalFaultError(SDBError.Fatal,
                    "Unsupported storage type code in DataSink.write(Datum, IndexSpaceID):"+storageType);
						break;
				}
			}
			
			
		} catch (IOException e){
			new FileError(SDBError.Recoverable, "IO Error in write(): "+e.getMessage());
		}
	}




	protected void open()
   {
		try {
        
            this.raFile=new RandomAccessFile(this.filename,"rw");
			
        } catch (IOException e){
        
            new FileError(SDBError.Recoverable,"Couldn't Open File:"+e);
        }	
	}

	
	
	protected void close()
   {	
		try {
	
            raFile.close();
            
		} catch (IOException e){
        
			new FileError(SDBError.Recoverable, "IO Error in close().");
		}
	}
	
	

    /** Map the IndexSpaceID to a single integer that serves as
    *  as an Datum offset into the physical file or network source. 
    *
    * @param IndexSpaceID dsid
    * @return int
    */
    protected final int computeOffset(IndexSpaceID dsid)
    {        
        int i,d=this.dimensionality,offset=0;
        int[] coords=dsid.refCoords();
        
        try{
            for(i=0;i<d;i++){
                offset+=coords[i]*axisWeights[i];
            }
            
            return offset;
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"ISID argument has the wrong dimensionality.");
            return 0;
        }
    }

    /** 
    * Maps an n-dimensional array of coordinates to an Datum offset in file
    * or array.
    *
    * @param int[] coordArray
    * @return int
    */
    protected final int computeOffset(int coordArray[])
    {
        int i,offset=0;
        int d=this.dimensionality;
       
        try{
            for(i=0;i<d;i++){
                offset+=coordArray[i]*axisWeights[i];
            }
            
            return offset;
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"Array argument has the wrong dimensionality.");
            return 0;
        }
    }
}
