/* 
    DataSourceBuilder.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
*/

//------------------- Edit history --------------------------------------
// 06/05/08 rdb: Made all Vectors generic
// 07/04/07 pjr: Trivial bug fix in makeComponentInfos()
// 11/28/05 rdb: Fixed defaulted field name in Composite.
//               Changed clearComponentMaps method to clearFieldMaps method.
//               Added Composite fieldName Vector since old code didn't work.
// 10/21/05 rdb: removed unused variables
//               put dummy references in main for method getFieldIDMappers()
//-----------------------------------------------------------------------

package edu.unh.sdb.datasource;
//import java.io.*;
import java.util.Vector;
import edu.unh.sdb.common.*;

//--------------------------DataSourceBuilder class ----------------------
/** Manufactures Logical DataSource objects such AttributeJoin and Blocked
DataSources.  Hold relevant data and validates before instantiating the appropriate
DataSource objects*/


abstract public class DataSourceBuilder
{
  /** table/list to hold the available datasources */
  protected DataSourceCollection dsList;
  
  /** the name of the pending datasource */
  private String dsName;
  
  /** the designated component Datasource indices into DataSourceList */
  protected Vector<Integer> componentIndices;
    
  /** the bounds for each component */
  protected Vector<ISBounds> componentBounds;
  
  /** denotes extent of Index Space for the pending datasource*/
  protected ISBounds compositeBounds; 

  /** interim lower bounds specifications */
  private Vector<Integer> componentLowerAxes;
    
  /** interim upper bounds specifications */
  private Vector<Integer> componentUpperAxes;
    
  /** datasource is not to be instantiated */
  protected boolean discardFlag;
    
  // declare the active file descriptor
  private FDLDescriptor currentFDLDescriptor;

  // the fieldID mappings for each components
  protected Vector<FieldIDMapper> componentMaps;
  
  /** the component field positions for mapping from component to composite */
  protected Vector<Integer> componentFieldPositions;
    
  /** the composite field positions for mapping from component to composite */
  protected Vector<Integer> compositeFieldPositions = null;
    
  /** the composite field names for mapping from component to composite */
  protected Vector<String> nameMapping = null;
  
  /** rdb added a single composite name vector; could have holes in it */
  protected Vector<String> compositeFieldNames = null;
  
  /******************* 
   * rdb: 11/28/05 Added dumb private class to encapsulate the 3 separate Vectors
   * that had been used. These were not correctly used. We need to save the field
   * mappings for each of the Components since it is a 2-pass process. Old code 
   * deleted the parallel Vectors after each step of the 1st pass. 
  private class FieldMapInfo 
  {
     public int  componentIndex;   // which attribute in component is it
     public int  compositeIndex;   // which attribute in composite
     public String compositeName;  // name of field in composite
  }
  
  // Need array of FieldMapInfo objects for the current component DS.
  private Vector currentFieldMappings = null;
  // And we need an array of FieldMappings arrays, one for each component
  private Vector allMappings = null;
  /********************************/
    
  /** Default Constructor 
  */
  public DataSourceBuilder( DataSourceCollection d, String n )
  {
    dsList = d;
    dsName = n;
    componentIndices = new Vector<Integer>();
    compositeBounds = null;
    componentBounds = new Vector<ISBounds>();
    clearComponentAxes();
    
    componentMaps = new Vector<FieldIDMapper>();
    //rdb componentFieldMappings
    //rdb currentFieldMappings  = new Vector(); 
    
    componentFieldPositions = new Vector<Integer>();
    compositeFieldPositions = new Vector<Integer>();
    nameMapping = new Vector<String>();
    
    //rdb positeFields = new Vector(); 
    compositeFieldNames = new Vector<String>();  // keep all composite names in one vector
   }
   
    /** Sets the datasource name to the String provided*/
    public void setDSName(String s)
    {
      dsName = s;
    }
  
    /** Returns the datasource name*/
    public String getDSName()
    {
      return dsName;
    }

    /** Sets the current File Descriptor Language object*/
    public void setCurrentFDLDescriptor(FDLDescriptor fdl)
    {
      currentFDLDescriptor = fdl;
    }
  
    /** Returns the current File Descriptor Language object*/
    public FDLDescriptor getCurrentFDLDescriptor()
    {
      return currentFDLDescriptor;
    }

    /** Returns the DataSource object mapped to the most recent compoent */
    protected DataSource getCurrentDataSource()
    {
      return dsList.getElementAt(
         componentIndices.lastElement().intValue());  
    }
   
  /** Sets the bounds of the Index Space for the pending datasource */
    public void setCompositeBounds(ISBounds bounds)
    {
      if (bounds!=null)
        compositeBounds=bounds.copy();
      else
        new IllegalArgumentError(SDBError.Recoverable,"DataSourceBuilder.setBounds() was passed a null reference");
    }
    
    /** Returns the bounds of the Index Space for the pending datasource.
    If no bounds were specified, the bounds of the first component datasource
    will be returned.
    */
    public ISBounds getCompositeBounds()
    {
      if (compositeBounds != null)
        return compositeBounds.copy();
      else
      {
        ISBounds[] isbArray = getAllComponentBounds();
        if (isbArray != null)
        {
          if (getAllComponentBounds()[0] != null)
            return getAllComponentBounds()[0].copy();
        }
      }
        
      System.err.println("No bounds specified for composite data source: "
              + getDSName());
      return null;
    }
    
    /** Returns an array containing the bounds of the Index Space for each component
    of the pending datasource.    */
    protected ISBounds[] getAllComponentBounds()
    {
      ISBounds[] componentBoundsArray = new ISBounds[componentBounds.size()];
        
      for (int i = 0; i < componentBounds.size(); i++)  //populate an array for use in constructor
        componentBoundsArray[i] = componentBounds.elementAt(i);
        
      return componentBoundsArray;
    }

    /** Returns an array containing the component datasources. */
    protected DataSource[] getComponentDataSources()
    {
      if ( componentIndices.size() < 1 )
        return null;
        
      DataSource[] componentSources = new DataSource[ componentIndices.size() ];
      
      int  j;
      for ( int i = 0; i < componentIndices.size(); i++)  //populate an array for use in constructor
      {
        j = componentIndices.elementAt(i).intValue();
        componentSources[ i ] = dsList.getElementAt( j );
      }
      
      return componentSources;
    }
    
    /** Returns the first DataSource in the component list
    or a null if no components have been specified. */
    protected DataSource getFirstComponentDataSource()
    {
      if ( componentIndices.size() < 1 )
        return null; // no datasources have been specified 

      return dsList.getElementAt( componentIndices.elementAt( 0 ).intValue() );
    }
    
  /** Converts upper and lower axes arrays into an ISBounds object **/
  public ISBounds makeComponentBounds()
  {
    //int[] lowerAxesArray, upperAxesArray;      
    //did user specify any bounds for this component?
    if (componentLowerAxes.size() < 1)
      return null; //no bounds, just do nothing

    return new ISBounds(
               new IntegerIndexSpaceID(vectorToIntArray(componentLowerAxes)), 
               new IntegerIndexSpaceID(vectorToIntArray(componentUpperAxes)));
  }

    /** makes an integer array out of a vector */
    protected static int[] vectorToIntArray(Vector v)
    {
      int[] array;
        
      if (v.size() == 0) //no bounds were specified
      {
        array = new int[1];
        array[0] = 0;
      }
      else
      {
        array = new int[v.size()];

        for (int i = 0; i < v.size(); i++)
            array[i] = ((Integer)v.elementAt(i)).intValue();
      }
        
      return array;
    }
    
  /** adds an upper and lower axes to an individual component bounds
  specification*/
  public void setComponentAxes(Integer l, Integer u)
  {
    componentLowerAxes.addElement(l);
    componentUpperAxes.addElement(u);
  }
    
  /** reinstantiates bounds pairs to start a new component **/
  private void clearComponentAxes()
  {
    componentLowerAxes = new Vector<Integer>();
    componentUpperAxes = new Vector<Integer>();
  }   
    
  /** returns an ISIDMapper for a component */
  protected ISIDMapper makeISIDMapper(int i)
  {
    if (i < componentBounds.size())
    {
      return new ISIDMapper(
            componentBounds.elementAt(i), 
           getCompositeBounds());
    }
    else 
       return null;
  }
  /** adds DataSource named by the String to the list of components
  with the bounds implied to be that of the datasource named by the String.*/
  public boolean addComponent(String s)
  {     
    if (dsList.indexOfName(s) != -1) //-1 means that the name was not on the list
    {           
      //take the bounds from the original datasource component
      ISBounds b = makeComponentBounds();  //try to make bounds from DDL specification
      clearComponentAxes();  //set up for a new
            
      if (b == null) //user did not specify bounds, so use the original ones
      {
        b = dsList.getDataSource(s).copyBounds();
      }
            
      if (b == null) //no bounds could be found for this object
      {
        System.err.println("No bounds specified for datasource: " + s);
        return false; //shouldn't happen
      }
            
      return addComponent(s, b);
    }
      
    return false; //shouldn't get here
  }

  /** adds DataSource named by the String to the list of components
  with the bounds specified.*/
  public boolean addComponent( String s, ISBounds b )
  {          
    if ( dsList.indexOfName(s) != -1 ) //-1 means that the name was not on the list
    {
      componentIndices.addElement( new Integer( dsList.indexOfName( s ) )); //add index to list
      componentBounds.addElement( b );
      clearFieldMaps();
      return true;
    }
        
    return false;
  }  

    /**completes a new FieldIDMapper for current component */
    public void processFinalComponentMap()
    {
        componentMaps.addElement( makeComponentFieldIDMapper() );
    } 
    
    /** store fieldname/position pairs for the component attributes **/
    public void setComponentFieldPair( String name, Integer compositePosition, 
                                                    String compositeName )
    {
       if ( compositeName == null || compositeName.equals( "" ))
          compositeName = name;
       
      int componentPosition = getComponentFieldPosition( name );
       
      if ( componentPosition >= 0 && compositePosition.intValue() >= 0 )
      {
         // save mapping for component to composite
        componentFieldPositions.addElement( new Integer( componentPosition ));
        compositeFieldPositions.addElement( compositePosition );
        nameMapping.addElement( compositeName );
        
        // also save the Composite name in appropriate position for composite.
        int index = compositePosition.intValue();
        // index may not come in order, so make sure Vector is allocated up 
        // to this position.
        for ( int i = compositeFieldNames.size(); i <= index; i++ )
           compositeFieldNames.add( "" );
        // now set the correct entry to this name.
        String oldName = compositeFieldNames.set( index, compositeName );
        if ( !oldName.equals( "" ) )
        {
           new SDBError( SDBError.Warning, "Composite field used multiple times: " +
                       oldName + "@" + index + ". Other name is: " + compositeName );
        }
      }
    }
    
    /** Store fieldname/position pairs for the component attributes. New name
    is defaulted to be the same as the original name.**/
    public void setComponentFieldPair( String name, Integer compositePosition )
    {
      setComponentFieldPair( name, compositePosition, name );       
    }
    
  /** Examines the current component datasource and returns the 
  position of the field named by "name".  If field name does not exist
  -1 will be returned */
  protected int getComponentFieldPosition( String name )
  {
    //get the datasource
    DataSource ds = getCurrentDataSource();
        
    if (getCurrentDataSource() != null)
    { //we have a datasource
      //get record descriptor
      RecordDescriptor rd = ds.getRecordDescriptor();
      if ( rd != null )
      {// we have a record descriptor
        for ( int i = 0; i < rd.size(); i++)
        {
          if ( rd.getAttribute(i).name().equals( name ))
            return i;
        }
      }
    }       
        
    return -1;
  }
    
  /** returns a FieldIDMapper for the current component **/
  protected FieldIDMapper makeComponentFieldIDMapper()
  {
    if (componentFieldPositions.isEmpty() ||
        compositeFieldPositions.isEmpty())  //user didn't specify fields
    { 
        
      //make up a FieldIDMapper with all fields from the component datasource
      if (getCurrentDataSource() != null)
      { //we have a componentdatasource, get record descriptor
        RecordDescriptor rd = getCurrentDataSource().getRecordDescriptor();
        int[] array; 
    
        if (rd != null)
        {// we have a record descriptor, make array for mapping
          if (rd.size() < 1)
          {
            array = new int[1];
            array[0] = 0;
          } 
          else 
          {
            array = new int[rd.size()];
            for (int i = 0; i < rd.size(); i++)
              array[i] = i;
          }
          return new FieldIDMapper(array, array);
        }
      }
        System.err.println("Could not make a FieldIDMapper.");
        return null;  //this shouldn't happen
    }
    else  //we have a field positions specified;
    { // ****later*** need to handle composite field names here 
        
      return new FieldIDMapper(
                vectorToIntArray(componentFieldPositions),
                vectorToIntArray(compositeFieldPositions));
    }
  }
   
  /** Creates an array of FieldIDMapper objects form the specified components 
   * Currently isn't used, but have kept it since we probably some day will need 
   * it
   */
  private FieldIDMapper[] getFieldIDMappers()
  {
      FieldIDMapper[] array;
        
      if ( componentMaps.isEmpty() ) // were specified
      {
        array = null;
      }
      else
      { //we have fieldIDMappers to return
        array = new FieldIDMapper[ componentMaps.size() ];

        for ( int i = 0; i < componentMaps.size(); i++)
            array[ i ] = componentMaps.elementAt( i );
      }
        
      return array;
  }    
    
  /** reinstantiates field ID pairs to start a new component **/
  private void clearFieldMaps()
  {
    componentFieldPositions.clear();
    compositeFieldPositions.clear();
    nameMapping.clear();     //rdb 11/28/05
  }
    
  /** instantiate a list of ComponentInfos to be used to 
    create the datasource */
  
  protected ComponentInfo[] makeComponentInfos()
  {
    ComponentInfo [] ci = new ComponentInfo[ componentMaps.size() ];
            
    for ( int i = 0; i < componentMaps.size(); i++)
    {
       DataSource ds = dsList.getElementAt( 
                                     (componentIndices.elementAt(i) ).intValue());
       FieldIDMapper fim = componentMaps.elementAt( i );
       
       ci[ i ] = new ComponentInfo( ds, fim, makeISIDMapper( i ) ); 
       
      /* set each field position to the new field name. */
      /* most of the time these "new names" will be same as old ones */
      /*** rdb rewrote 11/28/05
      for ( int j = 0; j < compositeFieldPositions.size(); j++ )
      {
        ci[i].setAttributeName(
            ((Integer)compositeFieldPositions.elementAt( j )).intValue(),
            (String)nameMapping.elementAt( j ));
      }
      /******************************/
      
      // PJR:  added small fix to properly use the components array in the loop 7/4/07
      int[] components = fim.getComponentFields();
      for ( int j = 0; j < components.length; j++ )
      {
         ci[i].setAttributeName( j, 
                           compositeFieldNames.get( fim.toComposite( components[j] )));
      }
    }     
    return ci;
  }
    /** instantiates an actual datasource **/
    public DataSource makeDataSource()
    {
       return null;
    }

 /** diagnostic utility method for dumping this class to a string **/
    public String toString()
    {
      String s = "Number of components: " ;

      if ( componentIndices == null )
        return s;
        
      int size = componentIndices.size();
      s = s + size;
      DataSource ds;
      Integer component;
        
      for (int i = 0; i < size; i++)
      {
        s = s + "\n" + i + ". ";
        component = componentIndices.elementAt(i);
            
        if (componentBounds.elementAt(i) != null)
          s = s + " component bounds" + componentBounds.elementAt(i).toString();
        else
          s = s + " component bounds are null ";
            
        if (component != null)
          s = s + component.intValue();
            
        s = s + "- ";
        ds = dsList.getElementAt(component.intValue());

        if (ds != null)
          s = s + ds.toString(); 
      }
        
      return s;
    }
    //------------ main DUMMY ----------------
    public static void main( String[] args )
    {
       // only purpose is to make a reference to getFieldIDMappers method.
       // Some day we may get around to using it
       DataSourceBuilder dummy = new AJDataSourceBuilder( null, "Dummy" );
       FieldIDMapper[] fims = dummy.getFieldIDMappers();
       System.out.println( fims.length );
    }
 }
