/* 
DataSourceCollection.java
Copyright 2002 University of New Hampshire
Durham, NH 03824
Authored by Denise Mitchell
*/

package edu.unh.sdb.datasource;
//import java.io.*;
import java.util.*;

/**
 * The DataSourceCollection class provides for a list of datasources
 * with customary list functionality.  The list was intended to be
 * populated either by reading a database or by processing some Data
 * Description Language.  The list is only allowed to grow.  Elements
 * are never deleted from the list.
 */

public class DataSourceCollection 
{
    /** The datasource list */
    private Vector<DataSource> dsList;
    
    /** holds index of next element in stored in the vector */
    private int next;
    
    public DataSourceCollection()
    {
        dsList = new Vector<DataSource>();
        next = 0;
    }
    
    /** returns the next DataSource object or null DataSource object if no more exists */
    public DataSource next()
    {
        next++;
        if (next <= dsList.size())
          return getElementAt(next-1);
        else
          return null;
    }
    
    /** Returns the DataSource at the designated index or null DataSource object if no more exists.
    Returns a null DataSource object if index does not exist. */
    public DataSource getElementAt(int i)
    {
        if (i < dsList.size())
            return dsList.elementAt(i);
        else 
            return null;
    }
    
    /** Returns the DataSource at the designated index or null DataSource object if no more exists.
    Returns a null DataSource object if index does not exist. This is a shorthand method to getElementAt() 
    added for compatibility with SDB naming conventions.*/
    public DataSource get(int i)
    {
        return getElementAt(i);
    }
    
    /** Add a DataSource to the end of the list */
    public void add(DataSource ds)
    {
        if (ds != null)
        {
            dsList.addElement(ds);
        }
        else
            System.err.println("Request to add a null data source to collection is being disregarded ");
    }
    
    /** Resets the index to beginning of the list */
    public void rewind()
    {
        next = 0;
    }
    
    /** Returns the integer index value for the DataSource denoted by string. Returns
    -1 if the DataSource name does not exist**/
    public int indexOfName(String s)
    {
        DataSource ds;
        
        int i;
        for (i = 0; i < dsList.size(); i++)
        {
            ds = dsList.elementAt(i);
            if (ds.getName().equals(s)) //names are the same so...
                return i; //return the matching index
        }
        return -1;
    }
    
    /** Returns true if DataSource named by String is in the list */
    public boolean nameExists(String s)
    {
        /* if it exists, return true else */
        int size = dsList.size();
        
        for (int i = 0; i < size; i++)
        {
            if (s.equals( getElementAt(i).getName()) )
                return true; 
        }
        
        return false; 
    }
    
    /** Returns the integer index of the datasource corresponding to the 
    DataSource named by the string.  If name does not exist, a null DataSource will
    be returned**/
    public int indexOf(String s)
    {
        return dsList.indexOf(s);
    }
    
    /** Returns the DataSource object corresponding to the DataSource named
    by the string **/
    public DataSource getDataSource(String s)
    {
        int size = dsList.size();
        DataSource ds;
        
        int i;
        for (i = 0; i < size; i++)
        {
            ds = dsList.elementAt(i);
            if (ds.getName().equals(s)) //names are the same so...
                return ds; //return the matching index
        }
        return null;
    }
    
    /** Returns the DataSource object corresponding to the DataSource named
    by the string.  This is a shorthand for getDataSource(String) to comply
    with SDB naming conventions.**/
    public DataSource get(String s)
    {
        return getDataSource(s);
    }
    
    /** Returns the number of elements stored this DataSource collection as an integer
    **/
    public int size()
    {
        return dsList.size();
    }
    
    /** adds an entire new collection onto the existing list **/
    /** if name exists in both collections, the new one is discarded **/
    public void merge(DataSourceCollection dsColl)
    {
        DataSource ds;
        
        for (int i = 0; i < dsColl.size(); i++)
        {
            ds = dsColl.get(i);
            
            if (nameExists(ds.getName() ))
            {
              System.out.println("DataSource : " + ds.getName() +
                    " already exists.  Ignoring subsequent definition: ");
            }
            else
            add(ds);
        }
    }
    
    /** for debugging - prints the total number of datasource collections stored.
    Later will give names of each datasource */
    public String toString()
    {
        String s = "Size of datasource collection: " + size();
        for (int i = 0; i < size(); i++)
        {
            s = s + "\n" + i + ". " + (dsList.elementAt(i)).getName(); 
        }
        return s;
        
    }
    
}

