package edu.unh.sdb.datasource;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.common.EulerAngles;
import edu.unh.sdb.common.FieldIDMapper;
import edu.unh.sdb.common.SDBError;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sourceforge.sizeof.SizeOf;

/**
 * Abstract class for a dynamic chunking data source that provides
 * methods that either read data from memory if the 
 * block that contained it is loaded, or loads the block, eventually discards an 
 * already loaded block in LRU fashion and then returns the data from the loaded 
 * block. Each time a block is accessed, it it moved at the begining of the LRU 
 * list.
 * @author dlipsa
 */
public class DynamicChunking extends DataSource
{

   /**
    * Creates a DynamicChunking object which views a 'dataSource' through a 
    * DynamicChunking memory manager and it has the same interface as a 
    * DataSource.
    */
   public DynamicChunking (DataSource dataSource)
   {
      super (dataSource.getName (), dataSource.getRecordDescriptor (),
         dataSource.getBounds ());
      this.dataSource = dataSource;
      iterationSubvolume = dataSource.getBounds ();
      blockTableOptimization = false;
      blockOrdering = new AxisOrdering (dataSource.dim ());
      lruListHeader = new Block (null, 0);
      lruListHeader.next = lruListHeader;
      lruListHeader.previous = lruListHeader;
   }

   public DataSource getDataSource ()
   {
      return dataSource;
   }

   public void setMemory (long memory)
   {
      long freeMem = getFreeMemoryForDynamicChunking ();
      if (freeMem < memory)
         memory = freeMem;
      this.memory = memory;
   }

   public long getMemory ()
   {
      return memory;
   }

   /**
    * If the application specifies the iteration subvolume, the block table
    * is set to cover only the subvolume not the entire dataset.
    * @param bounds
    */
   public void setIterationSubvolume (ISBounds bounds)
   {
      if (! blockTableOptimization)
      {
         iterationSubvolume = bounds;
         blockTableOptimization = true;
         logger.info ("Block Table Optimization: iterationSubvolume " + iterationSubvolume);
      }
   }
   
   /**
    * Uses the maximum block size given the amount of available memory.
    * Calculates the block size analytically for a slicing application.
    * @param eulerAngles
    * @param iterationSubvolume
    */
   public void setBlockSizeSliceOptimization (EulerAngles eulerAngles,
      ISBounds iterationSubvolume)
   {
      setIterationSubvolume (iterationSubvolume);
      IndexSpaceID dimensions = iterationSubvolume.getDimensions ();
      blockSizeSliceOptimization = new BlockSizeSliceOptimization (
         eulerAngles, dimensions);
      logger.info ("Block Size Slice Optimization: eulerAngles " + eulerAngles + 
         " iterationSubvolume " + iterationSubvolume);
   }

   public void setBlockSizeAdaptiveOptimization (
      ISBounds iterationSubvolume, String iterationId)
   {
      setIterationSubvolume (iterationSubvolume);
      if (blockDimensions == null)
         throw new IllegalStateException (
            "You need a default block size for block size adaptive optimization");
      blockSizeAdaptiveOptimization = new BlockSizeAdaptiveOptimization (
         iterationId, blockDimensions.getCoord (0));
   }   
   
   public ISBounds getIterationSubvolume ()
   {
      return iterationSubvolume;
   }
   
   /**
    * Sets up a callback function that is called each time a block is accessed.
    * @param blockAccessed function to be called, if null no function is called.
    */
   public void setBlockAccessedCallback (BlockAccessCallback blockAccessed)
   {
      this.callback = blockAccessed;
   }

   public void setBlockDimensions (IndexSpaceID dimensions)
   {
      if (partitionRatio != null)
         throw new IllegalStateException (
            "You cannot set both blockDimension and partitionRatio");         
      this.blockDimensions = dimensions;
   }   
   
   public void setPartitionRatio (IndexSpaceID partitionRatio)
   {
      if (blockDimensions != null)
         throw new IllegalStateException (
            "You cannot set both blockDimension and partitionRatio");
      this.partitionRatio = partitionRatio;
   }

   public IndexSpaceID getBlockDimensions ()
   {
      return blockDimensions;
   }

   public String getStatistics ()
   {
      return "page hit=" + blockHit +
         " page load=" + memoryBlockCount +
         " page replace=" + blockReplace;
   }

   /**
    * Activates the cache for the given memory and block size. If the memory
    * is 0, it frees any memory that was previously used, if any.
    */
   @Override
   public void activate ()
   {
      if (super.isActivated ())
         return;
      super.activate ();
      dataSource.activate ();
      setupSizeOf ();
      partitionRatio2blockDimensions ();
      if (memory == 0 || !dataSource.isActivated ())
      {
         deactivate ();
         return;
      }
      
      int voxelSize = dataSource.datum (iterationSubvolume.getLower ()).byteSize ();
      if (isOptimizedBlockSizeSlice ())
      {
         blockSizeSliceOptimization.setVoxelSize (voxelSize);
         blockSizeSliceOptimization.setAvailableMemory (memory);
         int blockEdge = blockSizeSliceOptimization.optimizeBlockSize ();
         if (blockEdge != 0)
         {
            for (int i = 0; i < blockDimensions.dim (); i++)
               blockDimensions.setCoord (i, blockEdge);
         }
      }
      else if (isOptimizedBlockSizeAdaptive ())
      {
         int blockEdge = blockSizeAdaptiveOptimization.readBlockEdge ();         
         for (int i = 0; i < blockDimensions.dim (); i++)
            blockDimensions.setCoord (i, blockEdge);         
         typeOfIteration = BlockSizeAdaptiveOptimization.TypeOfIteration.NO_THRASHING;
      }      
      isActivated = allocateBlockTable (voxelSize);
      if (isOptimizedBlockSizeAdaptive())
         blockSizeAdaptiveOptimization.initDiscoverThrashing(
            iterationSubvolume.getDimensions(), 
            blockDimensions.getCoord(0), blocks.length);
   }

   private boolean allocateBlockTable ( int voxelSize)
   {
      // calculate the dimensions of the data source in pages
      IndexSpaceID subvolumeDimensionsInPages = 
         iterationSubvolume.getDimensions ().setDimensionsInUnits (blockDimensions);
      axisWeightsForSubvolumeDimensionsInBlocks =
         blockOrdering.makeAxisWeights (subvolumeDimensionsInPages);

      int blockSize = voxelSize * (int) blockDimensions.volume ();
      int subvolumeBlockCount = (int) subvolumeDimensionsInPages.volume ();
      long memoryForBlockTable = BlockSizeSliceOptimization.estimateBlockTableMemory (subvolumeBlockCount);
      long memoryForABlock = BlockSizeSliceOptimization.estimateBlockMemory (blockSize);
      maxMemoryBlockCount = (int) 
              ((memory - memoryForBlockTable) / memoryForABlock);
      if (maxMemoryBlockCount <= 0)
      {
         DynamicChunking.logger.warning (
            "Not enough memory! Available memory is " + memory +
            ". Memory for block table is " + memoryForBlockTable + 
            ". Memory for a block is " + memoryForABlock);
         deactivate ();
         return false;
      }
      
      blocks = new Block[subvolumeBlockCount];
      DynamicChunking.logger.info (
         "Cache memory is " + memory + " which fits " + maxMemoryBlockCount + " blocks"+
         "\nBlock table has " + subvolumeBlockCount  + " entries, its memory is " + 
         SizeOf.sizeOf (blocks)  + " bytes which was estimated to " + 
         BlockSizeSliceOptimization.estimateBlockTableMemory (subvolumeBlockCount) + " bytes" +
         "\nBlock memory was estimated to " + BlockSizeSliceOptimization.estimateBlockMemory (blockSize));            
      return true;
   }
   
   
   /**
    * Dealocate all resources used for the DynamicChunking. If blockSizeAdaptiveOptimization
    * is enabled, it saves information about the current iteration which will
    * be used in future iterations.
    */
   @Override
   public void deactivate ()
   {
      if (isOptimizedBlockSizeAdaptive ())
      {
         blockSizeAdaptiveOptimization.writeIterationInfo (
            blockDimensions.getCoord (0), typeOfIteration);
      }
      super.deactivate ();
      memory = memoryBlockCount = blockHit = blockReplace = 0;
      lruListHeader.next = lruListHeader;
      lruListHeader.previous = lruListHeader;
      blocks = null;
   }

   public int getMaxMemoryBlockCount ()
   {
      return maxMemoryBlockCount;
   }

   @Override
   public void datum (Datum datum, IndexSpaceID dataId)
   {
      checkActivated ();
      if (DynamicChunking.logger.isLoggable (Level.FINER))
         DynamicChunking.logger.finer ("\ndataId=" + dataId);
      IndexSpaceID id = new IntegerIndexSpaceID (dataId);
      Block page = accessBlock (dataId, id);
      page.data.datum (datum, id);
   }

   /** not implemented */
   public void datum (Datum dtm, IndexSpaceID did, FieldIDMapper mp)
   {
      throw new UnsupportedOperationException ("Not yet implemented");
   }

   public void subblock (DataBlock dstBlock, ISBounds dstBounds,
      ISBounds srcBounds)
   {
      checkActivated ();
      if (DynamicChunking.logger.isLoggable (Level.FINE))
         logger.fine ("============== copy src: " +
            srcBounds + "to dst: " + dstBounds);
      if (!dstBounds.getDimensions ().equals (srcBounds.getDimensions ()))
      {
         String message = "dstBounds dimensions " + dstBounds.getDimensions () +
            " different than srcBounds dimensions " +
            srcBounds.getDimensions ();
         logger.warning (message);
         new SDBError (SDBError.Fatal, message);
         return;
      }
      ArrayList<Integer>[] srcStart = subblocksStart (srcBounds);
      ArrayList<Integer>[] dstStart = translateSubblockStart (
         srcStart, srcBounds.getLower (), dstBounds.getLower ());
      int[] lengths = new int[srcBounds.dim ()];
      // srcStart contains all start of subblocks and the last entry 
      // contains the end of a subblock. We want only to process only the 
      // start of subblocks.
      for (int i = 0; i < lengths.length; i++)
         lengths[i] = (srcStart[i].size () - 1);
      Iterator<int[]> it = new SubblockIndexIterator (lengths);
      ISBounds srcSubblockBounds = new ISBounds (
         new IntegerIndexSpaceID (srcBounds.getLower (), 0),
         new IntegerIndexSpaceID (srcBounds.getLower (), 0));
      ISBounds dstSubblockBounds = new ISBounds (srcSubblockBounds);
      while (it.hasNext ())
      {
         int[] startIndex = it.next ();         
         convertBlockStartIndex (startIndex, srcSubblockBounds, srcStart);
         convertBlockStartIndex (startIndex, dstSubblockBounds, dstStart);
         copySubblock (dstBlock, dstSubblockBounds, srcSubblockBounds);
      }
   }

   /** not implemented */
   public void subblock (DataBlock theBlock,
      ISBounds dstBounds, ISBounds srcBounds,
      FieldIDMapper mp)
   {
      if (mp == null || mp.isIdentityMapper ())
         subblock (theBlock, dstBounds, srcBounds);
      else
         throw new UnsupportedOperationException ("Not yet implemented");
   }

   public Iterator<ISBounds> getBlockBoundsIterator ()
   {
      return new BlockBoundsIterator ();
   }

   public boolean isOptimizedBlockTable ()
   {
      return blockTableOptimization;
   }

   public boolean isOptimizedBlockSizeSlice ()
   {
      return blockSizeSliceOptimization != null;
   }
   
   public boolean isOptimizedBlockSizeAdaptive ()
   {
      return blockSizeAdaptiveOptimization != null;
   }
   
   public interface BlockAccessCallback
   {
      void blockAccess (int hashCode, ISBounds blockBounds);
   }
   
   public void partitionRatio2blockDimensions ()
   {
      if (partitionRatio != null)
      {
         if (! blockTableOptimization)
            throw new IllegalStateException (
               "To specify partitionRatio you need blockTableOptimization");
         blockDimensions = new IntegerIndexSpaceID (
            iterationSubvolume.getDimension (0) / partitionRatio.getCoord (0),
            iterationSubvolume.getDimension (1) / partitionRatio.getCoord (1),
            iterationSubvolume.getDimension (2) / partitionRatio.getCoord (2));
      }
   }
   
       
   /**
    * For each coordinate (0,1,2) returns a list of dataSource relative IndexSpaceIDs
    * for the start of the each subblock. The start of each subblock happens at
    * a start of a dynamic chunking block. (Remember that dynamic chunking blocks
    * are relative to iterationSubvolume)
    * @param dataSourceRelativeBounds the subblock we are interested in
    * @return for each coordinate, it returns a list of IndexSpaceIDs
    */
   @SuppressWarnings("unchecked")
   private ArrayList<Integer>[] subblocksStart (ISBounds dataSourceRelativeBounds)
   {
      ArrayList<Integer>[] srcStart = new ArrayList[dataSourceRelativeBounds.dim ()];
      IndexSpaceID dataSourceRelativeLower = dataSourceRelativeBounds.getLower ();
      IndexSpaceID dataSourceRelativeUpper = dataSourceRelativeBounds.getUpper ();
      IndexSpaceID subvolumeRelativeLower = 
         new IntegerIndexSpaceID (dataSourceRelativeLower);
      dataSourceToSubvolumeAndBlock (dataSourceRelativeLower, 
         subvolumeRelativeLower, null);            
      IndexSpaceID lowerBlock = new IntegerIndexSpaceID (
         subvolumeRelativeLower).divide (blockDimensions);
      for (int i = 0; i < srcStart.length; i++)
      {
         srcStart[i] = new ArrayList<Integer> ();
         srcStart[i].add (dataSourceRelativeLower.getCoord (i));
         // newStart is dataSource relative
         int newStart = (lowerBlock.getCoord (i) + 1) *
            blockDimensions.getCoord (i) + iterationSubvolume.getLower (i);
         while (newStart <= dataSourceRelativeUpper.getCoord (i))
         {
            srcStart[i].add (newStart);
            newStart += blockDimensions.getCoord (i);
         }
         srcStart[i].add (dataSourceRelativeUpper.getCoord (i) + 1);
      }
      return srcStart;
   }

   @SuppressWarnings("unchecked")
   private ArrayList<Integer>[] translateSubblockStart (
      ArrayList<Integer>[] srcStart,
      IndexSpaceID srcLower, IndexSpaceID dstLower)
   {
      ArrayList<Integer>[] dstStart = new ArrayList[srcStart.length];
      for (int i = 0; i < srcStart.length; i++)
      {
         dstStart[i] = new ArrayList<Integer> ();
         Iterator it = srcStart[i].iterator ();
         while (it.hasNext ())
         {
            int val = ((Integer) it.next ()).intValue ();
            val -= srcLower.getCoord (i);
            val += dstLower.getCoord (i);
            dstStart[i].add (val);
         }
      }
      return dstStart;
   }

   private void remove (Block page)
   {
      page.previous.next = page.next;
      page.next.previous = page.previous;
   }

   private void addFirst (Block page)
   {
      // link the page at the begining of the list
      page.next = lruListHeader.next;
      page.previous = lruListHeader;
      lruListHeader.next.previous = page;
      lruListHeader.next = page;
   }

   private ISBounds dataSourceRelativeBlockBounds (int index)
   {
         int[] subvolumeRelativeBlockCoords =
            new int[axisWeightsForSubvolumeDimensionsInBlocks.length];
         blocks[index].data.getAxisOrdering ().computeCoords (
            index, subvolumeRelativeBlockCoords, axisWeightsForSubvolumeDimensionsInBlocks);
         IndexSpaceID subvolumeRelativeBlockCoordsId = 
            new IntegerIndexSpaceID (subvolumeRelativeBlockCoords);
         return dataSourceRelativeBlockBounds (subvolumeRelativeBlockCoordsId);
   }
   
   private ISBounds dataSourceRelativeBlockBounds (
      IndexSpaceID subvolumeRelativeBlockCoordsId)
   {
      int[] subvolumeRelativeBlockCoords = subvolumeRelativeBlockCoordsId.refCoords ();
      int[] blockDims = blockDimensions.refCoords ();
      ISBounds blockBounds = new ISBounds (blockDimensions);
      int[] lower = blockBounds.getLowerArray ();
      int[] upper = blockBounds.getUpperArray ();
      int[] dataSourceUpper = dataSource.getBounds ().getUpperArray ();
      for (int i = 0; i < lower.length; i++)
      {
         lower[i] = iterationSubvolume.getLower(i) +
            subvolumeRelativeBlockCoords[i] * blockDims[i];
         upper[i] = lower[i] + blockDims[i] - 1;
         if (upper[i] > dataSourceUpper[i])
            upper[i] = dataSourceUpper[i];
      }
      if (DynamicChunking.logger.isLoggable (Level.FINER))
         DynamicChunking.logger.finer (
            "\npageId=" + subvolumeRelativeBlockCoordsId + " pageBounds=" + blockBounds);
      return blockBounds;
   }

   private static void convertBlockStartIndex (
      int[] blockStartIndex, ISBounds block, ArrayList[] startSubblocks)
   {
      for (int i = 0; i < block.dim (); i++)
      {
         block.setLower (
            i, ((Integer) startSubblocks[i].get (blockStartIndex[i])).intValue ());
         block.setUpper (i,
            ((Integer) startSubblocks[i].get (
            blockStartIndex[i] + 1)).intValue () - 1);
      }
   }

   private void copySubblock (
      DataBlock dstBlock, ISBounds dstSubblockBounds,
      ISBounds dataSourceRelativeSrcSubblockBounds)
   {
      if (DynamicChunking.logger.isLoggable (Level.FINE))
         logger.fine (
            "copy src: " + dataSourceRelativeSrcSubblockBounds + 
            "to dst: " + dstSubblockBounds);
      IndexSpaceID blockRelativeLower = new IntegerIndexSpaceID (0, 0, 0);
      Block page = accessBlock (dataSourceRelativeSrcSubblockBounds.getLower (), blockRelativeLower);
      IndexSpaceID blockRelativeUpper = new IntegerIndexSpaceID (0, 0, 0);      
      dataSourceToSubvolumeAndBlock (dataSourceRelativeSrcSubblockBounds.getUpper (), null, blockRelativeUpper);
      page.data.subblock (
         dstBlock, dstSubblockBounds, new ISBounds (blockRelativeLower, blockRelativeUpper));
   }

   /**
    * Returns the block which contains 'dataSourceRelativeId' datum. Changes
    * 'blockRelativeId' to refer to the datum inside the block.
    */
   private Block accessBlock (
      IndexSpaceID dataSourceRelativeId, IndexSpaceID blockRelativeId)
   {
      IndexSpaceID subvolumeRelativeId = new IntegerIndexSpaceID (dataSourceRelativeId);
      dataSourceToSubvolumeAndBlock (dataSourceRelativeId, 
         subvolumeRelativeId, blockRelativeId);
      IndexSpaceID blockId =
         new IntegerIndexSpaceID (subvolumeRelativeId).divide (blockDimensions);
      int pageIndex = (int) AxisOrdering.computeOffset (
         blockId, axisWeightsForSubvolumeDimensionsInBlocks);
      // volume to page coordinates
      Block block = blocks[pageIndex];
      if (block == null)
      {
         ISBounds blockToRead = dataSourceRelativeBlockBounds (blockId);
         if (isOptimizedBlockSizeAdaptive () && 
            blockSizeAdaptiveOptimization.isThrashingIfReadBlock (
               blockToRead.getDimensions (), pageIndex))
         {
            typeOfIteration = BlockSizeAdaptiveOptimization.TypeOfIteration.THRASHING;
            logger.info ("" + BlockSizeAdaptiveOptimization.TypeOfIteration.THRASHING);
            throw new CacheThrashingError ();
         }         
         
         block = (memoryBlockCount < maxMemoryBlockCount) ? 
            createLRUBlock (pageIndex, blockToRead) : 
            replaceLRUBlock (pageIndex, blockToRead);
         currentBlockIndex++;
      }
      else
      {
         blockHit++;
         remove (block);
         addFirst (block);
      }
      if (callback != null)
         callback.blockAccess (block.index, dataSourceRelativeBlockBounds (block.index));
      return block;
   }

   private void dataSourceToSubvolumeAndBlock (
      IndexSpaceID dataSourceRelativeId, IndexSpaceID subvolumeRelativeId,
      IndexSpaceID blockRelativeId)
   {
      if (subvolumeRelativeId == null)
         subvolumeRelativeId = new IntegerIndexSpaceID (dataSourceRelativeId);
      else
         subvolumeRelativeId.set (dataSourceRelativeId);
      if (blockTableOptimization)
         subvolumeRelativeId.subtract (iterationSubvolume.getLower ());
      if (blockRelativeId == null)
         blockRelativeId = new IntegerIndexSpaceID (subvolumeRelativeId);
      else
         blockRelativeId.set (subvolumeRelativeId);
      blockRelativeId.mod (blockDimensions);
   }

   private Block createLRUBlock (int blockIndex, ISBounds blockToRead)
   {
      Block block = blocks[blockIndex] = new Block (
         dataSource.createDataBlock (new ISBounds (blockDimensions)),
         blockIndex);
      dataSource.subblock (block.data,
         new ISBounds (blockToRead.getDimensions ()), blockToRead);
      blockNumber++;
      if (DynamicChunking.logger.isLoggable (Level.FINE))
         DynamicChunking.logger.fine (
            "page number " + blockNumber +
            ": create for subblock=" + blockToRead);
      addFirst (block);
      memoryBlockCount++;
      if (memoryBlockCount == 1)
         DynamicChunking.logger.info ("Measured block memory: " + block.sizeOf ());
      if (DynamicChunking.logger.isLoggable (Level.FINEST) &&
         memoryBlockCount % 10 == 0)
      {
         /* used to find out the payload for a page */
         Runtime.getRuntime ().gc ();
         long newMemory = Runtime.getRuntime ().freeMemory ();
         DynamicChunking.logger.finest (
            "\nmemoryPageCount=" + memoryBlockCount +
            "\nfreeMemoryDifference=" + (freeMemory - newMemory));
         freeMemory = newMemory;         
      }
      return block;
   }

   private Block replaceLRUBlock (int pageIndex, ISBounds pageToRead)
   {
      Block pageToReplace = lruListHeader.previous;
      // store the page at the right index
      blocks[pageToReplace.index] = null;
      blocks[pageIndex] = pageToReplace;
      pageToReplace.index = pageIndex;
      remove (pageToReplace);
      addFirst (pageToReplace);
      dataSource.subblock (pageToReplace.data,
         new ISBounds (pageToRead.getDimensions ()), pageToRead);
      blockNumber++;
      if (DynamicChunking.logger.isLoggable (Level.FINE))
         DynamicChunking.logger.fine (
            "page number " + blockNumber +
            ": replace with subblock=" + pageToRead);
      blockReplace++;
      if (DynamicChunking.logger.isLoggable (Level.FINEST) &&
         blockReplace % 1000 == 0)
         DynamicChunking.logger.finest ("\npageReplace=" + blockReplace);
      return pageToReplace;
   }

   private void setupSizeOf ()
   {
      SizeOf.skipStaticField (true);
      SizeOf.skipFinalField (true);
   }
   
   public class CacheThrashingError extends Error
   {}
      
   private class BlockBoundsIterator implements Iterator<ISBounds>
   {

      BlockBoundsIterator ()
      {
         index = -1;
         incrementIndex ();
      }

      public boolean hasNext ()
      {
         return index < blocks.length;
      }

      public ISBounds next ()
      {
         ISBounds bounds = dataSourceRelativeBlockBounds (index);
         incrementIndex ();
         return bounds;
      }

      public void remove ()
      {
         throw new UnsupportedOperationException ();
      }

      private void incrementIndex ()
      {
         do
            index++;
         while (index < blocks.length && blocks[index] == null);
      }
      private int index;
   }
   
   private static long getFreeMemoryForDynamicChunking ()
   {
      final double DYNAMIC_CHUNKING_PERCENTAGE = 0.8;
      return (long)((Runtime.getRuntime ().freeMemory ()) * 
         DYNAMIC_CHUNKING_PERCENTAGE);
   }

   private static class SubblockIndexIterator implements Iterator<int[]>
   {

      SubblockIndexIterator (int[] length)
      {
         this.length = length;
         this.next = new int[length.length];
         this.cur = new int[length.length];
      }

      public boolean hasNext ()
      {
         return !carry;
      }

      public int[] next ()
      {
         System.arraycopy (next, 0, cur, 0, next.length);
         carry = true;
         for (int i = 0; i < next.length; i++)
            if (carry)
               if (next[i] < length[i] - 1)
               {
                  next[i]++;
                  carry = false;
                  break;
               }
               else
                  next[i] = 0;
         return cur;
      }

      public void remove ()
      {
         throw new UnsupportedOperationException ();
      }
      private int[] length;
      int[] next;
      int[] cur;
      boolean carry = false;
   }
      
   /**
    * Represents a block in a {@see DynamicChunking}.
    */
   static private class Block
   {
      Block (DataBlock dataBlock, int pageIndex)
      {
         this.data = dataBlock;
         this.index = pageIndex;
      }
      DataBlock data;
      int index;
      Block next;
      Block previous;
      
      long sizeOf ()
      {
         return SizeOf.sizeOf (this) + SizeOf.deepSizeOf (data);
      }      
   }
   
   private DataSource dataSource;
   private ISBounds iterationSubvolume;
   private boolean blockTableOptimization = false;
   /** The order in which you store the pages in the page table*/
   private AxisOrdering blockOrdering;
   private Block lruListHeader;
   private long memory = getFreeMemoryForDynamicChunking ();
   private IndexSpaceID blockDimensions;
   private IndexSpaceID partitionRatio;
   private int blockHit;
   private int blockReplace;
   private int maxMemoryBlockCount;
   private int memoryBlockCount;
   private int blockNumber;
   private Block[] blocks;
   private long[] axisWeightsForSubvolumeDimensionsInBlocks;
   private int currentBlockIndex = -1;
   /* used to experimentally determine BLOCK_PAYLOAD*/
   private long freeMemory;
   private BlockAccessCallback callback;
   static Logger logger =
      Logger.getLogger ("edu.unh.sdb.datasource.DynamicChunking");
   BlockSizeSliceOptimization blockSizeSliceOptimization;
   BlockSizeAdaptiveOptimization blockSizeAdaptiveOptimization;
   BlockSizeAdaptiveOptimization.TypeOfIteration typeOfIteration;
}
