/*
    ISBounds.java
    Copyright 2003 University of New Hampshire
    Durham, NH 03824
 
    This is the array implementation of ISBounds.
    ---------------------- Edit history as of 01/26/05 ------------------------
    09/20/07 dlipsa: replaced clone() with copy constructor or new/arraycopy
    12/22/05 dlipsa: Implemented longVolume using IndexSpaceID.volume
    10/21/05 rdb: Removed unused variables.
    10/02/05 rdb: Added code to "equalShape" method to return true if the
                  passed ISBounds and this are of different dimension, but
                  the higher dimension bounds has lengths of 1 along all
                  the extra axes. I think works, but it's likely that other
                  Granite code will have to change for this to be useful.
                  -Added try-catch blocks in 2 copyTo methods so better
                   error messages are generated.
    05/28/05 rdb: Remove deprecation on setLower and setUpper
    04/25/05 rdb: -Rewrote intersection( ISB ); improved performance by 400%
                  -added intersectWith( ISB ) as in place intersection (change this)
                   it is 4x as fast as the new intersection.
                  -deleted intersection( ISB, int[], int[] )
                  It was included because of much better performance than the
                  old intersection(ISB) for the null case. It was ugly as sin,
                  isn't needed now because the new version is just as good for
                  the null case -- and it wasn't used anywhere anyway.
    03/31/05 rdb: Made getLowerArray, getUpperArray, createAtOrigin public
                  Corrected copyAndMove and copyAndMoveTo code.
                  Added copyAndMoveCenterTo( isid | int[] )
                        moveCenterTo( isid | int[] )
                        createCentered( isid, isid )
                        createCentered( int[], int[] );
    03/30/05 pjr: Added move( int, int )
                  Added static createCentered( ISB, isid )
    03/20/05 rdb: Allowed getDimensions( int[] ) to have null argument.
    02/20/05 rdb: Restored getLower() and getUpper() as synonyms for
                           copyLower()    copyUpper().
    02/15/05 rdb: Completed the renaming of methods that I started earlier
         old                        new
        expand                    copyExpandBy
            increase( n )             copyIncreasedBy
            increaseByOne()           copyIncreasedBy( 1 )
            reduce( n )               copyReducedBy
            reduceByOne()             copyReducedBy( 1 )
            scale                     copyScaled
            shrink                    copyShrinkBy
 
            getLower                   copyLower
            getUpper                   copyUpper
            getLowerArray              copyLowerArray
            getUpperArray              copyUpperArray
 
          <package access>
            getLowerArrayReference     getLowerArray
            getUpperArrayReference     getUpperArray
 
 
    02/14/05 rdb: added makeRelativeTo( ISBounds ) update method.
    02/13/05 rdb: added constructor ISBounds( int, int[] )
    02/07/05 rdb: made ISBounds ( int[], int[] ) public and changed semantics so that
                      it clones arguments and doesn't selectively pick minimum
                      entry of each array.
                  I checked all uses and changed calls that cloned first.
                     Seemed to be random knowledge of clone vs non-clone in callers.
    02/03/05 rdb: Revised setUpper( int, int ) and setLower( int, int ) to
                        have index first then new value like all other sets in Granite
                  Added getLower( int ) and getUpper( int )
    01/31/05 rdb: Added getDimensionsArray()
                  Added getLowerArrayReference, made getLowerArray return copy
                  Added getUpperArrayReference, made getUpperArray return copy
    01/30/05 rdb: Added equalShape( ISBounds )
                  Removed error on dimensionality mismatch on equals and equalShape;
                     different dim are just not equal, but not error to compare
    01/26/05 rdb: Changed all methods that modify "this" to also return "this"
                      for convenience of use.
                  Added
                       set( ISBounds )                 to replace copy( ISBounds )
                       setAndMove( ISBounds, int[] )   to replace copyAndMove( ... )
                       setAndMoveTo( ISBounds, int[] ) to replace copyAndMoveTo( ... )
                       increase() as an in place update to replace increaseByOne(ISB)
                       decrease() as an in place replacement for decreaseByOne(ISB)
                  Deprecated copy( ISB ), copyAndMove, copyAndMoveTo
                  Deprecated increaseByOne( ISB ) and decreaseByOne( ISB )
 
                  Non-static
                     createAtOrigin();
                     createAt( IS );
                     createAt( int[] );
 
 
                  Deprecated
                    newBoundsAtOrigin( int[] ) -> createAtOrigin
                    newBoundsAtOrigin( IndexSpaceID ) -> createAtOrigin
                    copyAndMoveToOrigin -> createAtOrigin
 
                  There are still issues related to naming so that update and new
                     versions of the various methods can be more easily identified;
                     this is especially important now that I've made them all
                     return an ISBounds -- you can't tell from the return value
                     whether it is an update method or a "new" method.
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import java.io.*;

/**  The ISBounds class is meant to represent hypercubic subregions
 *   in an n-dimensional index space. Typically, a lower and upper corner
 *   point are given to the constructor,  although it is also possible
 *   to just specify the dimensions of the subregion, yielding an ISBounds
 *   of the proper size situated at the origin. The dimensionality of
 *   an ISBounds is taken from the dimensionality of the corner points,
 *   which are assumed to match.
 */
public class ISBounds  extends Bounds implements Cloneable, Serializable
{
   private static final long serialVersionUID = 0; // for serialization inheritance
   protected int [] lower,upper;
      
   /** This constructor takes two corner points (SpaceIDs)
    *  that define a hypercube of arbitrary dimensionality.
    *
    *  Notice that the getLower() and getUpper() methods may
    *  or may not return the same IndexSpaceID values given in this
    *  constructor.
    */
   public ISBounds ( IndexSpaceID corner1, IndexSpaceID corner2 )
   {
      int d = corner1.dim ();
      
      if ( d == corner2.dim ())
      {
         this.lower= new int[d];
         this.upper= new int[d];
         int[] c1 = corner1.refCoords ();
         int[] c2 = corner2.refCoords ();
         
         for ( int i=0; i < d; i++)
         {
            int c1Value = c1[i];
            int c2Value = c2[i];
            
            if ( c1Value <= c2Value)
            {
               lower[i] = c1Value;
               upper[i] = c2Value;               
            }
            else
            {
               lower[i] = c2Value;
               upper[i] = c1Value;
            }
         }
      }
      else
      {
         new DimensionalityError ( SDBError.Fatal,
            "Can't create ISBounds with two corners of different dimensionality.");
      }
   }
   
   
   
   /** This constructor takes a single IndexSpaceID
    * that defines the dimensions of a hypercube with
    *	lower coordinates at the origin. Notice
    * that the upper coordinates will each be one less
    * than the corresponding coordinate in the argument
    * IndexSpaceID.  This constructor takes the absolute
    * value of all coordinates in the argument.
    */
   public ISBounds ( IndexSpaceID dimensions )
   {
      int d=dimensions.dim ();
      
      this.lower = new int[d];
      this.upper = new int[d];
      int [] dimensionsArray=dimensions.refCoords ();
      //this.upper = dimensions.getCoords(new int[d]);
      
      for ( int i=0; i < this.lower.length; i++)
      {
         this.lower[i] = 0;
         this.upper[i] = Math.abs (dimensionsArray[i]) - 1;
      }
   }
   
   /** This constructor takes a single integer array
    *  that defines the dimensions of a hypercube with
    *  lower coordinates at the origin. Notice
    *  that the upper coordinates will each be one less
    *  than the corresponding coordinate in the argument
    *  array.  This constructor takes the absolute
    *  value of all coordinates in the argument. It is
    *  an error for any value in the dimensions array
    *  to be less than 1.
    */
   public ISBounds (int [] dimensions)
   {
      int d=dimensions.length;
      
      this.lower = new int[d];
      this.upper = new int[d];
      
      for ( int i=0; i < this.lower.length; i++)
      {
         this.lower[i]=0;
         if ( dimensions[i] <= 0)
         {
            new IllegalArgumentError ( SDBError.Recoverable,
               "Dimensions Array contains value less than 1 at position: "
               + i );
            dimensions[i] = 1;
         }
         this.upper[i]=Math.abs (dimensions[i]) - 1;
      }
   }
   
   /************************************/
   /**  This constructor takes two corner points expressed as int arrays
    * that define a hypercube of arbitrary dimensionality.
    */   
   public ISBounds ( int [] lower, int [] upper)
   {
      if ( lower.length != upper.length)
         new DimensionalityError ( SDBError.Recoverable,
            "Can't create ISBounds with two corners of different dimensionality.");
      else
      {
         this.lower = new int[lower.length];
         this.upper = new int[lower.length];
         System.arraycopy (lower, 0, this.lower, 0,   lower.length);
         System.arraycopy (upper, 0, this.upper, 0,   lower.length);
      }
   }

   
   /** Construct a copy of the argument bounds.
    * 09/20/07 new implementation that avoids clone();
    */
   public ISBounds ( ISBounds bounds )
   {      
      this (bounds.getLowerArray (), bounds.getUpperArray ());
   }
   
   /** This constructor just allocates space for the lower and upper arrays.*/
   protected ISBounds ( int dimensionality )
   {
      this.lower = new int[dimensionality];
      this.upper = new int[dimensionality];
   }
   
   /** This constructor allocates an ISBounds of the dimensionality
    *  defined by the first argument. The second argument is an array
    *  whose length is 2 times the dimensionality. The first dim entries
    *  are the lower bounds, the 2nd dim are the upper.
    */
   public ISBounds ( int dim, int[] bounds )
   {
      this.lower = new int[ dim ];
      this.upper = new int[ dim ];
      
      if ( bounds.length != 2*dim )
         new SDBError ( SDBError.Recoverable, "ISBounds( int, int[]) error: "
            + " array length must be exactly 2 * dimensionality.");
      else
         for ( int i = 0; i < dim; i++ )
         {
         this.lower[ i ] = bounds[ i ];
         this.upper[ i ] = bounds[ i + dim ];
         }
   }
   
   /** create a new ISBounds with lower bounds at the origin, and
    * upper bounds with the specified IndexSpaceID.
    * @deprecated Use createAtOrigin( isid )
    */
   public static ISBounds newBoundsAtOrigin ( IndexSpaceID id )
   {
      return new ISBounds ( new int[id.dim ()], (id.refCoords ().clone ()));
      
   }
   /** create a new ISBounds with lower bounds at the origin, and
    * upper bounds with the specified IndexSpaceID.
    */
   public static ISBounds createAtOrigin ( IndexSpaceID id )
   {
      // This implementation assumes that a freshly created array of ints
      // will be initialized to zero.
      return new ISBounds ( new int[id.dim ()], (id.refCoords ().clone ()));
   }
   
   /** Create a new ISBounds with lower bounds at the origin, and
    * upper bounds with coordinates corresponding
    * to the integer array argument.
    *
    * This method may elect to use the array argument directly,
    * without making a copy.
    *
    * This implementation assumes that a freshly created array of ints
    * will be initialized to zero.
    * @deprecated Use createAtOrigin( coords );
    */
   static ISBounds newBoundsAtOrigin (int [] coords)
   {
      return createAtOrigin ( coords );
   }
   
   /** Create a new ISBounds with lower bounds at the origin, and
    * upper bounds with coordinates corresponding
    * to the integer array argument.
    *
    * This method may elect to use the array argument directly,
    * without making a copy.
    *
    * This implementation assumes that a freshly created array of ints
    * will be initialized to zero.
    */
   public static ISBounds createAtOrigin (int [] coords)
   {
      return new ISBounds ( new int[coords.length], coords );
   }
   
   
   /** Create a new ISBounds with its center at the ISID specified
    * by the second argument and whose dimensions are the same as
    * the dimensions of the 1st argument.
    */
   
   public static ISBounds createCentered ( ISBounds shape, IndexSpaceID center )
   {
      return ISBounds.createCentered ( shape.getDimensions ().getCoords (),
         center.getCoords () );
   }
   
   /** Create a new ISBounds with its center at the ISID specified
    * by the second argument and whose dimensions are the same as
    * the dimensions of the 1st argument.
    */
   public static ISBounds createCentered ( IndexSpaceID shape, IndexSpaceID center )
   {
      return createCentered ( shape.getCoords (), center.getCoords () );
   }
   
   /** Create a new ISBounds with its center at the ISID specified
    * by the second argument and whose dimensions are given by the 1st argument.
    */
   public static ISBounds createCentered ( int[] shape, int[] center )
   {
      int      d = shape.length;
      ISBounds bounds = new ISBounds (d);
      
      for (int i=0; i<d; i++)
      {
         
         bounds.lower[i] = center[i] - shape[i] / 2;
         bounds.upper[i] = bounds.lower[i] + shape[i] - 1;
      }
      return bounds;
   }
   
   
   /** create a new ISBounds from this one with lower bounds at specified
    *  IndexSpace position and upper bounds defined by the size of this ISBounds.
    */
   final public ISBounds copyTo ( IndexSpaceID lower )
   {
      try
      {
         int upperArray[] = new int[ lower.dim () ];
         for ( int i = 0; i < lower.dim (); i++ )
            upperArray[i] = lower.getCoord (i) + this.upper[i] - this.lower[i];
         return new ISBounds ( lower.refCoords (), upperArray );
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Fatal,
            "ISBounds.copyTo( ISID ) failed on dimensionality mismatch");
         return null;
      }
   }
   
   /** create a new ISBounds from this one with lower bounds at specified
    *  position and upper bounds defined by the size of this ISBounds.
    */
   final public ISBounds copyTo ( int[] lowerArray )
   {
      try
      {
         int upperArray[] = new int[ lowerArray.length ];
         for ( int i = 0; i < lowerArray.length; i++ )
            upperArray[i] = lowerArray[i] + this.upper[i] - this.lower[i];
         return new ISBounds ( lowerArray, upperArray );
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Fatal,
            "ISBounds.copyTo( ISID ) failed on dimensionality mismatch");
         return null;
      }
   }
   
   /** create a new ISBounds from this one with lower bounds at the origin, and
    * upper bounds defined by the size of this ISBounds.
    *
    */
   final public ISBounds copyToOrigin ()
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.upper[i] = this.upper[i] - this.lower[i];
         result.lower[i]=0;
      }
      return result;
   }
   
   /** Return a copy of this ISBounds object
    */
   public final ISBounds copy ()
   {
      return new ISBounds ( this );
   }
   
   /** Return a copy of this ISBounds object
    */
   public Object clone ()
   {
      return new ISBounds ( this.lower.clone (), this.upper.clone ());
   }
   
   /** Set the lower and upper bounds of this object to refer
    * to the lower and upper bounds of the parameter. No
    * copy of the bounds are made.
    * "this" is returned for convenience of use.
    */
   final ISBounds assign (ISBounds b)
   {
      this.lower=b.lower;
      this.upper=b.upper;
      return this;
   }
   
   /** Set the lower and upper bounds of this object to refer
    * to the argument arrays. No copy of the arrays are made.
    *
    * In the interest of speed, no dimensionality checking is done,
    * so make sure that the argument arrays have the same length.
    *
    *      "this" is returned for convenience of use.
    * Package access only.
    */
   final ISBounds assign (int [] lower, int[] upper)
   {
      this.lower=lower;
      this.upper=upper;
      return this;
   }
   
   /** Set the lower bound of this object to refer
    * to the parameter. If the IndexSpaceID contains
    * an integer array, the ISBounds may
    * refer directly to that array.
    *
    * In the interest of speed, no dimensionality checking is done,
    * so make sure that the argument has a length matching
    * the dimensionality of the upper bounds.
    *
    *      "this" is returned for convenience of use.
    * Package access only.
    */
   final ISBounds assignLower ( IndexSpaceID lower )
   {
      this.lower=lower.refCoords ();
      return this;
   }
   
   /** Set the upper bound of this object to refer
    * to the parameter. If the IndexSpaceID contains
    * an integer array, the ISBounds may
    * refer directly to that array.
    *
    * In the interest of speed, no dimensionality checking is done,
    * so make sure that the argument has a length matching
    * the dimensionality of the lower bounds.
    *
    *      "this" is returned for convenience of use.
    * Package access only.
    */
   final ISBounds assignUpper (IndexSpaceID upper)
   {
      this.upper = upper.refCoords ();
      return this;
   }
   
   /** Set the lower bound of this object to refer
    * to the parameter. No copy of the array arguments are made.
    *
    * In the interest of speed, no dimensionality checking is done,
    * so make sure that the argument array has a length matching
    * the dimensionality of the upper bounds.
    *
    *      "this" is returned for convenience of use.
    * Package access only.
    */
   final ISBounds assignLower (int [] lower)
   {
      this.lower=lower;
      return this;
   }
   
   /** Set the upper bound of this object to refer
    * to the parameter. No copy of the array arguments are made.
    *
    * In the interest of speed, no dimensionality checking is done,
    * so make sure that the argument array has a length matching
    * the dimensionality of the lower bounds.
    *
    *      "this" is returned for convenience of use.
    * Package access only.
    *
    */
   final ISBounds assignUpper ( int [] upper )
   {
      this.upper=upper;
      return this;
   }
   
   /** Copy the upper and lower bounds of the argument ISBounds into this object.
    *  No new memory is allocated. Dimensionalities should match.
    *      "this" is returned for convenience of use.
    * @deprecated
    *     Use set( ISBounds )
    */
   public final ISBounds copy ( ISBounds b )
   {
      return set ( b );
   }
   
   /** Copy the upper and lower bounds of the argument ISBounds into this object.
    *  No new memory is allocated. Dimensionalities should match.
    *      "this" is returned for convenience of use.
    */
   public final ISBounds set ( ISBounds b )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            this.lower[i]=b.lower[i];
            this.upper[i]=b.upper[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Fatal,
            "Argument Bounds has too few dimensions.");
      }
      return this;
   }
   
   /** Set the bounds of this object so it is the same shape as the ISBounds
    *  argument, but translated to the location represented by the array argument.
    *  No new memory is allocated. Dimensionalities should match.
    *  This is an *absolute* move.
    *      "this" is returned for convenience of use.
    * @deprecated
    *    Use setAndMoveTo( ISBounds, int[] )
    */
   public final ISBounds copyAndMoveTo ( ISBounds b, int [] destination )
   {
      return setAndMoveTo ( b, destination );
   }
   
   /** Set the bounds of this object so it is the same shape as the ISBounds
    *  argument, but translated to the location represented by the array argument.
    *  No new memory is allocated. Dimensionalities should match.
    *  This is an *absolute* move.
    *      "this" is returned for convenience of use.
    */
   public final ISBounds setAndMoveTo ( ISBounds b, int [] destination )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            this.lower[i] = destination[i];
            this.upper[i] = destination[i] + (b.upper[i] - b.lower[i]) ;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Fatal,
            "Argument array has too few dimensions.");
      }
      return this;
   }
   
   /** Set the bounds of this object so it is the same shape as the ISBounds
    *  argument, but translated to the location represented by the array argument.
    *  No new memory is allocated. Dimensionalities should match.
    *  This is a *relative* move.
    *      "this" is returned for convenience of use.
    * @deprecated
    * Use setAndMove( ISBounds, int[] )
    */
   public final ISBounds copyAndMove ( ISBounds b, int [] distance )
   {
      return setAndMove ( b, distance );
   }
   
   /** Create a copy of this bounds moved by the relative amount specified
    *  by the argument.
    */
   public final ISBounds copyAndMove ( int [] distance )
   {
      return this.copy ().move ( distance );
   }
   
   /** Create a copy of this bounds moved the lower left corner to
    * the specified location.
    */
   public final ISBounds copyAndMoveTo ( int [] loc )
   {
      return this.copy ().moveTo ( loc );
   }
   
   /** Create a copy of this bounds moved the center to
    * the specified location.
    */
   public final ISBounds copyAndMoveCenterTo ( int [] loc )
   {
      return this.copy ().moveCenterTo ( loc );
   }
   
   /** Create a copy of this bounds moved the center to
    * the specified location.
    */
   public final ISBounds copyAndMoveCenterTo ( IndexSpaceID loc )
   {
      return this.copy ().moveCenterTo ( loc.getCoords () );
   }
   
   /** Set the bounds of this object so it is the same shape as the ISBounds
    *  argument, but translated to the location represented by the array argument.
    *  No new memory is allocated. Dimensionalities should match.
    *  This is a *relative* move.
    *      "this" is returned for convenience of use.
    */
   public final ISBounds setAndMove ( ISBounds b, int [] distance )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            this.lower[i] = b.lower[i] + distance[i];
            this.upper[i] = b.upper[i] + distance[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Fatal,
            "ISBounds or Argument array has too few dimensions.");
      }
      return this;
   }
   
   /** Copy the contents of the argument array into the lower coordinate for this bounds.
    *  No new memory is allocated.
    * "this" is returned for convenience of use.
    * @deprecated Use setLower( int [] )
    */
   public final ISBounds copyLower ( int [] lower )
   {
      return setLower ( lower );
   }
   
   /** Copy the contents of the argument array into the upper coordinate for this bounds.
    *  No new memory is allocated.
    * @deprecated Use setUpper( int [] )
    * "this" is returned for convenience of use.
    */
   public final ISBounds copyUpper (int [] upper)
   {
      return setUpper ( upper );
   }
   
   /** Set the lower bounds to the coordinates represented
    * by the IndexSpaceID argument.
    * "this" is returned for convenience of use.
    */
   public final ISBounds setLower ( IndexSpaceID lower )
   {
      if ( this.lower.length == lower.dim ())
         this.lower= lower.refCoords ().clone();
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** Set the upper bounds to the coordinates represented
    *  by the IndexSpaceID argument.
    * "this" is returned for convenience of use.
    */
   public final ISBounds setUpper ( IndexSpaceID upper )
   {
      if ( this.lower.length == upper.dim () )
         this.upper=upper.refCoords ().clone ();
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** Set the lower bounds to the coordinates represented
    *  by the integer array argument.
    * "this" is returned for convenience of use.
    */
   public final ISBounds setLower ( int [] lower )
   {
      if ( this.lower.length == lower.length)
         this.lower= lower.clone ();
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** Set the upper bounds to the coordinates represented
    *  by the integer array argument.
    * "this" is returned for convenience of use.
    */
   public final ISBounds setUpper ( int [] upper )
   {
      if ( this.lower.length == upper.length )
         this.upper=upper.clone ();
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** Set the upper bounds of a given axis specified in the argument
    * "this" is returned for convenience of use.
    */
   public final ISBounds setUpper ( int axis, int newValue )
   {
      if ( axis <= this.upper.length)
         this.upper[axis] = newValue;
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** set the lower bounds of a given axis specified in the argument
    * "this" is returned for convenience of use.
    */
   public final ISBounds setLower ( int axis, int newValue )
   {
      if ( axis <= this.lower.length )
         this.lower[axis] = newValue;
      else
         new DimensionalityError ( SDBError.Recoverable,
            "Argument has wrong dimensionality.");
      return this;
   }
   
   /** Return an IndexSpaceID representing the lower bounds.
    *
    */
   public final IntegerIndexSpaceID copyLower ()
   {
      return new IntegerIndexSpaceID ( this.lower );
   }
   
   /** Return an IndexSpaceID representing the lower bounds.
    *
    */
   public final IntegerIndexSpaceID getLower ()
   {
      return new IntegerIndexSpaceID ( this.lower );
   }
   
   /** Return an int representing the lower bound of the specified dimension.
    */
   public final int getLower ( int axis )
   {
      try
      {
         return this.lower[ axis ];
      }
      catch ( IndexOutOfBoundsException ex )
      {
         new SDBError ( SDBError.Recoverable, "axis argument invalid.");
         return 0;
      }
      
   }
   
   /** Return an int representing the upper bound of the specified dimension.*/
   public final int getUpper ( int axis )
   {
      try
      {
         return this.upper[ axis ];
      }
      catch ( IndexOutOfBoundsException ex )
      {
         new SDBError ( SDBError.Recoverable, "axis argument invalid.");
         return 0;
      }
   }
   
   /** Return an IndexSpaceID representing the upper bounds.
    *
    */
   public final IntegerIndexSpaceID copyUpper ()
   {
      return  new IntegerIndexSpaceID ( this.upper );
   }
   
   /** Return an IndexSpaceID representing the upper bounds.
    *
    */
   public final IntegerIndexSpaceID getUpper ()
   {
      return  new IntegerIndexSpaceID ( this.upper );
   }
   
   /** Return a reference to the integer array
    *  representing the upper bounds.
    */
   public final int [] getLowerArray ()
   {
      return this.lower;
   }
   
   /** Return a reference to the integer array
    *  representing the lower bounds.
    */
   public final int [] copyLowerArray ()
   {
      return this.lower.clone();
   }
   
   /** Return a copy of the integer array
    *  representing the upper bounds.
    *
    * public final int [] copyLowerArray()
    * {
    * return (int[])this.lower.clone();
    * }
    *
    * /** Copy the lower bound coordinates into the
    *	array argument. A reference to the argument
    * is returned.
    * @deprecated Use copyLowerArray( int [] )
    */
   public final int [] getLowerArray ( int [] arr )
   {
      return copyLowerArray ( arr );
   }
   
   /** Copy the lower bound coordinates into the
    *	array argument. A reference to the argument
    * is returned.
    */
   public final int [] copyLowerArray ( int [] arr )
   {
      for ( int i=0; i < this.lower.length; i++)
         arr[i] = this.lower[i];
      return arr;
   }
   
   /**  Return a reference to the integer array
    *  representing the upper bounds. Package access.
    */
   public final int [] getUpperArray ()
   {
      return  this.upper;
   }
   
   /** Return a copy of the integer array
    *  representing the upper bounds.
    */
   public final int [] copyUpperArray ()
   {
      return this.upper.clone();
   }
   
   /** Copy the upper bound coordinates into the
    * array argument. A reference to the argument
    * is returned.
    * @deprecated Use copyUpperArray( int [] )
    */
   public final int [] getUpperArray ( int [] arr )
   {
      return copyUpperArray ( arr );
   }
   
   /** Copy the upper bound coordinates into the
    * array argument. A reference to the argument
    * is returned.
    */
   public final int [] copyUpperArray ( int [] arr )
   {
      for ( int i=0; i < this.upper.length; i++)
         arr[i] = this.upper[i];
      return arr;
   }
   
   /** Return dimensionality of Bounds. */
   public final int dim ()
   {
      return this.lower.length;
   }
   
   /** Return true if parameter is equal to this object, false otherwise.
    *  In this case equality means that all bounds parameters are the same.
    */
   public boolean equals ( ISBounds b )
   {
      try
      {
         for ( int i=0; i < this.dim (); i++)
            if (   this.lower[i] != b.lower[i]
            || this.upper[i] != b.upper[i] )
               return false;
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         //rdb new DimensionalityError( SDBError.Recoverable,
         //"Dimensionality of Bounds argument does not match this Bounds object.");
         return false;
      }
      return true;
   }
   
   /** Return true if parameter has the same size in all dimensions as
    *  this object, false otherwise.
    * If the dimensionalities do not match, we'll try to see if shapes
    * still match because the higher dimensional bounds has length 1 in
    * all the "extra" dimensions, and the others match.
    */
   public boolean equalShape ( ISBounds b )
   {
      if ( this.dim () != b.dim () )
         return equalShapeAfterProject ( b );
      else
      {
         try
         {
            for ( int i=0; i < this.dim (); i++)
               if ( this.upper[i] - this.lower[i] != b.upper[i] - b.lower[i] )
                  return false;
         }
         catch ( ArrayIndexOutOfBoundsException e)
         {
            //rdb new DimensionalityError( SDBError.Recoverable,
            //"Dimensionality of Bounds argument does not match this Bounds object.");
            return false;
         }
         return true;
      }
   }
   
   /**
    * Returns true if the passed bounds has the same shape as this
    * bounds after the higher dimensioned bounds is projected along
    * any dimension that is of length 1. Of course, we have to project
    * the argument along any axis of length 1 as well.
    * @param b
    * @return
    */
   protected boolean equalShapeAfterProject ( ISBounds b )
   {
      // We are going to try to avoid actually doing the project operation.
      // As we are looping through the higher dimension bounds, we will
      // ignore any axes of length 1.
      ISBounds highDimBnds = this; // guess that this is > b
      ISBounds lowDimBnds =  b;
      if ( b.dim () < this.dim () )  // correct it if guess was wrong
      {
         highDimBnds = b;
         lowDimBnds = this;
      }
      int[] highDims = highDimBnds.getDimensionsArray ();
      int[] lowDims = lowDimBnds.getDimensionsArray ();
      
      boolean equal = true;
      int lo = 0;  // index of lower dimension bounds
      int hi = 0;  // index for higher dim bounds
      for (  hi = 0;
      equal && hi < highDims.length && lo < lowDims.length;
      hi++ )
      {
         if ( highDims[ hi ] == lowDims[ lo ] )
            lo++;
         else if ( highDims[ hi ] != 1 )
            equal = false;
      }
      for ( ; hi < highDims.length; hi++ ) // extra high dim axes must be 1
         equal &= highDims[ hi ] == 1;
      
      return equal && lo >= lowDims.length;
   }
   
   /** Return the total volume of the ISBounds as an integer.
    * "Volume" is defined to be the total number of elements
    * within the dataspace enclosed by the ISBounds.
    * If the volume does not fit into an int, an exception
    * is thrown.
    */
   public final int volume ()
   {
      long vol = longVolume ();
      int  ivol = (int) vol;
      if ( ivol != vol )
         new SDBError ( SDBError.AlwaysThrow,
            "DataCollection volume larger than int." );
      return ivol;
   }
   
   /** Return the total volume of the ISBounds as a long.
    * "Volume" is defined to be the total number of elements
    * within the dataspace enclosed by the ISBounds.
    */
   public final long longVolume ()
   {
      long volume = 1;
      for (int i = 0; i < lower.length; i++)
         volume *= (upper[i] - lower[i] + 1);
      return volume;
   }
   
   /** Return an IndexSpaceID with axis values denoting the size
    * of each corresponding dimension of the ISBounds.
    */
   public final IndexSpaceID getDimensions ()
   {
      return new IntegerIndexSpaceID (
         this.getDimensions ( new int[ this.lower.length ]));
   }
   
   /**
    * @deprecated Use copyIncreasedBy( int )
    */
   public final ISBounds increase ( int n )
   {
      return copyIncreasedBy ( n );
   }
   /** Return a new ISBounds that has upper coordinates that are increased
    * by the specified amount from the upper coordinates of this ISBounds.
    *
    * A convenience function.
    */
   public final ISBounds copyIncreasedBy ( int n )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.lower[i] = this.lower[i];
         result.upper[i] = this.upper[i] + n;
      }
      return result;
   }
   
   /** Return an int array with axis values denoting the size
    * of each corresponding dimension of the ISBounds.
    */
   public final int[] getDimensionsArray ()
   {
      return this.getDimensions ( new int[ this.lower.length ]);
   }
   
   /** Fill the argument array with coordinate values denoting the size
    * of each corresponding dimension of the ISBounds. A reference to the argument
    * array is returned as a convenience.
    *
    * A null argument is legal and results in the allocation of a correct
    *  length array. If the argument is not null, it should have a length
    *  corresponding to the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short,  a DimensionalityError will be created.
    */
   public final int [] getDimensions ( int[] dimensions )
   {
      if ( dimensions == null )
         dimensions = new int[ this.lower.length ];
      try
      {
         for ( int i=0; i < this.lower.length; i++ )
            dimensions[i] = this.upper[i] - this.lower[i] + 1;
      }
      catch ( ArrayIndexOutOfBoundsException e )
      {
         new DimensionalityError ( SDBError.Fatal,
            "Argument array has too few dimensions.");
      }
      
      return dimensions;
   }
   
   /** Return the length of the axis specified as an argument. */
   public final int getDimension ( int axis )
   {
      try
      {
         return this.upper[axis] - this.lower[axis] + 1;
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Axis argument is too large.");
         return -1;
      }
   }
   
   /** Returns a new ISBounds which is the projection of this ISBounds onto a
    * plane orthogonal to the specified axis. The dimensionality of the result
    * will therefore always be one less than the dimensionality of this object,
    * as long as the axis is valid for this ISBounds.
    *
    * If the specified axis is not valid, an error is created.
    */
   public ISBounds project ( int axis )
   {
      int d = this.lower.length;
      
      if ( axis >= d)
      {
         new IllegalArgumentError ( SDBError.Recoverable,
            "axis of projection is beyond the dimensionality of this ISBounds.");
         return this;
      }
      
      int [] newLower=new int[d-1];
      int [] newUpper=new int[d-1];
      
      int j=0;
      
      for ( int i=0; i < d; i++)
      {
         if ( i != axis)
         {
            newLower[j] = this.lower[i];
            newUpper[j] = this.upper[i];
            j++;
         }
      }
      return new ISBounds ( newLower, newUpper );
   }
   
   /** Returns a new ISBounds which is a slice of this ISBounds of thickness 1 in a plane orthogonal
    *	to the specified axis, having the given coordinate value for that axis. The
    * dimensionality of the result matches the dimensionality of this ISBounds, as long
    * as the axis is valid for this ISBounds.
    * If the specified axis is not valid, then an error is created.
    */
   public ISBounds slice ( int axis, int coordinate )
   {
      int d = this.lower.length;
      
      if ( axis >= d)
      {
         new IllegalArgumentError ( SDBError.Recoverable,
            "axis of slice is beyond the dimensionality of this ISBounds.");
         return this;
      }
      int [] newLower=new int[d];
      int [] newUpper=new int[d];
      
      for ( int i=0; i < this.lower.length; i++)
      {
         if ( i != axis)
         {
            newLower[i]=this.lower[i];
            newUpper[i]=this.upper[i];
         }
         else
         {
            newLower[i]=coordinate;
            newUpper[i]=coordinate;
         }
      }
      return new ISBounds ( newLower, newUpper );
   }
   
   /** Returns true if the argument IndexSpaceId represents the dimensions of
    *  a slice of this ISBounds of arbitrary thickness in the specified axis.
    */
   public boolean isSlice ( IndexSpaceID dimensionsID, int axis )
   {
      int d = this.lower.length;
      
      if ( axis >= d )
      {
         new IllegalArgumentError ( SDBError.Recoverable,
            "slice axis is beyond the dimensionality of this ISBounds.");
         return false;
      }
      int [] dimensions = dimensionsID.refCoords ();
      
      for ( int i=0; i < this.lower.length; i++)
      {
         if ( i != axis)
         {
            if ( dimensions[i] != this.upper[i] - this.lower[i] + 1 )
               return false;
         }
         else
         {
            if ( dimensions[i] < 1
               || (dimensions[i] > this.upper[i] - this.lower[i] + 1))
               return false;
         }
      }
      return true;
   }
   
   /** The two argument bounds are set to represent two portions of this ISBounds
    * split by a plane orthogonal to the splitAxis, at the coordinate value
    * contained in splitAxisCoord. This coordinate value must be equal to or
    * greater than the lower bound of this object in the split axis, and less
    * than the upper bound. Otherwise, an error is created.
    */
   public void split ( ISBounds lowerBounds, ISBounds upperBounds,
      int splitAxis, int splitAxisCoord)
   {
      int [] LBlower = lowerBounds.lower;
      int [] LBupper = lowerBounds.upper;
      
      int [] UBlower = upperBounds.lower;
      int [] UBupper = upperBounds.upper;
      
      try
      {
         for ( int i=0; i < this.lower.length; i++ )
         {
            LBlower[i] = this.lower[i];
            UBupper[i] = this.upper[i];
            
            if ( i == splitAxis)
            {
               if ( splitAxisCoord >= this.lower[i])
                  LBupper[i] = splitAxisCoord;
               else
               {
                  new IllegalArgumentError ( SDBError.Recoverable,
                     "SplitAxisCoord Argument is too small for this bounds.");
                  return;
               }
               if ( splitAxisCoord < this.upper[i])
                  UBlower[i] = splitAxisCoord + 1;
               else
               {
                  new IllegalArgumentError ( SDBError.Recoverable,
                     "SplitAxisCoord Argument is too large for this bounds.");
                  return;
               }
            }
            else
            {
               LBupper[i] = this.upper[i];
               UBlower[i] = this.lower[i];
            }
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Bounds dimensionalities do not match.");
         return;
      }
   }
   
   /** Fills the second argument with the coordinates of the specified
    *  vertex. The vertices are stored in a "bitwise" order. That is,
    *  vertex 0 consists of all lower coords. Vertex 1 is similar, except
    *  that it uses the upper coordinate for axis 0. In summary, the
    *  bits in the vertex number determine whether the lower or upper
    *  coordinate will be chosen for the vertex location.
    *
    *  If the coords array is null, a new array will be created.
    *
    *  A reference to the coords array is returned.
    */
   public int [] getVertex (int vertexNumber, int [] coords)
   {
      
      int d = this.lower.length;
      int mask = 0x01;
      
      if(coords == null)
      {
         
         coords = new int[d];
      }
      
      for(int i=0; i<d ; i++)
      {
         
         coords[i] = ( (vertexNumber & mask) !=0 ) ? this.upper[i] : this.lower[i];
         
         mask <<= 1;
      }
      
      return coords;
   }
   
   
   /** Fills the second argument with the coordinates of the specified
    *  vertex. The vertices are stored in a "bitwise" order. That is,
    *  vertex 0 consists of all lower coords. Vertex 1 is similar, except
    *  that it uses the upper coordinate for axis 0. In summary, the
    *  bits in the vertex number determine whether the lower or upper
    *  coordinate will be chosen for the vertex location.
    *
    *  If the IndexSpaceID is null, a new IntegerIndexSpaceID will be created.
    *
    *  A reference to the coords array is returned.
    */
   public IndexSpaceID getVertex (int vertexNumber, IndexSpaceID id)
   {
      
      int d = this.lower.length;
      int mask = 0x01;
      
      if(id == null)
      {
         
         id = new IntegerIndexSpaceID (new int[d]);
      }
      
      int [] coords = id.refCoords ();
      
      for(int i=0; i<d ; i++)
      {
         
         coords[i] = ( (vertexNumber & mask) !=0 ) ? this.upper[i] : this.lower[i];
         
         mask <<= 1;
      }
      
      id.assign (coords);
      
      return id;
   }
   
   
   /** Returns true if the argument bounds has the same shape as this ISBounds, and false otherwise.
    * No explicit check of dimensionality is done, and shape checking is only done up to and
    * including the highest dimension of this bounds. However, if the argument bounds is of
    * a lower dimensionality than this ISBounds, a DimensionalityError will be created. */
   public final boolean hasSameShape ( ISBounds b )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
            if ( (this.upper[i] - this.lower[i]) != (b.upper[i] - b.lower[i]) )
               return false;
         return true;
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Argument array has too few dimensions.");
         return false;
      }
   }
   
   /**
    * @deprecated Use copyExpandBy( int )
    */
   public final ISBounds expand ( int n )
   {
      return copyExpandBy ( n );
   }
   
   /** Modifies this ISBounds with upper and lower coordinates that have been
    * increased by a given multiple.  The size of the ISBounds is multiplied by
    * the factor, but the "center of gravity" remains unchanged.
    */
   public final void expandByFactor(int factor)
   {
       for(int i = 0; i < this.lower.length; i++)
       {
           this.upper[i] = ((this.upper[i] + 1) * factor) - 1;
           this.lower[i] = this.lower[i] * factor;
       }
   }
   
   /** Return a new ISBounds that has upper and lower coordinates that have been
    * expanded by a given multiple.  The size of the ISBounds is multiplied by
    * the factor, but the "center of gravity" remains unchanged.
    */
   public final ISBounds copyExpandByFactor(int factor)
   {
       ISBounds result = new ISBounds(this.lower.length);
       
       for(int i = 0; i < this.lower.length; i++)
       {
           result.upper[i] = ((this.upper[i] + 1) * factor) - 1;
           result.lower[i] = this.lower[i] * factor;
       }
       
       return result;
   }
   
   /** Return a new ISBounds that has upper and lower coordinates that are
    * moved outward by the specified distance. That is, the lower coordinates
    * are decreased, while the upper coordinates are increased. The "center of
    * gravity" is not changed.
    */
   public final ISBounds copyExpandBy ( int n )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.upper[i] = this.upper[i] + n;
         result.lower[i] = this.lower[i] - n;
      }
      return result;
   }
   
   /** Return a new ISBounds that has upper coordinates that are exactly
    * one less than the upper coordinates of this ISBounds.
    *
    * @deprecated Use copyReducedBy( 1 )
    */
   public final ISBounds reduceByOne ()
   {
      return copyReducedBy ( 1 );
   }
   
   /** Copies coordinates into the argument ISBounds such that
    * the lower coordinates are unchanged, but the upper coordinates
    * are reduced by one.
    *
    * A convenience function.
    *
    * @deprecated
    * rdb: I find the semantics very confusing. One would expect
    *  that the this object would be the one changed, not the argument object.
    *  Especially since our other similar methods either change "this" or
    *  return a new object (except decreaseByOne(ISB)).
    */
   public final void reduceByOne ( ISBounds b )
   {
      for ( int i=0; i< this.lower.length; i++)
      {
         b.lower[i] = this.lower[i];
         b.upper[i] = this.upper[i] - 1;
      }
   }
   
   /**
    * @deprecated Use copyReducedBy( int )
    */
   public final ISBounds reduce ( int n )
   {
      return copyReducedBy ( n );
   }
   
   /** Return a new ISBounds that has upper coordinates that are reduced
    * by the specified amount from the upper coordinates of this ISBounds.
    *
    * A convenience function.
    */
   public final ISBounds copyReducedBy ( int n )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.lower[i] = this.lower[i];
         result.upper[i] = this.upper[i] - n;
      }
      return result;
   }
   
   /** Modifies this ISBounds with upper and lower coordinates that have been
    * reduced by a given multiple.  The size of the ISBounds is divided by
    * the factor, but the "center of gravity" remains unchanged.
    */
   public final void shrinkByFactor(int factor)
   {
       for(int i = 0; i < this.lower.length; i++)
       {
           this.upper[i] = this.upper[i] / factor;
           this.lower[i] = this.lower[i] / factor;
       }
   }
   
      /** Return a new ISBounds that has upper and lower coordinates that have been
    * reduced by a given multiple.  The size of the ISBounds is divided by
    * the factor, but the "center of gravity" remains unchanged.
    */
   public final ISBounds copyShrinkByFactor(int factor)
   {
       ISBounds result = new ISBounds(this.lower.length);
       
       for(int i = 0; i < this.lower.length; i++)
       {
           result.upper[i] = this.upper[i] / factor;
           result.lower[i] = this.lower[i] / factor;
       }
       
       return result;
   }
   
   /** Return a new ISBounds that has upper and lower coordinates that are
    * moved inward by the specified distance. That is, the lower coordinates
    * are increased, while the upper coordinates are reduced. The "center of
    * gravity" is not changed.
    */
   public final ISBounds copyShrinkBy ( int n )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.upper[i] = this.upper[i] - n;
         result.lower[i] = this.lower[i] + n;
      }
      return result;
   }
   
   /** Return a new ISBounds that has upper coordinates that are exactly
    * one greater than the upper coordinates of this ISBounds.
    *
    *	@deprecated Use copyIncreasedBy( 1 )
    */
   public final ISBounds increaseByOne ()
   {
      return copyIncreasedBy ( 1 );
   }
   
   /** Copies coordinates into the argument ISBounds such that
    * the lower coordinates are unchanged, but the upper coordinates
    * are increased by one.
    *
    * A convenience function.
    * @deprecated
    * rdb: I find the semantics very confusing. One would expect
    *  that the this object would be the one changed, not the argument object.
    *  Why do we need this? It is exactly the same as b.increaseByOne();
    *  All other similar methods either change "this" or
    *  return a new object (except increaseByOne(ISB)).
    */
   public final void increaseByOne ( ISBounds b )
   {
      for ( int i=0; i < this.lower.length; i++)
      {
         b.lower[i] = this.lower[i];
         b.upper[i] = this.upper[i] + 1;
      }
   }
   
   /** Return a new ISBounds with dimensions that have been scaled by the
    * the argument array. The lower bounds are not affected, but each coordinate
    * of the upper bounds is modified to reflect the scaling.
    */
   public final ISBounds copyScaled ( double[] scales )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.lower[i] = this.lower[i];
         result.upper[i] = (int)( this.lower[i] + (this.upper[i] - this.lower[i]) * scales[i] + 0.5);
      }
      return result;
   }
   
   /** Return a new ISBounds with dimensions that have been scaled by the
    * the argument array. The lower bounds are not affected, but each coordinate
    * of the upper bounds is modified to reflect the scaling.
    */
   public final ISBounds copyScaled ( float[] scales )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.lower[i] = this.lower[i];
         result.upper[i] = (int)( this.lower[i] + (this.upper[i] - this.lower[i]) * scales[i] + 0.5);
      }
      return result;
   }
   
   /** Return a new ISBounds with dimensions that have been scaled by the
    * the argument array. The lower bounds are not affected, but each coordinate
    * of the upper bounds is modified to reflect the scaling.
    */
   public final ISBounds copyScaled ( int[] scales )
   {
      ISBounds result = new ISBounds ( this.lower.length );
      
      for ( int i=0; i < this.lower.length; i++)
      {
         result.lower[i] = this.lower[i];
         result.upper[i] = this.lower[i] + (this.upper[i] - this.lower[i]) * scales[i];
      }
      return result;
   }
   
   /** Translate the ISBounds object
    * in space by the value of the parameter IndexSpaceID. Each
    * coordinate of the IndexSpaceID is added to the coordinates of the
    * upper and lower bounds of this ISBounds.
    *
    *      "this" is returned for convenience of use.
    * This is a *relative* move.
    */
   public final ISBounds move ( IndexSpaceID distance )
   {
      return this.move ( distance.refCoords () );
   }
   
   
   /** Translate the ISBounds object
    * in space by the coordinates in the integer array argument. Each
    * coordinate is added to the coordinates of the
    * upper and lower bounds of this ISBounds.
    *
    *      "this" is returned for convenience of use.
    * This is a *relative* move.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short,  a DimensionalityError will will be created.
    */
   public final ISBounds move ( int [] distance )
   {
      try
      {
         for ( int i=0; i < this.lower.length ; i++)
         {
            this.lower[i]+=distance[i];
            this.upper[i]+=distance[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Argument array has too few dimensions.");
      }
      return this;
   }
   
   /** Translate the ISBounds object
    * in space along the specified axis by the specified distance.
    * The distance argument is added to the coordinates of the
    * upper and lower bounds of this ISBounds.
    *
    * If the specified axis is not less than the dimensionality
    * of this ISBounds, a dimensionality error is created.
    *
    * "this" is returned for convenience of use.
    * This is a *relative* move.
    *
    */
   public final ISBounds move ( int axis, int distance )
   {
      try
      {
         this.lower[axis]+=distance;
         this.upper[axis]+=distance;
         
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Axis value is too large.");
      }
      return this;
   }
   
   
   
   /** Translate the ISBounds object in space so that the
    * lower left corner is moved to the location specified
    * by the parameter IndexSpaceID.
    *
    * This is an *absolute* move.
    *
    * The IndexSpaceID should have the same dimensionality as the bounds
    * object. However, if the argument array is too long, the extra
    * elements will be ignored. If it is too short,
    * a DimensionalityError will will be created.
    *
    *      "this" is returned for convenience of use.
    */
   public final ISBounds moveTo ( IndexSpaceID location )
   {
      return moveTo ( location.refCoords () );
   }
   
   /** Translate the ISBounds object in space so that the
    * lower left corner is moved to the coordinates specified
    * int the integer array argument.
    *
    * This is an *absolute* move.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short, a DimensionalityError will be created.
    *
    * For convenience the object itself is returned as the result of the method.
    *
    */
   public final ISBounds moveTo ( int[] location )
   {
      try
      {
         for ( int i=0; i< this.lower.length; i++)
         {
            this.upper[i] += (location[i] - lower[i]);
            this.lower[i] = location[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable, "Argument array has too few dimensions.");
      }
      return this;
   }
   
   /** Translate the ISBounds object in space so that the
    * center is moved to the coordinates specified
    * in the integer array argument.
    *
    * This is an *absolute* move.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short, a DimensionalityError will be created.
    *
    * For convenience the object itself is returned as the result of the method.
    *
    */
   public final ISBounds moveCenterTo ( int[] location )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++ )
         {
            // Note that we must compute centerCoord this way, rather than
            // as (upper + lower)/2, so that we get the same round off here
            // as we do in createCentered methods
            int centerCoord = ( this.upper[i] + this.lower[i] + 1 ) / 2;
            this.lower[i] += location[i] - centerCoord;
            this.upper[i] += location[i] - centerCoord;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Argument array has too few dimensions.");
      }
      return this;
   }
   
   /** Translate the ISBounds object in space so that the
    * center is moved to the coordinates specified
    * in the integer array argument.
    *
    * This is an *absolute* move.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short, a DimensionalityError will be created.
    *
    * For convenience the object itself is returned as the result of the method.
    *
    */
   public final ISBounds moveCenterTo ( IndexSpaceID location )
   {
      return moveCenterTo ( location.getCoords () );
   }
   
   /** Translate this ISBounds in space so that the lower
    *   point is now at the origin. For convenience, this
    *   is returned as the value of the method.
    */
   public ISBounds moveToOrigin ()
   {
      for ( int i=0; i<this.lower.length; i++)
      {
         this.upper[i] -= this.lower[i];
         this.lower[i] = 0;
      }
      return this;
   }
   
   /** Translate this ISBounds in space so it is expressed in
    * the local coordinate system whose origin is at the
    * IndexSpaceId passed as an argument.
    */
   public ISBounds makeRelativeTo ( IndexSpaceID localOrigin )
   {
      try
      {
         for ( int i=0; i<this.lower.length; i++)
         {
            this.upper[i] -= localOrigin.getCoord ( i );
            this.lower[i] -= localOrigin.getCoord ( i );
         }
         return this;
      }
      catch ( Exception ex )
      {
         new SDBError ( SDBError.Recoverable,
            "ISBounds.makeRelativeTo: dimension mismatch: this dimension = " +
            this.lower.length + ". Argument is " + localOrigin.dim () );
         return this;
      }
   }
   
   /** Translate this ISBounds in space so it is expressed in
    * the local coordinate system whose origin is at the
    * IndexSpaceId passed as an argument.
    */
   public ISBounds makeRelativeTo ( int[] localOrigin )
   {
      try
      {
         for ( int i=0; i<this.lower.length; i++)
         {
            this.upper[i] -= localOrigin[i];
            this.lower[i] -= localOrigin[i];
         }
      }
      catch ( Exception ex )
      {
         new SDBError ( SDBError.Recoverable,
            "ISBounds.makeRelativeTo: dimension mismatch: this dimension = " +
            this.lower.length + ". Argument is " + localOrigin.length );
      }
      return this;
   }
   
   /** Returns a new ISBounds of the same shape as the argument
    * bounds, but moved to the origin.
    * @deprecated
    * Use bounds.copyToOrigin( )
    */
   final public static ISBounds copyAndMoveToOrigin ( ISBounds bounds )
   {
      return bounds.copyToOrigin ();
   }
   
   /** Updates this ISBounds object to be the
    * bounding volume of this ISBounds and the parameter
    * ISBounds. If the dimensionality of the parameter ISBounds
    * does not match the dimensionality of this ISBounds, then
    * a DimensionalityException is thrown.
    *      "this" is returned for convenience of use.
    * @deprecated
    * Use setToBoundingVolume( ISBounds )
    */
   final public ISBounds boundingVolume ( ISBounds db )
   {
      return setToBoundingVolume ( db );
   }
   /** Updates this ISBounds object to be the
    * bounding volume of this ISBounds and the parameter
    * ISBounds. If the dimensionality of the parameter ISBounds
    * does not match the dimensionality of this ISBounds, then
    * a DimensionalityException is thrown.
    *      "this" is returned for convenience of use.
    */
   public ISBounds setToBoundingVolume ( ISBounds db )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         { // for each axis
            // Only change the lower value for the current axis
            // if the value from db is smaller.
            this.lower[i] = (db.lower[i] < this.lower[i]) ? db.lower[i] : this.lower[i];
            
            // Only change the upper value for the current axis
            // if the value from db is larger.
            this.upper[i] = (db.upper[i] > this.upper[i]) ? db.upper[i] : this.upper[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Dimensionality of Bounds argument does not match this Bounds object.");
      }
      return this;
   }
   
   /**
    *  If the intersection of this with the argument is non-empty, the method returns
    *  true, and false otherwise.  This object is unchanged.
    */
   public boolean intersects ( ISBounds b2 )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            if ( (this.lower[i] < b2.lower[i] && this.upper[i] < b2.upper[i]) ||
               (b2.lower[i] < this.lower[i] && b2.upper[i] < this.upper[i]))
               return false;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Dimensionality of Bounds argument does not match this Bounds object.");
         return false;
      }
      return true; // The bounds intersect for all dimensions, so return true.
   }
   
   /** Returns the intersecton of this object with the ISBounds argument. If the intersection
    * doesn't exist, then return null. */
   public ISBounds intersection ( ISBounds b2 )
   {
      try
      {
         for ( int i=0; i<this.lower.length; i++)
         {
            if ( ( this.upper[i] < b2.lower[i] )
            || ( b2.upper[i] < this.lower[i] ))
               return null;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
         return null;
      }
      
      ISBounds newBnds = new ISBounds ( this.dim () );
      for ( int i=0; i<this.lower.length; i++)
      {
         newBnds.lower[i] = ( this.lower[i] > b2.lower[i] ) ? this.lower[i] : b2.lower[i];
         newBnds.upper[i] = ( this.upper[i] > b2.upper[i] ) ? b2.upper[i] : this.upper[i];
      }
      return newBnds;
   }
   
   /** Updates this ISBounds to the intersection of itself with the argument.
    * Returns this as a convenience. Throws an exception if the intersection
    * is null.
    */
   public ISBounds intersectWith ( ISBounds b2 )
   {
      try
      {
         for ( int i=0; i<this.lower.length; i++)
         {
            if ( ( this.upper[i] < b2.lower[i] )
            || ( b2.upper[i] < this.lower[i] ))
               new SDBError ( SDBError.AlwaysThrow, "intersectWith results in null space.");
            
            
            this.lower[i] = ( this.lower[i] > b2.lower[i] ) ? this.lower[i] : b2.lower[i];
            this.upper[i] = ( this.upper[i] > b2.upper[i] ) ? b2.upper[i] : this.upper[i];
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
         return null;
      }
      return this;
   }
   
   /** Returns the intersecton of this object with the ISBounds argument. If the intersection
    * doesn't exist, then return null. */
   private ISBounds pintersection ( ISBounds b2 )
   {
      //ISBounds newBnds = null;
      int maxlower[]=new int[ this.lower.length ];
      int minupper[]=new int[ this.lower.length ];
      
      try
      {
         for ( int i=0; i<this.lower.length; i++)
         {
            if ( (this.lower[i] < b2.lower[i] && this.upper[i] < b2.lower[i])
            || (b2.lower[i] < this.lower[i] && b2.upper[i] < this.lower[i])
            )
               return null;
            else
            {
               maxlower[i] = ( this.lower[i] > b2.lower[i] ) ? this.lower[i] : b2.lower[i];
               minupper[i] = ( this.upper[i] > b2.upper[i] ) ? b2.upper[i] : this.upper[i];
            }
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
         return null;
      }
      return new ISBounds ( maxlower, minupper );
   }
   
   /** Returns true if this Bounds completely contains the argument bounds. */
   public boolean contains ( ISBounds b2 )
   {
      try
      {
         for ( int i=0; i<this.lower.length; i++ )
         {
            if ( ! ( this.lower[i] <= b2.lower[i]
               && this.upper[i] >= b2.upper[i]) )
            {
               return false;
            }
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
         return false;
      }
      return true;
   }
   
   /** Returns true if this Bounds contains the coordinates represented in the integer array.
    *   Unlike the usual contains method, If the coordinates lie on an upper edge of the bounds,
    *   this method will return false. The array argument should have a length corresponding to the
    *   dimensionality of the bounds object. However, if the argument
    *   array is too long, the extra elements will be ignored. If it is
    *   too short,  a DimensionalityError will will be created.
    */
   public boolean containsExcludeUpper ( int [] coords )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            if ( (coords[i] < this.lower[i] ) ||
               (coords[i] >= this.upper[i] ) )
               return false;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Argument array has too few dimensions.");
         return false;
      }
      return true;
   }
   
   /** Returns true if this Bounds contains the coordinates represented in the IndexSpaceID.
    *   Unlike the usual contains method, If the coordinates lie on an upper edge of the bounds,
    *   this method will return false. The array argument should have a length corresponding to the
    *   dimensionality of the bounds object. However, if the argument
    *   array is too long, the extra elements will be ignored. If it is
    *   too short,  a DimensionalityError will will be created.
    */
   public boolean containsExcludeUpper ( IndexSpaceID id )
   {
      return this.containsExcludeUpper (id.refCoords ());
   }
   
   /** Returns true if this Bounds contains the coordinates contained in the integer array.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    *  array is too long, the extra elements will be ignored. If it is
    * too short,  a DimensionalityError will will be created.
    */
   public boolean contains ( IndexSpaceID id )
   {
      return this.contains (id.refCoords ());
   }
   
   /** Returns true if this Bounds contains the coordinates contained in the integer array.
    *
    * The array argument should have a length corresponding to the
    * dimensionality of the bounds object. However, if the argument
    * array is too long, the extra elements will be ignored. If it is
    * too short,  a DimensionalityError will will be created.
    */
   public boolean contains ( int [] coords )
   {
      try
      {
         for ( int i=0; i < this.lower.length; i++)
         {
            if ( (coords[i] < this.lower[i] ) ||
               (coords[i] > this.upper[i] ) )
               return false;
         }
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         new DimensionalityError ( SDBError.Recoverable,
            "Argument array has too few dimensions.");
         return false;
      }
      return true;
   }
   
   /** Return a String representing the object. */
   public String toString ()
   {
      return this.toString ("");
   }
   
   /** Return a String representing the object indented by the argument string. */
   public String toString ( String indent )
   {
      int d = this.lower.length;
      StringBuffer result=new StringBuffer (indent+"{ISBounds: l=<");
      
      for ( int i=0; i<d-1; i++)
         result.append (lower[i]+", ");
      result.append (lower[d-1]+">, u=<");
      
      for ( int i=0; i<d-1; i++)
         result.append (upper[i]+", ");
      result.append (upper[d-1]+">}");
      
      return new String (result);
   }
   
   //-------------------------- testCenter() -------------------------
   /* test center related methods added 3/31/05 */
   static void testCenter ()
   {
      System.out.println ( "============ Center tests ========================" );
      
      // isb center is at 3, 4, 5.5 = 3,4,5 or 3,4,6 << 3,4,6 >> is what works.
      ISBounds isb = new ISBounds (
         new int[] { 1, 2, 2 },
         new int[] { 5, 6, 9 } );
      int[]  center = new int[] { 3,2,7 };
      int[]  sizeArray = new int[] {5, 5, 8};
      IndexSpaceID isid = new IntegerIndexSpaceID ( center );
      IndexSpaceID size = new IntegerIndexSpaceID ( sizeArray );
      
      ISBounds isb2 = ISBounds.createCentered ( isb, isid );
      System.out.println ( "Should be: [1,0,3], [5,4,10]: " + isb2 );
      isb2 = ISBounds.createCentered ( size, isid );
      System.out.println ( "Should be: [1,0,3], [5,4,10]: " + isb2 );
      isb2 = ISBounds.createCentered ( sizeArray, center );
      System.out.println ( "Should be: [1,0,3], [5,4,10]: " + isb2 );
      
      isb2 = isb.copyAndMoveCenterTo ( center );
      System.out.println ( "Should be: [1,0,3], [5,4,10]: " + isb2 );
      isb2 = isb.copyAndMoveCenterTo ( isid );
      System.out.println ( "Should be: [1,0,3], [5,4,10]: " + isb2 );
      
      isb.moveCenterTo ( new int[] {1,1,10});
      System.out.println ( "Should be: [-1,-1,6], [3,3,13]: " + isb );
   }
   /** For debugging. Runs basic tests on the ISBounds class. */
   static public void main (String[] args)
   {
      testIntersect ();
      testBasics ();
      testCenter ();
   }
   //----------- test intersection timings --------------
   static private void testIntersect ()
   {
      int[] pt0 = new int[] { 0, 0, 0 };
      int[] pt1 = new int[] { 1, 1, 1 };
      //int[] pt2 = new int[] { 2, 2, 2 };
      int[] pt3 = new int[] { 3, 3, 3 };
      //int[] pt4 = new int[] { 4, 4, 4 };
      int[] pt5 = new int[] { 5, 5, 5 };
      
      ISBounds isb01 = new ISBounds ( pt0, pt1 );
      /*** not used
       * ISBounds isb02 = new ISBounds( pt0, pt2 );
       * ISBounds isb03 = new ISBounds( pt0, pt3 );
       * ISBounds isb04 = new ISBounds( pt0, pt4 );
       * ISBounds isb25 = new ISBounds( pt2, pt5 );
       * /*******/
      ISBounds isb05 = new ISBounds ( pt0, pt5 );
      ISBounds isb35 = new ISBounds ( pt3, pt5 );
      
      ISBounds out = null;
      
      int count = 10000;
      // primer for class loading
      for ( int i = 0; i < count; i++ )
         out = isb01.pintersection ( isb05 );
      if ( out == null ) // force reference to avoid warning
         out = null;
      count = 10000000;
      System.gc ();
      
      long start;
      
      System.out.println ( "--Non-Null test:    old1    new1    with'" );
      start = System.currentTimeMillis ();
      //for ( int i = 0; i < count; i++ ) out = isb01.pintersection( isb05 );
      float time1Old = (System.currentTimeMillis () - start)/ 1000.0f;
      System.gc ();
      
      start = System.currentTimeMillis ();
      for ( int i = 0; i < count; i++ )
         out = isb01.intersection ( isb05 );
      float time1New = (System.currentTimeMillis () - start)/ 1000.0f;
      System.gc ();
      
      start = System.currentTimeMillis ();
      for ( int i = 0; i < count; i++ )
         isb01 = isb01.intersectWith ( isb05 );
      float time3Old = (System.currentTimeMillis () - start)/ 1000.0f;
      System.out.println ( "                    "
         + time1Old + "    " + time1New + "    " + time3Old );
      System.gc ();
      
      System.out.println ( "--Null     test:    old1    new1    old3    new1'" );
      start = System.currentTimeMillis ();
      //for ( int i = 0; i < count; i++ ) out = isb01.pintersection( isb35 );
      time1Old = (System.currentTimeMillis () - start)/ 1000.0f;
      System.gc ();
      
      start = System.currentTimeMillis ();
      for ( int i = 0; i < count; i++ )
         out = isb01.intersection ( isb35 );
      time1New = (System.currentTimeMillis () - start)/ 1000.0f;
      System.gc ();
      
      start = System.currentTimeMillis ();
      //for ( int i = 0; i < count; i++ )  out = isb01.intersection( isb35, tmp1, tmp2 );
      time3Old = (System.currentTimeMillis () - start)/ 1000.0f;
      System.out.println ( "                    "
         + time1Old + "    " + time1New + "    " + time3Old );
      
   }
   
   //----------- basic -----------
   static private void testBasics ()
   {
      System.out.println ("\n**********************\n***ISBounds Tests***\n**********************");
      
      IndexSpaceID temp;
      
      IntegerIndexSpaceID id1=new IntegerIndexSpaceID (0);
      IntegerIndexSpaceID id2=new IntegerIndexSpaceID (7);
      
      IntegerIndexSpaceID id3=new IntegerIndexSpaceID (1,1);
      IntegerIndexSpaceID id4=new IntegerIndexSpaceID (6,6);
      
      IntegerIndexSpaceID id5=new IntegerIndexSpaceID (2,3);
      IntegerIndexSpaceID id6=new IntegerIndexSpaceID (0,7);
      
      IntegerIndexSpaceID id7=new IntegerIndexSpaceID (1,2,3);
      IntegerIndexSpaceID id8=new IntegerIndexSpaceID (4,1,7);
      IntegerIndexSpaceID id9=new IntegerIndexSpaceID (8,1,9);
      
      IntegerIndexSpaceID id10=new IntegerIndexSpaceID (3);
      
      System.out.println ("\n\n***ISBounds Constructor and create Tests***");
      
      ISBounds db1=new ISBounds (id1,id2);
      ISBounds db2=new ISBounds (id3,id4);
      ISBounds db3=new ISBounds (id5,id6);
      ISBounds db4=new ISBounds (id7,id8);
      ISBounds db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      ISBounds db6=new ISBounds (new IntegerIndexSpaceID (1,-1,0), new IntegerIndexSpaceID (6,5,4));
      ISBounds db7 = new ISBounds (new IntegerIndexSpaceID (0,3),new IntegerIndexSpaceID (0,3));
      ISBounds theCopy;
      
      System.out.print (" 1D ISBounds: Should be {(0),(7)}-->" +db1);
      System.out.print ("\n 2D ISBounds: Should be {(1,1),(6,6)}-->"+db2);
      System.out.print ("\n 2D ISBounds with coord swapping: Should be {(0,3),(2,7)}-->"+db3);
      System.out.print ("\n 3D ISBounds with coord swapping: Should be {(1,1,3),(4,2,7)}-->"+db4);
      System.out.println ();
      
      try
      {
         ISBounds foo = new ISBounds ( id1,id3 );
         System.out.println (" ***Error: A DimensionalityException should have " +
            "been thrown due to mismatched constructor arguments. "
            + foo );
      }
      catch (DimensionalityException e)
      {
         System.out.println (" DimensionalityException correctly thrown due to " +
            "mismatched constructor arguments.");
      }
      
        /*
        try {
            ISBounds bar=new ISBounds(id8,id9);
            System.out.println(" ***Error: A BoundsException should have been thrown due to a degenerate volume definition.");
        } catch (BoundsException e){
            System.out.println(" BoundsException correctly thrown due to a degenerate volume definition.");
        }
         */
      
      System.out.println ( "\n\n***Create Tests***" );
      theCopy = db2.copyTo ( id4 );
      System.out.print ( "3D createAt: Should be {(6,6),(11,11)}-->" + theCopy );
      
      System.out.println ( "\n\n***Copy Tests***" );
      
      theCopy=(ISBounds) db1.clone ();
      System.out.print (" 1D ISBounds clone(): Should be {(0),(7)}-->");
      System.out.println ( "*** " + theCopy);
      theCopy=(ISBounds) db2.clone ();
      System.out.print ("\n 2D ISBounds clone(): Should be {(1,1),(6,6)}-->"+theCopy);
      theCopy=(ISBounds) db4.clone ();
      System.out.print ("\n 3D ISBounds clone(): Should be {(1,1,3),(4,2,7)}-->"+theCopy);
      
      
      System.out.println ("\n\n***getLower Tests***");
      
      temp=db1.copyLower ();
      System.out.print (" 1D ISBounds getLower(): Should be (0)-->"+temp);
      temp=db2.copyLower ();
      System.out.print ("\n 2D ISBounds getLower(): Should be (1,1)}-->"+temp);
      temp=db4.copyLower ();
      System.out.print ("\n 3D ISBounds getLower(): Should be (1,1,3)-->"+temp);
      
      System.out.println ("\n\n***getUpper Tests***");
      
      temp=db1.copyUpper ();
      System.out.print (" 1D ISBounds getUpper(): Should be (7)-->"+temp);
      temp=db2.copyUpper ();
      System.out.print ("\n 2D ISBounds getUpper(): Should be (6,6)-->"+temp);
      temp=db4.copyUpper ();
      System.out.print ("\n 3D ISBounds getUpper(): Should be (4,2,7)-->"+temp);
      
      
      System.out.println ("\n\n***dim Tests***");
      
      System.out.print (" 1D ISBounds dim(): Should be 1-->"+db1.dim ());
      System.out.print ("\n 2D ISBounds dim(): Should be 2-->"+db2.dim ());
      System.out.print ("\n 3D ISBounds dim(): Should be 3-->"+db4.dim ());
      
      
      System.out.println ("\n\n***volume Tests***");
      
      System.out.print (" 1D ISBounds volume(): Should be 8-->"+db1.volume ());
      System.out.print ("\n 2D ISBounds volume(): Should be 36-->"+db2.volume ());
      System.out.print ("\n 3D ISBounds volume(): Should be 40-->"+db4.volume ());
      
      
      System.out.println ("\n\n***getDimensions Tests***");
      
      System.out.print (" 1D ISBounds getDimensions(): Should be (8)-->"+db1.getDimensions ());
      System.out.print ("\n 2D ISBounds getDimensions(): Should be (6,6)-->"+db2.getDimensions ());
      System.out.print ("\n 3D ISBounds getDimensions(): Should be (4,2,5)-->"+db4.getDimensions ());
      
      
      System.out.println ("\n\n***move Tests***");
      db1.move (id10);
      System.out.print (" 1D ISBounds move(): Should be {(3),(10)}-->"+db1);
      db3.move (id5);
      System.out.print ("\n 2D ISBounds move(): Should be {(2,6),(4,10)}-->"+db3);
      db4.move (id9);
      System.out.print ("\n 3D ISBounds move(): Should be {(9,2,12),(12,3,16)}-->"+db4);
      id8.negate ();
      db4.move (id8);
      System.out.print ("\n 3D ISBounds move() in negative direction: Should be {(5,1,5),(8,2,9)}-->"+db4);
      db4.move (id8);
      System.out.print ("\n 3D ISBounds move() in negative direction: Should be {(1,0,-2),(4,1,2)}-->"+db4);
      
      
      System.out.println ("\n\n***moveTo Tests***");
      db1.move (id10);
      System.out.print (" 1D ISBounds move(): Should be {(6),(13)}-->"+db1);
      
      
      db3.move (id5);
      System.out.print ("\n 2D ISBounds move(): Should be {(4,9),(6,13)}-->"+db3);
      
      db4.move (id9);
      System.out.print ("\n 3D ISBounds move(): Should be {(9,1,7),(12,2,11)}-->"+db4);
      
      
      
      System.out.println ("\n\n***BoundingVolume Tests***");
      
      db2.boundingVolume (db3);
      
      System.out.println (" 2D BoundingVolume Test: Should be {(1,1),(6,13)}-->"+db2);
      
      theCopy=(ISBounds) db5.clone ();
      theCopy.boundingVolume (db6);
      System.out.println (" 3D BoundingVolume Test: Should be {(1,-1,0),(6,5,6)}-->"+theCopy);
      
      
      System.out.println ("\n\n***Intersection Tests***");
      
      /** rdb: the following seems to be an older version of intersection
       *       that modified "this" rather than returning a new ISBounds.
       *
       * System.out.println("Using public boolean intersection(ISBounds b2)");
       * db2.intersection(db3);
       *
       * System.out.println(" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+db2);
       *
       *
       * theCopy=(ISBounds) db2.clone();
       * theCopy.move( new IntegerIndexSpaceID(10,10) );
       *
       * theCopy.intersection(db2);
       * System.out.println(" 2D Intersection Test: Should be {(0,0),(0,0)}-->"+theCopy);
       *
       * theCopy=(ISBounds) db2.clone();
       * theCopy.move(new IntegerIndexSpaceID(1,2));
       *
       * theCopy.intersection(db2);
       * System.out.println(" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+theCopy);
       *
       * theCopy=(ISBounds) db2.clone();
       * theCopy.move(new IntegerIndexSpaceID(-1,-2));
       *
       * theCopy.intersection(db2);
       * System.out.println(" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+theCopy);
       *
       * theCopy=new ISBounds(new IntegerIndexSpaceID(1,1),new IntegerIndexSpaceID(2,2));
       *
       * db7.intersection(theCopy);
       * System.out.println(" 2D Intersection Test: Should be {(0,0),(0,0)}-->"+db7);
       *
       * db7=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(3,3));
       * theCopy.intersection(db7);
       * System.out.println(" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+theCopy);
       *
       *
       * theCopy=(ISBounds) db6.clone();
       *
       * theCopy.intersection(db5);
       * System.out.println(" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+theCopy);
       *
       *
       * /********************/
      System.out.println ("Using ISBounds intersection(ISBounds b2)");
      
      db2=new ISBounds (new IntegerIndexSpaceID (4,9),new IntegerIndexSpaceID (6,13));
      db3=new ISBounds (new IntegerIndexSpaceID (4,9),new IntegerIndexSpaceID (6,13));
      db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      db6=new ISBounds (new IntegerIndexSpaceID (1,-1,0), new IntegerIndexSpaceID (6,5,4));
      
      ISBounds result=db2.intersection (db3);
      
      System.out.println (" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (10,10));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be null-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (1,2));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (-1,-2));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+result);
      
      theCopy=new ISBounds (new IntegerIndexSpaceID (1,1),new IntegerIndexSpaceID (2,2));
      db7 = new ISBounds (new IntegerIndexSpaceID (0,3),new IntegerIndexSpaceID (0,3));
      
      result=db7.intersection (theCopy);
      System.out.println (" 2D Intersection Test: Should be null-->"+result);
      
      db7=new ISBounds (new IntegerIndexSpaceID (0,0),new IntegerIndexSpaceID (3,3));
      result=theCopy.intersection (db7);
      System.out.println (" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+result);
      
      
      theCopy=(ISBounds) db6.clone ();
      
      result=theCopy.intersection (db5);
      System.out.println (" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+result);
      /*************/
      System.out.println ("Using ISBounds intersection3(ISBounds b2, int max[], int min[])");
      
      //int [] min2 = new int[2];
      //int [] max2 = new int[2];
      
      db2=new ISBounds (new IntegerIndexSpaceID (4,9),new IntegerIndexSpaceID (6,13));
      db3=new ISBounds (new IntegerIndexSpaceID (4,9),new IntegerIndexSpaceID (6,13));
      db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      db6=new ISBounds (new IntegerIndexSpaceID (1,-1,0), new IntegerIndexSpaceID (6,5,4));
      
      
      result=db2.intersection (db3);
      
      System.out.println (" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (10,10));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be null-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (1,2));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+result);
      
      theCopy=(ISBounds) db2.clone ();
      theCopy.move (new IntegerIndexSpaceID (-1,-2));
      
      result=theCopy.intersection (db2);
      System.out.println (" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+result);
      
      theCopy=new ISBounds (new IntegerIndexSpaceID (1,1),new IntegerIndexSpaceID (2,2));
      db7 = new ISBounds (new IntegerIndexSpaceID (0,3),new IntegerIndexSpaceID (0,3));
      
      result=db7.intersection (theCopy);
      System.out.println (" 2D Intersection Test: Should be null-->"+result);
      
      db7=new ISBounds (new IntegerIndexSpaceID (0,0),new IntegerIndexSpaceID (3,3));
      result=theCopy.intersection (db7);
      System.out.println (" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+result);
      
      //int [] min3=new int[3];
      //int [] max3=new int[3];
      
      theCopy=(ISBounds) db6.clone ();
      result=theCopy.intersection (db5);
      System.out.println (" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+result);
      
      /***************************************/
      System.out.println ( "------------ test return 'this'  ------------------");
      db1=new ISBounds (new IntegerIndexSpaceID (0,0),new IntegerIndexSpaceID (1,1));
      db2=new ISBounds (new IntegerIndexSpaceID (4,9),new IntegerIndexSpaceID (6,13));
      db3=new ISBounds (new IntegerIndexSpaceID (5,3),new IntegerIndexSpaceID (10,12));
      db4=new ISBounds (new IntegerIndexSpaceID (0,0,0), new IntegerIndexSpaceID (9,9,6));
      db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      db6=new ISBounds (new IntegerIndexSpaceID (1,-1,0), new IntegerIndexSpaceID (10,5,8));
      
      System.out.println ( " copy(ISB): both should be [4,9] to [6,13]" +
         db1.copy ( db2 ) + "     " + db1 );
      System.out.println ( " 2d boundingVolume: both should be [4,3] to [10,13]" +
         db3.boundingVolume ( db2 ) + "    " + db3 );
      System.out.println ( " 3d boundingVolume: both should be [0,-1,0] to [10,9,8]" +
         db6.boundingVolume ( db4 ) + "    " + db6 );
      System.out.println ( " copyAndMove(ISB): both should be [5,10] to [7,14]" +
         db1.copyAndMove ( db2, new int[] {1,1}) + "     " + db1 );
      System.out.println ( " copyAndMoveTo(ISB): both should be [1,1] to [3,5]" +
         db1.copyAndMoveTo ( db2, new int[] {1,1}) + "     " + db1 );
      System.out.println ( " move(): both should be [5,10] to [7,14]" +
         db2.move ( new int[] {1,1}) + "     " + db2 );
      System.out.println ( " moveTo(): both should be [1,1] to [3,5]" +
         db2.moveTo ( new int[] {1,1}) + "     " + db2 );
      System.out.println ( " moveToOrigin(): both should be [0,0] to [2,4]" +
         db2.moveToOrigin () + "     " + db2 );
      System.out.println ( " setLower(ISB): both should be [0,0,0] to [4,5,6]" +
         db5.setLower ( db4.copyLowerArray ()) + "     " + db5 );
      System.out.println ( " setUpper(ISB): both should be [0,0,0] to [9,9,6]" +
         db5.setUpper ( db4.copyUpperArray ()) + "     " + db5 );
      db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      System.out.println ( " setLower(ISB): both should be [0,0,0] to [4,5,6]" +
         db5.setLower ( db4.copyLower ()) + "     " + db5 );
      System.out.println ( " setUpper(ISB): both should be [0,0,0] to [9,9,6]" +
         db5.setUpper ( db4.copyUpper ()) + "     " + db5 );
      db5=new ISBounds (new IntegerIndexSpaceID (1,1,1), new IntegerIndexSpaceID (4,5,6));
      System.out.println ( " dopyLower(ISB): both should be [0,0,0] to [4,5,6]" +
         db5.copyLower ( db4.copyLowerArray ()) + "     " + db5 );
      System.out.println ( " copyUpper(ISB): both should be [0,0,0] to [9,9,6]" +
         db5.copyUpper ( db4.copyUpperArray ()) + "     " + db5 );
      
      
      /***************************************/
      
      System.out.println ();
      System.out.println ("Speed Test:  intersection() vs. intersection2() vs. intersection3() [in 3D]");
      System.out.println (" With non-null intersection:");
      
      long startTime,stopTime;
      ISBounds dst=new ISBounds (new IntegerIndexSpaceID (1,2,3),new IntegerIndexSpaceID (10,11,12));
      ISBounds src=new ISBounds (new IntegerIndexSpaceID (4,5,6),new IntegerIndexSpaceID (12,13,14));
      
      System.gc ();
      startTime=System.currentTimeMillis ();
      
      for ( int i=0; i<1000000; i++)
      {
         dst.intersects (src);
      }
      
      stopTime=System.currentTimeMillis ();
      
      System.out.println ("  Time for one Million intersects() calls: " + (stopTime-startTime) + " milliseconds.");
      
      System.gc ();
      
      startTime=System.currentTimeMillis ();
      /*** implementation deleted
       *
       * for ( int i=0; i<1000000; i++){
       * dst.intersection(src);
       * }
       *
       * stopTime=System.currentTimeMillis();
       *
       * System.out.println("  Time for one Million intersection(ISBounds) calls:" + (stopTime-startTime) + " milliseconds.");
       *
       * System.gc();
       *
       * startTime=System.currentTimeMillis();
       *
       * for ( int i=0; i<1000000; i++){
       * dst.intersection(src,min3,max3);
       * }
       *
       * stopTime=System.currentTimeMillis();
       *
       * System.out.println("  Time for one Million intersection(ISBounds, int[], int[]) calls:" + (stopTime-startTime) + " milliseconds.");
       *
       *
       * System.out.println(" With null intersection:");
       *
       *
       * dst=new ISBounds(new IntegerIndexSpaceID(1,2,3),new IntegerIndexSpaceID(10,11,12));
       * src=new ISBounds(new IntegerIndexSpaceID(9,12,11),new IntegerIndexSpaceID(15,16,17));
       *
       * System.gc();
       * startTime=System.currentTimeMillis();
       *
       * for ( int i=0; i<1000000; i++){
       * dst.intersects(src);
       * }
       *
       * stopTime=System.currentTimeMillis();
       *
       * System.out.println("  Time for one Million intersects() calls: " + (stopTime-startTime) + " milliseconds.");
       *
       * System.gc();
       *
       * startTime=System.currentTimeMillis();
       *
       * for ( int i=0; i<1000000; i++){
       * dst.intersection(src);
       * }
       *
       * stopTime=System.currentTimeMillis();
       *
       * System.out.println("  Time for one Million intersection(ISBounds) calls:" + (stopTime-startTime) + " milliseconds.");
       *
       * System.gc();
       *
       * startTime=System.currentTimeMillis();
       *
       * for ( int i=0; i<1000000; i++){
       * dst.intersection(src,min3,max3);
       * }
       *
       * stopTime=System.currentTimeMillis();
       *
       * System.out.println("  Time for one Million intersection(ISBounds, int[], int[]) calls:" + (stopTime-startTime) + " milliseconds.");
       *
       * System.gc();
       * /************************************/
      
      
      System.out.println ();
      System.out.println ("Speed Test:  assign() vs. copy()");
      
      
      
      dst=new ISBounds (new IntegerIndexSpaceID (1,2,3),new IntegerIndexSpaceID (10,11,12));
      src=new ISBounds (new IntegerIndexSpaceID (4,5,6),new IntegerIndexSpaceID (12,13,14));
      
      
      startTime=System.currentTimeMillis ();
      
      for ( int i=0; i<1000000; i++)
      {
         dst= src.copy ();
      }
      stopTime=System.currentTimeMillis ();
      
      System.out.println ("Time for one Million copies: " + (stopTime-startTime) + " milliseconds.");
      startTime=System.currentTimeMillis ();
      
      for ( int i=0; i<1000000; i++)
      {
         dst.assign (src);
      }
      stopTime=System.currentTimeMillis ();
      
      System.out.println ("Time for one Million assigns: " + (stopTime-startTime) + " milliseconds.");
   }
}

