/*    
 ISIDMapper.java
 Copyright 2001 University of New Hampshire
 Durham, NH 03824
 
 *------------------------- Edit history (as of 01/11/06) --------------------
 * 05/24/08 rdb: Added toString() method.
 * 01/10/06 rdb: Revised implementation to improve performance. If the mapping
 *               does not need scaling, we now never setup the multipliers array
 *               and avoid doing the scaling during mapping. It might have been
 *               a bit more elegant to use subclassing, but this recoding gives a
 *               measurable improvement without changing any other code: for ISID 
 *               versions of toComponent/toComposite that don't need scaling, 
 *               get about 20% improvement, for int[] versions it's closer to 50%. 
 *          rdb: Created versions of toComponent/toComposite that accept pre-created
 *               arguments to be filled in, rather than allocating a new result
 *               object -- these would provide another 50% improvement for no
 *               scaling and 40-45% improvement for the scaling version -- if
 *               they were used.
 *
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

/**
 * This class is used for mapping IndexSpaceIDs and ISBounds to/from ISBounds
 * defined within component and composite spaces.
 *  
 */

public class ISIDMapper
{
   ISBounds componentSubset;
   ISBounds compositeExtent;
   int dimensionality;
   double[] multipliers = null; //conversion factors from composite to component space
   int[] compositeLB; //
   int[] componentLB; //

   /**
    * Create an ISIDMapper. The first argument specifies the subset of the
    * component space that will be mapped to a region in the composite space.
    * The second argument specifies the extent of this region in the composite
    * space.
    */
   public ISIDMapper( ISBounds componentSubset, ISBounds compositeExtent )
   {
      int d = componentSubset.dim();
      this.dimensionality = d;
      // Clone the Bounds arguments.
      this.componentSubset = new ISBounds (componentSubset);

      if ( compositeExtent == null )
         this.compositeExtent = new ISBounds (componentSubset);
      else
      {
         this.compositeExtent = new ISBounds (compositeExtent);
         
         if ( d != compositeExtent.dim() )
         {
            new DimensionalityError( SDBError.Recoverable,
                                   "Component and Composite Bounds do not match." );
         }
      }
      this.initialize( d );
   }

   /** Create an ISIDMapper which does identity mapping. */
   public ISIDMapper( ISBounds bounds )
   {
      this( bounds, null );
      /****** rdb 01.10.06
        
      int d = bounds.dim();

      this.componentSubset = (ISBounds) bounds.clone();
      this.compositeExtent = (ISBounds) bounds.clone();

      this.dimensionality = d;

      this.compositeLB = this.compositeExtent.getLowerArray();
      this.componentLB = this.componentSubset.getLowerArray();

      this.multipliers = new double[d];
      for ( int i = 0; i < d; i++ )
         multipliers[i] = 1;
      /****************************/
   }

   /**
    * Build the multiplier array and set the two LB datamenbers. This method
    * assumes that the compositeExtent and componentSubset data members have
    * already been set and the space for the multiplier array has already been
    * allocated. The parameter specifies the dimensionality.
    */
   private void initialize( int dim )
   {
      this.compositeLB = this.compositeExtent.getLowerArray();
      this.componentLB = this.componentSubset.getLowerArray();
      
      if ( this.componentSubset.equalShape( this.compositeExtent ))
         this.multipliers = null;
      else  // need multipliers array
      {
         if ( this.multipliers == null || this.multipliers.length != dim )
            this.multipliers = new double[ dim ];
         
         for ( int i = 0; i < dim; i++ )
         {
            int[] compositeHB = this.compositeExtent.getUpperArray();
            int[] componentHB = this.componentSubset.getUpperArray();
            this.multipliers[i] = ( (double) ( componentHB[i] - this.componentLB[i] + 1 ))
                                    / ( compositeHB[i] - this.compositeLB[i] + 1 );
         }
      }
   }

   /**
    * Assign the argument bounds to the componentSubset of this ISIDMapper. No
    * copy of the bounds is made. Package access only.
    */
   void assignComponentSubset( ISBounds componentSubset )
   {
      this.componentSubset.assign( componentSubset );

      initialize( this.dimensionality );
   }

   /**
    * Assign the argument bounds to the compositeExtent of this ISIDMapper. No
    * copy of the bounds is made. Package access only.
    */
   void assignCompositeExtent( ISBounds compositeExtent )
   {

      this.compositeExtent.assign( compositeExtent );

      initialize( this.dimensionality );
   }

   /** Map the IndexSpaceID argument from component space to composite space. */
   public IndexSpaceID toComposite( IndexSpaceID componentID )
   {
      int pos;   // index position
      IndexSpaceID compositeID = componentID.copyID();

      if ( this.multipliers == null )
      {
         for ( int i = 0; i < this.dimensionality; i++ )
         {
            pos = componentID.getCoord( i ) - this.componentLB[i] + this.compositeLB[i];
            compositeID.setCoord( i, pos );
         }
      }
      else
      {
         for ( int i = 0; i < this.dimensionality; i++ )
         {
            pos = (int)(( componentID.getCoord( i ) - this.componentLB[i] ) 
                                          / multipliers[i] ) + this.compositeLB[i];
            compositeID.setCoord( i, pos );
         }
      }
      return compositeID;
   }

   /** Map the coords array argument from component space to composite space. */
   public int[] toComposite( int[] componentCoords )
   {
      int d = this.dimensionality;
      int[] compositeCoords = new int[d];

      for ( int i = 0; i < d; i++ )
      {
         if ( this.multipliers == null )
            compositeCoords[i] = componentCoords[i] - this.componentLB[i]
                                          + this.compositeLB[i];
         else
            compositeCoords[i] = (int)(( componentCoords[i] - this.componentLB[i] )
                                       / multipliers[i] + this.compositeLB[i] );
      }

      return compositeCoords;
   }

   private IndexSpaceID toCompositeRoundUp( IndexSpaceID componentID )
   {

      int i;
      IndexSpaceID compositeID = componentID.copyID();

      for ( i = 0; i < this.dimensionality; i++ )
      {
         // compositeID.setCoord(i,(int)
         // (0.5+componentID.getCoord(i)/multipliers[i]-offsets[i]));
         compositeID.setCoord( i, (int) ( ( componentID.getCoord( i )
               - this.componentLB[i] + 0.5 ) / multipliers[i] )
               + this.compositeLB[i] );
      }

      return compositeID;
   }

   /** Map the bounds argument from component space to composite space. */
   public ISBounds toComposite( ISBounds componentBounds )
   {
      IndexSpaceID lower, upper;

      lower = toComposite( componentBounds.getLower() );
      if ( this.multipliers == null )
         upper = toComposite( componentBounds.getUpper() );
      else
         upper = toCompositeRoundUp( componentBounds.getUpper() );

      return new ISBounds( lower, upper );

   }

   private IndexSpaceID toComponentRoundUp( IndexSpaceID compositeID )
   {
      IndexSpaceID componentID = compositeID.copyID();
      for ( int i = 0; i < this.dimensionality; i++ )
      {
         componentID.setCoord( i, (int) ( ( compositeID.getCoord( i )
               - this.compositeLB[i] + 0.5 )
               * multipliers[i] + this.componentLB[i] ) );
      }

      return componentID;

   }

   /** Map the IndexSpaceID argument from composite space to component space. */
   final public IndexSpaceID toComponent( IndexSpaceID compositeID )
   {
      return toComponent( null, compositeID );
      /**
      IndexSpaceID componentID = compositeID.copyID();

      for ( int i = 0; i < this.dimensionality; i++ )
      {
         componentID.setCoord( i,
               (int) ( ( compositeID.getCoord( i ) - this.compositeLB[i] )
                      * multipliers[i] + this.componentLB[i] ) );
                     // + this.componentLB[i] ) );
      }

      return componentID;
      /**/
   }

   /** Map the IndexSpaceID argument from composite space to component space. */
   public IndexSpaceID toComponent( IndexSpaceID componentID, IndexSpaceID compositeID )
   {
      int pos;    // index position
      if ( componentID == null || componentID.dim() != compositeID.dim() )
         componentID = compositeID.copyID();

      if ( multipliers == null )
      {  // no scaling needed
         for ( int i = 0; i < this.dimensionality; i++ )
         {
            pos = compositeID.getCoord( i ) - this.compositeLB[i] + this.componentLB[i];
            componentID.setCoord( i, pos );
         }
      }
      else  
      {  // scaling needed
         for ( int i = 0; i < this.dimensionality; i++ )
         {
            pos = (int)(( compositeID.getCoord( i ) - this.compositeLB[i] )
                                          * multipliers[i] + this.componentLB[i] );
            componentID.setCoord( i, pos );
         }
      }
      return componentID;
   }

   /** Map the coords array argument from composite space to component space. */
   final public int[] toComponent( int[] compositeCoords )
   {
      return toComponent( null, compositeCoords );
      /***
      int[] componentCoords = new int[compositeCoords.length];

      for ( int i = 0; i < this.dimensionality; i++ )
      {
         componentCoords[i] = (int) ( ( compositeCoords[i] - this.compositeLB[i] )
               //* multipliers[i] + this.componentLB[i] );
                 + this.componentLB[i] );
      }

      return componentCoords;
      /****/
   }

   /** Map the coords array argument from composite space to component space. 
    * Pass in the variable to be returned with component location. If the first argument
    * is not null and is of the right dimension, it will be filled and returned. 
    * Otherwise, a new array will be allocated, filled and returned.
    */
   public int[] toComponent( int[] componentCoords, int[] compositeCoords )
   {
      if ( componentCoords == null || componentCoords.length != compositeCoords.length )
         componentCoords = new int[ compositeCoords.length ];
      if ( this.multipliers == null )
      {
         for ( int i = 0; i < this.dimensionality; i++ )
         {
            componentCoords[i] = compositeCoords[i] - this.compositeLB[i] + this.componentLB[i];
         }
      }
      else
      {
         for ( int i = 0; i < this.dimensionality; i++ )
         {            
            componentCoords[i] = (int)(( compositeCoords[i] - this.compositeLB[i] )
                                          * multipliers[i] + this.componentLB[i] );
         }
      }
      return componentCoords;
   }

   /** Map the bounds argument from composite space to component space. */
   public ISBounds toComponent( ISBounds compositeBounds )
   {
      IndexSpaceID lower, upper;

      lower = toComponent( compositeBounds.getLower() );
      if ( this.multipliers == null )
         upper = toComponent( compositeBounds.getUpper() );
      else
         upper = toComponentRoundUp( compositeBounds.getUpper() );

      return new ISBounds( lower, upper );
   }
   
   /** Return a copy of the Composite Extent bounds */
   public ISBounds getCompositeExtent()
   {
      return (ISBounds) this.compositeExtent.clone();
   }

   /** Return a copy of the Component Subset bounds */
   public ISBounds getComponentSubset()
   {
      return (ISBounds) this.componentSubset.clone();
   }

   /** Return true if this ISIDMapper is an identity mapper */
   public boolean isIdentityMapper()
   {
      return this.componentSubset.equals( this.compositeExtent );
   }

   /** Check if the given IndexSpaceID is contained by this component */
   public boolean contains( IndexSpaceID id )
   {
      return this.compositeExtent.contains( id );
   }

   /** Return true if two ISIDMapper equal */
   public boolean equals( ISIDMapper mapper2 )
   {
      return ( mapper2.getComponentSubset().equals( componentSubset ) && mapper2
            .getCompositeExtent().equals( compositeExtent ) );
   }
   
   /** convert to a String representation for debugging */
   public String toString()
   {
	   return componentSubset.toString() + " ---> " + compositeExtent.toString() ;
   }

   /** For debugging. Runs basic tests on the ISIDMapper class. */
   static public void main( String[] args )
   {
      timeTestArray();
      timeTestID();
      
      /*********
      //Uniform Resolution Tests
      System.out.println( "***Uniform Resolution Tests***" );
      ISBounds component = new ISBounds( new IntegerIndexSpaceID( 0, 0 ),
            new IntegerIndexSpaceID( 3, 3 ) );
      ISBounds composite = new ISBounds( new IntegerIndexSpaceID( 10, 10 ),
            new IntegerIndexSpaceID( 13, 13 ) );

      ISIDMapper mapper = new ISIDMapper( component, composite );

      IndexSpaceID c1 = new IntegerIndexSpaceID( 2, 3 );
      IndexSpaceID c2 = new IntegerIndexSpaceID( 12, 13 );
      IndexSpaceID c3 = new IntegerIndexSpaceID( 0, 0 );
      IndexSpaceID c4 = new IntegerIndexSpaceID( 10, 10 );
      IndexSpaceID c5;
      ISBounds isb1 = new ISBounds( new IntegerIndexSpaceID( 1, 1 ),
            new IntegerIndexSpaceID( 2, 2 ) );
      ISBounds isb2 = new ISBounds( new IntegerIndexSpaceID( 10, 10 ),
            new IntegerIndexSpaceID( 12, 12 ) );

      System.out.println( "Component Bounds: " + component );
      System.out.println( "Composite Bounds: " + composite );
      System.out.println( "Component Point=" + c1 + " and Composite Point="
            + mapper.toComposite( c1 ) );
      System.out.println( "Composite Point=" + c2 + " and Component Point="
            + mapper.toComponent( c2 ) );

      System.out.println( "Component Point=" + c3 + " and Composite Point="
            + mapper.toComposite( c3 ) );
      System.out.println( "Composite Point=" + c4 + " and Component Point="
            + mapper.toComponent( c4 ) );

      System.out.println( "Component Bounds=" + isb1 + " and Composite Point="
            + mapper.toComposite( isb1 ) );
      System.out.println( "Composite Bounds=" + isb2 + " and Component Point="
            + mapper.toComponent( isb2 ) );

      //Multiresolution Tests
      System.out.println( "" );
      System.out.println( "\n***Multiresolution Tests***" );

      component = new ISBounds( new IntegerIndexSpaceID( 0, 0 ),
            new IntegerIndexSpaceID( 4, 4 ) );
      composite = new ISBounds( new IntegerIndexSpaceID( 10, 10 ),
            new IntegerIndexSpaceID( 19, 19 ) );

      mapper = new ISIDMapper( component, composite );
      System.out.println( "Component Bounds: " + component );
      System.out.println( "Composite Bounds: " + composite );
      System.out.println( "" );
      c1 = new IntegerIndexSpaceID( 2, 3 );
      c2 = new IntegerIndexSpaceID( 14, 16 );
      c3 = new IntegerIndexSpaceID( 0, 0 );
      c4 = new IntegerIndexSpaceID( 10, 10 );
      c5 = new IntegerIndexSpaceID( 15, 17 );

      System.out.println( "Component Point=" + c1 + " and Composite Point="
            + mapper.toComposite( c1 ) );
      System.out.println( "Composite Point=" + c2 + " and Component Point="
            + mapper.toComponent( c2 ) );
      System.out.println( "Composite Point=" + c5 + " and Component Point="
            + mapper.toComponent( c5 ) );

      System.out.println( "Component Point=" + c3 + " and Composite Point="
            + mapper.toComposite( c3 ) );
      System.out.println( "Composite Point=" + c4 + " and Component Point="
            + mapper.toComponent( c4 ) );
      isb1 = new ISBounds( new IntegerIndexSpaceID( 0, 0 ),
            new IntegerIndexSpaceID( 1, 1 ) );
      System.out.println( "Component Bounds=" + isb1 + " and Composite Point="
            + mapper.toComposite( isb1 ) );
      System.out.println( "Composite Bounds=" + isb2 + " and Component Point="
            + mapper.toComponent( isb2 ) );

      System.out.println( "" );
      /***************************************/
   }
   private static void timeTestID()
   {
      ISBounds component = new ISBounds( new IntegerIndexSpaceID( 0, 0 ),
                                    new IntegerIndexSpaceID( 3, 3 ) );
      ISBounds composite = new ISBounds( new IntegerIndexSpaceID( 10, 10 ),
                                    new IntegerIndexSpaceID( 13, 13 ) );

      ISIDMapper mapper = new ISIDMapper( component, composite );
      IndexSpaceID componentId = new IntegerIndexSpaceID( 1, 1 );
      
      IndexSpaceID  compositeId = new IntegerIndexSpaceID( 11, 11 );
      long start = System.currentTimeMillis();
      int  count = 10000000;
      for ( int i = 0; i < count; i++ )
      {
         //componentId = mapper.toComponent( compositeId );
         componentId = mapper.toComponent( componentId, compositeId );
         //compositeId = mapper.toComposite( componentId );
      }
      float time = ( System.currentTimeMillis() - start ) / 1000.0f;
      System.out.println( componentId );
      System.out.println( "Iterations: " + count + "   Secs: " + time );
   }

   private static void timeTestArray()
   {
      ISBounds component = new ISBounds( new IntegerIndexSpaceID( 0, 0 ),
                                    new IntegerIndexSpaceID( 3, 3 ) );
      ISBounds composite = new ISBounds( new IntegerIndexSpaceID( 10, 10 ),
                                    new IntegerIndexSpaceID( 13, 13 ) );

      ISIDMapper mapper = new ISIDMapper( component, composite );
      int[] componentPos = null;
      
      int[]  compositePos = new int[]{ 11, 11 };
      long start = System.currentTimeMillis();
      int  count = 10000000;
      for ( int i = 0; i < count; i++ )
      {
         //componentPos = mapper.toComponent( compositePos );
         componentPos = mapper.toComponent( componentPos, compositePos );
      }
      float time = ( System.currentTimeMillis() - start ) / 1000.0f;
      System.out.println( "[" + componentPos[ 0 ] + "," + componentPos[ 1 ] + "]" );
      System.out.println( "Iterations: " + count + "   Secs: " + time );
   }
 }