package edu.unh.sdb.datasource;

//import java.io.*;
import edu.unh.sdb.common.*;

/** Child class of IndexSpaceID. ISIterator can iterate through an IndexSpace*/

/* ------------------- Edit history as of 09/16/06 ----------------------------
 * 09/16/06 pjr: modified init() to avoid unnecessary allocation.
 * 10/21/05 rdb: Removed unused variables.
 * 04/26/05 rdb: added setSteps, moveTo( ISID ) and moveTo( int[] )
 * 04/04/05 pjr: removed refLower and refUpper in favor of refLowerArray() and
 *                  refUpperArray().
 * 04/03/05 rdb: added toOrigin() and toOrigin( isid )
 * 04/03/05 pjr: added refLower and refUpper
 * 02/07/05 rdb: Cleaned up constructors to localize code.
 * 01/28/05 rdb: Changed setSpaceBounds and getSpaceBounds to 
 *                  setBounds and getBounds, so names will be consistent with 
 *                  other similar methods in other classes. Deprecated the old names.
 * 12/14/04 rdb: Restored getAxisOrdering as synonym for getOrdering.
 *               This makes it consistent with other classes's methods.
 *               getOrdering should be deprecated.
 ----------------------------------------------------------------------------
 ***/
public class ISIterator extends IndexSpaceID implements Cloneable, SDBIterator
{
   private AxisOrdering ordering;
   private int[] axes;
   private boolean valid = false;
   private int[] steps;
   private int dim;
   private int[] ubounds;
   private int[] lbounds;

   /** Constructor taking an ISBounds, which specifies the IndexSpace */
   public ISIterator( ISBounds db )
   {
      this( db, null, null );
   }

   /** Constructor taking an ISBounds, which specifies the IndexSpace,
    * and the AxisOrdering which specifies the axis incrementation order 
    */
   public ISIterator( ISBounds db, AxisOrdering ord )
   {
      this( db, ord, null );
   }

   /** Constructor taking an ISBounds, which specifies the IndexSpace 
    *  and the steps array allowing ISIterator increment step other than 1 */
   public ISIterator( ISBounds db, int[] steps )
   {
       this( db, null, steps );
   }

   /** Constructor taking an ISBounds, AxisOrdering and steps array */
   public ISIterator( ISBounds db, AxisOrdering ord, int[] steps )
   {
      this.dim = db.dim();
      if ( ord == null )
         this.ordering = new AxisOrdering( this.dim );
      else
         this.ordering = ord;
      if ( steps != null )
      {
         if ( steps.length != this.dim )
         {
            new SDBError( SDBError.AlwaysThrow, "ISIterator: bad dimensions for steps "
                              + "argument: " + steps.length + " Needed: " + this.dim );           
            this.steps = new int [ this.dim ];
            for ( int i = 0; i < this.dim; i++ )
               this.steps[ i ] = 1;
         }
         else 
         {
            this.steps = new int[steps.length];
            System.arraycopy (steps, 0, this.steps, 0, steps.length);
         }
      }
      else
      {
         this.steps = new int [ this.dim ];
         for ( int i = 0; i < this.dim; i++ )
            this.steps[ i ] = 1;
      }
      this.axes = new int[this.dim];
      this.init( db );
   }

   private void init( ISBounds bounds )
   {
      int  dim = bounds.dim();

      if ( this.ubounds == null || dim != this.dim )
      {
         this.dim = dim;
         this.ubounds = new int[this.dim];
         this.lbounds = new int[this.dim];
         this.axes = new int[this.dim];

         if ( this.ordering == null )
         {
            this.ordering = new AxisOrdering( this.dim );
         }
      }

//      IndexSpaceID upper = bounds.copyUpper();
//      IndexSpaceID lower = bounds.copyLower();

      int [] upper = bounds.getUpperArray();
      int [] lower = bounds.getLowerArray();

      for ( int i = 0; i < this.dim; i++ )
      {
//         this.ubounds[i] = upper.getCoord( i );
//         this.lbounds[i] = lower.getCoord( i );

         this.ubounds[i] = upper[ i ];
         this.lbounds[i] = lower[ i ];



         this.axes[i] = lbounds[i];

         if ( this.steps[i] < 1 )
         {
            this.steps[i] = 1;
            new IllegalArgumentError( SDBError.Recoverable,
                  "Steps array has elements less than one." );
         }
         /***** rdb 4/2005: I thought this should be a check also. However,
          *     it may be necessary to not change it. The default steps defines
          *     the entire rod extent. If you want a subset, the step will be too
          *     large, but then change the subset and you want it to grow. Also
          *     given proper "gap" processing, the gap code may need this extent
          *     here. For now, I'm leaving the check out.
          * 
         else if ( this.steps[i] > this.ubounds[i] - this.lbounds[i] + 1 )
         {
            new SDBError( SDBError.NeverThrow, 
                              "Iterator step too large, reduced to max step. axis: "
                               + i + " old step: " + this.steps[i] +
                               "  new step: " + (this.ubounds[i] - this.lbounds[i] + 1));
            this.steps[i] = this.ubounds[i] - this.lbounds[i] + 1;
         }
         /************************/
      }
      valid = true;
   }

   /**Return an IntegerIndexSpaceID corresponding to the current position.*/
   public IndexSpaceID copyID()
   {
      return new IntegerIndexSpaceID( this.axes );
   }

   /** Return a copy of this ISIterator
    */
   public Object clone()
   {
      return this.copy();
   }

   /** Return a copy of this ISIterator.
    Has been replaced by clone() method.
    @deprecated
    */
   public IndexSpaceID copy()
   {
      ISIterator cpy;
      ISBounds db = new ISBounds( new IntegerIndexSpaceID( this.lbounds ),
                                  new IntegerIndexSpaceID( this.ubounds ) );
      cpy = new ISIterator( db, this.ordering );
      int axes[] = new int[this.axes.length];
      
      for ( int i = 0; i < this.axes.length; i++ )
      {
         axes[i] = this.axes[i];
      }
      cpy.axes = axes;
      cpy.steps = this.steps.clone();
      cpy.valid = this.valid;
      return cpy;
   }

   /**Return description of this ISIterator*/
   public String toString()
   {
      int i;
      String s = new String( "<" + axes[0] );
      for ( i = 1; i < this.axes.length; i++ )
      {
         s = s + "," + axes[i];
      }
      s = s + ">";
      return s;
   }

   /** Return the AxisOrdering for this iterator. 
    */
   public final AxisOrdering getAxisOrdering()
   {
      return this.ordering;
   }

   /** Return the AxisOrdering for this iterator. 
    *  @deprecated
    */
   public final AxisOrdering getOrdering()
   {
      return this.ordering;
   }

   /** Return the number of elements that lie along the specified
    axis in the iterator's bounds.*/
   public final int getAxisLength( int axis )
   {
      try
      {
         return this.ubounds[axis] - this.lbounds[axis] + 1;

      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Argument specifies an illegal axis." );
         return -1;
      }
   }

   /** Returns an array of ints representing the dimensions of 
    the space being iterated over.
    */
   public final int[] getAxisLengthArray()
   {
      int i, d = this.lbounds.length;
      int[] axisLengths = new int[d];

      for ( i = 0; i < d; i++ )
      {
         axisLengths[i] = this.ubounds[i] - this.lbounds[i] + 1;
      }
      return axisLengths;
   }

   /** Returns an IndexSpaceID representing the dimensions of 
    *  the space being iterated over.
    */
   public final IndexSpaceID getAxisLengths()
   {
      int i, d = this.lbounds.length;
      int[] axisLengths = new int[d];

      for ( i = 0; i < d; i++ )
      {
         axisLengths[i] = this.ubounds[i] - this.lbounds[i] + 1;
      }

      return new IntegerIndexSpaceID( axisLengths );
   }

   /** Map the current position of this iterator to an equivalent 
    * IndexSpaceID relative to the lower bound of this iterator 
    */
   public IndexSpaceID relativeLocation()
   {
      return relativeLocation( null );
   }

   /** Map the current position of this iterator to an equivalent 
    * IndexSpaceID relative to the lower bound of this iterator.
    * If the argument is not null, update the argument's position.
    * If the argument is null, create a new IndexSpaceID. 
    */
   public IndexSpaceID relativeLocation( IndexSpaceID isid )
   {
      if ( isid == null )
         isid = new IntegerIndexSpaceID( new int[ this.dim ] );
      int[] coord = isid.refCoords();
      
      for ( int i = 0; i < this.dim; i++ )
         coord[i] = this.refCoords()[i] - this.refLowerArray()[i];
      
      return isid;      
   }

   /** Reposition the bounds of this iterator so that its lower bound
    * is at the position specified by the argument. Although it only
    * makes sense to do this prior to an iteration, the current position
    * of the iterator is also updated based on the relative change of
    * the lower bound.
    */
   public void moveTo( IndexSpaceID newOrigin )
   {
      if ( newOrigin != null && newOrigin.dim() == this.dim() )
      {
         for ( int i = 0; i < this.dim; i++ )
         {
            int delta = newOrigin.getCoord( i ) - lbounds[i];
            ubounds[i] += delta;
            axes[i]    += delta;
            lbounds[i] += delta; // lbounds[i] = isid.getCoord( i );
         }       
      }
      else
         new SDBError( SDBError.AlwaysThrow, "Invalid argument to ISIterator.moveTo ");
   }

   /** Reposition the bounds of this iterator so that its lower bound
    * is at the position specified by the argument. Although it only
    * makes sense to do this prior to an iteration, the current position
    * of the iterator is also updated based on the relative change of
    * the lower bound.
    */
   public void moveTo( int [] newOrigin )
   {
      if ( newOrigin != null && newOrigin.length == this.dim() )
      {
         for ( int i = 0; i < this.dim; i++ )
         {
            int delta = newOrigin[i] - lbounds[i];
            ubounds[i] += delta;
            axes[i]    += delta;
            lbounds[i] += delta; // lbounds[i] = isid.getCoord( i );
         }       
      }
      else
         new SDBError( SDBError.AlwaysThrow, "Invalid argument to ISIterator.moveTo ");
   }

   /** Change the IndexSpace to which this ISIterator is bound 
    */
   public void setBounds( ISBounds bound )
   {
      this.init( bound );
   }

   /** Define a new steps array for the iterator 
    */
   public void setSteps( int[] newSteps )
   {
      this.steps = newSteps.clone();
      this.init( this.getBounds() ); // check validity of newSteps )
   }

   /**Change the IndexSpace to which this ISIterator is bound 
    * @deprecated   Use setBounds( ISBounds )
    */
   public void setSpaceBounds( ISBounds bound )
   {
      this.init( bound );
   }

   /**Return the ISBounds over which this ISIterator is iterating. 
    */
   public ISBounds getBounds()
   {
      return new ISBounds( this.lbounds, this.ubounds );
   }

   /**Return the ISBounds over which this ISIterator is iterating. 
    * @deprecated  
    * Use getBounds() 
    */
   public ISBounds getSpaceBounds()
   {

      return new ISBounds( this.lbounds, this.ubounds );
   }
   
  
   /** Return a reference to the lower bounds of the iteration space.
    *
    */
   public int [] refLowerArray(){
      
      return this.lbounds;
   }
   
   /** Return a reference to the lower bounds of the iteration space.
    *  
    */
   public int [] refUpperArray(){
      
      return this.lbounds;
   }
  
   /**Initilize this ISiterator*/
   public void init()
   {
      int i;

      for ( i = 0; i < this.dim; i++ )
      {
         this.axes[i] = lbounds[i];
         //            if(this.steps[i]<1){
         //                this.steps[i] = 1;
         //                new IllegalArgumentError(SDBError.Fatal,"Steps array has elements less than one.");
         //            }
      }
      valid = true;
   }

   /** Set the iterator value to the last valid value in the iteration.*/
   public void end()
   {

      int i;

      for ( i = 0; i < this.dim; i++ )
      {
         this.axes[i] = ubounds[i];
      }
      valid = true;
   }

   private void resetAxis( int axisNumber )
   {

      this.setCoord( axisNumber, this.lbounds[axisNumber] );
   }

   /** Get the value corresponding to the specified axis.*/
   public int getCoord( int axisNumber )
   {

      try
      {
         return axes[axisNumber];

      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new DimensionalityError( SDBError.Fatal,
               "Axis argument is out of bounds." );
         return 0;
      }

   }

   /** Get an array of axis values. This method returns a reference to the axes for this ISID, so we
    only allow package access.*/

   int[] refCoords()
   {

      return axes;
   }

   /** Directly set this ISIterator's value to the argument array. No copy is made, but a validity
    check is performed.
    */
   void assign( int[] coords )
   {

      int[] oldAxes = this.axes;
      this.axes = coords;

      try
      {
         int i, d = this.dim;
         this.valid = true;

         for ( i = 0; i < d && this.valid; i++ )
         {

            this.valid = ( this.axes[i] >= this.lbounds[i] )
                  && ( this.axes[i] <= this.ubounds[i] );
         }

      } catch ( ArrayIndexOutOfBoundsException e )
      {

         this.axes = oldAxes;
         new DimensionalityError( SDBError.Recoverable,
               "Array argument has wrong dimensionality." );
      }

   }

   /** Fill the array argument with the coordinate values for this ISID. For convenience,
    a reference to the argument array is returned. */
   public int[] getCoords( int[] coords )
   {

      int i, d = this.axes.length;

      try
      {

         for ( i = 0; i < d; i++ )
         {
            coords[i] = axes[i];
         }

      } catch ( ArrayIndexOutOfBoundsException e )
      {

         new DimensionalityError( SDBError.Recoverable,
               "Array argument has wrong dimensionality." );
      }
      return coords;
   }

   private boolean incAxis( int axisNumber, int value )
   {

      if ( this.ubounds[axisNumber] < value || this.lbounds[axisNumber] > value )
      {
         return false;
      }

      axes[axisNumber] = value;
      return true;
   }

   /** Set the specified axis to the specified value.*/
   public void setCoord( int axisNumber, int value )
   {

      if ( this.lbounds[axisNumber] > value || this.ubounds[axisNumber] < value )
      {
         new BoundsError( SDBError.Recoverable, "Argument is out of bounds." );
         return;
      }

      axes[axisNumber] = value;
   }

   /** Returns the dimensionality of this ISIterator. */
   public int dim()
   {
      return this.dim;
   }

   /**Return true if the current position is a valid position in the IndexSpace.*/
   public boolean valid()
   {
      return this.valid;
   }

   /**Increment the ISIterator to the next position.*/
   public boolean next()
   {
      int value = 0, axis = 0;

      if ( this.valid == false )
         return false;

      for ( int i = 0; i < this.dim; i++ )
      {
         axis = this.ordering.getAxis( this.dim - i - 1 );
         //value=this.getCoord(axis)+steps[axis];
         value = this.axes[axis] + steps[axis];
         if ( this.incAxis( axis, value ) )
            return true;
         else
         {
            if ( i == this.dim - 1 )
            {
               this.valid = false;
               for ( int j = 0; j < this.dim; j++ )
                  this.axes[j] = this.ubounds[j];
               return false;
            }
            else
               this.resetAxis( axis );
         }
      }
      return false;
   }

   /**Return true if this ISIterator can be incremented.*/
   public boolean hasMoreElements()
   {
      int i, axis0, axis;
      boolean equal = true;
      for ( i = 1; i < this.dim(); i++ )
      {
         axis = this.ordering.getAxis( this.dim() - i - 1 );
         if ( this.getCoord( axis ) != this.ubounds[axis] )
         {
            equal = false;
            break;
         }
      }
      axis0 = this.ordering.getAxis( this.dim() - 1 );
      if ( ( equal )
            && ( this.getCoord( axis0 ) + steps[axis0] > ubounds[axis0] ) )
         return false;

      for ( i = 0; i < this.dim; i++ )
      {
         if ( ( this.getCoord( i ) > ubounds[i] )
               || ( this.getCoord( i ) < lbounds[i] ) )
         {
            return false;
         }
      }

      return true;
   }

   /** Returns true if this IndexSpaceID is contained by the 
    ISBounds parameter, or false otherwise. 
    
    If the ISBounds has a diifferent dimensionality than
    this IndexSpaceID, then a DimensionalityException is 
    thrown.
    */
   public boolean containedBy( ISBounds db )
   {

      IndexSpaceID id0 = db.copyLower();
      IndexSpaceID id1 = db.copyUpper();

      return this.valid && id0.lessThanOrEqual( this )
            && id1.greaterThanOrEqual( this );
   }

   static public void main( String[] args )
   {

      IntegerIndexSpaceID lower1 = new IntegerIndexSpaceID( 0, 0 );
      IntegerIndexSpaceID upper1 = new IntegerIndexSpaceID( 4, 3 );
      //IntegerIndexSpaceID lower2 = new IntegerIndexSpaceID( 0, 0 );
      //IntegerIndexSpaceID upper2 = new IntegerIndexSpaceID( 1, 2 );

      ISBounds ISBounds1;
      //ISBounds ISBounds2;
      //int isorderingarray[] = { 1, 0 };
      int dsorderingarray[] = { 0, 1 };
      //AxisOrdering isordering = new AxisOrdering( isorderingarray );
      AxisOrdering dsordering = new AxisOrdering( dsorderingarray );
      ISBounds1 = new ISBounds( lower1, upper1 );
      //ISBounds2 = new ISBounds( lower2, upper2 );
      int steps[] = { 1, 1 };
      ISIterator iter1 = new ISIterator( ISBounds1, dsordering, steps );
      /*        
       for (int i=0;i<30;i++){
       System.out.println(iter1+"  ");
       iter1.next();
       
       }
       */
      for ( iter1.init(); iter1.valid(); iter1.next() )
      {

         System.out.println( iter1 );

      }

      System.out.println( "\nAfter loop:" );
      System.out.println( "iter1=" + iter1 );

      /*
       iter1.init();
       System.out.println(iter1);
       while (iter1.valid()){
       
       System.out.println(iter1);
       iter1.next();
       }
       //        for(iter1.init();iter1.valid();iter1.next()){
       //            System.out.println(iter1);
       //        }

       */
      System.out.println( "end" );

   } //end main()

}
