//
//  ISPartitioning.java
//  Unstructured
//
//  Created by Philip Rhodes on Tue Oct 01 2002.
//  Copyright (c) 2002 University of New Hampshire. All rights reserved.
//
//------------------- Edit history as of 12/14/04 ----------------------------
// 02/15/12 rdb: Moved volume() method from RegularISPartitioning to here
//                 (the parent) and simplified its implementation.
//               Moved partitionVolume() from RegularISPartitioning to here
// 02/14/12 rdb: Added a new instance variable:
//                 int[] partitionShape
//               This is created regardless of whether the ISPartitioning 
//               is created by specifying the partition shape or the 
//               partitioning shape. It might be argued that the partition shape
//               might only be known to a RegularISPartitioning object, but
//               the constructor that allowed this already existed.               
// 02/12/12 rdb: Added additional instance variables, so all data needed to
//               work with partitionings either from the partition shape or
//               partitioningBounds is available all the time, regardless of
//               how it was constructed. This is mostly useful in the child 
//               classes (now only RegularISPartitioning). This moved some
//               code from RegularISParitioning; corrected calculation of
//               partition shape so that it is correct if partitioning does
//               not partition space with all equal size partitions.
//          rdb: I renamed instance variables that represent the "sizes" of
//               blocks from "Dimensions" variables to "Shape" variables. 
//               The term "dimension" is so over-loaded, I like to use it as
//               little as possible.
//               Also, created synonyms for the public methods that return
//               "dimensions" information so the name component "Dimensions" can
//               be replaced by "Shape":
//                      getPartitionShape
//                      getPartitioningShape
// 02/01/12 rdb: added protected instance variables defining the partitioningBounds
//               and partitioningVolume; also made other instance variables
//               protected instead of package.
// 06/04/08 rdb: added some comments and minor reformatting
// 10/21/05 rdb: Removed unused variables.
//----------------------------------------------------------------------------

package edu.unh.sdb.datasource; 

/** 
 * An ISPartitioning provides a mapping between an N-Dimensional index space
 *  and an integer partitioning Space of the same dimensionality. 
 */

public abstract class ISPartitioning 
{
	protected int      dimensionality;
   
   protected ISBounds partitioningBounds;
   protected int      partitioningVolume; 
   
	protected int[]    partitioningLowerBounds;
	protected int[]    partitioningUpperBounds;
	protected int[]    partitioningShape;
   
   protected int[]    partitionShape; // shape of each partition

   protected ISBounds spaceBounds;
	protected int[]    spaceLowerBounds;
	protected int[]	 spaceUpperBounds;
   protected int[]    spaceShape;
	
   /**
    * Create an ISPartitioning from a Bounds describing the partitioning and a 
    * Bounds describing the data space.
    *  
    * @param partitioningBounds
    * @param spaceBounds
    */
   public ISPartitioning( ISBounds partitioningBnds, ISBounds spaceBnds )
   {
      dimensionality = spaceBnds.dim();
      setSpaceBounds( spaceBnds );
      setPartitioning( partitioningBnds );
   }
   /** 
    * The first argument specifies the shape of each partition, not the 
    *    dimensions of the partitioning space.
    * The second argument specifies the bounds of the data space. 
    */
   public ISPartitioning( IndexSpaceID partitionShape, ISBounds spaceBnds )
   {
      dimensionality = spaceBnds.dim();
      setSpaceBounds( spaceBnds );
      setPartitionShape( partitionShape );      
    }
    /**
     * Create an ISPartitioning
     * 
     * @param partitioningDims - the number of partitions in each dimension.
     *                                 So [8, 6, 5] means there are 8x6x5 partitions
     * @param spaceBounds - the data space bounds over which the partitioning occurs.
     */
    public ISPartitioning( int[] partitioningDims, ISBounds spaceBnds )
    {
      dimensionality= spaceBnds.dim();
      setSpaceBounds( spaceBnds );
      setPartitioning( new ISBounds( partitioningDims ));
    }

   /**
    * Define the space bounds for the partitioning.
    */
   protected void setSpaceBounds( ISBounds spaceBnds )
   {
      spaceBounds = spaceBnds.copy();
      spaceShape = spaceBounds.getDimensionsArray();
      spaceLowerBounds = spaceBounds.copyLowerArray();
      spaceUpperBounds = spaceBounds.copyUpperArray();
   }

   /**
    * Define the partitioning from an ISBounds.
    */
   protected void setPartitioning( ISBounds partitioningBnds )
   {
      partitioningBounds      = partitioningBnds.copy();
      partitioningVolume      = partitioningBounds.volume();
      partitioningLowerBounds = partitioningBounds.getLowerArray();
      partitioningUpperBounds = partitioningBounds.getUpperArray();
      partitioningShape       = partitioningBounds.getDimensionsArray();
   }
   /**
    * Define the partitioning by defining the shape of each partition
    */
   protected void setPartitionShape( IndexSpaceID shapeOfEachPartition )
   {
      partitioningLowerBounds = new int[ dimensionality ];
      partitioningUpperBounds = new int[ dimensionality ];
      partitioningShape       = new int[ dimensionality ];
      
      partitionShape          = shapeOfEachPartition.getCoords();

      for ( int d = 0; d < dimensionality; d++ )
      {
         partitioningShape[d] = ( spaceShape[d] - 1 ) / partitionShape[d] + 1;
         int p = (int) Math.ceil((float) spaceShape[d] / partitionShape[d] );
         if ( p != partitioningShape[d] )
            System.err.println( "--- ISPing old/new discrepancy for dim "
                                + d + "   " + p + " vs " + partitioningShape[d]);
         // old: partitioningShape[d] = (int) Math.ceil((float) spaceShape[d] / partitionShape[d] );
         /****************/
      }

      partitioningBounds = new ISBounds( partitioningShape );
      partitioningLowerBounds = partitioningBounds.getLowerArray();
      partitioningUpperBounds = partitioningBounds.getUpperArray();
      partitioningVolume = partitioningBounds.volume();
   }

     /**
     * @return the dimensionality of the partitioning
     */
    public final int dim()
    {   
        return this.dimensionality;
    }
    
    /**
     * @return the Bounds of the partitioning
     */
    public ISBounds getPartitioningBounds()
    {        
        return this.partitioningBounds;
    }

   /** 
    * Return true if the argument represents a partition coordinate that is 
    *  within the bounds of this partitioning.  This is a partitioning space operation,
    *  not an index space operation.
    */
   public final boolean inBounds( IndexSpaceID pid )
   {   
      return inBounds( pid.refCoords() );
   }

   /** 
    * Return true if the argument represents a partition coordinate that is 
    *  within the bounds of this partitioning.  This is a partitioning space operation,
    *  not an index space operation.
    */
   public final boolean inBounds( int [] coords )
   {
      for ( int i=0; i< this.dimensionality; i++ )
      {       
         if ( coords[i] < this.partitioningLowerBounds[i] 
             || coords[i] > this.partitioningUpperBounds[i])
         {
            return false;
         }
      }
      return true;
   }

    /**
     * Return the dimensions of the partitioning and fill in those dimensions
     * in the passed argument.
     * 
     * @param dimensions -- if not null, fill in with partitioning dimensions
     * @return the partitioning dimensions
     */
    public int [] getPartitioningDimensions( int [] dimensions )
    {
       return getPartitioningShape( dimensions );
    }
    /**
     * Return the dimensions of the partitioning 
     * 
     * @return the partitioning dimensions
     */
    public int [] getPartitioningShape()
    {
       return getPartitioningShape( null );
    }
    /**
     * Return the dimensions of the partitioning and fill in those dimensions
     * in the passed argument.
     * 
     * @param dimensions -- if not null, fill in with partitioning dimensions
     * @return the partitioning dimensions
     */
    public int [] getPartitioningShape( int [] dimensions )
    {
      if ( dimensions == null )
      {        
         dimensions = new int[ dimensionality ];
      }
      
      return this.getPartitioningBounds().getDimensions( dimensions );
    }
 
    /**
     * Return the bounds of the data space underlying the partitioning
     * @return data space bounds
     */
    public ISBounds getSpaceBounds()
    {    
         return new ISBounds( this.spaceLowerBounds, this.spaceUpperBounds );
    }
    
    /**
     * Create and return an iterator over the partitions that cover the 
     * data space bounds defined by the parameter.
     * 
     * @param subSpace - region of data space
     * @return - bounds iterator over the partitions
     */
    public abstract ISBoundsIterator getPartitionIterator( ISBounds subSpace );
	
    /**
     * Return in IndexSpace iterator over the partitions
     * @return index space iterator
     */
	 public ISIterator getISIterator()
	 {	
		return new ISIterator( this.getPartitioningBounds() );
	 }	
    
	/**
	 * Map a position in the data space to the partition that contains it.
	 * 
	 * @param dataISID - ISID in data space
	 * @return an ISID in the partitioning space
	 */
    public abstract IndexSpaceID map( IndexSpaceID dataISID );

    /**
     * Map a bounds in the data space to the bounds in the partitioning space
     * that contains the data bounds.
     * 
     * @param dataBounds - bounds in data space
     * @return  - bounds in partitioning space
     */
    public abstract ISBounds map( ISBounds dataBounds );
       
    /** 
     * Sets the argument ISBounds to the bounds corresponding to the partition ID. 
     */
    public abstract ISBounds getIndexedBounds( ISBounds     bounds, 
    		                                   IndexSpaceID partitioningIndex );
    
    /** Return a String representing the object. */
    public String toString()
    {     
        return this.toString( "" );
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent)
    {    
        StringBuffer result = new StringBuffer( indent + "dimensionality=" 
        		                                  + this.dimensionality + "\n" );
        
        result.append( indent + "partitioningbounds=" );
        result.append( this.getPartitioningBounds() + "\n" );
        result.append( indent + "spaceBounds=" );
        result.append( this.getSpaceBounds() );
        
        return new String( result );
    }
    
   /** 
    * Return the total number of partitions.
    */
   public long volume()
   {
      return partitioningBounds.volume();
      /*************** rdb: simplified old code below to above line
      long volume = 1;
      
      for(  int i=0; i< this.dimensionality; i++ )
      {         
      volume *= (int) Math.ceil(( float)this.spaceShape[i] / this.partitionShape[i]);
      }   
      return volume;
      /******************/
   }
   /** 
    * Return the volume of a single complete partition. Partial partitions
    * will not have this volume.
    */
   public long partitionVolume()
   { 
      long volume = 1;
      
      for(  int i = 0; i < this.dimensionality; i++ )
      {        
         volume *= this.partitionShape[i];
      }     
      return volume;
   }   
}
