/* 

    IndexSpaceID.java
    Copyright 2001 University of New Hampshire
    Durham, NH 03824
    
    ----------------- Edit history as of 03/20/05 ------------------------------
    01/02/06 dlipsa: change assign(IndexSpaceID id) to copy the information from id,
                     instead of pointing to its internal data structure.
    12/22/05 dlipsa: added methods: volume, linearIndex
    10/21/05 rdb: Deleted unused variables to avoid compiler warnings
    03/20/05 rdb: made int[] getCoords() public and provided default copy impl.
                  defined int[] refCoords() for package access reference to data.
                  
    3/30/05: pjr: added new add and subtract methods.
*/

package edu.unh.sdb.datasource;
import edu.unh.sdb.common.*;

/**
 * The IndexSpaceID class represents points in a DataSource index space
 */
public abstract class IndexSpaceID extends SpaceID{ 


    
    /** Makes a copy of an IndexSpaceID object that is guaranteed
      *  not to copy iterator information in the case that the
      *  object is actually an iterator.
      */
    public abstract IndexSpaceID copyID();
    
    
    /**
     * Sets (copies) 'this' object from the 'id' object.
     */
    final IndexSpaceID set (IndexSpaceID id)
    {
        int[] coords = id.getCoords(new int[ this.dim()]);
        
        this.assign (coords);
        return this;
    }
    
    /** Get the value corresponding to the specified axis. */
    public abstract int getCoord(int axisNumber);
       
    /** Fill the array argument with the coordinate values for this ISID. For convenience,
      *  a reference to the argument array is returned. 
      */
    public abstract int[] getCoords(int[] coords);
    
    /** Return the integer array representing this ISID. If possible, this
     *  will be a reference to a data member; hence, the package access.
     * @return
     */  
    abstract int[] refCoords();
    
    /** Return an array with the coordinate values for this ISID. This
     *  copies the values. 
     */
    public int[] getCoords( )
    {
       return getCoords( new int[ this.dim() ]);
    }
      
    
    /** Set this IndexSpaceID to the coordinate values contained in the array. 
      * The IndexSpaceID is allowed to use the array directly instead of copying
      * its contents to an internal array.
      *  Package access only.
      */
    abstract void assign(int[] coords);
    
//    /** Set the specified axis to the specified coordinate value. */
//    abstract public void  setCoord(int axisNumber,int value);
    

    /** Set the axes of this id to the axes of the parameter id.
      * The assign method may elect to refer to the data in
      * the parameter id, and not make a copy of this information.
      * 
      * Package access only.
      */       
    final IndexSpaceID assign(IndexSpaceID id){
        int [] idCoords=id.refCoords();

        this.assign(idCoords);
        return this;
    }
    
    
    
    /** For each axis of this IndexSpaceID, add the value
      *  of the corresponding axis of the parameter. 
      *
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      */
    public void add(IndexSpaceID id){
    

        int [] idCoords=id.refCoords();
        int [] coords=this.refCoords();
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                coords[i]+=idCoords[i];
            }
            
            this.assign(coords);
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable,"Argument ISID has dimensionality that's too low.");
        }
    }
    
    /*
     * Returns the volume (total number of elements) of the hypercube whose 
     * dimensions are specified by this IndexSpaceID.
     */
    public long volume()
    {
        int[] coords = refCoords();
        long volume = 1;

        for ( int i=0; i < coords.length; i++)
           volume *= coords[i];
        return volume;
    }
    
    /**
     * Sets this to the IndexSpaceID representing the dimensions in units of the 
     * hypercube whose dimensions in points are specified
     * by this IndexSpaceID.
     * A reference to 'this' is returned.
     */
    public IndexSpaceID setDimensionsInUnits (IndexSpaceID unit)
    {
        int[] coords = refCoords ();
        int[] unitCoords = unit.refCoords ();
        for (int i = 0; i < coords.length; i++)
            coords[i] = (int)Math.ceil ((double)coords[i]/unitCoords[i]);        
        this.assign (coords);
        return this;
    }           

    /** For the specified axis of this IndexSpaceID, add
      *  specified value. 
      *
      * If the specified axis is not less than the dimensionality
      * of this IndexSpaceID, a dimensionality error is created.
      */
    public void add(int axis, int value){

        int [] coords=this.getCoords();
        
        try{
        
            coords[axis]+=value;
            this.assign(coords);
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Axis argument is too large.");
        }
    }

    
    /** For the specified axis of this IndexSpaceID, subtract
      *  specified value. 
      *
      * If the specified axis is not less than the dimensionality
      * of this IndexSpaceID, a dimensionality error is created.
      */
    public void subtract(int axis, int value){

        int [] coords=this.getCoords();
        
        try{
        
            coords[axis]-=value;
            this.assign(coords);
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Axis argument is too large.");
        }
    }
    
    

    /** Arithmetically negate this IndexSpaceID's axis values. */
    public final void negate(){
    
        int [] coords=this.refCoords();
        int i, d=coords.length;       
                                                               
        for(i=0;i<d;i++){
            coords[i]= - coords[i];   
        }
        
        this.assign(coords);
    }





    /** For each axis of this IndexSpaceID, subtract the value
      *  of the corresponding axis of the parameter. 
      *  
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      */
    public IndexSpaceID subtract(IndexSpaceID id){

        
        int [] idCoords=id.refCoords();
        int [] coords=this.refCoords();
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                coords[i]-=idCoords[i];
            }
            
            this.assign(coords);
            return this;
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, 
                "Argument ISID has dimensionality that's too low.");
            return this;
        }
        
    }
    
    /** For each axis of this IndexSpaceID, subtract the value
      *  of the corresponding axis of the second argument, and 
      *  assign the result to the first argument.
      *  
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      */
    public void subtract(IndexSpaceID result, IndexSpaceID id){

        
        int [] idCoords=id.getCoords();
        int [] coords=this.getCoords();
        int [] resultCoords = result.getCoords();
        
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                resultCoords[i]= coords[i] - idCoords[i];
            }
            
            result.assign(resultCoords);
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable,"Axis argument is too large.");
        }
        
    }


    /** Scale this IndexSpaceID's axis values by the given integer. 
     *  A reference to 'this' is returned as a convenience.*/
    public final IndexSpaceID scale( int m){
    
        int [] coords=this.refCoords();
        int i, d=coords.length;       
                                                               
        for(i=0;i<d;i++){
            coords[i] *= m;   
        }
        
        this.assign(coords);
        
        return this;
    }
    
    /**
     * Pairwise multiply this IndexSpaceID with 'other' and store the result in this.
     * A reference to 'this' is returned.
     */
    public final IndexSpaceID multiply (IndexSpaceID other)
    {
        int[] coords = this.refCoords();
        int[] otherCoords = other.refCoords ();
        for (int i = 0; i < coords.length; i++)
            coords[i] *= otherCoords[i];
        assign (coords);
        return this;
    }

    
    /**
     * Pairwise divide this IndexSpaceID with 'other' and store the result 
     * in this. A reference to 'this' is returned.
     */
    public final IndexSpaceID divide (IndexSpaceID other)
    {
        int[] coords = this.refCoords();
        int[] otherCoords = other.refCoords ();
        for (int i = 0; i < coords.length; i++)
            coords[i] /= otherCoords[i];
        assign (coords);
        return this;
    }

    
    /**
     * Pairwise shift this IndexSpaceID with 'other' and store the result 
     * in this. A reference to 'this' is returned.
     */
    public final IndexSpaceID shift (IndexSpaceID other)
    {
        int[] coords = this.refCoords();
        int[] otherCoords = other.refCoords ();
        for (int i = 0; i < coords.length; i++)
            coords[i] = (coords[i] >> otherCoords[i]);
        assign (coords);
        return this;
    }
    
    
    /**
     * Calculate pairwise mod of this IndexSpaceID with 'other' and store the 
     * result in 'this'. A reference to 'this' is returned.
     */
    public final IndexSpaceID mod (IndexSpaceID other)
    {
        int[] coords = this.refCoords();
        int[] otherCoords = other.refCoords ();
        for (int i = 0; i < coords.length; i++)
            coords[i] %= otherCoords[i];
        assign (coords);
        return this;
    }
    
    
    
    

    /** Scale this IndexSpaceID's axis values by the given double. 
     *  Axis values are rounded to the nearest integer.
     *  A reference to 'this' is returned as a convenience.*/
    public final IndexSpaceID scale( double m){
    
        int [] coords=this.refCoords();
        int i, d=coords.length;       
                                                               
        for(i=0;i<d;i++){
            coords[i] = (int) (coords[i] * m + 0.5);   
        }
        
        this.assign(coords);
        
        return this;
    }



	/** Returns a new IndexSpaceID which is the projection of this IndexSpaceID onto a 
	 * plane orthogonal to the specified axis. The dimensionality of the result 
	 * will therefore always be one less than the dimensionality of this object, 
	 * as long as the axis is valid for this IndexSpaceID.
	 *
	 * If the specified axis is not valid, an error is created.
	 */
	public IndexSpaceID project( int axis )
	{
        int [] oldCoords = this.getCoords();
        int d = oldCoords.length;
		
		if ( axis >= d) 
		{
			new IllegalArgumentError( SDBError.Recoverable, 
			      "axis of projection is beyond the dimensionality of this IndexSpaceID.");
			return this;
		}
		
        int [] newCoords=new int[d-1];
				
		int j=0;
            
		for ( int i=0; i < d; i++)
		{
			if ( i != axis)
			{
				newCoords[j] = oldCoords[i];
				j++;
			}
		}
        
		return new IntegerIndexSpaceID(newCoords);
	}


    /** Returns true if this IndexSpaceID is contained by the 
      *  ISBounds parameter, or false otherwise. 
      *  
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      */
    public boolean containedBy(ISBounds b){

        int [] coords=this.refCoords();
        int [] blCoords = b.getLowerArray();
        int [] buCoords = b.getUpperArray();
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                
                if( blCoords[i] > coords [i]  ||  buCoords[i] < coords[i]){
                    return false;
                }
            }
            
            return true;
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable,"Argument Bounds has dimensionality that's too low.");
            return false;
        }
    }
    

    
    /** Return true if the argument is an IndexSpaceID representing the same
      *  coordinates as this object, false otherwise.
      */
    public boolean equals(Object obj){
    
        try{
        
            IndexSpaceID id=(IndexSpaceID) obj;

            int [] idAxes = id.refCoords();
            int [] axes = this.refCoords();
            int i, d=idAxes.length;   
                     
            if (d != axes.length){
                return false;
            }
            
            for(i=0;i<d;i++){
                if(axes[i] != idAxes[i]){
                    return false;
                }
            }
    
            return true;
        } catch (ClassCastException e){
        
            return false;
        }
    }
    

    

    /** A helper function that returns true if all the coordinates of this
      *  id are less than or equal to the corresponding coordinates of the parameter id. If any
      *  coordinate is greater than the parameter coordinate, then false is
      *  returned. 
      *  
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      *
      */
    public boolean lessThanOrEqual(IndexSpaceID id){
    
        int [] coords=this.refCoords();
        int [] idCoords=id.refCoords();
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                
                if( coords [i] > idCoords[i] ){
                    return false;
                }
            }
            
            return true;
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable,"Argument Bounds has dimensionality that's too low.");
            return false;
        }
    }
    
    /** A helper function that returns true if all the coordinates of this
      *  id are greater than or equal to the corresponding coordinates of the parameter id. If any
      *  coordinate is less than the parameter coordinate, then false is
      *  returned.         
      *  
      *  If the dimensionality of the parameter is not
      *  at least as large as the dimensionality of this object, 
      *  then a DimensionalityError is created.
      */
    public boolean greaterThanOrEqual(IndexSpaceID id){
    
        int [] coords=this.refCoords();
        int [] idCoords=id.refCoords();
        int i, d=coords.length;       
        
        try{
        
            for(i=0;i<d;i++){
                
                if( coords [i] < idCoords[i] ){
                    return false;
                }
            }
            
            return true;
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable,"Argument Bounds has dimensionality that's too low.");
            return false;
        }
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i,d=this.dim();
        String result=new String(indent+"<IndexSpaceID (");
        
        for(i=0;i<d-1;i++){
            result+=this.getCoord(i)+",";
        }
        
        result+=this.getCoord(i)+")>";
        
        return result;
    }    
   
    
    
    public static void main(String[] args){
    
        PackedIndexSpaceID dsid1=new PackedIndexSpaceID(3);
        PackedIndexSpaceID dsid2=new PackedIndexSpaceID(-2);
        PackedIndexSpaceID dsid2a=new PackedIndexSpaceID(4);
        
    
        PackedIndexSpaceID dsid3=new PackedIndexSpaceID(4,5);
        PackedIndexSpaceID dsid4=new PackedIndexSpaceID(new int [] {-3,-7});
        PackedIndexSpaceID dsid5=new PackedIndexSpaceID(6,7);
                
    
        PackedIndexSpaceID dsid6=new PackedIndexSpaceID(1,1,3);
        PackedIndexSpaceID dsid7=new PackedIndexSpaceID(new int [] {4,2,7});
        PackedIndexSpaceID dsid8=new PackedIndexSpaceID(new int [] {-1,-2,-3});

        
        IndexSpaceID result;
        
        
        System.out.println("***********************\n***IndexSpaceID Tests***\n***********************");
        
        System.out.println("\n***IndexSpaceID Constructor Tests***");
        
        System.out.println(" 1D IndexSpaceID: Should be (3)-->"+dsid1);
        System.out.println(" 1D IndexSpaceID, Negative axis values: Should be (-2)-->"+dsid2);
        
        System.out.println(" 2D IndexSpaceID: Should be (4,5)-->"+dsid3);
        System.out.println(" 2D IndexSpaceID: Should be (-3,-7)-->"+dsid4);
        
        System.out.println(" 3D IndexSpaceID: Should be (1,1,3)-->"+dsid6);
        System.out.println(" 3D IndexSpaceID: Should be (4,2,7)-->"+dsid7);
        System.out.println(" 3D IndexSpaceID: Should be (-1,-2,-3)-->"+dsid8);
        
        System.out.println();
        System.out.println("\n***IndexSpaceID getCoord Test***");
        
        System.out.println(" 1D IndexSpaceID: getCoord(0) Should be 3-->"+dsid1.getCoord(0));
        System.out.println(" 1D IndexSpaceID, getCoord(0) Should be -2-->"+dsid2.getCoord(0));

        System.out.println(" 2D IndexSpaceID: getCoord(0) Should be 4-->"+dsid3.getCoord(0));
        System.out.println(" 2D IndexSpaceID: getCoord(1) Should be 5-->"+dsid3.getCoord(1));
        
        System.out.println(" 2D IndexSpaceID: getCoord(0) Should be -3-->"+dsid4.getCoord(0));
        System.out.println(" 2D IndexSpaceID: getCoord(0) Should be -7-->"+dsid4.getCoord(1));
        
        
        System.out.println(" 3D IndexSpaceID: getCoord(0) Should be 4-->"+dsid7.getCoord(0));
        System.out.println(" 3D IndexSpaceID: getCoord(1) Should be 2-->"+dsid7.getCoord(1));
        System.out.println(" 3D IndexSpaceID: getCoord(2) Should be 7-->"+dsid7.getCoord(2));
        
        System.out.println(" 3D IndexSpaceID: getCoord(0)     Should be -1-->"+dsid8.getCoord(0));
        System.out.println(" 3D IndexSpaceID: getCoord(1) Should be -2-->"+dsid8.getCoord(1));
        System.out.println(" 3D IndexSpaceID: getCoord(2) Should be -3-->"+dsid8.getCoord(2));
        
        System.out.println();
        System.out.println("\n***IndexSpaceID setCoord Test***");
        
        dsid1.setCoord(0,14);
        System.out.println(" 1D IndexSpaceID: setCoord(0) Should be 14-->"+dsid1.getCoord(0));
        dsid1.setCoord(0,-5);
        System.out.println(" 1D IndexSpaceID, setCoord(0) Should be -5-->"+dsid1.getCoord(0));


        dsid3.setCoord(0,8);
        System.out.println(" 2D IndexSpaceID: setCoord(0) Should be 8-->"+dsid3.getCoord(0));
        dsid3.setCoord(1,17);
        System.out.println(" 2D IndexSpaceID: setCoord(1) Should be 17-->"+dsid3.getCoord(1));
        
        
        dsid4.setCoord(0,-11);
        System.out.println(" 2D IndexSpaceID: setCoord(0) Should be -11-->"+dsid4.getCoord(0));
        dsid4.setCoord(1,-54);
        System.out.println(" 2D IndexSpaceID: setCoord(1) Should be -54-->"+dsid4.getCoord(1));
        System.out.println(" 2D IndexSpaceID: (Negative Overwrite) getCoord(0) Should be -11-->"+dsid4.getCoord(0));

        
        dsid7.setCoord(0,12);
        System.out.println(" 3D IndexSpaceID: setCoord(0) Should be 12-->"+dsid7.getCoord(0));
        dsid7.setCoord(1,23);
        System.out.println(" 3D IndexSpaceID: setCoord(1) Should be 23-->"+dsid7.getCoord(1));
        dsid7.setCoord(2,9);
        System.out.println(" 3D IndexSpaceID: setCoord(2) Should be 9-->"+dsid7.getCoord(2));
        
        
        dsid8.setCoord(0,-7);
        System.out.println(" 3D IndexSpaceID: setCoord(0) Should be -7-->"+dsid8.getCoord(0));
        dsid8.setCoord(1,-12);
        System.out.println(" 3D IndexSpaceID: setCoord(1) Should be -12-->"+dsid8.getCoord(1));
        dsid8.setCoord(2,-27);
        System.out.println(" 3D IndexSpaceID: setCoord(2) Should be -27-->"+dsid8.getCoord(2));
        System.out.println(" 3D IndexSpaceID: (Negative Overwrite) getCoord(0) Should be -7-->"+dsid8.getCoord(0));
        System.out.println(" 3D IndexSpaceID: (Negative Overwrite) getCoord(1) Should be -12-->"+dsid8.getCoord(1));

        dsid1=new PackedIndexSpaceID(3);
        dsid3=new PackedIndexSpaceID(4,5);
        dsid4=new PackedIndexSpaceID(-3,-7);
        dsid7=new PackedIndexSpaceID(new int [] {4,2,7});
        dsid8=new PackedIndexSpaceID(-1,-2,-3);
        
        System.out.println();
        System.out.println("\n***IndexSpaceID Addition Test***");

        result= dsid1.copyID();
        result.add(dsid2);
        System.out.println(" 1D IndexSpaceID: (3) + (-2) Should be (1)-->"+result);

        result= dsid1.copyID();
        result.add(dsid2a);
        System.out.println(" 1D IndexSpaceID: (3) + (4) Should be (7)-->"+result);
        
        result= dsid3.copyID();
        result.add(dsid5);
        System.out.println(" 2D IndexSpaceID: (4,5) + (6,7) Should be (10,12)-->"+result);
                
        result= dsid3.copyID();
        result.add(dsid4);
        System.out.println(" 2D IndexSpaceID: (4,5) + (-3,-7) Should be (1,-2)-->"+result);


        result= dsid6.copyID();
        result.add(dsid7);
        System.out.println(" 3D IndexSpaceID: (1,1,3) + (4,2,7) Should be (5,3,10)-->"+result);

        result=dsid7.copyID();
        result.add(dsid8);
        System.out.println(" 3D IndexSpaceID: (4,2,7) + (-1,-2,-3) Should be (3,0,4)-->"+result);        
        
        System.out.println();
        System.out.println("***IndexSpaceID Subtraction Test***");
        
        result= dsid2a.copyID();
        result.subtract(dsid1);
        System.out.println(" 1D IndexSpaceID: (4) - (3) Should be (1)-->"+result);
        
        result= dsid1.copyID();
        result.subtract(dsid2);
        System.out.println(" 1D IndexSpaceID: (3) - (-2) Should be (5)-->"+result);
        
        result= dsid5.copyID();
        result.subtract(dsid3);
        System.out.println(" 2D IndexSpaceID: (6,7) - (4,5) Should be (2,2)-->"+result);
        
        result= dsid5.copyID();
        result.subtract(dsid4);
        System.out.println(" 2D IndexSpaceID: (6,7) - (-3,-7) Should be (9,14)-->"+result);        
        
        
        result= dsid7.copyID();
        result.subtract(dsid6);
        System.out.println(" 3D IndexSpaceID: (4,2,7) - (1,1,3) Should be (3,1,4)-->"+result);

        result= dsid8.copyID();
        result.subtract(dsid6);
        System.out.println(" 3D IndexSpaceID: (-1,-2,-3) - (1,1,3) Should be (-2,-3,-6)-->"+result);
        
        System.out.println();
        System.out.println("\n***IndexSpaceID containedBy() Test***");
        

        ISBounds db=new ISBounds(dsid2,dsid2a);
        System.out.println(" 1D (3) containedBy {(-2),(4)} should be true-->"+dsid1.containedBy(db));

        result=new PackedIndexSpaceID(-2);
        System.out.println(" 1D (-2) containedBy {(-2),(4)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(-3);
        System.out.println(" 1D (-3) containedBy {(-2),(4)} should be false-->"+result.containedBy(db));
        
        
        db=new ISBounds(dsid1,dsid2a);
        System.out.println(" 1D (-2) containedBy {(3),(4)} should be false-->"+dsid2.containedBy(db));

        result=new PackedIndexSpaceID(5);
        System.out.println(" 1D (5) containedBy {(3),(4)} should be false-->"+result.containedBy(db));


        result=new PackedIndexSpaceID(4);
        System.out.println(" 1D (4) containedBy {(3),(4)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(3);
        System.out.println(" 1D (3) containedBy {(3),(4)} should be true-->"+result.containedBy(db));




        db=new ISBounds(dsid3,dsid5);
        result=new PackedIndexSpaceID(5,5);
        System.out.println(" 2D (5,5) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(5,6);
        System.out.println(" 2D (5,6) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(5,7);
        System.out.println(" 2D (5,7) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(4,6);
        System.out.println(" 2D (4,6) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(6,6);
        System.out.println(" 2D (4,6) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(4,5);
        System.out.println(" 2D (4,5) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(6,7);
        System.out.println(" 2D (6,7) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(4,7);
        System.out.println(" 2D (4,7) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(6,5);
        System.out.println(" 2D (6,5) containedBy {(4,5),(6,7)} should be true-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(-1,6);
        System.out.println(" 2D (-1,6) containedBy {(4,5),(6,7)} should be false-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(7,6);
        System.out.println(" 2D (-1,6) containedBy {(4,5),(6,7)} should be false-->"+result.containedBy(db));
        
        db=new ISBounds(new PackedIndexSpaceID(-2,-3), new PackedIndexSpaceID(1,4));
        result=new PackedIndexSpaceID(-3,0);
        System.out.println(" 2D (-3,0) containedBy {(-2,-3),(1,4)} should be false-->"+result.containedBy(db));
        
        result=new PackedIndexSpaceID(-2,-4);
        System.out.println(" 2D (-2,-4) containedBy {(-2,-3),(1,4)} should be false-->"+result.containedBy(db));
        
    
    
        db=new ISBounds(dsid6,dsid7);
        result=new PackedIndexSpaceID(3,1,4);
        System.out.println(" 3D (3,1,4) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));
    
        result=new PackedIndexSpaceID(1,1,4);
        System.out.println(" 3D (1,1,4) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(3,2,7);
        System.out.println(" 3D (3,2,7) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(3,1,7);
        System.out.println(" 3D (3,1,7) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));
        


        result=new PackedIndexSpaceID(3,1,4);
        System.out.println(" 3D (3,1,4) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));
    
        result=new PackedIndexSpaceID(1,1,4);
        System.out.println(" 3D (1,1,4) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(3,2,7);
        System.out.println(" 3D (3,2,7) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(3,1,7);
        System.out.println(" 3D (3,1,7) containedBy {(1,1,3),(4,2,7)} should be true-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(5,1,7);
        System.out.println(" 3D (5,1,7) containedBy {(1,1,3),(4,2,7)} should be false-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(0,1,7);
        System.out.println(" 3D (0,1,7) containedBy {(1,1,3),(4,2,7)} should be false-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(2,1,8);
        System.out.println(" 3D (2,1,8) containedBy {(1,1,3),(4,2,7)} should be false-->"+result.containedBy(db));

        result=new PackedIndexSpaceID(2,-3,6);
        System.out.println(" 3D (2,-3,6) containedBy {(1,1,3),(4,2,7)} should be false-->"+result.containedBy(db));

        System.out.println("******Assign Tests*******");
        
        IndexSpaceID pid=new PackedIndexSpaceID(1,2,3);
        IndexSpaceID iid=new IntegerIndexSpaceID(5,6,7);
        
        pid.assign(iid);
        
        System.out.println("After pid.assign(iid), pid should be (5,6,7)-->"+pid);
        
        pid.assign(new int[]{1,2,3});
        iid.assign(pid);
        System.out.println("After iid.assign(pid), iid should be (1,2,3)-->"+iid);
        
        PackedIndexSpaceID pid2=new PackedIndexSpaceID(1,2,3);
        PackedIndexSpaceID pid3=new PackedIndexSpaceID(5,6,7);
        
        pid2.assign(pid3);
        System.out.println("After pid2.assign(pid3), pid2 should be (5,6,7)-->"+pid2);

        IntegerIndexSpaceID iid2=new IntegerIndexSpaceID(1,2,3);
        IntegerIndexSpaceID iid3=new IntegerIndexSpaceID(5,6,7);
        
        iid2.assign(iid3);
        System.out.println("After iid2.assign(iid3), iid2 should be (5,6,7)-->"+iid2);
    }
    
    
}

