/* 
    IntegerIndexSpaceID.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    ------------------ Edit history as of 03/20/05 -----------------------
    10/18/05 dlipsa: added IntegerIndexSpaceID(IndexSpaceID id, int value)
                     constructor.
    03/20/05 rdb: Allowed argument to getCoords( int[] ) to be null
                  Made getCoords() public.
                  Provided package access, refCoords() and updated all existing
                     uses of getCoords() to use refCoords().
*/
   
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

/** This class implements the IndexSpaceID class using an array of integers.
*/

public class IntegerIndexSpaceID extends IndexSpaceID {

    protected int[] axes;


    /** Constructs a one dimensional IntegerIndexSpaceID with axis value set 
        to the argument.
    */
    public IntegerIndexSpaceID(int i) {

        axes=new int[1];
        
        axes[0]=i;
    }

    /** Constructs a two dimensional IntegerIndexSpaceID with axis values set 
        to the given values.
    */
    public IntegerIndexSpaceID(int i, int j){

        axes=new int[2];
        
        axes[0]=i;
        axes[1]=j;
    }


    /** Constructs a three dimensional IntegerIndexSpaceID with axis values set 
        to the given values.
     */
    public IntegerIndexSpaceID(int i, int j, int k){

        axes=new int[3];
        
        axes[0]=i;
        axes[1]=j;
        axes[2]=k;
    }


    /** Constructs a IntegerIndexSpaceID with dimensionality matching the length of the
        parameter array. The axes will be set to the contents of the array. 
    */
    public IntegerIndexSpaceID(int [] axes)
    {    
        this.axes=new int[axes.length];
        System.arraycopy (axes, 0, this.axes, 0, axes.length);
    }
    
    /**
     * Constructs an IntegerIndexSpaceID with dimensionality matching 
     * <code>id</code> and all values set to <code>value</code>
     */
    public IntegerIndexSpaceID(IndexSpaceID id, int value)
    {
        axes = new int[id.dim ()];
        for (int i = 0; i < axes.length; i++)
            axes[i] = value;
    }
    
    
    /** Constructs an IntegerIndexSpaceID from the contents of the IndexSpaceID argument. 
    */
    public IntegerIndexSpaceID(IndexSpaceID id){
    
        this (id.refCoords());
    }

    /** Constructs an IntegerIndexSpaceID from the contents of the IntegerIndexSpaceID argument. 
    */
    public IntegerIndexSpaceID(IntegerIndexSpaceID id){
        this (id.axes);
    }
    

    /** Return a copy of this IndexSpaceID.*/
    public final IndexSpaceID copyID(){
        
        return new IntegerIndexSpaceID(this.axes);
    }

    /** Return a copy of this IndexSpaceID. */
    public Object clone(){
        
        return new IntegerIndexSpaceID(this.axes);
    }
    
    
    /**Set the axes of this id to the axes of the parameter id.
       The assign method may elect to refer to the data in
       the parameter id, and not make a copy of this information.
       
       Package Access only.
       */
    void assign(IntegerIndexSpaceID id){

        this.axes=id.axes;
    }    
    
    
    /**Set the axes of this id to the parameter array.
       The assign method may elect to refer to the data in
       the parameter id, and not make a copy of this information.
       
       Package Access only.       
       */
    final void assign(int[] axes){

        this.axes=axes;
    }    


    /** Get the value corresponding to the specified axis. If the specified axis 
        is out of bounds, a DimensionalityError is created. */
    public final int getCoord(int axisNumber){
        
        int axis;
        
        try {
            axis = axes[axisNumber];
        }
        catch (ArrayIndexOutOfBoundsException e) {
            new DimensionalityError(SDBError.Recoverable,"Argument specifies non-existent axis.");
            return -1;
        }
        
        return axis;

    }
    
    /** Get an array of axis values. This method returns the axes for this ISID, so we
        only allow package access.*/
    
    int[] refCoords()
    {
        return axes;
    }

    /** Fill the array argument with the coordinate values for this ISID. For convenience,
        a reference to the argument array is returned. null is valid for the
        argument and results in a new array being allocated and returned. */
    public int[] getCoords( int[] coords )
    {
        int d=this.axes.length;
        if ( coords == null )
           coords = new int[ d ];
        
        try{
            for ( int i=0; i< d; i++)
                coords[i]=axes[i];
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Argument array is the wrong length.");
        }
        return coords;
    }

    
    
    
    /** Set the specified axis to the specified value. If the specified axis 
        is too large, a DimensionalityError is created. */ 
    public void setCoord( int axisNumber,int value )
    {
        try
        {
            axes[axisNumber] = value;
            
        } catch (ArrayIndexOutOfBoundsException e)
        {
            new DimensionalityError(SDBError.Recoverable,"Argument specifies non-existent axis.");
        }
    }

    
    /** Returns the dimensionality of this IndexSpaceID. */
    public final int dim(){
    
        return this.axes.length;
    }

    /** Return true if parameter is equal to this object, false otherwise.
    */
    public boolean equals(Object obj){
    
        try{
        
            IndexSpaceID id=(IndexSpaceID) obj;

            int [] idAxes = id.refCoords();
            int i, d=idAxes.length;   
                     
            if (d != this.axes.length){
                return false;
            }
            
            for(i=0;i<d;i++){
                if(this.axes[i] != idAxes[i]){
                    return false;
                }
            }
    
            return true;
        } catch (ClassCastException e){
        
            return false;
        }
    }
    
    
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i,d=axes.length;
        String result=new String(indent+"<IntISID (");
        
        for(i=0;i<d-1;i++){
            result+=axes[i]+",";
        }
        
        result+=axes[i]+")>";
        
        return result;
    }    
    
    
    public static void main(String[] args){
        
        System.out.println("*****IntegerIndexSpaceID Tests*****");
        
        IntegerIndexSpaceID first=new IntegerIndexSpaceID(10,10);
        
        System.out.println("First ID is "+first);
        
        IntegerIndexSpaceID theClone=(IntegerIndexSpaceID) first.clone();

        System.out.println("The clone is "+theClone);

        theClone.setCoord(1,5);
        
        System.out.println("After theClone.setCoord(1,5): The clone is "+theClone);
        System.out.println("                        and the original is "+first);

        System.out.println();
        System.out.println("Speed Test:  assign() vs. copy()");
        
        int i;
        long startTime,stopTime;
        IntegerIndexSpaceID dst=new IntegerIndexSpaceID(1,2,3);
        IntegerIndexSpaceID src=new IntegerIndexSpaceID(4,5,6);
        

        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst=(IntegerIndexSpaceID) src.copyID();
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("Time for one Million copies: " + (stopTime-startTime) + " milliseconds.");
        
        
        
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.assign(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("Time for one Million assigns: " + (stopTime-startTime) + " milliseconds.");
        
        
        
        
        

        
    }

}
