//
//  LRUCache.java
//  CacheTest
//
//  Created by Philip Rhodes on Fri Feb 06 2004.
//  Copyright (c) 2004 University of New Hampshire. All rights reserved.
//
//--------------------------------------------------------------------------
//Edit history 
// 06/05/08 rdb: Made LinkedHashMap generic
// 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
//--------------------------------------------------------------------------

package edu.unh.sdb.datasource;

import java.util.*;

import edu.unh.sdb.common.*;

public class LRUCache extends BlockCache {
    private static final long serialVersionUID = 0; // for serialization inheritance

    long [] tagAxisWeights;          // Controls mapping of n-D block space to 1D cache tags
 
    int[] queryBoundsDimensions;
            
                
    // The following group of data members comprise the ISBounds
    // that is used to query the source indexable object.
                            
    int[] sQBLowerInts;
    int[] sQBUpperInts;
        
    ISBounds sourceQueryBounds;
    
    
    
    // The following group of data members comprise the ISBounds
    // that is used to query the cacheblock for block access. For
    // datum acess, datumCoords1 is used by itself.
    
    // Used for storing the intra-block datum coords. These arrays
    // are assigned into the bounds defined below.
    int [] datumCoords1;             
    int [] datumCoords2;
    
    // This bounds denotes the subblock to be read from 
    // a cache block into the argument block. 
    ISBounds subblockQueryBounds;

    ISBoundsIterator partitionIter;
    ISBounds subDstBounds;
    int [] distance;



    LRULinkedHashMap hash;
    LRUKey theKey;
    RegularISPartitioning partitioning;
    LRUCacheDataBlock lastBlock;

    LRUCache(Indexable source, ISBounds iterationBounds, IndexSpaceID blockDimensions,AxisOrdering cacheBlockOrdering, int numBlocks){

        super(source, iterationBounds, blockDimensions, numBlocks);
        
        LRUCacheDataBlock [] theBlocks;
        int d=this.dimensionality;
        RecordDescriptor rd = source.getRecordDescriptor();
        
        theBlocks = new LRUCacheDataBlock [numBlocks+1];
        
        
        for ( int i=0; i<numBlocks+1; i++){
        
            theBlocks[i] = new LRUCacheDataBlock(new ISBounds(blockDimensions), rd, cacheBlockOrdering, -1);
        }
        
        this.hash=new LRULinkedHashMap(theBlocks);
        this.partitioning = new RegularISPartitioning(new IntegerIndexSpaceID(blockDimensions), this.iterationSpaceBounds);
        this.theKey=new LRUKey(-1);
        
        // We're just trying to make a valid Iterator with the right dimensionality.
        this.partitionIter = new ISBoundsIterator(source.copyBounds(), source.copyBounds());
        this.subDstBounds = new ISBounds(d);
        this.distance = new int[d];
        
        
        // From UniBlockCache
        AxisOrdering partitioningOrdering=new AxisOrdering(d);
        
        this.queryBoundsDimensions = new int[d];
        
        this.datumCoords1=new int[d];
        this.datumCoords2=new int[d];

        
        this.tagAxisWeights=partitioningOrdering.makeAxisWeights(partitioningDimensions);

        this.sQBLowerInts=new int[d];
        this.sQBUpperInts=new int[d];

        sourceQueryBounds = new ISBounds(this.sQBLowerInts, this.sQBUpperInts);
        subblockQueryBounds = new ISBounds(this.datumCoords1, this.datumCoords1);
    }

    void lookup(Datum theDatum, IndexSpaceID id){
    
        int tag=0;

        int coordBlockIndex;
        int [] coords=id.refCoords();
        LRUCacheDataBlock theBlock=null;
         
        for ( int i=0; i<this.dimensionality ;i++)
        {
            coordBlockIndex = (coords[i] / cacheBlockDimensions[i]);
            this.sQBLowerInts[i] = coordBlockIndex * cacheBlockDimensions[i];
            this.datumCoords1[i] = coords[i] - this.sQBLowerInts[i];
            tag += coordBlockIndex * this.tagAxisWeights[i];
        }
           
        if ( lastBlock != null && lastBlock.getTag() == tag )
           theBlock = lastBlock;
        else 
           theBlock = this.findBlock(tag);
        
        if(theBlock != null)
        {
            //System.out.println("Cache Hit: tag="+tag);
            theBlock.datum( theDatum,this.datumCoords1);
        } 
        else 
        {
           //System.out.println("Cache Miss: tag="+tag);
           theBlock = this.hash.getOldest();
            
            for ( int i=0; i < this.dimensionality; i++) 
            {
               this.sQBUpperInts[i] = this.sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
            }
            
            this.sourceQueryBounds.assign(sQBLowerInts,sQBUpperInts);

            theSource.subblock(theBlock, this.cacheBlockBounds,sourceQueryBounds);
            theBlock.setTag(tag);
            
            theBlock.datum(theDatum,this.datumCoords1);
            
            this.hash.put(theBlock, theBlock);
            //System.out.println("LRUCache: Inserted "+theBlock.getTag());            
        }
        
        lastBlock = theBlock;
    }
        
    void lookup(BasicBlock queryBlock, ISBounds dstBounds, ISBounds userQueryBounds){


        int i, d=this.dimensionality;
        int lowerTag=0, upperTag=0;
        int lowerCoordBlockIndex, upperCoordBlockIndex;
        int [] lowerCoords=userQueryBounds.getLowerArray();
        int [] upperCoords=userQueryBounds.getUpperArray();
        LRUCacheDataBlock theBlock=null;
                
        userQueryBounds.getDimensions(this.queryBoundsDimensions);
        
        for(i=0;i<d;i++){

            lowerCoordBlockIndex = (lowerCoords[i] / this.cacheBlockDimensions[i]);
            upperCoordBlockIndex = (upperCoords[i] / this.cacheBlockDimensions[i]);
            
            this.sQBLowerInts[i] = lowerCoordBlockIndex * this.cacheBlockDimensions[i];
            //this.sQBUpperInts[i] = upperCoordBlockIndex * this.cacheBlockDimensions[i];
            
            this.datumCoords1[i] = lowerCoords[i] - this.sQBLowerInts[i];
            this.datumCoords2[i] = this.datumCoords1[i] + this.queryBoundsDimensions[i] - 1;
            
            lowerTag += lowerCoordBlockIndex * this.tagAxisWeights[i];
            upperTag += upperCoordBlockIndex * this.tagAxisWeights[i];
        }


        if(lowerTag == upperTag){

            if(lastBlock != null && lastBlock.getTag() == lowerTag){
                
                theBlock = lastBlock;
            } else {
            
                theBlock = this.findBlock(lowerTag);
            }
        
        
            subblockQueryBounds.assign(this.datumCoords1, this.datumCoords2);

            if(theBlock != null){
                //System.out.println("*Hit*");
                //AppMonitor.recordCacheHit();
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            } else {
                //System.out.println("*Miss*");
                //AppMonitor.recordCacheMiss();
                
                theBlock= this.hash.getOldest();

                
                for(i=0;i<d;i++){
                    
                    this.sQBUpperInts[i] = this.sQBLowerInts[i] + this.cacheBlockDimensions[i] - 1;
                    
                    this.sQBUpperInts[i] = (this.sQBUpperInts[i] > this.upperIterationSpaceCoords[i]) ? this.upperIterationSpaceCoords[i] : this.sQBUpperInts[i];
                }
                
                this.sourceQueryBounds.assign(sQBLowerInts,sQBUpperInts);
    
                theSource.subblock(theBlock, this.cacheBlockBounds,sourceQueryBounds);
                theBlock.setTag(lowerTag);
                
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
                
                this.hash.put(theBlock, theBlock);
//System.out.println("LRUCache: Inserted "+theBlock.getTag());

            }
            
        } else {

            //ISBoundsIterator partitionIter = this.partitioning.getPartitionIterator(userQueryBounds);
            partitionIter = this.partitioning.getPartitionIterator(userQueryBounds, partitionIter);
            //ISBounds subDstBounds = new ISBounds(d);
            int [] usbLower = userQueryBounds.getLowerArray();
            int [] dstLower = dstBounds.getLowerArray();
            //int [] distance = new int[d];
            
            for(i=0; i < d ; i++){
            
                distance[i] = dstLower[i] - usbLower[i];
            }
            
            
            for(partitionIter.init(); partitionIter.valid(); partitionIter.next()){
            
                subDstBounds.setAndMove(partitionIter, distance);
            
                //System.out.println(partitionIter);

                this.lookup(queryBlock, subDstBounds, partitionIter);
               
            }
            
            //System.out.println("*****");
            //System.out.println("You're in da weeds, Lorna! ;-)");
            
            //this.theSource.subblock(queryBlock, dstBounds, userQueryBounds);
        
        }
    
        lastBlock = theBlock;
    
    
    }
    
    void lookup(Datum theDatum, IndexSpaceID id, CacheContext context){

      int coordBlockIndex;
      int d = this.dimensionality;
      int [] coords=id.refCoords();
      LRUCacheDataBlock theBlock=null;
      boolean coherent = true;
         
         
      for ( int i=0; i<d; i++ ){
         
         coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i];
         
         context.datumCoords1[i] = coords[i] - (lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i]); // required
         
         if(coordBlockIndex != context.coordBlockIndex[i]){
         
            coherent = false;
         }
         
         context.coordBlockIndex[i] = coordBlockIndex;
      }
            
      if(!coherent){
      
         context.arrayIndex = 0;
         context.tag = 0;
         
         for ( int i=0; i<d; i++ ){
                        
            context.tag += context.coordBlockIndex[i] * this.tagAxisWeights[i];  // required
         }
      }
           
        if ( lastBlock != null && lastBlock.getTag() == context.tag )
           theBlock = lastBlock;
        else 
           theBlock = this.findBlock(context.tag);
        
        if(theBlock != null)
        {
            //System.out.println("Cache Hit: tag="+tag);
            theBlock.datum( theDatum, context.datumCoords1);
        } 
        else 
        {
           //System.out.println("Cache Miss: tag="+tag);
           theBlock = this.hash.getOldest();
            
            for ( int i=0; i<d; i++){
               context.sQBLowerInts[i] = lowerIterationSpaceCoords[i] + context.coordBlockIndex[i] * cacheBlockDimensions[i]; // required
               context.sQBUpperInts[i] = context.sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
                
            }
            
            context.sourceQueryBounds.assign(context.sQBLowerInts, context.sQBUpperInts);
            
            // The iterationSpaceBounds should already have been clipped against
            // the ds bounds, so clipping against it should be safe.
            context.sourceQueryBounds.intersectWith( iterationSpaceBounds );  

            theSource.subblock(theBlock, this.cacheBlockBounds, context.sourceQueryBounds);
            theBlock.setTag(context.tag);
            
            theBlock.datum(theDatum, context.datumCoords1);
            
            this.hash.put(theBlock, theBlock);
            //System.out.println("LRUCache: Inserted "+theBlock.getTag());            
        }
        
        lastBlock = theBlock;
    }
    
    
    

    private LRUCacheDataBlock findBlock(int tag){
        
        this.theKey.setKeyValue(tag);
        return this.hash.get(this.theKey);
    
    }
    
//    protected LRUCacheDataBlock load(){
//    
//        LRUCacheDataBlock oldest = this.hash.getOldest();
//        
//        this.source.subblock(oldest, dstBounds, subblockQueryBounds);
//        
//        return oldest;
//    }



    private class LRULinkedHashMap // rdb extends LinkedHashMap
               extends LinkedHashMap <LRUCacheDataBlock,LRUCacheDataBlock> 
    {
    
       private static final long serialVersionUID = 0; // for serialization inheritance
       LRUCacheDataBlock oldest;
        boolean removeEldest=false;
    
        public LRULinkedHashMap(LRUCacheDataBlock[] blocks){
        
            super(blocks.length - 1, 0.1f, true);
            
            int i, blocksInCache=blocks.length - 1;
            
            this.removeEldest = false;
            
            for(i=0; i<blocksInCache; i++){
            
                this.put(blocks[i], blocks[i]);
            }

            this.removeEldest = true;
            
            this.oldest = blocks[i];
        }
    
    
        protected boolean removeEldestEntry(Map.Entry eldest){
        

            if (this.removeEldest) {
//System.out.println("LRUCache: Removing "+((LRUCacheDataBlock) eldest.getValue()).getTag());
            
                this.oldest = (LRUCacheDataBlock) (eldest.getValue());
                return true;
            } else {
//System.out.println("Called removeEldestEntry:  returning false");
            
                return false;
            }
        }
        
        LRUCacheDataBlock getOldest(){
        
            return this.oldest;
        }
    }


    private class LRUKey {
    
        private int k;
        
        LRUKey(int k){
        
            this.k=k;
        }
        
        public void setKeyValue(int k){
        
            this.k = k;
        }
        
        public int hashCode(){
        
            return k;
        }
        
        public boolean equals(Object o){
        
            return this.k == o.hashCode();
        }
    }

    private class LRUCacheDataBlock extends CacheDataBlock {
    
        
        LRUCacheDataBlock(ISBounds isb, RecordDescriptor recordDesc, AxisOrdering cacheBlockOrdering, int tag){
        
            super(isb, recordDesc, cacheBlockOrdering);
            super.setTag(tag);
        }
        
        /** Use the inherited tag as the hashcode. This may violate the hashCode contract,
         *  since this value changes over the duration of the object's lifetime. So, we'll
         *  hide this ugliness in a private class.
         */
        public int hashCode(){
        
            return (int) this.tag;
        }
    }
    
    
    public static void main(String argv[]){
    
        
    
    }
}
