/* MRDataSource
 * 
 * Created on Mar 29, 2005
 *
 * ------------------ Edit history ------------------------------------
 *  06/06/08 rdb: Changed hierarchy ArrayList so that every entry is itself an
 *                   ArrayList<DSFile>
 *                Changed interface to support (in a crude way) multiple DS on
 *                   the same level -- just an index into the list at that level.
 *  11/29/05 rdb: revised file name handling; '@' is no longer part of filename url.
 *  11/27/05 rdb: started basic error support, getErrorDataSource, not complete
 *  10/21/05 rdb: added dumb code in main to reference things that cause compiler
 *                warnings, but I don't want to delete.
 *  06/01/05 rdb: modified to match FDLDescriptor refactoring
 *  03/29/05 rdb: created
 */
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.util.ArrayList;
import java.io.*;

/**
 * MRDataSource 
 *   Serves as an interface to a MultiResolution hierarchy of DataSources 
 * representing the same data at different resolution levels. 
 * 
 * This implementation is a primitive one that uses file name and file
 * directory structure conventions to define the hierarchy,
 * rather than using complete metadata support for defining the hierarchy
 * (as well as associated error datasources).
 * 
 * For this version we'll encode the MR nature of the data in the xfdl file
 * by prefixing the file name with with the "@" symbol. This will tell
 * DataSource.create( ... ) that it should create an MRDataSource.
 * 
 * FDL fileName format:  @rootdir/level0DataFileName
 * 
 * It is acceptable for the "rootdir" field to be '.';
 * it is acceptable for the "level0DataFileName" to be omitted; default is "data".
 * 
 * The filename (with the "@" removed) from the .xfdl file should be the
 * name of the root directory of the hierarchcy. This directory must have
 * an entry named "data". root/data can be a plain file or it can be a 
 * directory that contains the data file. (This is especially intended for
 * time series data sets that have many files, but could also be used for
 * storing multiple attributes in separate files -- though there is no 
 * support for this yet.)
 * 
 * If there is an error datasource associated with the data it should
 * be in the file "error", or in a file (or series of files) in the directory
 * called "error". Any lower resolution representations of this data must
 * be stored in other directories. With time series data, there could be
 * either a spatial wavelet or a time wavelet; eventually we want to support
 * adaptive resolution representations for which there could be multiple versions
 * for this data at different error tolerance levels. 
 * 
 * For the time being, this code supports the following conventions for the
 * lower resolution data:
 *   1. If there is just one subdirectory (other than "data" and "error"), we
 *      assume that it is the next lower resolution.
 *   2. If the directory name begins with "space" or is "s", then it is a spatial
 *      wavelet directory (intended to be used for time series).
 *   3. If the file name begins with "time" or is "t", then it is a time
 *      wavelet directory for a time series data source.
 * 
 * Each lower resolution subdirectory must follow the same conventions
 * as the root directory; i.e., it must contain a file or a directory named
 * "data" and may contain a file or directory named "error" and a may contain
 * a directory defining the next lower resolution of data.
 * 
 * Each level of resolution must have its own xfdl file. It can be at the 
 * same level as the data/error file or directories or it can be in the
 * data and error directories. 
 * 
 * One goal of this initial implementation is to allow a single .xfdl
 * file to be used for an entire hierarchy. This may prove to be of
 * marginal importance since new .xfdl files could be automatically generated
 * by the wavelet program. 
 * 
 * @author rdb
 */
public class MRDataSource extends DataSource
{
   //----------------- data members ----------------------------------
   ArrayList<ArrayList<DSFile>>  hierarchy 
   = new ArrayList<ArrayList<DSFile>>();  // Stores hierarchy of DataSources
   //   resolution directory levels
   int        curLevel  = 0;             // currently active level
   int        curWhich  = 0;             // currently active DS at current level
   DataSource curDS    = null;

   //----------------- DSFile class ----------------------------------
   /*
    * DSFile class encapsulates a DataSource and its associated File object
    * into a single entity to be saved in the hierarcy list.
    */
   class DSFile 
   {
      public DataSource ds;
      public File       file;
      DSFile( DataSource ds, File file )
      {
         this.ds = ds; this.file = file;
      }
   }

   /**
    * @param aname
    * @param fdld
    */
   public MRDataSource( String aname, FDLDescriptor fdld )
   {
      super( aname );
      this.fdlDescriptor = fdld;

      //String mrFilename = fdld.dataFileURL.substring( 1 ); // delete @

      String mrFilename = fdld.getDataFileURL();
      String rootFilename = mrFilename;
      String dataFilename = "data";

      int    slash = mrFilename.lastIndexOf( File.separatorChar );
      if ( slash > 0 )
      {
         dataFilename = mrFilename.substring( slash + 1 );
         rootFilename = mrFilename.substring( 0, slash );
      }

      fdld.setDataFileURL( mrFilename );
      //rootFilename = checkDataFileName( rootFilename, fdld.getRootDirectory() ); 

      File mrRootDir = new File( rootFilename );

      if ( !mrRootDir.isDirectory() )
      {
         new SDBError( SDBError.AlwaysThrow, "MultiRes hierarchy root must be a" +
               " directory. Specified name is not: " + mrRootDir );
         return;
      }

      dataFilename = rootFilename + File.separator + dataFilename;
      File dataFile = new File( dataFilename );
      if ( !dataFile.exists() )
      {
         new SDBError( SDBError.AlwaysThrow, "MRDataSource: Can't find data file: " 
               + dataFilename );
         return;         
      }
      this.fdlDescriptor.setDataFileURL( dataFilename );
      this.curDS = DataSource.create( this.name + ".root", this.fdlDescriptor );
      addToLevel( 0, this.curDS, dataFile );
      updateMembers();
   }

   //-------------------- MRDataSource methods -------------------------------
   /**
    * upDateMembers copies the key information from the current DS to 
    * data members of the MRDS.
    */
   protected void updateMembers()
   {
      this.recordDesc = this.curDS.recordDesc;
      this.axisWeights = this.curDS.axisWeights;
      this.bounds = this.curDS.bounds;
      this.dataOrder = this.curDS.dataOrder;
      this.dimensionality = this.curDS.dimensionality;
      this.isActivated = this.curDS.isActivated;
      this.numAttributes = this.curDS.numAttributes;
      this.axisOrdering = this.curDS.axisOrdering;
   }

   /**
    * Find data file
    */
   protected File getDataFile( File root )
   {
      File data = null;
      File rootData = new File( root + "/data" );
      if ( rootData.isFile() )
         data = rootData;
      else if ( rootData.isDirectory() )
      {
         File[] entries = rootData.listFiles();
         File[] files   = getFiles( entries );
         if ( files.length == 1 && files[0].isFile() )
            data = files[0];
         else
         {
            new SDBError( SDBError.AlwaysThrow, "MRDataSource: Ambiguous choice " 
                  + "for root data file in " + root );
         }
      }
      return data;
   }

   /**
    * Find fdl file with the prefix indicated by argument.
    * The suffix can be .fdl, .xfdl or .xml. 
    * If no fdl file is found, check if the prefix is a directory;
    * if so, look for the same kind of file name in the directory.
    *
    */
   private File getFDLFile( File root, String prefix )
   {
      String[] suffix = new String[] { ".xfdl", ".fdl", ".xml" };
      File fdl = null;
      boolean found = false;
      String   base = root + "/" + prefix;


      for ( int i = 0; i < suffix.length && !found; i++ )
      {
         fdl = new File( base + suffix[i] );
         found = fdl.exists();
      }

      if ( !fdl.exists() )
      {
         File dataDir = new File( base );
         if ( dataDir.isDirectory() )
         {
            fdl = getFDLFile( dataDir, prefix ); // look for same name at next level 
            if ( fdl == null )
               fdl = getFDLFile( dataDir, "" ); // look for hidden name (a la tsd)
         }
         else 
            fdl = null;
      }
      return fdl;
   }

   /**
    * createDS( File )
    */
   private DataSource createDS( File fdlFile, int level )
   {  
      System.out.println( fdlFile.getAbsolutePath() );
      System.out.println( fdlFile.getName() );
      DataSource ds = DataSource.create( this.name + "." + level, 
            fdlFile.getAbsolutePath() );
      addToLevel( level, ds, fdlFile );
      return ds;
   }

   /**
    * addToLevel( ds, level )
    */
   protected void addToLevel( int level, DataSource ds, File file )
   {
      DSFile dsf = new DSFile( ds, file );
      ArrayList<DSFile> levelList = null;
      if ( level < this.hierarchy.size() )
         levelList = this.hierarchy.get( level );
      if ( levelList != null )
         levelList.add( dsf );
      else   // already had a file at this level
      {
         levelList = new ArrayList<DSFile>();
         levelList.add( dsf );
         this.hierarchy.add( level, levelList );
      }
   }

   /**
    *  getDSFile( int level, int which )
    */
   private DSFile getDSFile( int level, int which )
   {
      DSFile dsf = null;
      try 
      {
         ArrayList<DSFile> levelList = this.hierarchy.get( level );
         dsf = levelList.get( which );
      } 
      catch ( IndexOutOfBoundsException iobex )
      {
         new SDBError( SDBError.Recoverable, "MRDataSource.getDSFile: "
               + " Nonexistent resolution choice " + which + " at level " + level );
      }     
      return dsf;
   }

   /**
    *  getDataSource() -- return current data source
    */
   public DataSource getDataSource( )
   {
      DSFile dsf = getDSFile( this.curLevel, this.curWhich );
      if ( dsf == null )
         return null;
      else 
         return dsf.ds;
   }

   /**
    *  getDataSource( int level )
    */
   public DataSource getDataSource( int level )
   {
      DSFile dsf = getDSFile( level, 0 );
      if ( dsf == null )
         return null;
      else 
         return dsf.ds;
   }
   /**
    *  getDataSource( int level, int which )
    */
   public DataSource getDataSource( int level, int which )
   {
      DSFile dsf = getDSFile( level, which );
      if ( dsf == null )
         return null;
      else 
         return dsf.ds;
   }

   /**
    *  getDataFile()
    */
   public File getDataFile()
   {
      DSFile dsf = getDSFile( this.curLevel, 0 );
      if ( dsf == null )
         return null;
      else 
         return dsf.file;
   }

   /**
    *  getDataFile( int level )
    */
   public File getDataFile( int level )
   {
      DSFile dsf = getDSFile( level, 0 );
      if ( dsf == null )
         return null;
      else 
         return dsf.file;
   }
   /**
    *  getDataFile( int level, int which )
    */
   public File getDataFile( int level, int which )
   {
      DSFile dsf = getDSFile( level, which );
      if ( dsf == null )
         return null;
      else 
         return dsf.file;
   }

   /**
    *  int curLevel()
    */
   public int curLevel()
   {
      return this.curLevel;
   }
   /**
    *  int curWhich()
    */
   public int curWhich()
   {
      return this.curWhich;
   }

   /**
    *  int getNumResolutionLevels()
    */
   public int getNumResolutionLevels()
   {
      return this.hierarchy.size();
   }
   /**
    *  int getNumDSatLevel( int )
    */
   public int getNumDSatLevel( int level )
   {
      try
      {
         return this.hierarchy.get( level ).size();
      }
      catch ( IndexOutOfBoundsException iobex )
      {
         return 0;
      }
   }

   /**
    *  int  setCurrentDS( int newLevel )
    * 
    * Set resolution level to the DS 0 on the specified level;  don't open 
    * the datasource.
    * 
    * returns true if requested DS exists
    */
   protected boolean setCurrentDS( int newLevel )
   {
      return setCurrentDS( newLevel, 0 );
   }
   /**
    *  int  setCurrentDS( int newLevel, int index )
    * 
    * Set resolution level to specified level and DS, but don't open the datasource
    * If the desired level or index does not exist, no change is made.
    * 
    * returns true if requested setting occurred. 
    */
   protected boolean setCurrentDS( int newLevel, int index )
   {
      DSFile dsf = getDSFile( newLevel, index );
      if ( dsf == null )
         return false;
      else
      {
         this.curLevel = newLevel;
         this.curWhich = index;
         return true;
      }
   }

   /**
    *  boolean finer( )
    *  Move to next finer data representation. Returns false if no finer available.
    */
   public boolean finer()
   {
      if ( this.curLevel > 0 )
         openDS( --this.curLevel );
      else
         return false;
      return true;
   }

   /**
    *  boolean changeResolution( int changeCode )
    * 
    * changeCode = 0 implies go to resolution level zero
    * otherwise new resolution level = old level + changeCode
    * 
    * changeCode < 0 implies make resolution finer by abs( changeCode ) levels
    * changeCode > 0 implies make resolution coarser by changeCode levels
    */
   public boolean changeResolution ( int changeCode )
   {
      int prevLevel = this.curLevel;
      int desiredLevel = 0;

      if ( changeCode == 0 )
      {
         desiredLevel = 0;
         setCurrentDS( 0 );
      }
      else
      {   
         desiredLevel = this.curLevel + changeCode;
         setCurrentDS( desiredLevel );
      }
      if ( prevLevel != this.curLevel )
         openDS( this.curLevel );

      return this.curLevel == desiredLevel;
   }

   /**
    *  boolean coarser( )
    *  Move to next coarser data representation. Returns false if no coarser available
    */
   public boolean coarser()
   {
      if ( this.curLevel + 1 < this.hierarchy.size() 
            || checkCoarser( this.curLevel ))
         openDS( ++this.curLevel );
      else 
         return false;
      return true;
   }

   /**
    *  int findCoarser( int level )
    * 
    * NOT YET IMPLEMENTED
    * 
    * Try to find all next coarser data representations in file system for
    * the next coarser level of resolution from that specified in the argument.
    * The "base" argument is the directory of the parent in the hierarchy;
    * the "level" argument is the level for the children, if any, of that parent.
    * 
    * Returns number of children created.
    */
   protected int findCoarser(  int level )
   {
      // Need to implement a tree tree ds rather than just a list
      // This routine will need to find every node at the specified
      // level and for each find all its children. Probably need
      // to pass a tree node rather than a level number.
      return 0;
   }

   /**
    * For a particular directory at the specified level find all the 
    * coarser representations for that directory.
    * 
    * NOT FINISHED
    * 
    * Returns the number of coarser representations found.
    * 
    * @param parent
    * @param level
    */
   protected int findCoarser(  File parent, int level )
   {
      File fdlFile  = null;
      int  count   = 0;
      File[] dirs = getDirs( parent.listFiles() );
      for ( int i = 0; i < dirs.length; i++ )
      {
         fdlFile = getFDLFile( dirs[i], "data" );
         if ( fdlFile != null )  // found fdl file, add this node to level list
         {
            DataSource newDS = createDS( fdlFile, level + 1 );
            if ( newDS != null )
               count++;
         }
      }         
      System.out.println( "New level" + level + " has " + count + "  new datasets" );
      return count;
   }

   /**
    *  void checkCoarser( int level )
    *  Try to open next coarser data representation.
    */
   protected boolean checkCoarser( int level )
   {
      File foundDir = null;
      File onlyOne = null;
      int  count   = 0;
      File curDir = getDataFile( level ).getParentFile();
      File[] dirs = getDirs( curDir.listFiles() );
      for ( int i = 0; i < dirs.length && foundDir == null; i++ )
      {
         String name = dirs[i].getName();
         if ( name.equals( "s" ) || name.equals( "t" )
               || name.startsWith( "space" ) || name.startsWith( "time" ) 
               || name.startsWith( "level" ) )
            foundDir = dirs[i];
         else if ( !name.equals( "error" ) && !name.equals( "data")
               && !name.startsWith( ".") )
         {
            count++;
            onlyOne = dirs[i];
         }
      }
      if ( foundDir == null && count == 1 )
         foundDir = onlyOne;
      if ( foundDir == null )
         return false;

      File fdlFile = getFDLFile( foundDir, "data" );
      if ( fdlFile == null )
         return false;

      DataSource newDS = createDS( fdlFile, level+1 );
      if ( newDS == null )
         return false;
      System.out.println( "New level: " + getDataFile( level+1 ) );
      return true;
   }

   /**
    *  Swap datasources. For now I'll deactivate and activate.
    */
   protected void openDS( int level )
   {
      this.curDS.deactivate();
      this.curDS = getDataSource( level );
      this.curLevel = level;
      this.curDS.activate();
      this.updateMembers();
   }

   /**
    * get list of plain files from a File array 
    */
   private File[] getFiles( File[] entries )
   {
      File[] ret = null;
      ArrayList<Object> temp = new ArrayList<Object>();
      for ( int i = 0; i < entries.length; i++ )
      {
         if ( entries[ i ].isFile() )
            temp.add( entries[i] );
      }

      ret = new File[ temp.size() ];
      for ( int i=0; i < ret.length; i++ )
         ret[ i ] = (File) temp.get( i ); 
      return ret;
   }

   /**
    * get list of directories from a File array 
    */
   private File[] getDirs( File[] entries )
   {
      File[] ret = null;
      ArrayList<Object> temp = new ArrayList<Object>();
      for ( int i = 0; i < entries.length; i++ )
      {
         if ( entries[ i ].isDirectory() )
            temp.add( entries[i] );
      }

      ret = new File[ temp.size() ];
      for ( int i=0; i < ret.length; i++ )
         ret[ i ] = (File) temp.get( i ); 
      return ret;
   }

   //-------------------- DataSource  methods ------------------------------------
   /**
    * @see edu.unh.sdb.datasource.DataSource#activate()
    */
   public void activate()
   {
      super.activate();
      this.curDS.activate();
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#datum(edu.unh.sdb.common.Datum, edu.unh.sdb.datasource.IndexSpaceID)
    */
   public void datum( Datum dtm, IndexSpaceID did )
   {
      this.curDS.datum( dtm, did );
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#datum(edu.unh.sdb.common.Datum, edu.unh.sdb.datasource.IndexSpaceID, edu.unh.sdb.common.FieldIDMapper)
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      this.curDS.datum( dtm, did, mp );
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#subblock(edu.unh.sdb.datasource.DataBlock, edu.unh.sdb.datasource.ISBounds, edu.unh.sdb.datasource.ISBounds)
    */
   public void subblock( DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
      this.curDS.subblock( theBlock, dstBounds, srcBounds );
   }

   public void subblock( DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, 
         FieldIDMapper mp )
   {
      this.curDS.subblock( theBlock, dstBounds, srcBounds, mp );
   }

   public static void main( String[] args )
   {
      MRDataSource ds = (MRDataSource) DataSource.create( "MRtest", args[0] );
      ds.activate();
      float[] data = ds.getFloats();
      System.out.println( ds.getName() + "@" + ds.curLevel + " " + data.length );
      //printData( "level 0", data );
      ds.coarser();
      data = ds.getFloats();
      System.out.println( ds.getName() + "@" + ds.curLevel + " " + ds.volume() );       
      //printData( "level 1", data );

      ds.coarser();
      data = ds.getFloats();
      System.out.println( ds.getName() + "@" + ds.curLevel + " " + ds.volume() );       
      //printData( "level 0", data );

   }
   public static void printData( String header, float[] data )
   {
      System.out.println( "-------------- " + header + " ---------------------");
      for ( int i = 0; i < data.length; i++ )
      {
         System.out.print( data[i] + "  " );
         if ( i % 8 == 7 ) System.out.println();
      }
      System.out.println( "------------------------------------------------------");
   }
}
