/* 
    MultiBlockCache.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    --------------------- Edit history as of 02/16/05 ------------------------
    10/21/05 rdb: deleted unused variables
    02/16/05 rdb: API changess
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import java.util.*;
//import java.io.*;

/**
	The MultiBlockCache is a BlockCache with an array of multiple CacheBlocks.
   
 */


class MultiBlockCache extends BlockCache{


    long [] blockAxisWeights;
    long [] tagAxisWeights;          // Controls mapping of n-D block space to 1D cache tags
    int [] cacheArrayDimensions;     // dimensionss of the cacheArray
     
    CacheDataBlock [] blockArray;
    int[] queryBoundsDimensions;
                
    // The following group of data members comprise the ISBounds
    // that is used to query the source indexable object.
                            
    int[] sQBLowerInts;
    int[] sQBUpperInts;
        
    ISBounds sourceQueryBounds;
    
    // The following group of data members comprise the ISBounds
    // that is used to query the cacheblock for block access. For
    // datum acess, datumCoords1 is used by itself.
    
    // Used for storing the intra-block datum coords
    int [] datumCoords1;             
    int [] datumCoords2;
    
    // This bounds denotes the subblock to be read from 
    // a cache block into the argument block. 
    ISBounds subblockQueryBounds;

    Object cacheLock = new Object(); // semaphore used in prefetching.

       
    MultiBlockCache( Indexable source, 
                     ISBounds iterationSpace, 
                     IndexSpaceID blockDimensions, 
                     IndexSpaceID cacheDimensions)
   {
        super(source, iterationSpace, blockDimensions, cacheDimensions);
        
        //ISBounds spaceBounds=source.copyBounds();
        int d=this.dimensionality;
        AxisOrdering partitioningOrdering=new AxisOrdering(d);
        AxisOrdering blockOrdering = new AxisOrdering(d);
        this.queryBoundsDimensions = new int[d];
        this.cacheArrayDimensions = cacheDimensions.getCoords();
        
        
        
        this.datumCoords1=new int[d];
        this.datumCoords2=new int[d];
        
        this.tagAxisWeights=partitioningOrdering.makeAxisWeights(this.partitioningDimensions);

        this.sQBLowerInts=new int[d];
        this.sQBUpperInts=new int[d];

        this.sourceQueryBounds = new ISBounds(this.sQBLowerInts, this.sQBUpperInts);
        this.subblockQueryBounds = new ISBounds(this.datumCoords1, this.datumCoords1);
        
        this.blockAxisWeights = blockOrdering.makeAxisWeights( cacheDimensions );
                
        this.blockArray = new CacheDataBlock[this.numCacheBlocks];
        RecordDescriptor rd = source.getRecordDescriptor();
        
        for ( int i = 0; i < this.numCacheBlocks; i++){
        
           this.blockArray[i] = new CacheDataBlock(this.cacheBlockBounds, rd);
        }

    }
   
    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.*/
   void lookup(Datum theDatum,IndexSpaceID id){
 
      int tag = 0;
      CacheDataBlock theBlock;
      int [] coords=id.refCoords();
      int d = this.dimensionality;
      int coordBlockIndex;
      int arrayIndex = 0;

//      int [] pcoords = new int[d];  // debugging

      for ( int i=0; i<d; i++ ){

         coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i];
         
//         pcoords[i] = coordBlockIndex;
         
         this.sQBLowerInts[i] = lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i];
         
         this.datumCoords1[i] = coords[i] - this.sQBLowerInts[i];
         
         arrayIndex +=  (coordBlockIndex % cacheArrayDimensions[i]) * this.blockAxisWeights[i];
         
         tag += coordBlockIndex * this.tagAxisWeights[i];
      }
        
      theBlock = this.blockArray[arrayIndex];
        
		if ( theBlock.getTag() == tag){
            //System.out.println("*Hit*");
            //AppMonitor.recordCacheHit();
            theBlock.datum(theDatum,this.datumCoords1);
        } else {
            System.out.println("*Miss*");
            //System.out.println("*Miss*: Retrieving tag number\t" + tag + "\tinto cache index "+ arrayIndex + " for ISID "+id+" with datumcoords ="+Arrays.toString(datumCoords1));
            
            //System.out.println(tag);
            //System.out.println(id);
            //AppMonitor.recordCacheMiss();
            for ( int i=0; i<d; i++){
                
                this.sQBUpperInts[i] = this.sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
                
            }
            
            this.sourceQueryBounds.assign(sQBLowerInts,sQBUpperInts);
            
            // The iterationSpaceBounds should already have been clipped against
            // the ds bounds, so clipping against it should be safe.
            this.sourceQueryBounds.intersectWith( iterationSpaceBounds );  

//            System.out.println("    *****Miss******: Cache block pcoords ="+ Arrays.toString(pcoords));
            
            theSource.subBasicBlock(theBlock, this.cacheBlockBounds, this.sourceQueryBounds);
            theBlock.setTag(tag);
            theBlock.datum(theDatum,this.datumCoords1);
        }
    }



    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block. This version provides scratch
      *  space for threaded operation, and exploits coherence where possible to reduce
      *  overhead.
      */
   void lookup(Datum theDatum,IndexSpaceID id, CacheContext context ){
 
      CacheDataBlock theBlock;
      int [] coords=id.refCoords();
      int d = this.dimensionality;
      int coordBlockIndex;
      boolean coherent = true;
      

      
      for ( int i=0; i<d; i++ ){
         
         coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i];
         
         context.datumCoords1[i] = coords[i] - (lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i]); // required
         
         if(coordBlockIndex != context.coordBlockIndex[i]){
         
            coherent = false;
         }
         
         context.coordBlockIndex[i] = coordBlockIndex;
      }
            
      if(!coherent){
      
         context.arrayIndex = 0;
         context.tag = 0;
         
         for ( int i=0; i<d; i++ ){
            
            context.arrayIndex +=  (context.coordBlockIndex[i] % cacheArrayDimensions[i]) * this.blockAxisWeights[i]; // required
            
            context.tag += context.coordBlockIndex[i] * this.tagAxisWeights[i];  // required
         }
      }
        
      theBlock = this.blockArray[context.arrayIndex];
        
		if ( theBlock.getTag() == context.tag ){
      
                  
         if(theBlock.pending){
         
            synchronized(this.cacheLock){ }
         }

         theBlock.datum(theDatum, context.datumCoords1);
      } else {
            System.out.println("*Miss*");
            //System.out.println("*Miss*: Retrieving tag number\t" + tag + "\tinto cache index "+ arrayIndex + " for ISID "+id+" with datumcoords ="+Arrays.toString(datumCoords1));
            
            //System.out.println(tag);
            //System.out.println(id);
            //AppMonitor.recordCacheMiss();
            for ( int i=0; i<d; i++){
               context.sQBLowerInts[i] = lowerIterationSpaceCoords[i] + context.coordBlockIndex[i] * cacheBlockDimensions[i]; // required
               context.sQBUpperInts[i] = context.sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
                
            }
            
            context.sourceQueryBounds.assign(context.sQBLowerInts, context.sQBUpperInts);
            
            // The iterationSpaceBounds should already have been clipped against
            // the ds bounds, so clipping against it should be safe.
            context.sourceQueryBounds.intersectWith( iterationSpaceBounds );  

//            System.out.println("    *****Miss******: Cache block pcoords ="+ Arrays.toString(pcoords));
            
            theSource.subBasicBlock(theBlock, this.cacheBlockBounds, context.sourceQueryBounds);
            theBlock.setTag(context.tag);
            theBlock.datum(theDatum, context.datumCoords1);
        }
    }



    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.*/
   void lookup(Datum theDatum,IndexSpaceID id, int [] sQBLowerInts, int [] sQBUpperInts, int [] datumCoords1, ISBounds sourceQueryBounds){
 
      int tag = 0;
      CacheDataBlock theBlock;
      int [] coords=id.refCoords();
      int d = this.dimensionality;
      int coordBlockIndex;
      int arrayIndex = 0;
      


//      int [] pcoords = new int[d];  // debugging

//      for ( int i=0; i<d; i++ ){
//         
//         coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i]; // required
//         
////         pcoords[i] = coordBlockIndex;
//         
//
//         sQBLowerInts[i] = lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i]; // required
//         
//         datumCoords1[i] = coords[i] - sQBLowerInts[i]; // required
//         
//         arrayIndex +=  (coordBlockIndex % cacheArrayDimensions[i]) * this.blockAxisWeights[i]; // required
//         
//         tag += coordBlockIndex * this.tagAxisWeights[i];  // required
//      }



      for ( int i=0; i<d; i++ ){
      
         coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i];
         
         datumCoords1[i] = coords[i] - (lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i]); // required
         
         arrayIndex +=  (coordBlockIndex % cacheArrayDimensions[i]) * this.blockAxisWeights[i]; // required
         
         tag += coordBlockIndex * this.tagAxisWeights[i];  // required
      }
      
        
      theBlock = this.blockArray[arrayIndex];
        
		if ( theBlock.getTag() == tag ){
      
                  
         if(theBlock.pending){
         
            synchronized(this.cacheLock){
//               try{
//               
//                  this.cacheLock.wait();
//               } catch(InterruptedException e){
//               
//                  new SDBError(SDBError.Warning, "MultiblockCache.lookup(): interrupted while waiting for pending block.");
//               }
            }
         }

         theBlock.datum(theDatum, datumCoords1);
      } else {
            System.out.println("*Miss*");
            //System.out.println("*Miss*: Retrieving tag number\t" + tag + "\tinto cache index "+ arrayIndex + " for ISID "+id+" with datumcoords ="+Arrays.toString(datumCoords1));
            
            //System.out.println(tag);
            //System.out.println(id);
            //AppMonitor.recordCacheMiss();
            for ( int i=0; i<d; i++){
               coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i]; // required
               sQBLowerInts[i] = lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i]; // required
               sQBUpperInts[i] = sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
                
            }
            
            sourceQueryBounds.assign(sQBLowerInts,sQBUpperInts);
            
            // The iterationSpaceBounds should already have been clipped against
            // the ds bounds, so clipping against it should be safe.
            sourceQueryBounds.intersectWith( iterationSpaceBounds );  

//            System.out.println("    *****Miss******: Cache block pcoords ="+ Arrays.toString(pcoords));
            
            theSource.subBasicBlock(theBlock, this.cacheBlockBounds, sourceQueryBounds);
            theBlock.setTag(tag);
            theBlock.datum(theDatum, datumCoords1);
        }
    }
    
    
   /** If the cacheblock corresponding to the partition specified by the id 
    *  is not present in the cache, then add the cacheblock to the stencil, 
    *  This information can 
    *  later be used for a bulk read.
    *  Package Access Only.
    */
   void addToStencil(DSBlockStencil stencil, IndexSpaceID id){
   
   
      int tag = 0;
      CacheDataBlock theBlock;
      int [] pcoords=id.refCoords(); // coordinates in partitioning space
      int d = this.dimensionality;
      //int coordBlockIndex;
      int arrayIndex = 0;
      
      if( !stencil.inPartitioningBounds(id) ){
      
         return;
      }
      

      for ( int i=0; i<d; i++ ){

         arrayIndex +=  (pcoords[i] % cacheArrayDimensions[i]) * this.blockAxisWeights[i];
         tag += pcoords[i] * this.tagAxisWeights[i];
      }
        
      theBlock = this.blockArray[arrayIndex];
        
		if ( theBlock.getTag() != tag){
      
//         try {
      
            if(!theBlock.pending && stencil.setElement(id, theBlock)) {   
            
//               System.out.println("Adding "+Arrays.toString(id.refCoords())+" to stencil.     tag="+tag + " arrayIndex="+arrayIndex);
            
               theBlock.setTag(tag);
               theBlock.pending = true;
            } else {
            
               System.out.println("  *** Conflict in addToStencil while adding "+Arrays.toString(id.refCoords())+" to stencil.     tag="+tag + " arrayIndex="+arrayIndex);
               
               if(theBlock.pending){
               
                  System.out.println("  ***     CacheBlock already pending with tag="+theBlock.getTag() + " pCoords=" + Arrays.toString(tagToPartitioningCoords(theBlock.getTag(), new int [3])));
               } else {
               
                  System.out.println("  ***     Stencil Element already in use.\n");
               }
            }
         
//         } catch (ArrayIndexOutOfBoundsException e){
//         
//            System.out.println("addToStencil(): OOB Exception");
//         }
      
      }

   }



   /** Ask the datasource to retrieve the data contained in each bounds specified
    *  by the stencil, and place it in the corresponding datablocks.
    */
   void prefetch(DSBlockStencil stencil){

      CacheDataBlock cdb;
      int size = stencil.size();
            
      if( size > 0 ) {         

         synchronized (this.cacheLock) {

            theSource.subblock(stencil);
            
            // Set pending flags to false, since I/O is now complete
            for( int i=0; i<size; i++){
            
               cdb = (CacheDataBlock) stencil.getIndexedBlock(i);
            
               cdb.pending=false;
            }
         
//            this.cacheLock.notify();
         }
      }
   }
    
    
    
    /** Retrieves the block from the Cache or directly from the datasource, refreshing the
        cache with a new block.*/
    void lookup(BasicBlock queryBlock, ISBounds dstBounds, ISBounds userQueryBounds){
 
        int i, d=this.dimensionality;
        int lowerTag=0, upperTag=0;
        int arrayIndex=0;
        int lowerCoordBlockIndex, upperCoordBlockIndex;
        int [] lowerCoords=userQueryBounds.getLowerArray();
        int [] upperCoords=userQueryBounds.getUpperArray();
        CacheDataBlock theBlock;
                
        userQueryBounds.getDimensions(this.queryBoundsDimensions);
        
        for(i=0;i<d;i++){

            lowerCoordBlockIndex = (lowerCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];
            upperCoordBlockIndex = (upperCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];

            
            this.sQBLowerInts[i] = lowerCoordBlockIndex * this.cacheBlockDimensions[i] + lowerIterationSpaceCoords[i];         
            
            this.datumCoords1[i] = lowerCoords[i] - this.sQBLowerInts[i];
            this.datumCoords2[i] = this.datumCoords1[i] + this.queryBoundsDimensions[i] - 1;
            
            arrayIndex +=  (lowerCoordBlockIndex % cacheArrayDimensions[i]) * this.blockAxisWeights[i];
            
            lowerTag += lowerCoordBlockIndex * this.tagAxisWeights[i];
            upperTag += upperCoordBlockIndex * this.tagAxisWeights[i];
        }
       
        
        
        theBlock = blockArray[arrayIndex];

        if(lowerTag == upperTag){
        
            subblockQueryBounds.assign(this.datumCoords1, this.datumCoords2);

            if(theBlock.getTag() == lowerTag){
                //System.out.println("*Hit*");
                //AppMonitor.recordCacheHit();
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            } else {
                //System.out.println("*Miss*: Retrieving block number " + lowerTag);
                //AppMonitor.recordCacheMiss();
                
                
                for(i=0;i<d;i++){
                    
                    this.sQBUpperInts[i] = this.sQBLowerInts[i] + this.cacheBlockDimensions[i] - 1;
                    
                    //this.sQBUpperInts[i] = (this.sQBUpperInts[i] > this.upperIterationSpaceCoords[i]) ? this.upperIterationSpaceCoords[i] : this.sQBUpperInts[i];
                }
                
                this.sourceQueryBounds.assign(sQBLowerInts,sQBUpperInts);
                
                
                // The iterationSpaceBounds should already have been clipped against
                // the ds bounds, so clipping against it should be safe.
                this.sourceQueryBounds.intersectWith( iterationSpaceBounds );  

    
                theSource.subBasicBlock(theBlock, this.cacheBlockBounds,sourceQueryBounds);
                theBlock.setTag(lowerTag);
                
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            }
            
        } else {
        
            //System.out.println("You're in da weeds, Lorna! ;-)");
            
            this.theSource.subBasicBlock(queryBlock, dstBounds, userQueryBounds);
        
        }
    }

    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(BasicBlock queryBlock, ISBounds userQueryBounds){
    
        this.lookup(queryBlock, queryBlock.getBounds(), userQueryBounds);
    }
    

}
