package edu.unh.sdb.datasource;

import java.util.Arrays;
import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

/**
 *
 * A plane that can be iterated through a {@link Volume} and
 * can be rotated and translated.
 * @author dlipsa
 */
public class Plane
{
   
   /** Creates a new instance of Plane */
   public Plane (double a, double b, double c, double d)
   {
      setEquation (a, b, c, d);
   }
   
   public Plane ()
   {}
   
   public Vector3d getNormal ()
   {
      return normal;
   }
   
   public void setEquation (double a, double b, double c, double d)
   {
      equation[0] = a;
      equation[1] = b;
      equation[2] = c;
      equation[3] = d;
      normal.set (equation);
   }
   
   public void set (int i, double x)
   {
      equation[i] = x;
      if (i < 3) normal.set (equation);
   }
   
   
   public double get (int i)
   {
      return equation[i];
   }
   
   
   public void startIteration (Volume volume)
   {
      volume.movePlaneBeginIteration (this);
   }
   
   public void prev (double step)
   {
      equation[3] += step;
   }
   
   public void next (double step)
   {
      equation[3] -= step;
   }
   
   public boolean validPosition (Volume volume)
   {
      return volume != null && volume.validPosition (this);
   }

   
   public void normalize ()
   {
      equation[3] /= normal.length ();
      normal.normalize ();
      setFromNormal ();
   }
   
   
   public void normalize (int numSteps)
   {
   
      float [] normalArray = new float[] {
         (float)normal.x, (float)normal.y, (float)normal.z};
   
      // This method will set any components of the normal that are sufficiently
      // small to zero.  The implementation assumes that the points are centered 
      // in the middle of a index space element.
      ObliquePlaneIterator.setEffectiveZeros(normalArray, numSteps);
   
      normal = new Vector3d(normalArray[0], normalArray[1], normalArray[2]);
   
      equation[3] /= normal.length ();
      normal.normalize ();
      setFromNormal ();
   }
   
   public void translate (Vector3d translation)
   {
      Vector3d n = getNormal ();
      equation[3] += n.dot (translation);
   }
         
   public void transform (Matrix3d rot)
   {
      rot.transform (normal);
      setFromNormal ();
   }
   
   @Override
   public String toString ()
   {
      return Arrays.toString (equation);
   }
   
   private void setFromNormal ()
   {
      equation[0] = normal.x;
      equation[1] = normal.y;
      equation[2] = normal.z;
   }
   
   private double[] equation = new double[4];
   private Vector3d normal = new Vector3d ();
}
