
//
//  PartitionedBlockIterator.java
//
//  Created by Xuan Tang  on Sun Feb 1 2004.
//  Copyright (c) 2004 UNH. All rights reserved.
//
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables or added dummy refs in main
//-----------------------------------------------------------------------

package edu.unh.sdb.datasource;

//import java.io.*;
import java.util.*;
import edu.unh.sdb.common.*;
//import edu.unh.sdb.datasource.*;


/**
    this class generates a BlockIterator that randomly moves within
    this ISBounds of the iteration data source. Iteration block does not
    partially overlap  with the datablock of previous iteration steps.
    If setEnumWindow( int size) is called, enumeration window will be used 
    during the iteration process. That means ISbounds of future iteration steps
    will be generated. 
    
    The iteration space is partitioned based on the size of iteration block. 
*/

//   NOTE : variable name started with "partSpace" represents variables in 
//   partitioned space. otherwise all the variables are defined in the
//   grid space. 

public class RandomBlockIterator // may extends dataBlock or iterator
{
    // following attributes should be private data members of a parent class
    // SDBIterator
    private DataBlock           iterationBlock;
    private Indexable           source;
    private IndexSpaceID        identityID, tempID;
    private ISBounds            iterationBounds, tempBounds;

    private ISBounds            blockBounds;  
 
    private int[]               sourceDim;
    private int[]               blockDim;
    private int[]               partitionDim;
   
    // random number generator
    private Random              random;

    private int                 numOfDim;              
    private int[]               spaceID;
    private int                 i;
 

    /** indicate if the enumeration window is open. if the window is 
        open, the iteration has to follow the iteration sequence within
        the enumberation window, otherwise the iteration follows the 
        path generated by the random indexSpaceID generator
     */
    private boolean             isWindowOpen;
    private ISBounds[]          enumerationWindow;
    private int                 activeWindowIndex;
    private int                 windowSize;
  
    private RegularISPartitioning partition;
    private ISBounds partSpaceBounds;
    
    private boolean windowReSized;

  /** The partSpaceBounds is the iterationBounds in the partitioned space
   */
  public RandomBlockIterator( ISBounds partitionSpaceBounds,
                                   ISBounds iterationBounds,
                                   Indexable iterationSource )
   {
       this.source = iterationSource;
       this.iterationBounds = iterationBounds.copy();
       this.partition = new RegularISPartitioning( partitionSpaceBounds,
                                                       iterationBounds );

       this.partSpaceBounds = partitionSpaceBounds.copy();
      
       this.blockBounds = this.partition.getGeometricBounds(); 
       this.isWindowOpen = false;
    
       this.numOfDim  = this.source.dim();
       this.spaceID   = new int[numOfDim];

       this.sourceDim = new int[numOfDim];
       this.blockDim  = new int[numOfDim];
       this.partitionDim = new int[numOfDim];
       this.tempID = new IntegerIndexSpaceID( this.partitionDim );
       this.partSpaceBounds.getDimensions(partitionDim);
       //System.out.println("block bounds is :" + this.blockBounds);

       this.iterationBlock = new BasicBlock( this.blockBounds,
                                            source.getRecordDescriptor() );
       this.random = new Random();
       windowReSized = false;
  }

   /** get the bounds of this block iterator
    */
    public ISBounds getIteratorBounds()
    {
         return this.blockBounds;
    }

    /** initialize the iterator, read data values from the source into 
        the iteration block.

        Currently I use BasicBlock as the container 
        BasicBlock uses the default axisOrdering 
    */
    public void init()
    {
        this.next();
    }

    /** Set the iteration data source 
     */
    public void setSource(DataSource iterationSource)
    {
        this.source = iterationSource;
    }

    /** set the enumberation window size */
    public void setWindowSize(int size)
    {
       this.windowSize = size;
       this.enumerationWindow = new ISBounds[size];
       this.activeWindowIndex =0;
       this.isWindowOpen = true;
       this.windowReSized = true;
    }
   
   /** move to next iteration step
    */
    public void next()
    {
        if(isWindowOpen)
            this.next_windowOpen();
        else
            this.next_windowClose();
     }
    /** return if the enumeration window is open or closed
     */
    public boolean isWindowOpen()
    {
        return isWindowOpen;
    }

    /** move the iterator to next iteration step, use enumeration window
     */
    private void next_windowOpen()
    {
        if( this.windowReSized || activeWindowIndex >= this.windowSize)
        {
           //tmpBounds = this.nextBounds();
           //System.out.println("queryBounds:" + tmpBounds ); 
           //System.out.println("dst Bounds :" + this.blockBounds);
           //source.subblock(iterationBlock, this.blockBounds, tmpBounds);
           //source.subblock(iterationBlock, this.blockBounds, this.nextBounds() );
           this.renewEnumWindow();
           activeWindowIndex =0;
           this.windowReSized = false;
        }
        source.subblock( iterationBlock,
                          this.blockBounds, 
                          enumerationWindow[activeWindowIndex++]
                        );
    }

    /** move the iterator to next iteration step with enumeration window closed
     */
    private void next_windowClose()
    {
         source.subblock(iterationBlock, this.blockBounds, this.nextBounds() );
     }


   /** preproces the iteration
    */
    public void renewEnumWindow()
    {
      for( int a =0; a< windowSize; a++){
        enumerationWindow[a] = this.nextBounds();
      }
      this.activeWindowIndex =0;
    }
/*

    public void renewEnumWindow()
    {
      int[] l = {0,0,0};
      int[] u = {2,2,2};
      ISBounds queryBounds = new ISBounds( new IntegerIndexSpaceID(l),
                                             new IntegerIndexSpaceID(u));
      int[] l2 = {0,0,0};
      int[] u2 = {2,2,2};
      ISBounds queryBounds = new ISBounds( new IntegerIndexSpaceID(l),
                                             new IntegerIndexSpaceID(u));
        
      this.getEnumWindow();
    }


*/
  public ISBounds[] getEnumWindow()
   {
      //this.renewEnumWindow();
      return enumerationWindow;
   }


  /** return the ISBounds of an indexed enumeration
      window element
   */
  public ISBounds getWindowElement(int windowIndex )
   {
       return enumerationWindow[windowIndex];
   }

 
  public int getEnumWindowIndex()
  {
      return activeWindowIndex;
  } 

  // this method is used for debugging 
/*   public ISBounds[] getEnumWindow()
   {
      //int[] l = {0,0,0};
      //int[] u = {2,2,2};
      //ISBounds queryBounds = new ISBounds( new IntegerIndexSpaceID(l),
      //                                       new IntegerIndexSpaceID(u));
    
      for( int a =0; a< windowSize; a++){
        enumerationWindow[a] = this.blockBounds;
      }
      activeWindowIndex =0;
      return enumerationWindow;
   }
*/


    /** get ISBounds of next iteration step
     */
    public ISBounds nextBounds()
    {
        this.getRandomISID(tempID);
        return partition.mapToSpace( tempID);
    } 
 

    /** generate a random IndexSpaceID in the partitioned space
     */
    private void getRandomISID( int[] indexSpaceIDs )
    {
        for( i=0 ;i< this.numOfDim ; i++)
         {
             indexSpaceIDs[i] = random.nextInt( partitionDim[i]) ;
         }
    }
    
    /** get a random IndexSpaceID in the partitioned space from
        random number  generator 
    */
    private void getRandomISID( IndexSpaceID id)
    {
        for( i=0 ;i< this.numOfDim ; i++)
         {
             id.setCoord(i,random.nextInt( partitionDim[i]) );
         }
    }
    public String toString()
    {
         return this.iterationBlock.toString();
    }

   
    public static void main( String[] args)
    {

      PhysicalDataSource ds =(PhysicalDataSource)DataSource.create(
                                                 "foo", "/home/csg/xtang/file/simplefile.bin.xml");
      ds.activate();

      RecordDescriptor rec = ds.getRecordDescriptor();
      System.out.println("rec " + rec);
      System.out.println("ds " + ds);

      int[] lower = {0,0,0};
      int[] upper = {3,3,3};
      ISBounds partBounds = new ISBounds( new IntegerIndexSpaceID(lower),
                                             new IntegerIndexSpaceID(upper));
      
      
      RandomBlockIterator iter = new RandomBlockIterator( partBounds,ds.getBounds(), ds );
      iter.setWindowSize(10);
      ISBounds iterBounds = iter.getIteratorBounds();

      for( int i =0; i< 10; i++)
      {
 
           System.out.println( iter.nextBounds() );
           iter.next();
           System.out.println( iter);
      }
 
      BlockListCacheDataSource cacheSource = new BlockListCacheDataSource(
                              "listCache", ds, ds.getBounds(), iterBounds, 10 );
      cacheSource.activate();
      cacheSource.setIterator(iter);

      for( int i =0; i< 10; i++)
      {
           System.out.println( iter.nextBounds() );
           iter.next();
           System.out.println( iter);
       }
 
      //------ following are dummy references to avoid compiler warnings
      RandomBlockIterator rbi = new RandomBlockIterator( null, null, null );
      if ( rbi.blockDim == null || rbi.tempBounds == null || rbi.spaceID == null
           || rbi.sourceDim == null || rbi.identityID == null 
           || rbi.blockBounds == null || rbi.iterationBounds == null )
         rbi.getRandomISID( (int[]) null );
    }
 }
    
