/*
 * LinesCuboidIterator.java
 *
 * Created on March 7, 2007, 10:35 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.SDBIterator;
import java.util.ArrayList;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Iterates through an arbitrary oriented cuboid. At each step a new ray
 * perpendicular on the base of the cuboid is returned. The iterations continues
 * untill the whole based is finished.
 * @author dlipsa
 */
public class RayCastingIterator implements SDBIterator
{
   
   /**
    * Start with a cuboid oriented along the Z axis, the base at z=-height/2 
    * and the top at z = height/2.  The 'center' of the cuboid is at origin.
    * Rotate the cuboid to get to 'initialOrientation'. Usually, this is a no-operation.
    * Rotate the cuboid with 'rotation' and then move the center to 'start'.
    */
   public RayCastingIterator (DataSource dataSource,
      Vector3d initialOrientation, 
      Matrix4d rot,
      Point3d start, 
      double baseX, double baseY, double height)
   {
      final Vector3d Z_AXIS = new Vector3d (0, 0, 1);
      final Vector3d Y_AXIS = new Vector3d (0, 1, 0);
      this.rotation = new Matrix4d ();
      SliceUtil.calculateRotation (Z_AXIS, initialOrientation, this.rotation);
      this.rotation.mul (rot, this.rotation);
      this.viewingDirection = new Vector3d (Z_AXIS);
      this.rotation.transform (this.viewingDirection);
      this.sliceNormal = new Vector3d (Y_AXIS);
      this.rotation.transform (this.sliceNormal);      
      this.start = start;
      this.baseX = baseX;
      this.baseY = baseY;
      this.height = height;
      this.dataSource = dataSource;
      beginX = - baseX / 2;
      endX = beginX + Math.floor (baseX / step) - 1;
      beginY = - baseY / 2;
      endY = beginY + Math.floor (baseY / step) - 1;
      iterationBoundingBox = calculateIterationBoundingBox ();
   }
   
   public void init ()
   {
      currentX = beginX;
      currentY = beginY;
      setCurrentPoints ();
      currentRay = new RayIterator(dataSource, begin, end, blockSize, step);                  
      DynamicChunking.logger.info (
         "RayIterator runtime class: " + currentRay.getClass ().toString () + 
         "\nblock size: " + blockSize +
         "\nstep: " + step +
         "\nbackToFront: " + backtoFront);
   }
   
   public boolean next ()
   {
      currentX += step;
      if (currentX > endX)
      {
         currentX = beginX;
         currentY += step;
         if (currentY > endY)
            return false;
      }
      setCurrentPoints ();
      currentRay = new RayIterator (dataSource, begin, end, blockSize, step);      
      return true;
   }
   
   public boolean valid ()
   {
      return currentX <= endX && currentY <= endY;
   }
   
   public boolean lastX ()
   {
      return currentX == endX;
   }
   
   public boolean lastY ()
   {
      return currentY == endY;
   }
   
   public void end ()
   {
      
   }
   
   public ISBounds getIterationBoundingBox ()
   {
      return iterationBoundingBox;
   }
   
   public RayIterator getCurrentRay ()
   {
      return currentRay;
   }
      
   public double getBaseX ()
   {
      return baseX;
   }
   
   public double getBeginX ()
   {
      return beginX;
   }
   
   public double getEndX ()
   {
      return endX;
   }
   
   public double getCurrentX ()
   {
      return currentX;
   }
   
   public double getBeginY ()
   {
      return beginY;
   }
   
   public double getEndY ()
   {
      return endY;
   }
   
   public double getCurrentY ()
   {
      return currentY;
   }
      
   public double getBaseY ()
   {
      return baseY;
   }
   
   public double getHeight ()
   {
      return height;      
   }
   
   public Point3d getStart ()
   {
      return start;
   }
   
   public Vector3d getBaseNormal ()
   {
      Vector3d v = new Vector3d (0, 0, 1);
      rotation.transform (v);
      return v;
   }
   
   public Matrix4d getRotation ()
   {
      return rotation;
   }
   
   public void setBlockSize (IndexSpaceID blockSize)
   {
      this.blockSize = blockSize;
   }

   public void setStep (double step)
   {
      this.step = step;
   }
   
   public double getStep ()
   {
      return step;
   }
   
   public boolean isBacktoFront ()
   {
      return backtoFront;
   }

   public void setBacktoFront (boolean backtoFront)
   {
      this.backtoFront = backtoFront;
   }   
   
   // returns a vector that points toward the viewer.
   public Vector3d getViewingDirection ()
   {
      return viewingDirection;
   }
   
   public Vector3d getSliceNormal ()
   {
      return sliceNormal;
   }
   
   public ArrayList<Point3d> getCornersDeterminingBb ()
   {
      return cornersDeterminingBb;
   }
   
   private void transform (Point3d p)
   {      
      rotation.transform (p);
      p.add (start);            
   }
   
   public void setDataSource (DataSource dataSource)
   {
      this.dataSource = dataSource;
   }   
   
   private void setCurrentPoints ()
   {
      double beginZ = -height / 2;
      double endZ = beginZ + height / step;
      if (backtoFront)
      {
         double temp = beginZ;
         beginZ = endZ;
         endZ = temp;
      }
      begin.set (currentX, currentY, beginZ);
      transform (begin);
      end.set (currentX, currentY, endZ);            
      transform (end);
   }

   private ISBounds calculateIterationBoundingBox ()
   {
      Point3d[] corners = {
         new Point3d (-baseX/2, -baseY/2, -height/2), 
         new Point3d (baseX/2, -baseY/2, -height/2),
         new Point3d (-baseX/2, baseY/2, -height/2), 
         new Point3d (baseX/2, baseY/2, -height/2),

         new Point3d (-baseX/2, -baseY/2, height/2), 
         new Point3d (baseX/2, -baseY/2, height/2),
         new Point3d (-baseX/2, baseY/2, height/2), 
         new Point3d (baseX/2, baseY/2, height/2)         
      };
      ISBounds iterationBb = new ISBounds (3);
      for (int axis = 0; axis < iterationBb.getLower ().dim (); axis++)
      {
         iterationBb.setLower (axis, Integer.MAX_VALUE);
         iterationBb.setUpper (axis, Integer.MIN_VALUE);
      }
      int[] corner = new int[3];
      for (int i = 0; i < corners.length; i++)
      {
         transform (corners[i]);
         corner[0] = (int)Math.round (corners[i].x);
         corner[1] = (int)Math.round (corners[i].y);
         corner[2] = (int)Math.round (corners[i].z);
         for (int axis = 0; axis < corner.length; axis++)
         {
            if (corner[axis] < iterationBb.getLower (axis))
            {
               iterationBb.setLower (axis, corner[axis]);
            }
            if (corner[axis] > iterationBb.getUpper (axis))
            {
               iterationBb.setUpper (axis, corner[axis]);
            }
         }
      }
      for (Point3d c : corners)
      {
         if (
            ((int)Math.round(c.x)) == iterationBb.getLower (0) ||
            ((int)Math.round(c.y)) == iterationBb.getLower (1) ||
            ((int)Math.round(c.z)) == iterationBb.getLower (2) ||
            ((int)Math.round(c.x)) == iterationBb.getUpper (0) ||
            ((int)Math.round(c.y)) == iterationBb.getUpper (1) ||
            ((int)Math.round(c.z)) == iterationBb.getUpper (2)
            )
         {
            cornersDeterminingBb.add (c);
         }
         else
         {
            DynamicChunking.logger.info ("corner NOT in BB: " + c);
         }            
      }
      DynamicChunking.logger.info ("Bounding Box: " + iterationBb);
      return iterationBb;
   }
   
   private Vector3d viewingDirection;
   private Vector3d sliceNormal;
   private ISBounds iterationBoundingBox;
   private ArrayList<Point3d> cornersDeterminingBb = new ArrayList<Point3d> (8);
   private DataSource dataSource;
   private double baseX, baseY, height;
   private Point3d start;
   private Matrix4d rotation;
   
   private Point3d begin = new Point3d ();
   private Point3d end = new Point3d ();
   private double beginX, beginY, endX, endY, currentX, currentY;
   private RayIterator currentRay;
   private IndexSpaceID blockSize = new IntegerIndexSpaceID(1,1,1);
   private double step = 1;
   private boolean backtoFront = false;
}
