/*
    RegDellDataSource.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/

//------------------- Edit history --------------------------------------
// 12/09/05 rdb: all createDataBlock calls changed to copy bounds to origin first.
// 11/13/05 rdb: removed uses of CompositeBlock
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------


/** 
 * RegCellDataSource class implements the features to access the data of a 
 * regular rectilinear cell data source. It contains a point datasource 
 * corresponding to the regular rectilinear cell data source. 
 */

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import edu.unh.sdb.datasource.*;
import edu.unh.sdb.lattice.*;
//import java.io.*;


public class RegCellDataSource extends CellDataSource implements CellIndexable{
    private DataSource ptDS; //store a point data source
    private DimResolution resolution;
       //store the dimensional resolution of a cell data source
    private int[] path;
       //store the path of a uniform cell data source in a ARRS tree

    /** Constructor sets all data members to copies of the
     *  corresponding members from a point DataSource corresponding
     *  to this cell data source.
     */
    public RegCellDataSource (DataSource ds) 
    { 
       super(ds); 
       this.ptDS = ds;
       this.bounds = this.ptDS.getBounds().copyReducedBy(1);
       this.resolution = null;
       this.path = null;
       //this.activate();
    }

    /** The DataSource will perform any initialization required
     *  in order to begin reading/processing data.
     */
    public void activate(){ this.ptDS.activate(); }


    /** Get the point data source relative to a cell data souce */
    public DataSource getPointDS(){ return this.ptDS; }

    /** Set the resolution */
    public void setDimResolution(DimResolution resolution)
    {
       this.resolution = resolution;
    }
 

    /** Set the path of an uniform cell datasource in a ARRS tree */
    public void setPath(int[] path)
    {
       this.path = path.clone();
    }
 
    /** Get the path of an uniform region cell datasource in a ARRS tree */
    public int[] getPath()
    {
       return this.path;
    }

    /** Get the dimensional resolution */
    public DimResolution getDimResolution()
    {
       return this.resolution;
    }

 
    /** Get the corners of a cell */
    public IndexSpaceID[] getCorners(IndexSpaceID cellISID)
    {
       int i;
       IndexSpaceID[] corners;
       ISBounds bound;

       if (!this.validID(cellISID)) {
          new IllegalArgumentError(SDBError.Fatal,"cellISID argument has the wrong value.");

          System.exit(1);
       }
 
       //get the bounds of the cell
       bound = this.getPointBounds(cellISID);

       //set the size of the array of the corners
       corners = new IndexSpaceID[1 << this.dim()];   

       //create a iterator object
       ISIterator iter = new ISIterator(bound);

       //traverse a block to get all the corner points
       i = 0;
       int num = 0;
       IndexSpaceID tempID;
       for (iter.init(); iter.valid(); iter.next()) {
          tempID = iter.copyID();
          if (this.cornerPointCheck(tempID, bound)) {
             corners[i] = tempID;
             i++;
          }
          num++;
       }

       return corners;
    }


    /* check if a cell id is valid */
    private boolean validID(IndexSpaceID id)
    {
       IndexSpaceID low, up;
       int d = this.bounds.dim();

       low = this.bounds.copyLower();
       up = this.bounds.copyUpper();
   
       if (d != id.dim()) 
          return false;

       for (int i = 0; i < d; i++) {
          if((low.getCoord(i)>id.getCoord(i)) || (up.getCoord(i)<id.getCoord(i)))
             return false;
       }

       return true;
    }

 
    /* check if a IndexSpaceID is a corner point. a helper for the function
      Cell(ISBounds bound) */
    private boolean cornerPointCheck(IndexSpaceID id, ISBounds b)
    {
       IndexSpaceID low, up;
       low = b.copyLower();
       up = b.copyUpper();

       for ( int i = 0; i < id.dim(); i++) {
          if ((id.getCoord(i)!=low.getCoord(i)) && (id.getCoord(i)!=up.getCoord(i)))
             return false;
       }
       return true;
    }
 

    /* Get the point bounds of a cell */
    private ISBounds getPointBounds(IndexSpaceID cellISID)
    {
       int d, i;
       int[] axes;
       IndexSpaceID low, up;
       ISBounds bound;

       if (!this.validID(cellISID)) {
          new IllegalArgumentError(SDBError.Fatal,"cellISID argument has the wrong value.");
          System.exit(1);
       }

       //get the low index id of the specified cell
       low = cellISID;

       d = cellISID.dim();
       axes = new int[d];
       for (i = 0; i < d; i++) {
          axes[i] = cellISID.getCoord(i) + 1;
       }

       //get the up index id of the specified cell
       up = new IntegerIndexSpaceID(axes);
       //get the bounds of the cell
       bound = new ISBounds(low, up);
      
       return bound;
    }


    /** Get the datums of the corners in a cell */
    public Datum[] getCornerDatums(IndexSpaceID cellISID)
    {
       IndexSpaceID[] index;
       int dim = cellISID.dim();
       Datum[] datums;

       if (!this.validID(cellISID)) {
          new IllegalArgumentError(SDBError.Fatal,"cellISID argument has the wrong value.");
          System.exit(1);
       }

       datums = new Datum[1 << dim];
       index = this.getCorners(cellISID);
       for ( int i = 0; i < index.length; i++) {
          datums[i] = this.ptDS.datum(index[i]);
       }

       return datums;
    }

    /** Get the point block of a cell */
    public DataBlock getCellData(IndexSpaceID cellISID)
    {
       ISBounds bound;

       if (!this.validID(cellISID)) {
          new IllegalArgumentError(SDBError.Fatal,"cellISID argument has the wrong value.");
          System.exit(1);
       }

       //get the bounds of the cell
       bound = this.getPointBounds(cellISID);
       return this.ptDS.subblock(bound);
    }

    /** Get a cell in the cell data source */
    public Cell getCell(IndexSpaceID cellISID)
    {
       if (!this.validID(cellISID)) {
          new IllegalArgumentError(SDBError.Fatal,"cellISID argument has the wrong value.");
          return null;
       }
      
       //create a cell id 
       ARRSCellID cellid = new ARRSCellID(cellISID);
       //get the base bounds of a cell
       ISBounds bd = this.getPointBounds(cellISID);
       //create a rectilinear cell with the cell id and the demensionality 
       ARRSCell cell = new ARRSCell(cellid, bd.dim());
       //set the base bounds to the rectilinear cell
       cell.setBaseBounds(bd);
       //get the corner data of the cell
       Datum[] corners = this.getCornerDatums(cellISID); 
       cell.setCellData(corners);
       Point[] points = new Point[4];
       for (int i = 0; i < 4; i++) 
          points[i] = new Point(corners[i]);

       cell.setCellVertices(points); 
       return cell;
    } 


    /** The DataSource will release any resources used during the
                readling or processing of data.*/
    public void deactivate()
    {  this.ptDS.deactivate(); }

    /** Return the Datum corresponding to the given IndexSpaceID. */ 
    public Datum datum(IndexSpaceID did) 
    {  return null; } 

    /** Return the Datum corresponding to the given IndexSpaceID. The Datum 
        will only contain attributes specified by the RecordSpec argument.
    */
    public Datum datum(IndexSpaceID did, RecordSpec spec)
    { 
       return null; 
    }


    /** PBR: Return by reference the Datum corresponding to the given
     *  IndexSpaceID.
     */
    public void datum(Datum dtm, IndexSpaceID did)
    { 
        return;
    }

    /** PBR: Return by reference the Datum corresponding to the given
     *  IndexSpaceID. Only datum fields referred to by the FieldIDMapper
     *  are modified.
     */
    public void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp)
    {
        return;
    }

    /** PBR: Return by reference the Datum corresponding to the given
     *  IndexSpaceID. Only datum fields referred to by the RecordSpec
     *  are modified.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec)
    { 
        return;
    }

    /** Return a new DataBlock that is a subset of this datablock. */
    public  DataBlock subblock( ISBounds db )
    {
       if (this.ptDS == null)
          return null;
       //get the point bounds corresponding to the cell bounds
       ISBounds bd = db.copyIncreasedBy(1);
       if (!this.ptDS.getBounds().contains(bd)) {
          new IllegalArgumentError(SDBError.Recoverable,"The value of the bounds is invalid.");
          return null;
       }
       
       //get the point data block corresponding to the cell bounds
       //rdb CompositeBlock ptblock;
       DataBlock ptblock;
       ISBounds bd2;
       //rdb ptblock = new CompositeBlock(bd, this.ptDS.getRecordDescriptor());
       ptblock = this.ptDS.getRecordDescriptor().createDataBlock( bd.copyToOrigin() );
       bd2 = ptblock.getBounds();
       this.ptDS.subblock(ptblock, bd2, bd);

       //create a uniform cell block
       URCellBlock ublock = new URCellBlock(db, this.recordDesc);
       //set a point data block to the uniform cell block
       ublock.setPointDataBlock(ptblock);
                                                                                
       return ublock;
    }

    /** Reads a subblock defined by srcBounds into the region of the 
        Datablock argument defined by dstBounds. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                         ISBounds srcBounds)
    {  
       ISBounds pointSrcBounds, pointDstBounds;
       //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;

       if (this.ptDS == null)
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try{
          cb = (URCellBlock)theBlock;
          if ( cb.getPointDataBlock() == null ) 
          {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rdb                                  this.ptDS.getRecordDescriptor());
             ptDataBlock = this.ptDS.getRecordDescriptor()
                          .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }
          else
             ptDataBlock = cb.getPointDataBlock();
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();

          this.ptDS.subblock( ptDataBlock, pointDstBounds, pointSrcBounds );
          cb.setPointDataBlock( ptDataBlock );
            
       } catch (ClassCastException e){
            new IllegalArgumentError( SDBError.Recoverable, 
                  "Non-CellBlock passed to CellDataSource.subblock()." );
         }
    }


    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                          ISBounds srcBounds, FieldIDMapper mp)
    { 
       ISBounds pointSrcBounds, pointDstBounds;
       //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;

       if ( this.ptDS == null )
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try{
          cb = (URCellBlock)theBlock;
          if (cb.getPointDataBlock() == null) 
          {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rdb                                 this.ptDS.getRecordDescriptor());
             ptDataBlock = this.ptDS.getRecordDescriptor() 
                                    .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }
          else
             ptDataBlock = cb.getPointDataBlock();
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();

          this.ptDS.subblock( ptDataBlock, pointDstBounds, pointSrcBounds, mp );
          cb.setPointDataBlock( ptDataBlock );

       } catch ( ClassCastException e ){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Non-CellBlock passed to CellDataSource.subblock().");
         }
    } 

    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                          ISBounds srcBounds, RecordSpec rs)
    { 
       ISBounds pointSrcBounds, pointDstBounds;
       //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;

       if (this.ptDS == null)
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try{
          cb = (URCellBlock)theBlock;
          if ( cb.getPointDataBlock() == null ) 
          {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rdb                                  this.ptDS.getRecordDescriptor());
             ptDataBlock = this.ptDS.getRecordDescriptor()
                                    .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }
          else
             ptDataBlock = cb.getPointDataBlock();
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();

          this.ptDS.subblock( ptDataBlock, pointDstBounds, pointSrcBounds, rs );
          cb.setPointDataBlock( ptDataBlock );

       } catch ( ClassCastException e ){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Non-CellBlock passed to CellDataSource.subblock().");
         }
    }


    /** Return a cell block with the specified resolution and the specified
     *  bounds 
     */
    public void urCellBlock( DataBlock ublock, ISBounds b, ISBounds res )
    {
     //System.out.println("start urCellBlock in RegCellDataSource"); //--------
	   //rdb CompositeBlock pblock, pblock2;
       DataBlock pblock, pblock2;
	   ISBounds pbounds, pbounds2;
	   DSApproximator approx = new DSApproximator();
       if (res.equals(this.bounds)) 
       {
          this.subblock(ublock, ublock.getBounds(), b);
       }
       else 
       {
	      pbounds = res.copyIncreasedBy(1);
	      ISIDMapper map = new ISIDMapper( pbounds, this.ptDS.getBounds() );
	      pbounds = b.copyIncreasedBy(1);
	      pbounds2 = map.toComposite( pbounds );
          //System.out.println("compsite bounds in urCellBlock in RegCellDS: "+pbounds2.toString()); //--------
	      //rdb pblock2 = new CompositeBlock(pbounds2, this.recordDesc);
          //rdb pblock2 = this.recordDesc.createDataBlock( pbounds2 );
          pblock2 = this.recordDesc.createDataBlock( pbounds2.copyToOrigin() );
	      this.ptDS.subblock( pblock2, pblock2.getBounds(), pbounds2 );
           //System.out.println(pblock2.toString()); //--------
	      pbounds.moveToOrigin();
	      //rdb pblock = (CompositeBlock)approx.approximate(pblock2, pbounds);
          pblock = (DataBlock) approx.approximate(pblock2, pbounds);
          //System.out.println(pblock.toString()); //--------

	      /*create a uniform cell block object corresponding to the specified
	      resolution and the specified bounds*/
	      ((URCellBlock)ublock).setPointDataBlock(pblock);
       }
    }


    /** Create a URCellBlock of a class appropriate for use with this DataSour
        using the specified bounds and this datasource's RecordDescripto
     */
    DataBlock createBlock(ISBounds bounds)
    {
      URCellBlock cblock = new URCellBlock( bounds,this.recordDesc );
      return cblock;
    }


    /** Create a CellBlock of a class appropriate for use with this DataSoure
        using the specified bounds and a new RecordDescriptor constructed from
        this datasource's RecordDescriptor and the FieldIDMapper.
     */
    DataBlock createBlock( ISBounds bounds, FieldIDMapper fid )
    {
       return new URCellBlock( bounds, this.recordDesc.create(fid) );
    }

    /** Create a CellBlock of a class appropriate for use with this DataSour
        using the specified bounds and a new RecordDescriptor constructe
        this datasource's RecordDescriptor and the RecordSpec.
     */
    DataBlock createBlock(ISBounds bounds, RecordSpec rs){

       return new URCellBlock(bounds,this.recordDesc.create(rs));
    }


    /** Return a String representing the object. */
    public String toString(){
       return this.toString("");
    }
                                                                                
    /** Return a String representing the object indented by the argument string.      */
    public String toString(String indent){
       int i;
       String indent2=indent+Globals.toStringIndent;
       StringBuffer result=new StringBuffer(indent+"{ RegCellDataSource:\n");
       result.append(indent2+"Bounds: ");
       result.append(this.bounds.toString()+"\n");
       result.append(indent2+"Point bounds: ");
       result.append(this.ptDS.getBounds().toString()+"\n");
       if (this.resolution != null) {
          result.append(this.resolution.toString(indent2)+"\n");
       }
       if (this.path != null) {
          result.append(indent2+"Path is:     <PATH (");
          for (i = 0; i < this.path.length-1; i++) {
             result.append(this.path[i]+", ");
          }
          result.append(this.path[this.path.length-1]+")>\n");
       }
       result.append(indent+"}");
       return new String(result);
    }
                                                                                

    /* Test the method of subblock. The command to run the program is:
       java RegCelldataSource fld-file data-file */
    /* static public void main(String[] args) {

   } */

}

