//
//  ThreadedBlockCache.java
//  
//
//  Created by Philip Rhodes on Mon Dec 08 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//

/* 
    ThreadedBlockCache.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    --------------------- Edit history as of 02/16/05 ------------------------
    10/21/05 rdb: deleted unused variables
    02/16/05 rdb: new granite API updates
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

/**
  *	The ThreadedBlockCache uses threaded IO to read blocks of data (hopefully) before they are needed.
  * This behavior is only possible when the user accesses the cache in the order specified by the iterator
  * given to the CacheMaker. 
  */


class ThreadedBlockCache extends BlockCache{
    
    CacheDataBlock[] theBlocks;
    int[] queryBoundsDimensions;
            
                
    // The following group of data members comprise the ISBounds
    // that is used to query the source indexable object.
                            
    int[] sQBLowerInts;
    int[] sQBUpperInts;
        
    ISBounds sourceQueryBounds;
    ISBounds queryBlockDstBounds;
    
    // The following group of data members comprise the ISBounds
    // that is used to query the cacheblock for block access. For
    // datum acess, datumCoords1 is used by itself.
    
    // Used for storing the intra-block datum coords. These arrays
    // are assigned into the bounds defined below.
    int [] datumCoords1;             
    int [] datumCoords2;
    
    // This bounds denotes the subblock to be read from 
    // a cache block into the argument block. 
    ISBounds subblockQueryBounds;
    
    int [] partitioningCoords; // used by load as scratch space
    
    
    protected CacheIOThread ioThread=null;
    protected int lastTag=-1;
    protected int numTags;
    
    
	/** Create a ThreadedBlockCache that caches the Indexable object passed as the
	  *	first argument. The second argument specifies the dimensions of the 
	  *	cache block used, and the third specifies the cache block axisOrdering.
	  */
    ThreadedBlockCache(Indexable source, ISBounds iterationSpace,IndexSpaceID blockDimensions,  AxisOrdering cacheBlockOrdering){
    
    
        super(source, iterationSpace, blockDimensions, 2, cacheBlockOrdering);

        int d=iterationSpace.dim();
        this.partitioningCoords=new int[d];
                
        this.queryBoundsDimensions = new int[d];
        
        this.datumCoords1=new int[d];
        this.datumCoords2=new int[d];

     
        this.numTags = (int) this.spacePartitioning.volume();
                 
        

        this.sQBLowerInts=new int[d];
        this.sQBUpperInts=new int[d];

        this.sourceQueryBounds = new ISBounds(this.sQBLowerInts, this.sQBUpperInts);
        this.subblockQueryBounds = new ISBounds(this.datumCoords1, this.datumCoords1);
        this.queryBlockDstBounds = new ISBounds(this.dimensionality);

        this.theBlocks=new CacheDataBlock[2];
        
        this.theBlocks[0]=new CacheDataBlock(this.cacheBlockBounds,
											source.getRecordDescriptor(),
                                            cacheBlockOrdering);
                                            
        this.theBlocks[1]=new CacheDataBlock(this.cacheBlockBounds,
											source.getRecordDescriptor(),
                                            cacheBlockOrdering);


    }
    
    
    
    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(Datum theDatum, IndexSpaceID id){
 
        int tag = 0; 
        int coordBlockIndex;
        int [] coords=id.refCoords();
        
        for ( int i=0; i<this.dimensionality; i++ ){

            coordBlockIndex = (coords[i] / cacheBlockDimensions[i]);
            
            this.datumCoords1[i] = coords[i] - (coordBlockIndex * cacheBlockDimensions[i]);
            
            tag += coordBlockIndex * this.tagAxisWeights[i];
        }
        
        
        CacheDataBlock theBlock=theBlocks[ tag % 2];
                
        if(theBlock.getTag()==tag){
        
            theBlock.datum(theDatum,this.datumCoords1);
        } else {
            
			if(	this.ioThread != null && 
				this.ioThread.tag==tag &&
				!this.ioThread.isDone ){
				
				try{
					this.ioThread.join();
				} catch(InterruptedException e) {
					
					new SDBError(SDBError.Recoverable, "**** Main thread was interrupted");
					this.lookup(theDatum, id);
					return;
				}
			} else {	
				this.load(tag);
			}
            
            theBlock.datum(theDatum,this.datumCoords1);
        }

        this.handleActivePrefetch(tag);

        
    }

 
    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(BasicBlock queryBlock, ISBounds dstBounds, ISBounds userQueryBounds){
 

        int i, d=this.dimensionality;
        int lowerTag=0, upperTag=0;
        int lowerCoordBlockIndex, upperCoordBlockIndex;
        int [] lowerCoords=userQueryBounds.getLowerArray();
        int [] upperCoords=userQueryBounds.getUpperArray();
        int splitAxis = -1;
        int splitAxisUpperBlockIndex=-1;
                
        userQueryBounds.getDimensions(this.queryBoundsDimensions);
        
        for(i=0;i<d;i++){

            lowerCoordBlockIndex = (lowerCoords[i] / this.cacheBlockDimensions[i]);
            upperCoordBlockIndex = (upperCoords[i] / this.cacheBlockDimensions[i]);


            if (upperCoordBlockIndex - lowerCoordBlockIndex == 1) {
            
                splitAxis = i;
                splitAxisUpperBlockIndex = upperCoordBlockIndex; 
            }

            //sharedDimension = (lowerCoordBlockIndex == upperCoordBlockIndex) ? i : sharedDimension;

            this.datumCoords1[i] = lowerCoords[i] - (lowerCoordBlockIndex * this.cacheBlockDimensions[i]);
            this.datumCoords2[i] = this.datumCoords1[i] + this.queryBoundsDimensions[i] - 1;
            
            lowerTag += lowerCoordBlockIndex * this.tagAxisWeights[i];
            upperTag += upperCoordBlockIndex * this.tagAxisWeights[i];
        }

        
        //System.out.println("lowerTag="+lowerTag + " upperTag="+upperTag);
        
        CacheDataBlock theBlock=theBlocks[ lowerTag % 2];
        
        if(lowerTag == upperTag){
        
            subblockQueryBounds.assign(this.datumCoords1, this.datumCoords2);

            if(theBlock.getTag() == lowerTag){
            
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            } else {
                    
                if(	this.ioThread != null && 
                    this.ioThread.tag==lowerTag &&
                    !this.ioThread.isDone ){
                    
                    try{
                        this.ioThread.join();
                    } catch(InterruptedException e) {
                        
                        new SDBError(SDBError.Recoverable, "**** Main thread was interrupted");
                        this.lookup(queryBlock, dstBounds, userQueryBounds);
                        return;
                    }
                } else {	
                    this.load(lowerTag);
                }
                
                theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            }
            
        } else {
            
            if (upperTag - lowerTag == 1){
                
                //System.out.println("lowerTag="+lowerTag + " upperTag="+upperTag);
                //System.out.println(" queryBounds="+ userQueryBounds);


                
                if(splitAxis != -1){
                
                    System.out.println("Neighboring blocks:  sharedDimension="+ splitAxis);
            
                    ThreadedSliceCache.SliceCacheISBounds.lowerTSCUtil(userQueryBounds, dstBounds, queryBlockDstBounds, subblockQueryBounds, splitAxis, splitAxisUpperBlockIndex * this.cacheBlockDimensions[splitAxis] - 1, this.cacheBlockDimensions[splitAxis]);

                    theBlock=this.theBlocks[lowerTag % 2]; 
                    
                    
    System.out.println("qbdb="+queryBlockDstBounds + " sbqb="+subblockQueryBounds);
                    
                    theBlock.psubblock(queryBlock, queryBlockDstBounds, subblockQueryBounds);
                    
                    
                    ThreadedSliceCache.SliceCacheISBounds.upperTSCUtil(userQueryBounds, dstBounds, queryBlockDstBounds, subblockQueryBounds, splitAxis, splitAxisUpperBlockIndex * this.cacheBlockDimensions[splitAxis] - 1, this.cacheBlockDimensions[splitAxis]);
                
                    theBlock=this.theBlocks[upperTag % 2];
                    
                    this.waitOnIOThread();
                    
                    theBlock.psubblock(queryBlock, queryBlockDstBounds, subblockQueryBounds);
                    
                } else {
                    
                    System.out.println("Consecutive tags, but not neighbors.");
                    
                    this.waitOnIOThread();
                    this.theSource.subblock(queryBlock, dstBounds, userQueryBounds);
                }
            
            } else {
            
                this.waitOnIOThread();
                this.theSource.subblock(queryBlock, dstBounds, userQueryBounds);
            }
        }
 
        
        this.handleActivePrefetch(lowerTag);
        
    }
    
    
	/** If any IO thread is running, wait for it to terminate.
	 *  
	 */
	void waitOnIOThread(){
	
		if(this.ioThread != null && this.ioThread.isAlive()){
		
			try{
			
				this.ioThread.join();
			} catch(InterruptedException e){
				
				new InternalFaultError(SDBError.Recoverable,"join() was interrupted in waitOnIOThread()");
			}
		}
	}
    
    private void handleActivePrefetch(int tag){
    
		
		if(tag != this.lastTag && this.numTags > 1){

			this.lastTag=tag;
                
         this.ioThread = new CacheIOThread(this, (tag + 1) % this.numTags);
			this.ioThread.start();
		}        
    
    }
    


   protected synchronized void load(int tag){

      CacheDataBlock theBlock=this.theBlocks[tag % 2];

      tagToBounds(tag,  this.sourceQueryBounds);

      // The iterationSpaceBounds should already have been clipped against
      // the ds bounds, so clipping against it should be safe.
      this.sourceQueryBounds.intersectWith( iterationSpaceBounds );  

      theSource.subBasicBlock(theBlock, this.cacheBlockBounds, sourceQueryBounds);
      theBlock.setTag(tag);
      }

	
	protected class CacheIOThread extends Thread{
	
      protected ThreadedBlockCache c;
      protected int tag;
		protected boolean isDone=false;
		
		CacheIOThread(ThreadedBlockCache c, int tag){
		
         super("ThreadedBlockCache IO Thread");
            
			this.c=c;
         this.tag=tag;
		}

        @Override
		public void run(){
		
			c.load(this.tag);
			
			this.isDone = true;
		}
	
	}

   
}
