/*
 * Created on May 30, 2005
 *
 * Issues 5/30
 *    Do we really want to return DataCollections, rather than DataSources?
 *    Do we want to activate a DS before returning it?
 */
package edu.unh.sdb.datasource;

import java.io.File;
import java.util.*;
import edu.unh.sdb.common.*;

/**
 * This class encapsulates time series data without imposing a particular
 * data access interface on the series itself. It is primarily intended to
 * support the TimeSeriesDataSource, but we thought there might be some
 * advantage to  provide a class to manage the members of the series without
 * imposing the DataSource interface onto the manager. For now, we'll try to
 * define the series as a series of DataCollections.
 *  
 * @author rdb
 *
 * ---------------------------- Edit history ----------------------------
 * 11/30/05 rdb: Fixed shared FDLDescriptor problem. Series DS no longer share FDLD
 * 10/21/05 rdb: added getName()
 * 06/01/05 rdb: Working version done
 * 05/30/05 rdb: created
 */
public class TimeSeries
{
   // --------------------------- TimeStep ----------------------------------
   private class TimeStep
   {
      public DataCollection data;   // DataCollection at the step
      public float          time; // time of the step
      public TimeStep( DataCollection dc, float t )
      {
         this.data = dc; this.time = t;
      }
   }

   private ArrayList<TimeStep> steps = null;    // array of TimeStep objects
   private FDLDescriptor fdld  = null;    // fdldescr associated with TimeSeries
   private String        name = null;
   
   // Filename pattern information 
   private String        prefix = null;
   private String        suffix = null;
   private String        pattern = null;

   /**
    * Create a time series of DataSources from an xfdlFile.
    */
   public TimeSeries( String name, String xfdlFile )
   {
      this.name = name;
      this.steps = new ArrayList<TimeStep>();
      this.fdld  = FDLDescriptor.create( xfdlFile );
      if ( this.fdld == null )
      {
         new SDBError( SDBError.Recoverable,
                       "TimeSeries error: Failed on file spec: " + xfdlFile );
      }
      makeSteps(); 
   }
   
   /**
    * Create a time series from an array of DataCollections
    */
   public TimeSeries( String name, DataCollection[] elements )
   {
      this( name, elements,  null );
   }
        
   /**
    * Create a time series from an array of DataCollections and a time mapping
    */
   public TimeSeries( String name, DataCollection[] elements, float[] times )
   {
      this.name = name;
      times = checkTimes( times, elements.length );
      this.steps = new ArrayList<TimeStep>( elements.length );
      for ( int i = 0; i < elements.length; i++ )
         this.steps.add( new TimeStep( elements[ i ], times[ i ] ));
   }
   
   //++++++++++++++++++++++++ utility methods ++++++++++++++++++++++++
   
   //-------------------- checkTimes ----------------------------------
   private float[] checkTimes( float[] times, int len )
   {
      if ( !goodTimes( times, len ))
      {
         times = new float[ len ];
         for ( int i = 0; i < len; i++ )
            times[ i ] =  i;
      }
      return times;
   }
   
   //--------------------- goodTimes ----------------------------------
   private boolean goodTimes( float[] times, int len )
   {
      if ( times == null )
         return false;
      if ( len != times.length )
      {
         new SDBError( SDBError.Warning, 
                       "TimeSeries error: length mismatch of series and times arrays");
         return false;
      }
      
      // check order
      for ( int i = 0; i < len-1; i++ )
      {
         if ( times[ i + 1 ] <= times[ i ] )
         {
            new SDBError( SDBError.Warning, 
                       "TimeSeries error: times array not ordered." );
            return false;
         }
      }
      return true;
   }
   
   //-------------- accessors -----------------------------------------
   /**
    * return time series name
    */
   public String getName()
   {
      return this.name;
   }
   
   /**
    * return number of steps in time series
    */
   public int getNumSteps()
   {
      return this.steps.size();
   }
   
   /**
    * Return an element in the series based on ordinal position
    */
   public DataCollection getStep( int i )
   {
      try 
      {
         TimeStep tsdata = this.steps.get( i );
         return tsdata.data;
      }
      catch ( ArrayIndexOutOfBoundsException abe )
      {
         return null;
      }
   }
   
   /**
    * Return the real time for an element in the series based on ordinal position
    */
   public float getTime( int i )
   {
      try 
      {
         TimeStep tsdata = this.steps.get( i );
         return tsdata.time;
      }
      catch ( ArrayIndexOutOfBoundsException abe )
      {
         new SDBError( SDBError.Recoverable, 
                      "TimeSeries error: invalid time step index: " + i );
         return 0.0f;
      }
   }
   
   /**
    * Close time series; deactivate all steps
    */
   public void deactivate()
   {
      for ( int i = 0; i < this.steps.size(); i++ )
      {
         TimeStep step = this.steps.get( i );
         if ( step.data instanceof DataSource )
            ((DataSource) step.data).deactivate();
      }
   }
   
   /*------------------ parseTimeField( String ) --------------------------
    * 
    */
   private void parseNameTemplate( String fileNamePattern )
   {       
      // look for time step coding key in file name
      // %6s   means 6 character time field
      // %6.2s means 6 integer chars, and 2 or more fractional fields
      // %6.s  means 6 characters of integral time and optional
      //      . followed by arbitrary size fractional part
      
      String widthField = null;       // width field, between '%' and 's'
      int minFractWidth = -1;         // min # fractional digits, -1 => none allowed
      int intWidth      = 0;          // number digits in integer part of widht
      int sPos          = -1;         // position of 's' wrt '%'
      int dotPos        = -1;         // position of '.' in widhtField if there is one
      
      int widthPos = fileNamePattern.indexOf( "%" );  // start of width specification
      
      if ( widthPos < 0 )
      {
         new SDBError( SDBError.Recoverable, "FDLDescriptor.parseTimeField: " 
                       + "Not a valid time series specification: " + fileNamePattern );
         return;
      }
      
      this.prefix = "";
      if ( widthPos > 0 )
         this.prefix = fileNamePattern.substring( 0, widthPos );
      
      sPos = fileNamePattern.substring( widthPos ).indexOf( "s" );
      if ( sPos < 0 )
      {
         new SDBError( SDBError.Recoverable, "FDLDescriptor.getTimeSeries: "
                          + "Invalid time width format. no 's': " + fileNamePattern );
         return;
      }
      
      this.suffix = "";
      if ( widthPos + sPos < fileNamePattern.length() - 1 )
         this.suffix = fileNamePattern.substring( widthPos + sPos + 1 );
      
      widthField = fileNamePattern.substring( widthPos + 1, widthPos + sPos );
      
      dotPos = widthField.indexOf( '.' );
      if ( dotPos == widthField.length() - 1 ) // '%n.s' form
      {
         widthField = widthField.substring( 0, widthField.length() - 2 );
         minFractWidth = 0;
      }
      else if ( dotPos >= 0 )                 // '%n.ms' form
      {
         String fractionField = widthField.substring( dotPos + 1 );
         widthField = widthField.substring( 0, dotPos );
         if ( fractionField.length() > 0 )
            try 
            {
               Integer fract = new Integer( fractionField );
               minFractWidth = fract.intValue();
            }
         catch ( NumberFormatException nfe )
         {
            new SDBError( SDBError.Recoverable, "FDLDescripter.getTimeSeries: " 
                                   + "Invalid time field spec: " + fractionField );                  
         }
         // else have '%ns' form
      }
      
      // now parse the integer part
      try
      {
         Integer w = new Integer( widthField );
         intWidth = w.intValue();
      }
      catch ( NumberFormatException nfe )
      {
         new SDBError( SDBError.Recoverable, "FDLDescripter.getTimeSeries: " 
                       + "Invalid time field spec: " + widthField );                  
      } 
      // create the pattern for the file search 
      // integer part must be exactly 'intWidth' digits
      this.pattern = "\\d{" + intWidth +"}"; 
     
      if ( minFractWidth == 0 )      // 0 or more fractional digits
         this.pattern += "\\.\\d*";
    
      else if ( minFractWidth > 0 )  // minFractWidth or more digits
         this.pattern += "\\.\\d{" + minFractWidth +",}";   
      // else no fractions allowed 
      
      this.pattern = this.prefix + this.pattern + this.suffix;
   }
   
   /**
    * Create DataSources for all steps of a time series object
    * 
    * The xfdl file name (actually getDataFileURL()) must encode both
    * the directory in which the time series exists as well as the pattern
    * for the names of the files in the series. Everything upto and including
    * the last / (or \)  is the directory; everything after is the pattern.
    * If there is no / (or \), the directory is .
    */
   private void makeSteps( )
   {
      String              baseName         = this.fdld.getDataFileURL();
      String              dir              = null;
      String              fileNamePattern  = null;   // the filename portion of url
      
      // extract directory specification
      int slashPos = baseName.lastIndexOf( File.separatorChar );
      if ( slashPos < 0 )
         dir = ".";
      else if ( slashPos == 0 )
      {
         dir = File.separator;
         fileNamePattern = baseName.substring( slashPos + 1 );
      }
      else
      {
         dir = baseName.substring( 0, slashPos );
         fileNamePattern = baseName.substring( slashPos + 1 );
      }
      parseNameTemplate( fileNamePattern );
      
      File tsDir = new File( dir );
      if ( tsDir == null || !tsDir.exists() )
      {
         String rootedDir = this.fdld.fdlFileRoot + File.separatorChar + dir;
         tsDir = new File( rootedDir );
         if ( tsDir == null )
         {
            new SDBError( SDBError.Recoverable, "FDLDescriptore.getTimeSeries: "
                          + "Can't find time series directory: " + dir );
         }
      }
      
      String[] tsFiles = tsDir.list();
      
      if ( tsFiles != null )       
      {
         DataSource    ds     = null;
         
         for ( int i = 0; i < tsFiles.length; i++ )
         {
            // created an FDLD to be used to define the individual datasources from the
            // FDLD for the series -- cannot reuse it
            FDLDescriptor dsFDLD = (FDLDescriptor) this.fdld.clone();
            dsFDLD.setRootDirectory( tsDir.getAbsolutePath() );
                        
            if ( tsFiles[ i ].matches( this.pattern ))
            {
               dsFDLD.setDataFileURL( tsFiles[i] );               
               ds = DataSource.create( tsFiles[i], dsFDLD );  
               this.steps.add( new TimeStep( ds, getTime( tsFiles[i] )));
            }
         }
      }
   }
      
   //-------------------- getTime -------------------------
   private float getTime( String fileName )
   {
      int    suffixStart = fileName.length() - this.suffix.length();
      String timeField = fileName.substring( this.prefix.length(), suffixStart );
      try
      {
         return (new Float( timeField )).floatValue();
      }
      catch ( NumberFormatException nfe )
      {
         new SDBError( SDBError.Fatal, "TimeSeries.getTIme: Invalid time field: " 
                                + timeField + ".  System error????" );                
      }
      return 0.0f;
   }
   
   //--------------------------------- main -------------------------------------
   public static void main( String[] args )
   {
      // args[0] must be an xfdl file defining time series
      TimeSeries ts = new TimeSeries( "testing", args[0] );
      for ( int i = 0; i < ts.getNumSteps(); i++ )
      {
         TimeStep step = ts.steps.get( i );
         System.out.println( step.time + ": " + ((DataSource)step.data).name );
      }
   }
}
