/* 
    UniBlockCache.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    ------------------------- Edit history as of 02/16/05 ------------------
    04/25/05 rdb: fixed problem in lookup in computing sourceQueryBounds when
                  cache block is not an even multiple of the source size.
    02/16/05 rdb: new granite API updates
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.util.*;
//import java.io.*;

/** 
 *  The UniBlockCache provides cache service for an underlying datasource. CacheDataSources
 *   are typically created by the CacheMaker class, which tunes the size of the cache blocks to
 *   a particular iteration and underlying datasource. As its name implies, the UniBlockCacheDataSource
 *   uses a cache with a single block. When a query is received, it will be satisfied from this cache
 *   block if possible. Otherwise, one of two things will be done to satisfy the query. If the query can
 *   be satisfied entirely by reading a different block into the cache, then the new block is read, and 
 *   the query satisfied. This is always the course of action for datum queries. For subblock queries, if 
 *   the query region is entirely contained in some cache block, then that block is read and the query
 *   satisfied from its contents. If the query region spans more than one cache block, the query is 
 *   satisfied directly from the datasource, and the cache block remains unchanged.
 */
class UniBlockCache extends BlockCache{

    long [] tagAxisWeights;          // Controls mapping of n-D block space to 1D cache tags
 
    
    CacheDataBlock theBlock;
    int[] queryBoundsDimensions;
            
                
    // The following group of data members comprise the ISBounds
    // that is used to query the source indexable object.
                            
    int[] sQBLowerInts;
    int[] sQBUpperInts;
        
    ISBounds sourceQueryBounds;
    
    
    
    // The following group of data members comprise the ISBounds
    // that is used to query the cacheblock for block access. For
    // datum acess, datumCoords1 is used by itself.
    
    // Used for storing the intra-block datum coords. These arrays
    // are assigned into the bounds defined below.
    int [] datumCoords1;             
    int [] datumCoords2;
    
    // This bounds denotes the subblock to be read from 
    // a cache block into the argument block. 
    ISBounds subblockQueryBounds;
    
/*    
    int [] partitioningDimensions;
    RegularISPartitioning partitioning;
    ISBoundsIterator partitioningIterator;
*/    
//  ------------------- Edit history --------------------------------------
//  10/21/05 rdb: removed unused variables; commented out rootId uses
// -----------------------------------------------------------------------
       
	/** Create a UniBlockCache that caches the Indexable object passed as the
	  *	first argument. The second argument specifies the dimensions of the 
	  *	cache block used, and the third specifies the cache block axisOrdering.
	  */
    UniBlockCache( Indexable source, ISBounds iterationSpace, 
                   IndexSpaceID blockDimensions, AxisOrdering cacheBlockOrdering){
    
        super( source, iterationSpace, blockDimensions, 1 );

        ISBounds spaceBounds = source.copyBounds();
        int d = spaceBounds.dim();
        AxisOrdering partitioningOrdering  = new AxisOrdering(d);
                
        //this.cacheBlockDimensions = this.getCacheBlockDimensions(new int[d]);
        this.queryBoundsDimensions = new int[d];
        
        this.datumCoords1 = new int[d];
        this.datumCoords2 = new int[d];
/*
        this.partitioningDimensions = new int[d];
        
        for(i=0; i<d; i++){
        
            this.partitioningDimensions[i] = spaceBounds.getDimension(i) / blockDimensions.getCoord(i);
        }
        
        this.partitioning=new RegularISPartitioning(new ISBounds(this.partitioningDimensions), spaceBounds);
        this.partitioningIterator=new ISBoundsIterator(spaceBounds, spaceBounds);
*/        
        
        this.tagAxisWeights = partitioningOrdering.makeAxisWeights(partitioningDimensions);

        this.sQBLowerInts=new int[d];
        this.sQBUpperInts=new int[d];

        sourceQueryBounds   = new ISBounds( this.sQBLowerInts, this.sQBUpperInts );
        subblockQueryBounds = new ISBounds( this.datumCoords1, this.datumCoords1 );
        
        this.theBlock=new CacheDataBlock(   this.cacheBlockBounds,
                                            source.getRecordDescriptor(),
                                            cacheBlockOrdering);

    }
    
    
    
    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(Datum theDatum, IndexSpaceID id){
 
        int tag=0;

        int d=this.dimensionality;
        int coordBlockIndex;
        int [] coords=id.refCoords();
        
        for ( int i=0; i<d; i++){

            coordBlockIndex = (coords[i]- lowerIterationSpaceCoords[i]) / cacheBlockDimensions[i];
            
            this.sQBLowerInts[i] = lowerIterationSpaceCoords[i] + coordBlockIndex * cacheBlockDimensions[i];
            
            this.datumCoords1[i] = coords[i] - this.sQBLowerInts[i];
            
            tag += coordBlockIndex * this.tagAxisWeights[i];
        }
        
        
        
        if ( theBlock.getTag()==tag ){
            //System.out.println("*Hit*");
            //AppMonitor.recordCacheHit();
            //this.theBlock.datum((ShortDatum) theDatum,this.datumCoords1);
            this.theBlock.datum( theDatum,this.datumCoords1);
        } else {
            //System.out.println("*Miss*");
            //AppMonitor.recordCacheMiss();
            
            for ( int i=0; i<d; i++ ){
                
                this.sQBUpperInts[i] = this.sQBLowerInts[i] + cacheBlockDimensions[i] - 1;
            }
            
            this.sourceQueryBounds.assign( sQBLowerInts, sQBUpperInts );
            
            
            // Clip the cache block to the datasource bounds. It is 
            // possible for the iteration space bounds to exceed the ds bounds
            // with oblique iteration, so we'll use the ds bounds instead.

            this.sourceQueryBounds.intersectWith( dsSpaceBounds );  
            
           //System.out.println("UBC: miss, loading "+ this.sourceQueryBounds);

            theSource.subBasicBlock( this.theBlock, this.cacheBlockBounds, this.sourceQueryBounds);
            
            this.theBlock.setTag( tag );
            this.theBlock.datum( theDatum,this.datumCoords1 );
        }
    }


    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(BasicBlock queryBlock, ISBounds userQueryBounds){
    
        this.lookup(queryBlock, queryBlock.getBounds(), userQueryBounds);
    }




 
    /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
      * the cache with a new block.
	  */
    void lookup(BasicBlock queryBlock, ISBounds dstBounds, ISBounds userQueryBounds){
 

        int d=this.dimensionality;
        int lowerTag=0, upperTag=0;
        int lowerCoordBlockIndex, upperCoordBlockIndex;
        int [] lowerCoords=userQueryBounds.getLowerArray();
        int [] upperCoords=userQueryBounds.getUpperArray();
                
        userQueryBounds.getDimensions(this.queryBoundsDimensions);
        
        for ( int i=0; i<d; i++ ){

            lowerCoordBlockIndex = (lowerCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];
            upperCoordBlockIndex = (upperCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];

            
            this.sQBLowerInts[i] = lowerCoordBlockIndex * this.cacheBlockDimensions[i] + lowerIterationSpaceCoords[i];
            //this.sQBUpperInts[i] = upperCoordBlockIndex * this.cacheBlockDimensions[i];
            
            this.datumCoords1[i] = lowerCoords[i] - this.sQBLowerInts[i];
            this.datumCoords2[i] = this.datumCoords1[i] + this.queryBoundsDimensions[i] - 1;
            
            lowerTag += lowerCoordBlockIndex * this.tagAxisWeights[i];
            upperTag += upperCoordBlockIndex * this.tagAxisWeights[i];
        }
         
        if ( lowerTag == upperTag ){
        
            subblockQueryBounds.assign( this.datumCoords1, this.datumCoords2 );

            if ( theBlock.getTag() == lowerTag ){
                //System.out.println("*Hit*");
                //AppMonitor.recordCacheHit();
                this.theBlock.subblock( queryBlock, dstBounds, subblockQueryBounds );
            } else {
                //System.out.println("*Miss*");
                //AppMonitor.recordCacheMiss();              
                
                for ( int i=0; i<d; i++){
                    
                    this.sQBUpperInts[i] = this.sQBLowerInts[i] + this.cacheBlockDimensions[i] - 1;
                    
                    //this.sQBUpperInts[i] = (this.sQBUpperInts[i] > this.upperIterationSpaceCoords[i]) ? this.upperIterationSpaceCoords[i] : this.sQBUpperInts[i];
                }
                               
                this.sourceQueryBounds.assign( sQBLowerInts, sQBUpperInts );
                
                
                // The iterationSpaceBounds should already have been clipped against
                // the ds bounds, so clipping against it should be safe.
                this.sourceQueryBounds.intersectWith( iterationSpaceBounds );  
    
                theSource.subBasicBlock(this.theBlock, this.cacheBlockBounds,sourceQueryBounds);
                this.theBlock.setTag(lowerTag);
                
                this.theBlock.subblock(queryBlock, dstBounds, subblockQueryBounds);
            }
            
        } else {
        
            System.out.println("You're in da weeds, Lorna! ;-)");
            
            this.theSource.subBasicBlock(queryBlock, dstBounds, userQueryBounds);       
        }
    }   


		//Added by Baoqiang Yan on 11/30/2007 to avoid the copying cost
		//while using cache iterator to load cache blocks
		CacheDataBlock getCacheDataBlockReference(){
			return this.theBlock;
		}
		
		//Added by Baoqiang Yan on 11/30/2007 to avoid the copying cost
		//while using cache iterator to load cache blocks
    void loadCacheBlock(ISBounds userQueryBounds){
 

        int d=this.dimensionality;
        int lowerTag=0, upperTag=0;
        int lowerCoordBlockIndex, upperCoordBlockIndex;
        int [] lowerCoords=userQueryBounds.getLowerArray();
        int [] upperCoords=userQueryBounds.getUpperArray();
                
        userQueryBounds.getDimensions(this.queryBoundsDimensions);
        
        for ( int i=0; i<d; i++ ){

            lowerCoordBlockIndex = (lowerCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];
            upperCoordBlockIndex = (upperCoords[i] - lowerIterationSpaceCoords[i]) / this.cacheBlockDimensions[i];

            
            this.sQBLowerInts[i] = lowerCoordBlockIndex * this.cacheBlockDimensions[i] + lowerIterationSpaceCoords[i];
            //this.sQBUpperInts[i] = upperCoordBlockIndex * this.cacheBlockDimensions[i];
            
            this.datumCoords1[i] = lowerCoords[i] - this.sQBLowerInts[i];
            this.datumCoords2[i] = this.datumCoords1[i] + this.queryBoundsDimensions[i] - 1;
            
            lowerTag += lowerCoordBlockIndex * this.tagAxisWeights[i];
            upperTag += upperCoordBlockIndex * this.tagAxisWeights[i];
        }
         
        if ( lowerTag == upperTag ){
        
            subblockQueryBounds.assign( this.datumCoords1, this.datumCoords2 );

            if ( theBlock.getTag() == lowerTag ){
                //System.out.println("*Hit*");
                //AppMonitor.recordCacheHit();
                //this.theBlock.subblock( queryBlock, dstBounds, subblockQueryBounds );
            } else {
                System.out.println("*Miss*");
                //AppMonitor.recordCacheMiss();              

                for ( int i=0; i<d; i++){
                    
                    this.sQBUpperInts[i] = this.sQBLowerInts[i] + this.cacheBlockDimensions[i] - 1;
                    
                    //this.sQBUpperInts[i] = (this.sQBUpperInts[i] > this.upperIterationSpaceCoords[i]) ? this.upperIterationSpaceCoords[i] : this.sQBUpperInts[i];
                }
                               
                this.sourceQueryBounds.assign( sQBLowerInts, sQBUpperInts );
                
                
                // The iterationSpaceBounds should already have been clipped against
                // the ds bounds, so clipping against it should be safe.
                this.sourceQueryBounds.intersectWith( iterationSpaceBounds );  
    
                theSource.subBasicBlock(this.theBlock, this.cacheBlockBounds,sourceQueryBounds);
                this.theBlock.setTag(lowerTag);

            }
            
        } else {
        
            System.out.println("You're in da weeds, Lorna! ;-)");
            
            //this.theSource.subBasicBlock(queryBlock, dstBounds, userQueryBounds);       
        }
    }   
}
