/*
  programer: Li Ye
  
  --------------------- Edit history as of 01/25/05 -------------------------
  10/21/05 rdb: removed unused variables
  07/15/05 rdb: added a getValues() since I made DataCollection.getValues abstract.
  02/20/05 rdb: updated RecordSpec to FieldIDMapper mapping
  01/26/05 rdb: revised subblock methods to make consistent with DataCollection and
                   to make sure that RecordSpec and FieldIDMapper args of null are ok.
                   Code not tested at all and the code is complicated enough that I'm
                   not very confident of the changes.
  01/25/05 rdb: Made this a child of DataSource
                used RecordDescriptor.createDataBlock() instead of explicit alloc
                Deleted DataBlock subblock( ISBounds ). Parent method is now efficient
                   enough; it now allocates a block and calls the reference version.

*/

package edu.unh.sdb.datasource;

//import java.io.*;
import edu.unh.sdb.common.*;

/**
VRIndexable provides different resolution view of the datasource.
It can be built on top of any datasource to view the datasource with
different resolution level. The resolution level is a integer number
which can be incremented and decrease by VRDS' function call.
Negative numbers mean the current resolution coarser than the datasource's
original resolution. It also means the current VRDS has less data points
than the original datasource.
Positive numbers' meaning is opposite.
*/

public class VRIndexable extends DataCollection
{
    protected DataCollection source_dc;
    protected ISIDMapper mapper;
    protected int res_level,base_resolution;
    
    protected ISBounds bounds;
    protected int dimensionality;
    protected RecordDescriptor recordDesc;
    
    /** Construct a VRDataSource from any source datasource.*/
    public VRIndexable( DataCollection dc )
    {
       super( dc.bounds, dc.recordDesc );
       this.source_dc = dc;
       this.mapper = new ISIDMapper(this.bounds, this.bounds);
       this.res_level = 0;
    }
	
    /**Construct a VRdataSource and set it to a certain resolution level.*/
    public VRIndexable( DataCollection dc, int base_res )
    {
       this( dc );
       if ( base_res > 0 ) 
           while ( this.res_level< base_res )
              this.incr();
       else if ( base_res < 0 ) 
          while ( this.res_level> base_res )
              this.decr();
    }
    
    /**Construct a VRdataSource and set its resolution corresponding
       to the given ISBounds argument.*/
    public VRIndexable( DataCollection dc, ISBounds res )
    {
       this( dc );
       this.mapper = new ISIDMapper( this.bounds, res );
    }

    /** Get source */
    public Indexable getSource()
    {
       return this.source_dc;
    }
   
    /** Get the mapper */
    public ISIDMapper getMapper()
    {
        return this.mapper;
    }
 
    /** Return by reference the Datum corresponding to the given 
        IndexSpaceID.
     */
    public void datum( Datum dtm, IndexSpaceID did)
    {
        source_dc.datum( dtm, this.mapper.toComponent(did) );
    }

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
    {
        source_dc.datum( dtm, this.mapper.toComponent(did), mp );
    }
 
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec)
    {
        source_dc.datum( dtm, this.mapper.toComponent(did), recSpec );
    }
    /** Return the Datum corresponding to the given IndexSpaceID and
	current resolution level, using the ISIDMapper. */
    public Datum datum(IndexSpaceID did) {
        IndexSpaceID dsid = this.mapper.toComponent(did);
        return source_dc.datum(dsid);
    }
   

    /** Return the attribute projected Datum corresponding to the given 
        IndexSpaceID and current resolution level, using the ISIDMapper. */
    public Datum datum(IndexSpaceID did,RecordSpec spec) {
        IndexSpaceID dsid = this.mapper.toComponent(did);
        return source_dc.datum(dsid,spec);
    }
	
    /** Increase Resolution Level by 1 */
    public boolean incr() {
        int[] upper=this.bounds.getUpperArray();
        int[] lower=this.bounds.getLowerArray();
        
        int d=this.dimensionality;
        
        this.res_level ++;
        
        for(int i=0; i<d; i++){
            upper[i] = 2*upper[i]+1;	//(upper[i]+1)*2-1;
        }
        
        this.bounds.assign(lower,upper);
        this.mapper = new ISIDMapper(source_dc.getBounds(),this.bounds);
        
        return true;
   } 
    
    
    /** Decrease Resolution Level by 1 */
    public boolean decr() {
        int i,d=this.dimensionality;
        int[] lower=this.bounds.getLowerArray();
        int[] upper=this.bounds.getUpperArray();
        
        // Couldn't we just check the upper coords?
        if (this.bounds.volume() == 1) {
            return false;
        }
        
        for(i=0; i<d; i++){

            upper[i] = (upper[i]-1)/2; //(upper[i]+1)/2-1;
        }
        
        this.res_level--;        
        this.bounds.assign(lower,upper);
        this.mapper = new ISIDMapper(source_dc.getBounds(),this.bounds);
        return true;
    }
    
    /** Return a Values array of all attributes in whatever type and order
     *  they are stored internally (either attribute order or point order 
     *  for Granite-defined objects). If the VRIndexable is a DataSource that is
     *  very large, this implementation could be a problem since it just makes a
     *  single DataBlock and does a getValues on that.
     */
    public Values getValues()
    {
       return this.subblock( getBounds() ).getValues();
    } 

    //------------------ subblock methods ------------------------------------
    
    /* Returns a DataBlock that is an attribute projection of a spatial 
        subset of the DataSource.Override parent method for inefficiency.
        rdb: this version should be handled by DataCollection version.
        */
    /*********************************************
     rdb:
     All this code should be unnecesssary given general implementation
     of DataCollection.subblock( ISBounds, RecordSpec ).
     However, the code here is not the same as subblock(DB, ISB, ISB, RS )
     so rather than just deleting it, I'm commenting it out. 
     Once tested, this could be deleted.
     **************
    public  DataBlock subblock( ISBounds db, RecordSpec rs )
    {
        DataBlock ret;
        ISIDMapper blk_mapper;
        ISIterator iter1;
        ISBounds srcBounds;
        ISBounds dstBounds;

        // check if the dimensionalities match
        if ( this.dim() != db.dim() )
        {
            new DimensionalityError(SDBError.Recoverable,
                  "ID Argument has wrong dimensionality.");
            return null;
        }
        
        // check if the db is inside the datasource bounds 
        
        if ( !this.bounds.contains( db ))
        {
           new BoundsError( SDBError.Recoverable,
                 "Given Bounds is not inside the DataSource.");
           return null;
        }

        dstBounds = ISBounds.createAtOrigin( db );
        ret = this.createDataBlock( dstBounds );

        if ( this.res_level == 0 ) 
        {
           ret = this.source_dc.createDataBlock( 
           //ret = new CompositeBlock( db, this.source_dc.getRecordDescriptor());
           this.source_dc.subblock( ret, ret.getBounds(), db, rs );
        }
        else
        {
           srcBounds = this.mapper.toComponent( dstBounds );
           
           iter1 = new ISIterator( dstBounds );
           blk_mapper = new ISIDMapper( srcBounds, dstBounds );
           
           // if the resolution level is greater than 0, VRDS will
           // retrieve the data as a block because the contiguous space
           // in VRDS is contiguous in the original datasource
           if ( this.res_level > 0 ) 
           {
              DataBlock temp_blk = this.recordDesc.createDataBlock( 
                                        ISBounds.createAtOrigin( srcBounds )); 
              srcBounds.moveToOrigin();
              while (iter1.valid())
              {
                 ret.setDatum( iter1, temp_blk.datum( blk_mapper.toComponent( iter1 )));
                 iter1.next();  
              }
           }
           
           // if the resolution level is less than 0, VRDS will
           // retrieve the data as single data points because the 
           // contiguous space in VRDS is not contiguous in the 
           // original datasource
           else if ( this.res_level < 0 ) 
           {
              while (iter1.valid()){
                 ret.setDatum( iter1, this.datum( blk_mapper.toComponent( iter1 ), rs));
                 iter1.next();
              }
           }
        }
        return ret;
    }
    /**************************************************************/
	
    /** Reads a subblock defined by srcBounds into the region of the Datablock 
        argument defined by dstBounds. */
    public void subblock(DataBlock blk, ISBounds dstBounds, ISBounds srcBounds)
    {
        this.subblock( blk, dstBounds, srcBounds, (RecordSpec)null );
    }


    /** Reads a subblock defined by srcISB into the region of the Datablock 
     * argument defined by dstISB. The RecordSpec argument specifies a subset 
     * of the DataSource attributes to be written into the DataBlock datums.
     */
    public void subblock( DataBlock blk, 
                          ISBounds dstISB, ISBounds srcISB, 
                          RecordSpec spec) 
    {
       subblock( blk, dstISB, srcISB, spec.getFieldIDMapper());
    }


    /** Reads a subblock defined by srcISB into the region of the Datablock 
        argument defined by dstISB. The FieldIDMapper argument determines the 
        mapping between datums in the datasource and datums in the DataBlock.
     */      
    public void subblock( DataBlock blk, 
                          ISBounds dstISB, ISBounds srcISB, 
                          FieldIDMapper mp) 
    {
       ISIDMapper ismapper;
       ISIterator iter;
       
       /* check if the dimensionalities match*/
       if ( this.dim() != blk.dim() )
       {
           new DimensionalityError(SDBError.Recoverable,
                 "ID Argument has wrong dimensionality.");
           return;
       }
       /* check if subbounds is validly inside entire bounds */
       if ( !this.bounds.contains( srcISB ))
       {
          new BoundsError( SDBError.Recoverable,
                "Given Bounds is not inside the DataSource.");
          return;
       }

       if ( this.res_level == 0 ) 
           this.source_dc.subblock( blk, dstISB, srcISB, mp );
       else
       {
          ISBounds srcBnds = this.mapper.toComponent( srcISB );
          if ( this.res_level > 0 ) 
          {
             DataBlock temp_blk = this.createDataBlock( srcBnds, mp ); 
             this.source_dc.subblock( temp_blk, temp_blk.getBounds(), srcBnds, mp );
             
             ismapper = new ISIDMapper( temp_blk.getBounds(), dstISB );
             
             iter = new ISIterator( dstISB );
             for( ; iter.valid(); iter.next())
                //blk.setDatum(iter,data[temp_blk.computeOffset(ismapper.toComponent(iter))]);
                blk.setDatum( iter, temp_blk.datum( ismapper.toComponent( iter )));
          }
          else if ( this.res_level < 0 ) 
          {
             ismapper = new ISIDMapper( srcISB, dstISB );
             iter = new ISIterator( dstISB );
             for ( ; iter.valid(); iter.next() )
                blk.setDatum( iter,source_dc.datum( ismapper.toComponent(iter), mp ));
          }
       }
    }
    

    /** Return the DataSource on which this VRDS is built */
    public Indexable getSourceIndexable() {
        return this.source_dc;
    }
    
    /** Return the description of the VRDS */
    public String toString(){
    
        StringBuffer result=new StringBuffer();
        result.append(super.toString());
        result.append("\n");
        
        result.append("DataSource Type: Variable Resolution Data Source\n");
        
        result.append("Current Resolution Level:");
        result.append(this.res_level);
        result.append("\n");
        return new String(result);
    }
    
    //static public void main(String[] args) throws IOException{
    static public void main(String[] args) {
       IntegerIndexSpaceID lower1 = new	IntegerIndexSpaceID(0,0);
       IntegerIndexSpaceID upper1 = new	IntegerIndexSpaceID(3,3);
       IntegerIndexSpaceID lower2 = new	IntegerIndexSpaceID(0,0);
       IntegerIndexSpaceID upper2 = new	IntegerIndexSpaceID(1,1);
       IntegerIndexSpaceID lower3 = new	IntegerIndexSpaceID(2,2);
       IntegerIndexSpaceID upper3 = new	IntegerIndexSpaceID(5,5);

       ISBounds ISBounds1,ISBounds2;
       ISBounds1 = new ISBounds(lower1,upper1);
       ISBounds2 = new ISBounds(lower2,upper2);
		//int isorderingarray[]={0,1};
		int dsorderingarray[]={0,1};
		//AxisOrdering isordering=new AxisOrdering(isorderingarray);
		AxisOrdering dsordering=new AxisOrdering(dsorderingarray);
		

		RecordDescriptor recordDesc; // 3 attributes
		Datum d;
		AttributeDescriptor[] attribs = new AttributeDescriptor[3];
		attribs[0]=new AttributeDescriptor("float","Carbon");
		attribs[1]=new AttributeDescriptor("float","Oxygen");
		attribs[2]=new AttributeDescriptor("float","Nitrogen");
		recordDesc = new RecordDescriptor(attribs);
		
		TextFileDataSource ds1,ds2;
		
		ds1=new TextFileDataSource("",recordDesc, ISBounds1,dsordering,"testdata/ar1.dat");
		ds2=new TextFileDataSource("", recordDesc, ISBounds2,dsordering,"testdata/ar2.dat");
         ds1.activate();
         ds2.activate();
		
		VRDataSource vrds = new VRDataSource("",ds1);

		ISIterator iter = new ISIterator(ds1.getBounds());
		
		System.out.print(iter);
		System.out.println(ds1.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			System.out.print(iter);
			d=ds1.datum(iter);
			System.out.println(d);
		}
		vrds.incr();
		ISBounds sub = new ISBounds(lower3,upper3);
		DataBlock blk = vrds.subblock(sub);
		sub.moveToOrigin();
		iter = new ISIterator(sub);
		
		System.out.print(iter);
		System.out.println(blk.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			System.out.print(iter);
			d=blk.datum(iter);
			System.out.println(d);
		}
		
		
		
		System.out.println("");
		vrds.decr();
		
		iter = new ISIterator(vrds.getBounds());
		
		System.out.print(iter);
		System.out.println(vrds.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			System.out.print(iter);
			d=vrds.datum(iter);
			System.out.println(d);
		}
		return;
		
		//ds2.activate();
	/*		ISIterator it = new ISIterator(ISBounds1,isordering);
			i=1;
			System.out.println("#"+i+" Datum");
			System.out.println(ds1.datum(it));
			while (it.hasMoreElements()) {
				it.next();
				System.out.println("#"+i+" Datum");
				d=ds1.datum(it);
				System.out.println(d);
				i++;
			
			}
	*/	
	/*
		//VRDataSource vrds = new VRDataSource(ds2,1);
		//System.out.println(vrds.decr());
		System.out.println(vrds.getBounds());
		ISIterator iter = new ISIterator(vrds.getBounds());
		i=1;
		System.out.println("#"+i+" Datum");
		System.out.println(vrds.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			System.out.println("#"+i+" Datum");
			d=vrds.datum(iter);
			System.out.println(d);
			i++;
		}
	*/
	
/*		System.out.println(vrds.decr());
		System.out.println(vrds.getBounds());
		iter = new ISIterator(vrds.getBounds());
		i=1;
		System.out.println("#"+i+" Datum");
		System.out.println(vrds.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			System.out.println("#"+i+" Datum");
			d=ds1.datum(iter);
			System.out.println(d);
			i++;
		}
*/		

/********* test blocked-vrds *************

IntegerIndexSpaceID offset1 = new IntegerIndexSpaceID(0,0);
IntegerIndexSpaceID offset2 = new IntegerIndexSpaceID(0,4);
VRDataSource vrds = new VRDataSource(ds2,1);
BlockLayout bl[] = new BlockLayout[2];
bl[0] = new BlockLayout(ds1,offset1);
bl[1] = new BlockLayout(vrds,offset2);
BlockedDataSource bds = new BlockedDataSource(bl);
VRDataSource vds = new VRDataSource(bds,-1);
ISBounds db_test= vds.getBounds();
ISIterator iter = new ISIterator(db_test);
		System.out.println(db_test);
		i=1;
		System.out.println("#"+i+" Datum");
		System.out.println(vds.datum(iter));
		while (iter.hasMoreElements()) {
			iter.next();
			i++;
			System.out.println("#"+i+" Datum");
			d=vds.datum(iter);
			System.out.println(d);
			
		}
/********** test subblock **************/
/*		VRDataSource vrds1 = new VRDataSource(ds1,0);
		System.out.println(vrds1.incr());
		PackedIndexSpaceID lower_test = new	PackedIndexSpaceID(0,0);
		PackedIndexSpaceID upper_test = new	PackedIndexSpaceID(3,3);
		ISBounds db_test=new ISBounds(lower_test,upper_test);
		iter = new ISIterator(db_test);
		System.out.println(db_test);
		DataBlock blk = vrds1.subblock(db_test);
		if (blk.datum(iter) == null) {
		 System.out.println("NULL");
		}
		else {
			System.out.println(blk.datum(iter));
		}
		while (iter.hasMoreElements()) {
			iter.next();
			
			Datum dd=blk.datum(iter);
			if (dd == null) {
		 		System.out.println("NULL");
			}
			else {
				System.out.println(dd);
			}
			
		}
*/

	} //end main()


}
