/*
 * Volume.java
 * ---------------------------------------------------
 * Edit history (since 06/05/08)
 * -----------------------------
 * 06/05/08 rdb: makde ArrayList generic
 */
package edu.unh.sdb.datasource;

import java.util.ArrayList;
import java.util.Arrays;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 *
 * A volume that can be intersected with a {@link Plane}.
 * @author dlipsa
 */
public class Volume
{
   /** Creates a new instance of VolumeInfo */
   public Volume (Point3d lower, Point3d upper)
   {
      this.lower = lower;lower.get (l);
      this.upper = upper;upper.get (u);
      maxDiagonal = calculateMaxDiagonal ();
      rayStart = new double[][][]
      {
         {   // x axis
            {lower.x, lower.y, lower.z},
            {lower.x, upper.y, lower.z},
            {lower.x, upper.y, upper.z},
            {lower.x, lower.y, upper.z},
         },
         {   // y axis
            {lower.x, lower.y, lower.z},
            {lower.x, lower.y, upper.z},
            {upper.x, lower.y, upper.z},
            {upper.x, lower.y, lower.z},
         },
         {   // z axis
            {lower.x, lower.y, lower.z},
            {lower.x, upper.y, lower.z},
            {upper.x, upper.y, lower.z},
            {upper.x, lower.y, lower.z},
         }
      };
      rayVector = new double[][]
      {
         {upper.x-lower.x, 0, 0},
         {0, upper.y-lower.y, 0},
         {0, 0, upper.z-lower.z}
      };
   }
   
   public Volume (ISBounds bounds)
   {
      this (new Point3d (bounds.getLowerArray ()[0], bounds.getLowerArray ()[1], bounds.getLowerArray ()[2]),
              new Point3d (bounds.getUpperArray ()[0], bounds.getUpperArray ()[1], bounds.getUpperArray ()[2]));
   }
   
   
   public double getMaxDiagonal ()
   {
      return maxDiagonal;
   }
   
   public Point3d[] calculateIntersectionPoints (Plane plane)
   {
      Vector3d n = plane.getNormal ();
      double d = -plane.get (3);
      ArrayList<Point3d> intersectionPoints =
              new ArrayList<Point3d>(12);
      
      // for all 3 directions
      for(int i = 0; i <3; i++)
      {
         Vector3d dv = new Vector3d (rayVector[i]);
         dv.normalize ();
         // for all 4 points
         for(int j = 0; j < 4; j++)
         {
            Vector3d p0 = new Vector3d (rayStart[i][j]);
            Point3d intersection =
                    SliceUtil.segmentPlaneIntersection (p0, dv, 0, rayVector[i][i], n, d);
            if (intersection != null)
               intersectionPoints.add (intersection);
         }
      }
      if (intersectionPoints.size () < 3)
         return null;
      else
         return intersectionPoints.toArray (new Point3d[0]);
   }
   
   public Point3d getLower ()
   {
      return lower;
   }
   
   public Point3d getUpper ()
   {
      return upper;
   }
   
   
   public double getLengthX ()
   {
      return Math.abs (upper.x - lower.x);
   }
   
   public double getLengthY ()
   {
      return Math.abs (upper.y - lower.y);
   }
   
   public double getLengthZ ()
   {
      return Math.abs (upper.z - lower.z);
   }
   
   public double[] getLowerArray ()
   {
      return l;
   }
   
   public double[] getUpperArray ()
   {
      return u;
   }
   
   @Override
   public String toString ()
   {
      return "(" + lower.toString () + ", " + upper.toString () + ")";
   }
   
   /**
    * Translate 'plane' along its normal such that it intersects the
    * volume at the farthest point (opposed/in the direction) to where the normal
    * to the plane points to.
    * To move the plane to the begining of the iteration pass (plane, l, u), to
    * move the plane to the end of the iteration pass (plane, u, l).
    */
   private static void movePlaneBeginEndIteration (Plane plane, double[] l, double[] u)
   {
      double[] p = new double[3];
      Vector3d n = plane.getNormal ();
      for (int i = 0; i < 3; i++)
         p[i] = corner (plane.get (i), u[i], l[i]);
      plane.set (3, - new Vector3d (p).dot (n));
   }
   
   void movePlaneBeginIteration (Plane plane)
   {
      movePlaneBeginEndIteration (plane, l, u);
   }
   
   void movePlaneEndIteration (Plane plane)
   {
      movePlaneBeginEndIteration (plane, u, l);
   }
   
   
   /**
    * See {@link ObliquePlaneIteratorInterface#getSpaceShapeArray}
    */
   int[] getSpaceShapeArray (Plane plane, int maxProjectionAxis)
   {
      Vector3d n = plane.getNormal ();
      Vector3d dv = new Vector3d (rayVector[maxProjectionAxis]);
      dv.normalize ();
      double oldD = plane.get (3);
      
      movePlaneBeginIteration (plane);double dBegin = -plane.get (3);
      movePlaneEndIteration (plane);double dEnd = -plane.get (3);
      
      double max = Double.MIN_VALUE, min = Double.MAX_VALUE;
      // for all 4 points
      for(int j = 0; j < 4; j++)
      {
         Vector3d p0 = new Vector3d (rayStart[maxProjectionAxis][j]);
         double val = SliceUtil.get (
                 SliceUtil.segmentPlaneIntersection (p0, dv, Double.NEGATIVE_INFINITY,
                 Double.POSITIVE_INFINITY, n, dBegin), maxProjectionAxis);
         max = Math.max (max, val);
         min = Math.min (min, val);
         
         val = SliceUtil.get (
                 SliceUtil.segmentPlaneIntersection (p0, dv, Double.NEGATIVE_INFINITY,
                 Double.POSITIVE_INFINITY, n, dEnd), maxProjectionAxis);
         max = Math.max (max, val);
         min = Math.min (min, val);
      }
      
      plane.set (3, oldD);
      int[] spaceShapeArray = new int[3];
      for (int i = 0; i < 3; i++)
         spaceShapeArray[i] = (int)((i == maxProjectionAxis) ? max - min : u[i] - l[i]);
      SliceUtil.logger.info ("maxProjectionAxis=" + maxProjectionAxis + 
         " spaceShapeArray=" + Arrays.toString (spaceShapeArray));
      return spaceShapeArray;
   }
   
   /**
    * Returns true if 'plane' intersects the volume, false otherwise
    * 3d Math Primer ..., Dunn & Parberry, ch 13.10, Intersection of AABB and
    * Plane
    */
   boolean validPosition (Plane plane)
   {
      if (plane == null)
         return false;
      double[] min = new double[3];
      double[] max = new double[3];
      for (int i = 0; i < 3; i++)
      {
         if (plane.get (i) > 0)
         {
            min[i] = l[i];
            max[i] = u[i];
         }
         else
         {
            min[i] = u[i];
            max[i] = l[i];
         }
      }
      Vector3d minV = new Vector3d (min);
      Vector3d maxV = new Vector3d (max);
      Vector3d n = plane.getNormal ();
      double d = - plane.get (3);
      double dotMaxN = maxV.dot (n);
      double dotMinN = minV.dot (n);
      if ( dotMaxN < d || dotMinN > d)
         return false;
      else
         return true;
   }
   
   private static double corner (double v, double l, double u)
   {
      return (v < 0) ? l : ((v < SliceUtil.ZERO) ? (l + u) / 2 : u);
   }
   
   private double calculateMaxDiagonal ()
   {
      return Math.sqrt (
              Math.pow (upper.x-lower.x, 2) +
              Math.pow (upper.y-lower.y, 2) +
              Math.pow (upper.z-lower.z, 2));
   }
   
   private Point3d lower, upper;
   private double maxDiagonal;
   private double[][][] rayStart;
   private double[][] rayVector;
   private double[] l = new double[3];
   private double[] u = new double[3];
}