/*
 * 
 * XDDLReader.java Author: Denise D. Mitchell 2002 University of New Hampshire
 * Durham, NH 03824
 * 
 */
//------------------- Edit history --------------------------------------
// 08/25/08 kjr: Made fdl file locating smarter to look in same directory
//                as the ddl file if it can't find it.
// 11/29/05 rdb: Revised handling of creation of component DataSources to go through
//                DataSource.create. Main advantage is to access the smarter
//                file location access included in that process. It also simplifies
//                this code quite a bit. For now I just commented out old code.
//               I haven't tested whether it will handle nested Composite DS.
//              -Added isDefault to test for empty String value
// 10/21/05 rdb: removed many blank lines; removed unused variables
//                forced references in main to unused variables we want to keep
// -----------------------------------------------------------------------
package edu.unh.sdb.datasource;

import java.util.*;
import java.io.File;
import java.nio.*;
import org.w3c.dom.*;
import edu.unh.sdb.common.*;

public class XDDLReader extends XMLReader
{

   /** declare a table/list to hold the actual datasources */
   protected DataSourceCollection dsList;

   /** the FDLDescriptor that will hold all data for current FDL */
   protected FDLDescriptor fdlDescriptor;

   /** interim lower bounds specifications */
   protected Vector<Integer> lowerAxes;

   /** interim upper bounds specifications */
   protected Vector<Integer> upperAxes;

   // declare the name of the pending datasource
   protected String dsName;

   // declare the ISBounds for the pending datasource
   protected ISBounds isBounds;

   // declare the AxisOrdering for the pending datasource
   protected AxisOrdering axisOrdering;

   // declare the DataSourceBuilder to process logical datasources
   protected DataSourceBuilder dsBuilder;

   /** code denoting a TextFileDataSource */
   private static String textType = "ascii";

   /** code denoting a BinaryFileDataSource */
   private static String binaryType = "binary";

   protected static Hashtable endiannessList = new Hashtable();

   /** for PublicDS */
   protected DataSource[] publicDSArray;
   protected ArrayList<String> publicDSNameList = new ArrayList<String>();

   /**
    * Takes an argument name of a file or URL as a String and a verbose
    * indicator
    */

   public XDDLReader( String inf, boolean v )
   {
      super( inf, v );
   }

   /**
    * Takes an argument name of a file or URL as a String and sets verbose to
    * false
    */
   public XDDLReader( String inf )
   {
      super( inf, false );
   }

   /** Instantiates the DataSourceCollection object to hold new parser info */
   /** Sets all pending values to null and sets the join flag to FALSE */
   protected void setDefaults()
   {
      if ( dsList == null )
         dsList = new DataSourceCollection();
      keyElement = "DataSource"; // this is the repeating element in the list
      dsName = null;
      lowerAxes = new Vector<Integer>();
      upperAxes = new Vector<Integer>();
      isBounds = null;
      axisOrdering = null;
      dsBuilder = null;
   }

   /** main method used for testing the class */
   static public void main( String[] argv )
   {
      System.out.println( textType + " " + binaryType ); // just to avoid compiler warning
      if ( argv.length != 1 )
      {
         System.err.println( "Missing fileName." );
         System.exit( 1 );
      }
      try
      {
         XDDLReader theApp = new XDDLReader( argv[0], true ); // instantiate
                                                               // the
                                                               // application
         if ( theApp == null )
         {
            theApp.processBlockedComponent( null ); // force reference to
                                                      // avoid warning
            System.exit( -1 );
         }
         // XDDLReader theApp = new XDDLReader(argv[0], false); // instantiate
         // the application
         // theApp.parse(); // Instantiate the parser &Parse the file
      } catch ( Exception e )
      {
         e.printStackTrace();
         System.exit( 1 );
      }
      System.exit( 0 );
   }

   /** process the main document in the xml file */
   protected boolean processDoc()
   {
      // Document doc = parser.getDocument();
      // Document is well-formed
      // Traverse all the children of the root element

      for ( Node child = doc.getDocumentElement().getFirstChild();
                 child != null; 
		 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "DataSource" ) )
            {
               processDataSource( child );
            }
            else if ( child.getNodeName().equals( "PublicDS" ) )
            {
               processPublicDS( child );
            } else if (child.getNodeName().equals( "lrucache") ) {
            
               //System.out.println("Found lrucache");
               processCache(child);
            }
         }
      } // end "for" loop, no more children nodes
      if ( verbose )
      {
         System.out.println( "final list of DataSources: " +
         getDataSources() );
      }
      finalProcessPublicDS();
      return true;
   }
   /** processes a single DataSource within the parsed tree * */
   private void processDataSource( Node n )
   {
      if ( verbose )
      {
         System.out.println( "  Processing Datasource Node: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) +
                                   " Type: " +
         makeChildrenText( n.getAttributes().getNamedItem( "type" ) ) );
      }

      dsName = makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) );
      if ( nameExists( makeChildrenText( n.getAttributes()
                                    .getNamedItem( "dsName" ) ) ) )
      { // discard if the datasource name is already in the table
         System.err.print( "Datasource: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) +
                " is a duplicate definition. Discarding this definition." );
         return;
      }

      if ( isPhysical( n ) || isDefault( n, "type") )
         processPhysicalDataSource( n ); // process as a physical file
      else // process as a composite file
         processCompositeDataSource( n );
   }


   /** An Extremely quick and dirty implementation that handles an LRU Cache. We
    *  should design and implement a general mechanism for all cache types.*/
   private void processCache( Node n )
   {

         String cdsName = makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) );
         String childName = makeChildrenText( n.getAttributes().getNamedItem( "childname" ) );
         int numBlocks = makeChildrenFloat(n.getAttributes().getNamedItem( "numblocks" ) ).intValue();
         
//         System.out.println( "  Processing Cache Node:" 
//                              + " dsName: "+ cdsName 
//                              + " numblocks: " + numBlocks
//                              + " childname: " + childName);
      
      
         ChunkedDataSource childDS = (ChunkedDataSource) dsList.getDataSource(childName);
   
         LRUCacheDataSource cds = new LRUCacheDataSource(cdsName, childDS ,numBlocks);
   
         addDataSource(cds);
         
         
   }


   /**
    * Returns true/false if the name exists in the
    * 
    * datasource collection
    */
   public boolean nameExists( String newDSName )
   {
      return dsList.nameExists( newDSName );
   }

   /** Adds a new datasource to the existing collection */
   private void addDataSource( DataSource ds )
   {
      if ( verbose )
         System.out.println( "ADDING DS: " + ds.getName() );
      getDataSources().add( ds );
   }

   /**
    * determines whether the datasource is being defined as a physical file
    * 
    * or a composite file such as an attibute joined or a blocked datasource.
    */
   private boolean isPhysical( Node n )
   {
      if ( makeChildrenText( n.getAttributes().getNamedItem( "type" ) )
                                    .equalsIgnoreCase( "physical" ) )
         return true;
      else
         return false;
   }

   /**
    * determines if the value of the node is an empty string
    */
   private boolean isDefault( Node n, String item )
   {
      if ( makeChildrenText( n.getAttributes().getNamedItem( item ) )
                                    .equals( "" ) )
         return true;
      else
         return false;
   }

   /**
    * determines whether the datasource is being defined as an attibute joined
    * datasource.
    */
   private boolean isAttributeJoined( Node n )
   {
      if ( makeChildrenText( n.getAttributes().getNamedItem( "type" ) )
                                    .equalsIgnoreCase( "joined" ) )
         return true;
      else
         return false;
   }

   /**
    * determines whether the datasource is being defined as an attibute joined
    * datasource.
    */
   private boolean isBlocked( Node n )
   {
      if ( makeChildrenText( n.getAttributes().getNamedItem( "type" ) )
                                    .equalsIgnoreCase( "blocked" ) )
         return true;
      else
         return false;
   }

   /**
    * determines whether the datasource is being defined as an attibute joined
    * datasource.
    */
   private boolean isVarRes( Node n )
   {
      if ( makeChildrenText( n.getAttributes().getNamedItem( "type" ) )
                                    .equalsIgnoreCase( "vares" ) )
         return true;
      else
         return false;
   }

   /** process node for a physical file including calling the XFDLReader */
   private void processPhysicalDataSource( Node n )
   {
      Node dsNameNode = n.getAttributes().getNamedItem( "dsName" );
      Node fileNameNode = n.getAttributes().getNamedItem( "fileName" );
      
      // if file specified by fileName doesn't exist, check and see if it does
      // exist when used as relative path from the directory that the .ddl is in.
      File file = new File(fileNameNode.getNodeValue());
      if(!file.exists()){
    	  file = new File(new File(inputFileName).getParent(), file.getName());
    	  if(file.exists()){
    		  fileNameNode.setNodeValue(file.getAbsolutePath());
    	  }
      }      
      
      if ( verbose )
      {
         System.out.println( "    Processing Physical DataSource: " 
           + makeChildrenText( dsNameNode )  + " for fileName: " 
           + makeChildrenText( fileNameNode )); 
      }
      // rdb new code 11/29/05
      // create a DataSource for the component
      DataSource component = DataSource.create( dsName, fileNameNode.getNodeValue() );
      if ( component != null )
         addDataSource( component );
      else
         new SDBError( SDBError.Warning, "Unable to create DataSource from " +
                                       fileNameNode.getNodeValue() );
     
      /********* old code ***************************
      XFDLReader fdlReader;
      if ( makeChildrenText( n.getAttributes().getNamedItem( "fileName" ) ) != null )
      { // there is an FDL fileName so call the FDLReader
         fdlReader = new XFDLReader(
             makeChildrenText( n.getAttributes().getNamedItem( "fileName" ) ), verbose );
         
         if ( fdlReader != null )
            
            return; // the reader couldn't process FDL so abort this
                     // datasource.
      }
      else
      { // there is not an FDL fileName specified so abort this datasource
         System.err.println( "Physical DataSource " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) +
                                    " does not have an FDL file URL" );
         return;
      }

      if ( physicalDataSourceValid( fdlReader ) ) // if the descriptor is valid
          createPhysicalDataSource( fdlReader ); // create DS and add to table
      /************************************/
   }

   /************ deleted rdb 11/29/05 *************************
   /** creates a physical datasource *
   private void createPhysicalDataSource( XFDLReader fdl )
   {
      if ( verbose )
      {
         System.out.println( "Creating textType DataSource: " +
                             fdl.getFDLDescriptor().getDataFileType() );
      }
      if ( fdl.getFDLDescriptor().getDataFileType().equals( textType ) )
      {
         if ( verbose )
         {
            System.out.println( "      Creating textType DataSource: " );
         }
         addDataSource( makeTextFileDataSource( fdl ) );
      }
      else if ( fdl.getFDLDescriptor().getDataFileType().equals( binaryType ) )
      {
         if ( verbose )
         {
            System.out.println( "      Creating binaryType DataSource: " );
         }
         addDataSource( makeBinaryFileDataSource( fdl ) );
      }
      if ( verbose )
      {
         System.out.println(
	     "list of DataSources after createPhysicalDataSource: " 
	      + getDataSources() ); 
      }
   }
   /***********************************************************/
   
   /** Returns a TextFileDataSource of file datasource. * 
    * 
    * rdb delete 11/29/05
    *
   private TextFileDataSource makeTextFileDataSource( XFDLReader fdl )
   {
      TextFileDataSource ds = new TextFileDataSource( getDSName(),
                 fdl.getFDLDescriptor().getRecordDescriptor(),
                 fdl.getFDLDescriptor().getBounds(),
                 fdl.getAxisOrdering(),
                 fdl.getFDLDescriptor().getDataFileURL() );
      if ( verbose )
      {
         if ( ds == null )
            System.out.println( "BinaryDS: is null" );
         else
            System.out.println( "BinaryFDS: " + ds.getName() );
      }
      return ds;
   }
  /***********************************************************/

   /** Returns a BinaryFileDataSource of file datasource. * 
    * 
    * rdb delete 11/29/05
    *
   private BinaryFileDataSource makeBinaryFileDataSource( XFDLReader fdl )
   {
      BinaryFileDataSource ds = null;
      /*
       * BinaryFileDataSource ds = new BinaryFileDataSource (
       * 
       * getDSName(),
       * 
       * fdl.getFDLDescriptor().getRecordDescriptor(),
       * 
       * fdl.getFDLDescriptor().getBounds(),
       * 
       * fdl.getAxisOrdering(),
       * 
       * fdl.getFDLDescriptor().getDataFileURL(),
       * 
       * fdl.getFDLDescriptor().isPointOrder());
       */
      /**** rdb 11/29/05 
      ds = BinaryFileDataSource.createBinaryFileDataSource(
                   getDSName(),
                   fdl.getFDLDescriptor().getRecordDescriptor(),
                   fdl.getFDLDescriptor().getBounds(),
                   fdl.getAxisOrdering(),
                   fdl.getFDLDescriptor().getDataFileURL(),
                   fdl.getFDLDescriptor().getByteOrder(),
                   fdl.getFDLDescriptor().getFieldLayout() );
      /*************************************
      ds = (BinaryFileDataSource) DataSource.create( getDSName(), fdl.getFDLDescriptor());
      ds.setAxisOrdering( fdl.getAxisOrdering() );
      
      if ( verbose )
      {
         if ( ds == null )
            System.out.println( "BFDS: is null" );
         else
            System.out.println( "BFDS: " + ds.getName() );
      }
      if ( ds != null )
      {
         ds.setFileOffset( fdl.getFileOffset() );
         endiannessList.put( getDSName(), fdl.getByteOrder() );
      }
      if ( verbose )
      {
         System.err.println( "Offset=" + fdl.getFileOffset() );
         System.err.println( "Endianness=" + fdl.getByteOrder() );
      }
      return ds;
   }
/***********************************************************/

   /**
    * determine whether we have all of the data needed for a physical datasource
    * 
    * and report on any errors
    * rdb deleted 11/29/05
   private boolean physicalDataSourceValid( XFDLReader fdl )
   {
      return true;
   }
   /*****************************/

   private void processCompositeDataSource( Node n )
   {
      if ( isAttributeJoined( n ) )
         processAttributeJoinedDataSource( n );
      else if ( isBlocked( n ) )
         processBlockedDataSource( n );
      else if ( isVarRes( n ) )
         processVRDataSource( n );
      else
         new SDBError( SDBError.Warning, "Unknown datasource type: " +
                 makeChildrenText( n.getAttributes().getNamedItem( "type" ) ) );
      return;
   }

   /**
    * process node for a attribute joined datasource
    * 
    * including processing components
    */
   private void processAttributeJoinedDataSource( Node n )
   {
      if ( verbose )
      {
         System.out.println( "    Processing Attribute Joined DataSource: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) );
      }
      // does the name exist in the table?
      // yes, warn of duplicate and abort this datasource
      // no, continue.
      dsBuilder = new AJDataSourceBuilder( dsList, dsName );// start a
                                                            // datasource
                                                            // builder
      for ( Node child = n.getFirstChild(); 
		 child != null;
                 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "Component" ) )
            {
               processComponent( child );
            }
            else if ( child.getNodeName().equals( "Bounds" ) )
            {
               processCompositeBoundsPair( child );
            }
         }
      } // end "for" loop, no more components to process nodes
      processFinalCompositeBounds();
      addDataSource( getDataSourceBuilder().makeDataSource() );
      if ( verbose )
         System.out.println( "datasource: " + 
			 getDataSourceBuilder().makeDataSource() .toString() );
   }

   // ** process the details of an individual component
   private void processComponent( Node n )
   {
      addComponent( n ); // establish the component
      // check for child nodes for bounds and other attributes
      for ( Node child = n.getFirstChild(); 
		 child != null;
                 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "Bounds" ) )
            {
               processComponentBoundsPair( child );
            }
            else if ( child.getNodeName().equals( "Map" ) )
            {
               processComponentFieldPair( child );
            }
         }
      } // end "for" loop, no more components to process nodes
      getDataSourceBuilder().processFinalComponentMap();
   }

   /** process the component upper and lower bounds pairs */
   private void processComponentBoundsPair( Node n )
   {
      getDataSourceBuilder().setComponentAxes(
      makeChildrenInteger( n.getAttributes().getNamedItem( "lower" ) ),
      makeChildrenInteger( n.getAttributes().getNamedItem( "upper" ) ) );
   }

   /**
    * process the field ID Mapping within a component of an AJDS
    * 
    * XML example: <Map fieldName="carbon1" position="1" />
    */
   private void processComponentFieldPair( Node n )
   {
      if ( verbose )
         System.out.println( "inside ProcessComponentFieldPair" );
      getDataSourceBuilder().setComponentFieldPair(
            makeChildrenText( n.getAttributes().getNamedItem( "fieldName" ) ),
            makeChildrenInteger( n.getAttributes().getNamedItem( "position" ) ),
            makeChildrenText( n.getAttributes().getNamedItem( "newName" ) )
      );
   }

   /**
    * process node for a Blocked DataSource including
    * 
    * processing invidivual components
    */
   private void processBlockedDataSource( Node n )
   {
      if ( verbose )
      {
         System.out.println( "    Processing Blocked DataSource: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) );
      }
      // does the name exist in the table?
      // yes, warn of duplicate and abort this datasource
      // no, continue.
      dsBuilder = new BlockedDataSourceBuilder( dsList, dsName );// start a
                                                                  // datasource
                                                                  // builder
      for ( Node child = n.getFirstChild(); 
		 child != null;
                 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "Component" ) )
            {
               processComponent( child );
            }
            else if ( child.getNodeName().equals( "Bounds" ) )
            {
               processCompositeBoundsPair( child );
            }
         }
      } // end "for" loop, no more components to process nodes

      processFinalCompositeBounds();
      addDataSource( getDataSourceBuilder().makeDataSource() );
      if ( verbose )
         System.out.println( "datasource: "
                   + getDataSourceBuilder().makeDataSource().toString() );
   }

   /** process all of the components of a blocked datasource */
   private void processBlockedComponent( Node n )
   {
      addComponent( n ); // establish the component
      // check for child nodes for bounds and other attributes
      for ( Node child = n.getFirstChild(); 
                 child != null; 
                 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "Bounds" ) )
            {
               processComponentBoundsPair( child );
            }
         }
      } // end "for" loop, no more components to process nodes
   }

   /**
    * process node for a Variable Resolution DataSource including
    * 
    * processing one invidivual components
    */
   private void processVRDataSource( Node n )
   {
      if ( verbose )

      {
         System.out.println( "    Processing Variable Resolution DataSource: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) );
      }
      // does the name exist in the table?
      // yes, warn of duplicate and abort this datasource
      // no, continue.
      dsBuilder = new VRDataSourceBuilder( dsList, dsName );// start a
                                                            // datasource
                                                            // builder
      for ( Node child = n.getFirstChild(); 
		 child != null;
                 child = child.getNextSibling() )
      {
         // When child is an element
         if ( child instanceof Element )
         {
            if ( child.getNodeName().equals( "Component" ) )
            {
               processComponent( child );
            }
            // only a single component is allowed, no bounds, no field mapping,
            // no axis axisOrdering
         }
      } // end "for" loop, no more components to process nodes
      addDataSource( getDataSourceBuilder().makeDataSource() );
      if ( verbose )
         System.out.println( "datasource: "
                       + getDataSourceBuilder().makeDataSource().toString() );
   }

   /** processes a single component of a composite datasource */
   private void addComponent( Node n )
   {
      if ( verbose )
      {
         System.out.println( "Component dsName: " +
         makeChildrenText( n.getAttributes().getNamedItem( "dsName" ) ) );
      }
      if ( dsBuilder.addComponent( makeChildrenText( n.getAttributes()
                                    .getNamedItem( "dsName" ) ) ) != true )
         System.err.println( "Component request discarded." );
      return;
   }

   /** */
   private void processCompositeBoundsPair( Node n )
   {
      lowerAxes.addElement( makeChildrenInteger(
      n.getAttributes().getNamedItem( "lower" ) ) );
      upperAxes.addElement( makeChildrenInteger(
      n.getAttributes().getNamedItem( "upper" ) ) );
   }

   /****************************************************************************
    * instantiate an ISBounds using the pairs gathered in the DDL
    * 
    * for the composite datasource
    ***************************************************************************/
   private void processFinalCompositeBounds()
   {
      // int[] lowerAxesArray, upperAxesArray;
      // did user specify any bounds for this datasource?
      if ( lowerAxes.size() < 1 )
         return; // no bounds, just do nothing
      getDataSourceBuilder().setCompositeBounds( 
               new ISBounds(
                   new IntegerIndexSpaceID( vectorToIntArray( lowerAxes ) ),
                   new IntegerIndexSpaceID( vectorToIntArray( upperAxes ) ) ) );
   }

   /** returns an FDLDescriptor object containing all of the files attributes */
   public FDLDescriptor getFDLDescriptor()
   {
      return fdlDescriptor;
   }

   /**
    * Sets the bounds of the Index Space for this datasource. If a null
    * reference is
    * 
    * passed, then an IllegalArgumentException is thrown.
    * 
    */
   public void setBounds( ISBounds bounds )
   {
      if ( bounds != null )
      {
         isBounds = bounds.copy();
      }
      else
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "XDDLReader.setBounds() was passed a null reference" );
      }
   }
   /**
    * Returns the bounds of the Index Space for this
    * 
    * FDLDescriptor.
    * 
    */
   public ISBounds getBounds()
   {
      if ( isBounds != null )
         return isBounds.copy();
      else return null;
   }

   /** Returns the datasource name */
   public String getDSName()
   {
      return dsName;
   }

   /** returns the DataSourceBuilder to a new builder * */
   public DataSourceBuilder getDataSourceBuilder()
   {
      return dsBuilder;
   }

   /** Returns the list of Datasources */
   public DataSourceCollection getDataSources()
   {
      return dsList;
   }

   /** Returns the endianness of the datasource specified by dsName */
   public ByteOrder getByteOrder( String dsName )
   {
      if ( !endiannessList.containsKey( dsName ) )
      {
         new XMLReaderError( SDBError.Recoverable, "the datasource " + dsName
                                       + " is not found" );
      }
      return (ByteOrder) endiannessList.get( dsName );
   }

   /**
    * Added by Gang Lu supporting for public DataSource processes a single
    * PublicDS within the parsed tree
    */

   private void processPublicDS( Node n )
   {
      String name = makeChildrenText(
                ( n.getAttributes().getNamedItem( "dsName" ) ) ).trim();
      if ( publicDSNameList == null )
         publicDSNameList = new ArrayList<String>();
      publicDSNameList.add( name );

      if ( verbose )
      {
         System.out.println( "  Adding PublicDS: " + name );
      }
   }

   private void finalProcessPublicDS()
   {
      if ( publicDSNameList != null )
      {
         int i;
         int size = publicDSNameList.size();
         DataSource[] tmpDSArray = new DataSource[size];
         DataSource ds = null;

         if ( verbose )
            System.out.println( "Final process of PublicDS" );

         for ( i = 0; i < size; i++ )
         {
            ds = dsList.getDataSource( publicDSNameList.get( i ) );

            if ( ds != null )
            {
               tmpDSArray[i] = ds;
               if ( verbose )
                  System.out.println( "PublicDS: " + ds.getName() + 
				  " is added to the publicDSArray" );
            }
         }

         if ( tmpDSArray.length == size )
            publicDSArray = tmpDSArray;
         else
         {
            publicDSArray = new DataSource[tmpDSArray.length];
            for ( i = 0; i < publicDSArray.length; i++ )
               publicDSArray[i] = tmpDSArray[i];
         }
      }
   }

   public DataSource getPublicDS( String name )
   {
      if ( publicDSArray != null )
      {
         for ( int i = 0; i < publicDSArray.length; i++ )
         {
            if ( publicDSArray[i].getName().equals( name ) )
               return publicDSArray[i];
         }
      }
      return null;
   }

   public DataSource[] getAllPublicDS()
   {
      return publicDSArray;
   }
}
