/* 
    XTDLReader.java
    Developed by Denise Mitchell, Summer 2002
*/
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables; commented out rootId uses
//-----------------------------------------------------------------------
package edu.unh.sdb.datasource;
import org.w3c.dom.*;
import edu.unh.sdb.common.*;

public class XTDLReader extends XMLReader 
{
  /** declare a table/list to hold the actual type objects */
  private TypeTable typeList;

  /** flag to denote style of type being created (simple, ranged, record, etc) */
  private int userTypeStyle;

  /** denotes a simple UserType */
  public static final int SIMPLE_TYPE = 0;

  /** denotes a RandedType */
  public static final int RANGED_TYPE = 1;

  /** denotes a RecordType */
  public static final int RECORD_TYPE = 2;

  /** declare the name of the pending Type */
  private String typeName;

  /** declare the pending storage type */
  private String storageType;

  /** declare the pending default value */
  private String defaultValue;

  /** declare the pending range lower range */
  private String lowerRange;

  /** declare the pending range upper range */
  private String upperRange; 

  /** declare the pending range exception value */
  private String exceptionValue; 

  /** declare the DataSourceBuilder to process logical datasources */
  private UserTypeBuilder builder;

/** Takes an argument name of a file or  URL as a String and a verbose indicator*/
  public XTDLReader( String inf, boolean v) 
  {
    super(inf, v); //start the parser with the verbose flag
  }

  /** Takes an argument name of a file and default to non-verbose **/ 
  public XTDLReader(String inf)
  {
    super(inf); //start the parser with the verbose flag turned off
  }

  /** Sets all pending values to null  */
  protected void setDefaults()
  {
    if (typeList == null)
            typeList = new TypeTable(); 

    typeName = "";
    storageType = "";
    defaultValue = "";
    lowerRange = "";
    upperRange = "";
    exceptionValue = "";
    builder = null;         
  }

  /** main method used for testing the class */
  static public void main(String[] argv)
  {        
    if (argv.length != 1) {
      System.err.println("Missing filename.");
      System.exit(1);
      }
        
    try {
      XTDLReader theApp = new XTDLReader(argv[0], true); // instantiate the application
      if ( theApp == null ) //  just to avoid compile warnings for unreferenced vars
      {
         if ( theApp.defaultValue == null || theApp.userTypeStyle == 0 )
             System.exit( -1 );
      }
    } catch (Exception e)
      {
        e.printStackTrace();
        System.exit(1);
      }
        System.exit(0);
    }
     
  /** process the main document in the xml file */
  protected boolean processDoc()
  {
      //Document doc = parser.getDocument();
        // Document is well-formed
        // Traverse all the children of the root element
      for (Node child = doc.getDocumentElement().getFirstChild();
               child != null; child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {
          if (child.getNodeName().equals("Type"))
          {
            processType(child);
          }
        }
       } // end "for" loop, no more children nodes
        return true;
}

    /** processes a single type node **/
  private void processType(Node n)
  {
    if(verbose)
    {
      System.out.println("  Processing Type - name: " +
          makeChildrenText(n.getAttributes().getNamedItem("typeName")) +
            " storageType: " +
          makeChildrenText(n.getAttributes().getNamedItem("storageType")) +
            " default: " +
          makeChildrenText(n.getAttributes().getNamedItem("default"))
          );
    }
    builder = null; //clear the builder before starting a new one  
    userTypeStyle = SIMPLE_TYPE; //assume all types are simple user times until
                            //other elements are detected.
    typeName = makeChildrenText(n.getAttributes().getNamedItem("typeName"));
    storageType = makeChildrenText(n.getAttributes().getNamedItem("storageType"));
    defaultValue = makeChildrenText(n.getAttributes().getNamedItem("default"));
    Node child = n.getFirstChild();
      while (child != null)
      {  // There is a child
        if (child instanceof Element) //and it is an element
        {
          processComplexUserType(child);
          break;
        }
        else
          child = child.getNextSibling();
      }

     if (builder == null)
         processSimpleType();
     if (builder != null)
       addUserType(builder.makeUserType());
  }
  /** Dispatches the process for subnodes depending on the user type */
  private void processComplexUserType(Node child)
  {
    if (child.getNodeName().equals("Range"))
    {
      processRangedType(child);
    }
    else if (child.getNodeName().equals("Field"))
    {
      processRecordType(child);
    }
  }

  /** adds a completed type to the TypeTable **/ 
  public void addUserType(UserType t)
  {
     if (verbose)
        System.out.println("ADDING Type: " + t.getName());

      int tNum = typeList.insert(builder.makeUserType());
        
     if (verbose)
        System.out.println("Added Type: " + typeList.get(tNum) + " as # " + tNum);
  }

/**  Processes a type that is not a Ranged or Record Type**/
private void processSimpleType()
{
    if(verbose)
    {
      System.out.println("   Within processSimpleType " );
    }

    builder = new UserTypeBuilder(typeName, storageType);
    return;
  }

/**  Processes the range attributes**/
private void processRangedType(Node n)
{
    if(verbose)
    {
      System.out.println("   Within ProcessRangedType: Lower: " +
          makeChildrenText(n.getAttributes().getNamedItem("lower")) + 
          " Upper: " + makeChildrenText(n.getAttributes().getNamedItem("upper")) +
          " exception: " + makeChildrenText(n.getAttributes().getNamedItem("exception"))
            );
    }

    if (makeChildrenText(n.getAttributes().getNamedItem("lower")) != null)
    {  //there is a lower range
       lowerRange = makeChildrenText(n.getAttributes().getNamedItem("lower"));
    }
    if (makeChildrenText(n.getAttributes().getNamedItem("upper")) != null)
    {  //there is an upper range
       upperRange = makeChildrenText(n.getAttributes().getNamedItem("upper"));
    }

    if (makeChildrenText(n.getAttributes().getNamedItem("exception")) != null)
    {  //there is an exception value
       exceptionValue = makeChildrenText(n.getAttributes().getNamedItem("exception"));
    }

    builder = new RangedTypeBuilder(typeName, storageType, lowerRange, upperRange, exceptionValue);
    return;
  }

/**  Processes a type that is Record Type including all of it's FIELD nodes**/
  private void processRecordType(Node n)
  {
    if(verbose)
    {
      System.out.println("   Within processRecordType " );
    }

    RecordDescriptorBuilder tempType = new RecordDescriptorBuilder(typeName);
    while (n != null)
    {       
      if (n instanceof Element)
      {
         if (n.getNodeName().equals("Field"))
          {
            if(verbose)
            {
              System.out.println("       adding field: " + 
                    makeChildrenText(n.getAttributes().getNamedItem("fieldName")) +
                    " of type: " + makeChildrenText(n.getAttributes().getNamedItem("fieldType"))
                    );
            }
        
       //** make up the attribute **/
      tempType.addField( 
              makeChildrenText(n.getAttributes().getNamedItem("fieldName")) ,
              makeChildrenText(n.getAttributes().getNamedItem("fieldType")) );
          }
      }
        
      n = n.getNextSibling();  //get the next sibling for another field definition
    }
    
    //we have processed all of the siblings.
    builder = tempType;
    return;
  }

  /** Returns the Type name*/
  public String getTypeName()
  {
    return typeName;
  }
   
  /** returns the DataSourceBuilder to a new builder **/
  public UserTypeBuilder getBuilder()
  {
    return builder;
  }
  
  /** returns the TypeTable collection class **/
  public TypeTable getTypes()
  {
    return typeList;
  }
}
