//
//  ArrayApproximator.java
//  Unstructured
//
//  Created by Philip Rhodes on Wed Sep 25 2002.
//  Copyright (c) 2002 University of New Hampshire. All rights reserved.
//
//--------------------------------------------------------------------------
//Edit history (since 01/03/05
// 12/07/05 rdb Fixed use of FieldIDMapper to correctly map result.
// 03/04/05 rdb Changed "approximate" method to adjust weights to insure sum to 1
//              Added option to use FieldIDMapper as well as RecordSpec and combined
//                   code to avoid replication.
// 01/03/05 rdb Replaced new FloatDatum( float ) with 
//                       (FloatDatum) Datum.createDatum( float )
//
//--------------------------------------------------------------------------

package edu.unh.sdb.lattice; 

import edu.unh.sdb.common.*;

/** The ArrayApproximator class is meant for use with the ArrayCell class.
 *
 */
public class ArrayApproximator extends Approximator 
{
    public ArrayApproximator()
    {
    }
    
    /**
     *  Return an interpolated a Datum containing interpolated values of the 
     * attributes at the specified Point inside the cell.
     */
    final public Datum approximate( Cell c, Point p )
    {
       return approximate( c, p, (FieldIDMapper) null );
    }

    /**
     *  Return an interpolated a Datum containing interpolated values of the 
     * a subset of attributes (specified by the RecordSpec) at the specified 
     * Point inside the cell.
     */
    final public Datum approximate( Cell c, Point p, RecordSpec rs )
    {
       return approximate( c, p, rs.getFieldIDMapper() );
    }
    /**
     *  Return an interpolated a Datum containing interpolated values of the 
     * a subset of attributes (specified by the FieldIDMapper) at the specified 
     * Point inside the cell.
     */
    public Datum approximate( Cell c, Point p, FieldIDMapper fim )
    {
        // see http://scijou.eou.edu/scijou98/node88.html for the math
        // behind this algorithm
        //
        // rdb: code is somewhat flawed. It works "correctly" if the Cell is defined
        // with points at each of the corners of the cell, but that isn't required 
        // -- we let a cell be defined by any # of points -- one possible user error 
        // might be to define it by 2 opposite corners. The problem with the 
        // interpolation done here is that the sum of the weights for each of the 2  
        // points is NOT 1. This could be fixed by computing the sum of weights and
        // then scale up the final data value by 1/sum.
        // Probably we should force all corner points to be specified. Until we do
        // that, however, I've fixed this code to handle those cases more correctly.
        
        int     d              = c.dim();
        float[] pointCoords    = p.getCoords();
        float[] vertexCoords;
        Point[] vertices       = c.getVertexCoords();
        Datum[] data           = c.getData();
        int     numVertices    = vertices.length;
        int     numAttributes;
        float[] cellDimensions = ((ArrayCell)c).getDimensions();
 
        float   weight;
        float   sumOfWeights = 0;    
        float   epsilon      = 0.001f;   // don't need weight to sum precisely to 1
        
        if ( fim == null )
           numAttributes = data[0].size();
        else
           numAttributes = fim.size();

        float[] resultArray    = new float[ numAttributes];
         
        for ( int i=0; i<numVertices; i++)
        {
            vertexCoords = vertices[i].getCoords();
            weight = 1;
            
            for ( int j=0; j<d; j++ )
            {
               weight *= 1 - Math.abs( pointCoords[j]-vertexCoords[j])
                                           / cellDimensions[j];
            }
                        
            sumOfWeights += weight;
            for ( int k=0; k<numAttributes; k++ )
            {
               if ( fim == null )
                  resultArray[k] += weight * data[i].getFloat(k);
               else
                  resultArray[ fim.toDestination( k ) ] 
                               += weight * data[i].getFloat( fim.toSource( k ));
            }
        }
        
        // check for appropriate normalization
        if ( sumOfWeights < 1.0f - epsilon )
        {
           float scale = 1.0f / sumOfWeights;
           for ( int i = 0; i < numAttributes; i++ )
              resultArray[ i ] *= scale;
        }
        return new FloatDatum( resultArray );
    }
    /*************
    public Datum approximate(Cell c, Point p, RecordSpec s)
    {
        //return this.approximate(c, p).subset(s);
        
        // see http://scijou.eou.edu/scijou98/node88.html for the math
        // behind this algorithm
        
        int i, j, k, d=c.dim();
        float product;
        float [] pointCoords=p.getCoords();
        float [] vertexCoords;
        Point [] vertices=c.getVertexCoords();
        Datum [] data=c.getData();
        int numVertices=vertices.length;
        int numAttributes=s.size();
        float resultArray[]=new float[numAttributes];
        float [] cellDimensions=((ArrayCell)c).getDimensions();
        int [] fieldNumbers=s.getFieldNumbers();
        
        for(i=0;i<numVertices;i++){
            
            vertexCoords=vertices[i].getCoords();
            product=1;
            
            for(j=0;j<d;j++){
            
                 product *= 1 - Math.abs(pointCoords[j]-vertexCoords[j])/(cellDimensions[j]);
            }
            
            
           for(k=0; k<numAttributes; k++){
           
                resultArray[k] += product * data[i].getFloat(fieldNumbers[k]);
           }

        }
        
        return new FloatDatum(resultArray);
    }
    /*********************************/
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        return new String( indent+"ArrayApproximator:" );
    }
    
    public static void main( String[] args ){
    
    
        System.out.println("*************ArrayApproximator Tests*************\n");
        ArrayApproximator appr=new ArrayApproximator();        
        
        System.out.println("--1D Tests--");
        Datum [] data=new FloatDatum [] {
              (FloatDatum)Datum.create(2.0f), (FloatDatum)Datum.create(3.0f)};        
        Point [] vertices=new Point [] {new Point(0.0), new Point(1.0)};
        Point queryPoint;
        
        ArrayCell c=new ArrayCell(vertices, data );


    
        System.out.println("Vertices at: ");
        System.out.println("\t" + vertices[0] + "with data value " + data[0] + ".");
        System.out.println("\t" + vertices[1] + "with data value " + data[1] + ".");

        System.out.println();

        queryPoint=new Point(0.0);

        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.0}-->"+appr.approximate(c,queryPoint));
        
        queryPoint=new Point(0.25);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.25}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(0.5);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.5}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(0.75);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.75}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(1.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {3.0}-->"+appr.approximate(c,queryPoint));
                            
                            
/*                            
        queryPoint=new Point(2.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {4.0}-->"+appr.approximate(c,queryPoint));

System.exit(0);
*/
        System.out.println("\n");
        vertices= new Point [] {new Point(6.0), new Point(8.0)};
        c=new ArrayCell(vertices, data );
        
        System.out.println("Vertices at: ");
        System.out.println("\t" + vertices[0] + "with data value " + data[0] + ".");
        System.out.println("\t" + vertices[1] + "with data value " + data[1] + ".");

        System.out.println();

        queryPoint=new Point(6.0);

        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.0}-->"+appr.approximate(c,queryPoint));
        
        queryPoint=new Point(6.5);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.25}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(7.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.5}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(7.5);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.75}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(8.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {3.0}-->"+appr.approximate(c,queryPoint));


        System.out.println("\n");
        vertices= new Point [] {new Point(-3.0), new Point(1.0)};
        c=new ArrayCell(vertices, data );
        
        System.out.println("Vertices at: ");
        System.out.println("\t" + vertices[0] + "with data value " + data[0] + ".");
        System.out.println("\t" + vertices[1] + "with data value " + data[1] + ".");

        System.out.println();

        queryPoint=new Point(-3.0);

        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.0}-->"+appr.approximate(c,queryPoint));
        
        queryPoint=new Point(-2.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.25}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(-1.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.5}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(0.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {2.75}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(1.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {3.0}-->"+appr.approximate(c,queryPoint));
    

        /***********************
        
                2D Tests
                
        ***********************/

        System.out.println("\n");
        System.out.println("--2D Tests--");
        /* rdb
        data=new FloatDatum[] { 	new FloatDatum(2.0f), 
                            new FloatDatum(3.0f), 
                            new FloatDatum(4.0f), 
                            new FloatDatum(5.0f)
                         }; 
        /*** end rdb ******/
        data=new FloatDatum[] { 	
              (FloatDatum)Datum.create(2.0f), 
              (FloatDatum)Datum.create(3.0f), 
              (FloatDatum)Datum.create(4.0f), 
              (FloatDatum)Datum.create(5.0f)
           }; 
                         
        vertices= new Point[] {	new Point(0.0, 0.0),
                                new Point(0.0, 2.0),
                                new Point(2.0, 2.0), 
                                new Point(2.0, 0.0)
                              };
        
        Datum result = null;
        
        c=new ArrayCell(vertices, data );
    
        System.out.println("Vertices at: ");
        System.out.println("\t" + vertices[0] + "with data value " + data[0] + ".");
        System.out.println("\t" + vertices[1] + "with data value " + data[1] + ".");
        System.out.println("\t" + vertices[2] + "with data value " + data[2] + ".");
        System.out.println("\t" + vertices[3] + "with data value " + data[3] + ".");

        System.out.println();

        queryPoint=new Point(0.0,0.0);

        result = appr.approximate( c, queryPoint );
        System.out.println("Approximate value at " + queryPoint +
                            " should be {2.0}-->" + result);
        
        queryPoint=new Point(0.0, 2.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {3.0}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(2.0, 2.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {4.0}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(2.0, 0.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {5.0}-->"+appr.approximate(c,queryPoint));

        queryPoint=new Point(1.0, 1.0);
        
        System.out.println("Approximate value at "+queryPoint+
                            " should be {3.5}-->"+appr.approximate(c,queryPoint));
        
        System.out.println("\n--- Testing multiple attributes -----" );
        
        Datum[] datums = new Datum[ 4 ];
        for ( int i = 0; i < 4; i++ )
        {
           float off = i*10;
           datums[i] = new FloatDatum( new float[] { off, off+1, off+2, off+3, off+4});
        }
        ArrayCell ac = new ArrayCell( vertices, datums );
        
        System.out.println("Vertices at: ");
        System.out.println("\t" + vertices[0] + "with data value " + datums[0] + ".");
        System.out.println("\t" + vertices[1] + "with data value " + datums[1] + ".");
        System.out.println("\t" + vertices[2] + "with data value " + datums[2] + ".");
        System.out.println("\t" + vertices[3] + "with data value " + datums[3] + ".");

        System.out.println();
        
        queryPoint = new Point( 1.0, 1.0 );
        result = appr.approximate( ac, queryPoint );
        System.out.println( "Approximate value at " + queryPoint +
              " should be {15, 16, 17, 18, 19 }-->" + result );
        
        
        System.out.println("\n--- Testing RecordSpec ---- " );
        RecordSpec rs = new RecordSpec( new int[] {1,3,4} );
        result = appr.approximate( ac, queryPoint, rs );
        System.out.println( "Selecting {1,3,4} Approximate value at " + queryPoint +
              " should be {16, 18, 19}-->" + result );
        result = appr.approximate( ac, queryPoint, new RecordSpec( new int[] {2,0}));
        System.out.println( "Selecting {2,0} Approximate value at " + queryPoint +
              " should be {17, 15}-->" + result );
        
       
             
    }
}
