/*
    GBoundsGapIterator.java
    Copyright Oct. 2002 University of New Hampshire
    Durham, NH 03824

    This program is to code the GBoundsGapIterator class, which traverses 
    the specified IndexSpace with the specified shape and the specified
    steps.

    use the while loop or for loop to traverse:

    GBoundsGapIterator iter = new GBoundsGapIterator(dsBounds, shape, gaps);
    
    while (iter.valid()) {
        System.out.println(iter.toString());
        iter.next();
    }

    or 

    for(iter.init(); iter.valid(); iter.next()) {
        System.out.println(iter.toString());
    }

*/
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------


package edu.unh.sdb.lattice;
//import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;


public class GBoundsGapIterator extends GBoundsIterator implements Cloneable, SDBIterator
{

    private AxisOrdering ordering;
	
    private boolean valid = false;
	
    private float[] shape;
    private float[] steps;
	private float[] start;
	
    private float[] ubounds;
    private float[] lbounds;
	
	private int[] partitionIndex;
	private int[] pUBounds;
	
	//private float epsilon = 0.001f;
    //private float epsilon = 0.0001f;
	
    //------------------------------------------------------------------------    
    /**	Construct an GBoundsGapIterator that will traverse the given space with an
     *   an GBounds of the given shape (dimensions). The last argument specifies
     *   the desired gaps between GBounds in each dimension.  A gaps array of all
     *   zeros means the generated GBounds all abut exactly. If a gap value is
     *   positive, then there will be that many rows/columns between the generated
     *   GBounds in that dimension. If any of the gaps are negative, then the 
     *   generated GBounds will overlap in the corresponding dimension.  ( e.g. 
     *   a gap of -1 means the Bounds will overlap by 1 row/column.)
     *   
     *   Values in the shape array must be >= 1, and values in the gaps array must
     *   allow the iterator to make forward progress. In particular, for negative
     *   gap values, the magnitude must be less than the shape value of the same
     *   dimension. 
     * 
     */
    public GBoundsGapIterator(GBounds space, float shape[], float gaps[]) {

        super(space, shape);

        this.ordering = new AxisOrdering(space.dim());

		setupAndCheck(shape, gaps);    
		        
        this.init(space);
     }

    public GBoundsGapIterator(GBounds space, float shape[], float gaps[], AxisOrdering ordering) {

        super(space, shape);

        this.ordering = ordering.copy();
		
		setupAndCheck(shape, gaps);    
		        
        this.init(space);
     }


    public GBoundsGapIterator(GBounds space, Point start, float shape[], float gaps[]) {

        super(space, shape);

        this.ordering = new AxisOrdering(this.dim());
		
		setupAndCheck(shape, gaps);
		
        this.init(space, start);
     }

    public GBoundsGapIterator(GBounds space, Point start, float shape[], float gaps[], AxisOrdering ordering) {

        super(space, shape);

        this.ordering = ordering.copy();
		
		setupAndCheck(shape, gaps);
		
        this.init(space, start);
     }


	/* Allocates space for private arrays, and checks the shape and gaps arrays for improper values. */
	private final void setupAndCheck(float shape[], float gaps[]){

		int i, d=this.dim();
		
        this.shape = new float[d];        
        this.steps = new float[d];
		this.start = new float[d];
		
		this.partitionIndex = new int[d];      
		  
		this.pUBounds = new int[d];
		
        /* check shape and set the value of shape */
        for( i = 0; i < d; i++ ){
		
           if ( shape[i] <= 0.0 ) {
               new IllegalArgumentError(SDBError.Fatal,"Shape dimensions must be > 0.");
           }

           this.shape[i] = shape[i];
        }
      
        // Build the steps array, and check for illegal gap values.
        for(i=0; i<d; i++){
        
            this.steps[i] = this.shape[i] + gaps[i];
            
            if(this.steps[i] <= 0.0){
                new IllegalArgumentError(SDBError.Fatal,"A value in the gap array is too large.");
            }
        }
	
	}


    //------------------------------------------------------------------------	
    /* this is a help function for constructor, which is to set the Geometric
       bounds and initial positon of the shape to traverse. */ 
    private void init(GBounds bounds) {
		
        this.ubounds = bounds.upper.clone();
        this.lbounds = bounds.lower.clone();
        
		int i, d=this.lower.length;
		float [] spaceDimensions = bounds.getDimensions(new float[d]);

		for(i=0; i<d; i++){
		
			this.start[i] = this.lbounds[i];
			this.pUBounds[i] = (int) Math.floor(spaceDimensions[i] / this.steps[i]) - 1;
		}
		
    }
	

     
   
    //------------------------------------------------------------------------	
    /* this is a help function for constructor, which is to set the Geometric
       bounds and initial positon of the shape to traverse. */ 
    private void init(GBounds bounds, Point start) {
		
        this.ubounds = bounds.upper.clone();
        this.lbounds = bounds.lower.clone();
        
		int i, d=this.lower.length;
        //float startCoord;
		float [] spaceDimensions = bounds.getDimensions(new float[d]);
		float [] startCoords=start.getCoords();
		
		for(i=0; i<d; i++){
		
						
			if (startCoords[i] >= this.lbounds[i]){
			
				this.start[i] =  this.lbounds[i] - (this.steps[i] - (startCoords[i] - this.lbounds[i]) % this.steps[i]);
				
			} else {
			
				this.start[i] =  this.lbounds[i] - ( this.lbounds[i] - startCoords[i]) % this.steps[i];
				
			}
			
			this.pUBounds[i] = (int) Math.ceil((spaceDimensions[i] + (this.lbounds[i] - this.start[i])) / this.steps[i]) - 1;
		}
		
		this.init();        
    }
  

    //------------------------------------------------------------------------
    /** Initialize this GBoundsGapIterator */
    public void init() { 
	
       int i, d=this.lower.length;
       
        for ( i = 0; i < d; i++) {
        
            this.lower[i] = this.start[i];
			this.upper[i] = this.start[i] + this.shape[i];
            
			this.lower[i] = ( this.lower[i] >= this.lbounds[i]) ? this.lower[i] : this.lbounds[i];
			this.upper[i] = ( this.upper[i] <= this.ubounds[i]) ? this.upper[i] : this.ubounds[i];
			
			this.partitionIndex[i]=0;
        }
		
        this.valid = true;

     }

    
    
    

    /** Return an IndexSpaceID representing the iterator's steps.*/
    public final Point getSteps(){

        return new Point(this.steps);
    }
    
    
    /** Return this iterator's internal steps array.*/
    final float [] getStepsArray(){

        return this.steps;
    }


    //------------------------------------------------------------------------
    /* this is a help function for next() function, which reset the axis 
       if the axis is outside its bound */
    private void resetAxis(int axisNumber){
    
		this.partitionIndex[axisNumber]=0;
	
		//this.pLower[axisNumber] = this.lbounds[axisNumber];
		//this.pUpper[axisNumber] = this.pLower[axisNumber] + this.shape[axisNumber];
    }
   


    //------------------------------------------------------------------------
    /* this is a help fuction for next() function, which is to set a 
       specified value to the specified axis if the value is in bounds */
    private boolean incrementAxis(int axisNumber){

	
		if(this.partitionIndex[axisNumber] == pUBounds[axisNumber]){
			
			return false;
			
		} else {
		
			this.partitionIndex[axisNumber]++;
			return true;
		}
	
		/*
		if( this.pUpper[axisNumber] >= this.ubounds[axisNumber]){

            return false;
			
        } else {
		
			this.pLower[axisNumber] = this.pUpper[axisNumber];
			this.pUpper[axisNumber] += steps[axisNumber];
			
			float distance = this.ubounds[axisNumber] - this.pUpper[axisNumber];
			
			if(distance < this.epsilon && distance > 0){
			
				this.pUpper[axisNumber] = this.ubounds[axisNumber];
			}
			
			return true;
		}
		*/
    }
   

    /* Copy the values of pLower and pUpper to the GBounds data members,
	   clipping the values to the iterator space. */
	private final void copyToGBounds(int lastChangedAxisIndex){
	
		int axis;
    
	//System.out.println("**** last changed axis="+lastChangedAxisIndex);
	
	
		for ( int i=0; i<=lastChangedAxisIndex; i++){
			axis = this.ordering.getAxis( this.lower.length - i - 1 );
			//axis=this.ordering.getNthLeastSignificantAxis(i);
			
	//oldupper = this.upper[axis];
	
			this.lower[axis] = this.start[axis] + this.partitionIndex[axis] * this.steps[axis];
			this.upper[axis] = this.lower[axis] + this.shape[axis];
	
//	if(oldupper != this.lower[axis]){
//	
//		System.out.println("**** gap is "+(this.lower[axis] - oldupper));
//	}
		
			this.lower[axis] = ( this.lower[axis] >= this.lbounds[axis]) ? this.lower[axis] : this.lbounds[axis];
			this.upper[axis] = ( this.upper[axis] <= this.ubounds[axis]) ? this.upper[axis] : this.ubounds[axis];
		}
	}
	
    //------------------------------------------------------------------------ 
    /** Return true if the current position is a valid position in the 
       IndexSpace. */
    public boolean valid(){
	
        return this.valid;
    }
  
 
    //------------------------------------------------------------------------ 
    /** Moving the GBoundsGapIterator to the next position.*/
    public boolean next() {
        int axis = 0;
        int d=this.lower.length;
        
        if (this.valid == false){
            return false;
        }
        
        int i = 0;
        for ( i = 0; i < d; i++) {
            axis = this.ordering.getAxis( d - i - 1 );
			//axis=this.ordering.getNthLeastSignificantAxis(i);
			
//if(axis != 	this.ordering.getNthLeastSignificantAxis(i)){
//
//	System.out.println("****** axes are different!");
//
//}		
            if (this.incrementAxis(axis)) {
                
				this.copyToGBounds(i);
                return true;
                
            } else {
				if (i == d-1) {

                    this.valid = false;

                    return false;
					
                } else {
           
                    this.resetAxis(axis);
                }
            }
        }
//System.out.println("*****End of Next:");
		this.valid=false;
		this.copyToGBounds(i-1);
        return false;
    }


    /** Not implemented.*/
    public void end(){
    
        new SDBError(SDBError.Fatal, "Not implemented");
    }


    //------------------------------------------------------------------------
    /** Return a simple GBounds corresponding to the current value of this iterator. */
    public GBounds copyBounds()
    {
        return new GBounds( lower.clone(), upper.clone());
    }


    //------------------------------------------------------------------------
    /** Return a copy of this GBoundsGapIterator */
    public Object clone() {
    
        int i, d=this.lower.length;
        
        GBounds space = new GBounds( this.lbounds.clone(), this.ubounds.clone() );

        GBoundsGapIterator cpy = new GBoundsGapIterator(space, this.shape, this.steps); 
        
        cpy.valid=this.valid;
        
        for(i=0; i < d; i++){
            cpy.lower[i] = this.lower[i];
            cpy.upper[i] = this.upper[i];
        }
         
        return cpy;
    }
    
        /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }
	
   
    //------------------------------------------------------------------------ 
    /** Return description of this GBoundsGapIterator */
    public String toString(String indent){
    
        String indent2=indent+Globals.toStringIndent;
        StringBuffer result=new StringBuffer(indent+"{ GBoundsGapIterator:\n");
        
        result.append(indent2+"valid="+this.valid+"\n");
        
        result.append(indent2+"Iterator value is: ");
        result.append(super.toString("")+"\n");
        
        
        result.append(indent2+"Iterator Bounds are:");
        result.append(new GBounds(new Point(lbounds), 
                                   new Point(ubounds))+"\n");
        result.append(indent2+"Iterator Shape is: ");
        result.append(new Point(this.shape)+"\n");
        result.append(indent2+"Iterator Steps are: ");
        result.append(new Point(this.steps)+"\n");
        
        
        result.append(indent+"}");
        
        return new String(result);
    }
               
    /** For debugging. Runs basic tests on the GBoundsGapIterator class. */
    static public void main(String[] args){ 
/*    
        System.out.println("*****Clone Test*****");
        GBounds space=new GBounds(new float[]{0,0}, new float[]{30,30});        
        GBoundsGapIterator iter1=new GBoundsGapIterator(space, new float[]{10, 10}, new float[]{0,-1} );
        GBoundsGapIterator iter2=(GBoundsGapIterator)iter1.clone();
        
        for(iter1.init(); iter1.valid(); iter1.next()){
            
            System.out.println(iter1);
        }
        
        System.out.println("\nAfter loop:");
        System.out.println("iter1="+iter1);
        System.out.println("iter2="+iter2);
        
        
        
        
       //System.exit(0);
        
        float[] lower = {0,0,0};
        float[] upper = {5,9,7};
 
        float[] shape, steps;
        shape = new float[3];
        steps = new float[3];
    
        shape[0] = 2;
        shape[1] = 2;
        shape[2] = 2;
        steps[0] = 1;
        steps[1] = 2;
        steps[2] = 3;
        System.out.println( "-------shape[0]: " + shape[0] );
        System.out.println( "-------shape[1]: " + shape[1] );
        System.out.println( "-------shape[3]: " + shape[2] );

        GBounds bound = new GBounds(lower, upper);                   
        GBoundsGapIterator iter = new GBoundsGapIterator(bound,shape,steps);
        GBoundsGapIterator cpy = (GBoundsGapIterator)iter.clone();

        System.out.println( "------------Iterator------------ \n"  );
        System.out.println( "---Iterator dim: " + iter.dim() );
        System.out.println( "---iterator getBounds: " + iter.copyBounds());
        System.out.println( "---iterator lower: " + iter.getLower().toString());
        System.out.println( "---iterator upper: " + iter.getUpper().toString());

        for(iter.init(); iter.valid(); iter.next()){
            System.out.println(iter.toString());
            System.out.println("lower: " + iter.getLower().toString());
            System.out.println("upper: " + iter.getUpper().toString());
        }
        
        System.out.println();
        System.out.println("*************");
        System.out.println("After Loop:");

        System.out.println( "---iterator getBounds: " + iter.copyBounds());
        System.out.println( "---iterator lower: " + iter.getLower().toString());
        System.out.println( "---iterator upper: " + iter.getUpper().toString());


//System.exit(1);


        System.out.println( "" );
        System.out.println( "--------------copy------------------ \n"  );
        System.out.println( "---copy dim: " + cpy.dim() );
        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );

        for(cpy.init(); cpy.valid(); cpy.next()){
            System.out.println(cpy.toString());
            System.out.println("lower: " + cpy.getLower().toString());
            System.out.println("upper: " + cpy.getUpper().toString());
        }

        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );

        System.out.println( "" );
        cpy = (GBoundsGapIterator)iter.clone();
        System.out.println( "--------------copy again---------------- \n"  );
        System.out.println( "---copy dim: " + cpy.dim() );
        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );
*/                
    }
                                
}




