/*
    GBoundsIterator.java
    Copyright Oct. 2002 University of New Hampshire
    Durham, NH 03824

    This program is to code the GBoundsIterator class, which traverses 
    the specified IndexSpace with the specified shape and the specified
    steps.

    use the while loop or for loop to traverse:

    GBoundsIterator iter = new GBoundsIterator(dsBounds, shape, steps);
    
    while (iter.valid()) {
        System.out.println(iter.toString());
        iter.next();
    }

    or 

    for(iter.init(); iter.valid(); iter.next()) {
        System.out.println(iter.toString());
    }

*/


package edu.unh.sdb.lattice;
//import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;


public class GBoundsIterator extends GBounds implements Cloneable, SDBIterator
{

    private AxisOrdering ordering;
	
    private boolean valid = false;
	
    private float[] shape;
    //private float[] steps;
	private float[] start;
	
    private float[] ubounds;
    private float[] lbounds;
	
	private int[] partitionIndex;
	private int[] pUBounds;
	
	//private float epsilon = 0.001f;
    //private float epsilon = 0.0001f;
	
    //------------------------------------------------------------------------    
    /**	Construct an GBoundsIterator that will traverse the given space with an
     *   an GBounds of the given shape (dimensions). 
     *   
     *   Values in the shape array must be >= 1.      
     */
    public GBoundsIterator(GBounds space, float shape[]) {

        super(space.dim());

        this.ordering = new AxisOrdering(this.dim());
		
		setupAndCheck(shape);    
		        
        this.init(space);
     }


    public GBoundsIterator(GBounds space, IndexSpaceID dimensions) {
		
        super(space.dim());
		
		int d=space.dim();
		
        this.ordering = new AxisOrdering(d);
		
		setupAndCheck(space,  dimensions.getCoords(new int[d]));    
		        
        // We don't call an init() because the above method takes care of that stuff.
     }


    public GBoundsIterator(GBounds space, int[] dimensions) {
		
        super(space.dim());
		
		int d=space.dim();
		
        this.ordering = new AxisOrdering(d);
		
		setupAndCheck(space, dimensions.clone());    
		        
        // We don't call an init() because the above method takes care of that stuff.
     }


    public GBoundsIterator(GBounds space, float shape[], AxisOrdering ordering) {

        super(space.dim());

        this.ordering = ordering.copy();
				
		setupAndCheck(shape);    
		        
        this.init(space);
     }




    public GBoundsIterator(GBounds space, Point start, float shape[]) {

        super(space.dim());

        this.ordering = new AxisOrdering(this.dim());
		
		setupAndCheck(shape);
		
        this.init(space, start);
     }

    public GBoundsIterator(GBounds space, Point start, float shape[], AxisOrdering ordering) {

        super(space.dim());

        this.ordering = ordering.copy();
				
		setupAndCheck(shape);
		
        this.init(space, start);
     }



	/* Allocates space for private arrays, and checks the shape array for improper values. */
	private final void setupAndCheck(float shape[]){

		int i, d=this.dim();
		
        this.shape = new float[d];        
		this.start = new float[d];
		
		this.partitionIndex = new int[d];      
		  
		this.pUBounds = new int[d];
		
        /* check shape and set the value of shape */
        for( i = 0; i < d; i++ ){
		
           if ( shape[i] <= 0.0 ) {
               new IllegalArgumentError(SDBError.Fatal,"Shape dimensions must be > 0.");
           }

           this.shape[i] = shape[i];
        }
      	
	}


	/* Allocates space for private arrays, and initializes the shape array based on the pUBounds array. */
	private final void setupAndCheck(GBounds bounds, int [] dimensions){

		int i, d=bounds.dim();
		
		this.pUBounds=dimensions;

        this.ubounds = bounds.upper.clone();
        this.lbounds = bounds.lower.clone();

		
        this.shape = new float[d];        
		this.start = new float[d];
		
		this.partitionIndex = new int[d];      
		  		
        
        for( i = 0; i < d; i++ ){
		
			this.shape[i] = (this.ubounds[i] - this.lbounds[i]) / dimensions[i];
			
			
			this.pUBounds[i]--; // Caution: pUBounds and dimensions are the same array.

			this.start[i] = this.lbounds[i];
			
			    
        }
		
      	this.init(); 
	}

    /* Set the iterator space bounds and initial position. */ 
    private void init(GBounds bounds) {
		
        this.ubounds = bounds.upper.clone();
        this.lbounds = bounds.lower.clone();
        
		int i, d=this.lower.length;
		float [] spaceDimensions = bounds.getDimensions(new float[d]);

		for(i=0; i<d; i++){
		
			this.start[i] = this.lbounds[i];
			this.pUBounds[i] = (int) Math.floor(spaceDimensions[i] / this.shape[i]) - 1;
		}
		
    }
	
    
     
   
	
    /* Set the iterator space bounds and initial position of the GBounds. */ 
    private void init(GBounds bounds, Point start) {
		
        this.ubounds = bounds.upper.clone();
        this.lbounds = bounds.lower.clone();
        
		int i, d=this.lower.length;
		float [] spaceDimensions = bounds.getDimensions(new float[d]);
		float [] startCoords=start.getCoords();
		
		for(i=0; i<d; i++){
		
						
			if (startCoords[i] > this.lbounds[i]){
			
				this.start[i] =  this.lbounds[i] - (this.shape[i] - (startCoords[i] - this.lbounds[i]) % this.shape[i]);
				
			} else {
			
				this.start[i] =  this.lbounds[i] - ( this.lbounds[i] - startCoords[i]) % this.shape[i];
				
			}
			
			this.pUBounds[i] = (int) Math.ceil((spaceDimensions[i] + (this.lbounds[i] - this.start[i])) / this.shape[i]) - 1;
            
		}
		
		this.init();        
    }

  
    /** Set space bounds, starting position, and shape of GBounds after construction. The 
      * iterator is reset to the first iteration value.
      */
    final void set(GBounds space, float [] startCoords, float [] shape){

        try{
        
            int i, d=space.dim();
            float spaceDimension;
            
            for( i = 0; i < d; i++ ){
            
                this.shape[i] = shape[i];
                this.lbounds[i] = space.lower[i];
                this.ubounds[i] = space.upper[i];
                
                spaceDimension = space.getDimension(i);
                
                
                if (startCoords[i] > this.lbounds[i]){
                
                    this.start[i] =  this.lbounds[i] - (this.shape[i] - (startCoords[i] - this.lbounds[i]) % this.shape[i]);
                    
                } else {
                
                    this.start[i] =  this.lbounds[i] - ( this.lbounds[i] - startCoords[i]) % this.shape[i];
                    
                }
                
                this.pUBounds[i] = (int) Math.ceil((spaceDimension + (this.lbounds[i] - this.start[i])) / this.shape[i]) - 1;
    
            }
            
            this.init(); 
        
        } catch(Exception e){
        
            int d = space.dim();
            
            this.lower = new float[d];
            this.upper = new float[d];
            
            this.ordering=new AxisOrdering(d);
            
            this.setupAndCheck(shape);
            this.init(space, new Point(startCoords));
            
            new IllegalArgumentError(SDBError.Recoverable, "One or more of the arguments had too large a dimensionality for this iterator.");
        } 
        
	}

    
    /** Initialize this GBoundsIterator so that its value is set to the first GBounds in the iteration. */
    public void init() { 
	
       int i, d=this.lower.length;
       
        for ( i = 0; i < d; i++) {
        
            this.lower[i] = this.start[i];
			this.upper[i] = this.start[i] + this.shape[i];
            
			this.lower[i] = ( this.lower[i] >= this.lbounds[i]) ? this.lower[i] : this.lbounds[i];
			this.upper[i] = ( this.upper[i] <= this.ubounds[i]) ? this.upper[i] : this.ubounds[i];
			
			this.partitionIndex[i]=0;
        }
		
        this.valid = true;

     }

    /** Return the AxisOrdering object for this iterator.*/
    public final AxisOrdering getAxisOrdering(){
    
        return this.ordering;
    }

    /** Return the length of this iterator's space along the specified axis. */
    public final float getAxisLength(int axis){
    
        try{
        
            return this.ubounds[axis]-this.lbounds[axis];
            
        } catch( ArrayIndexOutOfBoundsException e){
        
            new IllegalArgumentError(SDBError.Recoverable, "Argument specifies an illegal axis.");
            return -1;
        }
    }


    /** Returns an array of ints representing the dimensions of 
     *   the space being iterated over.
     */
    public final float[] getAxisLengthArray(){
    
        int i,d=this.lbounds.length;
        float [] axisLengths=new float[d];
        
        
        for(i=0; i<d; i++){
            
            axisLengths[i]=this.ubounds[i]-this.lbounds[i];
        }
        
        return axisLengths;
    }

    
    /** Return an IndexSpaceID representing the shape of the iterator's blocks.*/
    public final Point getBlockShape(){

        return new Point(this.shape);
    }
    
    
    /** Return this iterator's internal shape array, representing the shape
      *  of the iterator's blocks.*/
    final float [] getBlockShapeArray(){

        return this.shape;
    }

    

    //------------------------------------------------------------------------
    /* this is a help function for next() function, which reset the axis 
     *  if the axis is outside its bound */
    private void resetAxis(int axisNumber){
    
		this.partitionIndex[axisNumber]=0;
	
		//this.pLower[axisNumber] = this.lbounds[axisNumber];
		//this.pUpper[axisNumber] = this.pLower[axisNumber] + this.shape[axisNumber];
    }
   


    //------------------------------------------------------------------------
    /* this is a help fuction for next() function, which is to set a 
     *  specified value to the specified axis if the value is in bounds */
    private boolean incrementAxis(int axisNumber){

	
		if(this.partitionIndex[axisNumber] == pUBounds[axisNumber]){
			
			return false;
			
		} else {
		
			this.partitionIndex[axisNumber]++;
			return true;
		}
	
		/*
		if( this.pUpper[axisNumber] >= this.ubounds[axisNumber]){

            return false;
			
        } else {
		
			this.pLower[axisNumber] = this.pUpper[axisNumber];
			this.pUpper[axisNumber] += steps[axisNumber];
			
			float distance = this.ubounds[axisNumber] - this.pUpper[axisNumber];
			
			if(distance < this.epsilon && distance > 0){
			
				this.pUpper[axisNumber] = this.ubounds[axisNumber];
			}
			
			return true;
		}
		*/
    }
   

    /* Copy the values of pLower and pUpper to the GBounds data members,
	 *  clipping the values to the iterator space. */
	private final void copyToGBounds(int lastChangedAxisIndex){
	
		int i, d=this.lower.length;
		int axis;
    
	
	
		for(i=0; i<=lastChangedAxisIndex; i++){
			axis = this.ordering.getAxis( d - i - 1 );
			//axis=this.ordering.getNthLeastSignificantAxis(i);
			
			
			if(this.partitionIndex[axis] != 0){
			
				//this.lower[axis] = this.start[axis] + this.partitionIndex[axis] * this.steps[axis];
				this.lower[axis]=this.upper[axis];
				this.upper[axis] = this.start[axis] + (this.partitionIndex[axis] + 1) * this.shape[axis];
				
				this.upper[axis] = ( this.partitionIndex[axis] != this.pUBounds[axis]) ? this.upper[axis] : this.ubounds[axis];
			
			} else {
				this.lower[axis]=this.lbounds[axis];
				this.upper[axis] = this.start[axis] + this.shape[axis];
			}
	
	
			
			
			
		}
	}
	
    //------------------------------------------------------------------------ 
    /** Return true if the current position is a valid position in the 
      * IndexSpace. */
    public boolean valid(){
	
        return this.valid;
    }
  
 
    //------------------------------------------------------------------------ 
    /** Moving the GBoundsIterator to the next position.*/
    public boolean next() {
        int axis = 0;
        int d=this.lower.length;
        
        if (this.valid == false){
            return false;
        }
        
        int i;
        for ( i = 0; i < d; i++) {
            axis = this.ordering.getAxis( d - i - 1 );
			//axis=this.ordering.getNthLeastSignificantAxis(i);
			
if(axis != 	this.ordering.getNthLeastSignificantAxis(i)){

	System.out.println("****** axes are different!");

}		
            if (this.incrementAxis(axis)) {
                
				this.copyToGBounds(i);
                return true;
                
            } else {
				if (i == d-1) {

                    this.valid = false;

                    return false;
					
                } else {
           
                    this.resetAxis(axis);
                }
            }
        }
System.out.println("*****End of Next:");
		this.valid=false;
		this.copyToGBounds(i-1);
        return false;
    }


    /** Not implemented.*/
    public void end(){
    
        new SDBError(SDBError.Fatal, "Not implemented");
    }

    
    //------------------------------------------------------------------------
    /** Return a simple GBounds corresponding to the current value of this iterator. */
    public GBounds copyBounds()
    {
        return new GBounds( lower.clone(), upper.clone());
    }


    //------------------------------------------------------------------------
    /** Return a copy of this GBoundsIterator */
    public Object clone() {
    
        int i, d=this.lower.length;
        
		
        GBounds space = new GBounds( this.lbounds.clone(), this.ubounds.clone() );

        GBoundsIterator cpy = new GBoundsIterator(space, new Point(this.start),this.shape, this.ordering); 
        
        cpy.valid=this.valid;
        
        for(i=0; i < d; i++){
            cpy.lower[i] = this.lower[i];
            cpy.upper[i] = this.upper[i];
			cpy.partitionIndex[i] = this.partitionIndex[i];
			cpy.pUBounds[i] = this.pUBounds[i];
        }
         
        return cpy;
    }
    
        /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }
	
   
    //------------------------------------------------------------------------ 
    /** Return description of this GBoundsIterator */
    public String toString(String indent){
    
        String indent2=indent+Globals.toStringIndent;
        StringBuffer result=new StringBuffer(indent+"{ GBoundsIterator:\n");
        
        result.append(indent2+"valid="+this.valid+"\n");
        
        result.append(indent2+"Iterator value is: ");
        result.append(super.toString("")+"\n");
        
        
        result.append(indent2+"Iterator Bounds are:");
        result.append(new GBounds(new Point(lbounds), 
                                   new Point(ubounds))+"\n");
        result.append(indent2+"Iterator Shape is: ");
        result.append(new Point(this.shape)+"\n");
        
        
        result.append(indent+"}");
        
        return new String(result);
    }
               
    /** For debugging. Runs basic tests on the GBoundsIterator class. */
    static public void main(String[] args){ 
/*    
        System.out.println("*****Clone Test*****");
        GBounds space=new GBounds(new float[]{0,0}, new float[]{30,30});        
        GBoundsIterator iter1=new GBoundsIterator(space, new float[]{10, 10}, new float[]{0,-1} );
        GBoundsIterator iter2=(GBoundsIterator)iter1.clone();
        
        for(iter1.init(); iter1.valid(); iter1.next()){
            
            System.out.println(iter1);
        }
        
        System.out.println("\nAfter loop:");
        System.out.println("iter1="+iter1);
        System.out.println("iter2="+iter2);
        
        
        
        
       //System.exit(0);
        
        float[] lower = {0,0,0};
        float[] upper = {5,9,7};
 
        float[] shape, steps;
        shape = new float[3];
        steps = new float[3];
    
        shape[0] = 2;
        shape[1] = 2;
        shape[2] = 2;
        steps[0] = 1;
        steps[1] = 2;
        steps[2] = 3;
        System.out.println( "-------shape[0]: " + shape[0] );
        System.out.println( "-------shape[1]: " + shape[1] );
        System.out.println( "-------shape[3]: " + shape[2] );

        GBounds bound = new GBounds(lower, upper);                   
        GBoundsIterator iter = new GBoundsIterator(bound,shape,steps);
        GBoundsIterator cpy = (GBoundsIterator)iter.clone();

        System.out.println( "------------Iterator------------ \n"  );
        System.out.println( "---Iterator dim: " + iter.dim() );
        System.out.println( "---iterator getBounds: " + iter.copyBounds());
        System.out.println( "---iterator lower: " + iter.getLower().toString());
        System.out.println( "---iterator upper: " + iter.getUpper().toString());

        for(iter.init(); iter.valid(); iter.next()){
            System.out.println(iter.toString());
            System.out.println("lower: " + iter.getLower().toString());
            System.out.println("upper: " + iter.getUpper().toString());
        }
        
        System.out.println();
        System.out.println("*************");
        System.out.println("After Loop:");

        System.out.println( "---iterator getBounds: " + iter.copyBounds());
        System.out.println( "---iterator lower: " + iter.getLower().toString());
        System.out.println( "---iterator upper: " + iter.getUpper().toString());


//System.exit(1);


        System.out.println( "" );
        System.out.println( "--------------copy------------------ \n"  );
        System.out.println( "---copy dim: " + cpy.dim() );
        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );

        for(cpy.init(); cpy.valid(); cpy.next()){
            System.out.println(cpy.toString());
            System.out.println("lower: " + cpy.getLower().toString());
            System.out.println("upper: " + cpy.getUpper().toString());
        }

        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );

        System.out.println( "" );
        cpy = (GBoundsIterator)iter.clone();
        System.out.println( "--------------copy again---------------- \n"  );
        System.out.println( "---copy dim: " + cpy.dim() );
        System.out.println( "---copy getBounds: " + cpy.copyBounds() );
        System.out.println( "---copy lower: " + cpy.getLower().toString() );
        System.out.println( "---copy upper: " + cpy.getUpper().toString() );
*/                
		GBounds b1=new GBounds(new Point(0,0), new Point(8,8));
		//float [] pDimensions=new float[]{1.14286f, 1.14286f};
		//float [] steps=new float[]{0f,0f};
		
		//GBoundsIterator gIter1=new GBoundsIterator(b1, pDimensions, steps); 
		GBoundsIterator gIter1=new GBoundsIterator(b1, new int[]{7,7});
		
		float missedArea=0;
		float lastX=0, lastY=0; 
		float x,y, height, width=b1.getUpper().getCoord(1)-b1.getLower().getCoord(1);
		int count=0;
		
        for(gIter1.init(); gIter1.valid(); gIter1.next()){
			
			x=gIter1.getLower().getCoord(1);
			y=gIter1.getLower().getCoord(0);

			height=gIter1.getUpper().getCoord(0)-gIter1.getLower().getCoord(0);
			
			if(lastX < x){
				missedArea += ( x - lastX ) * height;
			}

			if(lastY < y){
				missedArea += ( y - lastY ) * width;
			}
			
			lastX=gIter1.getUpper().getCoord(1);
			lastY=gIter1.getUpper().getCoord(0);
			
			System.out.println(gIter1);
			
			count++;
		}
        
		System.out.println("missed area="+missedArea);
		System.out.println("count="+count);
    }
                                
}




