//
//  Lattice.java
//  Unstructured
//
//  Created by Philip Rhodes on Wed Sep 25 2002.
//  Copyright (c) 2002 University of New Hampshire. All rights reserved.
//
// --------------- Edit history since 01/10/05 -----------------------------
//
// 01/10/05 rdb: removed direct allocation of children of Datum

package edu.unh.sdb.lattice; 

//import java.io.*;
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;




public class Lattice {

    protected Topology topology;
    protected Geometry geometry;
    protected Approximator approximator;
    protected String name;
	protected boolean isCellLattice;
	protected Datum nullDatum;
    
    public Lattice(String name, Topology t, Geometry g, Approximator a){
    
        this.topology=t;
        this.geometry=g;
        this.approximator=a;
        this.name=name;
		this.isCellLattice=g.isCellGeometry();
		
		this.nullDatum=t.getRecordDescriptor().createDatum();
    }
    
    
    Lattice(String name){
    
        this.topology=null;
        this.geometry=null;
        this.approximator=null;
        this.name=name;
    }

    public void activate(){
        
        this.topology.activate();
		
    }



    public Datum datum(Point p){
 
        IndexSpaceID partitionID = this.geometry.map(p);
		
		if(this.isCellLattice){
		
			Cell c = topology.findEnclosingCell(partitionID, p);
			
			if(c != null){
				return approximator.approximate(c,p);
			} else {
				return this.nullDatum;
			}
		
		} else {
		
			return this.topology.datum(partitionID);
		
		}
    }


	/** Set the datum returned when no data is available for a 
		point in the domain. The argument is not copied.*/
	public void setDefaultDatum(Datum d){
	
		this.nullDatum=d;
	}

    /** Return the lattice name. */
    public String getName(){
    
        return name;
    }
    
	/** Return the Bounds of the Geometry.*/
	public GBounds getGBounds(){
	
		return this.geometry.getBounds();
	}
	
	/** Return the Bounds of the Topology.*/
	public ISBounds getTBounds(){
	
	
		return this.topology.getBounds();
	
	}


	/** Return a RecordDescriptor for datums returned by this lattice.*/														
	public RecordDescriptor getRecordDescriptor(){
	
		return this.topology.getRecordDescriptor();
	}
	
	/** Return the number of attributes for this lattice.*/
	public int getNumAttributes(){
	
		return this.topology.getRecordDescriptor().size();
	}
	
	
	Topology getTopology(){
	
		return this.topology;
	}
	
	/** Return the dimensionality of the lattice geometry.*/
	int getGeometryDim(){
	
		return this.geometry.dim();
	}
	
	/** Return the dimensionality of the lattice topology.*/
	int getTopologyDim(){
	
		return this.topology.dim();
	}
	
	
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){

        StringBuffer result=new StringBuffer(indent+"{ Lattice: name="+this.name+"\n");
        String indent2=indent+Globals.toStringIndent;
        
        result.append(indent2+"Topology=\n"+this.topology.toString(indent2)+"\n");
        result.append(indent2+"Geometry=\n"+this.geometry.toString(indent2)+"\n");
        result.append(indent2+"Approximator=\n"+this.approximator.toString(indent2)+"\n");
        
        result.append("}");
        
        return new String(result);
    }

}
