/* 
	Point.java
	Copyright 2000 University of New Hampshire
	Durham, NH 03824
	
	----------------- Edit history as of 01/10/05 -------------------------
	05/28/05 rdb: removed deprecation on Point ( float[], int, int )	
	01/10/05 rdb: 
	              Deprecated
	                   assign( float[], int start, int copyCount )
	              Changed the semantics of 
	                   Point(  float[], int start, int copyCount )
	               and implemented
	                   setCoords( float[], int start, int copyCount )
	              so that "start" is now a direct index into the array, rather
	              than a number of an increment of the "copyCount". This makes
	              it consistent with the new semantics of similar methods in the
	              common package.
*/

package edu.unh.sdb.lattice; 

import edu.unh.sdb.common.*;

/**
 * Represent a location in the geometry
 *   
 */

//public abstract class Point extends GeomEntity{
public class Point extends SpaceID{
    
    protected float axes[];
    
    /** default constructor. Does nothing */
    Point(){}
    
    /** Create a point with n dimensions.*/
    public Point(int n){
    
        this.axes=new float[n];
    }
    
    /** Create a one dimensional point with the given axis value.*/
    public Point(double x){
    
        this.axes=new float[1];
        this.axes[0]=(float) x;
    }
    
    /** Create a two dimensional point with the given axis values.*/
    public Point(double x,double y){
    
        this.axes=new float[2];
        
        this.axes[0]=(float) x;
        this.axes[1]=(float) y;
    }
    
    /** Create a three dimensional point with the given axis values.*/
    public Point(double x,double y,double z){
    
        this.axes=new float[3];
        
        this.axes[0]=(float) x;
        this.axes[1]=(float) y;
        this.axes[2]=(float) z;
    }
    
    /** Create point with the axis values given in a Datum*/
    public Point(Datum datum){
        
        int i,d=datum.size();
        
        this.axes=new float[d];
        
        for(i=0;i<d;i++){
            this.axes[i]=datum.getFloat(i);
        }
    }
    
    /** Create point with the coord values given in an array of floats.*/
    public Point(float [] coords){
    
        this.axes= coords.clone();
	}

	/** Create a point with the coord values interleaved in an array of floats.
		The coordinates are taken from the array beginning at the position
		specified by the 2nd argument. numCoords values will be copied.
	*/
	public Point( float [] coords, int startIndex, int numCoords ){
	
      this.axes = new float[ numCoords ];
      //rdb int startIndex=pointIndex*numCoords;
		
       try {
          for (int i=0; i<numCoords; i++){
             this.axes[i] = coords[startIndex + i];
          }
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Point constructor: insufficient data." );
       }
	}

    /** Create a copy of the Point parameter.*/
    public Point(Point p){

        this.axes= p.axes.clone();
    }

	/** Returns a new datum that contains the point axes as values. */
	public Datum makeDatum(){

		return new FloatDatum(this.axes);
	}

     /** Get the value corresponding to the specified axis. */
    public final int getIntAxis(int axisNumber){
        
        return (int) getCoord(axisNumber);
    }
        
    /** Get the value corresponding to the specified axis. */
    public final float getFloatAxis(int axisNumber){
    
        return axes[axisNumber];
    }
    
    
    /** Get the value corresponding to the specified axis. */
    public float getCoord(int axisNumber){
    
        return axes[axisNumber];
    }

    /** Copy the coordinates into the argument array. A reference
		to the argument is returned. */
    public float[] getCoords(float [] axes){
    
		int i, d=axes.length;
		
		try {
		
			for(i=0; i<d; i++){
			
				axes[i]=this.axes[i];
			}
		} catch (ArrayIndexOutOfBoundsException e){
		
			new IllegalArgumentError(SDBError.Recoverable, "Argument array is too short.");
		}
        return axes;
	}
    
	/** Returns a reference to the axes array. Package access only.*/
	float [] getCoords(){
	
		return axes;
	}
    
    /** Set the specified axis to the specified value. */
    public void setCoord(int axisnumber,float value){
    
        axes[axisnumber]=  value;
    }
    
    /** Set the specified axis to the specified value. */
    public final void setCoord(int axisnumber,int value){
    
        axes[axisnumber]= value;
    }
    
    /** Return a copy of this point. */
    public Object clone(){
    
        Point n=new Point();
        
        n.axes=new float[this.axes.length];
        System.arraycopy(this.axes,0,n.axes,0,this.axes.length);
        
        return n;
    }
    
    /** Returns the dimensionality of this Point. */
    public final int dim(){
        
        return axes.length;
    }
    
    /**************************************/
    
    /**Set the axes of this id to the axes of the parameter id.
        The assign method may elect to refer to the data in
        the parameter id, and not make a copy of this information.*/
    public void assign( Point id ){
    
        int i,d=id.dim();
        
        for(i=0;i<d;i++){
            this.setCoord(i,id.getCoord(i));
        }
    }
    
 	/** Assign new coordinate values to an existing Point of the same dimensionality.
		The coord values are assumed to be interleaved in an array of floats.
		The start position is computed by multiplying 2nd arg by the
		3rd argument.The 3rd arg specifies how many to copy.
		If this object is of too small a dimensionality to hold the new coordinates, 
		an error is created. 
		
		@deprecated
		Use setCoords( float[], int startIndex, int numCoords ).
		Note that the semantics of the start position is different in setCoords.
	*/
	public void assign( float [] coords, int pointIndex, int numCoords )
	{
	   setCoords( coords, pointIndex*numCoords, numCoords );
	}
    
 	/** Assign new coordinate values to an existing Point of the same dimensionality.
	 * The coord values are assumed to be interleaved in an array of floats.
	 * The coordinates are taken from the array beginning at the position
	 * specified by the 2nd argument; the 3rd arg specifies how many to copy.
	 * If this object is of too small a dimensionality to hold the new coordinates, 
	 * an error is created. 
	 */
	public void setCoords( float [] coords, int startIndex, int numCoords )
	{
		try 
		{
			for(int i=0; i<numCoords; i++)
			{
				this.axes[i]=coords[startIndex + i];
			}
		} catch (ArrayIndexOutOfBoundsException e){
			new DimensionalityError(SDBError.Recoverable, 
			      "Tried to assign to a Point of different dimensionality.");
		}
	}
    
    /** For each axis of this Point, add the value
        of the corresponding axis of the parameter. 
    
            If the dimensionality of the parameter does not
        match the dimensionality of this object, then
        a DimensionalityError is created.
    */
    public void add(Point dsd){
    
        int i,d=dsd.dim();
    
        
        if(d != this.dim()) {
            new DimensionalityError(SDBError.Recoverable,"Can't add IDs of different dimensionality.");
        }
        
        for(i=0;i<d;i++){
            this.setCoord(i,this.getCoord(i)+dsd.getCoord(i)); 
        }
    }
    
    /** Arithmetically negate this Point's axis values. */
    public void negate(){
    
        int i,d=this.dim();
                                                                
        for(i=0;i<d;i++){
            this.setCoord(i,-this.getCoord(i));    
        }
    }
    
    /** For each axis of this Point, subtract the value
        of the corresponding axis of the parameter. 
        
        If the dimensionality of the parameter does not
        match the dimensionality of this object, then
        a DimensionalityError is created.
    */
    public void subtract(Point dsd){
    
        int i,d=dsd.dim();
        
        
        if(d != this.dim()) {
            new DimensionalityError(SDBError.Recoverable,"Can't subtract IDs of different dimensionality.");
        }
        
        for(i=0;i<d;i++){
            this.setCoord(i,this.getCoord(i)-dsd.getCoord(i)); 
        }
        
    }
    
    /** Returns true if this Point is contained by the 
        ISBounds parameter, or false otherwise. 
        
        If the ISBounds has a different dimensionality than
        this Point, then a DimensionalityError is 
        created.
    */
    public boolean containedBy(GBounds db){
    
        Point id0=db.getLower();
        Point id1=db.getUpper();
        
        return id0.lessThanOrEqual(this) && id1.greaterThanOrEqual(this);
    }
    
    /** Return true if parameter is equal to this object, false otherwise.
    
        If the argument has a different dimensionality than
        this Point, then a DimensionalityError is 
        created.*/
    public boolean equal(Point id){
    
        int i, d=id.dim();
    
        
        if(d != this.dim()) {
            new DimensionalityError(SDBError.Recoverable,"Can't compare IDs of different dimensionality.");
        }
        
        for(i=0;i<d;i++){
            if(this.getCoord(i) != id.getCoord(i)){
                return false;
            }
        }
    
        return true;
    }
    
    
    /** A helper function that returns true if all the coordinates of this
        id are less than or equal to the corresponding coordinates of the parameter id. If any
        coordinate is greater than the parameter coordinate, then false is
        returned. If the dimensionality of the parameter does not match the dimensionality
        of this object, a DimensionalityError is created.
    */
    public boolean lessThanOrEqual(Point dsd){
    
        int i, d=dsd.dim();
    
        
        if(d != this.dim()) {
            new DimensionalityError(SDBError.Recoverable,"Can't compare IDs of different dimensionality.");
        }
    
        for(i=0;i<d;i++){
            if(this.getCoord(i) > dsd.getCoord(i)){
                return false;
            }
        }
        
        return true;
    }
    
    /** A helper function that returns true if all the coordinates of this
        id are greater than or equal to the corresponding coordinates of the parameter id. If any
        coordinate is less than the parameter coordinate, then false is
        returned. If the dimensionality of the parameter does not match the dimensionality
        of this object, a DimensionalityError is created.
    */
    public boolean greaterThanOrEqual(Point dsd){
        int i, d=dsd.dim();
    
        if(d != this.dim()) {
            new DimensionalityError(SDBError.Recoverable,"Can't compare IDs of different dimensionality.");
        }
    
        for(i=0;i<d;i++){
            if(this.getCoord(i) < dsd.getCoord(i)){
                return false;
            }
        }
        return true;
    }
    
    /**************************************/
    /** return true if this == p */
    boolean equals(Point p){
    
        return false;
    }
    
    /** Compute the euclidean distance between this point and the argument.
        *  The computation is done within the dimensionality of this point.
        */
    public float distance(Point p){
        int i,d=this.axes.length;
        float dist,distance=0;
        
        for(i=0;i<d;i++){
            dist=(this.axes[i]-p.axes[i]);
            distance+= dist * dist;
        }
        
        return (float) Math.sqrt(distance);
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i,d=this.dim();
        String result=new String(indent+"<Point (");
        
        for(i=0;i<d-1;i++){
            result+=this.getCoord(i)+",";
        }
        
        result+=this.getCoord(i)+")>";
        
        return result;
    }    
    
    /** Print the point coordinates */
    public void print(){
    
        System.out.println(this);
    }
}