//
//  TriangleCell.java
//  Unstructured 
//
//  Created by Philip Rhodes on Wed Jul 17 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
//

/** ----------------- Edit history since 01/10/05 ------------------------------
 *
 * 01/10/05 rdb: Edited Datum.assign and Point constructor and assign
 *               calls to reflect changed semantics.
 *
 * ----------------------------------------------------------------------------
 */

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

public class TriangleCell extends UnstructuredCell{
   
    /** For debugging. */
    public TriangleCell(Point p1, Point p2, Point p3){
    
        super(new Point [] {p1,p2,p3}, null);
    }

	protected TriangleCell(){
	
		super(null, null);
	}
	
    public TriangleCell(int numAttributes, int dimensionality){
	
		super(numAttributes, dimensionality, 3);
	}
 
    public TriangleCell(Point [] coords, Datum [] data){
    
        super(coords, data);
    
        if(coords.length!=3 || data.length!=3){
            new IllegalArgumentError(SDBError.Fatal,"TriangleCell constructor was given a vertices or data array with length not equal to three.");
        }
    
    }


	public TriangleCell(	float [] coordFloats, float [] dataFloats,  int numDataAttributes, 
							int dimensionality, int triangleIndex){
	
		super( numDataAttributes, dimensionality, 3 );
		
		Point [] coords=new Point[3];
		Datum [] data=new FloatDatum[3];
		
		int index;
		for ( int i=0; i < 3; i++){
			
		    index = triangleIndex + i;
		    //rdb next 2 lines edited to reflect new semantics of the constructors
		    //    note that this would be more efficient to have 2 separate index
		    //    variables, do the two * operations outside the loop and increment
		    //    the 2 variables by dimensionality and numDataAttributes resp.
		    //    For now, I wanted the simplest mapping until it is sure that we 
		    //    will keep this semantics change.
			//rdb coords[i] = new Point(coordFloats, index, dimensionality);
			//rdb [i] = new FloatDatum(dataFloats, index, numDataAttributes);
			coords[i] = new Point(coordFloats, index*dimensionality, dimensionality);
			data[i] = new FloatDatum(dataFloats, index*numDataAttributes, numDataAttributes);
		}
		
		this.setVertexCoords(coords);
		this.setData(data);
	}
    
	public TriangleCell( int [] vIndices, float [] coordFloats, float [] dataFloats, 
	                     int numDataAttributes, int dimensionality, int triangleIndex){

		super( numDataAttributes, dimensionality, 3 );
		
		Point [] coords=new Point[3];
		Datum [] data=new FloatDatum[3];
		int      index;
		
		for ( int i=0; i < 3; i++){
			
			index = vIndices[ triangleIndex*3+i ];	
			
			//rdb next 2 lines modified to account for new semantics of these constr.
			//rdb coords[i] = new Point(coordFloats, index, dimensionality);
			//rdb data[i]   = new FloatDatum(dataFloats, index, numDataAttributes);
			coords[i] = new Point(coordFloats, index*dimensionality, dimensionality);
			data[i]   = new FloatDatum(dataFloats, index*numDataAttributes, numDataAttributes);
		}
		this.setVertexCoords(coords);
		this.setData(data);
	}

	public TriangleCell(		DataSource vIndicesDS, 
								DataSource coordsDS, 
								RecordSpec vIndicesRS, 
								RecordSpec coordsRS,
								RecordSpec dataRS, 
								int dimensionality,
								IndexSpaceID cellID){   
	
		
		super(vIndicesDS, coordsDS, vIndicesRS, coordsRS, dataRS, dimensionality, 3, cellID);
	}

	public void assign(int [] vIndices, float [] coordFloats, float [] dataFloats, 
	                   int numDataAttributes, int dimensionality, int triangleIndex){
		
		int i, index;
		
		for(i=0; i < 3; i++){
			
			index = vIndices[triangleIndex*3+i];

			//rdb: 2 lines changed to reflect new semantics of constructors.
			//rdb this.coords[i].assign( coordFloats, index, dimensionality );
			//rdb ((FloatDatum)this.data[i]).assign( dataFloats, index, numDataAttributes );
			
			this.coords[i].setCoords( coordFloats, index*dimensionality, dimensionality );
			this.data[i].setFloats( dataFloats, index*numDataAttributes, numDataAttributes );
		}
	}
    
    public boolean intersects(Cell c){
    
        if(c instanceof TriangleCell){
        
            new IllegalArgumentError(SDBError.Fatal,"Triangle-Triangle Intersection not implemented.");
            return false;
        
        } else {
            new IllegalArgumentError(SDBError.Fatal,"Generalized Intersection not implemented.");
            return false;
        }
    }





    /** Returns true if the Point p is inside the TriangleCell, false otherwise. 
        Assumes 2D space for now.*/

    public boolean inside(Point p){
        
        
        
        float x=p.getCoord(0);
        float y=p.getCoord(1);
        
        float t; 	        
        float yval;         
        
        float vax,vay,vbx,vby;
        boolean intersectsUpward=false;
        boolean intersectsDownward=false;
        
        
        int i=0;
        int edge=0;
        
        while(i<3 && !intersectsUpward){
        
            vax=this.coords[edge].getCoord(0);
            vay=this.coords[edge].getCoord(1);
            
            vbx=this.coords[(edge+1) % 3].getCoord(0);
            vby=this.coords[(edge+1) % 3].getCoord(1);
            
            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsUpward=(y<yval);
                }
            }
            
            i++;
            edge++;
        }
        
        i=0;
        edge=0;
        
        while(i<3 && !intersectsDownward){
        
            vax=this.coords[edge].getCoord(0);
            vay=this.coords[edge].getCoord(1);
            
            vbx=this.coords[(edge+1) % 3].getCoord(0);
            vby=this.coords[(edge+1) % 3].getCoord(1);
            
            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsDownward=(y>yval);
                }
            }
            
            i++;
            edge++;
        }
    
        return intersectsUpward && intersectsDownward;
    
    }


    /** Returns true if the Point p is inside the TriangleCell, false otherwise. 
		This static method expects the points to be stored in an interleaved array
		of floats. P0x, P0y, P1x, P1y, P2x, P2y... the last argument specifies
		the index of the TriangleCell's zero'th vertex. 
        Assumes 2D space for now.*/

    public static boolean inside(Point p, float[] coords, int startIndex){
        
        
        
        float x=p.getCoord(0);
        float y=p.getCoord(1);
        
        float t; 	        
        float yval;         
        
        float vax,vay,vbx,vby;
        boolean intersectsUpward=false;
        boolean intersectsDownward=false;
        
        
        int i=0;
        int edge=0;
        
        while(i<3 && !intersectsUpward){

            vax=coords[startIndex + edge*2];
            vay=coords[startIndex + edge*2+1];
            
            vbx=coords[startIndex + ((edge+1) % 3)*2];
            vby=coords[startIndex + ((edge+1) % 3)*2+1];
			
            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsUpward=(y<yval);
                }
            }
            
            i++;
            edge++;
        }
        
        i=0;
        edge=0;
        
        while(i<3 && !intersectsDownward){

            vax=coords[startIndex + edge*2];
            vay=coords[startIndex + edge*2+1];
            
            vbx=coords[startIndex + ((edge+1) % 3)*2];
            vby=coords[startIndex + ((edge+1) % 3)*2+1];

            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsDownward=(y>yval);
                }
            }
            
            i++;
            edge++;
        }
    
        return intersectsUpward && intersectsDownward;
    }


    /** Returns true if the Point p is inside the TriangleCell, false otherwise. 
		This static method expects the point coords to be stored in an interleaved array
		of floats. P0x, P0y, P1x, P1y, P2x, P2y... the third argument is an array of
		integers that specifies the location of coordinate pairs in the coords array for
		a collection of cells. The last argument specifies the location of a particular
		cell's vertex indices in the index array.
        Assumes 2D space for now.*/

    public static boolean inside(Point p, int [] vIndices, float[] coords,  int triangleIndex){
        
        
        
        float x=p.getCoord(0);
        float y=p.getCoord(1);
        
        float t; 	        
        float yval;         
        
        float vax,vay,vbx,vby;
        boolean intersectsUpward=false;
        boolean intersectsDownward=false;
        
        
        int i=0;
        int edge=0;
        int startIndex=triangleIndex*3;
		
        while(i<3 && !intersectsUpward){

            vax=coords[vIndices[startIndex + edge]*2];
            vay=coords[vIndices[startIndex + edge]*2+1];
            
            vbx=coords[vIndices[startIndex + (edge+1) % 3]*2];
            vby=coords[vIndices[startIndex + (edge+1) % 3]*2+1];
			
            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsUpward=(y<yval);
                }
            }
            
            i++;
            edge++;
        }
        
        i=0;
        edge=0;
        
        while(i<3 && !intersectsDownward){

            vax=coords[vIndices[startIndex + edge]*2];
            vay=coords[vIndices[startIndex + edge]*2+1];
            
            vbx=coords[vIndices[startIndex + (edge+1) % 3]*2];
            vby=coords[vIndices[startIndex + (edge+1) % 3]*2+1];

            //Check for equal x values
            
            if(vax==vbx){
                if(x==vax && ((y>=vay && y<=vby) || (y>=vby && y<=vay))){
                    return true;
                }
            } else {
                
                t=(x-vax)/(vbx-vax);
                
                if(t>=0 && t<=1){
                    yval=vay+t*(vby-vay);
                        
                    if(yval==y){
                        return true;
                    }
        
                    intersectsDownward=(y>yval);
                }
            }
            
            i++;
            edge++;
        }
    
        return intersectsUpward && intersectsDownward;
    }
    
    
    public GBounds boundingBox(){
    
        int i,j,d=this.dimensionality;
        float[] min=new float[d];
        float[] max=new float[d];
        
        float[][] vertexArray=new float[3][];
        
        vertexArray[0]=coords[0].getCoords();
        vertexArray[1]=coords[1].getCoords();
        vertexArray[2]=coords[2].getCoords();


        // for each dimension
        for(i=0;i<d;i++){
        
                    
            // find the max and min...
            
            // assume 0th vertex is both max and
            // min value for the current dimension
            min[i]=vertexArray[0][i];
            max[i]=vertexArray[0][i];
            
            // Now see if the other vertices
            // have smaller or bigger values
            for(j=1;j<3;j++){
			
                if (vertexArray[j][i] < min[i]){
                    min[i]=vertexArray[j][i];
                }

                if (vertexArray[j][i] > max[i]){
                    max[i]=vertexArray[j][i];
                }            
            }
        }
        
        return new GBounds(min, max);
    }
	
	
	public static GBounds boundingBox(int[] vIndices, float[] coords, int dimensionality, int triangleIndex){
    
        int i,j;
        float[] min=new float[dimensionality];
        float[] max=new float[dimensionality];
        
        float[][] vertexArray=new float[3][dimensionality];
        
		int arrayIndex=triangleIndex*3;
		
		for(i=0; i<dimensionality; i++){
		
			vertexArray[0][i] = coords[ vIndices[arrayIndex    ] * dimensionality + i];
			vertexArray[1][i] = coords[ vIndices[arrayIndex + 1] * dimensionality + i];
			vertexArray[2][i] = coords[ vIndices[arrayIndex + 2] * dimensionality + i];
		}
		

        // for each dimension
        for(i=0;i<dimensionality;i++){
        
                    
            // find the max and min...
            
            // assume 0th vertex is both max and
            // min value for the current dimension
            min[i]=vertexArray[0][i];
            max[i]=vertexArray[0][i];
            
            // Now see if the other vertices
            // have smaller or bigger values
            for(j=1;j<3;j++){
			
                if (vertexArray[j][i] < min[i]){
                    min[i]=vertexArray[j][i];
                }

                if (vertexArray[j][i] > max[i]){
                    max[i]=vertexArray[j][i];
                }            
            }
        }
        
        return new GBounds(min, max);
	}
	
	
	/** Returns true if this TriangleCell is contained by the GBounds argument, false otherwise.*/
	public boolean containedBy(GBounds bounds){

        int i,j,d=this.coords[0].dim();
        
        float[][] vertexArray=new float[3][];
		
		float[] lower=bounds.getLowerArray();
		float[]	upper=bounds.getUpperArray();
        
        vertexArray[0]=coords[0].getCoords();
        vertexArray[1]=coords[1].getCoords();
        vertexArray[2]=coords[2].getCoords();


        // for each dimension
        for(i=0;i<d;i++){
        
			// for each vertex
            for(j=0;j<3;j++){
                if (vertexArray[j][i] < lower[i] || vertexArray[j][i] > upper[i] ){
                   return false;
				}
            }
        }
			
		return true;
	}



    public boolean intersects(GBounds b){
		
		int i,j;
        Point[] cellVertices=this.getVertexCoords();
        byte [] outcodes=new byte[3];
		
		Point lowerPoint=b.getLower();
		float [] lower=lowerPoint.getCoords();
		
		Point upperPoint=b.getUpper();
		float [] upper=upperPoint.getCoords();
    
        if(lower.length!=2){
            new IllegalArgumentError(SDBError.Fatal,"GBounds.intersects(): GBounds is not 2D.");
        }

        // We could make this faster with a single function call,
        // which avoids getting the bounds axis values repeatedly.
        for(i=0;i<3;i++){
            outcodes[i]=outcode(b, cellVertices[i]);
        }

        byte oc1, oc2, ocOR;
        int numClearlyOutside=0;
        
        for(i=0;i<3;i++){
        
            j=(i+1)%3;
            
            oc1=outcodes[i];
            oc2=outcodes[j];
            
            // Trivially accept triangles with vertices inside the
            // rectangle
            if(oc1==0 || oc2==0){
                return true;
            }
        
            // Test for lines that cross straight through from 
            // left to right, or top to bottom.
            ocOR=(byte) (oc1 | oc2);
            if(ocOR==0x03 || ocOR==0x0C){
                return true;
            }
            
            // Count the number of edges that are clearly
            // outside the rectangle
            if((oc1 & oc2) !=0) {
                numClearlyOutside++;
            } else {
                // The edge is not clearly outside, so we'll have to do
                // some line intersections. 
            
                // If one vertex has the center left outcode, then we
                // can try intersecting the edge with the left rectangle 
                // boundary.
                if(oc1 == 0x01 || oc2 ==0x01) { // left edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    lowerPoint,
                                    new Point(lower[0], upper[1]),
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x02 || oc2==0x02){ // right edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(upper[0], lower[1]),
                                    upperPoint,
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x04 || oc2==0x04){ // bottom edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    lowerPoint,
                                    new Point(upper[0], lower[1]),
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(oc1==0x08 || oc2==0x08){ // top edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(lower[0], upper[1]),
                                    upperPoint,
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(ocOR==0x0F) { //diagonals
                
                    // Here, we handle the case where the vertices lie in opposite 
                    // corner outcodes.  First, we'll see if the edge intersects with 
					// the left side.
					
					
					if(intersection_helper(	cellVertices[i],
									cellVertices[j],
									lowerPoint,
									new Point(lower[0], upper[1]),
									vertical)						
					){
						return true;
					}
					
					// Now, we have to see which of the two kinds of diagonals we have,
					// and either test the top or bottom sides of the rectangle.
					
					if (oc1==0x09 || oc2==0x09){
						// The line must go from upper left to lower right,
						// so we'll test the top side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										new Point(lower[0], upper[1]),
										upperPoint,
										horizontal)
							
						){
							return true;
						}
					} else {
						// The line must go from lower left to upper right,
						// so we'll test the bottom side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										lowerPoint,
										new Point(upper[0], lower[1]),
										horizontal)
							
						){
							return true;
						}
					}
					
				
                }
            } //if
        }


        if(numClearlyOutside==3){ 
            return false;
        } else {
            //return inside(this.lower,cellVertices);
            return this.inside(lowerPoint);

        }

    }




    public static boolean intersects(GBounds b, int[] vIndices, float[] coords, int triangleIndex){
		
		int i,j;
        
        byte [] outcodes=new byte[3];
		
		Point lowerPoint=b.getLower();
		float [] lower=lowerPoint.getCoords();
		
		Point upperPoint=b.getUpper();
		float [] upper=upperPoint.getCoords();
    
	
		Point[] cellVertices=new Point[3];
	
		int startIndex=3*triangleIndex;
		
		for(i=0; i<3; i++){
			
             //rdb next line modified to reflect new semantics for Point constructor.
		    //rdb cellVertices[i]=new Point(coords, vIndices[startIndex + i], 2);
			cellVertices[i]=new Point(coords, vIndices[startIndex + i]*2, 2);
		}
	
	
        if(lower.length!=2){
            new IllegalArgumentError(SDBError.Fatal,"GBounds.intersects(): GBounds is not 2D.");
        }

        // We could make this faster with a single function call,
        // which avoids getting the bounds axis values repeatedly.
        for(i=0;i<3;i++){
            outcodes[i]=outcode(b, cellVertices[i]);
        }

        byte oc1, oc2, ocOR;
        int numClearlyOutside=0;
        
        for(i=0;i<3;i++){
        
            j=(i+1)%3;
            
            oc1=outcodes[i];
            oc2=outcodes[j];
            
            // Trivially accept triangles with vertices inside the
            // rectangle
            if(oc1==0 || oc2==0){
                return true;
            }
        
            // Test for lines that cross straight through from 
            // left to right, or top to bottom.
            ocOR=(byte) (oc1 | oc2);
            if(ocOR==0x03 || ocOR==0x0C){
                return true;
            }
            
            // Count the number of edges that are clearly
            // outside the rectangle
            if((oc1 & oc2) !=0) {
                numClearlyOutside++;
            } else {
                // The edge is not clearly outside, so we'll have to do
                // some line intersections. 
            
                // If one vertex has the center left outcode, then we
                // can try intersecting the edge with the left rectangle 
                // boundary.
                if(oc1 == 0x01 || oc2 ==0x01) { // left edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    lowerPoint,
                                    new Point(lower[0], upper[1]),
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x02 || oc2==0x02){ // right edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(upper[0], lower[1]),
                                    upperPoint,
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x04 || oc2==0x04){ // bottom edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    lowerPoint,
                                    new Point(upper[0], lower[1]),
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(oc1==0x08 || oc2==0x08){ // top edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(lower[0], upper[1]),
                                    upperPoint,
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(ocOR==0x0F) { //diagonals
                
                    // Here, we handle the case where the vertices lie in opposite 
                    // corner outcodes.  First, we'll see if the edge intersects with 
					// the left side.
					
					
					if(intersection_helper(	cellVertices[i],
									cellVertices[j],
									lowerPoint,
									new Point(lower[0], upper[1]),
									vertical)						
					){
						return true;
					}
					
					// Now, we have to see which of the two kinds of diagonals we have,
					// and either test the top or bottom sides of the rectangle.
					
					if (oc1==0x09 || oc2==0x09){
						// The line must go from upper left to lower right,
						// so we'll test the top side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										new Point(lower[0], upper[1]),
										upperPoint,
										horizontal)
							
						){
							return true;
						}
					} else {
						// The line must go from lower left to upper right,
						// so we'll test the bottom side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										lowerPoint,
										new Point(upper[0], lower[1]),
										horizontal)
							
						){
							return true;
						}
					}
					
				
                }
            } //if
        }


        if(numClearlyOutside==3){ 
            return false;
        } else {
            //return inside(this.lower,cellVertices);
            return inside(lowerPoint,vIndices,  coords, triangleIndex);

        }

    }


    
    /*  intersection_helper()
    
        This private helper function will return true if the line v1-v2 intersects the line b1-b2, and 
        false otherwise. The last argument specifies whether b1-b2 is horizontal or vertical, using
        the contants defined below.
          It is important that b1 < b2 for the axis being checked. If not, incorrect results will be 
        generated. Also, b1-b2 is meant to be one side of a GBounds, so it is currently assumed to be
        either horizontal or vertical, depending on the last argument. However, v1-v2 can have any
        orientation. 
    */
    private static final int vertical=0;
    private static final int horizontal=1;
    
    private static final boolean intersection_helper(Point v1, Point v2, Point b1, Point b2,int vh){
    
        float t;
        
		
        if(vh==vertical){
            float x0=v1.getCoord(0);
            float x1=v2.getCoord(0);
                    
            t=(b1.getCoord(0)-x0)/(x1-x0);
            
            if( t < 0 || t > 1){
				System.out.println("Foo");
                return false;
            } else {
                float y0=v1.getCoord(1);
                float y1=v2.getCoord(1);
                float y=y0+t*(y1-y0);
				
                if(y < b1.getCoord(1) || y > b2.getCoord(1)){ // Problem here // seems alright
                    return false;
                } else {
                    return true;
                }
            }
        } else { //Horizontal

            float y0=v1.getCoord(1);
            float y1=v2.getCoord(1);
                    
            t=(b1.getCoord(1)-y0)/(y1-y0);
            
            if( t < 0 || t > 1 ){
				System.out.println("Foo");
                return false;
            } else {
                float x0=v1.getCoord(0);
                float x1=v2.getCoord(0);
                float x=x0+t*(x1-x0);
            
                if(x < b1.getCoord(0) || x > b2.getCoord(0)){
                    return false;
                } else {
                    return true;
                }
            }
        
        
        }
		
    }
 
 
	/* 	1001	1000	1010
		0001	0000	0010
		0101	0100	0110
	*/
     
    private static final byte outcode(GBounds bounds, Point p){
    
        float x=p.getCoord(0);
        float y=p.getCoord(1);
		
		float [] upper=bounds.getUpperArray();
		float [] lower=bounds.getLowerArray();
		
            
        return (byte)(((y > upper[1]?1:0) << 3) |
                      ((y < lower[1]?1:0) << 2) |  
                      ((x > upper[0]?1:0) << 1) |
                      ((x < lower[0]?1:0)     ) );
    }

	
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        StringBuffer result=new StringBuffer(indent+"TriangleCell: ");
        result.append(super.toString(""));
        
        return new String(result);
    }
    
    public static void main(String argv[]){
    
        // Test 1
        System.out.println("Test 1: all true.\n");
    
        TriangleCell c1=new TriangleCell(new Point(4, 4),new Point(12, 12), new Point(20,4));
        
        Point p1=new Point(4,4);
        Point p2=new Point(12,4);
        Point p3=new Point(20,4);
        Point p4=new Point(8,8);
        Point p5=new Point(12,8);
        Point p6=new Point(16,8);
        Point p7=new Point(12,12);
        Point p8=new Point(8,4);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));


        // Test 2
        System.out.println("\n\nTest 2: all false.\n");
    
        c1=new TriangleCell(new Point(4, 4),new Point(12, 12), new Point(20,4));
        
        p1=new Point(2,2);
        p2=new Point(4,2);
        p3=new Point(12,2);
        p4=new Point(22,2);
        p5=new Point(4,12);
        p6=new Point(20,12);
        p7=new Point(12,16);
        p8=new Point(16,16);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));



        // Test 3
        System.out.println("\n\nTest 3: all true.\n");

        c1=new TriangleCell(new Point(4, 12),new Point(12, 4), new Point(20,12));
        
        p1=new Point(12,4);
        p2=new Point(8,8);
        p3=new Point(12,8);
        p4=new Point(16,8);
        p5=new Point(4,12);
        p6=new Point(8,12);
        p7=new Point(12,12);
        p8=new Point(20,12);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));


        // Test 4
        System.out.println("\n\nTest 4: all true.\n");

        c1=new TriangleCell(new Point(4, 4),new Point(4, 14), new Point(14, 4));
        
        p1=new Point(4,4);
        p2=new Point(9,4);
        p3=new Point(14,4);
        p4=new Point(12,5);
        p5=new Point(7,7);
        p6=new Point(4,9);
        p7=new Point(9,9);
        p8=new Point(4,14);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));

        // Test 5
        System.out.println("\n\nTest 5: all true.\n");

        c1=new TriangleCell(new Point(4, 4),new Point(14, 14), new Point(14, 4));
        
        p1=new Point(4,4);
        p2=new Point(9,4);
        p3=new Point(14,4);
        p4=new Point(11,7);
        p5=new Point(9,9);
        p6=new Point(14,9);
        p7=new Point(12,12);
        p8=new Point(14,14);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));


        // Test 6
        System.out.println("\n\nTest 6: all false.\n");

        c1=new TriangleCell(new Point(4, 4),new Point(4, 14), new Point(14, 4));
        
        p1=new Point(4,2);
        p2=new Point(9,2);
        p3=new Point(14,2);
        p4=new Point(2,4);
        p5=new Point(16,4);
        p6=new Point(2,10);
        p7=new Point(10,10);
        p8=new Point(14,10);
        //Point p9=new Point(2,14);
        //Point p10=new Point(4,16);
        
        System.out.println("c1.inside(p1)="+c1.inside(p1));
        System.out.println("c1.inside(p2)="+c1.inside(p2));
        System.out.println("c1.inside(p3)="+c1.inside(p3));
        System.out.println("c1.inside(p4)="+c1.inside(p4));
        System.out.println("c1.inside(p5)="+c1.inside(p5));
        System.out.println("c1.inside(p6)="+c1.inside(p6));
        System.out.println("c1.inside(p7)="+c1.inside(p7));
        System.out.println("c1.inside(p8)="+c1.inside(p8));


    }

}
