/* 
    XLDLReader.java
    Created by Denise Mitchell October 2002.
    University of New Hampshire
    Durham, NH 03824

    processes XML for Lattice metadata
*/
//------------------- Edit history --------------------------------------
// 06/05/08 rdb: makde ArrayList and Vectors generic
// 10/21/05 rdb: removed many blank lines; removed unused variables
//                forced references in main to unused variables we want to keep
//-----------------------------------------------------------------------

package edu.unh.sdb.lattice;
import java.util.*;
import org.w3c.dom.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

public class XLDLReader extends XMLReader{

  /** declare a table/list to hold the actual datasources */
  private LatticeCollection latticeList;
    
  /** declare a table/list to hold the actual datasources */
  private DataSourceCollection dsList;
  /** declare a list to hold the FDLGroup of a lattice */
  private ArrayList<FDLGroup> fdlGroupList = new ArrayList<FDLGroup>();

  /** declare a list to hold LatticeBuilder */
  private ArrayList<LatticeBuilder> latticeBuilderList = new ArrayList<LatticeBuilder>();
  /** interim lower bounds specifications */
  private Vector<Float> lowerAxes;
    
  /** interim upper bounds specifications */
  private Vector<Float> upperAxes;

  /** declare the root datasource object */
  private DataSource rootDS;
    
  /** declare the ISBounds for the pending datasource */
  private ISBounds isBounds;
    
  /** declare the DataSourceBuilder to process logical datasources */
  private LatticeBuilder latticeBuilder;


/**  Takes an argument name of a file or URL as a String, a verbose indicator
    and a data source list*/
  public XLDLReader( String inf, boolean v, DataSourceCollection ds) 
  {
    super(inf, v);

    if (ds != null)
        dsList = ds;
    else
        dsList = new DataSourceCollection();
  }
    
/**  Takes an argument name of a file or URL as a String and a verbose indicator*/
  public XLDLReader( String inf, boolean v) 
  {
    super(inf, v);
  }
    
/**  Takes an argument name of a file or URL as a String and sets verbose to false*/
  public XLDLReader( String inf) 
  {
    super(inf);
  }
  
  /** Instantiates the LatticeCollection object to hold new parser info*/
  /** Sets all pending values to null and sets the join flag to FALSE */ 
  protected void setDefaults()
  {
    if (latticeList == null)
            latticeList = new LatticeCollection(); 
    resetBoundsAxes();
  }

  /** main method used for testing the class */
  static public void main(String[] argv)
  {        
    if (argv.length != 1) {
      System.err.println("Missing fileName.");
      System.exit(1);
      }
        
    try {
      XLDLReader theApp = new XLDLReader(argv[0], true); // instantiate the application
      if ( theApp == null ) // use this to force references to unused things we want
      {
         if ( theApp.isBounds == null || theApp.dataSourceValid( null ) )
            theApp.makeIntBaseArray( 0 );
      }
  //    XLDLReader theApp = new XLDLReader(argv[0], false); // instantiate the application
        
    } catch (Exception e)
      {
        e.printStackTrace();
        System.exit(1);
      }
        System.exit(0);
    }
     
  /** process the main document in the xml file */
  protected boolean processDoc()
  {
      //Document doc = parser.getDocument();

        // Document is well-formed
            
        // Traverse all the children of the root element
        
 
      for (Node child = doc.getDocumentElement().getFirstChild();
               child != null; child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {
          if (child.getNodeName().equals("Lattice"))
          {
            processLattice(child);
          }
        }
            
       } // end "for" loop, no more children nodes
        
        if(verbose)
        {
          System.out.println("final list of Lattices: " +
                getLattices());
        }

        return true;
}

  /** main processing of each individual lattice element **/
  private void processLattice(Node n)
  {
    
    if(verbose)
    {
      System.out.println("  Processing Lattice Node: " +
          makeChildrenText(n.getAttributes().getNamedItem("latName")) );
    }
        
    latticeBuilder = new LatticeBuilder(
             makeChildrenText(n.getAttributes().getNamedItem("latName")));
        
    if(nameExists(makeChildrenText(n.getAttributes().getNamedItem("latName"))) )
    { //discard if the lattice name is already in the table
      System.err.print( "Lattice: " + 
            makeChildrenText(n.getAttributes().getNamedItem("latName")) +
         " is a duplicate definition. Discarding this definition.");
      return;
    }
        
    for (Node child = n.getFirstChild(); child != null;
            child = child.getNextSibling())
    {
      // When child is an element
      if ( child instanceof Element )
      {            
        if (child.getNodeName().equals("Geometry"))
        {
          processGeometry(child, "lattice");
        }
          else if (child.getNodeName().equals("Topology"))
        {
          processTopology(child);
          if ( fdlGroupList != null && fdlGroupList.size() > 0 )
           finalProcessFDLGroup();

        }
        else if (child.getNodeName().equals("Approximator"))
        {
          processApproximator(child);
        }
      }
     } // end "for" loop, no more components to process nodes
    //check Topology's Geometry, if it is omitted, set it to the lattice's Geometry
    latticeBuilder.checkTopology();
    if ( latticeBuilderList == null )
        latticeBuilderList= new ArrayList<LatticeBuilder>();
    latticeBuilderList.add( latticeBuilder );
     /** make the lattice and add to the list. **/
    latticeList.add(latticeBuilder.makeLattice());
  }// end process lattice method
    
    
  /** Returns true/false if the name exists in the
    datasource collection */
  public boolean nameExists(String newDSName)
  {
    return latticeList.nameExists(newDSName);
  }
        
  /** Adds a new lattice to the existing collection */
  public void addLattice(Lattice ls)
  {
     if (verbose)
        System.out.println("ADDING Lattice: " + ls.getName());

     getLattices().add(ls);
  }



  /** main processing of the Geometry element **/
  private void processGeometry(Node n, String type)
  {
    if (verbose)
      System.out.println("Processing a " + type + " geometry.");
    String gType = makeChildrenText(n.getAttributes().getNamedItem(
                                     "type")).trim();
    if (verbose)
      System.out.println("Geometry type: " + type);
    if (gType.toUpperCase().equals("UNSTRUCTURED_GEOMETRY"))
    {
          latticeBuilder.setGeometryType(LatticeBuilder.UNSTRUCTURED_GEOMETRY, type);
    }
    if (gType.toUpperCase().equals("REGULARCELL_GEOMETRY"))
    {
          latticeBuilder.setGeometryType(LatticeBuilder.REGULARCELL_GEOMETRY, type);
    }
    if (gType.toUpperCase().equals("REGULARPOINT_GEOMETRY"))
    {
          latticeBuilder.setGeometryType(LatticeBuilder.REGULARPOINT_GEOMETRY, type);
    }

     String partitioningDimensions = makeChildrenText(n.getAttributes().getNamedItem(
                                     "partitioningDimensions")).trim();
     processPartitioningDimensions(partitioningDimensions, type);
    for (Node child = n.getFirstChild(); child != null;
            child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
          if (child.getNodeName().equals("GBounds"))
          {
            processBounds(child, type);
          }
        }

       } // end "for" loop, no more components to process nodes  
        
  }
    
    /** processes the DomainBounds element **/
  private void processBounds(Node n, String type)
  {
      if (verbose)
        System.out.println("within processBounds.");
   
      resetBoundsAxes();
        
      for (Node child = n.getFirstChild(); child != null;

            child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
          {
            processBoundsPair(child);
          }
         }        
       } // end "for" loop, no more components to process nodes
      String nodeName = n.getNodeName();
      if(nodeName.equals("GBounds"))
        processFinalGBounds(type);
      /*
      if(nodeName.equals("ISBounds"))
          processFinalISBounds(type);
       */
    }

    /** Processes a single DomainBounds axis pair*/
    private void processBoundsPair(Node n)
    { 
      if (verbose)
        System.out.println("within processBoundsPair.");
        
      lowerAxes.addElement(makeChildrenFloat(
         n.getAttributes().getNamedItem("lower")));
      upperAxes.addElement(makeChildrenFloat(
         n.getAttributes().getNamedItem("upper")));
    }
  
  /** instantiate a GBounds for the LatticeBuilder using the pairs gathered for the geometry**/
  private void processFinalGBounds(String type)
  {
    if (verbose)
      System.out.println("within processFinalGBounds.");
        
      //did user specify any bounds for this geometry?
      if (lowerAxes.size() < 1)
          return; //no bounds, just do nothing
     if (verbose)
     {
        System.out.println("corner 1" + new Point(vectorToFloatArray(lowerAxes)));
        System.out.println("corner 2: " + new Point(vectorToFloatArray(upperAxes)));
     }
        
      latticeBuilder.setGBounds(new GBounds(
      vectorToFloatArray(lowerAxes), vectorToFloatArray(upperAxes)), type );
  }
    
  /** instantiate a ISBounds using the pairs gathered for the topology within a partition**/
/*
  private void processFinalISBounds(String type)
  {
    if (verbose)
      System.out.println("within processFinalISBounds.");
        
      //did user specify any bounds for this geometry?
      if (lowerAxes.size() < 1)
          return; //no bounds, just do nothing
        
      latticeBuilder.setISBounds(new ISBounds(
          new IntegerIndexSpaceID(vectorToIntArray(lowerAxes)), 
          new IntegerIndexSpaceID(vectorToIntArray(upperAxes)) ), type);
  }
*/
  /** process partitioningDimensions */
  private void processPartitioningDimensions(String partitionStr, String type)
  {
        if (verbose)
        System.out.println("within processPartitioningDimensions." + partitionStr);
        int[] partition = processArrayOfInt(partitionStr);
        if(partition == null)
        {
            //partitonStr is an empty string
            partition=new int[0];
        }
        latticeBuilder.setISBounds(new ISBounds( new IntegerIndexSpaceID(partition) ), type); 
  }
  private int[] makeIntBaseArray(int size)
  {
      int[] ret= null;
      if(size > 0)
      {
          ret = new int[size];
          for(int i=0; i<size; i++)
              ret[i]=0;
      }
      return ret;
  }
  /** deduct 1 from each element */
  /*
  private int[] makeIntUpperArray(int[] partition)
  {
      int[] ret= null;
      if(partition != null)
      {
          int size= partition.length;
          ret = new int[size];
          for(int i=0; i<size; i++)
              ret[i]=partition[i]-1;
      }
      return ret;
  }
   */
  /** initializes the vectors used for accumulating bounds upper/lower axes **/
  private void resetBoundsAxes()

  {
    lowerAxes = new Vector<Float>();
    upperAxes = new Vector<Float>();
  }
  
    
  /** main processing of the Topology element **/
  private void processTopology(Node n)
  {
    if (verbose)
      System.out.println("Processing a Topology.");
    
    int dimensionality = Integer.parseInt(makeChildrenText(n.getAttributes().getNamedItem("dimensionality")).trim());
    if(verbose)
        System.out.println("Dimension is " + dimensionality);
      latticeBuilder.setDimensionality(dimensionality);
    
    for (Node child = n.getFirstChild(); child != null;
            child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
          if (child.getNodeName().toUpperCase().equals("RECTILINEARTOPOLOGY"))
          {
              latticeBuilder.setTopologyType(LatticeBuilder.RECTILINEAR_TOPOLOGY);
              processRectlinearTopology(child);
          }
          if (child.getNodeName().toUpperCase().equals("SPLITUNSTRUCTUREDTOPOLOGY"))
          {
              latticeBuilder.setTopologyType(LatticeBuilder.UNSTRUCTURED_TOPOLOGY);
              processSplitUnstructuredTopology(child);
          }
         }
       } // end "for" loop, no more components to process nodes  
      
  }
    /** main processing of RectlinearTopology */
  private void processRectlinearTopology(Node n)
  {
      if (verbose)
        System.out.println("Processing a RectlinearTopology.");
      for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
          if (child.getNodeName().equals("DataSource"))
          {     
              processDataSource(child);
          }
          if (child.getNodeName().equals("Geometry"))
          {
              processTopologyGeometry(child);
          }
         }
       } // end "for" loop, no more components t
  }
  
  /** process node for a root Data Source including calling the XDDLReader */
    private void processDataSource(Node n)
    {
    XDDLReader ddlReader;
    String dsName;
    if(verbose)
    {
      System.out.println("    Processing DataSource: " +
          makeChildrenText(n.getAttributes().getNamedItem("dsName")) + 
          " for fileName: " + makeChildrenText(n.getAttributes().getNamedItem("fileName")));
    }
    //Process a new DDL file if necessary
    if (makeChildrenText(n.getAttributes().getNamedItem("fileName")).length() > 0)
    {  //there is a DDL fileName so call the DDLReader
      ddlReader = new XDDLReader(
      makeChildrenText(n.getAttributes().getNamedItem("fileName")), verbose);
      if (ddlReader != null)
      {
         //merge the current DSCollection with the one just read
        if (dsList != null)
          dsList.merge(ddlReader.getDataSources());
        else
            dsList = ddlReader.getDataSources();
      }
    }
    
    dsName = makeChildrenText(n.getAttributes().getNamedItem("dsName"));

    if(dsName != null)   //if the user specified a data source name
    {   // does the root datasource exist in the DS collection from the ddlreader
      if (getDataSources() != null)
         latticeBuilder.setTopologyDS(getDataSources().getDataSource(dsName));
      else
         System.err.println("Root DataSource: " + dsName + " not found in DataSource Collection");
    }
    else
    {
      System.err.println ("Lattice: " + latticeBuilder.getName()
                +  " has no data source specified");
      return;
    }
    return;
  }

    /** skeleton for method to validate datasources at a later date **/
  private boolean dataSourceValid(XDDLReader ddl) 
  {
     return true;
  }
  
  /** main processing of the TopologyGeometry */
  private void processTopologyGeometry(Node n)
  {
      if(verbose)
          System.out.println("Processing Topology Geomentry");
      //Currently this Geometry is the same as that of Lattice
      processGeometry(n, "topology");
  }
  
   /** main processing of SplitUnstructuredTopology */
  private void processSplitUnstructuredTopology(Node n)
  {
      if (verbose)
        System.out.println("Processing a SplitUnstructuredTopology.");
      //RecordSpec is specified as "1 2 3"
        int[] axis = null;
        String rsString=makeChildrenText(n.getAttributes().getNamedItem("vertexIndicesRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setVertexIndicesRecordSpec(new RecordSpec(axis));
        else
            System.out.println("vertexIndicesRecordSpec is null");
        if(verbose)
            System.out.println("vertexIndicesRecordSpec is " + rsString);
  
        axis = null;
        rsString=makeChildrenText(n.getAttributes().getNamedItem("coordinateRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setCoordinateRecordSpec(new RecordSpec(axis));
        else
            System.out.println("coordinateRecordSpec is null");
        if(verbose)
            System.out.println("coordinateRecordSpec is " + rsString);
        
        axis = null;
        rsString=makeChildrenText(n.getAttributes().getNamedItem("dataRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setDataRecordSpec(new RecordSpec(axis));
        else
            System.out.println("dataRecordSpec is null");
        if(verbose)
            System.out.println("dataRecordSpec is " + rsString);
        String tmp=makeChildrenText(n.getAttributes().getNamedItem("pointInfoFDLFileName")).trim();
        latticeBuilder.setPointInfoFDLFileName(tmp);
        if(verbose)
            System.out.println("pointInfoFDLFileName is " + latticeBuilder.getPointInfoFDLFileName( ));
        tmp=makeChildrenText(n.getAttributes().getNamedItem("vIndicesFDLFileName")).trim();
        latticeBuilder.setVIndicesFDLFileName(tmp);
        if(verbose)
            System.out.println("vIndicesFDLFileName is " + latticeBuilder.getVIndicesFDLFileName( ));
       tmp=makeChildrenText(n.getAttributes().getNamedItem("coordinateFDLFileName")).trim();
        latticeBuilder.setCoordinateFDLFileName(tmp);
        if(verbose)
            System.out.println("coordinateFDLFileName is " + latticeBuilder.getCoordinateFDLFileName( ));
      for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
            /*
          if (child.getNodeName().equals("TopologyDescriptor"))
          {     
              processTopologyDescriptor(child);
          }
          */
            if (child.getNodeName().equals("FDLGroup"))
            {
              processFDLGroup(child);
            }
          if (child.getNodeName().equals("Geometry"))
          {
              processTopologyGeometry(child);
          }
         }
       } // end "for" loop, no more components t
  }
  
  /** main processing of the TopologyDescriptor element **/
/**
  private void processTopologyDescriptor(Node n)
  {
    if (verbose)
      System.out.println("Processing a TopologyDescriptor.");
    //RecordSpec is specified as "1 2 3"
        int[] axis = null;
        String rsString=makeChildrenText(n.getAttributes().getNamedItem("vertexIndicesRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setVertexIndicesRecordSpec(new RecordSpec(axis));
        else
            System.out.println("vertexIndicesRecordSpec is null");
        if(verbose)
            System.out.println("vertexIndicesRecordSpec is " + rsString);
  
        axis = null;
        rsString=makeChildrenText(n.getAttributes().getNamedItem("coordinateRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setCoordinateRecordSpec(new RecordSpec(axis));
        else
            System.out.println("coordinateRecordSpec is null");
        if(verbose)
            System.out.println("coordinateRecordSpec is " + rsString);
        
        axis = null;
        rsString=makeChildrenText(n.getAttributes().getNamedItem("dataRecordSpec")).trim();
        if(rsString != null)
            axis=processArrayOfInt(rsString);
        if(axis != null)
            latticeBuilder.setDataRecordSpec(new RecordSpec(axis));
        else
            System.out.println("dataRecordSpec is null");
        if(verbose)
            System.out.println("dataRecordSpec is " + rsString);

    for (Node child = n.getFirstChild(); child != null;
            child = child.getNextSibling())
      {
        // When child is an element
        if (child instanceof Element)
        {            
          if (child.getNodeName().equals("FDLGroup"))
          {
              processFDLGroup(child);
          }
         }
       } // end "for" loop, no more components to process nodes  
  }
  **/
  
   /** main processing of the FDLGroup element **/
  private void processFDLGroup(Node n)
  {
        if (verbose)
            System.out.println("Processing a FDLGroup.");
        FDLGroup fdlGroup = new FDLGroup();
        fdlGroup.vertexIndicesFileName = makeChildrenText(n.getAttributes().getNamedItem("vertexIndicesFileName")).trim();
        fdlGroup.coordinateFileName = makeChildrenText(n.getAttributes().getNamedItem("coordinateFileName")).trim();
        fdlGroup.dataFileName = makeChildrenText(n.getAttributes().getNamedItem("dataFileName")).trim();
        if(verbose)
        {
            System.out.println("vertexIndicesFileName is " + fdlGroup.vertexIndicesFileName);
            System.out.println("coordinateFileName is " + fdlGroup.coordinateFileName);
            System.out.println("dataFileName is " + fdlGroup.dataFileName);
        }
        if(fdlGroupList == null)
            fdlGroupList = new ArrayList<FDLGroup>();
        fdlGroupList.add(fdlGroup);
  }
  
  /** add FDLGroup to LatticeBuilder */
  private void finalProcessFDLGroup()
  {
      if(fdlGroupList != null)
      {
          int size = fdlGroupList.size();
          if(size > 0)
          {
              FDLGroup[] fdlGroupArray = new FDLGroup[size];
              for(int i=0; i < size; i++)
                  fdlGroupArray[i] = fdlGroupList.get(i);
              latticeBuilder.setFDLGroup(fdlGroupArray);
              fdlGroupList = new ArrayList<FDLGroup>();
          }
      }
  }

  /** main processing of the Approximator element **/
  private void processApproximator(Node n)
  {
    if (verbose)
      System.out.println("Processing an approximator.");
        
    if (makeChildrenText(n.getAttributes().getNamedItem("type")).equalsIgnoreCase("AreaTriangleApproximator"))
      latticeBuilder.setApproximatorType(LatticeBuilder.AREA_TRIANGLE_APPROXIMATOR);
    else // default is to use the array cell approximator
      latticeBuilder.setApproximatorType(LatticeBuilder.ARRAY_APPROXIMATOR);
  }
    
  /** Returns the datasource name*/
  public String getDSName()
  {
    return rootDS.getName();
  }
  /** Returns the list of LatticeBuilder */   
  public ArrayList getLatticeBuilders()
  {
      return latticeBuilderList;
  }
  /** Returns the current LatticeBuilder */
  public LatticeBuilder getCurrentLatticeBuilder()
  {
      return latticeBuilder;
  }
  /** Returns the LatticeBuilder with the lattice name specified */
  public LatticeBuilder getLatticeBuilder(String name)
  {
      LatticeBuilder lb= null;
      if(latticeBuilderList != null && !latticeBuilderList.isEmpty())
      {
          int size = latticeBuilderList.size();
          for(int i=0; i < size; i++)
          {
              LatticeBuilder tmp = latticeBuilderList.get(i);
              if(tmp.getName().equals(name))
              {
                  lb=tmp;
                  break;
              }
          }
      }
      return lb;
  }

  /** Returns the list of Datasources */
  public LatticeCollection getLattices()
  {
    return latticeList;
  }
    
  /** Returns the list of Datasources */
  public DataSourceCollection getDataSources()
  {
    return dsList;
  }
   
}
