/********************** CommandProcessor.java *****************************
 *  Release 1 as unpackaged version.
 *  (identified as such 01/20/05, though no major changes made since 11/01).
 *
 *  Provides a line-based interface for a simple command parser. 
 *  We use the StreamTokenizer to do this, but we modify the standard
 *  options so that:
 *      Comments are initiated by #
 *      All characters are "word" characters, except " and ' that
 *      define quoted strings.
 *
 *    After parsing, a command is executed via a CommandAction class
 *    Defines exceptions to report problems instead of output errors
 *       CommandDefException occurs when bad command definition found
 *       CommandException occurs when the parsing phase discovers an 
 *          invalid command.
 *       ParameterException occurs when the parsing phase discovers an
 *          incorrect parameter type.
 *
 * rdb   10/00
 * 06/04/08  Made all Vectors generic
 * 11/23/05  fixed basic error in readLine; eof not properly returned. Only shows
 *           up if last command in input file has too few arguments; only showed up
 *           when we added the parseCommand( String ) method which creates an input
 *           stream from a string.
 * 11/17/05  added parseCommand( String ), for parsing one command at a time
 *           where command does not come from a file 
 * 05/11/05  EOF error fixed. If eof occurs without eol, the tokens on 
 *           last line were lost. This was hard to make happen with line 
 *           editors, happens often with gui editors. 
 * 10/21/05  removed unused variables; force phony reference in main to keep one
 *             without generating compiler warning.
 * 11/26/01  Changed readLine so returns true if any tokens read. Now
 *           works correctly with StringReader
 * 11/01/01  Add setEOFAction call.
 * 10/26/01  Add comment options. Define comment char
 *           setCommentAction( CommandAction )
 *  9/29/01  Added constructors 
 *               (String      filename)
 *               (InputStream input)
 *           Added newReader( String )
 *                 newReader( InputStream )
 *  8/10/01  Added support for multiword commands
 *  6/12/01  Added heuristics to avoid multiple match error
 *  6/10/01  Added no arg constructor.  newReader doesn't delete defs.
 *  5/29/01  Changed test for # required args to use numReqArgs
 *           Fixed getLine to use StreamTokenizer.
 *           Fixed very old bug in which CommandSpec.init not called.
 *           Constructor now accept Reader, not InputStreamReader.
 *  5/ 8/01  Previous change caused error on blank line
 *  4/26/01  Added hack to handle -string. It only handles it by
 *           turning it into two separate String tokens: - string
 *  2/10/01  Fixed a bug when argument type specifies 0 or more args,
 *           such as ".*" or "f*". Wouldn't recognize 0 argument input.
 *  2/ 6/01  Added "." as legal argument specification meaning any type.
 *  2/ 5/01  Added "processAllCommands" as synonym for "readAllCommands"
 */
package edu.unh.sdb.util.CommandProcessor;

import java.io.*;
import java.util.Vector;

//++++++++++++++++++++++ CommandProcessor ++++++++++++++++++++++++++++
public class CommandProcessor 
{
  private StreamTokenizer   tok = null;  // tokenizer to use for input
  private boolean           numberParse = false; // true if parse #'s
  private Reader            rdr;       // Reader asssoc w/ tok
  private Vector<CommandSpec> legalCmds; // vector of valid commands
  private char              commentStart = '#';     // default comment char
  private String            commentStartStr = "#"; 
  private CommandAction     commentAction = null; 
  private CommandAction     eofAction = null; 

  //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  //------------------ CommandProcessor( ) ------------------
  public CommandProcessor( )
  {
    //------ create Vector to save legal command specifications
    legalCmds = new Vector<CommandSpec>();
  }
  //------------------ CommandProcessor( String filename ) -----------
  public CommandProcessor( String filename ) throws IOException
  {
    this();
    //------ create tokenizer.
    if ( !newReader( filename ) )
      if ( filename == null )
        throw new NullPointerException("CommandProcessor: bad String" );
      else
        throw new IOException( "CommandProcessor can't open:" 
                                                       + filename );
  }
  //------------------ CommandProcessor( Reader ) ------------------
  public CommandProcessor( Reader rdr )
  {
    this();
    newReader( rdr );
  }
  //------------------ CommandProcessor( InputStream )----------------
  public CommandProcessor( InputStream input )
  {
    this( new InputStreamReader( input ) );
  }
  //------------- newReader( Reader ) ----------------------------
  public void newReader( Reader rdr )
  //
  {
    //------ create tokenizer
    this.rdr = rdr;
    this.tok = new StreamTokenizer( rdr );
    initTokenizer( );
  }
  //------------- newReader( InputStream ) ---------------------------
  public void newReader( InputStream input )
  //
  {
    newReader( new InputStreamReader( input ));
  }
  //------------- newReader( String filename ) ----------------------
  public boolean newReader( String filename ) 
  //
  {
     Reader rdr = null;
     try
     {
        rdr = new BufferedReader( 
                   new InputStreamReader( 
                     new FileInputStream( filename )));
     }
        catch( Exception e )
        {
           return false;
        }
     newReader( rdr );
     return true;
  }
  //------------- processAllCommands( ) ------------------------------
  // Add because of documentation error
  public void processAllCommands() throws IOException
  {
     readAllCommands();
  }
  //------------- commentChar( char ) ------------------------------
  // Allow user to specify comment chars for the StreamTokenizer
  public void commentChar( char cChar ) 
  {
     this.commentStart = cChar;
     this.commentStartStr = new String( new char[]{ cChar });
     initTokenizer();
  }
  //------------- setCommentAction( CommandAction ) --------------------------
  // Allow user to specify comment chars for the StreamTokenizer
  public void setCommentAction( CommandAction comAction ) 
  {
     this.commentAction = comAction;
     initTokenizer();
  }
  //------------- setEOFAction( CommandAction ) --------------------------
  // Allow user to specify comment chars for the StreamTokenizer
  public void setEOFAction( CommandAction comAction ) 
  {
     this.eofAction = comAction;
  }
  //------------- readAllCommands( ) ---------------------------------
  //
  public void readAllCommands() throws IOException
  {
    boolean more = true;
    while ( more )
       try
       {
         more = nextCommand();
       }
       catch ( CommandException badCmd )
       {
         System.err.println("***Command Error: " + badCmd.getMessage());
       }
       catch ( ParameterException badParm )
       {
         System.err.println("***Param Error:" + badParm.getMessage());
       }
       catch ( EOFException ioex )
       {
         if ( this.eofAction != null )
            more = eofAction.execute( new Vector() );
         else
            more = false;
         System.err.println( "***Unexpected EOF: " + ioex.getMessage());
       }
       catch ( IOException ioex )
       {
         System.err.println( "***IOException: " + ioex.getMessage());
       }
  }
  //--------------------- getLine( ) -------------------------
  // redefine StreamTokenizer and read upto the next EOL 
  // as raw text.
  //
  public String getLine( )  
  {
     StringBuffer line = new StringBuffer();

     tok.resetSyntax();
     tok.eolIsSignificant( true );
     try {
         while ( this.tok.nextToken() != StreamTokenizer.TT_EOL )
         {
            char ch = (char) this.tok.ttype;
            if ( ch == ('\n' ) )
               System.err.println( "getLine: got cr" );
            line.append( (char) this.tok.ttype );
         }
     }
       catch ( IOException ioex ) {
          // assume it's eof, but we only care if we've gotten nothing
         System.err.println( "getLine got io ex" );
         if ( line.length() == 0 )
            line = null;
       }
     initTokenizer();  // return tokenizer to normal state
     if ( line == null )
        return null;
     else 
        return new String( line );
  }
  
  //--------------------- parseCommand( ) ----------------------------
  public boolean parseCommand( String command ) throws IOException
  {
     this.tok = new StreamTokenizer( new StringReader( command ));
     initTokenizer();
     return nextCommand();
  }

 
  //--------------------- nextCommand( ) ----------------------------
  public boolean nextCommand() throws IOException
  //
  // nextCommand parses the input looking for a valid command.
  //    Commands must start at the beginning of a line, but the
  //       arguments can spread over multiple lines.
  //    Blank lines and lines with only comments are skipped over.
  //    If the command and its arguments are valid, it is executed
  //        using the CommandAction object specified at defineCommand
  //        time and nextCommand returns the value returnded by the 
  //        execute method. If the command is not valid, a 
  //        CommandException is thrown. 
  //    If eof is detected, false is returned.
  {
    Arguments   myArgs = new Arguments(); // for the arguments
    Vector      args = null;
 
    boolean     retVal = true;
    CommandSpec cmd = null;
    boolean     moreData = readLine( myArgs );

    if ( myArgs.size() > 0 )
    {
       Object  firstToken = myArgs.get( 0 );

       if ( !( firstToken instanceof String ) )
          throw new CommandException( "first token not a String: " 
                                          + firstToken );
       //        check for comment line
       else if ( ((String)myArgs.get( 0 )).equals( this.commentStartStr ) 
                 && this.commentAction != null )
          this.commentAction.execute( myArgs.getArgs() );
       else 
       {
          Vector cmds = findCommands((String)( firstToken ));
          checkMultiWordCmds( cmds, myArgs );
          if ( cmds.size() == 0 )
            throw new CommandException( "Invalid command: " + firstToken );
          else
          {
             selectCmd( cmds, myArgs );
             if ( cmds.size() > 1 )
                cmds = checkAmbiguity( cmds );
             if ( cmds.size() >= 1 )
             {
                cmd = (CommandSpec) cmds.get(0);
                args = convertArgs( cmd, myArgs );
                retVal = cmd.action.execute( args );
                if ( cmds.size() > 1 )
                {
                   String msg = "Ambiguous command invocation: " 
                         + ". Used: " + cmd + "\nOther matches";
                   for ( int i = 1; i < cmds.size(); i++ )
                      msg += ": " + ( cmds.get( i ) );
                   throw new CommandException( msg );
                }
             }
             else if ( cmds.size() == 0 )
                throw new CommandException( "Bad args for command: " +
                                             firstToken );
          }
       }
    }

    if ( !moreData )
    {
       // eof occurred
       if ( this.eofAction != null )
          retVal = eofAction.execute( new Vector() );
       else
          retVal = false;
    }
    return retVal;
  }
  //------ defineCommand( String cmd, String parms, CommandAction ) ---
  public void defineCommand(String cmd, String parms, CommandAction act)
                                         throws CommandDefException
  {
     CommandSpec newSpec = new CommandSpec( cmd, parms, act );
     legalCmds.add( newSpec );
     if ( !this.numberParse && newSpec.hasNumberArg() )
     {
        this.numberParse = true;
        if ( this.tok != null )
           this.tok.parseNumbers();
     }
  }
  //+++++++++++++++++++++++ private methods +++++++++++++++++++++++++++
  //------------- initTokenizer( ) -------------------
  private void initTokenizer( ) 
  {
    if ( this.tok != null )
    {
      this.tok.resetSyntax();  // turn off number parsing for now
      if ( this.numberParse )
         this.tok.parseNumbers();
      this.tok.parseNumbers();
      this.tok.wordChars( '!' /* 33 */, '~' /* 126 */ );
      this.tok.whitespaceChars( '\u0000', '\u0020' );
      if ( this.commentStart != 0 && commentAction == null )
         this.tok.commentChar( this.commentStart );     // # is comment start
      else 
         this.tok.ordinaryChar( this.commentStart );
      this.tok.quoteChar( '\'' );      // catch single quoted strings
      this.tok.quoteChar( '"' );       // catch double quoted strings
      this.tok.eolIsSignificant(true); //  want eol as a token
    }
  }
  //--------------------- readLine( Arguments args ) -------------------
  private boolean readLine( Arguments args ) throws IOException
  //
  // readLine parses the next line of input into tokens which it returns
  //            in a vector.
  //    Blank lines and lines with only comments are skipped over.
  //    If eof is detected, false is returned.
  //    otherwise true is returned
  {
    boolean     more = true;            // the loop control variable
    boolean     eof = false;
    int         startCount = args.size();
    while ( more ) 
    {
      if ( this.tok.nextToken() == StreamTokenizer.TT_WORD ) 
      {
         // is a word, so field 'sval' is a String that holds the word
         args.add( new String( this.tok.sval ));
      }
      else if ( this.tok.ttype == StreamTokenizer.TT_NUMBER )
      {
         // if it's a number, the 'nval' field has the value as a float
         // Let's check to see if the value is actually an integer
         //   if so, we'll return an Integer.
         if ( this.tok.nval == (int) this.tok.nval  )
            args.add( new Integer( (int) this.tok.nval ));
         else
            args.add( new Float( this.tok.nval ) );
      }
      else if (  ( this.tok.ttype == '"' ) 
              || ( this.tok.ttype == '\'' ) ) 
         args.add( new String( this.tok.sval ));
      else if ( this.tok.ttype == StreamTokenizer.TT_EOL )
      {
         if ( args.size() > startCount ) 
           more = false;
      }
      else if ( this.tok.ttype == StreamTokenizer.TT_EOF ) 
      {
         more = false;
         eof = true;
      }
      else if ( this.commentAction != null 
              && this.tok.ttype == this.commentStart )
      {  
         if ( args.size() == 0 ) // comment starts the line
         {
            args.add( this.commentStartStr );
            String line = getLine();
            args.add( line );
            more = false;
         }
         else     // finish off the command before generating comment
         {
            this.tok.pushBack();   // put comment symbol back in
            more = false;
         }
      }
      else // any ordinary character will be turned into a string
      {
         char charArray[] = new char[1];
         charArray[0] = (char) this.tok.ttype;
         args.add( new String( charArray ));
      }
    }
    return !eof && args.size() > 0;
  }
  //------------- checkAmbiguity( Vector ) ---------------------------
  private Vector checkAmbiguity( Vector cmds ) 
  {
     // We'll disambiguate by sorting on numPromote. Fewer promotions
     // will be chosen. We'll do a pretty dumb sort. First scan to
     // find min numPromote value, the do it again to delete all entries
     // with more promotions.
     CommandSpec cmd;
     int minPromote = 999999;
     for ( int i = 0; i < cmds.size(); i++ )
     {
         cmd = (CommandSpec) cmds.get( i );
         if ( cmd.numPromote < minPromote )
            minPromote = cmd.numPromote;
     }
     for ( int i = cmds.size() - 1; i >= 0; i-- )
        if (( (CommandSpec)cmds.get( i ) ).numPromote > minPromote )
        {
          cmds.removeElementAt( i );
        }
     return cmds;
  }
  //------------- findCommands( String ) ------------------------------
  // find all potential commands based on matching the arg to first word
  // of the command.
  private Vector<CommandSpec> findCommands( String word ) 
  {
     Vector<CommandSpec> allCmds = new Vector<CommandSpec>();

     for ( int i=0; i < legalCmds.size(); i++ ) 
     {
        CommandSpec cmd = legalCmds.get( i );
        if ( cmd.partialMatch( word ))
        {
           cmd.init();        //re-initialize matching fields 
           allCmds.add( cmd );
        }
     }
     return allCmds;
  }
  //---------- typeErrorString( int num, String expectedType ) ---------
  private void typeErrorString( String cmd, int num, String expectedType ) 
  {
    throw new ParameterException( "Command: " + cmd 
                           + " --> Bad type for argument: " 
                           + num +  " " + expectedType + " needed." );
  }
  //--------- checkMultiWordCmds( Vector cmds, Arguments args ) ------
  // If any of the potential commands are multiword commands, check the 
  // arguments to see if the remaining words match. If they do, those
  // arguments will be folded into the command name and other commands 
  // will be removed. I.e., multiword commands take precedence over
  // arguments to commands with fewer words.

  private void  checkMultiWordCmds( Vector cmds, Arguments args )  
                                                  throws IOException
  {
     CommandSpec cmd;
     int maxNumWords = 1;
     for ( int i = 0; i < cmds.size(); i++ )
     {
        cmd = (CommandSpec) cmds.get( i );
        if ( cmd.numCmdWords() >= maxNumWords && cmd.fullMatch( args ) )
           maxNumWords = cmd.numCmdWords();
     }
     // now delete any commands whose command word count don't match
     //    longest found. 
     for ( int i = cmds.size() - 1; i >= 0; i-- )
     {
         cmd = (CommandSpec) cmds.get( i );
         if ( cmd.numCmdWords() != maxNumWords )
             cmds.removeElementAt( i );
     }
     if ( maxNumWords > 1 )
     {
        // now fix up the args array. change args[0] to have full name
        //    remove the command words from the remaining args
        // args array has
        ///         0    1         n-1     n  n+1  n+2
        //        cmd0 cmd1 ... cmd(n-1) arg0 arg1 arg2 ...
        // change it to
        //        "cmd0 cmd1 ... cmd(n-1)" arg0 arg1 ...
        // This means deleting args[1] ... args[n-1]
        // but after we delete args[1], the old args[2] is now args[1]
        // so we really just need to delete args[1] n-1 times.
        cmd = (CommandSpec) cmds.get( 0 );
        args.setElementAt( cmd.getCommand(), 0 );
        for ( int i = 0; i < cmd.numCmdWords() - 1; i++ )
           args.removeElementAt( 1 );
     }
  }
  //----------- selectCmd( Vector cmds, Arguments args ) -------------
  // Find all commands whose name and arguments match  the input.
  // If more than one, then choose according to max # matched arguments.
  // This could still result in >1 match, however. We let the caller 
  // decide what to do then.

  private void  selectCmd( Vector cmds, Arguments args )  
                                                  throws IOException
  {
     Vector<Object> outArgs = new Vector<Object>(); // we'll copy arguments here
     CommandSpec cmd = null;
     boolean     completeMatch = false;
     boolean     done = false;

     outArgs.add( args.get( 0 ) );      // copy the command

     // for each argument in input, check validity.  to handle 
     // argument-less commands we make a pre-pass through the tests 
     // to see if one of the potential commands is an argument-less 
     // command.  We do this by starting argnum at 0
     int argNum = 0;
     while ( !done ) // check each actual arg til complete match or fail
     {
        completeMatch = false;
        for ( int i = 0; i < cmds.size(); i++ )
        {
           cmd = (CommandSpec) cmds.get( i );
           if ( argNum == 0 )
           {
              if ( cmd.numReqParms == 0 )
              {
                 completeMatch = true;
                 cmd.status = '+';
              }
           }
           else if ( cmd.status != '-' ) // only check still valid cmds
           {
              char  argType = args.getType( argNum );
              cmd.status = cmd.checkArg( argNum - 1, argType );
              if ( cmd.status == '+' )
                  completeMatch = true;
           }
        } // end for
        // delete failed entries - go backwards through the vector since
        // if went forward deletions would change the indexes.
        for ( int i = cmds.size() - 1; i >= 0; i-- )
        {
           cmd = (CommandSpec) cmds.get( i );
           if ( cmd.status == '-' )
               cmds.removeElementAt( i );
        }
        if ( argNum == args.size() - 1 ) // used all args read so far?
        {
           if  ( completeMatch        // done if we have a match
              || cmds.size() == 0 )   // or if no more candidates
              done = true;
           else 
           {
              boolean moreData = readLine( args );
              //11/23/05 if ( args.size() == 0 && !moreData )
              if ( !moreData )
              {
                 System.err.println( "CommandProcessor: " 
                              + "EOF while reading arguments for command: " 
                              + cmd.getCommand() + ". Missing "
                              + ( cmd.numReqParms = cmd.numMatch ) + " arguments.");
                 done = true;
              }
           }
        }
        argNum++;
    } // while

    // delete incomplete entries. need to go backwards 
    int maxNumMatch = 0;
    for ( int i = cmds.size() - 1; i >= 0; i-- )
    {
       //System.out.println( cmds.get( i ) );
       cmd = (CommandSpec)cmds.get( i );
       if ( cmd.status != '+' )
          cmds.removeElementAt( i );
       else if ( cmd.numMatch > maxNumMatch )
          maxNumMatch = cmd.numMatch;
    }
    // now delete entries that don't have maxNumMatch arg matching
    for ( int i = cmds.size() - 1; i >= 0; i-- )
    {
       cmd = (CommandSpec)cmds.get( i );
       if ( cmd.numMatch < maxNumMatch )
          cmds.removeElementAt( i );
    }
  }
  //-------- convertArgs( CommandSpec, Arguments inArgs )------------
  // This method is called after we determine that the arguments will
  // match the parameter specification of the command, but here we do
  // conversions of types that are necessary. This is because
  // StreamTokenizer produces only Float or String tokens. Essentially,
  // we sometimes need to convert Floats to either Integer or Double.
  // Modified for assignment 4. readLine now makes the test for integral
  // input values and produces Integer in the array it returns. To make
  // this routine more general, it no longer assumes that its input is
  // coming from StreamTokenizer -- it makes logical choices based on
  // whatever input it gets as long as the input is one of the 4 types:
  // Integer, Float, Double, String.

  private Vector  convertArgs( CommandSpec cmd, Arguments inArgs ) 
                                     throws IOException
  {
     Vector<Object>  outArgs = new Vector<Object>(); // we'll copy arguments here

     outArgs.add( inArgs.get( 0 ) );      // copy the command

     for ( int argIndex = 1; argIndex < inArgs.size() ; argIndex++ )
     {
        Object thisObj = inArgs.get( argIndex );
        if ( inArgs.size() > 0 )
          switch ( cmd.parmType( argIndex - 1 ) )
          {
          case '.':  // any argument type is legal
                outArgs.add( thisObj );
                break;
          case 'n':  // Number expected - can take Float or Integer
          case 'f':  // float expected - can take Float or Integer
                if ( thisObj instanceof Float ) 
                   outArgs.add( thisObj );
                else if ( thisObj instanceof Integer ) // cvt to Float
                  outArgs.add( new Float(((Integer) thisObj).floatValue()));
                else
                  typeErrorString(cmd.getCommand(), argIndex, "Float");
                break;
          case 'd':  // double expected - Double, Float or Integer ok
                //System.out.println( "-- checking 'd' ---" );
                if ( thisObj instanceof Double ) 
                   outArgs.add( thisObj );
                else if ( thisObj instanceof Float ) // ok, convert
                   outArgs.add( new Double(((Float) thisObj).floatValue()));
                else if ( thisObj instanceof Integer ) // ok, convert
                   outArgs.add( new Double(((Integer) thisObj).floatValue()));
                else
                   typeErrorString( cmd.getCommand(), argIndex, "Double" );
                break;
          case 'i':  // want int, Double/Float allowed if val int
                if ( thisObj instanceof Integer ) 
                   outArgs.add( thisObj );
                else if ( thisObj instanceof Number ) 
                {
                   float f =  ((Number) thisObj).floatValue();
                   if ( ( (int) f ) == f ) // it's an int
                       outArgs.add( new Integer( (int) f ) );
                   else
                      typeErrorString( cmd.getCommand(), argIndex, "Integer" );
                }
                else
                    typeErrorString( cmd.getCommand(), argIndex, "Integer" );
                break;
          case 's':  // String expected
                //System.out.println( "-- checking 's' ---" );
                if ( thisObj instanceof String ) 
                   outArgs.add( thisObj );
                else
                   typeErrorString( cmd.getCommand(), argIndex, "String" );
                break;
          case '?':  // we've gone beyond the number valid args
                     // Don't signal any error and no conversion.
                     // and don't copy input arg to output arg
                break;
          default:   // signal argument type specification error
                System.err.println( 
                        "***** CommandProcessor: Argument type error: " 
                        + cmd.parms.charAt( argIndex ));
                break;
          }
     }
     //System.out.println( "CheckArgs: " + outArgs );
     return outArgs;
  }
  //---------------------- main( String args[] ) ---------------------
  //
  // In principal, this is a standalone test harness for this class.
  // In practice, this shows the frame work and does not really serve
  // as a reasonable test harness for CommandProcessor.

  public static void main( String args[] ) throws IOException
  {
     System.out.println( "CommandProcessor.main" );
     CommandProcessor cmp = new CommandProcessor( 
                         new InputStreamReader( System.in ));
     /********/
     cmp.defineCommand( "+", "ii*", new CommandAction() );
     cmp.defineCommand( "ad", "dd*", new CommandAction() );
     cmp.defineCommand( "af", "ff*", new CommandAction() );
     /**************/
     /********* Test multiword commands **********************/
     cmp.defineCommand( "Two words", "ii", new CommandAction() );
     cmp.defineCommand( "Three words now", "iii", new CommandAction() );
     cmp.defineCommand( "Three", "iii", new CommandAction() );
     cmp.defineCommand( "  Add   spaces ", "i", new CommandAction() );
     cmp.defineCommand( "Add", "si", new CommandAction() );
     
     //cmp.readAllCommands();
     if ( cmp.rdr == null ) // force dumb reference to rdr so avoid compiler warning
        cmp.rdr = null;
     cmp.parseCommand( "Add string" );
  }
}
