///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          U N I V E R S I T Y    O F    N E W    H A M P S H I R E         //
//                                                                           //
//                       COMPUTER  SCIENCE  DEPARTMENT                       //
//                                                                           //
//                    Graphics And Visualization Research                    //
//                                                                           //
//          --------------------------------------------------------         //
//  File: GraniteRemoteIF.cpp                                                //
//                                                                           //
//  The GraniteRemoteIF provides a nice wrapper for functionality provided   //
//  by the RemoteConnection class.                                           // 
//                                                                           //
//                                                                           //
//  Programmer:  Jon Frain                                                   //
//  Creation Date: 7/19/09                                                   //
//                                                                           //
//  History:                                                                 //
///////////////////////////////////////////////////////////////////////////////


#include "GraniteRemoteIF.h"

//----------------------------------------------------------------------------
// Constructor
//
// args:  RemoteConnection *rc - all communication with granite remote server
//                               is sent through this class pointer
//
//----------------------------------------------------------------------------
GraniteRemoteIF::GraniteRemoteIF(char* name, int port) {
   remoteConn = new RemoteConnection( name, port );
}


//----------------------------------------------------------------------------
// Destructor
//
// args:  None
//
//----------------------------------------------------------------------------
GraniteRemoteIF::~GraniteRemoteIF(){

}

 
//----------------------------------------------------------------------------
// CreateDataSource
//
// args: char* dsName - Name to give this new datasource
//       char* dsPath - location of the xfdl file for the datasource
//                        
// purpose: creation of the granite data source
//
// returns: integer ID of the datasource as used by Granite
//----------------------------------------------------------------------------
int GraniteRemoteIF::createDataSource(char* dsName, char* dsPath){
    char message[1024];
    sprintf(message,"dsCreate %s %s\n",dsName,dsPath);
    return (remoteConn->sendStringGetInt(message));
} 



//----------------------------------------------------------------------------
// subblock
//
// args: int dsID - datasource ID for which the subblock is applied to
//       int start1, int start2, int start3 - 3D start location of subblock
//       int end1,   int end2,   int end3   - 3D end location fo subblock
//                        
// purpose: creates a subblock within a preexisting datasource
//
// returns: integer ID of the subblock as used by Granite
//----------------------------------------------------------------------------
int   GraniteRemoteIF::subblock(int dsID, int start1, int start2, int start3,
                                          int end1,   int end2,   int end3)
{
    char message[1024];
    sprintf(message,"subblock %d %d %d %d %d %d %d\n",dsID, start1, start2, 
            start3, end1, end2, end3);
    return (remoteConn->sendStringGetInt(message));
}


//----------------------------------------------------------------------------
// getFloat
//
// args:  int dsID - datasource ID to retrieve from         
//        int attrID - attribute to retrieve value for 
//                    (0 if single attribute data source)
//        int loc1, loc2, loc3 - 3D location of value to retrieve
//               
// purpose: Retrieve a single float value from a datasource
//
// returns: float - value retreived from datasource
//----------------------------------------------------------------------------
float GraniteRemoteIF::getFloat(int dsID, int attrID, 
                                int loc1, int loc2, int loc3)
{
    char message[1024];
    sprintf(message,"getFloat %d %d %d %d %d\n",dsID, attrID, loc1, loc2, loc3);
    return (remoteConn->sendStringGetFloat(message));
}



//----------------------------------------------------------------------------
// getFloatsPointOrder
//
// args:  int dsID - datasource ID to retrieve from
//        float* arr - preallocated array to be filled by this retrieval
//        int szInBytes - number of bytes the preallocated array takes up
//                        
// purpose: Fill an array of float values from a datasource - fills
//          array in point order
//
// returns: none - modifies argument array
//----------------------------------------------------------------------------
void GraniteRemoteIF::getFloatsPointOrder(int dsID, float* arr, int szInBytes)
{
    // Frain - the method called for this seems to need a space before '\n'
    //       - I don't have time at the moment to try and change that
    remoteConn->sendStringGetRawData(arr, szInBytes, "getFloatsByPoint %d \n",
                                     dsID);
}


//----------------------------------------------------------------------------
// getFloatsAttrOrder
//
// args:  int dsID - datasource ID to retrieve from
//        float* arr - preallocated array to be filled by this retrieval
//        int szInBytes - number of bytes the preallocated array takes up
//                        
// purpose: Fill an array of float values from a datasource - fills
//          array in attribute order
//
// returns: none - modifies argument array
//----------------------------------------------------------------------------
void GraniteRemoteIF::getFloatsAttrOrder(int dsID, float* arr, int szInBytes)
{
    // Frain - the method called for this seems to need a space before '\n'
    //       - I don't have time at the moment to try and change that
    remoteConn->sendStringGetRawData(arr, szInBytes, "getFloatsByAttr %d \n",
                                     dsID);
}


//------------------------------------------------------------------------
// getByteOrder
//
// args:  int dsID - datasource ID to retrieve from         
//               
// purpose: Retrieve the byte order from a datasource
//
// returns: int - Byte Order
//                0 - Big Endian
//                1 - Little Endian
//
//------------------------------------------------------------------------
int GraniteRemoteIF::getByteOrder(int dsID){
    char message[1024];
    sprintf(message,"getByteOrder %d \n",dsID);
    return (remoteConn->sendStringGetInt(message));
}


//------------------------------------------------------------------------
// setByteOrder
//
// args:  int dsID - datasource ID to retrieve from  
//        int order       
//                0 - Big Endian
//                1 - Little Endian
//               
// purpose: Set the byte order for a datasource
//
//------------------------------------------------------------------------
void GraniteRemoteIF::setByteOrder(int dsID, char* order){
    char message[1024];
    sprintf(message,"setByteOrder %d %s \n",dsID,order);
    remoteConn->sendStringGetInt(message);
}
    

//------------------------------------------------------------------------
// getDimensions
//------------------------------------------------------------------------
int* GraniteRemoteIF::getDimensions(int dsID, int* sz){
    char message[1024];

    // first need to get # of dimensions for the data source 
    sprintf(message,"getNumDimensions %d \n",dsID);
    *sz = remoteConn->sendStringGetInt(message);

    // method needs size in bytes
    *sz *= sizeof(int);

     int* arr = (int*)malloc(*sz);
    //  now request the dimensions
    if (sz != 0) {
       sprintf(message,"getDimensions %d \n",dsID);
       remoteConn->sendStringGetRawData(arr, *sz, message, dsID);
       *sz /= sizeof(int);
       return arr;
    }
    else {
      return NULL;
    }
}


//------------------------------------------------------------------------
// getNumAttributes
//------------------------------------------------------------------------
int GraniteRemoteIF::getNumAttributes(int dsID){
    char message[1024];

    // first need to get # of dimensions for the data source 
    sprintf(message,"getNumAttributes %d \n",dsID);
    return (remoteConn->sendStringGetInt(message));
}

//------------------------------------------------------------------------
// getAttributeName
//------------------------------------------------------------------------
char** GraniteRemoteIF::getAttributeNames(int dsID, int* sz){
    char message[1024];
    

    // first need to get # of dimensions for the data source 
    sprintf(message,"getAttributeNames %d \n",dsID);
    char* arr = remoteConn->sendStringGetString(message);

    fprintf(stderr,"       --->arr = %s\n", arr);

    // allocate space for 200 string pointers
    char** strArray = (char**)malloc(sizeof(char*) * 200);

    char *token;
    int i=0;
      
    token = strtok (arr," ");
    while (token != NULL)
    {
       strArray[i] = token;
       i++;    
       token = strtok (NULL, " ");    
    }
    *sz = i;

    return strArray;
}



//------------------------------------------------------------------------
// disableRemoteByteSwapping
//------------------------------------------------------------------------
void GraniteRemoteIF::disableRemoteByteSwapping(){
   remoteConn->disableByteSwapping();
}


//------------------------------------------------------------------------
// enableRemoteByteSwapping
//------------------------------------------------------------------------
void GraniteRemoteIF::enableRemoteByteSwapping(){
   remoteConn->enableByteSwapping();
}
