//
//  CacheDataSource.java
//  CacheTest
//
//  Created by Philip Rhodes on Tue Apr 15 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
/*------------------------- Edit history as of 01/30/05 ----------------------
 * 
 * 01/30/05 rdb: removed methods now implemented by DataCollection, such as:
 *                   Datum datum( id ) and Datum datum( id, recordspec );
 */



package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;


/** This is the abstract base class for all datasources that provide caching service for 
 *  an underlying datasource. 
 */


public abstract class CacheDataSource extends DataSource
{
    DataSource theSource;
    ISBounds   iterationSpace;
    Cache      theCache;

    /** Create a CacheDataSource with a Cache associated with the argument datasource.
      * The third argument specifies the DataSource subset that is cached.
      */
	public CacheDataSource(String name, DataSource ds, ISBounds iterationSpace)
	{
      super( name, ds.getRecordDescriptor(), ds.getBounds() );
        
      this.theSource = ds;
      this.iterationSpace = new ISBounds(iterationSpace);
	}

   /** When this method is called, this subblock methods of this collection will 
    *  read data as bytes, without performing any type conversion. This is meant 
    *  to be used by servers, since they send raw bytes over the network anyway,
    *  and any conversion is wasted effort.
    */
   public void useBytesOnly(){
   
      this.theSource.useBytesOnly();
   }



	/** Calls activate() for the datasource backing this cache. This method should be 
      * overridden by child classes to create the cache object.
      */	
	public void activate()
	{
        super.activate();
        if (! theSource.isActivated())            
           theSource.activate();
	}

	/** Calls deactivate() for the datasource backing this cache.
	  */	
    public void deactivate()
    {
        super.deactivate();
        if ( theSource.isActivated() )
            theSource.deactivate();
        theCache=null;
	}
			
    /** Return a reference to the DataSource that backs this CacheDataSource.*/
    public DataSource getSource()
    {
        return this.theSource;
    }
    
    /** Returns a copy of the iteration space expressed as an ISBounds.*/
    public ISBounds getIterationSpace(){
    
         return new ISBounds(this.iterationSpace);
    }
   
   /** Load the data at the given index into the provided datum. */
    public	void datum(Datum datum, IndexSpaceID did)
    {
        if (datum.size() != this.getNumAttributes())
        {
            new IllegalArgumentError( SDBError.Recoverable,
                               "Datum is not the right size for this datasource." );
            return;
        }
        theCache.lookup( datum, did );
    }

   /** Load the data at the given index into the provided datum. No argument
    *  validation is performed. Package Access only.
    */
   void pdatum(Datum datum, IndexSpaceID did)
    {

        theCache.lookup( datum, did );
    }


   /** Load the data at the given index into the provided datum. No argument
    *  validation is performed. This version provides scratch
    *  space for threaded operation, and exploits coherence where possible to reduce
    *  overhead. Package Access only. 
    */
   void pdatum(Datum datum, IndexSpaceID did, CacheContext context )
    {

        theCache.lookup( datum, did, context );
    }
    
    
    
    /** Get the Datum and fill its values in the passed Datum at positions
      * specified by the corresponding RecordSpec.
      */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec)
    {
        new InternalFaultError(SDBError.Recoverable,"Unimplemented method.");
    }

    /** Get the Datum and fill its values in the passed Datum at positions
      * specified by the corresponding FieldIDMapper.
      */
    public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
    {
        new InternalFaultError(SDBError.Recoverable,"Unimplemented method.");
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
      *  defined by dstBounds. 
      */
    public void subblock(DataBlock queryBlock, ISBounds dstBounds, ISBounds srcBounds)
    {
        theCache.lookup((BasicBlock) queryBlock, dstBounds, srcBounds);
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
      *  defined by dstBounds. The FieldIDMapper argument determines the mapping between 
      *  datums in the datasource and datums in the DataBlock. 
      */    
    public void subblock( DataBlock theBlock, 
                          ISBounds dstBounds, ISBounds srcBounds, 
                          FieldIDMapper mp)
    {
       if ( mp == null )
          subblock( theBlock, dstBounds, srcBounds );
       else
       {
          new SDBError( SDBError.Warning, "CacheDataSource: Incomplete (inefficient) " +
          		"implementation of subblock( ..., FieldIDMapper | RecordSpec)" );
          DataBlock temp = theBlock.createDataBlock( srcBounds );
          subblock( temp, dstBounds, srcBounds );
          temp.subblock( theBlock, srcBounds, mp );
       }
    }
    
}
